// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using DGNetwork;
using DGNetwork.Services;

namespace DGNetwork.AddIns.AutoUpdate
{
	public sealed class UpdateService
	{
		#region Singleton
		private static UpdateService instance = new UpdateService();
		
		public static UpdateService Instance {
			get {
				return instance;
			}
		}
		
		private UpdateService()
		{
		}
		#endregion
		
		public string GetAppPath()
		{
			Assembly assembly = Assembly.GetEntryAssembly();
			return Path.GetDirectoryName(assembly.Location);
		}
		
		UpdateForm updateForm;
		
		public bool IsUpdateAllowed()
		{
			string appPath = GetAppPath().ToLower().Replace('\\', '/');
			if (appPath.EndsWith("bin/debug") || appPath.EndsWith("bin/release")) {
				Service.Message.ShowMessage("${res:AutoUpdate.DoesNotWorkFromSource}");
				if (!CommandLine.Instance.HasFlag("test"))
					return false;
				if (!Service.Message.AskQuestion("In test mode, update can run even though you built from source, but you should be aware of the consequences.\nContinue?"))
					return false;
			}
			return true;
		}
		
		public void StartUpdate()
		{
			if (updateForm != null) {
				updateForm.BeginInvoke(new MethodInvoker(updateForm.Activate));
				return;
			}
			updateForm = new UpdateForm();
			updateForm.Show();
			updateForm.Closed += new EventHandler(UpdateFormClosed);
		}
		
		private void UpdateFormClosed(object sender, EventArgs e)
		{
			updateForm = null;
		}
		
		public void StartUpdateHelperApplication(ArrayList archives)
		{
			string thisPath = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
			string helperApp = Path.Combine(thisPath, "ApplyUpdate.exe");
			StringBuilder b = new StringBuilder();
			foreach (UpdateArchive archive in archives) {
				if (b.Length > 0)
					b.Append(' ');
				b.Append('"');
				b.Append(archive.GetTempFileName());
				b.Append("\" \"");
				b.Append(archive.ExtractTo);
				b.Append('"');
				if (string.Compare(archive.ExtractTo, thisPath, true) == 0) {
					ExtractHelper(archive.GetTempFileName(), helperApp);
				}
			}
			ProcessStartInfo info = new ProcessStartInfo(helperApp, b.ToString());
			info.WorkingDirectory = GetAppPath();
			Process.Start(info);
		}
		
		void ExtractHelper(string archive, string helperPath)
		{
			// we don't have SharpZipLib in range, but the old helper contains everything we
			// need
			byte[] assemblyData;
			using (FileStream fs = new FileStream(helperPath, FileMode.Open, FileAccess.Read)) {
				assemblyData = new byte[fs.Length];
				int pos = 0;
				int c;
				do {
					c = fs.Read(assemblyData, pos, assemblyData.Length - pos);
					pos += c;
				} while (c > 0);
			}
			Assembly assembly = Assembly.Load(assemblyData);
			Type inputStreamType = assembly.GetType("ICSharpCode.SharpZipLib.Zip.ZipInputStream");
			Type entryType = assembly.GetType("ICSharpCode.SharpZipLib.Zip.ZipEntry");
			using (FileStream fs = new FileStream(archive, FileMode.Open, FileAccess.Read)) {
				using (Stream input = (Stream)Activator.CreateInstance(inputStreamType, new object[] { fs })) {
					object entry;
					while ((entry = inputStreamType.InvokeMember("GetNextEntry", BindingFlags.InvokeMethod, null, input, null)) != null) {
						string name = (string)entryType.InvokeMember("Name", BindingFlags.GetProperty, null, entry, null);
						if (name == "ApplyUpdate.new.exe") {
							using (FileStream output = new FileStream(helperPath, FileMode.Create, FileAccess.Write)) {
								CopyStream(input, output);
							}
						}
						return;
					}
					throw new ApplicationException("ApplyUpdate.new.exe not found!");
				}
			}
		}
		
		void CopyStream(Stream input, Stream output)
		{
			byte[] buf = new byte[4096];
			int c;
			do {
				c = input.Read(buf, 0, buf.Length);
				if (c > 0) {
					output.Write(buf, 0, c);
				}
			} while (c > 0);
		}
	}
}
