// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace DGNetwork.AddIns.AutoUpdate
{
	public struct VersionNumber
	{
		int major, minor, build, revision;
		
		public VersionNumber(int major, int minor)
		{
			this.major = major;
			this.minor = minor;
			this.build = 0;
			this.revision = 0;
		}
		
		public VersionNumber(int major, int minor, int build)
		{
			this.major = major;
			this.minor = minor;
			this.build = build;
			this.revision = 0;
		}
		
		public VersionNumber(int major, int minor, int build, int revision)
		{
			this.major = major;
			this.minor = minor;
			this.build = build;
			this.revision = revision;
		}
		
		public static VersionNumber Parse(string text)
		{
			string[] numbers = text.Split('.');
			switch (numbers.Length) {
				case 2:
					return new VersionNumber(int.Parse(numbers[0]), int.Parse(numbers[1]));
				case 3:
					return new VersionNumber(int.Parse(numbers[0]), int.Parse(numbers[1]), int.Parse(numbers[2]));
				case 4:
					return new VersionNumber(int.Parse(numbers[0]), int.Parse(numbers[1]), int.Parse(numbers[2]), int.Parse(numbers[3]));
				default:
					throw new FormatException("Argument is not a valid VersionNumber");
			}
		}
		
		/// <summary>Gets/Sets the first number in the version (Major.Minor.Build.Revision)</summary>
		public int Major {
			get { return major; }
			set { major = value; }
		}
		/// <summary>Gets/Sets the second number in the version (Major.Minor.Build.Revision)</summary>
		public int Minor {
			get { return minor; }
			set { minor = value; }
		}
		/// <summary>Gets/Sets the third number in the version (Major.Minor.Build.Revision)</summary>
		public int Build {
			get { return build; }
			set { build = value; }
		}
		/// <summary>Gets/Sets the last number in the version (Major.Minor.Build.Revision)</summary>
		public int Revision {
			get { return revision; }
			set { revision = value; }
		}
		
		public override string ToString()
		{
			if (build == 0 && revision == 0)
				return String.Format("{0}.{1}", major, minor);
			if (revision == 0)
				return String.Format("{0}.{1}.{2}", major, minor, build);
			return String.Format("{0}.{1}.{2}.{3}", major, minor, build, revision);
		}
		
		public override bool Equals(object obj)
		{
			if (!(obj is VersionNumber)) return false;
			VersionNumber myVersionNumber = (VersionNumber)obj;
			return myVersionNumber == this;
		}
		
		public override int GetHashCode()
		{
			unchecked {
				return 27 * (27 * (27 * major.GetHashCode() ^ minor.GetHashCode()) ^ build.GetHashCode()) ^ revision.GetHashCode();
			}
		}
		
		#region Operators
		public static bool operator ==(VersionNumber a, VersionNumber b)
		{
			return a.major == b.major && a.minor == b.minor && a.build == b.build && a.revision == b.revision;
		}
		
		public static bool operator !=(VersionNumber a, VersionNumber b)
		{
			return a.major != b.major || a.minor != b.minor || a.build != b.build || a.revision != b.revision;
		}
		
		public static bool operator >(VersionNumber a, VersionNumber b)
		{
			if (a.major > b.major)
				return true;
			if (a.major == b.major) {
				if (a.minor > b.minor)
					return true;
				if (a.minor == b.minor) {
					if (a.build > b.build)
						return true;
					if (a.build == b.build) {
						return a.revision > b.revision;
					}
				}
			}
			return false;
		}
		
		public static bool operator <(VersionNumber a, VersionNumber b)
		{
			if (a.major < b.major)
				return true;
			if (a.major == b.major) {
				if (a.minor < b.minor)
					return true;
				if (a.minor == b.minor) {
					if (a.build < b.build)
						return true;
					if (a.build == b.build) {
						return a.revision < b.revision;
					}
				}
			}
			return false;
		}
		
		public static bool operator >=(VersionNumber a, VersionNumber b)
		{
			if (a.major > b.major)
				return true;
			if (a.major == b.major) {
				if (a.minor > b.minor)
					return true;
				if (a.minor == b.minor) {
					if (a.build > b.build)
						return true;
					if (a.build == b.build) {
						return a.revision >= b.revision;
					}
				}
			}
			return false;
		}
		
		public static bool operator <=(VersionNumber a, VersionNumber b)
		{
			if (a.major < b.major)
				return true;
			if (a.major == b.major) {
				if (a.minor < b.minor)
					return true;
				if (a.minor == b.minor) {
					if (a.build < b.build)
						return true;
					if (a.build == b.build) {
						return a.revision <= b.revision;
					}
				}
			}
			return false;
		}
		#endregion
	}
}
