// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.IO;
using System.Net;
using System.Xml;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using DGNetwork.Services;
using DGNetwork.Gui;
using ICSharpCode.Core.Services;

namespace DGNetwork.AddIns.Discovery
{
	public class RegisterDialogSink
	{
		private Control control = null;
		private PictureBox confirmationImage = null;
		private Button registerButton = null;
		private Label statusLabel = null;
		private TextBox usernameTextBox = null;
		private TextBox mailTextBox = null;
		private TextBox passwordTextBox = null;
		private TextBox password2TextBox = null;
		private TextBox confirmCodeTextBox = null;
		internal Wizard wizard;
		
		string registerUrl;
		Bitmap confirmBitmap;
		string username, mail, password, confirmCode;
		int userid;
		
		private void ControlLoad(object sender, EventArgs e)
		{
			new MethodInvoker(GetImage).BeginInvoke(null, null);
		}
		
		private void GetImage()
		{
			try {
				SetStatus("LoadInfo");
				XmlDocument doc = new XmlDocument();
				doc.Load(URLs.Register);
				XmlElement e = doc.DocumentElement;
				if (e.Name == "error") {
					Service.Message.ShowError(e.InnerText);
					if (!control.IsDisposed)
						control.Invoke(new MethodInvoker(wizard.Back));
					return;
				}
				if (e.Name != "register") {
					Service.Message.ShowError("unknown root tag " + e.Name);
					if (!control.IsDisposed)
						control.Invoke(new MethodInvoker(wizard.Back));
					return;
				}
				registerUrl = e["targetpage"].InnerText;
				string imageUrl = e["image"].InnerText;
				SetStatus("GetImage");
				using (WebClient wc = new WebClient()) {
					wc.Headers["User-Agent"] = Service.Parse("${AppIdentification}");
					using (Stream s = wc.OpenRead(imageUrl)) {
						confirmBitmap = new Bitmap(s);
					}
				}
				if (!control.IsDisposed)
					control.Invoke(new MethodInvoker(FinishedGet));
			} catch (WebException) {
				SetStatus("ErrorInit");
				if (!control.IsDisposed)
					control.Invoke(new MethodInvoker(ErrorInitInvoked));
			} catch (Exception ex) {
				SetStatus("ErrorInit");
				Service.Message.ShowError(ex);
				if (!control.IsDisposed)
					control.Invoke(new MethodInvoker(wizard.Back));
			}
		}
		
		private void ErrorInitInvoked()
		{
			Service.Message.ShowError("${res:Discovery.ErrorInit}");
			wizard.Back();
		}
		
		private void FinishedGet()
		{
			confirmationImage.Image = confirmBitmap;
			registerButton.Enabled = true;
			SetStatus("ClickRegister");
		}
		
		private delegate void void_string(string a1);
		
		private void SetStatus(string txt)
		{
			if (control.IsDisposed)
				return;
			if (control.InvokeRequired)
				control.Invoke(new void_string(SetStatus), new object[] {txt} );
			else
				statusLabel.Text = Service.Parse("${res:Discovery." + txt + "}");
		}
		
		public void RegisterButtonClick(object sender, EventArgs e)
		{
			registerButton.Enabled = false;
			username = usernameTextBox.Text.Trim();
			mail = mailTextBox.Text.Trim();
			password = passwordTextBox.Text.Trim();
			confirmCode = confirmCodeTextBox.Text.Trim();
			if (password != password2TextBox.Text.Trim()) {
				RegisterError("${res:Discovery.MistypedPassword}");
			} else if (username == "") {
				RegisterError("${res:Discovery.NeedUsername}");
			} else if (password == "") {
				RegisterError("${res:Discovery.NeedPassword}");
			} else if (!Regex.IsMatch(confirmCode, "^[0-9a-zA-Z]+$")) {
				RegisterError("${res:Discovery.InvalidConfirmCode}");
			} else {
				new MethodInvoker(Register).BeginInvoke(null, null);
			}
		}
		
		private void Register()
		{
			try {
				SetStatus("Registering");
				HttpWebRequest r = (HttpWebRequest)HttpWebRequest.Create(registerUrl);
				r.Timeout = 10000;
				r.Method = "POST";
				r.Referer = URLs.Register;
				r.UserAgent = Service.Parse("${AppIdentification}");
				r.ContentType = "application/x-www-form-urlencoded";
				StringBuilder b = new StringBuilder();
				Encoding iso8859 = Encoding.GetEncoding("ISO8859-1");
				b.Append("username=");
				b.Append(Helper.EscapePart(username, iso8859));
				b.Append("&email=");
				b.Append(Helper.EscapePart(mail, iso8859));
				b.Append("&password=");
				b.Append(Helper.EscapePart(password, iso8859));
				b.Append("&confirmcode=");
				b.Append(confirmCode);
				byte[] data = Encoding.ASCII.GetBytes(b.ToString());
				r.ContentLength = data.Length;
				using (Stream s = r.GetRequestStream()) {
					s.Write(data, 0, data.Length);
				}
				using (WebResponse res = r.GetResponse()) {
					using (Stream s = res.GetResponseStream()) {
						XmlDocument doc = new XmlDocument();
						doc.Load(s);
						XmlElement xml = doc.DocumentElement;
						if (xml.Name == "errors") {
							string txt = "";
							foreach (XmlElement sub in xml) {
								txt += sub.InnerText + "\n";
							}
							RegisterError(txt);
						} else if (xml.Name == "ok") {
							userid = XmlConvert.ToInt32(xml.GetAttribute("userid"));
							if (!control.IsDisposed)
								control.Invoke(new MethodInvoker(RegisterSuccess));
						} else {
							RegisterError("unknown root tag " + xml.Name);
						}
					}
				}
			} catch (Exception ex) {
				SetStatus("ErrorRegister");
				Service.Message.ShowError(ex);
				RegisterError();
			}
		}
		
		private void RegisterSuccess()
		{
			try {
				Properties.Username = username;
				Properties.Password = password;
				Properties.UserID = userid;
				Service.Message.ShowMessage("${res:Discovery.RegisterSuccess}");
				wizard.Container.NextWizard();
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		private void RegisterError()
		{
			if (control.InvokeRequired)
				control.Invoke(new MethodInvoker(RegisterError));
			else
				registerButton.Enabled = true;
		}
		
		private void RegisterError(string txt)
		{
			if (control.InvokeRequired) {
				control.Invoke(new void_string(RegisterError), new object[] {txt});
			} else {
				statusLabel.Text = Service.Parse(txt);
				Service.Message.ShowMessage(txt);
				RegisterError();
			}
		}
	}
}
