// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Text;
using System.Threading;
using System.Xml;
using DGNetwork.Search;
using DGNetwork.Contacts;
using DGNetwork.Services;
using Grunwald.P2PLib.DGNetwork;

namespace DGNetwork.AddIns.Discovery
{
	#region Manager
	public class SearchManager : AbstractManager
	{
		public SearchManager() : base(null, "DG Network")
		{
			
		}
		
		public override bool SupportsCriteria(ICriteria criteria)
		{
			switch (criteria.Name) {
				case "Username":
				case "EMail":
				case "ICQ":
					return true;
				default:
					return false;
			}
		}
		
		public override ISearch BeginSearch(ISearchType searchType)
		{
			UserSearch search = new UserSearch();
			search.Start(searchType);
			return search;
		}
	}
	#endregion
	
	#region Search
	public class UserSearch : DefaultSearch
	{
		private delegate void void_void();
		ISearchType searchType;
		
		public void Start(ISearchType searchType)
		{
			this.searchType = searchType;
			new void_void(Run).BeginInvoke(null, null);
		}
		
		private void AppendCriteria(StringBuilder b, string varname, string criteria)
		{
			ICriteria crit = searchType.Criterias[criteria];
			if (crit == null)
				return;
			string val = crit.Value as string;
			if (val != null && val.Length > 0) {
				b.Append(varname);
				b.Append('=');
				b.Append(Helper.EscapePart(val));
				b.Append('&');
			}
		}
		
		private void AppendNumCriteria(StringBuilder b, string varname, string criteria)
		{
			ICriteria crit = searchType.Criterias[criteria];
			if (crit == null)
				return;
			string val = crit.Value as string;
			if (val != null) {
				val = val.Replace(" ", "");
				if (val.Length > 0) {
					b.Append(varname);
					b.Append('=');
					b.Append(val);
					b.Append('&');
				}
			}
		}
		
		private void DoSearch()
		{
			StringBuilder url = new StringBuilder(URLs.Search);
			url.Append('?');
			int len = url.Length;
			AppendCriteria(url, "username", "Username");
			AppendCriteria(url, "mail", "EMail");
			AppendNumCriteria(url, "icq", "ICQ");
			if (url.Length == len) {
				return;
			}
			XmlDocument doc = new XmlDocument();
			try {
				doc.Load(url.ToString());
			} catch (XmlException ex) {
				AddResult(new ErrorResult("DG Network Discovery", ex));
				return;
			} catch (System.Net.WebException ex) {
				AddResult(new ErrorResult("DG Network Discovery", ex));
				return;
			}
			XmlElement e = doc.DocumentElement;
			if (e.Name == "error") {
				Service.Message.ShowError(e.InnerText);
				return;
			}
			if (e.Name != "results") {
				Service.Message.ShowError("unknown root tag " + e.Name);
				return;
			}
			foreach (XmlElement res in e) {
				DiscoverySearchResult result;
				try {
					result = new DiscoverySearchResult(e, res);
				} catch (Exception ex) {
					Service.Message.ShowError("error while parsing search results: " + ex.Message);
					return;
				}
				AddResult(result);
			}
		}
		
		private void Run()
		{
			try {
				DoSearch();
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			} finally {
				Finish();
			}
		}
	}
	#endregion
	
	#region SearchResult
	public class DiscoverySearchResult : DefaultResult
	{
		int discoveryID;
		
		public int DiscoveryID {
			get {
				return discoveryID;
			}
		}
		
		public DiscoverySearchResult(XmlElement root, XmlElement e)
		{
			discoveryID = XmlConvert.ToInt32(e.GetAttribute("id"));
			this.Name = e.GetAttribute("name");
			string from = e.GetAttribute("from");
			if (from != null && from.Length > 0) {
				this.Columns.Add(new ResultColumn("${res:SearchDialog.Location}", from));
			}
			if (e.HasAttribute("lastvisit")) {
				int lastVisit = XmlConvert.ToInt32(e.GetAttribute("lastvisit"));
				DateTime visitTime = DateTime.Now.AddSeconds(-lastVisit);
				this.Columns.Add(new ResultColumn("${res:SearchDialog.LastVisit}", visitTime.ToString()));
			}
			if (e.HasAttribute("icq")) {
				int icq = XmlConvert.ToInt32(e.GetAttribute("icq"));
				this.Columns.Add(new ResultColumn("ICQ#", icq.ToString()));
			}
			this.ContextMenuTreePath = "/Context/Discovery/UserResult";
			this.CanAddToList = true;
			int distance = Properties.Network.GetDistance(discoveryID);
			if (distance >= 0) {
				this.Icon = "Icons.Status.Unknown";
				Packet p = new Packet();
				p.Header.Source = Properties.UserID;
				p.Header.Destination = new int[] { discoveryID };
				p.Target = "GetStatus";
				SendResult res = Properties.Network.Send(p);
				res.CallOnAnswer(new AnswerEventHandler(StatusAnswerReceived));
			} else {
				this.Icon = "Icons.Status.Offline";
			}
		}
		
		private void StatusAnswerReceived(object sender, AnswerEventArgs e)
		{
			this.Icon = "Icons.Status." + (string)e.Answer.Values.GetValue(0, typeof(string));
			OnResultChanged(EventArgs.Empty);
		}
		
		public override void DoubleClick()
		{
			Service.Browser.OpenWindow(URLs.ViewProfile + "?userid=" + discoveryID.ToString(System.Globalization.NumberFormatInfo.InvariantInfo));
		}
		
		public override void AddToList()
		{
			foreach (IContact contact in Service.Contact.Contacts) {
				DiscoveryContact c = contact as DiscoveryContact;
				if (c != null && c.DiscoveryID == this.DiscoveryID) {
					Service.Message.ShowMessage("${res:Discovery.ContactAlreadyInList}");
					return;
				}
			}
			Service.Contact.Contacts.Add(new DiscoveryContact(this));
		}
	}
	#endregion
}
