// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Globalization;
using DGNetwork.Chat;
using DGNetwork.FileTransfer;
using DGNetwork.Services;

namespace DGNetwork.AddIns.History
{
	public class HistoryFilter : IMessageFilter, ITempFileManager
	{
		bool hasWarned = false;
		
		void WarnUser()
		{
			if (hasWarned) return;
			hasWarned = true;
			Service.Message.ShowWarningFormatted("${res:History.IllegalPath}", new string[] { Properties.Path });
		}
		
		string GetExtension(LoggingMode mode)
		{
			switch (mode)
			{
				case LoggingMode.ToHtml:
					return ".html";
				case LoggingMode.ToText:
					return ".txt";
				case LoggingMode.None:
					return null;
				default:
					throw new ArgumentException("Invalid logging mode");
			}
		}
		
		static bool initialized;
		
		static void Initialize()
		{
			if (initialized)
				return;
			initialized = true;
			Service.Timer.Tick += new EventHandler(TimerTick);
			Service.Timer.Unload += new EventHandler(TimerUnload);
		}
		
		static void TimerUnload(object sender, EventArgs e)
		{
			lock (typeof(HistoryFilter)) {
				if (writer != null) {
					writer.Close();
					writer = null;
				}
			}
		}
		
		static void TimerTick(object sender, EventArgs e)
		{
			if ((Service.Timer.Ticks % 60) == 0) {
				lock (typeof(HistoryFilter)) {
					if (writer != null)
						writer.Flush();
				}
			}
		}
		
		static string filename;
		static StreamWriter writer;
		
		public void Filter(Message message, MessageModes messageMode, object owner)
		{
			lock (typeof(HistoryFilter)) {
				LoggingMode mode = Properties.Mode;
				string extension = GetExtension(mode);
				if (extension == null)
					return;
				string path = Properties.Path;
				try {
					if (!Directory.Exists(path)) {
						WarnUser();
						return;
					}
				} catch (Exception) {
					WarnUser();
					return;
				}
				path = Path.Combine(path, message.Time.ToString(@"yyyy\\MMMM\\dd"));
				if (!Directory.Exists(path))
					Directory.CreateDirectory(path);
				string newFilename = Path.Combine(path, Service.Parse("${res:History.LogFileName}").Replace("{0}", message.Time.Hour.ToString()) + extension);
				if (filename != newFilename) {
					oldRoom = null;
					if (writer != null)
						writer.Close();
					Initialize();
					filename = newFilename;
					bool newFile = !File.Exists(filename);
					Stream stream = new FileStream(filename, FileMode.Append);
					writer = new StreamWriter(stream);
					if (newFile) {
						if (mode == LoggingMode.ToHtml) {
							string header = BrowserDisplay.GetHeader();
							string dataPath = Service.Property.DataDirectory;
							dataPath = dataPath.Replace('\\', '/');
							if (!dataPath.EndsWith("/"))
								dataPath += '/';
							header = header.Replace("href=\"", "href=\"file://" + dataPath);
							header = header.Replace("src=\"", "src=\"file://" + dataPath);
							writer.Write(header);
							writer.WriteLine(Parse("<h1>${res:History.LogFileTitle}</h1>", message.Time));
							writer.WriteLine(Parse("<p>${res:History.LogFileSubtitle}</p>", message.Time));
						} else {
							writer.WriteLine(Parse("${res:History.LogFileTitle}", message.Time));
							writer.WriteLine(Parse("${res:History.LogFileSubtitle}", message.Time));
							writer.WriteLine();
							writer.WriteLine();
						}
					}
				}
				RenderMessage(message, mode);
			}
		}
		
		string Parse(string text, DateTime time)
		{
			return Service.Parse(text, new string[,] {
			                     	{"MessageTime", time.ToShortDateString()},
			                     	{"MessageTime.Hour", time.Hour.ToString()}
			                     });
		}
		
		IChatRoom oldRoom;
		
		void RenderMessage(Message message, LoggingMode mode)
		{
			// since we use the DisplayFilter on it, we must use a copy of the message
			message = message.Clone();
			IChatRoom room = message.Room;
			if (room != null && room != oldRoom) {
				oldRoom = room;
				if (mode == LoggingMode.ToHtml) {
					writer.Write("<h2>");
					writer.Write(Converter.TextToHtml(Service.Parse(room.Title)));
					writer.WriteLine("</h2>");
				} else {
					writer.WriteLine();
					writer.Write(Service.Parse(room.Title));
					writer.WriteLine(":");
				}
			}
			if (mode == LoggingMode.ToHtml) {
				Service.Chat.DisplayFilter.Filter(message, MessageModes.Html, this);
				writer.Write("<p class=\"entry\">");
				writer.Write(message.Time.ToString("HH:mm:ss"));
				writer.Write(" ");
				writer.Write(HtmlDisplay.RenderMessageContent(message));
				writer.WriteLine("</p>");
			} else {
				Service.Chat.DisplayFilter.Filter(message, MessageModes.Text, this);
				writer.Write(message.Time.ToString("HH:mm:ss"));
				writer.Write(" ");
				writer.WriteLine(message.ToString());
			}
		}
		
		public string GetTempFileName(string filename)
		{
			string path = Path.GetDirectoryName(HistoryFilter.filename);
			path = Path.Combine(path, "files");
			if (!Directory.Exists(path))
				Directory.CreateDirectory(path);
			return Path.Combine(path, filename);
		}
		
		public string GetRelativeName(string absolutePath)
		{
			string path = Path.GetDirectoryName(HistoryFilter.filename).ToLower();
			if (absolutePath.ToLower().StartsWith(path)) {
				return absolutePath.Substring(path.Length + 1);
			} else {
				return absolutePath;
			}
		}
	}
}
