// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Threading;

namespace Grunwald.NOscar
{
	/// <summary>
	/// The AutomaticConnection is the highest level of connections available in NOscar.
	/// It manages login and reconnecting automatically.
	/// For it's receive loop, it (unlike all other connections) creates an own thread.
	/// </summary>
	public sealed class AutomaticConnection
	{
		string loginServer;
		int port;
		string username, password;
		
		public AutomaticConnection(string loginServer, int port, string username, string password)
		{
			if (loginServer == null || username == null || password == null)
				throw new ArgumentNullException();
			this.loginServer = loginServer;
			this.port = port;
			this.username = username;
			this.password = password;
		}
		
		ConnectionState state = ConnectionState.Dead;
		
		public ConnectionState State {
			get {
				return state;
			}
		}
		
		public event EventHandler StateChanged;
		
		void SetState(ConnectionState state)
		{
			if (this.state != state) {
				this.state = state;
				if (StateChanged != null)
					StateChanged(this, EventArgs.Empty);
			}
		}
		
		Thread thread;
		
		public bool Running {
			get {
				return thread != null;
			}
		}
		
		public void Start(ThreadExceptionEventHandler handler)
		{
			if (handler == null)
				throw new ArgumentNullException("handler");
			if (thread != null)
				throw new InvalidOperationException("Thread is already running");
			this.handler = handler;
			thread = new Thread(new ThreadStart(Run));
			thread.Name = "Oscar Thread";
			thread.IsBackground = true;
			stopping = false;
			thread.Start();
		}
		
		OscarConnection connection;
		ThreadExceptionEventHandler handler;
		
		void Run()
		{
			try {
				while (!stopping) {
					RunLoop();
					if (!stopping)
						Thread.Sleep(1000);
				}
			} catch (Exception ex) {
				handler(this, new ThreadExceptionEventArgs(ex));
			} finally {
				stopping = false;
				if (connection != null) {
					connection.Close();
					connection = null;
				}
			}
		}
		
		void RunLoop()
		{
			SetState(ConnectionState.Connecting);
			
			ConnectionBuilder.Result res;
			try {
				res = ConnectionBuilder.BuildConnection(loginServer, port, username, password);
			} catch (System.Net.Sockets.SocketException) {
				SetState(ConnectionState.Dead);
				return;
			}
			if (!res.Success) {
				SetState(ConnectionState.Dead);
				return;
			}
			SetState(ConnectionState.Connected);
			connection = new OscarConnection(res.Connection);
			connection.StateChanged += new EventHandler(ConnectionStateChanged);
			connection.SnacReceived += new SnacEventHandler(OnSnacReceived);
			connection.Run();
			connection = null;
			res.Connection.Close();
		}
		
		void ConnectionStateChanged(object sender, EventArgs e)
		{
			if (sender != connection) return;
			SetState(connection.State);
		}
		
		volatile bool stopping;
		
		public void Stop()
		{
			if (stopping) return;
			stopping = true;
			if (connection != null)
				connection.Close();
		}
		
		public void Send(Snac snac)
		{
			if (state != ConnectionState.Active)
				throw new InvalidOperationException("Connection is not active!");
			connection.Send(snac);
		}
		
		/// <summary>
		/// Gets the connection responsible for the specified family.
		/// </summary>
		public OscarConnection GetConnection(ushort family)
		{
			return connection;
		}
		
		public event SnacEventHandler SnacReceived;
		
		void OnSnacReceived(object sender, SnacEventArgs e)
		{
			if (SnacReceived != null)
				SnacReceived(this, e);
		}
	}
}
