// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Net;
using System.Net.Sockets;

namespace Grunwald.NOscar
{
	/// <summary>
	/// This class automatically logs into the login server and then builds a connection with the main server.
	/// </summary>
	public sealed class ConnectionBuilder
	{
		private ConnectionBuilder() {}
		
		public static Result BuildConnection(IPEndPoint endPoint, string username, string password)
		{
			return BuildConnection(new TcpClient(endPoint), username, password);
		}
		
		public static Result BuildConnection(string hostname, int port, string username, string password)
		{
			return BuildConnection(new TcpClient(hostname, port), username, password);
		}
		
		static Result BuildConnection(TcpClient client, string username, string password)
		{
			AuthResponseSnac snac = new Helper(new SnacConnection(client.GetStream()), username, password).Run();
			if (snac == null)
				return null;
			if (snac.ErrorCode == AuthResponseErrorCode.Success) {
				string host = snac.Server;
				int pos = host.LastIndexOf(':');
				TcpClient newClient;
				Console.WriteLine("Building connection to BOS...");
				if (pos > 0)
					newClient = new TcpClient(host.Substring(0, pos), int.Parse(host.Substring(pos + 1)));
				else
					newClient = new TcpClient(host, FlapConnection.DefaultPort);
				Console.WriteLine("Connection to BOS has been created!");
				FlapConnection connection = new FlapConnection(newClient.GetStream());
				connection.SendLoginChain.SetData(0x0006, snac.Cookie);
				return new Result(connection, snac);
			} else {
				return new Result(snac);
			}
		}
		
		private class Helper
		{
			SnacConnection loginConnection;
			string username, password;
			AuthResponseSnac responseSnac;
			
			public Helper(SnacConnection loginConnection, string username, string password)
			{
				this.loginConnection = loginConnection;
				this.password = password;
				this.username = username;
			}
			
			public AuthResponseSnac Run()
			{
				loginConnection.StateChanged += new EventHandler(OnStateChange);
				loginConnection.Run();
				return responseSnac;
			}
			
			void OnStateChange(object sender, EventArgs e)
			{
				if (loginConnection.State == ConnectionState.Active) {
					AuthenticationSequence auth = new AuthenticationSequence(loginConnection);
					auth.Start(username, password);
					auth.Completed += new SnacEventHandler(AuthComplete);
					auth.Error += new EventHandler(AuthError);
				}
			}
			
			void AuthComplete(object sender, SnacEventArgs e)
			{
				responseSnac = (AuthResponseSnac)e.Snac;
			}
			
			void AuthError(object sender, EventArgs e)
			{
				throw new NOscarException("Unknown error during authorization (possibly timeout)");
			}
		}
		
		public class Result
		{
			AuthResponseSnac responseSnac;
			FlapConnection connection;
			
			public FlapConnection Connection {
				get {
					return connection;
				}
			}
			
			public AuthResponseSnac ResponseSnac {
				get {
					return responseSnac;
				}
			}
			
			public bool Success {
				get {
					return connection != null;
				}
			}
			
			public AuthResponseErrorCode ErrorCode {
				get {
					return responseSnac.ErrorCode;
				}
			}
			
			public Result(AuthResponseSnac responseSnac)
			{
				if (responseSnac == null)
					throw new ArgumentNullException("responseSnac");
				this.responseSnac = responseSnac;
			}
			
			public Result(FlapConnection connection, AuthResponseSnac responseSnac) : this(responseSnac)
			{
				this.connection = connection;
			}
		}
	}
}
