// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	/// <summary>
	/// A Flap is the basic data packet sent over the Oscar protocol.
	/// It consists of a channel number, a sequence number and the data.
	/// </summary>
	/// <seealso cref="FlapConnection"/>
	/// <seealso cref="FlapReader"/>
	/// <seealso cref="FlapWriter"/>
	public sealed class Flap
	{
		FlapChannel channel;
		ushort sequenceNumber;
		byte[] data;
		
		/// <summary>
		/// The channel of the FLAP.
		/// See <see cref="FlapChannel">FlapChannel documentation</see>.
		/// </summary>
		/// <seealso cref="FlapChannel"/>
		public FlapChannel Channel {
			get {
				return channel;
			}
			set {
				channel = value;
			}
		}
		
		/// <summary>The data body of the FLAP.</summary>
		public byte[] Data {
			get {
				return data;
			}
			set {
				if (value == null)
					throw new ArgumentNullException("data");
				data = value;
			}
		}
		
		/// <summary>
		/// The sequence number of the FLAP.
		/// The <see cref="FlapWriter"/> and <see cref="FlapConnection"/> automatically assign
		/// sequence numbers.
		/// </summary>
		public ushort SequenceNumber {
			get {
				return sequenceNumber;
			}
			set {
				sequenceNumber = value;
			}
		}
		
		/// <summary>Creates a new Flap object with the specified values.</summary>
		public Flap(FlapChannel channel, byte[] data)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			this.channel = channel;
			this.data = data;
			this.sequenceNumber = 0;
		}
		
		/// <summary>Creates a new Flap object with the specified values.</summary>
		public Flap(FlapChannel channel, ushort sequenceNumber, byte[] data)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			this.channel = channel;
			this.data = data;
			this.sequenceNumber = sequenceNumber;
		}
	}
	
	/// <summary>
	/// Event handler used when receiving a <see cref="Flap"/>.
	/// </summary>
	/// <seealso cref="Flap"/>
	/// <seealso cref="FlapConnection"/>
	public delegate void FlapEventHandler(object sender, FlapEventArgs e);
	
	/// <summary>
	/// Event arguments used when receiving a <see cref="Flap"/>.
	/// </summary>
	/// <seealso cref="Flap"/>
	/// <seealso cref="FlapConnection"/>
	public class FlapEventArgs : EventArgs
	{
		Flap flap;
		
		/// <summary>
		/// The FLAP that has been received.
		/// </summary>
		public Flap Flap {
			get {
				return flap;
			}
		}
		
		/// <summary>Creates a new FlapEventArgs object.</summary>
		public FlapEventArgs(Flap flap)
		{
			if (flap == null)
				throw new ArgumentNullException("flap");
			this.flap = flap;
		}
	}
}
