// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	/// <summary>
	/// The FlapConnection runs a receive loop on a <see cref="FlapReader"/> and wraps the methods
	/// of the <see cref="FlapWriter"/>.
	/// Moreover, it automatically sends the negotiation FLAP.
	/// </summary>
	/// <seealso cref="Flap"/>
	/// <seealso cref="FlapReader"/>
	/// <seealso cref="FlapWriter"/>
	public sealed class FlapConnection : IDisposable
	{
		public const int DefaultPort = 1590;
		
		FlapReader reader;
		FlapWriter writer;
		
		ConnectionState state = ConnectionState.Connected;
		
		public ConnectionState State {
			get {
				return state;
			}
		}
		
		public event EventHandler StateChanged;
		
		void SetState(ConnectionState state)
		{
			if (this.state != state) {
				this.state = state;
				if (StateChanged != null) {
					StateChanged(this, EventArgs.Empty);
				}
			}
		}
		
		public void Close()
		{
			SetState(ConnectionState.Dead);
			reader.Dispose();
			writer.Dispose();
		}
		
		void IDisposable.Dispose()
		{
			this.Close();
		}
		
		public FlapConnection(Stream stream)
		{
			reader = new FlapReader(stream);
			writer = new FlapWriter(stream);
		}
		
		/// <summary>
		/// Runs the receiving loop of the connection.
		/// </summary>
		/// <exception cref="IOException">Used for all cases of IO-failure.
		/// (e.g. connection closed when data was expected)</exception>
		/// <exception cref="InvalidFlapException">Used when an invalid flap was received.</exception>
		public void Run()
		{
			if (state != ConnectionState.Connected)
				throw new InvalidOperationException();
			try {
				SetState(ConnectionState.Negotiating);
				IAsyncResult result = WriteLoginFlap();
				Flap flap = reader.Read();
				if (flap == null)
					throw new IOException("Connection was closed, but welcome FLAP was expected.");
				if (flap.Channel != FlapChannel.Login)
					throw new InvalidFlapException("Expected: FlapChannel.Login, but got: FlapChannel." + flap.Channel.ToString());
				if (flap.Data.Length > 4)
					throw new InvalidFlapException("Expected: Login flap has at least 4 bytes of data.");
				if (NetworkBitConverter.ToUInt32(flap.Data, 0) != 1)
					throw new InvalidFlapException("Expected: Flapchannel has version number 1.");
				receivedLoginChain = new TlvChain(flap.Data, 4, flap.Data.Length - 4);
				writer.EndWrite(result);
				SetState(ConnectionState.Active);
				while ((flap = reader.Read()) != null) {
					if (ReceiveFlap != null) {
						ReceiveFlap(this, new FlapEventArgs(flap));
					}
					if (state == ConnectionState.Dead)
						break;
				}
			} finally {
				SetState(ConnectionState.Dead);
			}
		}
		
		TlvChain sendLoginChain = new TlvChain();
		TlvChain receivedLoginChain;
		
		public TlvChain ReceivedLoginChain {
			get {
				return receivedLoginChain;
			}
		}
		
		public TlvChain SendLoginChain {
			get {
				return sendLoginChain;
			}
		}
		
		IAsyncResult WriteLoginFlap()
		{
			byte[] data = new byte[4 + sendLoginChain.GetByteLength()];
			using (MemoryStream ms = new MemoryStream(data, true)) {
				// send flap version 1
				ms.WriteByte(0);
				ms.WriteByte(0);
				ms.WriteByte(0);
				ms.WriteByte(1);
				// send extra login data
				sendLoginChain.WriteBytes(ms);
			}
			return writer.BeginWrite(FlapChannel.Login, data, null, null);
		}
		
		public event FlapEventHandler ReceiveFlap;
		
		/// <summary>
		/// Sends the specified Flap.
		/// The sequence number is ignored but the correct sequence number is used instead.
		/// </summary>
		public void SendFlap(Flap flap)
		{
			if (state != ConnectionState.Active)
				throw new InvalidOperationException();
			writer.Write(flap);
		}
		
		/// <summary>
		/// Sends the specified Flap.
		/// </summary>
		public void SendFlap(FlapChannel channel, byte[] data)
		{
			if (state != ConnectionState.Active)
				throw new InvalidOperationException();
			writer.Write(channel, data);
		}
		
		public IAsyncResult BeginSendFlap(FlapChannel channel, byte[] data, AsyncCallback callback, object state)
		{
			if (this.state != ConnectionState.Active)
				throw new InvalidOperationException();
			return writer.BeginWrite(channel, data, callback, state);
		}
		
		public void EndSendFlap(IAsyncResult result)
		{
			writer.EndWrite(result);
		}
	}
}
