// Copyright (C) 2004  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	/// <summary>
	/// The FlapReader is able to read a sequence of FLAPs from a stream.
	/// </summary>
	/// <seealso cref="Flap"/>
	/// <seealso cref="FlapConnection"/>
	public sealed class FlapReader : IDisposable
	{
		Stream baseStream;
		bool ownsStream;
		
		public FlapReader(Stream stream) : this(stream, true) { }
		
		public FlapReader(Stream stream, bool ownsStream)
		{
			if (stream == null)
				throw new ArgumentNullException("stream");
			this.baseStream = stream;
			this.ownsStream = ownsStream;
		}
		
		/// <summary>
		/// Closes the FlapReader.
		/// If the FlapReader owns the <see cref="BaseStream"/>, the base stream is closed.
		/// </summary>
		public void Dispose()
		{
			if (ownsStream)
				((IDisposable)baseStream).Dispose();
		}
		
		/// <summary>
		/// Gets the underlying stream.
		/// </summary>
		public Stream BaseStream {
			get {
				return baseStream;
			}
		}
		
		/// <summary>
		/// Reads the next flap from the channel.
		/// Returns null when the connection was closed.
		/// </summary>
		/// <exception cref="IOException">Used for all cases of IO-failure.
		/// (e.g. connection closed when data was expected)</exception>
		/// <exception cref="InvalidFlapException">Used when an invalid flap was received.</exception>
		public Flap Read()
		{
			byte[] header = ReadBytes(6, true);
			if (header == null)
				return null;
			if (header[0] != 0x2a)
				throw new InvalidFlapException("Identification byte was invalid.");
			FlapChannel channel = (FlapChannel)header[1];
			ushort seqnum = NetworkBitConverter.ToUInt16(header, 2);
			ushort length = NetworkBitConverter.ToUInt16(header, 4);
			byte[] data = ReadBytes(length, false);
			return new Flap(channel, seqnum, data);
		}
		
		byte[] ReadBytes(int count, bool allowNull)
		{
			byte[] data = new byte[count];
			int pos = 0;
			int c;
			while (pos < count) {
				try {
					c = baseStream.Read(data, pos, count - pos);
				} catch (ObjectDisposedException) {
					c = 0;
				} catch (IOException) {
					c = 0;
				}
				if (c == 0) {
					if (allowNull && pos == 0)
						return null;
					else
						throw new IOException("Connection closed, but only " + pos + " of " + count + " bytes have been read.");
				}
				pos += c;
			}
			return data;
		}
	}
}
