// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	/// <summary>
	/// The FlapWriter is able to write FLAPs to a stream.
	/// </summary>
	/// <remarks>
	/// This class supports multi-threading by locking the base stream whenever 
	/// </remarks>
	/// <seealso cref="Flap"/>
	/// <seealso cref="FlapConnection"/>
	public sealed class FlapWriter : IDisposable
	{
		Stream baseStream;
		bool ownsStream;
		ushort sequenceNumber;
		
		public FlapWriter(Stream stream) : this(stream, true) { }
		
		public FlapWriter(Stream stream, bool ownsStream)
		{
			if (stream == null)
				throw new ArgumentNullException("stream");
			this.baseStream = stream;
			this.ownsStream = ownsStream;
			writeDelegate = new WriteDelegate(Write);
		}
		
		/// <summary>
		/// Closes the FlapWriter.
		/// If the FlapWriter owns the <see cref="BaseStream"/>, the base stream is closed.
		/// </summary>
		public void Dispose()
		{
			if (ownsStream)
				((IDisposable)baseStream).Dispose();
		}
		
		/// <summary>
		/// Gets the underlying stream.
		/// </summary>
		public Stream BaseStream {
			get {
				return baseStream;
			}
		}
		
		/// <summary>
		/// Writes the specified Flap to the stream. The sequence number is ignored but the next
		/// sequence number of this FlapWriter is used instead.
		/// </summary>
		public void Write(Flap flap)
		{
			if (flap == null)
				throw new ArgumentNullException("flap");
			Write(flap.Channel, flap.Data);
		}
		
		/// <summary>Writes the specified data into the FlapChannel.</summary>
		public void Write(FlapChannel channel, byte[] data)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			if (data.Length > ushort.MaxValue)
				throw new ArgumentException("data.Length must be <= " + ushort.MaxValue);
			lock (baseStream) {
				byte[] header = new byte[6 + data.Length];
				header[0] = 0x2a;
				header[1] = (byte)channel;
				unchecked {
					sequenceNumber += 1;
				}
				NetworkBitConverter.PutBytes(header, 2, sequenceNumber);
				NetworkBitConverter.PutBytes(header, 4, (ushort)data.Length);
				data.CopyTo(header, 6);
				baseStream.Write(header, 0, header.Length);
			}
		}
		
		/// <summary>Waits for all pending Flaps to be written and calls Flush() on the base stream.</summary>
		public void Flush()
		{
			lock (baseStream) {
				baseStream.Flush();
			}
		}
		
		delegate void WriteDelegate(FlapChannel channel, byte[] data);
		
		WriteDelegate writeDelegate;
		
		public IAsyncResult BeginWrite(FlapChannel channel, byte[] data, AsyncCallback callback, object state)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			if (data.Length > ushort.MaxValue)
				throw new ArgumentException("data.Length must be <= " + ushort.MaxValue);
			return writeDelegate.BeginInvoke(channel, data, callback, state);
		}
		
		public void EndWrite(IAsyncResult result)
		{
			writeDelegate.EndInvoke(result);
		}
	}
}
