// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	#region Enums
	/// <summary>
	/// Specifies the type of the meta info request.
	/// </summary>
	public enum IcqMetaInfoType : ushort
	{
		/// <summary>
		/// Request offline messages.
		/// </summary>
		OfflineMessageRequest  = 0x003c,
		/// <summary>
		/// Delete offline messages.
		/// </summary>
		DeleteOfflineMessages  = 0x003e,
		
		OfflineMessageResponse = 0x0042,
		EndOfOfflineMessages = 0x0042,
		/// <summary>
		/// Request some kind of meta information.
		/// </summary>
		MetaInformationRequest = 0x07d0,
		MetaInformationResponse = 0x07da
	}
	
	public enum IcqMetaRequestSubType : ushort
	{
		SetUserBasicInfo = 0x03ea,
		SetUserWorkInfo  = 0x03f3,
		SetUserMoreInfo  = 0x03fd,
		SetUserNotesInfo = 0x0406,
		SetUserExtendedEMailInfo = 0x040b,
		SetUserInterestsInfo     = 0x0410,
		SetUserAffilationsInfo   = 0x041a,
		SetUserPermissionsInfo   = 0x0424,
		ChangeUserPassword       = 0x042e,
		SetUserHomepageCategory  = 0x0442,
		RequestFullUserInfo      = 0x04b2,
		RequestShortUserInfo     = 0x04ba,
		UnregisterUser           = 0x04c4,
		RequestFullUserInfo2     = 0x04f0,
		SearchByDetailsPlain  = 0x0515,
		SearchByUinPlain      = 0x051f,
		SearchByEMailPlain    = 0x0529,
		SearchWhitepagesPlain = 0x0533,
		SearchByDetailsPlainWildcard  = 0x053d,
		SearchByEMailPlainWildcard    = 0x0547,
		SearchWhitepagesPlainWildcard = 0x0551,
		SearchByUinTlv      = 0x0569,
		SearchByEMailTlv    = 0x0573,
		SearchWhitepagesTlv = 0x055f,
		SearchRandomChatUser = 0x074e,
		RequestXmlServerVariable = 0x0898,
		SendRegistrationStatsReport = 0x0aa5,
		SendShortcutBarStatsReport  = 0x0aaf,
		SaveInfoTlv    = 0x0c3a,
		SendSms        = 0x1482,
		SendSpamReport = 0x2008
	}
	
	public enum IcqMetaResponseSubType : ushort
	{
		ProcessingError = 0x0001,
		SetUserBasicInfoAck = 0x0064,
		SetUserWorkInfoAck  = 0x006e,
		SetUserMoreInfoAck  = 0x0078,
		SetUserNotesInfoAck = 0x0082,
		SetUserExtendedEMailInfoAck = 0x0087,
		SetUserInterestsInfoAck     = 0x008c,
		
		/// <summary>Acknowledge user affilations info. WARNING: Has the same value as SendSmsAck.</summary>
		SetUserAffilationsInfoAck   = 0x0096,
		/// <summary>Acknowledge sms receipt. WARNING: Has the same value as SetUserAffilationsInfo.</summary>
		SendSmsAck                  = 0x0096,
		
		SetUserPermissionsInfoAck   = 0x00a0,
		ChangeUserPasswordAck       = 0x00aa,
		UnregisterUserAck           = 0x00b4,
		SetUserHomepageCategoryAck  = 0x00be,
		
		UserBasicInfoReply = 0x00c8,
		UserWorkInfoReply  = 0x006e,
		UserMoreInfoReply  = 0x0078,
		UserNotesInfoReply = 0x0082,
		UserExtendedEMailInfoReply = 0x0087,
		UserInterestsInfoReply     = 0x008c,
		UserAffilationsInfoReply   = 0x0096,
		ShortUserInfoReply         = 0x0104,
		UserHomepageCategoryReply  = 0x010e,
		
		SearchUserFoundReply = 0x1a4,
		SearchLastUserFoundReply = 0x1ae,
		
		RegistrationStatsReportAck  = 0x0302,
		RandomChatUserReply    = 0x0366,
		XmlServerVariableReply = 0x08a2,
		SaveInfoTlvAck    = 0x0c3f,
		SendSpamReportAck = 0x2012
	}
	#endregion
	
	#region IcqMetaInfoSnac
	/// <summary>
	/// This snac is used only with ICQ, not with AIM. It is used for ICQ-specific features
	/// like offline messages and user search.
	/// </summary>
	/// <remarks>This snac cannot be autogenerated because it has a special format.</remarks>
	public class IcqMetaInfoRequestSnac : IcqMetaInfoSnac
	{
		/// <summary>The SNAC subtype normally used for the IcqMetaInfoRequestSnac.</summary>
		public static ushort SnacType = 0x0002;
		
		/// <summary>Creates a new IcqMetaInfoRequestSnac.</summary>
		public IcqMetaInfoRequestSnac(int ownerUin, IcqMetaInfoType infoType, byte[] infoData)
			: base(ownerUin, infoType, infoData, SnacType)
		{
		}
		
		/// <summary>Creates a new IcqMetaInfoRequestSnac.</summary>
		public IcqMetaInfoRequestSnac(int ownerUin, IcqMetaRequestSubType infoSubType, byte[] infoData)
			: this(ownerUin, IcqMetaInfoType.MetaInformationRequest, infoData)
		{
			this.InfoSubType = infoSubType;
		}
		
		/// <summary>Loads the IcqMetaInfoRequestSnac from the byte array.</summary>
		public IcqMetaInfoRequestSnac(byte[] data, int index, int count) : base(data, index, count) { }
		
		/// <summary>
		/// Gets/Sets the information sub type.
		/// This property is only used when <see cref="InfoType"/> is
		/// <see cref="IcqMetaInfoType">IcqMetaInfoType.MetaInformationRequest</see>.
		/// </summary>
		public new IcqMetaRequestSubType InfoSubType {
			get {
				return (IcqMetaRequestSubType)base.InfoSubType;
			}
			set {
				base.InfoSubType = (ushort)value;
			}
		}
	}
	
	/// <summary>
	/// This snac is used only with ICQ, not with AIM. It is used for ICQ-specific features
	/// like offline messages and user search.
	/// </summary>
	/// <remarks>This snac cannot be autogenerated because it has a special format.</remarks>
	public class IcqMetaInfoResponseSnac : IcqMetaInfoSnac
	{
		/// <summary>The SNAC subtype normally used for the IcqMetaInfoRequestSnac.</summary>
		public static ushort SnacType = 0x0003;
		
		/// <summary>Creates a new IcqMetaInfoResponseSnac.</summary>
		public IcqMetaInfoResponseSnac(int ownerUin, IcqMetaInfoType infoType, byte[] infoData)
			: base(ownerUin, infoType, infoData, SnacType)
		{
		}
		
		/// <summary>Creates a new IcqMetaInfoResponseSnac.</summary>
		public IcqMetaInfoResponseSnac(int ownerUin, IcqMetaResponseSubType infoSubType, byte[] infoData)
			: this(ownerUin, IcqMetaInfoType.MetaInformationResponse, infoData)
		{
			this.InfoSubType = infoSubType;
		}
		
		/// <summary>Loads the IcqMetaInfoResponseSnac from the byte array.</summary>
		public IcqMetaInfoResponseSnac(byte[] data, int index, int count) : base(data, index, count) { }
		
		/// <summary>
		/// Gets/Sets the information sub type.
		/// This property is only used when <see cref="InfoType"/> is
		/// <see cref="IcqMetaInfoType">IcqMetaInfoType.MetaInformationResponse</see>.
		/// </summary>
		public new IcqMetaResponseSubType InfoSubType {
			get {
				return (IcqMetaResponseSubType)base.InfoSubType;
			}
			set {
				base.InfoSubType = (ushort)value;
			}
		}
	}
	
	/// <summary>
	/// This snac is used only with ICQ, not with AIM. It is used for many ICQ-specific features
	/// like offline messages and user search.
	/// </summary>
	/// <remarks>This snac cannot be autogenerated because it has a special format.</remarks>
	public class IcqMetaInfoSnac : TlvBasedSnac
	{
		/// <summary>The family id normally used for IcqMetaInfoSnacs.</summary>
		public static ushort SnacFamily = 0x0015;
		
		static ushort activeSequenceNumber;
		
		/// <summary>Creates a new IcqMetaInfoSnac.</summary>
		public IcqMetaInfoSnac(int ownerUin, IcqMetaInfoType infoType, byte[] infoData, ushort snacType)
			: base(SnacFamily, snacType)
		{
			lock (typeof(IcqMetaInfoSnac)) {
				if (activeSequenceNumber == ushort.MaxValue)
					activeSequenceNumber = 0;
				else
					activeSequenceNumber += 1;
				this.InfoSequenceNumber = activeSequenceNumber;
			}
			this.OwnerUin = ownerUin;
			this.InfoType = infoType;
			this.InfoData = infoData;
		}
		
		/// <summary>Loads the IcqMetaInfoSnac from the byte array.</summary>
		public IcqMetaInfoSnac(byte[] data, int index, int count) : base(data, index, count)
		{
			if (!BitConverter.IsLittleEndian)
				throw new NotImplementedException("IcqMetaInfoSnac is only implemented for little endian machines!");
			byte[] b = Chain[0x0001];
			ownerUin = LEBitConverter.ToInt32(b, 2);
			infoType = (IcqMetaInfoType)LEBitConverter.ToUInt16(b, 6);
			infoSequenceNumber = LEBitConverter.ToUInt16(b, 8);
			if (infoType == IcqMetaInfoType.MetaInformationRequest || infoType == IcqMetaInfoType.MetaInformationResponse) {
				infoSubType = LEBitConverter.ToUInt16(b, 10);
				infoData = SliceArray(b, 12, b.Length - 12);
			} else {
				infoData = SliceArray(b, 10, b.Length - 10);
			}
		}
		
		byte[] SliceArray(byte[] data, int start, int length)
		{
			byte[] b = new byte[length];
			Array.Copy(data, start, b, 0, length);
			return b;
		}
		
		public override int GetByteLength()
		{
			Chain.Remove(1);
			return base.GetByteLength() + GetDataLength();
		}
		
		int GetDataLength()
		{
			if (infoType == IcqMetaInfoType.MetaInformationRequest)
				return 12 + infoData.Length;
			else
				return 10 + infoData.Length;
		}
		
		public override void WriteBytes(Stream stream)
		{
			byte[] data = new byte[GetDataLength()];
			LEBitConverter.PutBytes(data, 0, (ushort)(data.Length - 2));
			LEBitConverter.PutBytes(data, 2, ownerUin);
			LEBitConverter.PutBytes(data, 6, (ushort)infoType);
			LEBitConverter.PutBytes(data, 8, infoSequenceNumber);
			if (infoType == IcqMetaInfoType.MetaInformationRequest) {
				LEBitConverter.PutBytes(data, 10, infoSubType);
				infoData.CopyTo(data, 12);
			} else {
				infoData.CopyTo(data, 10);
			}
			Chain.SetData(1, data);
			base.WriteBytes(stream);
		}
		
		int ownerUin;
		
		/// <summary>The user id of the owner of this meta information snac.</summary>
		public int OwnerUin {
			get {
				return ownerUin;
			}
			set {
				ownerUin = value;
			}
		}
		
		IcqMetaInfoType infoType;
		
		/// <summary>The type of this meta information snac.</summary>
		public IcqMetaInfoType InfoType {
			get {
				return infoType;
			}
			set {
				infoType = value;
			}
		}
		
		ushort infoSequenceNumber;
		
		/// <summary>The meta information sequence number.</summary>
		public ushort InfoSequenceNumber {
			get {
				return infoSequenceNumber;
			}
			set {
				infoSequenceNumber = value;
			}
		}
		
		ushort infoSubType;
		
		/// <summary>
		/// Gets/Sets the information sub type.
		/// This property is only used when <see cref="InfoType"/> is
		/// <see cref="IcqMetaInfoType">IcqMetaInfoType.MetaInformationRequest</see>
		/// or <see cref="IcqMetaInfoType">IcqMetaInfoType.MetaInformationResponse</see>.
		/// </summary>
		public ushort InfoSubType {
			get {
				return infoSubType;
			}
			set {
				infoSubType = value;
			}
		}
		
		byte[] infoData;
		
		/// <summary>
		/// Gets/Sets the request data.
		/// </summary>
		public byte[] InfoData {
			get {
				return infoData;
			}
			set {
				if (value == null)
					throw new ArgumentNullException("infoData");
				infoData = value;
			}
		}
		
		public override string ToString()
		{
			if (infoType == IcqMetaInfoType.MetaInformationRequest || infoType == IcqMetaInfoType.MetaInformationResponse) {
				return String.Format("[{2} RequestID = {0:x} InfoSubType = {1:x}]",
				                     this.RequestID, infoSubType, GetType().Name);
			} else {
				return String.Format("[{2} RequestID = {0:x} InfoType = {1}]",
				                     this.RequestID, infoType, GetType().Name);
			}
		}
	}
	#endregion
}
