// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Security.Cryptography;

namespace Grunwald.NOscar
{
	/// <summary>
	/// This sequence runs the following sequences in parallel:
	/// <see cref="IcbmRequestParameterInfoSequence"/>, <see cref="RateLimitInitializationSequence"/>,
	/// <see cref="LocationRightsRequestSequence"/>, <see cref="RequestSsiRightsSequence"/>
	/// and <see cref="RequestSsiDataSequence"/>.
	/// </summary>
	public sealed class AutoInitializationSequence
	{
		IcbmRequestParameterInfoSequence s1;
		RateLimitInitializationSequence  s2;
		LocationRightsRequestSequence    s3;
		RequestSsiRightsSequence         s4;
		RequestSsiDataSequence           s5;
			
		int okCount;
		const int SequenceCount = 5;
		
		public AutoInitializationSequence(SnacConnection connection)
		{
			s1 = new IcbmRequestParameterInfoSequence(connection);
			s1.Completed += new SnacEventHandler(SequenceCompleted);
			s1.Error += new EventHandler(SequenceError);
			s2 = new RateLimitInitializationSequence(connection);
			s2.Completed += new SnacEventHandler(SequenceCompleted);
			s2.Error += new EventHandler(SequenceError);
			s3 = new LocationRightsRequestSequence(connection);
			s3.Completed += new SnacEventHandler(SequenceCompleted);
			s3.Error += new EventHandler(SequenceError);
			s4 = new RequestSsiRightsSequence(connection);
			s4.Completed += new SnacEventHandler(SequenceCompleted);
			s4.Error += new EventHandler(SequenceError);
			s5 = new RequestSsiDataSequence(connection);
			s5.Completed += new SnacEventHandler(SequenceCompleted);
			s5.Error += new EventHandler(SequenceError);
		}
		
		public void Start()
		{
			isError = false;
			okCount = 0;
			s1.Start();
			s2.Start();
			s3.Start();
			s4.Start();
			s5.Start();
		}
		
		bool isError;
		
		public event EventHandler Error;
		public event EventHandler Completed;
		
		void SequenceError(object sender, EventArgs e)
		{
			if (isError) return;
			isError = true;
			if (Error != null)
				Error(this, e);
		}
		
		void SequenceCompleted(object sender, SnacEventArgs e)
		{
			okCount += 1;
			if (okCount == SequenceCount) {
				if (Completed != null)
					Completed(this, e);
			}
		}
	}
	
	/// <summary>
	/// This sequence is used to request the <see cref="IcbmParameterInfo">IcbmParameterInfo structure</see>.
	/// </summary>
	public sealed class IcbmRequestParameterInfoSequence : SnacSequence
	{
		public IcbmRequestParameterInfoSequence(SnacConnection connection) : base(connection) {}
		
		public void Start()
		{
			SendSnac(new IcbmRequestParameterInfoSnac(), new SnacEventHandler(Step1));
		}
		
		void Step1(object sender, SnacEventArgs e)
		{
			IcbmParameterInfoSnac s = e.Snac as IcbmParameterInfoSnac;
			if (s != null) {
				Connection.Data.IcbmParameterInfo = s.ParameterInfo;
				OnCompleted(e);
			} else {
				OnError(e);
			}
		}
	}
	
	/// <summary>
	/// This sequence is used to get the initial rate limit values.
	/// </summary>
	public sealed class RateLimitInitializationSequence : SnacSequence
	{
		public RateLimitInitializationSequence(SnacConnection connection) : base(connection) {}
		
		public void Start()
		{
			SendSnac(new RateInfoRequestSnac(), new SnacEventHandler(Step1));
		}
		
		void Step1(object sender, SnacEventArgs e)
		{
			RateInfoSnac response = e.Snac as RateInfoSnac;
			if (response == null) {
				OnError(e);
			} else {
				ushort[] classes = new ushort[response.InfoBlocks.Length];
				for (int i = 0; i < classes.Length; ++i) {
					classes[i] = response.InfoBlocks[i].RateClass;
				}
				Connection.Data.RateCommandsBlocks = response.CommandsBlocks;
				Connection.Data.RateInfoBlocks = response.InfoBlocks;
				Connection.Send(new RateAckSnac(classes));
				OnCompleted(e);
			}
		}
	}
	
	/// <summary>
	/// This sequence is used to request the location rights.
	/// </summary>
	public sealed class LocationRightsRequestSequence : SnacSequence
	{
		public LocationRightsRequestSequence(SnacConnection connection) : base(connection) {}
		
		public void Start()
		{
			SendSnac(new LocationRightsRequestSnac(), new SnacEventHandler(Step1));
		}
		
		void Step1(object sender, SnacEventArgs e)
		{
			LocationRightsSnac s = e.Snac as LocationRightsSnac;
			if (s != null) {
				Connection.Data.MaxLocationInfoLength = s.MaxInfoLength;
				OnCompleted(e);
			} else {
				OnError(e);
			}
		}
	}
	
	/// <summary>
	/// This sequence is used to request the ssi rights.
	/// </summary>
	public sealed class RequestSsiRightsSequence : SnacSequence
	{
		public RequestSsiRightsSequence(SnacConnection connection) : base(connection) {}
		
		public void Start()
		{
			SendSnac(new SsiRightsRequestSnac(), new SnacEventHandler(Step1));
		}
		
		void Step1(object sender, SnacEventArgs e)
		{
			SsiRightsSnac s = e.Snac as SsiRightsSnac;
			if (s != null) {
				// TODO: Store data in ConnectionData.
				OnCompleted(e);
			} else {
				OnError(e);
			}
		}
	}
	
	/// <summary>
	/// This sequence is used to request the ssi data.
	/// </summary>
	public sealed class RequestSsiDataSequence : SnacSequence
	{
		public RequestSsiDataSequence(SnacConnection connection) : base(connection) {}
		
		public void Start()
		{
			SendSnac(new SsiDataRequestSnac(), new SnacEventHandler(Step1));
		}
		
		void Step1(object sender, SnacEventArgs e)
		{
			SsiDataSnac s = e.Snac as SsiDataSnac;
			if (s != null) {
				Connection.Data.SsiItems = s.Items;
				OnCompleted(e);
			} else {
				OnError(e);
			}
		}
	}
}
