// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Text;

namespace Grunwald.NOscar
{
	public sealed class MessageBlock : IByteConvertible
	{
		ArrayList fragments = new ArrayList();
		
		public MessageBlock(string text)
		{
			fragments.Add(new MessageFragment(5, new byte[] { 1 }));
			AddText(text);
		}
		
		public MessageBlock(byte[] data) : this(data, 0, data.Length) {}
		
		public MessageBlock(byte[] data, int offset, int length)
		{
			int pos = offset;
			int end = offset + length;
			while (pos < end) {
				MessageFragment f = new MessageFragment(data, pos, end - pos);
				pos += f.GetByteLength();
				fragments.Add(f);
			}
		}
		
		public void SetText(string text)
		{
			for (int i = 0; i < fragments.Count; ++i) {
				MessageFragment f = (MessageFragment)fragments[i];
				if (f.Identifier == 1 && f.Version == 1)
					fragments.Remove(i--);
			}
			AddText(text);
		}
		
		void AddText(string text)
		{
			byte[] b = Tlv.Encoding.GetBytes(text);
			byte[] data = new byte[4 + b.Length];
			b.CopyTo(data, 4);
			fragments.Add(new MessageFragment(1, data));
		}
		
		public string GetText()
		{
			StringBuilder b = new StringBuilder();
			foreach (MessageFragment f in fragments) {
				if (f.Identifier == 1 && f.Version == 1) {
					// text block
					byte[] data = f.Data;
					ushort charset = NetworkBitConverter.ToUInt16(data, 0);
					// byte 2+3 = charsubset, significance is unknown
					Encoding encoding;
					switch (charset) {
						case 0:
							encoding = Encoding.ASCII;
							break;
						case 2:
							encoding = Encoding.BigEndianUnicode;
							break;
						default:
							// 3 is code for ISO-8859-1, which is always stored in Tlv.Encoding.
							// use it also for all unknown encodings.
							encoding = Tlv.Encoding;
							break;
					}
					b.Append(encoding.GetString(data, 4, data.Length - 4));
				}
			}
			return b.ToString();
		}
		
		/// <summary>
		/// Gets the total length this object would need in the byte array.
		/// </summary>
		public int GetByteLength()
		{
			int length = 0;
			foreach (MessageFragment c in fragments) {
				length += c.GetByteLength();
			}
			return length;
		}
		
		/// <summary>
		/// Writes the byte representation of this object into the specified stream.
		/// </summary>
		public void WriteBytes(Stream stream)
		{
			foreach (MessageFragment c in fragments) {
				c.WriteBytes(stream);
			}
		}
		
		/// <summary>
		/// Gets the MessageBlock as byte array.
		/// </summary>
		public byte[] ToByteArray()
		{
			byte[] b = new byte[GetByteLength()];
			using (MemoryStream ms = new MemoryStream(b, true)) {
				WriteBytes(ms);
			}
			return b;
		}
	}
}
