// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Text;

namespace Grunwald.NOscar
{
	/// <summary>
	/// Some small portions of the OSCAR protocol use data structures from the old Mirabilis ICQ protocol.
	/// They use little endian instead of network byte order and a different format to store strings.
	/// This TlvChain stores data in a format compatible to these data structures.
	/// </summary>
	[Serializable]
	public sealed class MirabilisTlvChain : IEnumerable, IByteConvertible
	{
		ListDictionary hash = new ListDictionary(); // ushort -> byte[]
		
		/// <summary>
		/// Creates a new, empty MirabilisTlvChain.
		/// </summary>
		public MirabilisTlvChain() {}
		
		/// <summary>
		/// Creates a new MirabilisTlvChain reading from the byte array.
		/// </summary>
		public MirabilisTlvChain(byte[] data) : this(data, 0, data.Length) {}
		
		/// <summary>
		/// Creates a new MirabilisTlvChain reading from the byte array.
		/// </summary>
		public MirabilisTlvChain(byte[] data, int index, int count)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			int pos = index;
			int end = index + count;
			while (pos < end) {
				ushort type = LEBitConverter.ToUInt16(data, pos);
				pos += 2;
				ushort length = LEBitConverter.ToUInt16(data, pos);
				pos += 2;
				byte[] value = new byte[length];
				Array.Copy(data, pos, value, 0, length);
				pos += length;
				hash.Add(type, value);
			}
		}
		
		/// <summary>
		/// Gets/Sets the value associated with the specified type.
		/// </summary>
		public byte[] this [ushort type] {
			get {
				return (byte[])hash[type];
			}
			set {
				if (value == null)
					hash.Remove(type);
				else
					hash[type] = value;
			}
		}
		
		IEnumerator IEnumerable.GetEnumerator()
		{
			return hash.Values.GetEnumerator();
		}
		
		/// <summary>
		/// Gets the value associated with the specified type.
		/// </summary>
		public byte[] GetData(ushort type)
		{
			return this[type];
		}
		
		/// <summary>
		/// Gets the value associated with the specified type.
		/// </summary>
		public void SetData(ushort type, byte[] value)
		{
			this[type] = value;
		}
		
		/// <summary>
		/// Removes the data associated with the specified type.
		/// </summary>
		public void Remove(ushort type)
		{
			this[type] = null;
		}
		
		/// <summary>
		/// Gets the string representation of the tlv of the specified type.
		/// </summary>
		public string GetString(ushort type)
		{
			return MirabilisTlv.DecodeString(this[type]);
		}
		
		/// <summary>
		/// Sets the string representation of the tlv of the specified type.
		/// </summary>
		public void SetString(ushort type, string value)
		{
			this[type] = MirabilisTlv.EncodeString(value);
		}
		
		/// <summary>
		/// Gets the short representation of the tlv of the specified type.
		/// </summary>
		public ushort GetShort(ushort type)
		{
			byte[] data = this[type];
			if (data == null) return 0;
			return LEBitConverter.ToUInt16(data, 0);
		}
		
		/// <summary>
		/// Sets the string representation of the tlv of the specified type.
		/// </summary>
		public void SetShort(ushort type, ushort value)
		{
			this[type] = LEBitConverter.GetBytes(value);
		}
		
		/// <summary>
		/// Gets the short representation of the tlv of the specified type.
		/// </summary>
		public uint GetInt(ushort type)
		{
			byte[] data = this[type];
			if (data == null) return 0;
			return LEBitConverter.ToUInt32(data, 0);
		}
		
		/// <summary>
		/// Sets the string representation of the tlv of the specified type.
		/// </summary>
		public void SetInt(ushort type, uint value)
		{
			this[type] = LEBitConverter.GetBytes(value);
		}
		
		/// <summary>
		/// Gets the total length of this MirabilisTlvChain.
		/// </summary>
		public int GetByteLength()
		{
			int l = 0;
			foreach (byte[] val in hash.Values) {
				l += 4 + val.Length;
			}
			return l;
		}
		
		/// <summary>
		/// Writes the byte representation of this MirabilisTlvChain into the specified stream.
		/// </summary>
		public void WriteBytes(Stream stream)
		{
			foreach (DictionaryEntry de in hash) {
				byte[] val = (byte[])de.Value;
				MirabilisTlv.Write((ushort)de.Key, val, stream);
			}
		}
		
		/// <summary>
		/// Gets this TlvChain as byte array.
		/// </summary>
		public byte[] ToByteArray()
		{
			byte[] b = new byte[GetByteLength()];
			using (MemoryStream ms = new MemoryStream(b, true)) {
				WriteBytes(ms);
			}
			return b;
		}
	}
	
	/// <summary>
	/// This is a helper class for writing TLVs in the old Mirabilis format.
	/// </summary>
	/// <seealso cref="MirabilisTlvChain"/>
	public sealed class MirabilisTlv
	{
		private MirabilisTlv() {}
		
		/// <summary>
		/// Encodes a string as null-terminated string preceded by length as little endian.
		/// </summary>
		public static byte[] EncodeString(string value)
		{
			if (value == null)
				return null;
			int len = Tlv.Encoding.GetByteCount(value) + 1;
			byte[] b = new byte[len + 2];
			LEBitConverter.PutBytes(b, 0, (ushort)len);
			Tlv.Encoding.GetBytes(value).CopyTo(b, 2);
			return b;
		}
		
		/// <summary>
		/// Decodes a string stored as null-terminated string preceded by length as little endian.
		/// </summary>
		public static string DecodeString(byte[] data)
		{
			if (data == null)
				return null;
			else
				return Tlv.Encoding.GetString(data, 2, LEBitConverter.ToUInt16(data, 0) - 1);
		}
		
		/// <summary>
		/// Decodes a string stored as null-terminated string preceded by length as little endian.
		/// The position pointer is set to the next element in the data structure.
		/// </summary>
		public static string ReadString(byte[] data, ref int pos)
		{
			int length = LEBitConverter.ToUInt16(data, pos) - 1;
			string val = Tlv.Encoding.GetString(data, pos + 2, length);
			pos += 3 + length;
			return val;
		}
		
		/// <summary>
		/// Writes the TLV to the stream.
		/// </summary>
		public static void Write(ushort type, byte[] data, Stream target)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			byte[] header = new byte[4];
			LEBitConverter.PutBytes(header, 0, type);
			LEBitConverter.PutBytes(header, 2, (ushort)data.Length);
			target.Write(header, 0, header.Length);
			target.Write(data, 0, data.Length);
		}
	}
}
