// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace Grunwald.NOscar
{
	/// <summary>
	/// <para>Negotiates the versions of specific families that are supported by client and server.</para>
	/// <para>This sequence is special because it does not use snac responses but listens directly to
	/// the connection.
	/// The ProtocolNegotiationSequence must be initialized before <see cref="SnacConnection.Run">SnacConnection.Run</see>
	/// is called. It registers with the SnacConnection in such a way that all unexpected snacs result in
	/// an error. When the sequence has completed, it deregisters from the SnacConnection and raises
	/// the <see cref="Completed">Completed event</see>.</para>
	/// <para>Because of this special behavior, it does not inherit from <see cref="SnacSequence"/></para>
	/// </summary>
	public sealed class ProtocolNegotiationSequence
	{
		SnacConnection connection;
		
		public ProtocolNegotiationSequence(SnacConnection connection)
		{
			if (connection.State != ConnectionState.Connected)
				throw new ArgumentException("The ProtocolNegotiationSequence can only be used on connections that are not already running.");
			this.connection = connection;
			connection.SnacReceived += new SnacEventHandler(OnSnacReceived);
		}
		
		void Stop()
		{
			connection.SnacReceived -= new SnacEventHandler(OnSnacReceived);
		}
		
		void OnSnacReceived(object sender, SnacEventArgs e)
		{
			if (e.Snac is ServerReadySnac)
				Process((ServerReadySnac)e.Snac);
			else if (e.Snac is ServerVersionsSnac)
				Process((ServerVersionsSnac)e.Snac);
			else
				OnError(EventArgs.Empty);
		}
		
		void Process(ServerReadySnac snac)
		{
			ushort[] families = snac.Families;
			int count = 0;
			for (int i = 0; i < families.Length; ++i) {
				if (GetFamilyVersion(families[i]) != 0)
					count += 1;
			}
			FamilyVersion[] versions = new FamilyVersion[count];
			int pos = 0;
			for (int i = 0; i < families.Length; ++i) {
				ushort ver = GetFamilyVersion(families[i]);
				if (ver != 0) {
					versions[pos++] = new FamilyVersion(families[i], ver);
				}
			}
			ClientVersionsSnac cvs = new ClientVersionsSnac(versions);
			cvs.RequestID = 0;
			connection.Send(cvs);
		}
		
		ushort GetFamilyVersion(ushort family)
		{
			switch (family) {
				case 0x01:
					return 4;
				case 0x02:
				case 0x03:
				case 0x04:
				case 0x06:
				case 0x08:
				case 0x09:
				case 0x0a:
				case 0x0b:
					return 1;
				case 0x13:
					return 3;
				default:
					return 0;
			}
		}
		
		void Process(ServerVersionsSnac snac)
		{
			connection.Data.ServerVersions = snac.Versions;
			OnCompleted(EventArgs.Empty);
		}
		
		public event EventHandler Error;
		
		void OnError(EventArgs e)
		{
			Stop();
			if (Error != null)
				Error(this, e);
		}
		
		public event EventHandler Completed;
		
		void OnCompleted(EventArgs e)
		{
			Stop();
			if (Completed != null)
				Completed(this, e);
		}
	}
}
