// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	/// <summary>
	/// A SNAC is the data format sent over FLAP channel 2.
	/// This is the base class for all Snacs.
	/// </summary>
	[Serializable]
	public abstract class Snac : IByteConvertible
	{
		ushort family;
		ushort subType;
		ushort flags;
		uint requestID;
		TlvChain extraData;
		
		static uint requestNum;
		
		public Snac()
		{
			lock (typeof(Snac)) {
				if (requestNum == int.MaxValue)
					requestNum = 0;
				this.requestID = ++requestNum;
			}
		}
		
		public Snac(ushort family, ushort subType) : this()
		{
			this.family = family;
			this.subType = subType;
		}
		
		/// <summary>
		/// Loads the Snac from the byte array.
		/// </summary>
		public Snac(byte[] data, int index, int count)
		{
			family = NetworkBitConverter.ToUInt16(data, index);
			subType = NetworkBitConverter.ToUInt16(data, index + 2);
			flags = NetworkBitConverter.ToUInt16(data, index + 4);
			requestID = NetworkBitConverter.ToUInt32(data, index + 6);
			if ((flags & 0x8000) == 0x8000) {
				int length = NetworkBitConverter.ToUInt16(data, index + 10);
				extraData = new TlvChain(data, index + 12, length);
			}
		}
		
		/// <summary>
		/// Gets/Sets the family id of this SNAC.
		/// </summary>
		public ushort Family {
			get { return family; }
			set { family = value; }
		}
		
		/// <summary>
		/// Gets/Sets the flags of this SNAC.
		/// </summary>
		public ushort Flags {
			get { return flags; }
			set { flags = value; }
		}
		
		/// <summary>
		/// Gets/Sets the request id of this SNAC.
		/// </summary>
		public uint RequestID {
			get { return requestID; }
			set { requestID = value; }
		}
		
		/// <summary>
		/// Gets/Sets the sub type of this SNAC.
		/// </summary>
		public ushort SubType {
			get { return subType; }
			set { subType = value; }
		}
		
		/// <summary>
		/// Gets/Sets the extra data tlv chain. In most cases, this is null.
		/// </summary>
		public TlvChain ExtraData {
			get {
				return extraData;
			}
			set {
				if (value == null)
					flags &= 0x7FFF;
				else
					flags |= 0x8000;
				extraData = value;
			}
		}
		
		/// <summary>
		/// Gets the total length this SNAC would need as byte array.
		/// </summary>
		public virtual int GetByteLength()
		{
			if (extraData == null)
				return 10;
			else
				return 12 + extraData.GetByteLength();
		}
		
		/// <summary>
		/// Writes the byte representation of this object into the specified stream.
		/// </summary>
		public virtual void WriteBytes(Stream stream)
		{
			byte[] data = new byte[10];
			NetworkBitConverter.PutBytes(data, 0, family);
			NetworkBitConverter.PutBytes(data, 2, subType);
			NetworkBitConverter.PutBytes(data, 4, flags);
			NetworkBitConverter.PutBytes(data, 6, requestID);
			stream.Write(data, 0, data.Length);
			if (extraData != null) {
				stream.Write(NetworkBitConverter.GetBytes((ushort)extraData.GetByteLength()), 0, 2);
				extraData.WriteBytes(stream);
			}
		}
		
		public override string ToString()
		{
			return String.Format("[{0}: Flags = {1:x}, RequestID = {2:x}]",
			                     GetType().Name,
			                     requestID,
			                     subType);
		}
	}
	
	/// <summary>TLV based snac</summary>
	[Serializable]
	public class TlvBasedSnac : Snac
	{
		TlvChain chain;
		
		/// <summary>
		/// Creates a new DataSnac.
		/// </summary>
		public TlvBasedSnac()
		{
			chain = new TlvChain();
		}
		
		/// <summary>
		/// Creates a new DataSnac.
		/// </summary>
		public TlvBasedSnac(ushort family, ushort subType) : base(family, subType)
		{
			chain = new TlvChain();
		}
		
		/// <summary>
		/// Loads the Snac from the byte array.
		/// </summary>
		public TlvBasedSnac(byte[] data, int index, int count) : base(data, index, count)
		{
			int baseLen = base.GetByteLength();
			chain = new TlvChain(data, index + baseLen, count - baseLen);
		}
		
		public TlvChain Chain {
			get {
				return chain;
			}
		}
		
		public override int GetByteLength()
		{
			return base.GetByteLength() + chain.GetByteLength();
		}
		
		public override void WriteBytes(Stream stream)
		{
			base.WriteBytes(stream);
			chain.WriteBytes(stream);
		}
	}
	
	/// <summary>A DataSnac is a snac that contains just a byte array.</summary>
	[Serializable]
	public class DataSnac : Snac
	{
		byte[] data;
		
		/// <summary>
		/// Creates a new DataSnac.
		/// </summary>
		public DataSnac() {}
		
		/// <summary>
		/// Loads the Snac from the byte array.
		/// </summary>
		public DataSnac(byte[] data, int index, int count) : base(data, index, count)
		{
			int baseLen = base.GetByteLength();
			this.data = new byte[count - baseLen];
			Array.Copy(data, index + baseLen, this.data, 0, this.data.Length);
		}
		
		/// <summary>
		/// Gets/Sets the data field of this SNAC.
		/// </summary>
		public byte[] Data {
			get { return data; }
			set {
				if (value == null)
					throw new ArgumentNullException("value");
				data = value;
			}
		}
		
		/// <summary>
		/// Gets the total length this SNAC would need as byte array.
		/// </summary>
		public override int GetByteLength()
		{
			return base.GetByteLength() + data.Length;
		}
		
		/// <summary>
		/// Writes the byte representation of this object into the specified stream.
		/// </summary>
		public override void WriteBytes(Stream stream)
		{
			base.WriteBytes(stream);
			stream.Write(data, 0, data.Length);
		}
		
		public override string ToString()
		{
			return String.Format("[{4}: Family = {0:x}, SubType = {3:x}, Flags = {1:x}, RequestID = {2:x}]",
			                     Family,
			                     Flags,
			                     RequestID,
			                     SubType,
			                     GetType().Name);
		}
	}
}
