// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;

namespace Grunwald.NOscar
{
	public class SnacConnection : IDisposable
	{
		FlapConnection flapConnection;
		SnacFactory snacFactory = new SnacFactory();
		
		ConnectionState state = ConnectionState.Connected;
		
		public ConnectionState State {
			get {
				return state;
			}
		}
		
		ConnectionData data = new ConnectionData();
		
		/// <summary>
		/// Gets the object storing information about this connection.
		/// </summary>
		public ConnectionData Data {
			get {
				return data;
			}
		}
		
		public event EventHandler StateChanged;
		
		protected virtual void OnStateChanged(EventArgs e)
		{
			if (StateChanged != null) {
				StateChanged(this, e);
			}
		}
		
		void SetState(ConnectionState state)
		{
			if (this.state != state) {
				this.state = state;
				OnStateChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler Closing;
		public event EventHandler Closed;
		
		protected virtual void OnClosed(EventArgs e)
		{
			if (Closed != null) {
				Closed(this, e);
			}
		}
		
		protected virtual void OnClosing(EventArgs e)
		{
			if (Closing != null) {
				Closing(this, e);
			}
		}
		
		public void Close()
		{
			OnClosing(EventArgs.Empty);
			flapConnection.Close();
			OnClosed(EventArgs.Empty);
		}
		
		void IDisposable.Dispose()
		{
			this.Close();
		}
		
		public SnacConnection(Stream stream) : this(new FlapConnection(stream)) {}
		
		public SnacConnection(FlapConnection flapConnection)
		{
			if (flapConnection == null)
				throw new ArgumentNullException("flapConnection");
			if (flapConnection.State != ConnectionState.Connected)
				throw new ArgumentException("The flap connection specified must be in state 'Connected'.");
			this.flapConnection = flapConnection;
			flapConnection.StateChanged += new EventHandler(FlapConnectionStateChanged);
			flapConnection.ReceiveFlap += new FlapEventHandler(FlapConnectionReceiveFlap);
		}
		
		void FlapConnectionStateChanged(object sender, EventArgs e)
		{
			if (flapConnection.State == ConnectionState.Dead) {
				Close();
			} else if (flapConnection.State == ConnectionState.Active) {
				FinishedNegotiation();
			}
		}
		
		/// <summary>
		/// Is called when the SnacConnection has finished negotiation and higher level
		/// connections may start their own negotiation.
		/// If not overridden, this sets the state to active.
		/// </summary>
		protected virtual void FinishedNegotiation()
		{
			SetState(ConnectionState.Active);
		}
		
		void FlapConnectionReceiveFlap(object sender, FlapEventArgs e)
		{
			switch (e.Flap.Channel) {
				case FlapChannel.KeepAlive:
					break;
				case FlapChannel.Snac:
					ReceiveSnac(e.Flap.Data);
					break;
				case FlapChannel.ConnectionClose:
					Console.WriteLine("Close-Flap: " + Tlv.Encoding.GetString(e.Flap.Data));
					this.Close();
					break;
				default:
					throw new InvalidFlapException("Unknown channel " + e.Flap.Channel.ToString());
			}
		}
		
		void ReceiveSnac(byte[] data)
		{
			ReceiveSnac(snacFactory.CreateSnac(data, 0, data.Length));
		}
		
		public event SnacEventHandler SnacReceived;
		
		void ReceiveSnac(Snac snac)
		{
			Console.WriteLine("<- {0}", snac);
			OnSnacReceived(new SnacEventArgs(snac));
		}
		
		protected virtual void OnSnacReceived(SnacEventArgs e)
		{
			if (SnacReceived != null) {
				SnacReceived(this, e);
			}
		}
		
		public void Run()
		{
			if (state != ConnectionState.Connected)
				throw new InvalidOperationException();
			SetState(ConnectionState.Negotiating);
			flapConnection.Run();
		}
		
		public virtual void Send(Snac snac)
		{
			Console.WriteLine("-> {0}", snac);
			using (MemoryStream ms = new MemoryStream(snac.GetByteLength())) {
				snac.WriteBytes(ms);
				flapConnection.SendFlap(FlapChannel.Snac, ms.ToArray());
			}
		}
	}
	
	public delegate void SnacEventHandler(object sender, SnacEventArgs e);
	
	public class SnacEventArgs : EventArgs
	{
		Snac snac;
		
		public Snac Snac {
			get {
				return snac;
			}
		}
		
		public SnacEventArgs(Snac snac)
		{
			this.snac = snac;
		}
	}
}
