// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Text;

namespace Grunwald.NOscar
{
	/// <summary>
	/// A TLV is a data structure containing only a block of data and a "data type" code.
	/// (The name comes from <b>T</b>ype, <b>L</b>ength, <b>V</b>alue.)<br/>
	/// A TlvChain is a series of TLVs.
	/// </summary>
	/// <remarks>
	/// Unfortunately, data type codes (type) do not identify the "type" of the
	/// data enclosed in a traditional sense. Instead, TLV types are only relevant
	/// for the command or data structure in which they are present.<br/>
	/// For example, a TLV of type 0x0001 contains a US-ASCII-encoded screenname in
	/// the authorization SNAC command, but a TLV of type 0x0001 in the security
	/// information part of one's UserInfo contains the user's X.509v3 certificate.
	/// </remarks>
	[Serializable]
	public sealed class TlvChain : IEnumerable, IByteConvertible
	{
		ListDictionary hash = new ListDictionary(); // ushort -> byte[]
		
		/// <summary>
		/// Creates a new, empty TlvChain.
		/// </summary>
		public TlvChain() {}
		
		/// <summary>
		/// Creates a new TlvChain reading from the byte array.
		/// </summary>
		public TlvChain(byte[] data) : this(data, 0, data.Length) {}
		
		/// <summary>
		/// Creates a new TlvChain reading from the byte array.
		/// </summary>
		public TlvChain(byte[] data, int index, int byteCount)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			int pos = index;
			int end = index + byteCount;
			while (pos < end) {
				ushort type = NetworkBitConverter.ToUInt16(data, pos);
				pos += 2;
				ushort length = NetworkBitConverter.ToUInt16(data, pos);
				pos += 2;
				byte[] value = new byte[length];
				Array.Copy(data, pos, value, 0, length);
				pos += length;
				hash.Add(type, value);
			}
		}
		
		/// <summary>
		/// Creates a new TlvChain reading from the byte array.
		/// Unlike the constructur, this methods takes the number of TLVs instead of the number of bytes.
		/// </summary>
		public static TlvChain FromData(byte[] data, int index, int tlvCount)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			TlvChain chain = new TlvChain();
			int pos = index;
			for (int i = 0; i < tlvCount; ++i) {
				ushort type = NetworkBitConverter.ToUInt16(data, pos);
				pos += 2;
				ushort length = NetworkBitConverter.ToUInt16(data, pos);
				pos += 2;
				byte[] value = new byte[length];
				Array.Copy(data, pos, value, 0, length);
				pos += length;
				chain.hash.Add(type, value);
			}
			return chain;
		}
		
		/// <summary>
		/// Gets/Sets the value associated with the specified type.
		/// </summary>
		public byte[] this [ushort type] {
			get {
				return (byte[])hash[type];
			}
			set {
				if (value == null)
					hash.Remove(type);
				else
					hash[type] = value;
			}
		}
		
		IEnumerator IEnumerable.GetEnumerator()
		{
			return hash.Values.GetEnumerator();
		}
		
		/// <summary>
		/// Gets the value associated with the specified type.
		/// </summary>
		public byte[] GetData(ushort type)
		{
			return this[type];
		}
		
		/// <summary>
		/// Gets the value associated with the specified type.
		/// </summary>
		public void SetData(ushort type, byte[] value)
		{
			this[type] = value;
		}
		
		static byte[] empty = new byte[0];
		
		/// <summary>
		/// Creates an empty value for the specified type.
		/// </summary>
		public void SetFlag(ushort type)
		{
			this[type] = empty;
		}
		
		/// <summary>
		/// Removes the data associated with the specified type.
		/// </summary>
		public void Remove(ushort type)
		{
			this[type] = null;
		}
		
		/// <summary>
		/// Gets the string representation of the tlv of the specified type.
		/// </summary>
		public string GetString(ushort type)
		{
			byte[] data = this[type];
			if (data == null) return null;
			return Tlv.Encoding.GetString(data);
		}
		
		/// <summary>
		/// Sets the string representation of the tlv of the specified type.
		/// </summary>
		public void SetString(ushort type, string value)
		{
			if (value == null)
				this[type] = null;
			else
				this[type] = Tlv.Encoding.GetBytes(value);
		}
		
		/// <summary>
		/// Gets the short representation of the tlv of the specified type.
		/// </summary>
		public ushort GetShort(ushort type)
		{
			byte[] data = this[type];
			if (data == null) return 0;
			return NetworkBitConverter.ToUInt16(data, 0);
		}
		
		/// <summary>
		/// Sets the string representation of the tlv of the specified type.
		/// </summary>
		public void SetShort(ushort type, ushort value)
		{
			this[type] = NetworkBitConverter.GetBytes(value);
		}
		
		/// <summary>
		/// Gets the short representation of the tlv of the specified type.
		/// </summary>
		public uint GetInt(ushort type)
		{
			byte[] data = this[type];
			if (data == null) return 0;
			return NetworkBitConverter.ToUInt32(data, 0);
		}
		
		/// <summary>
		/// Sets the string representation of the tlv of the specified type.
		/// </summary>
		public void SetInt(ushort type, uint value)
		{
			this[type] = NetworkBitConverter.GetBytes(value);
		}
		
		/// <summary>
		/// Gets the total length of this TlvChain.
		/// </summary>
		public int GetByteLength()
		{
			int l = 0;
			foreach (byte[] val in hash.Values) {
				l += 4 + val.Length;
			}
			return l;
		}
		
		/// <summary>
		/// Gets the number of elements in this TlvChain.
		/// </summary>
		public int Count {
			get {
				return hash.Count;
			}
		}
		
		/// <summary>
		/// Writes the byte representation of this TlvChain into the specified stream.
		/// </summary>
		public void WriteBytes(Stream stream)
		{
			foreach (DictionaryEntry de in hash) {
				byte[] val = (byte[])de.Value;
				Tlv.Write((ushort)de.Key, val, stream);
			}
		}
		
		/// <summary>
		/// Gets this TlvChain as byte array.
		/// </summary>
		public byte[] ToByteArray()
		{
			byte[] b = new byte[GetByteLength()];
			using (MemoryStream ms = new MemoryStream(b, true)) {
				WriteBytes(ms);
			}
			return b;
		}
	}
	
	/// <summary>
	/// This is a helper class for writing TLVs.
	/// </summary>
	/// <remarks>
	/// A TLV is a data structure containing only a block of data and a "data type" code.
	/// (The name comes from <b>T</b>ype, <b>L</b>ength, <b>V</b>alue.)<br/>
	/// Unfortunately, data type codes (type) do not identify the "type" of the
	/// data enclosed in a traditional sense. Instead, TLV types are only relevant
	/// for the command or data structure in which they are present.<br/>
	/// For example, a TLV of type 0x0001 contains a US-ASCII-encoded screenname in
	/// the authorization SNAC command, but a TLV of type 0x0001 in the security
	/// information part of one's UserInfo contains the user's X.509v3 certificate.
	/// </remarks>
	public sealed class Tlv
	{
		private Tlv() {}
		
		// I don't know what encoding to use, I think anything US-ASCII compatible will do it
		internal static Encoding Encoding = Encoding.GetEncoding("ISO8859-1");
		
		/// <summary>
		/// Writes the TLV to the stream.
		/// </summary>
		public static void Write(ushort type, byte[] data, Stream target)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			byte[] header = new byte[4];
			NetworkBitConverter.PutBytes(header, 0, type);
			NetworkBitConverter.PutBytes(header, 2, (ushort)data.Length);
			target.Write(header, 0, header.Length);
			target.Write(data, 0, data.Length);
		}
	}
}
