// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Threading;
using DGNetwork.Chat;
using DGNetwork.Services;
using Grunwald.NOscar;

namespace DGNetwork.AddIns.ICQ
{
	/// <summary>
	/// Wraps an ICQ network.
	/// </summary>
	public class IcqNetwork : IDisposable
	{
		string server;
		int port;
		int icqNr;
		string password;
		
		public int IcqNr {
			get {
				return icqNr;
			}
		}
		
		IcqNetworkPage page;
		
		public IcqNetworkPage Page {
			get {
				return page;
			}
		}
		
		public IcqNetwork(string server, int icqNr, string password, IcqNetworkPage page)
		{
			int pos = server.IndexOf(':');
			if (pos < 0) {
				this.server = server;
				this.port = FlapConnection.DefaultPort;
			} else {
				this.server = server.Substring(0, pos);
				this.port = int.Parse(server.Substring(pos + 1));
			}
			this.icqNr = icqNr;
			this.password = password;
			this.page = page;
			lock (ActiveNetworks) {
				ActiveNetworks.Add(this);
			}
			Service.Chat.StatusChanged += new StatusChangedEventHandler(ChatStatusChanged);
			AcceptStatus(Service.Chat.Status);
		}
		
		void ChatStatusChanged(object sender, StatusChangedEventArgs e)
		{
			AcceptStatus(e.NewStatus);
		}
		
		void AcceptStatus(string newStatus)
		{
			if (newStatus == "Offline")
				Disconnect();
			else
				Connect();
		}
		
		public void Dispose()
		{
			lock (ActiveNetworks) {
				ActiveNetworks.Remove(this);
			}
			Service.Chat.StatusChanged -= new StatusChangedEventHandler(ChatStatusChanged);
			Disconnect();
		}
		
		AutomaticConnection connection;
		internal static ArrayList ActiveNetworks = new ArrayList();
		
		public AutomaticConnection Connection {
			get {
				return connection;
			}
		}
		
		void Connect()
		{
			if (connection != null) return;
			connection = new AutomaticConnection(server, port, icqNr.ToString(), password);
			connection.SnacReceived += new SnacEventHandler(OnSnacReceived);
			connection.Start(new ThreadExceptionEventHandler(OnInternalError));
		}
		
		void Disconnect()
		{
			AutomaticConnection c = connection;
			connection = null;
			if (c != null) {
				c.Stop();
			}
		}
		
		void OnInternalError(object sender, ThreadExceptionEventArgs e)
		{
			Service.Message.ShowError(e.Exception);
			try {
				Disconnect();
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		void OnSnacReceived(object sender, SnacEventArgs e)
		{
			IncomingIcbmSnac snac = e.Snac as IncomingIcbmSnac;
			if (snac != null) {
				try {
					ReceiveIcbm(snac);
				} catch (Exception ex) {
					Service.Message.ShowError(ex);
				}
			}
		}
		
		void ReceiveIcbm(IncomingIcbmSnac snac)
		{
			string source = snac.UserInfo.Screenname;
			int uin = int.Parse(source);
			Message msg = new Message();
			msg.NotifyUser = true;
			if (snac.Channel == 1) {
				MessageBlock msgData = new MessageBlock(snac.DataChain.GetData(2));
				msg.Text = msgData.GetText();
			} else if (snac.Channel == 4) {
				Service.Message.ShowMessage("Incoming ICBM from " + source + " on channel " + snac.Channel.ToString());
				//throw new NotImplementedException("Sorry, receiving channel 4 ICBMs is not implemented.");
				//} else {
				//return;
			}
			IcqUser user = IcqUser.GetUser(page, uin);
			msg.Nickname = user.Name;
			user.PrivateChatRoom.ReceiveMessage(msg);
		}
	}
}
