// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using DGNetwork.Search;
using DGNetwork.Services;
using Grunwald.NOscar;

namespace DGNetwork.AddIns.ICQ
{
	public class IcqSearchManager : AbstractManager
	{
		public IcqSearchManager() : base(Icons.Online, "ICQ")
		{
			
		}
		
		public override bool SupportsCriteria(ICriteria criteria)
		{
			switch (criteria.Name) {
				case "Username":
				case "EMail":
				case "ICQ":
					return true;
				default:
					return false;
			}
		}
		
		public override ISearch BeginSearch(ISearchType searchType)
		{
			ArrayList searches = new ArrayList();
			lock (IcqNetwork.ActiveNetworks) {
				foreach (IcqNetwork network in IcqNetwork.ActiveNetworks) {
					AutomaticConnection c = network.Connection;
					if (c == null || c.State != ConnectionState.Active) {
						DefaultSearch search = new DefaultSearch();
						DefaultResult result = new DefaultResult();
						result.Icon = Icons.Offline;
						result.Name = "Not connected with ICQ";
						search.AddResult(result);
						search.Finish();
						searches.Add(search);
					} else {
						searches.Add(new IcqSearch(network));
					}
				}
			}
			foreach (ISearch s in searches) {
				if (s is IcqSearch)
					((IcqSearch)s).Start(searchType);
			}
			if (searches.Count == 0)
				return null;
			if (searches.Count == 1)
				return (ISearch)searches[0];
			MultiSearch ms = new MultiSearch();
			foreach (ISearch s in searches) {
				ms.AddSearch(s);
			}
			return ms;
		}
	}
	
	public class IcqSearch : DefaultSearch
	{
		IcqNetwork network;
		
		public IcqSearch(IcqNetwork network)
		{
			this.network = network;
		}
		
		ushort id;
		
		public override void Abort()
		{
			Stop();
			base.Abort();
		}
		
		AutomaticConnection conn;
		
		public void Start(ISearchType searchType)
		{
			MirabilisTlvChain c = new MirabilisTlvChain();
			
			IcqMetaRequestSubType subType;
			
			string icq = searchType.Criterias["ICQ"].Value as string;
			if (icq != null) icq = icq.Replace(" ", "");
			if (icq != null && icq.Length > 0) {
				subType = IcqMetaRequestSubType.SearchByUinTlv;
				c.SetInt(0x0136, uint.Parse(icq));
			} else {
				subType = IcqMetaRequestSubType.SearchWhitepagesTlv;
				
				string nickname = searchType.Criterias["Username"].Value as string;
				if (nickname != null && nickname.Length > 0)
					c.SetString(0x0154, nickname);
				
				string email = searchType.Criterias["EMail"].Value as string;
				if (email != null && email.Length > 0)
					c.SetString(0x015e, email);
				
				if (c.GetByteLength() == 0) {
					// no criteria set -> return nothing
					this.Finish();
					return;
				}
			}
			
			IcqMetaInfoRequestSnac snac = new IcqMetaInfoRequestSnac(network.IcqNr, subType, c.ToByteArray());
			id = snac.InfoSequenceNumber;
			conn = network.Connection;
			conn.SnacReceived += new SnacEventHandler(OnReceive);
			conn.Send(snac);
		}
		
		void Stop()
		{
			conn.SnacReceived -= new SnacEventHandler(OnReceive);
		}
		
		void OnReceive(object sender, SnacEventArgs e)
		{
			IcqMetaInfoResponseSnac snac = e.Snac as IcqMetaInfoResponseSnac;
			if (snac == null)
				return;
			if (snac.InfoSequenceNumber == id) {
				byte[] data = snac.InfoData;
				IcqMetaInfoSearchResult result = null;
				if (data[0] == 0x0a) { // success byte
					if (snac.InfoSubType == IcqMetaResponseSubType.SearchUserFoundReply) {
						result = new IcqMetaInfoSearchResult(data, 3, data.Length - 3);
					} else if (snac.InfoSubType == IcqMetaResponseSubType.SearchLastUserFoundReply) {
						result = new IcqMetaInfoSearchResult(data, 3, data.Length - 7);
					} else {
						Stop();
						this.Finish();
						return;
					}
				}
				if (result != null) {
					this.AddResult(new IcqSearchResult(network, result));
				}
				if (snac.InfoSubType == IcqMetaResponseSubType.SearchLastUserFoundReply) {
					Stop();
					this.Finish();
				}
			}
		}
	}
	
	public class IcqSearchResult : DefaultResult
	{
		int uin;
		IcqNetwork network;
		
		public IcqSearchResult(IcqNetwork network, IcqMetaInfoSearchResult result)
		{
			this.network = network;
			uin = (int)result.Uin;
			this.Name = result.Nickname;
			this.Icon = (result.OnlineStatus == 1) ? Icons.Online : Icons.Offline;
			this.Columns.Add(new ResultColumn("ICQ#", uin.ToString()));
			if (result.FirstName.Length > 0 && result.LastName.Length > 0) {
				this.Columns.Add(new ResultColumn("${res:SearchDialog.RealName}", result.FirstName + " " + result.LastName));
			} else if (result.FirstName.Length > 0) {
				this.Columns.Add(new ResultColumn("${res:SearchDialog.RealName}", result.FirstName));
			} else if (result.LastName.Length > 0) {
				this.Columns.Add(new ResultColumn("${res:SearchDialog.RealName}", result.LastName));
			}
			if (result.EMail.Length > 0)
				this.Columns.Add(new ResultColumn("${res:SearchDialog.EMail}", result.EMail));
		}
		
		public override void DoubleClick()
		{
			Service.Chat.OpenChat(IcqUser.GetUser(network.Page, uin, this.Name).PrivateChatRoom);
		}
	}
}
