// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using DGNetwork.Chat;
using DGNetwork.Services;
using Meebey.SmartIrc4net;

namespace DGNetwork.AddIns.IRC
{
	public class ChannelChatRoom : AbstractChatRoom
	{
		ArrayList users = new ArrayList();
		IrcClient irc;
		string name;
		
		public IrcClient IrcClient {
			get {
				return irc;
			}
		}
		
		public string Name {
			get {
				return name;
			}
		}
		
		public override bool DisposeOnLeave {
			get {
				return true;
			}
		}
		
		public ChannelChatRoom(IrcClient irc, string name)
		{
			this.irc = irc;
			this.name = name;
			this.Title = name;
			this.CanSend = false;
		}
		
		bool joined = false;
		
		public bool Joined {
			get {
				return joined;
			}
		}
		
		public void Join()
		{
			if (joined)
				throw new InvalidOperationException("The ChannelChatRoom is already joined.");
			joined = true;
			ReceiveMessageAndNotify("Joining the channel...");
			irc.OnChannelActiveSynced += new IrcEventHandler(ChannelActiveSynced);
			irc.OnChannelPassiveSynced += new IrcEventHandler(ChannelPassiveSynced);
			irc.OnChannelAction += new ActionEventHandler(ChannelAction);
			irc.OnChannelMessage += new IrcEventHandler(ChannelMessage);
			irc.OnChannelNotice += new IrcEventHandler(ChannelNotice);
			irc.OnJoin += new JoinEventHandler(IrcJoin);
			irc.OnPart += new PartEventHandler(IrcPart);
			irc.OnQuit += new QuitEventHandler(IrcQuit);
			irc.OnNickChange += new NickChangeEventHandler(IrcNickChange);
			irc.OnVoice += new VoiceEventHandler(IrcVoice);
			irc.OnDevoice += new DevoiceEventHandler(IrcDevoice);
			irc.OnKick += new KickEventHandler(IrcKick);
			irc.OnTopicChange += new TopicChangeEventHandler(IrcTopicChange);
			irc.RfcJoin(name);
		}
		
		protected override void OnDisposed(EventArgs e)
		{
			irc.OnChannelActiveSynced -= new IrcEventHandler(ChannelActiveSynced);
			irc.OnChannelPassiveSynced -= new IrcEventHandler(ChannelPassiveSynced);
			irc.OnChannelAction -= new ActionEventHandler(ChannelAction);
			irc.OnChannelMessage -= new IrcEventHandler(ChannelMessage);
			irc.OnChannelNotice -= new IrcEventHandler(ChannelNotice);
			irc.OnJoin -= new JoinEventHandler(IrcJoin);
			irc.OnPart -= new PartEventHandler(IrcPart);
			irc.OnQuit -= new QuitEventHandler(IrcQuit);
			irc.OnNickChange -= new NickChangeEventHandler(IrcNickChange);
			irc.OnVoice -= new VoiceEventHandler(IrcVoice);
			irc.OnDevoice -= new DevoiceEventHandler(IrcDevoice);
			irc.OnKick -= new KickEventHandler(IrcKick);
			irc.OnTopicChange -= new TopicChangeEventHandler(IrcTopicChange);
			
			// Part is only called by IrcServerContact through the Disposed-event.
			base.OnDisposed(e);
		}
		
		public void Part()
		{
			if (!joined)
				return;
			irc.RfcPart(name);
			joined = false;
		}
		
		bool firstJoin = true;
		
		private void ChannelActiveSynced(object sender, IrcEventArgs e)
		{
			if (e.Data.Channel != name)
				return;
			if (firstJoin) {
				this.UserListVisible = true;
				this.CanSend = true;
				Channel channel = irc.GetChannel(name);
				ReceiveMessageAndNotify("Channel was joined, the topic is " + channel.Topic + " (" + channel.Users.Count + " of " + channel.UserLimit + " users)");
				firstJoin = false;
			}
			UpdateUsers();
		}
		
		private void ChannelPassiveSynced(object sender, IrcEventArgs e)
		{
			if (e.Data.Channel == name)
				UpdateUsers();
		}
		
		private void UpdateUsers()
		{
			Channel channel = irc.GetChannel(name);
			bool dirty = false;
			lock (this.Users) {
				foreach (IrcChannelUser user in this.Users.ToArray()) {
					if (channel.Users[user.Name] == null) {
						dirty = true;
						this.Users.Remove(user);
					}
				}
				foreach (ChannelUser user in channel.Users.Values) {
					bool ok = false;
					foreach (IrcChannelUser ircUser in this.Users) {
						if (ircUser.Name == user.Nick) {
							ok = true;
							break;
						}
					}
					if (!ok) {
						dirty = true;
						this.Users.Add(new IrcChannelUser(irc, user));
					}
				}
			}
			if (dirty) {
				this.OnUsersChanged(EventArgs.Empty);
			}
		}
		
		private void ChannelAction(object sender, ActionEventArgs e)
		{
			if (e.Data.Channel == name) {
				Message msg = new Message(false);
				msg.Text = e.ActionMessage;
				msg.Nickname = e.Data.Nick;
				msg.IsAction = true;
				this.ReceiveMessage(msg);
			}
		}
		
		private void ChannelMessage(object sender, IrcEventArgs e)
		{
			if (e.Data.Channel == name)
				ReceiveMessageAndNotify(e.Data.Nick, e.Data.Message);
		}
		
		private void ChannelNotice(object sender, IrcEventArgs e)
		{
			if (e.Data.Channel == name)
				ReceiveMessageAndNotify("Notice", e.Data.Message);
		}
		
		private void IrcJoin(object sender, JoinEventArgs e)
		{
			if (e.Channel == name)
				ReceiveMessage(e.Who + " has joined " + e.Channel);
		}
		
		private void IrcPart(object sender, PartEventArgs e)
		{
			if (e.Channel == name)
				ReceiveMessage(e.Who + " has left " + e.Channel + " (" + e.PartMessage + ")");
		}
		
		private void IrcQuit(object sender, QuitEventArgs e)
		{
			if (e.Data.Channel == name)
				ReceiveMessage(e.Who + " has quit: " + e.QuitMessage);
		}
		
		private void IrcNickChange(object sender, NickChangeEventArgs e)
		{
			if (e.Data.Channel == name)
				ReceiveMessageAndNotify(e.OldNickname + " is now known as " + e.NewNickname);
		}
		
		private void IrcVoice(object sender, VoiceEventArgs e)
		{
			if (e.Channel == name)
				ReceiveMessage(e.Who + " gives voice to " + e.Whom);
		}
		
		private void IrcDevoice(object sender, DevoiceEventArgs e)
		{
			if (e.Channel == name)
				ReceiveMessage(e.Who + " takes voice from " + e.Whom);
		}
		
		private void IrcKick(object sender, KickEventArgs e)
		{
			if (e.Channel == name)
				ReceiveMessage(e.Who + " kicks " + e.Whom + ": " + e.KickReason);
		}
		
		private void IrcTopicChange(object sender, TopicChangeEventArgs e)
		{
			if (e.Channel == name)
				ReceiveMessage(e.Who + " sets topic to " + e.NewTopic);
		}
		
		public override ArrayList Users {
			get {
				return users;
			}
		}
		
		public override IMessageSendResult SendMessage(Message message)
		{
			return IrcService.SendMessage(irc, name, message);
		}
		
		public void ReceiveMessageAndNotify(string text)
		{
			ReceiveMessageAndNotify(null, text);
		}
		
		public void ReceiveMessage(string text)
		{
			Message msg = new Message(false);
			msg.Text = text;
			msg.Nickname = null;
			msg.NotifyUser = false;
			this.ReceiveMessage(msg);
		}
		
		public void ReceiveMessageAndNotify(string from, string text)
		{
			Message msg = new Message(false);
			msg.Text = text;
			msg.Nickname = from;
			this.ReceiveMessage(msg);
		}
	}
}
