// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using DGNetwork.Services;
using ICSharpCode.Core.AddIns.Codons;

namespace DGNetwork.AddIns.IRC
{
	#region IdentServer
	/// <summary>
	/// Provides an identification server according to RFC 1413.
	/// </summary>
	public class IdentServer : IDisposable
	{
		public const int DefaultPort = 113;
		
		TcpListener listener;
		Thread thread;
		
		public IdentServer()
		{
			listener = new TcpListener(IPAddress.Any, DefaultPort);
		}
		
		public IdentServer(int port)
		{
			listener = new TcpListener(IPAddress.Any, port);
		}
		
		bool running;
		
		public bool Start()
		{
			if (running)
				return true;
			try {
				listener.Start();
			} catch (SocketException) {
				return false;
			}
			thread = new Thread(new ThreadStart(Run));
			thread.IsBackground = true;
			thread.Name = "Identification server thread";
			running = true;
			thread.Start();
			return true;
		}
		
		private void Run()
		{
			try {
				while (running) {
					while (listener.Pending()) {
						BeginProcess(listener.AcceptTcpClient());
					}
					Thread.Sleep(100);
				}
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			} finally {
				listener.Stop();
				running = false;
			}
		}
		
		private delegate void void_TcpClient(TcpClient tcpClient);
		
		private void BeginProcess(TcpClient client)
		{
			new void_TcpClient(DoProcess).BeginInvoke(client, null, null);
		}
		
		private void DoProcess(TcpClient client)
		{
			try {
				Process(client);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			} finally {
				client.Close();
			}
		}
		
		/// <summary>
		/// Processes requests from a TcpClient
		/// </summary>
		public void Process(TcpClient client)
		{
			Process(client.GetStream());
		}
		
		/// <summary>
		/// Processes requests on the specified stream (should be a NetworkStream)
		/// </summary>
		public void Process(Stream stream)
		{
			StreamReader reader = new StreamReader(stream);
			StreamWriter writer = new StreamWriter(stream);
			writer.AutoFlush = true;
			string line;
			try {
				while ((line = reader.ReadLine()) != null) {
					writer.WriteLine(Process(line));
				}
			} catch (IOException) {
				// ignore receive/send errors
			}
		}
		
		/// <summary>
		/// Processes the specified request line. Returns the response the server would send.
		/// </summary>
		public string Process(string line)
		{
			if (line == null)
				throw new ArgumentNullException("line");
			line = line.Trim();
			string errorPrefix = line + " : ERROR : ";
			int pos = line.IndexOf(',');
			if (pos <= 0)
				return errorPrefix + "UNKNOWN-ERROR";
			string portOnServer = line.Substring(0, pos).Trim();
			string portOnClient = line.Substring(pos + 1).Trim();
			StringBuilder b = new StringBuilder();
			b.Append(portOnServer);
			b.Append(", ");
			b.Append(portOnClient);
			b.Append(" : USERID : UNIX : ");
			b.Append(Environment.UserName);
			/*try {
				string machineName = Environment.MachineName;
				b.Append('@');
				b.Append(machineName);
			} catch {
				// ignore errors
			}*/
			return b.ToString();
		}
		
		public void Stop()
		{
			if (running) {
				running = false;
				thread.Join();
			}
		}
		
		void IDisposable.Dispose()
		{
			Stop();
		}
	}
	#endregion
	
	#region Autostart server
	public class StartIdentServer : AbstractCommand
	{
		public override void Run()
		{
			IdentServer identServer = new IdentServer();
			identServer.Start();
		}
	}
	#endregion
}
