// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Xml;
using DGNetwork.Chat;
using DGNetwork.Services;
using DGNetwork.Contacts;
using ICSharpCode.Core.Properties;

namespace DGNetwork.AddIns.IRC
{
	public class IrcServerContact : FixedContactList, IXmlConvertable
	{
		string server, port;
		
		#region Constructor
		/// <summary>Should only be used to read contact from xml.</summary>
		public IrcServerContact()
		{
		}
		
		public IrcServerContact(string server, string port)
		{
			this.server = server;
			this.port = port;
			Init();
		}
		
		private void Init()
		{
			this.ContextMenuTreePath = "/Context/Contacts/IRCServer";
			UpdateIcon();
		}
		#endregion
		
		private void UpdateIcon()
		{
			if (this.Connected)
				this.Icon = "Icons.IrcOnline";
			else
				this.Icon = "Icons.IrcOffline";
		}
		
		#region Properties
		public string Port {
			get {
				return port;
			}
		}
		
		public string Server {
			get {
				return server;
			}
		}
		#endregion
		
		#region To/From Xml
		public object FromXmlElement(XmlElement element)
		{
			Service.Contact.LoadCommonProperties(element, this);
			server = element.GetAttribute("server");
			port = element.GetAttribute("port");
			Init();
			XmlElement channels = element["channels"];
			if (channels != null) {
				foreach (XmlElement channel in channels) {
					this.Contacts.Add(new ChannelContact(channel.GetAttribute("name")));
				}
			}
			return this;
		}
		
		public XmlElement ToXmlElement(XmlDocument doc)
		{
			XmlElement element = Service.Contact.GetElementWithCommonProperties(doc, this);
			element.SetAttribute("server", server);
			element.SetAttribute("port", port);
			XmlElement channels = doc.CreateElement("channels");
			foreach (ChannelContact channelContact in this.Contacts) {
				XmlElement channel = doc.CreateElement("channel");
				channel.SetAttribute("name", channelContact.Name);
				channels.AppendChild(channel);
			}
			element.AppendChild(channels);
			return element;
		}
		#endregion
		
		#region Connect/Disconnect
		ServerChatRoom room = new ServerChatRoom();
		Connection connection;
		
		public void Connect()
		{
			if (connection == null) {
				connection = new Connection(server, 6667);
				connection.Disposed += new EventHandler(ConnectionDisposed);
				room.Title = this.Name;
				room.Connection = connection;
				Service.Chat.OpenChat(room);
				room.ReceiveMessage("Connecting to server...", false);
				connection.Connect();
			}
			UpdateIcon();
		}
		
		public void Disconnect()
		{
			connection.Dispose();
		}
		
		private void ConnectionDisposed(object sender, EventArgs e)
		{
			if (connection == sender)
				connection = null;
			foreach (ChannelChatRoom room in new ArrayList(openChannels.Values)) {
				room.Dispose();
			}
			UpdateIcon();
		}
		
		public bool Connected {
			get {
				return connection != null;
			}
		}
		#endregion
		
		#region Channel Management
		public void AddChannel()
		{
			string name = Service.Message.ShowInputBox("${res:IRC.AddChannelText}", "${res:IRC.AddChannel}", "#name");
			if (name == null)
				return;
			this.Contacts.Add(new ChannelContact(name));
		}
		
		public ChannelContact FindChannel(string name)
		{
			foreach (ChannelContact c in this.Contacts) {
				if (c.Name == name)
					return c;
			}
			return null;
		}
		
		Hashtable openChannels = new Hashtable();
		
		public void OpenChannel(string name)
		{
			ChannelContact c = FindChannel(name);
			if (c != null)
				c.Icon = "Icons.IrcOnline";
			if (openChannels.Contains(name)) {
				Service.Chat.OpenChat((ChannelChatRoom)openChannels[name], true);
			} else {
				ChannelChatRoom room = new ChannelChatRoom(connection.IrcClient, name);
				room.Disposed += new EventHandler(ChannelRoomDisposed);
				openChannels.Add(name, room);
				Service.Chat.OpenChat(room, true);
				room.Join();
			}
		}
		
		private void ChannelRoomDisposed(object sender, EventArgs e)
		{
			ChannelChatRoom room = (ChannelChatRoom)sender;
			if (this.Connected && room.Joined)
				room.Part();
			openChannels.Remove(room.Name);
			ChannelContact c = FindChannel(room.Name);
			if (c != null)
				c.Icon = "Icons.IrcOffline";
		}
		#endregion
		
		public override bool CanClone {
			get {
				return false;
			}
		}
		
		public override IContact Clone()
		{
			throw new NotSupportedException();
		}
		
		protected override void OnNameChanged(EventArgs e)
		{
			room.Title = this.Name;
			base.OnNameChanged(e);
		}
	}
}
