// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.Text;
using DGNetwork.Chat;
using DGNetwork.Services;
using Meebey.SmartIrc4net;

namespace DGNetwork.AddIns.IRC
{
	public sealed class IrcService
	{
		private IrcService() { }
		
		public static IMessageSendResult SendMessage(IrcClient irc, string target, Message message)
		{
			string text = message.Text.Trim();
			Message textOnlyMessage = new Message(message.Sending);
			textOnlyMessage.Text = text;
			textOnlyMessage.Nickname = message.Nickname;
			bool isCommand = text.StartsWith("/");
			MessageSendStatus status;
			string resultText = null;
			if (isCommand) {
				status = ExecuteCommand(irc, text, target, ref resultText, textOnlyMessage);
			} else {
				irc.SendMessage(SendType.Message, target, text);
				status = MessageSendStatus.Success;
			}
			DefaultMessageSendResult result = new DefaultMessageSendResult(textOnlyMessage);
			result.ResultText = resultText;
			result.SetResult(status);
			return result;
		}
		
		static string lastCommand;
		
		public static MessageSendStatus ExecuteCommand(IrcClient irc, string command, string target, ref string result)
		{
			return ExecuteCommand(irc, command, target, ref result, null);
		}
		
		private static MessageSendStatus ExecuteCommand(IrcClient irc, string command, string target, ref string result, Message message)
		{
			string previousCommand = lastCommand;
			lastCommand = command;
			int pos = command.IndexOf(' ');
			string cmd;
			string para;
			if (pos < 0) {
				cmd = command.Substring(1);
				para = null;
			} else {
				cmd = command.Substring(1, pos - 1);
				para = command.Substring(pos + 1);
			}
			switch (cmd.ToLower()) {
				case "action":
				case "describe":
				case "me":
					if (message != null)
						message.IsAction = true;
					irc.SendMessage(SendType.Action, target, para);
					return MessageSendStatus.Success;
				case "!": // Recalls the last command typed in any window.
					if (previousCommand == null || previousCommand[2] == '!') {
						result = Service.Parse("${res:IRC.NoPreviousCommand}", cmd);
						return MessageSendStatus.CommandFailed;
					} else {
						return ExecuteCommand(irc, previousCommand, target, ref result);
					}
				case "nick":
					irc.RfcNick(para);
					return MessageSendStatus.CommandExecuted;
				case "away":
					if (para == null) {
						irc.RfcAway();
						result = "${res:IRC.AwayTurnedOff}";
					} else {
						irc.RfcAway(para);
						result = "${res:IRC.AwayTurnedOn}";
					}
					return MessageSendStatus.CommandExecuted;
				case "disconnect":
				case "quit":
					if (para == null)
						irc.RfcQuit();
					else
						irc.RfcQuit(para);
					return MessageSendStatus.CommandExecuted;
				case "invite":
					return NicknameChannelCommand(irc, cmd.ToLower(), para, target, ref result);
				default:
					result = Service.Parse("${res:IRC.UnknownCommand}", cmd);
					return MessageSendStatus.CommandFailed;
			}
		}
		
		static MessageSendStatus NicknameChannelCommand(IrcClient irc, string command, string para, string target, ref string result)
		{
			string[] parameters = SplitParameters(para);
			string user = null;
			string channel = null;
			if (parameters.Length == 2) {
				user = parameters[0];
				channel = parameters[1];
			} else if (parameters.Length == 1) {
				if (target.StartsWith("#")) {
					user = parameters[0];
					channel = target;
				}
			}
			if (user != null && user.Length > 0 && !user.StartsWith("#")) {
				if (channel != null && channel.StartsWith("#")) {
					if (NicknameChannelCommandInternal(irc, command, user, channel, ref result))
						return MessageSendStatus.CommandExecuted;
					else
						return MessageSendStatus.CommandFailed;
				}
			}
			if (target.StartsWith("#"))
				result = Service.Parse("${res:IRC.WrongSyntax}", "/" + command + " nickname [#channel]");
			else
				result = Service.Parse("${res:IRC.WrongSyntax}", "/" + command + " nickname #channel");
			return MessageSendStatus.CommandFailed;
		}
		
		static bool NicknameChannelCommandInternal(IrcClient irc, string command, string user, string channel, ref string result)
		{
			switch (command) {
				case "invite":
					return Invite(irc, user, channel, ref result);
				default:
					throw new ArgumentException("Unknown command /" + command);
			}
		}
		
		static bool Invite(IrcClient irc, string user, string channel, ref string result)
		{
			irc.RfcInvite(user, channel);
			result = Service.StringParser.Parse("${res:IRC.InviteOK}", new string[,] {{"Nick", user}, {"Channel", channel}});
			return true;
		}
		
		static string[] SplitParameters(string parameterList)
		{
			ArrayList col = new ArrayList();
			StringBuilder current = new StringBuilder();
			int mode = 0; // 0 = beginning; 1 = normal; 2 = double quotes; 3 = single quotes
			for (int i = 0; i < parameterList.Length; i++) {
				char c = parameterList[i];
				if (mode == 0) {
					if (c == '"')
						mode = 2;
					else if (c == '\'')
						mode = 3;
					else if (!Char.IsWhiteSpace(c)) {
						mode = 1;
						current.Append(c);
					}
				} else if (
				           (mode == 1 && Char.IsWhiteSpace(c)) ||
				           (mode == 2 && c == '"') ||
				           (mode == 3 && c == '\'')
				          ) {
					col.Add(current.ToString());
					current.Length = 0;
					mode = 0;
				} else {
					current.Append(c);
				}
			}
			if (current.Length > 0)
				col.Add(current.ToString());
			return (string[])col.ToArray(typeof(string[]));
		}
	}
}
