// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.Diagnostics;
using System.Collections;
using System.Reflection;
using System.Windows.Forms;

using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.AddIns.Codons;
using ICSharpCode.Core.Properties;
using ICSharpCode.Core.AddIns.Conditions;
using DGNetwork.Services;
using DGNetwork.Gui;
using DGNetwork.Gui.Menu;

namespace DGNetwork.AddIns.Codons
{
	[CodonName("MenuItem")]
	public class MenuItemCodon : AbstractCodon
	{
		[XmlMemberAttribute("label", IsRequired=true)]
		string label       = null;
		
		[XmlMemberAttribute("description")]
		string description = null;
		
		[XmlMemberArrayAttribute("shortcut", SeparatorString = "|")]
		string[] shortcut    = null;
		
		[XmlMemberAttribute("icon")]
		string icon        = null;
		
		[XmlMemberAttribute("link")]
		string link        = null;
		
		public string Link {
			get {
				return link;
			}
			set {
				link = value;
			}
		}
		
		public override bool HandleConditions {
			get {
				return true;
			}
		}
		
		public string Label {
			get {
				return label;
			}
			set {
				label = value;
			}
		}
		
		public string Description {
			get {
				return description;
			}
			set {
				description = value;
			}
		}
		
		public string Icon {
			get {
				return icon;
			}
			set {
				icon = value;
			}
		}
		
		public string[] Shortcut {
			get {
				return shortcut;
			}
			set {
				shortcut = value;
			}
		}
		
		/// <summary>
		/// Creates an item with the specified sub items. And the current
		/// Condition status for this item.
		/// </summary>
		public override object BuildItem(object owner, ArrayList subItems, ConditionCollection conditions)
		{
			CommandBarItem newItem = null;
			if (Label == "-") {
				newItem = new SdMenuSeparator(conditions, owner);
			} else if (Link != null) {
				newItem = new SdMenuCommand(conditions, null, Label,  new LinkCommand(Link));
			} else {
				object o = null;
				if (Class != null) {
					o = AddIn.CreateObject(Class);
				}
				if (o != null) {
					if (o is IMenuCommand) {
						IMenuCommand menuCommand = (IMenuCommand)o;
						menuCommand.Owner = owner;
						menuCommand.MenuItemID = this.ID;
						if (o is ICheckableMenuCommand) {
							newItem = new SdMenuCheckBox(conditions, owner, Label, (ICheckableMenuCommand)menuCommand);
						} else {
							newItem = new SdMenuCommand(conditions, owner, Label, menuCommand);
						}
					} else if (o is ISubmenuBuilder) {
						return o;//((ISubmenuBuilder)o).BuildSubmenu(owner);
					} else {
						Service.Message.ShowMessage("unknown class: " + o.GetType().FullName + "\nexpected MenuCommand or SubmenuBuilder");
					}
				}
			}
			if (newItem == null) {
				SdMenu newMenu = new SdMenu(conditions, owner, Label);
				if (subItems != null && subItems.Count > 0) {
					foreach (object item in subItems) {
						if (item != null) {
							newMenu.SubItems.Add(item);
						}
					}
				}
				newItem = newMenu;
			}
			System.Diagnostics.Debug.Assert(newItem != null);
			
			if (Icon != null) {
				if (Icon != null && Icon.IndexOf("${") >= 0) {
					if (newItem is SdMenuCommand)
						((SdMenuCommand)newItem).DynamicImage = Icon;
					else if (newItem is SdMenu)
						((SdMenu)newItem).DynamicImage = Icon;
					else
						Service.Message.ShowError("Cannot use dynamic icon for " + newItem.GetType().FullName);
				} else {
					try {
						newItem.Image = Service.Resource.GetBitmap(Icon);
					} catch (ResourceNotFoundException ex) {
						Service.Message.ShowError(ex.Message);
					}
				}
			}
			
			if (newItem is SdMenuCommand) {
				((SdMenuCommand)newItem).Description = description;
			}
			
			if (Shortcut != null && newItem is SdMenuCommand) {
				try {
					foreach (string key in this.shortcut) {
						((SdMenuCommand)newItem).Shortcut |= (System.Windows.Forms.Keys)Enum.Parse(typeof(System.Windows.Forms.Keys), key.Trim());
					}
				} catch (Exception) {
					((SdMenuCommand)newItem).Shortcut = System.Windows.Forms.Keys.None;
				}
			}
			newItem.IsEnabled = true; //action != ConditionFailedAction.Disable;
			return newItem;
		}
	}
	
	internal class LinkCommand : AbstractMenuCommand {
		string link;
		public LinkCommand(string link) {
			this.link = link;
		}
		public override void Run() {
			Service.Browser.OpenWindow(link);
		}
	}
}
