// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Grunwald.P2PLib;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;
using DGNetwork.Services;
using DGNetwork.Chat.Parser;

namespace DGNetwork.Chat
{
	/// <summary>
	/// HtmlDisplay is an abstract class that translates the various display methods
	/// into html.
	/// </summary>
	public abstract class HtmlDisplay : AbstractDisplay
	{
		public const string ScrollCode = "<script>a();</script>";
		public const string TimedScrollCode = "<script>setTimeout('a();', 500);</script>";
		
		public abstract void SendHTML(string html);
		
		public override void DisplayTitle(string title)
		{
			SendHTML("<h2>" + Service.Parse(title) + "</h2>" + ScrollCode);
		}
		
		ArrayList joinableRooms = new ArrayList();
		
		public override void DisplayMessage(Message message)
		{
			Service.Chat.DisplayFilter.Filter(message, MessageModes.Html, this);
			StringBuilder b = new StringBuilder("<p class=\"entry\">");
			if (message.Room != null && message.Room != this.Room) {
				if (!joinableRooms.Contains(message.Room))
					joinableRooms.Add(message.Room);
				b.Append("<span class=\"othergroup\" onClick=\"location.href='join:");
				b.Append(message.Room.GetHashCode().ToString("x"));
				b.Append("'\">");
				b.Append(Converter.TextToHtml(Service.Parse(message.Room.Title)));
				b.Append("&gt;</span> ");
			}
			RenderMessageContent(message, b);
			b.Append("</p>" + ScrollCode);
			SendHTML(b.ToString());
		}
		
		static void RenderMessageContent(Message message, StringBuilder b)
		{
			if (message.IsAction)
				b.Append("*** ");
			string name = message.Nickname;
			if (name != null) {
				b.Append("<span class=\"user\">");
				if (name.Length == 0)
					name = "NoName";
				b.Append(Converter.TextToHtml(name));
				if (message.IsAction)
					b.Append("</span>\n");
				else
					b.Append("</span>:\n");
			}
			b.Append(message.Html);
		}
		
		public static string RenderMessage(Message message)
		{
			StringBuilder b = new StringBuilder("<p class=\"entry\">");
			RenderMessageContent(message, b);
			b.Append("</p>");
			return b.ToString();
		}
		
		public static string RenderMessageContent(Message message)
		{
			StringBuilder b = new StringBuilder();
			RenderMessageContent(message, b);
			return b.ToString();
		}
		
		public override void DisplaySendError(string message)
		{
			SendHTML("<p class=\"error\">" + Service.Parse(message) + "</p>" + ScrollCode);
		}
		
		public override void DisplayText(string message)
		{
			SendHTML("<p class=\"entry\">" + Service.Parse(message) + "</p>" + ScrollCode);
		}
		
		protected string GetSystemText(string message)
		{
			return "<p><b>" + Service.Parse(message) + "</b></p>" + ScrollCode;
		}
		
		public void HasNavigatedTo(string url)
		{
			string decoded = Helper.Unescape(url);
			string beg = "javascript:void(docmd('/";
			if (decoded.StartsWith(beg)) {
				decoded = decoded.Substring(beg.Length, decoded.Length - beg.Length - 3);
				decoded = Helper.Unescape(decoded);
				
				Service.Chat.RunCommand(decoded, this.Room, this);
			} else if (decoded.StartsWith("join:")) {
				int code = int.Parse(decoded.Substring(5), System.Globalization.NumberStyles.HexNumber);
				foreach (IChatRoom room in joinableRooms) {
					if (room.GetHashCode() == code) {
						if (!room.IsDisposed)
							this.Room = room;
						break;
					}
				}
			} else if (decoded.StartsWith("mailto:")) {
				Service.Browser.WriteMail(decoded.Substring(7));
			} else {
				Service.Browser.OpenWindow(url);
			}
		}
	}
	
	/// <summary>
	/// BrowserDisplay the abstract base class for chat displays that use a web browser control.
	/// </summary>
	public abstract class BrowserDisplay : HtmlDisplay
	{
		Timer timer;
		
		public BrowserDisplay()
		{
			requestHandler = new RequestHandler();
			requestHandler.Display = this;
			requestHandler.RequestURL = "/system/browserdisplay/stream?" + Helper.Random.Next().ToString();
			Service.P2P.Server.RequestHandlers.Add(requestHandler);
			timer = new Timer();
			timer.Interval = 5000;
			timer.Tick += new EventHandler(TimerTick);
			timer.Enabled = true;
		}
		
		protected override void OnDisposed(EventArgs e)
		{
			if (requestHandler != null) {
				requestHandler.Close();
				Service.P2P.Server.RequestHandlers.Remove(requestHandler);
				requestHandler = null;
			}
			if (timer != null) {
				timer.Dispose();
				timer = null;
			}
			base.OnDisposed(e);
		}
		
		StringBuilder buffer = new StringBuilder();
		
		private void TimerTick(object sender, EventArgs e)
		{
			requestHandler.Send(" ", false);
		}
		
		public override void SendHTML(string html)
		{
			lock (requestHandler) {
				buffer.Append(html.Replace(HtmlDisplay.ScrollCode, "")); // HACK!!!
				requestHandler.Send(html);
			}
		}
		
		#region HTML Header
		static SortedList namespaces; // string -> string
		static ArrayList header;      // string
		
		public static ArrayList Header {
			get {
				if (header == null) {
					header = new ArrayList();
					// meta description
					header.Add("<meta name=\"description\" content=\"DG Network chat session\">");
					header.Add("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">");
					
					// stylesheet
					header.Add("<link rel=\"stylesheet\" type=\"text/css\" href=\"style.css\">");
					
					// default javascript
					header.Add("<script type=\"text/javascript\" src=\"script.js\"></script>");
					
					// MathPlayer for IE
					// TODO: This slows down loading by 20 seconds when internet connection is not available
					//header.Add("<object id=\"mathplayer\" classid=\"clsid:32F66A20-7614-11D4-BD11-00104BD3F987\"></object><?import namespace=\"mml\" implementation=\"#mathplayer\"?>");
					
					// Math javascript
					header.Add("<script type=\"text/javascript\" src=\"ASCIIMathML.js\"></script>");
					header.Add("<script type=\"text/javascript\">AMinitSymbols(); mathcolor='black';</script>");
				}
				return header;
			}
		}
		
		public static SortedList Namespaces {
			get {
				if (namespaces == null) {
					namespaces = new SortedList();
					namespaces.Add("mml", "http://www.w3.org/1998/Math/MathML");
				}
				return namespaces;
			}
		}
		
		public static string GetHeader()
		{
			return GetHeader("DG Network Chat");
		}
		
		public static string GetHeader(string title)
		{
			StringBuilder b = new StringBuilder();
			b.Append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" ");
			b.Append("\"http://www.w3.org/TR/html4/loose.dtd\">\n");
			b.Append("<html");
			foreach (DictionaryEntry e in Namespaces) {
				b.Append(" xmlns:");
				b.Append((string)e.Key);
				b.Append("=");
				b.Append(HtmlHelper.EncodeAttribute((string)e.Value));
			}
			b.Append(">\n");
			b.Append("<head><title>");
			b.Append(title);
			b.Append("</title>\n");
			foreach (string line in Header) {
				b.Append(line);
				b.Append('\n');
			}
			b.Append("</head><body>");
			return b.ToString();
		}
		#endregion
		
		#region Handle http connection
		public string URL {
			get {
				if (requestHandler == null)
					throw new ObjectDisposedException("BrowserDisplay");
				StringBuilder b = new StringBuilder("http://127.0.0.1");
				int port = Service.P2P.Server.Port;
				if (port != 80) {
					b.Append(':');
					b.Append(port);
				}
				b.Append(requestHandler.RequestURL);
				return b.ToString();
			}
		}
		
		RequestHandler requestHandler;
		
		private class RequestHandler : IRequestHandler
		{
			public BrowserDisplay Display;
			public string RequestURL;
			public HttpRequest Request;
			public IConnection Connection;
			
			int IRequestHandler.GetPriority(HttpRequest request)
			{
				if (request.FullRequest == RequestURL)
					return 100;
				else
					return int.MinValue;
			}
			
			void IRequestHandler.Handle(HttpRequest request)
			{
				Console.WriteLine("Accepting new browser connection...");
				lock (this) {
					if (this.Request != null) {
						Send(Display.GetSystemText("${res:Chat.BrowserConnectionReplaced}"));
						Close();
						Display.buffer.Append(Display.GetSystemText("${res:Chat.BrowserConnectionRecreated}"));
					}
					this.Request = request;
					this.Connection = request.Connection;
					HttpResponse response = new HttpResponse();
					response.Server = Service.Parse("${AppIdentification}");
					response.ContentType = "text/html; charset=UTF-8";
					response.SendHeaders(this.Connection.Writer, false);
					Send(BrowserDisplay.GetHeader());
					StringBuilder buffer = Display.buffer;
					if (buffer.Length < 10000) {
						byte[] b = new byte[10000];
						for (int i = 0; i < b.Length; ++i)
							b[i] = 32; // Fill with spaces
						for (int i = 99; i < b.Length; i += 100) {
							b[i - 1] = 13; // CR
							b[i] = 10; // LF
						}
						this.Connection.Stream.Write(b, 0, b.Length);
						this.Connection.Stream.Flush();
					}
					if (buffer.Length > 0) {
						Send(buffer.ToString());
						Send(HtmlDisplay.TimedScrollCode);
					}
				}
			}
			
			public void Close()
			{
				if (this.Connection == null)
					return;
				if (this.Connection.Connected)
					this.Connection.Close();
				this.Connection = null;
				this.Request = null;
			}
			
			public void Send(string html)
			{
				Send(html, true);
			}
			
			public void Send(string html, bool newLine)
			{
				if (this.Connection == null)
					return;
				if (this.Connection.Connected) {
					try {
						if (newLine)
							this.Connection.Writer.WriteLine(html);
						else
							this.Connection.Writer.Write(html);
						this.Connection.Writer.Flush();
					} catch (IOException) {
					} catch (ObjectDisposedException) {
						// send error -> connection closed
						this.Connection = null;
						this.Request = null;
					}
				} else {
					this.Connection = null;
					this.Request = null;
				}
			}
		}
		#endregion
	}
}
