// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;
using DGNetwork.Services;

namespace DGNetwork.Chat
{
	public class ChatForm : Form, IHasRoom
	{
		IChatRoom room;
		ChatPanel panel;
		bool isActiveWindow = true;
		
		public ChatForm(IChatRoom room)
		{
			this.Icon = Service.Resource.GetIcon("Icons.App");
			this.MinimumSize = new Size(200, 100);
			this.Size = new Size(500, 400);
			panel = new ChatPanel();
			panel.Dock = DockStyle.Fill;
			this.Controls.Add(panel);
			this.Activated += new EventHandler(ChatFormActivated);
			this.Deactivate += new EventHandler(ChatFormDeactivate);
			this.Room = room;
			panel.RoomChanged += new EventHandler(PanelRoomChanged);
			
			if (room == null) // RoomChanged didn't fire, so refresh title manually
				RoomTitleChanged(null, null);
			
			this.KeyPreview = true;
		}
		
		protected override void OnKeyDown(KeyEventArgs e)
		{
			if (e.Control)
				e.Handled = panel.ProcessShortcut(e.KeyData);
			base.OnKeyDown(e);
		}
		
		public ChatPanel ChatPanel {
			get {
				return panel;
			}
		}
		
		public IChatRoom Room {
			get {
				return room;
			}
			set {
				if (room == value)
					return;
				SetRoomInternal(value);
				RoomTitleChanged(null, null);
				panel.Room = value;
				OnRoomChanged(EventArgs.Empty);
			}
		}
		
		private void DetachEventHandler()
		{
			if (room != null) {
				room.TitleChanged -= new EventHandler(RoomTitleChanged);
				room.Disposed -= new EventHandler(RoomDisposed);
				if (isActiveWindow)
					room.IsActiveWindow = false;
			}
		}
		
		private void SetRoomInternal(IChatRoom value)
		{
			DetachEventHandler();
			room = value;
			if (room != null) {
				room.TitleChanged += new EventHandler(RoomTitleChanged);
				room.Disposed += new EventHandler(RoomDisposed);
				room.IsActiveWindow = isActiveWindow;
			}
		}
		
		public event EventHandler RoomChanged;
		
		protected virtual void OnRoomChanged(EventArgs e)
		{
			if (RoomChanged != null) {
				RoomChanged(this, e);
			}
		}
		
		private void PanelRoomChanged(object sender, EventArgs e)
		{
			this.Room = panel.Room;
		}
		
		private void RoomTitleChanged(object sender, EventArgs e)
		{
			if (room == null)
				this.Text = "DG Network";
			else
				this.Text = Service.Parse(room.Title) + " - DG Network";
		}
		
		private void RoomDisposed(object sender, EventArgs e)
		{
			Invoke(new MethodInvoker(this.Close));
		}
		
		private void ChatFormActivated(object sender, EventArgs e)
		{
			isActiveWindow = true;
			if (room != null)
				room.IsActiveWindow = true;
		}
		
		private void ChatFormDeactivate(object sender, EventArgs e)
		{
			isActiveWindow = false;
			if (room != null)
				room.IsActiveWindow = false;
		}
		
		protected override void OnClosed(EventArgs e)
		{
			DetachEventHandler();
			base.OnClosed(e);
		}
	}
}
