// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;
using DGNetwork.Gui;
using DGNetwork.Services;

namespace DGNetwork.Chat
{
	public class ChatPanel : UserControl, IHasRoom
	{
		DisplayPanel displayPanel;
		EditBoxPanel editBoxPanel;
		Splitter splitter;
		Panel bottomPanel;
		
		public ChatPanel()
		{
			displayPanel = new DisplayPanel();
			splitter = new Splitter();
			bottomPanel = new Panel();
			editBoxPanel = new EditBoxPanel();
			displayPanel.RoomChanged += new EventHandler(DisplayPanelRoomChanged);
			editBoxPanel.RoomChanged += new EventHandler(EditBoxPanelRoomChanged);
			editBoxPanel.SendFinished += new SendFinishEventHandler(EditBoxPanelSendFinished);
			displayPanel.DisplayChanged += new EventHandler(DisplayPanelDisplayChanged);
			editBoxPanel.EditBoxChanged += new EventHandler(EditBoxPanelEditBoxChanged);
			displayPanel.Dock = DockStyle.Fill;
			this.Controls.Add(displayPanel);
			splitter.Dock = DockStyle.Bottom;
			splitter.Height = 4;
			this.Controls.Add(splitter);
			editBoxPanel.Dock = DockStyle.Fill;
			bottomPanel.Controls.Add(editBoxPanel);
			CreateToolBar();
			bottomPanel.Dock = DockStyle.Bottom;
			this.Controls.Add(bottomPanel);
		}
		
		CommandBar toolBar;
		
		private void CreateToolBar()
		{
			toolBar = ToolBarService.CreateToolBar("/Chat/ToolBar", this);
			toolBar.Dock = DockStyle.Top;
			bottomPanel.Controls.Add(toolBar);
		}
		
		private void RecreateToolBar()
		{
			bottomPanel.Controls.Remove(toolBar);
			toolBar.Dispose();
			CreateToolBar();
		}
		
		public bool ProcessShortcut(Keys key)
		{
			foreach (object o in toolBar.Items) {
				CommandBarButtonBase button = o as CommandBarButtonBase;
				if (button != null) {
					if (button.Shortcut == key) {
						button.PerformClick();
						return true;
					}
				}
			}
			// when control is pressed, but shift wasn't, we try try again with control and shift
			if ((key & Keys.Control) != Keys.Control)
				return false;
			if ((key & Keys.Shift) == Keys.Shift)
				return false;
			key |= Keys.Shift;
			foreach (object o in toolBar.Items) {
				CommandBarButtonBase button = o as CommandBarButtonBase;
				if (button != null) {
					if (button.Shortcut == key) {
						button.PerformClick();
						return true;
					}
				}
			}
			return false;
		}
		
		public DisplayPanel DisplayPanel {
			get {
				return displayPanel;
			}
		}
		
		public EditBoxPanel EditBoxPanel {
			get {
				return editBoxPanel;
			}
		}
		
		private void DisplayPanelDisplayChanged(object sender, EventArgs e)
		{
			RecreateToolBar();
		}
		
		private void EditBoxPanelEditBoxChanged(object sender, EventArgs e)
		{
			RecreateToolBar();
		}
		
		private void EditBoxPanelSendFinished(object sender, SendFinishEventArgs e)
		{
			IMessageSendResult result = e.Result;
			if (result.Status == MessageSendStatus.Success) {
				displayPanel.Display.DisplayMessage(result.Message);
			} else if (result.Status == MessageSendStatus.CommandExecuted) {
				if (result.ResultText != null)
					displayPanel.Display.DisplayText(result.ResultText);
			} else {
				if (result.ResultText == null)
					displayPanel.Display.DisplaySendError("${res:Chat.Error." + result.Status.ToString() + "}");
				else
					displayPanel.Display.DisplaySendError(result.ResultText);
			}
		}
		
		#region ChatRoom
		IChatRoom room;
		
		public IChatRoom Room {
			get {
				return room;
			}
			set {
				if (room == value)
					return;
				room = value;
				displayPanel.Room = room;
				editBoxPanel.Room = room;
				OnRoomChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler RoomChanged;
		
		protected virtual void OnRoomChanged(EventArgs e)
		{
			if (RoomChanged != null) {
				RoomChanged(this, e);
			}
		}
		
		private void DisplayPanelRoomChanged(object sender, EventArgs e)
		{
			this.Room = displayPanel.Room;
		}
		
		private void EditBoxPanelRoomChanged(object sender, EventArgs e)
		{
			this.Room = editBoxPanel.Room;
		}
		#endregion
	}
}
