// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using DGNetwork.Services;

namespace DGNetwork.Chat
{
	public interface IChatRoom : IDisposable
	{
		string Title { get; }
		event EventHandler TitleChanged;
		
		bool CanSend { get; }
		event EventHandler CanSendChanged;
		
		bool UserListVisible { get; }
		event EventHandler UserListVisibleChanged;
		
		bool IsActiveWindow { get; set; }
		event EventHandler IsActiveWindowChanged;
		
		IMessageSendResult SendMessage(Message message);
		
		ArrayList Users { get; }
		
		/// <summary>Is raised when the contents of <see cref="Users"/> have changed.</summary>
		event EventHandler UsersChanged;
		
		void ReceiveMessage(Message message);
		event MessageReceiveEventHandler Receive;
		ArrayList Messages { get; }
		
		/// <summary>Gets if the display should call dispose on the room when the window is closed.</summary>
		bool DisposeOnLeave { get; }
		
		//void Dispose(); // inherited from IDisposable
		bool IsDisposed { get; }
		event EventHandler Disposed;
	}
	
	public abstract class AbstractChatRoom : IChatRoom
	{
		public abstract IMessageSendResult SendMessage(Message message);
		public abstract ArrayList Users { get; }
		
		/// <summary>Is raised when the contents of <see cref="Users"/> have changed.</summary>
		public event EventHandler UsersChanged;
		
		protected virtual void OnUsersChanged(EventArgs e)
		{
			if (UsersChanged != null) {
				UsersChanged(this, e);
			}
		}
		
		#region IsActiveWindow
		bool isActiveWindow;
		public event EventHandler IsActiveWindowChanged;
		
		public bool IsActiveWindow {
			get {
				return isActiveWindow;
			}
			set {
				if (isActiveWindow == value)
					return;
				isActiveWindow = value;
				OnIsActiveWindowChanged(EventArgs.Empty);
			}
		}
		
		protected virtual void OnIsActiveWindowChanged(EventArgs e)
		{
			if (IsActiveWindowChanged != null) {
				IsActiveWindowChanged(this, e);
			}
		}
		#endregion
		
		#region UserListVisible
		bool userListVisible;
		public event EventHandler UserListVisibleChanged;
		
		public bool UserListVisible {
			get {
				return userListVisible;
			}
			set {
				if (userListVisible == value)
					return;
				userListVisible = value;
				OnUserListVisibleChanged(EventArgs.Empty);
			}
		}
		
		protected virtual void OnUserListVisibleChanged(EventArgs e)
		{
			if (UserListVisibleChanged != null) {
				UserListVisibleChanged(this, e);
			}
		}
		#endregion
		
		#region Dispose
		bool isDisposed;
		
		public bool IsDisposed {
			get {
				return isDisposed;
			}
		}
		
		public void Dispose()
		{
			if (isDisposed)
				return;
			isDisposed = true;
			OnDisposed(EventArgs.Empty);
		}
		
		public event EventHandler Disposed;
		
		protected virtual void OnDisposed(EventArgs e)
		{
			if (Disposed != null) {
				Disposed(this, e);
			}
		}
		
		public virtual bool DisposeOnLeave {
			get {
				return false;
			}
		}
		#endregion
		
		#region Title
		string title;
		public event EventHandler TitleChanged;
		
		public string Title {
			get {
				return title;
			}
			set {
				if (title == value)
					return;
				title = value;
				OnTitleChanged(EventArgs.Empty);
			}
		}
		
		protected virtual void OnTitleChanged(EventArgs e)
		{
			if (TitleChanged != null) {
				TitleChanged(this, e);
			}
		}
		#endregion
		
		#region CanSend
		bool canSend = true;
		public event EventHandler CanSendChanged;
		
		public bool CanSend {
			get {
				return canSend;
			}
			set {
				if (canSend == value)
					return;
				canSend = value;
				OnCanSendChanged(EventArgs.Empty);
			}
		}
		
		protected virtual void OnCanSendChanged(EventArgs e)
		{
			if (CanSendChanged != null) {
				CanSendChanged(this, e);
			}
		}
		#endregion
		
		#region Receive Message
		ArrayList messages = new ArrayList();
		
		public ArrayList Messages {
			get {
				return messages;
			}
		}
		
		public event MessageReceiveEventHandler Receive;
		
		protected virtual void OnReceive(MessageReceiveEventArgs e)
		{
			if (Receive != null) {
				Receive(this, e);
			}
		}
		
		public void ReceiveMessage(Message message)
		{
			message.Room = this;
			try {
				Service.Chat.ReceiveFilter.Filter(message, MessageModes.All, this);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
			MessageReceiveEventArgs e = new MessageReceiveEventArgs(message, this);
			OnReceive(e);
			if (message.NotifyUser)
				DGNetwork.Services.Service.Chat.NotifyUser(e);
			if (e.AddToCache) {
				messages.Add(message);
			}
		}
		#endregion
	}
	
	#region MessageReceiveEventHandler / MessageReceiveEventArgs
	public delegate void MessageReceiveEventHandler(object sender, MessageReceiveEventArgs e);
	
	public class MessageReceiveEventArgs : EventArgs
	{
		Message message;
		IChatRoom room;
		
		public Message Message {
			get {
				return message;
			}
		}
		
		public IChatRoom Room {
			get {
				return room;
			}
		}
		
		bool addToCache = true;
		
		public bool AddToCache {
			get {
				return addToCache;
			}
			set {
				addToCache = value;
			}
		}
		
		public MessageReceiveEventArgs(Message message, IChatRoom room)
		{
			if (message == null)
				throw new ArgumentNullException("message");
			if (room == null)
				throw new ArgumentNullException("room");
			this.message = message;
			this.room = room;
		}
	}
	#endregion
}
