// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.Services;
using DGNetwork.Services;
using DGNetwork.Gui;

namespace DGNetwork.Chat
{
	/// <summary>
	/// Service class providing chat functionality.
	/// </summary>
	public class ChatService : AbstractService
	{
		#region InitializeService
		public ChatService()
		{
			this.Status = "Offline";
		}
		
		public override void InitializeService()
		{
			LoadDisplays();
			LoadFilters();
			ICSharpCode.Core.Properties.PropertyLink link;
			link = Service.Property.CreateLink("ChatService.Nick", this, "Nickname");
			link.UseEvent = true;
			Service.Timer.Tick += new EventHandler(AutoAwayTick);
			base.InitializeService();
		}
		#endregion
		
		#region Display and EditBox Management
		ArrayList displays;
		IDisplayFactory activeDisplay;
		
		public ArrayList Displays {
			get {
				return displays;
			}
		}
		
		public IDisplayFactory ActiveDisplay {
			get {
				return activeDisplay;
			}
			set {
				if (activeDisplay == value)
					return;
				activeDisplay = value;
				Service.Property.SetProperty("ChatService.ActiveDisplay", activeDisplay.Name);
				if (ActiveDisplayChanged != null)
					ActiveDisplayChanged(this, EventArgs.Empty);
			}
		}
		
		public event EventHandler ActiveDisplayChanged;
		
		ArrayList editBoxes;
		IEditBoxFactory activeEditBox;
		
		public IEditBoxFactory ActiveEditBox {
			get {
				return activeEditBox;
			}
			set {
				if (activeEditBox == value)
					return;
				activeEditBox = value;
				Service.Property.SetProperty("ChatService.ActiveEditBox", activeEditBox.Name);
				if (ActiveEditBoxChanged != null)
					ActiveEditBoxChanged(this, EventArgs.Empty);
			}
		}
		
		public event EventHandler ActiveEditBoxChanged;
		
		public ArrayList EditBoxes {
			get {
				return editBoxes;
			}
		}
		
		private void LoadDisplays()
		{
			string name;
			displays = Helper.BuildItems("/Chat/Display", this);
			if (displays.Count == 0)
				displays.Add(new DefaultDisplayFactory());
			activeDisplay = (IDisplayFactory)displays[0];
			name = Service.Property.GetProperty("ChatService.ActiveDisplay", activeDisplay.Name);
			foreach (IDisplayFactory display in displays) {
				if (display.Name == name) {
					activeDisplay = display;
					break;
				}
			}
			editBoxes = Helper.BuildItems("/Chat/EditBox", this);
			if (editBoxes.Count == 0)
				editBoxes.Add(new DefaultEditBoxFactory("default", "default", typeof(PlainTextEditBox)));
			activeEditBox = (IEditBoxFactory)editBoxes[0];
			name = Service.Property.GetProperty("ChatService.ActiveEditBox", activeEditBox.Name);
			foreach (IEditBoxFactory editBox in editBoxes) {
				if (editBox.Name == name) {
					activeEditBox = editBox;
					break;
				}
			}
		}
		#endregion
		
		#region Username
		string nickname = "";
		
		public string Nickname {
			get {
				return nickname;
			}
			set {
				if (value == null)
					value = "";
				else
					value = value.Trim();
				if (nickname == value)
					return;
				nickname = value;
				if (NicknameChanged != null)
					NicknameChanged(this, EventArgs.Empty);
			}
		}
		
		public event EventHandler NicknameChanged;
		#endregion
		
		#region Message Filters
		MessageFilterChain displayFilter, receiveFilter, sendFilter;
		
		public MessageFilterChain DisplayFilter {
			get {
				return displayFilter;
			}
		}
		
		public MessageFilterChain ReceiveFilter {
			get {
				return receiveFilter;
			}
		}
		
		public MessageFilterChain SendFilter {
			get {
				return sendFilter;
			}
		}
		
		private void LoadFilters()
		{
			displayFilter = (MessageFilterChain)Helper.BuildItem("/Chat/Filter/Display", this);
			receiveFilter = (MessageFilterChain)Helper.BuildItem("/Chat/Filter/Receive", this);
			sendFilter    = (MessageFilterChain)Helper.BuildItem("/Chat/Filter/Send", this);
		}
		#endregion
		
		#region Commands
		ArrayList commands;
		
		public bool RunCommand(string command, IChatRoom chatRoom, IDisplay display)
		{
			int pos = command.IndexOf(' ');
			if (pos < 0)
				return RunCommand(command, "", chatRoom, display);
			else
				return RunCommand(command.Substring(0, pos), command.Substring(pos + 1).Trim(), chatRoom, display);
		}
		
		public bool RunCommand(string command, string parameters, IChatRoom chatRoom, IDisplay display)
		{
			if (commands == null)
				commands = Helper.BuildItems("/Chat/Commands", this);
			string lcommand = command.ToLower();
			lock (commands) {
				foreach (IChatCommand cmd in commands) {
					if (cmd.Name.ToLower() == lcommand) {
						StringWriter writer = new StringWriter();
						bool success = cmd.Run(parameters, chatRoom, writer);
						string result = writer.ToString();
						if (result.Length > 0 && display != null) {
							if (success)
								display.DisplayText(result);
							else
								display.DisplaySendError(result);
						}
						return success;
					}
				}
			}
			if (display != null)
				display.DisplaySendError(Service.Parse("${res:Chat.UnknownCommand}", command));
			return false;
		}
		#endregion
		
		public bool UseSounds {
			get {
				return (bool)Service.Property.GetProperty("ChatService.UseSounds", true);
			}
		}
		
		public bool SendOnEnter {
			get {
				return (bool)Service.Property.GetProperty("ChatService.SendOnEnter", false);
			}
		}
		
		#region Manage ChatForm
		ArrayList chatForms = new ArrayList();
		
		private delegate void OpenChatDelegate(IChatRoom room, bool forceNewWindow);
		
		public void OpenChat(IChatRoom room)
		{
			OpenChat(room, false);
		}
		
		public void OpenChat(IChatRoom room, bool forceNewWindow)
		{
			if (Service.Window.InvokeRequired)
				Service.Window.BeginInvoke(new OpenChatDelegate(OpenChatInvoked), room, forceNewWindow);
			else
				OpenChatInvoked(room, forceNewWindow);
		}
		
		public bool IsChatFormOpen(IChatRoom room)
		{
			lock (chatForms) {
				return FindChatForm(room) != null;
			}
		}
		
		private ChatForm FindChatForm(IChatRoom room)
		{
			foreach (ChatForm form in chatForms) {
				if (form.Room == room)
					return form;
			}
			return null;
		}
		
		private void OpenChatInvoked(IChatRoom room, bool forceNewWindow)
		{
			if (mustVisitRooms.Contains(room)) {
				mustVisitRooms.Remove(room);
				for (int i = 0; i < missedMessages.Count; ++i) {
					if (((Message)missedMessages[i]).Room == room)
						missedMessages.RemoveAt(i--);
				}
			}
			OpenChatInternal(room, forceNewWindow);
		}
		
		private void OpenChatInternal(IChatRoom room, bool forceNewWindow)
		{
			ChatForm frm;
			lock (chatForms) {
				frm = FindChatForm(room);
				if (frm != null) {
					frm.Activate();
					return;
				}
				if (!forceNewWindow) {
					frm = Service.Window.MainWindow as ChatForm;
					if (frm != null) {
						frm.Room = room;
						frm.Activate();
						return;
					}
				}
				frm = new ChatForm(room);
				if (chatForms.Count == 0) {
					Service.Window.ManageWindowSize("ChatWindow", frm);
				} else {
					Service.Window.RestoreWindowSize("ChatWindow", frm);
					if (frm.WindowState == FormWindowState.Normal) {
						Rectangle bounds = frm.Bounds;
						bounds.X += Helper.Random.Next(-75, 75);
						bounds.Y += Helper.Random.Next(-75, 75);
						frm.Bounds = Service.Window.MoveToVisibleScreenArea(bounds);
					}
				}
				RegisterChatFormInternal(frm);
			}
			frm.Show();
			frm.Activate();
		}
		
		/// <summary>
		/// Registers the specified chat form. Use this method to make DG Network able
		/// to find your self-created ChatForms when OpenChat or NotifyChat is called.
		/// </summary>
		public void RegisterChatForm(ChatForm form)
		{
			lock (chatForms) {
				RegisterChatFormInternal(form);
			}
		}
		
		void RegisterChatFormInternal(ChatForm form)
		{
			chatForms.Add(form);
			form.Closed += new EventHandler(ChatFormClosed);
		}
		
		private void ChatFormClosed(object sender, EventArgs e)
		{
			chatForms.Remove((ChatForm)sender);
		}
		
		/// <summary>
		/// Notifies the user that a chat room has received a message.
		/// </summary>
		public void NotifyUser(MessageReceiveEventArgs e)
		{
			if (e == null)
				throw new ArgumentNullException("e");
			// do not use BeginInvoke because that might display the wrong number of messages.
			// and because NotifyUser could change e.AddToCache
			if (Service.Window.InvokeRequired)
				Service.Window.Invoke(new NotifyUserDelegate(NotifyUserInternal), e);
			else
				NotifyUserInternal(e);
		}
		
		private delegate void NotifyUserDelegate(MessageReceiveEventArgs e);
		
		private void NotifyUserInternal(MessageReceiveEventArgs e)
		{
			IChatRoom room = e.Room;
			Message message = e.Message;
			ChatForm frm = FindChatForm(room);
			if (frm == null) {
				frm = Service.Window.MainWindow as ChatForm;
				if (frm != null && frm.Room != room && message.Room == room && e.AddToCache) {
					e.AddToCache = false;
					frm.ChatPanel.DisplayPanel.Display.DisplayMessage(e.Message);
				}
			}
			if (frm != null) {
				if (!room.IsActiveWindow) {
					if (UseSounds)
						Service.Message.PlaySound("sound/MsgIn.wav");
					Service.Window.Flash(frm);
				}
			} else {
				if (!mustVisitRooms.Contains(room))
					mustVisitRooms.Add(room);
				missedMessages.Add(e.Message);
				if (this.Status == "DND")
					return;
				Service.StringParser.Properties["Username"] = message.Nickname;
				string text = message.Text;
				if (text.Length > 55)
					text = text.Substring(0, 50).Trim() + "...";
				if (message.IsAction)
					text = message.Nickname + " " + text;
				Service.StringParser.Properties["MessageText"] = text;
				int count = room.Messages.Count;
				if (count == 0) {
					if (UseSounds)
						Service.Message.PlaySound("sound/MsgIn.wav");
					Service.TNA.MainIcon.ShowBalloon(BalloonIcon.Info, "${res:Chat.GotMessageBalloonText}", "${res:Chat.GotMessageBalloonTitle}", 15000,
					                                 new EventHandler(BalloonClick), new EventHandler(BalloonTimeout));
				} else {
					Service.StringParser.Properties["MessageCount"] = (count + 1).ToString();
					Service.TNA.MainIcon.ShowBalloon(BalloonIcon.Info, "${res:Chat.GotMessagesBalloonText}", "${res:Chat.GotMessagesBalloonTitle}", 15000,
					                                 new EventHandler(BalloonClick), new EventHandler(BalloonTimeout));
				}
			}
		}
		
		ArrayList mustVisitRooms = new ArrayList();
		ArrayList missedMessages = new ArrayList();
		
		private void BalloonClick(object sender, EventArgs e)
		{
			if (!Service.Window.ActiveMainWindowBuilder.UseAsChatWindow)
				Service.Window.ShowMainWindow();
			ChatForm mainWindow = Service.Window.MainWindow as ChatForm;
			if (mainWindow == null) {
				foreach (IChatRoom room in mustVisitRooms) {
					if (!room.IsDisposed)
						OpenChatInternal(room, false);
				}
			} else {
				// When all missed messages are from one room, open that room
				// otherwise open no room but display all messages
				IChatRoom room = null;
				foreach (Message message in missedMessages) {
					if (room == null) {
						room = message.Room;
					} else {
						if (room != message.Room) {
							room = null;
							break;
						}
					}
				}
				if (room == null) {
					foreach (Message message in missedMessages) {
						mainWindow.ChatPanel.DisplayPanel.Display.DisplayMessage(message);
					}
				} else {
					mainWindow.Room = room;
				}
			}
			missedMessages.Clear();
			mustVisitRooms.Clear();
		}
		
		/// <summary>
		/// Opens rooms that have received a notification.
		/// </summary>
		/// <returns>Returns if there were rooms to open.</returns>
		public bool OpenNotifiedRooms()
		{
			if (mustVisitRooms.Count == 0)
				return false;
			if (Service.Window.InvokeRequired)
				Service.Window.BeginInvoke(new EventHandler(BalloonClick), null, null);
			else
				BalloonClick(null, null);
			return true;
		}
		
		private void BalloonTimeout(object sender, EventArgs e)
		{
			// We can ignore this, outstanding rooms are opened with the next double click.
		}
		#endregion
		
		#region Own user status
		/// <summary>
		/// Gets if the user is away or out.
		/// </summary>
		public bool IsAway {
			get {
				return status == "Away" || status == "Out";
			}
		}
		
		string status;
		
		public string Status {
			get {
				return status;
			}
			set {
				if (status == value)
					return;
				autoAwayActive = false;
				string oldStatus = status;
				status = value;
				Service.StringParser.Properties["Status"] = value;
				OnStatusChanged(new StatusChangedEventArgs(oldStatus, value));
			}
		}
		
		public event StatusChangedEventHandler StatusChanged;
		
		protected virtual void OnStatusChanged(StatusChangedEventArgs e)
		{
			if (StatusChanged != null) {
				StatusChanged(this, e);
			}
		}
		#endregion
		
		#region Auto-Away
		DateTime lastMouseMove;
		Point lastMousePosition;
		bool autoAwayActive;
		
		private void AutoAwayTick(object sender, EventArgs e)
		{
			DateTime now = DateTime.Now;
			if (now < lastMouseMove)
				lastMouseMove = now;
			Point mouse = Control.MousePosition;
			if (Math.Abs(mouse.X - lastMousePosition.X) + Math.Abs(mouse.Y - lastMousePosition.Y) >= 3) {
				// Mouse has been moved
				lastMouseMove = now;
				if (autoAwayActive)
					this.Status = "Online";
			} else {
				bool autoAwayPossible = (bool)Service.Property.GetProperty("ChatService.AutoAway", true);
				if (autoAwayPossible) {
					if (lastMouseMove.AddMinutes(5) < now) {
						if (lastMouseMove.AddMinutes(15) < now)
							this.Status = "Out";
						else
							this.Status = "Away";
						autoAwayActive = true;
					}
				}
			}
			lastMousePosition = mouse;
		}
		#endregion
	}
	
	public delegate void StatusChangedEventHandler(object sender, StatusChangedEventArgs e);
	
	public class StatusChangedEventArgs : EventArgs
	{
		string oldStatus, newStatus;
		
		public string NewStatus {
			get {
				return newStatus;
			}
		}
		
		public string OldStatus {
			get {
				return oldStatus;
			}
		}
		
		public StatusChangedEventArgs(string oldStatus, string newStatus)
		{
			this.newStatus = newStatus;
			this.oldStatus = oldStatus;
		}
	}
}
