// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;
using DGNetwork.Services;

namespace DGNetwork.Chat
{
	public interface IDisplay : IDisposable, IHasRoom
	{
		Control Control { get; }
		
		//IChatRoom Room { get; set; }    // inherited from IHasRoom
		//event EventHandler RoomChanged; // inherited from IHasRoom
		
		void DisplayTitle(string title);
		void DisplayMessage(Message message);
		void DisplaySendError(string message);
		void DisplayText(string message);
		
		//void Dispose(); // inherited from IDisposable
		bool IsDisposed { get; }
		event EventHandler Disposed;
	}
	
	public abstract class AbstractDisplay : IDisplay
	{
		public abstract Control Control { get; }
		public abstract void DisplayTitle(string title);
		public abstract void DisplayMessage(Message message);
		public abstract void DisplaySendError(string message);
		public abstract void DisplayText(string message);
		
		#region Dispose
		bool isDisposed;
		
		public bool IsDisposed {
			get {
				return isDisposed;
			}
		}
		
		public void Dispose()
		{
			if (isDisposed)
				return;
			isDisposed = true;
			LeaveRoom();
			OnDisposed(EventArgs.Empty);
		}
		
		public event EventHandler Disposed;
		
		protected virtual void OnDisposed(EventArgs e)
		{
			if (Disposed != null) {
				Disposed(this, e);
			}
		}
		#endregion
		
		IChatRoom room;
		
		public IChatRoom Room {
			get {
				return room;
			}
			set {
				if (room == value)
					return;
				LeaveRoom();
				room = value;
				EnterRoom();
				OnRoomChanged(EventArgs.Empty);
			}
		}
		
		private void EnterRoom()
		{
			if (room == null)
				return;
			this.DisplayTitle(room.Title);
			foreach (Message message in room.Messages) {
				this.DisplayMessage(message);
			}
			room.Messages.Clear();
			room.Receive += new MessageReceiveEventHandler(RoomReceiveMessage);
		}
		
		private void LeaveRoom()
		{
			if (room == null)
				return;
			room.Receive -= new MessageReceiveEventHandler(RoomReceiveMessage);
		}
		
		private void RoomReceiveMessage(object sender, MessageReceiveEventArgs e)
		{
			e.AddToCache = false;
			this.DisplayMessage(e.Message);
		}
		
		public event EventHandler RoomChanged;
		
		protected virtual void OnRoomChanged(EventArgs e)
		{
			if (RoomChanged != null) {
				RoomChanged(this, e);
			}
		}
	}
	
	public class DummyDisplay : AbstractDisplay
	{
		public override void DisplayTitle(string title) { }
		
		public override void DisplayMessage(Message message)
		{
			if (!message.Sending)
				Service.Message.ShowMessageFormatted("${res:Chat.DummyDisplayMessage}", new string[] { message.Text });
		}
		
		public override void DisplaySendError(string message)
		{
			Service.Message.ShowMessage(message);
		}
		
		public override void DisplayText(string message)
		{
			Service.Message.ShowMessage(message);
		}
		
		public DummyDisplay()
		{
			Label lbl = new Label();
			lbl.Text = Service.Parse("${res:Chat.DummyDisplayText}");
			lbl.TextAlign = ContentAlignment.MiddleCenter;
			control = lbl;
		}
		
		Control control;
		
		public override Control Control {
			get {
				return control;
			}
		}
	}
}
