// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using DGNetwork.Services;

namespace DGNetwork.Chat
{
	public class EditBoxPanel : UserControl, IHasRoom
	{
		IEditBox editBox;
		Panel buttonPanel;
		Button sendButton;
		
		#region Constructor
		public EditBoxPanel()
		{
			buttonPanel = new Panel();
			buttonPanel.Dock = DockStyle.Right;
			buttonPanel.Size = new Size(68, 24);
			sendButton = new Button();
			sendButton.Anchor = AnchorStyles.None;
			sendButton.Size = new Size(64, 24);
			sendButton.Location = new Point(2, 0);
			sendButton.Text = Service.Parse("${res:Chat.SendButton}");
			sendButton.FlatStyle = FlatStyle.System;
			sendButton.Enabled = false;
			sendButton.Click += new EventHandler(SendButtonClick);
			buttonPanel.Controls.Add(sendButton);
			this.Controls.Add(buttonPanel);
			this.Height = 60;
			editBox = CreateEditBox();
			Service.Chat.ActiveEditBoxChanged += new EventHandler(ActiveEditBoxChanged);
		}
		#endregion
		
		#region Dispose
		protected override void Dispose(bool disposing)
		{
			if (disposing) {
				Service.Chat.ActiveEditBoxChanged -= new EventHandler(ActiveEditBoxChanged);
				RemoveRoomHandler();
			}
			base.Dispose(disposing);
		}
		#endregion
		
		#region SendButton
		private void SendButtonClick(object sender, EventArgs e)
		{
			if (!sendButton.Enabled)
				return;
			IMessageSendResult result = editBox.Send();
			if (result != null)
				result.CallOnFinish(new EventHandler(SendResultFinish));
			editBox.Focus();
		}
		
		public event SendFinishEventHandler SendFinished;
		
		private void SendResultFinish(object sender, EventArgs e)
		{
			if (Service.Chat.UseSounds)
				Service.Message.PlaySound("sound/MsgSent.wav");
			OnSendFinished(new SendFinishEventArgs((IMessageSendResult)sender));
		}
		
		protected virtual void OnSendFinished(SendFinishEventArgs e)
		{
			if (SendFinished != null) {
				SendFinished(this, e);
			}
		}
		
		private void CanSendChanged(object sender, EventArgs e)
		{
			if (room == null)
				sendButton.Enabled = false;
			else
				sendButton.Enabled = room.CanSend && editBox.CanSend;
		}
		#endregion
		
		#region SizeChange
		bool fullButton = false;
		
		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged(e);
			int height = buttonPanel.Height;
			if (height > 90) {
				if (fullButton)
					return;
				else
					fullButton = true;
				sendButton.Anchor = AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right | AnchorStyles.Bottom;
				buttonPanel.Width = 24;
				sendButton.Size = new Size(24, height - 8);
				sendButton.Location = new Point(0, 4);
				string text = Service.Parse("${res:Chat.SendButton}");
				StringBuilder b = new StringBuilder();
				bool ok = false;
				for (int i = 0; i < text.Length; ++i) {
					char c = text[i];
					if (c != '&') {
						if (ok)
							b.Append("\r\n");
						else
							ok = true;
					}
					b.Append(c);
				}
				sendButton.Text = b.ToString();
			} else {
				if (!fullButton)
					return;
				else
					fullButton = false;
				sendButton.Anchor = AnchorStyles.None;
				buttonPanel.Width = 68;
				sendButton.Size = new Size(64, 24);
				sendButton.Location = new Point(2, (height - 24) / 2);
				sendButton.Text = Service.Parse("${res:Chat.SendButton}");
			}
		}
		#endregion
		
		#region EditBox
		public IEditBox EditBox {
			get {
				return editBox;
			}
		}
		
		private IEditBox CreateEditBox()
		{
			IEditBox box = Service.Chat.ActiveEditBox.Create();
			box.Room = this.Room;
			box.CanSendChanged += new EventHandler(CanSendChanged);
			box.RoomChanged += new EventHandler(EditBoxRoomChanged);
			box.SendTriggered += new EventHandler(SendButtonClick);
			Control ctl = box.Control;
			ctl.Dock = DockStyle.Fill;
			this.Controls.Add(ctl);
			this.Controls.SetChildIndex(ctl, 0);
			return box;
		}
		
		public event EventHandler EditBoxChanged;
		
		private void ActiveEditBoxChanged(object sender, EventArgs e)
		{
			IChatRoom oldRoom = this.Room;
			Control ctl = editBox.Control;
			this.Controls.Remove(ctl);
			ctl.Dispose();
			editBox = CreateEditBox();
			if (EditBoxChanged != null)
				EditBoxChanged(this, e);
		}
		
		private void ControlKeyDown(object sender, KeyEventArgs e)
		{
			
		}
		#endregion
		
		#region Room Property
		IChatRoom room;
		
		public IChatRoom Room {
			get {
				return room;
			}
			set {
				if (room == value)
					return;
				RemoveRoomHandler();
				room = value;
				if (room == null) {
					sendButton.Enabled = false;
				} else {
					CanSendChanged(null, null);
					room.CanSendChanged += new EventHandler(CanSendChanged);
				}
				OnRoomChanged(EventArgs.Empty);
				editBox.Room = room;
			}
		}
		
		private void EditBoxRoomChanged(object sender, EventArgs e)
		{
			this.Room = editBox.Room;
		}
		
		private void RemoveRoomHandler()
		{
			if (room != null)
				room.CanSendChanged -= new EventHandler(CanSendChanged);
		}
		
		public event EventHandler RoomChanged;
		
		protected virtual void OnRoomChanged(EventArgs e)
		{
			if (RoomChanged != null) {
				RoomChanged(this, e);
			}
		}
		#endregion
	}
	
	#region SendFinishEventHandler / SendFinishEventArgs
	public delegate void SendFinishEventHandler(object sender, SendFinishEventArgs e);
	
	public class SendFinishEventArgs : EventArgs
	{
		IMessageSendResult result;
		
		public IMessageSendResult Result {
			get {
				return result;
			}
		}
		
		public SendFinishEventArgs(IMessageSendResult result)
		{
			this.result = result;
		}
	}
	#endregion
}
