// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Globalization;
using System.Text;
using System.Xml;
using ICSharpCode.Core.Properties;
using DGNetwork.Services;

namespace DGNetwork.Chat
{
	public class Message : IXmlConvertable, ICloneable
	{
		#region Text, Html, Rtf and automatic conversions
		string text, html, rtf;
		
		public string Html {
			get {
				if (html == null) {
					if (rtf != null)
						html = Converter.RtfToHtml(rtf);
					else if (text != null)
						html = Converter.TextToHtml(text);
				}
				return html;
			}
			set {
				html = value;
			}
		}
		
		public string Rtf {
			get {
				if (rtf == null) {
					if (html != null)
						rtf = Converter.HtmlToRtf(html);
					else if (text != null)
						rtf = Converter.TextToRtf(text);
				}
				return rtf;
			}
			set {
				rtf = value;
			}
		}
		
		public string Text {
			get {
				if (text == null) {
					if (html != null)
						text = Converter.HtmlToText(html);
					else if (text != null)
						text = Converter.RtfToText(rtf);
				}
				return text;
			}
			set {
				text = value;
			}
		}
		#endregion
		
		public Message() : this(false)
		{
		}
		
		public Message(bool sending)
		{
			this.sending = sending;
			if (sending)
				this.nickname = Service.Chat.Nickname;
		}
		
		DateTime time = DateTime.Now;
		
		/// <summary>
		/// Gets/Sets the time when the message was created.
		/// </summary>
		public DateTime Time {
			get {
				return time;
			}
			set {
				time = value;
			}
		}
		
		IChatRoom room;
		
		/// <summary>
		/// Gets/Sets the room this message is received from / sent to.
		/// </summary>
		public IChatRoom Room {
			get {
				return room;
			}
			set {
				room = value;
			}
		}
		
		bool notifyUser = true;
		
		/// <summary>
		/// Gets/Sets if the user should be notified when this message is received.
		/// </summary>
		public bool NotifyUser {
			get {
				return notifyUser;
			}
			set {
				notifyUser = value;
			}
		}
		
		bool sending;
		
		/// <summary>
		/// Gets/Sets if this object is used to send a message (true).
		/// Otherwise, the object is used to receive a message.
		/// </summary>
		public bool Sending {
			get {
				return sending;
			}
			set {
				sending = value;
			}
		}
		
		bool isAction;
		
		/// <summary>
		/// Gets/Sets if the message is an action.
		/// </summary>
		public bool IsAction {
			get {
				return isAction;
			}
			set {
				isAction = value;
			}
		}
		
		string nickname;
		
		/// <summary>
		/// Gets/Sets the nickname of the sender of the message.
		/// </summary>
		public string Nickname {
			get {
				return nickname;
			}
			set {
				nickname = value;
			}
		}
		
		Hashtable extraData = new Hashtable(); // string -> IXmlConvertable
		
		public Hashtable ExtraData {
			get {
				return extraData;
			}
		}
		
		#region XmlConvertable
		public object FromXmlElement(XmlElement element)
		{
			if (element.Name != "Message")
				throw new BaseException("The xml element is not a message.");
			rtf = GetElement(element, "Rtf");
			html = GetElement(element, "Html");
			text = GetElement(element, "Text");
			nickname = GetElement(element, "Nick");
			sending = GetBooleanElement(element, "Sending");
			isAction = GetBooleanElement(element, "IsAction");
			notifyUser = GetBooleanElement(element, "NotifyUser");
			string timeString = GetElement(element, "Time");
			if (timeString != null && timeString.Length > 0)
				time = DateTime.Parse(timeString, CultureInfo.InvariantCulture);
			this.extraData.Clear();
			XmlElement extraData = element["Data"];
			if (extraData != null) {
				foreach (XmlElement child in extraData) {
					string key = child.GetAttribute("key");
					string typeName = child.GetAttribute("type");
					if (typeName == null || typeName.Length == 0) {
						ExtraData.Add(key, null);
						continue;
					}
					try {
						Type type = Helper.GetType(typeName);
						if (type != null) {
							IXmlConvertable obj = Activator.CreateInstance(type) as IXmlConvertable;
							if (obj != null) {
								ExtraData.Add(key, obj.FromXmlElement((XmlElement)child.FirstChild));
							}
						}
					} catch (Exception ex) {
						Service.Message.ShowError(ex);
					}
				}
			}
			return this;
		}
		
		private bool GetBooleanElement(XmlElement parent, string title)
		{
			string v = GetElement(parent, title);
			if (v == null)
				return false;
			else
				return XmlConvert.ToBoolean(v);
		}
		
		private string GetElement(XmlElement parent, string title)
		{
			XmlElement e = parent[title];
			if (e == null)
				return null;
			else
				return e.InnerText;
		}
		
		public XmlElement ToXmlElement(XmlDocument doc)
		{
			XmlElement msg = doc.CreateElement("Message");
			AddElement(doc, msg, "Rtf", rtf);
			AddElement(doc, msg, "Html", html);
			AddElement(doc, msg, "Text", text);
			AddElement(doc, msg, "Nick", nickname);
			AddElement(doc, msg, "Sending", sending);
			AddElement(doc, msg, "IsAction", isAction);
			AddElement(doc, msg, "NotifyUser", notifyUser);
			AddElement(doc, msg, "Time", time.ToString(CultureInfo.InvariantCulture));
			if (ExtraData.Count > 0) {
				XmlElement extraData = doc.CreateElement("Data");
				foreach (DictionaryEntry e in ExtraData) {
					XmlElement entry = doc.CreateElement("Entry");
					entry.SetAttribute("key", (string)e.Key);
					IXmlConvertable val = e.Value as IXmlConvertable;
					if (val != null) {
						entry.SetAttribute("type", val.GetType().FullName);
						entry.AppendChild(val.ToXmlElement(doc));
					}
					extraData.AppendChild(entry);
				}
				msg.AppendChild(extraData);
			}
			return msg;
		}
		
		private void AddElement(XmlDocument doc, XmlElement element, string title, bool val)
		{
			AddElement(doc, element, title, XmlConvert.ToString(val));
		}
		
		private void AddElement(XmlDocument doc, XmlElement element, string title, string val)
		{
			if (val == null)
				return;
			XmlElement e = doc.CreateElement(title);
			e.InnerText = val;
			element.AppendChild(e);
		}
		#endregion
		
		#region ICloneable
		object ICloneable.Clone()
		{
			return this.Clone();
		}
		
		public virtual Message Clone()
		{
			Message copy = (Message)this.MemberwiseClone();
			copy.extraData = new Hashtable();
			foreach (DictionaryEntry entry in extraData) {
				copy.extraData.Add(TryClone(entry.Key), TryClone(entry.Value));
			}
			return copy;
		}
		
		object TryClone(object o)
		{
			ICloneable c = o as ICloneable;
			if (c != null)
				return c.Clone();
			else
				return o;
		}
		#endregion
		
		public override string ToString()
		{
			StringBuilder b = new StringBuilder();
			if (this.IsAction)
				b.Append("*** ");
			b.Append(this.Nickname);
			if (!this.IsAction)
				b.Append(':');
			b.Append(' ');
			b.Append(this.Text);
			return b.ToString();
		}
	}
}
