// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.Text;

namespace DGNetwork.Chat.Parser
{
	public class HtmlFilter : IMessageFilter
	{
		Hashtable tags;
		
		public HtmlFilter()
		{
			tags = CollectionsUtil.CreateCaseInsensitiveHashtable();
		}
		
		public Hashtable Tags {
			get {
				return tags;
			}
		}
		
		ArrayList replacements = new ArrayList();
		
		public ArrayList Replacements {
			get {
				return replacements;
			}
		}
		
		/// <summary>
		/// Filters the message. The method can change the message's properties.
		/// </summary>
		/// <param name="message">
		/// The message being filtered.
		/// </param>
		/// <param name="mode">
		/// The mode parameter determines the message mode(s) that will be used further.
		/// This filter is only active when the mode contains
		/// </param>
		public void Filter(Message message, MessageModes mode, object owner)
		{
			if ((mode & MessageModes.Html) == 0)
				return;
			this.message = message;
			this.owner = owner;
			message.Html = Process(message.Html);
		}
		
		bool extendedSyntax;
		
		/// <summary>
		/// Gets/Sets if extended syntax ("chat code") is allowed (true) or the filter
		/// is restricture to pure html (false).
		/// </summary>
		public bool ExtendedSyntax {
			get {
				return extendedSyntax;
			}
			set {
				extendedSyntax = value;
			}
		}
		
		Message message;
		
		public Message Message {
			get {
				return message;
			}
		}
		
		object owner;
		
		public object Owner {
			get {
				return owner;
			}
		}
		
		public string Process(string input)
		{
			StringBuilder b = new StringBuilder();
			FilterTo(input, b);
			return b.ToString();
		}
		
		private bool ContainsTag(Stack stack, TagInfo info)
		{
			foreach (TagInfo t in stack) {
				if (t.Name.ToLower() == info.Name.ToLower())
					return true;
			}
			return false;
		}
		
		public void FilterTo(string input, StringBuilder b)
		{
			char lastchar;
			char c = ' ';
			ITag tag;
			Stack tagStack = new Stack();
			StringBuilder textBlock = new StringBuilder();
			bool acceptWhitespace = false;
			for (int i = 0; i < input.Length; ++i) {
				lastchar = c;
				c = input[i];
				switch (c) {
					case '<':
						TagInfo tagInfo = HtmlHelper.ParseTag(input, i, extendedSyntax);
						bool ok = false;
						if (tagInfo != null && tagInfo.IsEndTag) {
							// End tag
							if (ContainsTag(tagStack, tagInfo)) {
								// Close all tags before current one
								FlushTextBlock(ref textBlock, b, tagStack);
								TagInfo stored;
								do {
									stored = (TagInfo)tagStack.Pop();
									if (stored.Name != tagInfo.Name) {
										tag = (ITag)tags[stored.Name];
										if (tag != null) {
											TagInfo n = tagInfo.CreateEndTag();
											n.Name = stored.Name;
											tag.WriteTag(b, n, this);
										}
									}
								} while (stored.Name.ToLower() != tagInfo.Name.ToLower());
								// Close current tag
								tag = (ITag)tags[tagInfo.Name];
								if (tag != null)
									tag.WriteTag(b, tagInfo, this);
								ok = true;
							}
						} else if (tagInfo != null) {
							// Start tag
							
							tag = (ITag)tags[tagInfo.Name];
							if (tag != null) {
								TagType type = tag.GetTagType(tagInfo, this);
								if (type == TagType.Simple || type == TagType.Block) {
									FlushTextBlock(ref textBlock, b, tagStack);
									tag.WriteTag(b, tagInfo, this);
									ok = true;
									if (type == TagType.Block)
										tagStack.Push(tagInfo);
								} else if (type == TagType.Content) {
									FlushTextBlock(ref textBlock, b, tagStack);
									string endTag = tagInfo.StartChar + "/" + tagInfo.Name + tagInfo.EndChar;
									i = tagInfo.EndPos + 1;
									int endpos = input.ToLower().IndexOf(endTag.ToLower(), i);
									// opening content-tag without proper endtag?
									// everything till end of string is content!
									if (endpos < 0)
										endpos = input.Length;
									tagInfo.EndPos = Math.Min(endpos + endTag.Length, input.Length) - 1;
									string content = input.Substring(i, endpos - i);
									if (!extendedSyntax)
										content = HtmlHelper.DecodeEntities(content);
									tag.WriteContentTag(b, tagInfo, content, this);
									ok = true;
								}
							}
						}
						if (ok) {
							i = tagInfo.EndPos;
						} else {
							textBlock.Append(c);
						}
						acceptWhitespace = true;
						break;
					case '[':
						if (extendedSyntax)
							goto case '<';
						else
							goto default;
					case '`':
						if (extendedSyntax) {
							FlushTextBlock(ref textBlock, b, tagStack);
							b.Append("<formula code=\"");
							for (i += 1; i < input.Length; ++i) {
								if (input[i] == '`')
									break;
								b.Append(HtmlHelper.EncodeEntities(input[i].ToString()));
							}
							b.Append("\">");
							acceptWhitespace = false;
						} else {
							textBlock.Append(c);
						}
						break;
					case '&':
						textBlock.Append(HtmlHelper.ParseEntity(input, ref i));
						acceptWhitespace = true;
						break;
					case '\t':
						if (extendedSyntax)
							textBlock.Append("    ");
						else if (acceptWhitespace)
							textBlock.Append(' ');
						acceptWhitespace = false;
						break;
					default:
						if (char.IsWhiteSpace(c)) {
							if (extendedSyntax)
								textBlock.Append(c);
							else if (acceptWhitespace)
								textBlock.Append(' ');
							acceptWhitespace = false;
						} else {
							textBlock.Append(c);
							acceptWhitespace = true;
						}
						break;
				}
			}
			FlushTextBlock(ref textBlock, b, tagStack);
			while (tagStack.Count > 0) {
				TagInfo tagInfo = (TagInfo)tagStack.Pop();
				tag = (ITag)tags[tagInfo.Name];
				if (tag != null) {
					TagInfo n = tagInfo.CreateEndTag();
					n.StartPos = input.Length;
					n.EndPos = input.Length;
					tag.WriteTag(b, n, this);
				}
			}
		}
		
		private void FlushTextBlock(ref StringBuilder textBlock, StringBuilder b, Stack tagStack)
		{
			DoReplacements(textBlock.ToString(), b, tagStack);
			textBlock = new StringBuilder();
		}
		
		private void DoReplacements(string textBlock, StringBuilder b, Stack tagStack)
		{
			IReplacementMatch best;
			int bestIndex = int.MaxValue;
			int index = 0;
			do {
				best = null;
				foreach (IReplacement r in replacements) {
					IReplacementMatch m = r.Match(textBlock, index, textBlock.Length - index, tagStack);
					if (m != null) {
						int mIndex = m.Index;
						if (mIndex <= bestIndex) {
							if (mIndex == bestIndex && m.Length <= best.Length) {
								// ignore this one: same index but shorter length
							} else {
								best = m;
								bestIndex = best.Index;
							}
						}
					}
				}
				if (best != null) {
					if (bestIndex > index) {
						HtmlHelper.EncodeText(textBlock.Substring(index, bestIndex - index), b);
					}
					b.Append(best.Replacement);
					index = bestIndex + best.Length;
				}
			} while (best != null);
			if (index < textBlock.Length) {
				HtmlHelper.EncodeText(textBlock.Substring(index), b);
			}
		}
	}
}
