// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Text;

namespace DGNetwork.Chat.Parser
{
	/// <summary>
	/// Static class with helper methods for parsing HTML.
	/// </summary>
	public class HtmlHelper
	{
		private HtmlHelper() {}
		
		#region GetMatchingChar
		public static char GetMatchingChar(char c)
		{
			switch (c) {
				case '<':
					return '>';
				case '[':
					return ']';
				case '>':
					return '<';
				case ']':
					return '[';
				case '(':
					return ')';
				case ')':
					return ')';
				case '{':
					return '}';
				case '}':
					return '{';
				case '"':
				case '\'':
					return c;
				default:
					throw new ArgumentException("There is no matching char to " + c);
			}
		}
		#endregion
		
		#region ParseTag
		public static TagInfo ParseTag(string html, int pos)
		{
			return ParseTag(html, pos, false);
		}
		
		public static TagInfo ParseTag(string html, int pos, bool allowDirectAttribute)
		{
			char startChar = html[pos];
			char endChar = GetMatchingChar(startChar);
			char c;
			if (pos == html.Length - 1)
				return null;
			TagInfo tag = new TagInfo();
			tag.StartChar = startChar;
			tag.EndChar = endChar;
			tag.StartPos = pos;
			if (html[pos + 1] == '/') {
				tag.IsEndTag = true;
				pos += 1;
			}
			// Get name
			StringBuilder name = new StringBuilder();
			for (int i = pos + 1; i < html.Length; ++i) {
				c = html[i];
				if (char.IsWhiteSpace(c))
					break;
				if (c == endChar) {
					tag.EndPos = i;
					break;
				}
				if (c == '=') {
					if (!allowDirectAttribute || tag.IsEndTag)
						return null;
					tag.Name = name.ToString();
					i += 1;
					tag.DirectAttribute = ParseAttributeValue(html, ref i, startChar, endChar);
					if (tag.DirectAttribute == null) // return value of null means error
						return null;
					tag.EndPos = i;
					return tag;
				}
				if (char.IsLetterOrDigit(c))
					name.Append(c);
				else
					return null;
			}
			if (name.Length == 0)
				return null;
			if (tag.EndPos == 0 && tag.IsEndTag)
				return null;
			tag.Name = name.ToString();
			
			if (tag.EndPos > 0) {
				return tag;
			}
			
			// Get parameters
			name.Length = 0;
			bool requireSpace = true;
			for (int i = pos + 1 + tag.Name.Length; i < html.Length; ++i) {
				c = html[i];
				if (char.IsWhiteSpace(c)) {
					if (name.Length > 0) {
						tag.Attributes.Add(name.ToString(), "true");
						name.Length = 0;
					}
					requireSpace = false;
					continue;
				}
				if (requireSpace)
					break;
				if (char.IsLetterOrDigit(c)) {
					name.Append(c);
				} else if (c == endChar) {
					tag.EndPos = i;
					return tag;
				} else if (c == '=') {
					if (name.Length == 0)
						break;
					i += 1;
					string val = ParseAttributeValue(html, ref i, startChar, endChar);
					if (val == null)
						break;
					tag.Attributes.Add(name.ToString(), val);
					name.Length = 0;
					i -= 1;
				}
			}
			return null;
		}
		#endregion
		
		#region ParseAttributeValue
		/// <summary>
		/// Parses an attribute value. pos points to the location where the attribute starts,
		/// or to whitespace in front of the start. After the call, pos points to the next
		/// character after the attribute.<br/>
		/// startChar is the starting char of the tag (e.g. &lt;), endChar the ending char (&gt;)
		/// </summary>
		public static string ParseAttributeValue(string html, ref int pos, char startChar, char endChar)
		{
			// Ignore WhiteSpace at the beginning
			for (; pos < html.Length; ++pos) {
				if (!char.IsWhiteSpace(html[pos]))
					break;
			}
			if (pos == html.Length)
				return null;
			char c = html[pos];
			if (c == '"' || c == '\'') {
				string val = ParseQuotedValue(html, ref pos);
				pos += 1;
				return val;
			} else {
				// Get unquoted value
				StringBuilder val = new StringBuilder();
				for (; pos < html.Length; ++pos) {
					c = html[pos];
					if (char.IsWhiteSpace(c) || c == endChar)
						break;
					if (c == startChar)
						return null;
					val.Append(c);
				}
				return val.ToString();
			}
		}
		#endregion
		
		#region ParseQuotedValue
		/// <summary>
		/// Parses a quoted value. When calling this method, pos must point to the
		/// starting quote character. After the call, pos points to the closing
		/// quote chararacter.
		/// In case of errors (end of string without closing quote), the method returns
		/// null and sets pos to the error location (end of the string).
		/// Inside the quoted values, html entities are evaluated.
		/// </summary>
		public static string ParseQuotedValue(string html, ref int pos)
		{
			char quote = html[pos];
			StringBuilder b = new StringBuilder();
			char c;
			for (pos += 1; pos < html.Length; ++pos) {
				c = html[pos];
				if (c == quote)
					return b.ToString();
				if (c == '&') {
					b.Append(ParseEntity(html, ref pos));
				} else if (c == '<' || c == '>') {
					// HTML Syntax error, such characters have to be encoded as &lt; / &gt;
					return null;
				} else if (char.IsWhiteSpace(c)) {
					if (!char.IsWhiteSpace(html[pos - 1]))
						b.Append(' ');
				} else {
					b.Append(c);
				}
			}
			return null;
		}
		#endregion
		
		#region ParseEntity
		/// <summary>
		/// Parses and decodes the entity starting at pos in the html string.
		/// pos must be pointing to the '&amp;' chararacter, after the method call,
		/// it will point to the last character in the entity (normally the ';').
		/// </summary>
		public static string ParseEntity(string html, ref int pos)
		{
			StringBuilder b = new StringBuilder("&");
			char c;
			for (pos += 1; pos < html.Length; ++pos) {
				c = html[pos];
				if (c == ';')
					return DecodeEntity(b.ToString(1, b.Length - 1)); // drop the & when calling decode
				if (c == '<' || c == '&' || char.IsWhiteSpace(c))
					break;
				b.Append(c);
			}
			pos -= 1; // go back to last character in the entity
			return b.ToString();
		}
		#endregion
		
		#region DecodeEntity
		/// <summary>
		/// Decodes a html entity (&amp;entityName;). Specify only the entityName, not the &amp; and ;!
		/// WARNING: This method does not support the full entity set, and doesn't intend to do.
		/// It just decodes the basic entities lt, gt, amp, quot and numeric entities.
		/// </summary>
		public static string DecodeEntity(string entityName)
		{
			switch (entityName)
			{
				case "lt":
					return "<";
				case "gt":
					return ">";
				case "amp":
					return "&";
				case "quot":
					return "\"";
				case "nbsp":
					return " ";
				default:
					if (entityName.StartsWith("#")) {
						try {
							int number = int.Parse(entityName.Substring(1));
							return ((char)number).ToString();
						}
						catch (FormatException) { }
						catch (OverflowException) { }
					}
					return "&" + entityName + ";";
			}
		}
		
		/// <summary>
		/// Decodes all complete html entities from the text.
		/// WARNING: This method does not support the full entity set, and doesn't intend to do.
		/// It just decodes the basic entities lt, gt, amp, quot and numeric entities.
		/// </summary>
		public static string DecodeEntities(string text)
		{
			if (text == null)
				return null;
			// "a &amp; <b> &lt; c &lt d" becomes "a & <b> < c &lt d"
			StringBuilder b = new StringBuilder();
			for (int i = 0; i < text.Length; ++i) {
				char c = text[i];
				if (c == '&')
					b.Append(ParseEntity(text, ref i));
				else
					b.Append(c);
			}
			return b.ToString();
		}
		#endregion
		
		#region EncodeEntities
		/// <summary>
		/// Encodes the special html characters (&lt;, &gt;, &amp;) into entities.
		/// </summary>
		public static string EncodeEntities(string text)
		{
			if (text == null)
				return null;
			StringBuilder b = new StringBuilder();
			DoEncode(text, b, false);
			return b.ToString();
		}
		
		/// <summary>
		/// Encodes the special html characters (&lt;, &gt;, &amp;) into entities.
		/// </summary>
		public static void EncodeEntities(string text, StringBuilder targetBuilder)
		{
			if (targetBuilder == null)
				throw new ArgumentNullException("targetBuilder");
			if (text == null)
				return;
			DoEncode(text, targetBuilder, false);
		}
		
		/// <summary>
		/// Encodes the text to html.
		/// </summary>
		public static string EncodeText(string text)
		{
			if (text == null)
				return null;
			StringBuilder b = new StringBuilder();
			DoEncode(text, b, true);
			return b.ToString();
		}
		
		/// <summary>
		/// Encodes the text to html.
		/// </summary>
		public static void EncodeText(string text, StringBuilder targetBuilder)
		{
			if (targetBuilder == null)
				throw new ArgumentNullException("targetBuilder");
			if (text == null)
				return;
			DoEncode(text, targetBuilder, true);
		}
		
		/// <summary>
		/// Encodes the text as attribute (enclosed in quotes).
		/// </summary>
		public static string EncodeAttribute(string text)
		{
			StringBuilder b = new StringBuilder("\"");
			DoEncode(text, b, false);
			b.Append("\"");
			return b.ToString();
		}
		
		private static void DoEncode(string text, StringBuilder b, bool encodeWhitespace)
		{
			if (text == null)
				throw new ArgumentNullException("text");
			for (int i = 0; i < text.Length; ++i) {
				char c = text[i];
				switch (c) {
					case ' ':
						if (encodeWhitespace &&
						    ((i > 0 && Char.IsWhiteSpace(text[i - 1])) ||
						     (i < text.Length - 1 && Char.IsWhiteSpace(text[i + 1])))) {
							b.Append("&nbsp;");
						} else {
							b.Append(' ');
						}
						break;
					case '\t':
						if (encodeWhitespace)
							b.Append("&nbsp;&nbsp;&nbsp;&nbsp;");
						else
							b.Append('\t');
						break;
					case '\n':
						b.Append('\n');
						if (encodeWhitespace)
							b.Append("<br>");
						break;
					case '<':
						b.Append("&lt;");
						break;
					case '>':
						b.Append("&gt;");
						break;
					case '&':
						b.Append("&amp;");
						break;
					case '"':
						b.Append("&quot;");
						break;
					default:
						b.Append(c);
						break;
				}
			}
		}
		
		/// <summary>
		/// Encodes the integer value as attribute (without quotes).
		/// </summary>
		public static string EncodeAttribute(long value)
		{
			return System.Xml.XmlConvert.ToString(value);
		}
		#endregion
	}
}
