// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;
using DGNetwork.Chat.Parser;

namespace DGNetwork.Chat.Parser
{
	public class Smilie : DefaultReplacement
	{
		public Smilie(string code, string url)
			: base (HtmlHelper.EncodeEntities(code),
			        "<img src=" + HtmlHelper.EncodeAttribute(url) + ">")
		{
		}
	}
	
	public interface IReplacement
	{
		IReplacementMatch Match(string text, int startIndex, int length, Stack tagStack);
	}
	
	public interface IReplacementMatch
	{
		int Index { get; }
		int Length { get; }
		string Replacement { get; }
	}
	
	public class DefaultReplacementMatch : IReplacementMatch
	{
		int index, length;
		
		protected DefaultReplacementMatch(int index, int length)
		{
			if (index < 0)
				throw new ArgumentOutOfRangeException("index", index, "index cannot be negative");
			if (length < 0)
				throw new ArgumentOutOfRangeException("length", length, "length cannot be negative");
			this.index = index;
			this.length = length;
		}
		
		public DefaultReplacementMatch(int index, int length, string replacement) : this(index, length)
		{
			if (replacement == null)
				throw new ArgumentNullException("replacement");
			this.gotReplacement = true;
			this.replacement = replacement;
		}
		
		public int Index {
			get {
				return index;
			}
		}
		
		public int Length {
			get {
				return length;
			}
		}
		
		bool gotReplacement;
		string replacement;
		
		public string Replacement {
			get {
				if (!gotReplacement) {
					gotReplacement = true;
					replacement = GetReplacement();
				}
				return replacement;
			}
		}
		
		public virtual string GetReplacement()
		{
			throw new NotImplementedException();
		}
	}
	
	public class DefaultReplacement : IReplacement
	{
		string code;
		string replaceText;
		bool isRegex;
		
		public DefaultReplacement(string code, string replaceText) : this(code, replaceText, false) { }
		
		public DefaultReplacement(string code, string replaceText, bool isRegex)
		{
			if (code == null || code.Length == 0)
				throw new ArgumentException("code is null or empty");
			this.code = code;
			this.isRegex = isRegex;
			this.replaceText = replaceText;
		}
		
		public string Code {
			get {
				return code;
			}
		}
		
		public bool IsRegex {
			get {
				return isRegex;
			}
		}
		
		public string ReplaceText {
			get {
				return replaceText;
			}
		}
		
		Regex expression;
		
		public virtual IReplacementMatch Match(string text, int startIndex, int length, Stack tagStack)
		{
			foreach (TagInfo info in tagStack) {
				if (string.Compare(info.Name, "a", true) == 0) // do not replace inside URLs
					return null;
			}
			if (this.IsRegex) {
				if (expression == null)
					expression = new Regex(code);
				Match m = expression.Match(text, startIndex, length);
				if (m != null && m.Success)
					return new DefaultReplacementMatch(m.Index, m.Length, m.Result(replaceText));
				else
					return null;
			} else {
				int pos = text.IndexOf(code, startIndex, length);
				if (pos >= 0) {
					return new DefaultReplacementMatch(pos, code.Length, replaceText);
				} else {
					return null;
				}
			}
		}
		
		public override bool Equals(object obj)
		{
			if (!(obj is DefaultReplacement)) return false;
			if (this == obj) return true;
			DefaultReplacement myReplacement = (DefaultReplacement)obj;
			if (code != myReplacement.code) return false;
			if (replaceText != myReplacement.replaceText) return false;
			if (isRegex != myReplacement.isRegex) return false;
			return true;
		}
		
		public override int GetHashCode()
		{
			return (code != null ? code.GetHashCode() : 0) ^
				(replaceText != null ? replaceText.GetHashCode() : 0) ^
				isRegex.GetHashCode();
		}
	}
}
