// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Xml;

using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.AddIns.Codons;
using ICSharpCode.Core.AddIns.Conditions;
using ICSharpCode.Core.Services;
using DGNetwork.Services;

namespace DGNetwork.Chat.Parser
{
	[CodonName("SmilieReplacements")]
	public class SmilieCodon : AbstractCodon
	{
		[XmlMemberAttribute("filemask", IsRequired = true)]
		string filemask = null;
		
		/// <summary>
		/// Creates an item with the specified sub items. And the current
		/// Condition status for this item.
		/// </summary>
		public override object BuildItem(object owner, ArrayList subItems, ConditionCollection conditions)
		{
			if (subItems.Count > 0) {
				Service.Message.ShowErrorFormatted("${res:Codon.NoSubItems}", this.Name);
			}
			ArrayList replacements = new ArrayList();
			LoadDirectory(replacements, Path.Combine(Service.Property.DataDirectory, "smilies"), "smilies");
			replacements.Sort(new ReplacementCodeLengthComparer());
			return replacements.ToArray();
		}
		
		private class ReplacementCodeLengthComparer : IComparer
		{
			public int Compare(object x, object y)
			{
				DefaultReplacement a = x as DefaultReplacement;
				DefaultReplacement b = y as DefaultReplacement;
				if (a == b) // same replacement or both null
					return 0;
				if (a == null)
					return -1;
				if (b == null)
					return 1;
				if (a.Code.Length < b.Code.Length)
					return 1;
				else if (a.Code.Length > b.Code.Length)
					return -1;
				else
					return 0;
			}
		}
		
		#region LoadDirectory
		private void LoadDirectory(ArrayList list, string path, string relPath)
		{
			LoadDirectory(list, new DirectoryInfo(path), relPath);
		}
		
		private void LoadDirectory(ArrayList list, DirectoryInfo path, string relPath)
		{
			if ((path.Attributes & FileAttributes.Hidden) != 0)
				return;
			foreach (FileInfo file in path.GetFiles(filemask)) {
				LoadFile(list, file.FullName, relPath);
			}
			foreach (DirectoryInfo subdir in path.GetDirectories()) {
				LoadDirectory(list, subdir, relPath + "/" + subdir.Name);
			}
		}
		
		private void LoadFile(ArrayList list, string file, string path)
		{
			try {
				XmlDocument doc = new XmlDocument();
				doc.Load(file);
				LoadSet(list, doc.DocumentElement, path);
			} catch (XmlException ex) {
				Service.Message.ShowError("There was an error loading " + file + ":\n" + ex.Message);
			} catch (UnknownElementException ex) {
				Service.Message.ShowError("The file " + file + " contains an unknown element: " + ex.Element.Name);
			} catch (MissingAttributeException ex) {
				Service.Message.ShowError("The element " + ex.Element.Name + " in the file " + file + " needs the attribute " + ex.Attribute);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		#endregion
		
		#region LoadSet
		private void LoadSet(ArrayList list, XmlElement element, string path)
		{
			switch (element.Name) {
				case "messaging-emoticon-map":
					// Kopete-like file
					LoadEmoticonMap(list, element, path);
					break;
				case "sets":
					foreach (XmlElement child in element) {
						LoadSet(list, child, path);
					}
					break;
				case "set":
					LoadDGNSet(list, element, path);
					break;
				default:
					throw new UnknownElementException(element);
			}
		}
		
		private string GetRequiredAttribute(XmlElement element, string attribute)
		{
			string res = element.GetAttribute(attribute);
			if (res == null || res.Length == 0)
				throw new MissingAttributeException(element, attribute);
			return res;
		}
		#endregion
		
		#region LoadDGNSet
		private void LoadDGNSet(ArrayList list, XmlElement element, string path)
		{
			string url = element.GetAttribute("url");
			if (url == null || url.Length == 0) {
				url = path;
			}
			if (!url.EndsWith("/"))
				url += "/";
			foreach (XmlNode node in element) {
				XmlElement e = node as XmlElement;
				if (e == null)
					continue;
				IReplacement s = null;
				string temp;
				switch (e.Name) {
					case "smilie":
						s = new Smilie(e.InnerText, url + GetRequiredAttribute(e, "file"));
						break;
					case "alternativeLocation":
						temp = GetRequiredAttribute(e, "url");
						if (!temp.EndsWith("/"))
							temp += "/";
						string[] date = GetRequiredAttribute(e, "date").Split('/');
						if (date.Length == 2) {
							if (DateTime.Now.Day == int.Parse(date[1]) && DateTime.Now.Month == int.Parse(date[0]))
								url = temp;
						} else if (date.Length == 3) {
							if (DateTime.Now.Day == int.Parse(date[2]) && DateTime.Now.Month == int.Parse(date[1]) && DateTime.Now.Year == int.Parse(date[0]))
								url = temp;
						} else {
							throw new Exception("Invalid date format in " + e.Name);
						}
						break;
					case "replace":
						s = new DefaultReplacement(GetRequiredAttribute(e, "find"), e.InnerText);
						break;
					case "regex":
						s = new DefaultReplacement(GetRequiredAttribute(e, "find"), e.InnerText, true);
						break;
					default:
						throw new UnknownElementException(e);
				}
				
				if (s != null) {
					//foreach (Replacement old in smilies) {
					//	if (old.Code == s.Code)
					//		throw new DoubleCodeException(code);
					//}
					list.Add(s);
				}
			}
		}
		#endregion
		
		#region LoadEmoticonMap
		private void LoadEmoticonMap(ArrayList list, XmlElement element, string path)
		{
			string url = element.GetAttribute("url");
			if (url == null || url.Length == 0) {
				url = path;
			}
			if (!url.EndsWith("/"))
				url += "/";
			foreach (XmlNode node in element) {
				XmlElement emoticon = node as XmlElement;
				if (emoticon == null)
					continue;
				if (emoticon.Name != "emoticon")
					throw new UnknownElementException(emoticon);
				string file = emoticon.GetAttribute("file");
				if (file == null || file == "")
					throw new MissingAttributeException(emoticon, "file");
				string fullfile = url + file + ".png";
				foreach (XmlElement code in emoticon) {
					if (code.Name != "string")
						throw new UnknownElementException(code);
					Smilie s = new Smilie(code.InnerText, fullfile);
					//foreach (Replacement old in smilies) {
					//	if (old.Code == s.Code)
					//		throw new DoubleCodeException(code);
					//}
					list.Add(s);
				}
			}
		}
		#endregion
		
		#region Exceptions
		private class DoubleCodeException : BaseException
		{
			XmlElement code;
			
			public DoubleCodeException(XmlElement code)
			{
				this.code = code;
			}
			
			public XmlElement Code {
				get {
					return code;
				}
			}
		}
		
		private class MissingAttributeException : BaseException
		{
			XmlElement element;
			string attribute;
			
			public MissingAttributeException(XmlElement element, string attribute)
			{
				this.element = element;
				this.attribute = attribute;
			}
			
			public XmlElement Element {
				get {
					return element;
				}
			}
			
			public string Attribute {
				get {
					return attribute;
				}
			}
		}
		
		private class UnknownElementException : BaseException
		{
			XmlElement element;
			
			public UnknownElementException(XmlElement element)
			{
				this.element = element;
			}
			
			public XmlElement Element {
				get {
					return element;
				}
			}
		}
		#endregion
	}
}
