// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;
using DGNetwork.AddIns.Codons;
using DGNetwork.Services;
using DGNetwork.Chat.Parser;

namespace DGNetwork.Chat
{
	public class RichTextEditBox : AbstractEditBox
	{
		RichTextBox textBox;
		
		public RichTextEditBox()
		{
			textBox = new RichTextBox();
			textBox.AcceptsTab = true;
			textBox.TextChanged += new EventHandler(TextBoxTextChanged);
			textBox.KeyDown     += new KeyEventHandler(TextBoxKeyDown);
			textBox.Font = ChatProperties.Font;
			textBox.ForeColor = ChatProperties.Color;
			textBox.AllowDrop = true;
			textBox.DragEnter += new DragEventHandler(TextBoxDrag);
			textBox.DragDrop  += new DragEventHandler(TextBoxDrop);
			this.CanSend = false;
			lastSendText.Add("");
		}
		
		private void TextBoxTextChanged(object sender, EventArgs e)
		{
			this.CanSend = textBox.Text.Trim().Length > 0;
		}
		
		#region Drag'n'Drop
		private void TextBoxDrag(object sender, DragEventArgs e)
		{
			try {
				e.Effect = DragDropService.GetEffects(e.Data, this);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		private void TextBoxDrop(object sender, DragEventArgs e)
		{
			try {
				InsertText(DragDropService.GetText(e.Data, this));
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		#endregion
		
		#region KeyDown
		ArrayList lastSendText = new ArrayList();
		int lastSendTextBrowseIndex = 0;
		
		private void TextBoxKeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Enter) {
				if (e.Control || (!e.Shift && Service.Chat.SendOnEnter)) {
					e.Handled = true;
					OnSendTriggered(EventArgs.Empty);
				}
			}
			
			if (e.KeyData == (Keys.Control | Keys.V)) {
				string s = "";
				s = DragDropService.GetText(Clipboard.GetDataObject(), this);
				if (s != null) {
					textBox.SelectedText = s;
					e.Handled = true;
				}
			}
			
			if (e.Control || (textBox.SelectionStart == 0) &&
			    (textBox.SelectionLength == textBox.TextLength)) {
				if (e.KeyCode == Keys.Up) {
					// "Ebene" nach oben bewegen
					e.Handled = true;
					if (lastSendTextBrowseIndex == 0)
						return;
					if (lastSendTextBrowseIndex == lastSendText.Count - 1) {
						// speichere Text in aktueller "Ebene"
						lastSendText[lastSendTextBrowseIndex] = textBox.Rtf;
					}
					textBox.Rtf = (string) lastSendText[--lastSendTextBrowseIndex];
					textBox.SelectAll();
				}
				if (e.KeyCode == Keys.Down) {
					e.Handled = true;
					if (lastSendTextBrowseIndex == lastSendText.Count - 1)
						return;
					textBox.Rtf = (string) lastSendText[++lastSendTextBrowseIndex];
					textBox.SelectAll();
				}
			}
		}
		#endregion
		
		#region Toolbar Commands
		public abstract class FontStyleCommand : AbstractCheckableMenuCommand
		{
			FontStyle style;
			
			public FontStyleCommand(FontStyle style)
			{
				this.style = style;
			}
			
			public override void Run()
			{
				Font font = box.SelectionFont;
				if ((font.Style & style) == style) {
					box.SelectionFont = new Font(font, font.Style & ~style);
					this.IsChecked = false;
				} else {
					box.SelectionFont = new Font(font, font.Style | style);
					this.IsChecked = true;
				}
			}
			
			ChatPanel chatPanel;
			RichTextBox box;
			
			protected override void OnOwnerChanged(EventArgs e)
			{
				if (chatPanel != null)
					chatPanel.EditBoxPanel.EditBoxChanged -= new EventHandler(EditBoxChanged);
				chatPanel = (ChatPanel)this.Owner;
				if (chatPanel != null)
					chatPanel.EditBoxPanel.EditBoxChanged += new EventHandler(EditBoxChanged);
				EditBoxChanged(null, null);
				base.OnOwnerChanged(e);
			}
			
			private void EditBoxChanged(object sender, EventArgs e)
			{
				if (box != null)
					box.SelectionChanged -= new EventHandler(BoxSelectionChanged);
				if (chatPanel != null && chatPanel.EditBoxPanel.EditBox != null)
					box = chatPanel.EditBoxPanel.EditBox.Control as RichTextBox;
				else
					box = null;
				if (box != null)
					box.SelectionChanged += new EventHandler(BoxSelectionChanged);
				BoxSelectionChanged(null, null);
			}
			
			private void BoxSelectionChanged(object sender, EventArgs e)
			{
				if (box != null) {
					Font font = box.SelectionFont;
					if (font != null)
						this.IsChecked = (font.Style & style) == style;
				}
			}
		}
		
		public class BoldCommand : FontStyleCommand
		{
			public BoldCommand() : base(FontStyle.Bold) {}
		}
		
		public class ItalicCommand : FontStyleCommand
		{
			public ItalicCommand() : base(FontStyle.Italic) {}
		}
		
		public class UnderlineCommand : FontStyleCommand
		{
			public UnderlineCommand() : base(FontStyle.Underline) {}
		}
		#endregion
		
		#region Send Message
		HtmlFilter sendFilter;
		
		public override IMessageSendResult Send()
		{
			Message message = new Message(true);
			message.Room = this.Room;
			message.Text = this.Text;
			message.Rtf = this.Rtf;
			if (sendFilter == null) {
				sendFilter = (HtmlFilter)Helper.BuildItem("/Chat/Filter/Send/ChatCodeToHtml", this);
			}
			RtfConverter converter = new RtfConverter();
			converter.Parser = sendFilter;
			message.Html = converter.Convert(textBox.Rtf, textBox.Text);
			// this.sendFilter is just the HtmlFilter, but now we need to
			// execute all filters.
			Service.Chat.SendFilter.Filter(message, MessageModes.Text | MessageModes.Rtf | MessageModes.Html, this);
			IMessageSendResult result = this.Room.SendMessage(message);
			if (result != null) {
				lastSendText[lastSendText.Count - 1] = textBox.Rtf;
				this.Text = "";
				lastSendText.Add(textBox.Rtf);
				lastSendTextBrowseIndex = lastSendText.Count - 1;
			}
			return result;
		}
		#endregion
		
		#region Interface implementation
		public override Control Control {
			get {
				return textBox;
			}
		}
		
		public override void Focus()
		{
			textBox.Focus();
		}
		
		public override string Rtf {
			get {
				return textBox.Rtf;
			}
			set {
				textBox.Rtf = value;
			}
		}
		
		public override string Text {
			get {
				return textBox.Text;
			}
			set {
				textBox.Text = value;
			}
		}
		
		public override string Html {
			get {
				return Converter.RtfToHtml(this.Rtf);
			}
			set {
				this.Rtf = Converter.HtmlToRtf(value);
			}
		}
		
		public override Font Font {
			get {
				return textBox.SelectionFont;
			}
			set {
				if (textBox.TextLength == 0) {
					textBox.Font = value;
					ChatProperties.Font = value;
				}
				textBox.SelectionFont = value;
			}
		}
		
		public override void InsertText(string text)
		{
			if (text == null)
				return;
			if (textBox.SelectionLength > 0)
				textBox.SelectionLength = 0;
			textBox.SelectedText = text;
		}
		
		public override Color Color {
			get {
				return textBox.SelectionColor;
			}
			set {
				if (textBox.TextLength == 0) {
					textBox.ForeColor = value;
					ChatProperties.Color = value;
				}
				textBox.SelectionColor = value;
			}
		}
		#endregion
	}
}
