// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using DGNetwork.Services;
using DGNetwork.Gui;

namespace DGNetwork.Chat
{
	public class RoomUserView : UserControl
	{
		ListView listView;
		ColumnHeader header;
		ImageListManager imageListManager;
		
		public RoomUserView()
		{
			imageListManager = new ImageListManager();
			listView = new ListView();
			listView.View = View.Details;
			listView.HeaderStyle = ColumnHeaderStyle.None;
			listView.MultiSelect = false;
			listView.Dock = DockStyle.Fill;
			listView.SmallImageList = imageListManager.ImageList;
			header = new ColumnHeader();
			listView.Columns.Add(header);
			this.Controls.Add(listView);
			listView.AllowDrop = true;
			listView.DragEnter += new DragEventHandler(ListViewDrag);
			listView.DragDrop  += new DragEventHandler(ListViewDrop);
		}
		
		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged(e);
			header.Width = this.Width - 6;
		}
		
		protected override void Dispose(bool disposing)
		{
			if (disposing)
				this.Room = null;
			base.Dispose(disposing);
		}
		
		
		IChatRoom room;
		
		public IChatRoom Room {
			get {
				return room;
			}
			set {
				if (room == value)
					return;
				if (this.InvokeRequired)
					throw new Grunwald.Gui.InvokeRequiredException();
				if (room != null) {
					room.UserListVisibleChanged -= new EventHandler(RoomUserListVisibleChanged);
					room.UsersChanged -= new EventHandler(RoomUsersChanged);
					if (room.DisposeOnLeave)
						room.Dispose();
				}
				room = value;
				if (room != null) {
					room.UserListVisibleChanged += new EventHandler(RoomUserListVisibleChanged);
					room.UsersChanged += new EventHandler(RoomUsersChanged);
				}
				RefreshVisible();
				RefreshUsers();
			}
		}
		
		private void RoomUserListVisibleChanged(object sender, EventArgs e)
		{
			this.BeginInvoke(new MethodInvoker(RefreshVisible));
		}
		
		private void RefreshVisible()
		{
			try {
				bool visible = (room == null) ? false : room.UserListVisible;
				if (this.Visible != visible)
					this.Visible = visible;
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		private void RoomUsersChanged(object sender, EventArgs e)
		{
			this.BeginInvoke(new MethodInvoker(RefreshUsers));
		}
		
		private void ClearList()
		{
			foreach (ListViewItem item in listView.Items) {
				IUser user = (IUser)item.Tag;
				user.NameChanged -= new EventHandler(UserNameChanged);
				user.IconChanged -= new EventHandler(UserIconChanged);
			}
			listView.Items.Clear();
		}
		
		private void RefreshUsers()
		{
			try {
				ClearList();
				if (room == null)
					return;
				lock (room.Users) {
					foreach (IUser user in room.Users) {
						ListViewItem item = new ListViewItem(user.Name);
						item.Tag = user;
						item.ImageIndex = imageListManager[user.Icon];
						listView.Items.Add(item);
						user.NameChanged += new EventHandler(UserNameChanged);
						user.IconChanged += new EventHandler(UserIconChanged);
					}
				}
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		private void UserNameChanged(object sender, EventArgs e)
		{
			this.BeginInvoke(new MethodInvoker(RefreshUsers));
		}
		
		private void UserIconChanged(object sender, EventArgs e)
		{
			this.BeginInvoke(new MethodInvoker(RefreshUsers));
		}
		
		private void ListViewDrag(object sender, DragEventArgs e)
		{
			try {
				e.Effect = DragDropService.GetEffects(e.Data, this);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		private void ListViewDrop(object sender, DragEventArgs e)
		{
			try {
				DragDropService.GetText(e.Data, this);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
	}
}
