// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Globalization;
using System.Text;
using DGNetwork.Chat.Parser;

namespace DGNetwork.Chat
{
	/// <summary>
	/// Konvertiert RTF-Code in Chatcode.
	/// </summary>
	public class RtfConverter
	{
		HtmlFilter parser;
		
		/// <summary>
		/// Gets/Sets the parser to use to get information about the tags to ignore format
		/// changes in them.
		/// </summary>
		public HtmlFilter Parser {
			get {
				return parser;
			}
			set {
				parser = value;
			}
		}
		
		#region Convert
		
		/// <summary>
		/// Konvertiert den angegebenen RTF-Code in Chatcode
		/// </summary>
		/// <param name="rtf">Der RTF-Code (RichTextBox.Rtf)</param>
		/// <param name="clearText">Die Klartextentsprechung des RTF (RichTextBox.Text)</param>
		public string Convert(string rtf, string clearText)
		{
			StringBuilder b = new StringBuilder();
			// Input: rtf
			// Output: b
			rtf = rtf.TrimEnd(' ', '\r', '\n', '\0');
			if (rtf[0] != '{' || rtf[rtf.Length - 1] != '}')
				throw new ArgumentException("Invalid RTF source.", "rtf");
			
			Parse(rtf.Substring(1, rtf.Length - 2), b, clearText);
			if (b.Length > 1) {
				if (b[b.Length - 1] == '\n')
					b.Remove(b.Length - 1, 1);
			}
			return b.ToString();
		}
		
		#endregion
		
		#region GetWord
		
		bool isLetter(char c)
		{
			// all RTF control words must be lowercase
			return c >= 'a' && c <= 'z';
		}
		
		/// <summary>
		/// Ruft das Wort an der Stelle ab
		/// </summary>
		/// <param name="rtf">der RTF-Code</param>
		/// <param name="pos">Vor dem Aufruf die Position des \, nach dem Aufruf
		/// die Position des letzten Zeichen des Kontrollwortes</param>
		/// <param name="para">Der numerische Parameter. Wenn kein Parameter vorhanden
		/// ist, ist dies 1!!!</param>
		string GetWord(string rtf, ref int pos, out int para)
		{
			if (rtf[pos] != '\\')
				throw new ArgumentException("GetWord position must point to an \\ character!", "pos");
			pos++;
			char c = rtf[pos];
			
			if (c == '\'') {
				// \'xx ist was ganz spezielles
				string hex = rtf.Substring(pos + 1, 2);
				pos += 2;
				para = int.Parse(hex, NumberStyles.HexNumber);
				return "'";
			}
			
			// Control chars arbeiten ohne Delimiter
			if (!isLetter(c)) {
				para = 1;
				return "" + c;
			}
			
			// Control word
			string word = "" + c;
			string spara = "";
			bool inPara = false;
			for (pos++; pos < rtf.Length; pos++) {
				c = rtf[pos];
				if (!inPara) {
					if (isLetter(c)) {
						word += c;
					} else if (c == '-' || char.IsDigit(c)) {
						spara += c;
						inPara = true;
					} else {
						break;
					}
				} else {
					if (char.IsDigit(c))
						spara += c;
					else
						break;
				}
			}
			if (c != ' ')
				pos--; // nur das Leerzeichen gehört als Delimiter mit zum Word
			// Console.WriteLine("GetWord returns " + word + " with para=" + spara);
			if (inPara) // es gibt einen Parameter
				para = int.Parse(spara);
			else
				para = 1;
			return word;
		}
		
		#endregion
		
		#region StateContainer
		
		private class StateContainer
		{
			public bool Bold {
				get { return bold; }
				set {
					Changed = true;
					bold = value;
				}
			}
			
			public bool Strike {
				get { return strike; }
				set {
					Changed = true;
					strike = value;
				}
			}
			
			public bool Italic {
				get { return italic; }
				set {
					Changed = true;
					italic = value;
				}
			}
			
			public bool Underline {
				get { return underline; }
				set {
					Changed = true;
					underline = value;
				}
			}
			
			public string Font {
				get { return font; }
				set {
					Changed = true;
					font = value;
				}
			}
			
			public int CharSet = 0;
			bool bold, italic, underline, strike;
			string font = "Verdana";
			int size = 24;
			
			public int Size {
				get { return size; }
				set {
					Changed = true;
					size = value;
				}
			}
			
			//public bool LineBegon = true;
			public string CloseFont = "";
			public string oldMainFont = null;
			public bool Changed = true;
			public int UnicodeSkipChars = 1;
			public string StartWord;
			Color color = Color.Black;
			
			public Color Color {
				get { return color; }
				set {
					Changed = true;
					color = value;
				}
			}
			
			public StateContainer(string startWord)
			{
				StartWord = startWord;
			}
			
			public Stack HtmlTagStack = new Stack();
			public int EndOfCurrentTag = -1;
			public string HtmlContentTag = null;
			
			public bool IsMainLayer(int clearTextPos)
			{
				return HtmlTagStack.Count == 0
					&& HtmlContentTag == null
					&& EndOfCurrentTag < clearTextPos;
			}
		}
		
		#endregion
		
		#region SetFont (in HTML)
		
		void SetFont(StateContainer state, StringBuilder b)
		{
			if (SkipCharCount > 0)
				return;
			if (!state.Changed)
				return;
			string mainfont = Converter.FontToHtml(state.Font, state.Size/2.0f, state.Color);
			b.Append(state.CloseFont);
			if (mainfont != state.oldMainFont) {
				if (state.oldMainFont != null)
					b.Append("</font>");
				b.Append(mainfont);
				state.oldMainFont = mainfont;
			}
			state.CloseFont = "";
			if (state.Bold) {
				state.CloseFont = "</b>" + state.CloseFont;
				b.Append("<b>");
			}
			if (state.Italic) {
				state.CloseFont = "</i>" + state.CloseFont;
				b.Append("<i>");
			}
			if (state.Underline) {
				state.CloseFont = "</u>" + state.CloseFont;
				b.Append("<u>");
			}
			if (state.Strike) {
				state.CloseFont = "</s>" + state.CloseFont;
				b.Append("<s>");
			}
			state.Changed = false;
		}
		
		#endregion
		
		#region "Skip Chars - Write"
		
		/// <summary>
		/// Die Anzahl der (Klartext-)Zeichen, die am Anfang der Ausgabe übersprungen werden sollen.
		/// </summary>
		public int SkipCharCount = 0;
		
		void Write(StringBuilder b, char c, StateContainer state)
		{
			clearTextPos++;
			if (SkipCharCount > 0) {
				SkipCharCount--;
				return;
			}
			b.Append(c);
		}
		
		void Write(StringBuilder b, string txt)
		{
			clearTextPos += txt.Length;
			if (SkipCharCount > 0)
				SkipCharCount -= txt.Length;
			else
				b.Append(txt);
		}
		
		void WriteChar(StringBuilder b, string txt)
		{
			clearTextPos++;
			if (SkipCharCount > 0)
				SkipCharCount--;
			else
				b.Append(txt);
		}
		
		#endregion
		
		#region Work
		Encoding encoding = Encoding.GetEncoding(1252);
		
		string[] specialchars = {
			"lquote", "rquote", "ldblquote", "rdblquote",
			"bullet", "endash", "emdash"
		} ;
		
		void Work(StateContainer state, string word, int para, StringBuilder b, bool shouldSetFont, string clearText)
		{
			switch (word) {
				case "red":
					if (state.StartWord != "colortbl")
						return;
					state.Color = Color.FromArgb(para, 0, 0);
					break;
				case "green":
					if (state.StartWord != "colortbl")
						return;
					state.Color = Color.FromArgb(state.Color.R, para, 0);
					break;
				case "blue":
					if (state.StartWord != "colortbl")
						return;
					ColorTable.Add(Color.FromArgb(state.Color.R, state.Color.G, para));
					break;
				case "cf":
					if (ColorTable == null)
						return;
					state.Color = (Color) ColorTable[para];
					break;
				case "f":
					if (FontTable == null)
						return;
					if (FontTable.Count <= para)
						return;
					state.Font = (string) FontTable[para];
					state.CharSet = (int) CharSetTable[para];
					break;
				case "fcharset":
					state.CharSet = para;
					break;
				case "fs":
					state.Size = para;
					break;
				case "b":
					state.Bold = para != 0;
					break;
				case "strike":
					state.Strike = para != 0;
					break;
				case "i":
					state.Italic = para != 0;
					break;
				case "ul":
					state.Underline = true;
					break;
				case "ulnone":
					state.Underline = false;
					break;
				case "{":
				case "}":
				case "\\":
					if (shouldSetFont)
						SetFont(state, b);
					Write(b, word);
					break;
				case "row":
				case "line":
				case "par":
				case "\r":
				case "\n":
					WriteChar(b, "\n");
					break;
				case "~":
				case "emspace":
				case "enspace":
				case "cell":
					WriteChar(b, " ");
					break;
				case "tab":
					WriteChar(b, "    ");
					break;
				case "-":
				case "_":
					if (shouldSetFont)
						SetFont(state, b);
					Write(b, '-', state);
					break;
				case "u":
					if (shouldSetFont)
						SetFont(state, b);
					if (para < 0)
						para += 65536;
					Write(b, (char) para, state);
					parseSkipChars = state.UnicodeSkipChars;
					break;
				case "uc":
					state.UnicodeSkipChars = para;
					break;
				case "'":
					if (shouldSetFont)
						SetFont(state, b);
					if (state.CharSet != 0) {
						Write(b, clearText[clearTextPos], state);
						parseSkipChars = 1;
					} else {
						Write(b, encoding.GetString(new byte[] { (byte) para } )[0], state);
					}
					break;
				default:
					for (int i = 0; i < specialchars.Length; i++) {
						if (word == specialchars[i]) {
							if (shouldSetFont)
								SetFont(state, b);
							WriteChar(b, string.Format("&#{0:d};", 148 + i));
							break;
						}
					}
					break;
			}
		}
		
		#endregion
		
		#region ClearTextPos
		int clearTextPos = 0;
		int parseSkipChars = 0;
		
		void correctClearTextPos(char c, string clearText)
		{
			if (clearText[clearTextPos] == c)
				return; // alles OK
			int pos = clearText.IndexOf(c, clearTextPos, 10); // 10 max. Fehler
			if (pos > 0) {
				clearTextPos = pos;
				return;
			}
			/*System.IO.FileStream fs = new System.IO.FileStream("c:\\crash.dat", System.IO.FileMode.Create);
			System.IO.StreamWriter w = new System.IO.StreamWriter(fs, Encoding.Unicode);
			w.Write(clearText);
			w.Close();
			fs.Close();*/
			throw new Exception("RTFConverter::correctClearTextPos\r\n" +
			                    clearText.Insert(clearTextPos, ">>>"));
		}
		
		#endregion
		
		#region Parse
		ArrayList ColorTable;
		ArrayList FontTable;
		ArrayList CharSetTable;
		
		/// <summary>
		/// Parses an RTF Section and writes the results to the StringBuilder.
		/// </summary>
		StateContainer Parse(string rtf, StringBuilder b, string clearText)
		{
			parseSkipChars = 0;
			if (rtf[0] != '\\')
				throw new ArgumentException("RTF section must start with \\!\r\n" + rtf, "rtf");
			int pos = 0;
			int para;
			string startWord = GetWord(rtf, ref pos, out para);
			//  /* hat eine spezielle Bedeutung, wer sie nicht implementieren
			//  möchte sollte die gesamte Sektion ignorieren
			if (startWord == "*")
				return null;
			if (startWord == "rtf")
				clearTextPos = 0;
			switch (startWord) {
				case "rtf":
					break;
				case "colortbl":
					ColorTable = new ArrayList();
					ColorTable.Add(Color.Black);
					break;
				case "fonttbl": // font table
					FontTable = new ArrayList();
					CharSetTable = new ArrayList();
					break;
				case "f": // font definition
					break;
				default:
					return null;
			}
			StateContainer state = new StateContainer(startWord);
			
			#region Parse Content
			for (pos++; pos < rtf.Length; pos++) {
				char c = rtf[pos];
				if (c == '{') {
					int startpos = pos + 1; // after the {
					
					#region Search for the end of the new section
					int layers = 0;
					for (pos++; pos < rtf.Length; pos++) {
						c = rtf[pos];
						if (c == '\\') {
							// \ kann in RTF verwendet werden, um } zu escapen
							pos++;
						} else if (c == '{') {
							layers++;
						} else if (c == '}') {
							layers--;
							if (layers < 0)
								break;
						}
					}
					if (pos == rtf.Length)
						throw new ArgumentException("RTF section end not found", "rtf");
					#endregion
					
					if (startWord == "fonttbl")
						b = new StringBuilder();
					StateContainer substate;
					substate = Parse(rtf.Substring(startpos, pos - startpos), b, null);
					if (startWord == "fonttbl") {
						string font = b.ToString();
						if (font.EndsWith(";"))
							font = font.Substring(0, font.Length - 1);
						FontTable.Add(font);
						CharSetTable.Add(substate.CharSet);
						state.CharSet = 0;
					}
				} else if (c == '\\') {
					string word = GetWord(rtf, ref pos, out para);
					bool shouldSetFont = state.Changed && state.IsMainLayer(clearTextPos);
					if (parseSkipChars > 0) {
						parseSkipChars--;
					} else {
						Work(state, word, para, b, shouldSetFont, clearText);
					}
				} else if (c == '}') {
					throw new Exception("unexpected } in \r\n" + rtf);
				} else {
					#region Normales Text-Zeichen
					if (startWord == "f" && c != '\r' && c != '\n') {
						b.Append(c);
						continue;
					}
					if (startWord != "rtf")
						continue;
					if (parseSkipChars > 0) {
						parseSkipChars--;
						continue;
					}
					
					#region HTML-Tag-Start
					if ((c == '<' || c == '[') && state.EndOfCurrentTag < clearTextPos) {
						correctClearTextPos(c, clearText);
						// HTML-Tag ?
						TagInfo info = HtmlHelper.ParseTag(clearText, clearTextPos);
						if (info != null) {
							string cmd = info.Name;
							if (info.IsEndTag && state.HtmlContentTag == info.Name) {
								state.EndOfCurrentTag = info.EndPos;
								state.HtmlContentTag = null;
							} else if (state.HtmlContentTag == null) {
								state.EndOfCurrentTag = info.EndPos;
								if (info.IsEndTag && state.HtmlTagStack.Contains(cmd)) {
									// I'm porting this over to the new DG Network and uhmmm...
									// I think that really is intended to be an empty loop!
									while ((string) state.HtmlTagStack.Pop() != cmd) {
									}
								} else if (!info.IsEndTag) {
									ITag tag = (ITag)parser.Tags[info.Name];
									if (tag != null) {
										TagType t = tag.GetTagType(info, parser);
										if (t == TagType.Invalid)
											state.EndOfCurrentTag = -1; // gilt nicht als Tag
										if (t == TagType.Block)
											state.HtmlTagStack.Push(cmd); // ab auf den Stack
										if (t == TagType.Content)
											state.HtmlContentTag = info.Name;
									}
								}
							}
						}
					}
					#endregion
					
					if (c == '\r' || c == '\n') {
						// ignore newline
					} else {
						//Console.WriteLine("Appending " + c);
						if (state.Changed && state.IsMainLayer(clearTextPos))
							SetFont(state, b);
						Write(b, c, state);
					}
					#endregion
				}
			}
			#endregion
			
			return state;
		}
		
		#endregion
	}
}
