// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;

namespace DGNetwork.Contacts
{
	public interface IContact : ICloneable, DGNetwork.Gui.Menu.IItemWithContextMenu
	{
		bool CanClone { get; }
		
		string Name { get; set; }
		event EventHandler NameChanged;
		
		string Icon { get; set; }
		event EventHandler IconChanged;
		
		/// <summary>
		/// The update method is called every second by the tree view showing the contact.
		/// It should be used to update the name and icon if necessary, but if the class
		/// where the contact is getting the information from supports events when these
		/// information change, you should use that events instead of update.
		/// </summary>
		void Update();
		
		/// <summary>
		/// User has executed double click on the contact.
		/// </summary>
		void DoubleClick();
		
		/// <summary>
		/// Gets if the specified contact list can be a parent of this contact.
		/// </summary>
		bool IsValidParent(IContactList parent);
		
		/// <summary>
		/// Gets the font style of this contact when the currently opened room is <paramref name="currentRoom"/>.
		/// </summary>
		System.Drawing.FontStyle GetFontStyle(DGNetwork.Chat.IChatRoom currentRoom);
	}
	
	#region DefaultContact
	public class DefaultContact : IContact
	{
		public DefaultContact()
		{
		}
		
		public DefaultContact(IContact contact)
		{
			this.name = contact.Name;
			this.icon = contact.Icon;
		}
		
		/// <summary>
		/// Gets if the specified contact list can be a parent of this contact.
		/// </summary>
		public virtual bool IsValidParent(IContactList parent)
		{
			return true;
		}
		
		#region Name
		string name;
		
		public string Name {
			get {
				return name;
			}
			set {
				if (name == value)
					return;
				name = value;
				OnNameChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler NameChanged;
		
		protected virtual void OnNameChanged(EventArgs e)
		{
			if (NameChanged != null) {
				NameChanged(this, e);
			}
		}
		#endregion
		
		#region Icon
		string icon;
		
		public string Icon {
			get {
				return icon;
			}
			set {
				if (icon == value)
					return;
				icon = value;
				OnIconChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler IconChanged;
		
		protected virtual void OnIconChanged(EventArgs e)
		{
			if (IconChanged != null) {
				IconChanged(this, e);
			}
		}
		#endregion
		
		#region ContextMenu
		string contextMenuTreePath;
		
		public string ContextMenuTreePath {
			get {
				return contextMenuTreePath;
			}
			set {
				contextMenuTreePath = value;
			}
		}
		
		object DGNetwork.Gui.Menu.IItemWithContextMenu.Owner {
			get {
				return this;
			}
		}
		#endregion
		
		#region Update
		/// <summary>
		/// The update method is called every second by the tree view showing the contact.
		/// It should be used to update the name and icon if necessary, but if the class
		/// where the contact is getting the information from supports events when these
		/// information change, you should use that events instead of update.
		/// Don't forget to call update on your base class.
		/// </summary>
		public virtual void Update()
		{
			// default contact doesn't need update...
		}
		#endregion
		
		public virtual void DoubleClick()
		{
		}
		
		#region Clone
		/// <summary>
		/// Gets if this contact can be cloned.
		/// </summary>
		public virtual bool CanClone {
			get {
				return true;
			}
		}
		
		/// <summary>
		/// Clones the current contact.
		/// If not overridden, this is done using reflection to invoke the copy constructor
		/// of the current type.
		/// </summary>
		public virtual IContact Clone()
		{
			Type t = this.GetType();
			ConstructorInfo c = t.GetConstructor(new Type[] {t});
			if (c == null)
				throw new NotSupportedException(t.FullName + " doesn't support cloning: it doesn't have a copy constructor");
			return (IContact)c.Invoke(new object[] {this});
		}
		
		object ICloneable.Clone()
		{
			return this.Clone();
		}
		#endregion
		
		/// <summary>
		/// Gets the font style of this contact when the currently opened room is <paramref name="currentRoom"/>.
		/// </summary>
		public virtual System.Drawing.FontStyle GetFontStyle(DGNetwork.Chat.IChatRoom currentRoom)
		{
			return System.Drawing.FontStyle.Regular;
		}
		
		public override string ToString()
		{
			return String.Format("[{0}: Name = {1}]", GetType().Name, this.Name);
		}
	}
	#endregion
	
	#region ContactEventHandler / ContactEventArgs
	public delegate void ContactEventHandler(object sender, ContactEventArgs e);
	
	public class ContactEventArgs : EventArgs
	{
		IContact contact;
		
		public IContact Contact {
			get {
				return contact;
			}
		}
		
		int index;
		
		/// <summary>
		/// Gets the index the element was inserted at.
		/// This property is -1 if the element was deleted.
		/// </summary>
		public int Index {
			get {
				return index;
			}
		}
		
		public ContactEventArgs(IContact contact) {
			this.contact = contact;
			this.index = -1;
		}
		
		public ContactEventArgs(IContact contact, int index) {
			this.contact = contact;
			this.index = index;
		}
	}
	#endregion
}
