// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace DGNetwork.Contacts
{
	public interface IContactCollectionEventSender
	{
		void BeginUpdate();
		
		void EndUpdate();
		
		void RaiseAdded(IContact contact, int index);
		
		void RaiseRemoved(IContact contact);
		
		void RaiseMoved(IContact contact, int index);
		
		void CheckContact(IContact contact);
	}
	
	/// <summary>
	///   A collection that stores <see cref='IContact'/> objects.
	/// </summary>
	[Serializable()]
	public class ContactCollection : CollectionBase
	{
		IContactCollectionEventSender parent;
		
		public ContactCollection()
		{
		}
		
		public ContactCollection(IContactCollectionEventSender parent)
		{
			this.parent = parent;
		}
		
		private void CheckValue(IContact contact)
		{
			if (Contains(contact))
				throw new ArgumentException("The specified element is already added to the collection.");
			parent.CheckContact(contact);
		}
		
		/// <summary>
		///   Represents the entry at the specified index of the <see cref='IContact'/>.
		/// </summary>
		/// <param name='index'>The zero-based index of the entry to locate in the collection.</param>
		/// <value>The entry at the specified index of the collection.</value>
		/// <exception cref='ArgumentOutOfRangeException'><paramref name='index'/> is outside the valid range of indexes for the collection.</exception>
		public IContact this[int index] {
			get {
				return ((IContact)(List[index]));
			}
			set {
				IContact old = (IContact)List[index];
				if (old == value) return;
				CheckValue(value);
				if (parent == null) {
					List[index] = value;
				} else {
					parent.BeginUpdate();
					try {
						parent.RaiseRemoved(old);
						List[index] = value;
						parent.RaiseAdded(value, index);
					} finally {
						parent.EndUpdate();
					}
				}
			}
		}
		
		/// <summary>
		///   Adds a <see cref='IContact'/> with the specified value to the
		///   <see cref='ContactCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='IContact'/> to add.</param>
		/// <returns>The index at which the new element was inserted.</returns>
		/// <seealso cref='ContactCollection.AddRange'/>
		public int Add(IContact val)
		{
			CheckValue(val);
			int pos = List.Add(val);
			if (parent != null) parent.RaiseAdded(val, pos);
			return pos;
		}
		
		public void Move(IContact val, int index)
		{
			int pos = IndexOf(val);
			if (pos < 0)
				throw new ArgumentException("The specified element is not in this collection.");
			if (pos == index)
				return;
			int targetIndex;
			if (pos < index) {
				targetIndex = index - 1;
				// move element down
				//  -> move all other elements up
				for (int i = pos; i < targetIndex; ++i) {
					List[i] = List[i + 1];
				}
			} else {
				targetIndex = index;
				// move element up
				// -> move all other elements down
				for (int i = pos; i > targetIndex; --i) {
					List[i] = List[i - 1];
				}
			}
			List[targetIndex] = val; // set target element
			if (parent != null) parent.RaiseMoved(val, index);
		}
		
		/// <summary>
		///   Copies the elements of an array to the end of the <see cref='ContactCollection'/>.
		/// </summary>
		/// <param name='val'>
		///    An array of type <see cref='IContact'/> containing the objects to add to the collection.
		/// </param>
		/// <seealso cref='ContactCollection.Add'/>
		public void AddRange(IContact[] val)
		{
			if (parent != null) parent.BeginUpdate();
			try {
				for (int i = 0; i < val.Length; i++) {
					this.Add(val[i]);
				}
			} finally {
				if (parent != null) parent.EndUpdate();
			}
		}
		
		/// <summary>
		///   Adds the contents of another <see cref='ContactCollection'/> to the end of the collection.
		/// </summary>
		/// <param name='val'>
		///    A <see cref='ContactCollection'/> containing the objects to add to the collection.
		/// </param>
		/// <seealso cref='ContactCollection.Add'/>
		public void AddRange(ContactCollection val)
		{
			if (parent != null) parent.BeginUpdate();
			try {
				for (int i = 0; i < val.Count; i++) {
					this.Add(val[i]);
				}
			} finally {
				if (parent != null) parent.EndUpdate();
			}
		}
		
		/// <summary>
		///   Gets a value indicating whether the
		///    <see cref='ContactCollection'/> contains the specified <see cref='IContact'/>.
		/// </summary>
		/// <param name='val'>The <see cref='IContact'/> to locate.</param>
		/// <returns>
		/// <see langword='true'/> if the <see cref='IContact'/> is contained in the collection;
		///   otherwise, <see langword='false'/>.
		/// </returns>
		/// <seealso cref='ContactCollection.IndexOf'/>
		public bool Contains(IContact val)
		{
			return List.Contains(val);
		}
		
		/// <summary>
		///   Copies the <see cref='ContactCollection'/> values to a one-dimensional <see cref='Array'/> instance at the
		///    specified index.
		/// </summary>
		/// <param name='array'>The one-dimensional <see cref='Array'/> that is the destination of the values copied from <see cref='ContactCollection'/>.</param>
		/// <param name='index'>The index in <paramref name='array'/> where copying begins.</param>
		/// <exception cref='ArgumentException'>
		///   <para><paramref name='array'/> is multidimensional.</para>
		///   <para>-or-</para>
		///   <para>The number of elements in the <see cref='ContactCollection'/> is greater than
		///         the available space between <paramref name='arrayIndex'/> and the end of
		///         <paramref name='array'/>.</para>
		/// </exception>
		/// <exception cref='ArgumentNullException'><paramref name='array'/> is <see langword='null'/>. </exception>
		/// <exception cref='ArgumentOutOfRangeException'><paramref name='arrayIndex'/> is less than <paramref name='array'/>'s lowbound. </exception>
		/// <seealso cref='Array'/>
		public void CopyTo(IContact[] array, int index)
		{
			List.CopyTo(array, index);
		}
		
		/// <summary>
		///    Returns the index of a <see cref='IContact'/> in
		///       the <see cref='ContactCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='IContact'/> to locate.</param>
		/// <returns>
		///   The index of the <see cref='IContact'/> of <paramref name='val'/> in the
		///   <see cref='ContactCollection'/>, if found; otherwise, -1.
		/// </returns>
		/// <seealso cref='ContactCollection.Contains'/>
		public int IndexOf(IContact val)
		{
			return List.IndexOf(val);
		}
		
		/// <summary>
		///   Inserts a <see cref='IContact'/> into the <see cref='ContactCollection'/> at the specified index.
		/// </summary>
		/// <param name='index'>The zero-based index where <paramref name='val'/> should be inserted.</param>
		/// <param name='val'>The <see cref='IContact'/> to insert.</param>
		/// <seealso cref='ContactCollection.Add'/>
		public void Insert(int index, IContact val)
		{
			CheckValue(val);
			List.Insert(index, val);
			if (parent != null) parent.RaiseAdded(val, index);
		}
		
		/// <summary>
		///   Removes a specific <see cref='IContact'/> from the <see cref='ContactCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='IContact'/> to remove from the <see cref='ContactCollection'/>.</param>
		/// <exception cref='ArgumentException'><paramref name='val'/> is not found in the Collection.</exception>
		public void Remove(IContact val)
		{
			List.Remove(val);
			if (parent != null) parent.RaiseRemoved(val);
		}
		
		// TODO: RemoveAt and Clear don't throw events.
	}
}
