// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace DGNetwork.Contacts
{
	/// <summary>
	/// Description of ContactList.
	/// </summary>
	public interface IContactList : IContact
	{
		ContactCollection Contacts { get; }
		
		bool Expanded { get; set; }
		event EventHandler ExpandedChanged;
		bool IsValidChild(IContact contact);
		
		event ContactEventHandler ContactAdded;
		event ContactEventHandler ContactRemoved;
		event EventHandler ContactListChanged;
	}
	
	public class DefaultContactList : DefaultContact, IContactList, IContactCollectionEventSender
	{
		ContactCollection contacts;
		
		public DefaultContactList()
		{
			contacts = new ContactCollection(this);
		}
		
		public DefaultContactList(IContactList contact) : base(contact)
		{
			contacts = new ContactCollection(this);
			this.expanded = contact.Expanded;
			foreach (IContact child in contact.Contacts) {
				this.Contacts.Add((IContact)child.Clone());
			}
		}
		
		public ContactCollection Contacts {
			get {
				return contacts;
			}
		}
		
		public virtual bool IsValidChild(IContact contact)
		{
			return true;
		}
		
		public override bool CanClone {
			get {
				foreach (IContact contact in this.Contacts) {
					if (!contact.CanClone)
						return false;
				}
				return true;
			}
		}
		
		/// <summary>
		/// The update method is called every second by the tree view showing the contact.
		/// It should be used to update the name and icon if necessary, but if the class
		/// where the contact is getting the information from supports events when these
		/// information change, you should use that events instead of update.
		/// Don't forget to call update on your base class.
		/// This update method calls update on all child contacts.
		/// </summary>
		public override void Update()
		{
			foreach (IContact contact in contacts) {
				contact.Update();
			}
			base.Update();
		}
		
		#region Expanded
		bool expanded;
		
		public bool Expanded {
			get {
				return expanded;
			}
			set {
				if (expanded == value)
					return;
				expanded = value;
				OnExpandedChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler ExpandedChanged;
		
		protected virtual void OnExpandedChanged(EventArgs e)
		{
			if (ExpandedChanged != null) {
				ExpandedChanged(this, e);
			}
		}
		#endregion
		
		#region Contact Events
		public event ContactEventHandler ContactAdded;
		public event ContactEventHandler ContactRemoved;
		public event EventHandler ContactListChanged;
		
		protected virtual void OnContactAdded(ContactEventArgs e)
		{
			if (ContactAdded != null) {
				ContactAdded(this, e);
			}
		}
		
		protected virtual void OnContactRemoved(ContactEventArgs e)
		{
			if (ContactRemoved != null) {
				ContactRemoved(this, e);
			}
		}
		
		protected virtual void OnContactListChanged(EventArgs e)
		{
			if (ContactListChanged != null) {
				ContactListChanged(this, e);
			}
		}
		#endregion
		
		#region IContactCollectionEventSender interface implementation
		bool listChangedEventActive = true;
		bool listChangedEventDirty = false;
		
		void IContactCollectionEventSender.BeginUpdate()
		{
			listChangedEventActive = false;
		}
		
		void IContactCollectionEventSender.EndUpdate()
		{
			if (listChangedEventDirty) {
				OnContactListChanged(EventArgs.Empty);
				listChangedEventDirty = false;
			}
			listChangedEventActive = true;
		}
		
		void IContactCollectionEventSender.CheckContact(IContact contact)
		{
			CheckContactBeforeAdd(contact);
		}
		
		/// <summary>
		/// Checks a contact before adding it to the contacts collection.
		/// If not overridden, this method uses ValidChild to do the test.
		/// </summary>
		protected virtual void CheckContactBeforeAdd(IContact contact)
		{
			if (!this.IsValidChild(contact))
				throw new ArgumentException("Child contact is not valid");
		}
		
		void IContactCollectionEventSender.RaiseAdded(IContact contact, int index)
		{
			OnContactAdded(new ContactEventArgs(contact, index));
			if (listChangedEventActive) {
				OnContactListChanged(EventArgs.Empty);
			} else {
				listChangedEventDirty = true;
			}
		}
		
		void IContactCollectionEventSender.RaiseRemoved(IContact contact)
		{
			OnContactRemoved(new ContactEventArgs(contact));
			if (listChangedEventActive) {
				OnContactListChanged(EventArgs.Empty);
			} else {
				listChangedEventDirty = true;
			}
		}
		
		void IContactCollectionEventSender.RaiseMoved(IContact contact, int index)
		{
			if (listChangedEventActive) {
				OnContactListChanged(EventArgs.Empty);
			} else {
				listChangedEventDirty = true;
			}
		}
		#endregion
	}
}
