// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Runtime.Serialization;
using System.Text;
using System.Xml;
using System.Xml.Serialization;

using ICSharpCode.Core.Properties;
using ICSharpCode.Core.Services;

using DGNetwork.Gui;
using DGNetwork.Services;

namespace DGNetwork.Contacts
{
	public class ContactService : AbstractService
	{
		DefaultContactList root;
		
		/// <summary>
		/// Returns the invisible root node of the contact tree.
		/// </summary>
		public IContactList Root {
			get {
				return root;
			}
		}
		
		/// <summary>
		/// Returns the <see cref="ContactCollection"/>. This is equal to
		/// <c>ContactService.Root.Contacts</c>.
		/// </summary>
		public ContactCollection Contacts {
			get {
				return root.Contacts;
			}
		}
		
		#region RecursedContacts
		public IEnumerable RecursedContacts {
			get {
				return new RecursedContactCollection(root.Contacts);
			}
		}
		
		private class RecursedContactCollection : IEnumerable
		{
			ContactCollection col;
			
			public RecursedContactCollection(ContactCollection col)
			{
				this.col = col;
			}
			
			public IEnumerator GetEnumerator()
			{
				return new RecursedContactEnumerator(col);
			}
		}
		
		private class RecursedContactEnumerator : IEnumerator
		{
			Stack stack = new Stack();
			ContactCollection col;
			IEnumerator active;
			
			public RecursedContactEnumerator(ContactCollection col)
			{
				this.col = col;
				Reset();
			}
			
			object current;
			
			public object Current {
				get {
					if (active == null || current == null)
						throw new InvalidOperationException();
					return current;
				}
			}
			
			public void Reset()
			{
				stack.Clear();
				active = col.GetEnumerator();
				current = null;
			}
			
			public bool MoveNext()
			{
				if (active == null)
					return false;
				if (active.MoveNext()) {
					current = active.Current;
					if (current is IContactList) {
						IContactList l = (IContactList)current;
						stack.Push(active);
						active = l.Contacts.GetEnumerator();
					}
					return true;
				} else {
					if (stack.Count > 0) {
						active = (IEnumerator)stack.Pop();
						return MoveNext();
					} else {
						return false;
					}
				}
			}
		}
		#endregion
		
		#region Load/Save Contact List
		public override void InitializeService()
		{
			ContactListSerializer cls = (ContactListSerializer)Service.Property.GetProperty("Contacts", new ContactListSerializer(this));
			root = cls.ContactList;
			root.ContextMenuTreePath = "/Context/Contacts/TreeView";
			Service.Timer.Tick += new EventHandler(TimerTick);
		}
		
		private void TimerTick(object sender, EventArgs e)
		{
			root.Update();
		}
		
		public XmlElement ToXmlElement(XmlDocument doc, IContact contact)
		{
			if (contact is IXmlConvertable) {
				return ((IXmlConvertable)contact).ToXmlElement(doc);
			} else if (contact.GetType() == typeof(DefaultContactList)) {
				DefaultContactList list = (DefaultContactList)contact;
				XmlElement element = GetElementWithCommonProperties(doc, "ContactList", list);
				foreach (IContact c in list.Contacts) {
					element.AppendChild(ToXmlElement(doc, c));
				}
				return element;
			} else if (contact.GetType() == typeof(DefaultContact)) {
				return GetElementWithCommonProperties(doc, "Contact", contact);
			} else {
				throw new BaseException("Can't convert " + contact.GetType().FullName + " to xml because it doesn't implement IXmlConvertable!");
			}
		}
		
		/// <summary>
		/// Loads all child elements of the specified node into the
		/// <see cref="ContactCollection"/>.
		/// </summary>
		public void LoadChildElements(XmlElement parent, ContactCollection contacts)
		{
			foreach (XmlNode node in parent) {
				IContact contact = FromXmlElement(node as XmlElement);
				if (contact != null) {
					contacts.Add(contact);
				}
			}
		}
		
		/// <summary>
		/// Converts the specified XmlElement into a contact.
		/// If element is null, this method returns null.
		/// Exceptions are displayed to the user, then the method returns null.
		/// </summary>
		public IContact FromXmlElement(XmlElement element)
		{
			IContact contact;
			string elementName = element.Name;
			switch (elementName) {
				case "Contact":
					contact = new DefaultContact();
					LoadCommonProperties(element, contact);
					return contact;
				case "ContactList":
					contact = new DefaultContactList();
					LoadChildElements(element, ((IContactList)contact).Contacts);
					LoadCommonProperties(element, contact);
					return contact;
				case "DGNetwork.Discovery.DiscoveryContact":
					elementName = "DGNetwork.AddIns.Discovery.DiscoveryContact";
					goto default;
				default:
					Type type = Helper.GetType(elementName);
					if (type == null) {
						//Service.Message.ShowErrorFormatted("${res:Global.ClassNotFound}", element.Name);
						type = typeof(UnknownContact);
					}
					if (!typeof(IXmlConvertable).IsAssignableFrom(type)) {
						Service.Message.ShowError(elementName + " does not implement IXmlConvertible!");
						return null;
					}
					IXmlConvertable obj;
					try {
						obj = (IXmlConvertable)Activator.CreateInstance(type);
					} catch (Exception ex) {
						Service.Message.ShowError(ex);
						return null;
					}
					object obj2 = obj.FromXmlElement(element);
					if (obj2 == null)
						return null;
					if (!(obj2 is IContact)) {
						Service.Message.ShowError(elementName + " was converted to " + obj2.GetType().FullName + ", but that class does not implement IContact.");
						return null;
					}
					return (IContact)obj2;
			}
		}
		
		public void LoadCommonProperties(XmlElement element, IContact contact)
		{
			string val;
			contact.Name = element.GetAttribute("name");
			if (contact.Icon == null)
				contact.Icon = element.GetAttribute("icon");
			IContactList list = contact as IContactList;
			if (list != null) {
				val = element.GetAttribute("expanded");
				if (val.Length > 0)
					list.Expanded = XmlConvert.ToBoolean(val);
			}
			if (contact.GetType() == typeof(DefaultContactList) || contact.GetType() == typeof(DefaultContact)) {
				DefaultContact dc = (DefaultContact)contact;
				dc.ContextMenuTreePath = element.GetAttribute("contextMenu");
			}
		}
		
		public XmlElement GetElementWithCommonProperties(XmlDocument doc, IContactList contact)
		{
			return GetElementWithCommonProperties(doc, contact.GetType().FullName, contact);
		}
		
		public XmlElement GetElementWithCommonProperties(XmlDocument doc, string title, IContactList contact)
		{
			XmlElement element = GetElementWithCommonProperties(doc, title, (IContact)contact);
			element.SetAttribute("expanded", XmlConvert.ToString(contact.Expanded));
			return element;
		}
		
		public XmlElement GetElementWithCommonProperties(XmlDocument doc, IContact contact)
		{
			return GetElementWithCommonProperties(doc, contact.GetType().FullName, contact);
		}
		
		public XmlElement GetElementWithCommonProperties(XmlDocument doc, string title, IContact contact)
		{
			XmlElement element = doc.CreateElement(title);
			element.SetAttribute("name", contact.Name);
			element.SetAttribute("icon", contact.Icon);
			if (contact.GetType() == typeof(DefaultContactList) || contact.GetType() == typeof(DefaultContact))
				element.SetAttribute("contextMenu", contact.ContextMenuTreePath);
			return element;
		}
		
		private class ContactListSerializer : IXmlConvertable
		{
			DefaultContactList contactList = new DefaultContactList();
			ContactService service;
			
			public ContactListSerializer(ContactService service) {
				this.service = service;
			}
			
			public DefaultContactList ContactList {
				get {
					return contactList;
				}
			}
			
			public object FromXmlElement(XmlElement element)
			{
				contactList.Contacts.Clear();
				service.LoadChildElements(element, contactList.Contacts);
				return this;
			}
			
			public XmlElement ToXmlElement(XmlDocument doc)
			{
				return service.ToXmlElement(doc, contactList);
			}
		}
		#endregion
	}
}
