// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;
using Grunwald.Gui.Controls;
using DGNetwork.Chat;
using DGNetwork.Gui;
using DGNetwork.Services;

namespace DGNetwork.Contacts
{
	public class ContactTreeView : UserControl
	{
		RearrangeTreeView treeView = new RearrangeTreeView();
		IContactList root;
		TreeNode rootNode;
		ImageListManager imageListManager = new ImageListManager();
		
		#region Constructor
		public ContactTreeView()
		{
			root = Service.Contact.Root;
			treeView.ShowRootLines = false;
			treeView.AutoTreeNodeManagement = false;
			treeView.SelectedImageIndex = 99999;
			treeView.ImageIndex = 99999;
			treeView.ImageList = imageListManager.ImageList;
			treeView.AfterExpand   += new TreeViewEventHandler(TreeViewAfterExpand);
			treeView.AfterCollapse += new TreeViewEventHandler(TreeViewAfterCollapse);
			treeView.MouseDown   += new MouseEventHandler(TreeViewMouseDown);
			treeView.MouseUp     += new MouseEventHandler(TreeViewMouseUp);
			treeView.DoubleClick += new EventHandler(TreeViewDoubleClick);
			rootNode = AddChild(null, root);
			foreach (TreeNode node in rootNode.Nodes) {
				treeView.Nodes.Add(node);
			}
			
			// Work around Windows Forms bug: vertical scroll bar shown even when it isn't needed
			// WHIDBEY : see if this is still needed
			treeView.Width = 1;
			treeView.CreateControl();
			// End workround
			
			treeView.Dock = DockStyle.Fill;
			this.Controls.Add(treeView);
		}
		#endregion
		
		#region Dispose
		protected override void Dispose(bool disposing)
		{
			try {
				if (disposing) {
					imageListManager.Dispose();
					ReleaseMonitor(root);
					DisposeFonts();
				}
			} finally {
				base.Dispose(disposing);
			}
		}
		#endregion
		
		#region Child management
		Hashtable nodeHash = new Hashtable(); // IContact -> TreeNode
		
		private TreeNode AddChild(ITreeTag parent, IContact c)
		{
			lock (c) {
				int image = imageListManager[c.Icon];
				TreeNode node = new TreeNode(Service.StringParser.Parse(c.Name), image, image);
				nodeHash[c] = node;
				node.Tag = new ContactWrapperTag(parent, c);
				if (c == root)
					treeView.RootTag = (ITreeTag)node.Tag;
				c.NameChanged += new EventHandler(ContactNameChanged);
				c.IconChanged += new EventHandler(ContactIconChanged);
				SetFont(node, c);
				IContactList l = c as IContactList;
				if (l != null) {
					if (l.Expanded)
						node.Expand();
					l.ContactAdded   += new ContactEventHandler(ContactAdded);
					l.ContactRemoved += new ContactEventHandler(ContactRemoved);
					foreach (IContact sub in l.Contacts) {
						node.Nodes.Add(AddChild(node.Tag as ITreeTag, sub));
					}
				}
				return node;
			}
		}
		
		private void ReleaseMonitor(IContact c)
		{
			lock (c) {
				nodeHash[c] = null;
				c.NameChanged -= new EventHandler(ContactNameChanged);
				c.IconChanged -= new EventHandler(ContactIconChanged);
				IContactList l = c as IContactList;
				if (l != null) {
					l.ContactAdded   -= new ContactEventHandler(ContactAdded);
					l.ContactRemoved -= new ContactEventHandler(ContactRemoved);
					foreach (IContact sub in l.Contacts) {
						ReleaseMonitor(sub);
					}
				}
			}
		}
		
		private void DoInvoke(Delegate method, params object[] args)
		{
			this.Invoke(method, args);
		}
		
		private void ContactAdded(object sender, ContactEventArgs e)
		{
			if (this.InvokeRequired) {
				DoInvoke(new ContactEventHandler(ContactAdded), sender, e);
				return;
			}
			TreeNode parent = (TreeNode)nodeHash[sender];
			if (parent == null)
				throw new Exception("TreeNode not found");
			TreeNode newNode = AddChild(parent.Tag as ITreeTag, e.Contact);
			if (parent == rootNode) {
				treeView.Nodes.Insert(e.Index, newNode);
			} else {
				parent.Nodes.Insert(e.Index, newNode);
			}
			treeView.SelectedNode = newNode;
		}
		
		private void ContactRemoved(object sender, ContactEventArgs e)
		{
			if (this.InvokeRequired) {
				DoInvoke(new ContactEventHandler(ContactRemoved), sender, e);
				return;
			}
			TreeNode treeNode = (TreeNode)nodeHash[e.Contact];
			treeNode.Remove();
			ReleaseMonitor(e.Contact);
		}
		
		private void ContactNameChanged(object sender, EventArgs e)
		{
			if (this.InvokeRequired) {
				DoInvoke(new EventHandler(ContactNameChanged), sender, e);
				return;
			}
			IContact n = (IContact)sender;
			((TreeNode)nodeHash[n]).Text = Service.StringParser.Parse(n.Name);
		}
		
		private void ContactIconChanged(object sender, EventArgs e)
		{
			//if (!(sender is IContact))
			//	throw new ArgumentException("sender is not IContact, but " + sender.GetType().FullName);
			if (this.InvokeRequired) {
				DoInvoke(new EventHandler(ContactIconChanged), sender, e);
				return;
			}
			int image = imageListManager[((IContact)sender).Icon];
			TreeNode treeNode = (TreeNode)nodeHash[sender];
			treeNode.ImageIndex = image;
			treeNode.SelectedImageIndex = image;
		}
		#endregion
		
		#region Reformatting according to currently opened room
		IHasRoom roomOwner;
		
		public IHasRoom RoomOwner {
			get {
				return roomOwner;
			}
			set {
				if (roomOwner != value) {
					if (roomOwner != null)
						roomOwner.RoomChanged -= new EventHandler(RoomOwnerRoomChanged);
					roomOwner = value;
					if (roomOwner != null)
						roomOwner.RoomChanged += new EventHandler(RoomOwnerRoomChanged);
					RoomOwnerRoomChanged(null, null);
				}
			}
		}
		
		private void DisposeFonts()
		{
			if (roomOwner != null)
				roomOwner.RoomChanged -= new EventHandler(RoomOwnerRoomChanged);
			for (int i = 0; i < fonts.Length; ++i) {
				if (fonts[i] != null)
					fonts[i].Dispose();
			}
		}
		
		private void RoomOwnerRoomChanged(object sender, EventArgs e)
		{
			Service.Window.BeginInvoke(new MethodInvoker(RefreshFontStyles));
		}
		
		private void RefreshFontStyles()
		{
			foreach (DictionaryEntry e in nodeHash) {
				SetFont((TreeNode)e.Value, (IContact)e.Key);
			}
		}
		
		Font[] fonts = new Font[16];
		
		private void SetFont(TreeNode node, IContact contact)
		{
			IChatRoom room = (roomOwner == null) ? null : roomOwner.Room;
			FontStyle style = contact.GetFontStyle(room);
			int num = (int)style;
			if (fonts[num] == null)
				fonts[num] = new Font(treeView.Font, style);
			if (node.NodeFont != fonts[num]) {
				node.NodeFont = fonts[num];
				node.Text = node.Text; // Winforms Workaround: cause TreeNode to recalculate its width
			}
		}
		#endregion
		
		#region Expand / Collapse handler
		private void TreeViewAfterExpand(object sender, TreeViewEventArgs e)
		{
			IContactList list = ((ContactWrapperTag)e.Node.Tag).Contact as IContactList;
			if (list != null)
				list.Expanded = true;
		}
		
		private void TreeViewAfterCollapse(object sender, TreeViewEventArgs e)
		{
			IContactList list = ((ContactWrapperTag)e.Node.Tag).Contact as IContactList;
			if (list != null)
				list.Expanded = false;
		}
		#endregion
		
		#region MultiSelect
		public bool MultiSelect {
			get {
				return false;
				//return treeView.MultiSelect;
			}
			set {
				if (value)
					throw new NotSupportedException("Multiselect is not supported");
				/*if (treeView.MultiSelect == value) return;
				treeView.MultiSelect = value;
				OnMultiSelectChanged(EventArgs.Empty);*/
			}
		}
		
		public event EventHandler MultiSelectChanged;
		
		protected virtual void OnMultiSelectChanged(EventArgs e)
		{
			if (MultiSelectChanged != null) {
				MultiSelectChanged(this, e);
			}
		}
		#endregion
		
		#region Control property forwarding (background, font)
		protected override void OnBackColorChanged(EventArgs e)
		{
			treeView.BackColor = this.BackColor;
			base.OnBackColorChanged(e);
		}
		
		protected override void OnBackgroundImageChanged(EventArgs e)
		{
			treeView.BackgroundImage = this.BackgroundImage;
			base.OnBackgroundImageChanged(e);
		}
		
		protected override void OnFontChanged(EventArgs e)
		{
			treeView.Font = this.Font;
			base.OnFontChanged(e);
		}
		
		protected override void OnForeColorChanged(EventArgs e)
		{
			treeView.ForeColor = this.ForeColor;
			base.OnForeColorChanged(e);
		}
		#endregion
		
		#region Mouse events
		private void TreeViewDoubleClick(object sender, EventArgs e)
		{
			((ContactWrapperTag)treeView.SelectedNode.Tag).Contact.DoubleClick();
		}
		
		private void TreeViewMouseDown(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left && e.Clicks < 2) {
				TreeNode node = treeView.GetNodeAt(e.X, e.Y);
				if (e.X < 16 && node != null && node.Parent == null) {
					if (node.IsExpanded)
						node.Collapse();
					else
						node.Expand();
				}
			}
		}
		
		private void TreeViewMouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Right) {
				TreeNode node = treeView.GetNodeAt(e.X, e.Y);
				IContact contact;
				if (node == null)
					contact = root;
				else
					contact = ((ContactWrapperTag)node.Tag).Contact;
				Service.Menu.ShowContextMenu(contact.Owner, contact.ContextMenuTreePath, treeView, e.X, e.Y);
			}
		}
		#endregion
	}
}
