// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using Grunwald.Gui.Controls;

namespace DGNetwork.Contacts
{
	/// <summary>
	/// A wrapper to class for the RearrangeTreeView that makes the contact tree
	/// rearrangeable.
	/// </summary>
	public class ContactWrapperTag : ITreeTag
	{
		#region ITreeTag implementation
		ITreeTag parent;
		IContact contact;
		
		/// <summary>
		/// Creates a new ContactWrapperTag object for the specified contact
		/// </summary>
		public ContactWrapperTag(ITreeTag parent, IContact contact)
		{
			if (contact == null)
				throw new ArgumentNullException("contact");
			this.contact = contact;
			this.parent = parent;
		}
		
		/// <summary>
		/// Gets the parent page of the wrapped page.
		/// </summary>
		public ITreeTag Parent {
			get {
				return parent;
			}
			set {
				parent = value;
			}
		}
		
		public IContact Contact {
			get {
				return contact;
			}
		}
		
		/// <summary>
		/// Gets if this node can be dragged around.
		/// </summary>
		public bool CanDrag {
			get {
				return parent != null;
			}
		}
		
		/// <summary>
		/// Gets if this tree node can accept the specified node as a child.
		/// </summary>
		public bool AcceptChild(ITreeTag node, int index)
		{
			IContactList list = contact as IContactList;
			ContactWrapperTag tag = node as ContactWrapperTag;
			if (list == null || tag == null)
				return false;
			else
				return list.IsValidChild(tag.Contact) && tag.Contact.IsValidParent(list);
		}
		
		/// <summary>
		/// Adds a child to this tree node.
		/// </summary>
		public void AddChild(ITreeTag node, int index)
		{
			IContact child = ((ContactWrapperTag)node).Contact;
			((IContactList)contact).Contacts.Insert(index, child);
		}
		
		/// <summary>
		/// Moves a child inside this tree node.
		/// </summary>
		public void MoveChild(ITreeTag node, int index)
		{
			IContact child = ((ContactWrapperTag)node).Contact;
			((IContactList)contact).Contacts.Move(child, index);
		}
		
		/// <summary>
		/// Removes a child from this tree node.
		/// </summary>
		public void RemoveChild(ITreeTag node)
		{
			IContact child = ((ContactWrapperTag)node).Contact;
			((IContactList)contact).Contacts.Remove(child);
		}
		
		/// <summary>
		/// Gets a copy of this object.
		/// </summary>
		public object Clone()
		{
			return new ContactWrapperTag(null, (IContact)contact.Clone());
		}
		
		/// <summary>
		/// Gets if the object can be cloned.
		/// </summary>
		public bool CanClone {
			get {
				return contact.CanClone;
			}
		}
		
		/// <summary>
		/// Gets if this node can be delete using the
		/// <see cref="RearrangeTreeView.Delete">Delete method</see>.
		/// </summary>
		public bool CanDelete {
			get {
				return true;
			}
		}
		
		/// <summary>
		/// Deletes this tree node.<br/>
		/// This method is responsible for calling <see cref="RemoveChild"/> on its
		/// parent!
		/// </summary>
		public void Delete()
		{
			if (!CanDelete)
				throw new InvalidOperationException("delete not allowed");
			parent.RemoveChild(this);
		}
		#endregion
	}
}
