// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Xml;
using DGNetwork.Gui.Menu;
using DGNetwork.Services;
using ICSharpCode.Core.Properties;

namespace DGNetwork.Contacts
{
	/// <summary>
	/// An unknown contact is used when the class of the contact could not be found.
	/// The unknown contact supports saving the xml element back again so the contact
	/// is not lost when DG Network is started without a certain plugin one time.
	/// </summary>
	public sealed class UnknownContact : IContact, IXmlConvertable
	{
		XmlElement element;
		
		public object FromXmlElement(XmlElement element)
		{
			this.element = element;
			return this;
		}
		
		public XmlElement ToXmlElement(XmlDocument doc)
		{
			return (XmlElement)doc.ImportNode(element, true);
		}
		
		public string Name {
			get {
				return "${res:Contacts.UnknownContact}";
			}
			set {
				throw new NotSupportedException();
			}
		}
		
		public event EventHandler NameChanged;
		
		public string Icon {
			get {
				return null;
			}
			set {
				throw new NotSupportedException();
			}
		}
		public event EventHandler IconChanged;
		
		private void OnIconChanged(EventArgs e)
		{
			if (IconChanged != null) {
				IconChanged(this, e);
			}
		}
		
		private void OnNameChanged(EventArgs e)
		{
			if (NameChanged != null) {
				NameChanged(this, e);
			}
		}
		
		public void Update()
		{
		}
		
		public void DoubleClick()
		{
			Service.Message.ShowMessageFormatted("${res:Contacts.UnknownContact}", "${res:Contacts.UnknownContactText}", new string[] {element.Name});
		}
		
		/// <summary>
		/// Gets if the contact can be cloned.
		/// An unknown contact can NEVER be cloned, this method always returns false.
		/// </summary>
		public bool CanClone {
			get {
				return false;
			}
		}
		
		object ICloneable.Clone()
		{
			throw new NotSupportedException();
		}
		
		object IItemWithContextMenu.Owner {
			get {
				return this;
			}
		}
		
		public string ContextMenuTreePath {
			get {
				return null;
			}
		}
		
		/// <summary>
		/// Gets if the specified contact list can be a parent of this contact.
		/// </summary>
		public bool IsValidParent(IContactList parent)
		{
			return true;
		}
		
		/// <summary>
		/// Gets the font style of this contact when the currently opened room is <paramref name="currentRoom"/>.
		/// </summary>
		public System.Drawing.FontStyle GetFontStyle(DGNetwork.Chat.IChatRoom currentRoom)
		{
			return System.Drawing.FontStyle.Italic;
		}
	}
}
