// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using DGNetwork.Services;
using Grunwald.P2PLib;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;
using Grunwald.P2PLib.DGNetwork;

namespace DGNetwork.DGN
{
	/// <summary>
	/// This class helps building direct TCP-IP connections to other dg network users.
	/// </summary>
	public sealed class Connector : IRequestHandler, IDisposable
	{
		// error codes: 0 + hostname as value #0 = Please connect to me
		//              0 without values         = I can't accept incoming connections
		//              1                        = I'll make a push connection
		
		
		public static void MessageHandler(Network n, PacketEventArgs e)
		{
			IPEndPoint endPoint = Service.P2P.Server.InternetEndPoint;
			if (!Service.P2P.ServerAbilityTester.LastResult)
				endPoint = null;
			// when push is not possible: don't push
			// when push is necessary:    do push
			// otherwise, 50/50 chance push/don't push
			if (e.Packet.Values.Count > 0 && (endPoint != null || Helper.Random.Next(2) == 1)) {
				// I'll push
				e.Answer.ErrorCode = 1;
				BeginPushConnection((string)e.Packet.Values.GetValue(0, typeof(string)));
			} else {
				if (endPoint != null)
					e.Answer.Values.AddValue(endPoint.ToString());
			}
		}
		
		delegate void void_string(string url);
		
		static void BeginPushConnection(string url)
		{
			new void_string(PushConnection).BeginInvoke(url, null, null);
		}
		
		static void PushConnection(string url)
		{
			try {
				Uri uri = new Uri(url);
				EndPoint endPoint = new IPEndPoint(IPAddress.Parse(uri.Host), uri.Port);
				IConnection conn;
				try {
					conn = ConnectionFactory.DefaultFactory.BuildConnection(endPoint);
				} catch (BuildConnectionException) {
					return;
				}
				if (conn == null) {
					Service.Message.ShowError("ConnectionFactory returned null");
					return;
				}
				StringBuilder b = new StringBuilder();
				b.AppendFormat("GET {0} HTTP/1.1\r\n", uri.PathAndQuery);
				b.AppendFormat("Host: {0}\r\n", endPoint);
				b.Append(Service.Parse("User-Agent: DG Network ${exe:ProductVersion}${AppBuildName}\r\n"));
				try {
					conn.Writer.WriteLine(b.ToString());
					conn.Writer.Flush();
				} catch (System.IO.IOException) {
					return;
				}
				Service.P2P.Server.AcceptRequests(conn);
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		/// <summary>
		/// Deactivates the connector. You need to call this method to remove the listening
		/// handler from the server module!
		/// </summary>
		public void Dispose()
		{
			Service.P2P.Server.RequestHandlers.Remove(this);
		}
		
		string code;
		
		/// <summary>
		/// Creates a new Connector.
		/// </summary>
		public Connector()
		{
			code = "/system/connector/push/" + Helper.Random.Next().ToString("X");
			Service.P2P.Server.RequestHandlers.Add(this);
		}
		
		/// <summary>
		/// Starts building the connection.
		/// </summary>
		public void Connect(string networkIdentifier, int userid)
		{
			Connect(Service.DGN.GetNetwork(networkIdentifier), userid);
		}
		
		/// <summary>
		/// Starts building the connection.
		/// In case of success the Connected event is raised.
		/// In case of error (timeout etc.) nothing happens. (TODO: Fire event on error)
		/// </summary>
		public void Connect(Network network, int userid)
		{
			if (network == null)
				throw new ArgumentNullException("network");
			Packet p = new Packet();
			p.Target = "Connector";
			p.Header.Source = network.UserID;
			p.Header.Destination = new int[] { userid };
			IPEndPoint endPoint = Service.P2P.Server.InternetEndPoint;
			if (!Service.P2P.ServerAbilityTester.LastResult)
				endPoint = null;
			if (endPoint != null) {
				p.Values.AddValue("http://" + endPoint.ToString() + code);
			}
			SendResult res = network.Send(p);
			res.CallOnAnswer(new AnswerEventHandler(ReceiveAnswer));
		}
		
		/// <summary>
		/// Is called when the connection can't be built because nobody can accept
		/// incoming connections. TODO: Convert this into a general failed event.
		/// </summary>
		public event EventHandler NoServer;
		
		/// <summary>
		/// Is called when instead of an incoming connection (using PUSH techniques)
		/// a plain old outgoing connection is created.
		/// You can use this method to show progress to the user, but remember
		/// it isn't called in all cases (not with PUSH connections).
		/// </summary>
		public event EventHandler BuildingConnection;
		
		/// <summary>
		/// Is called when the connection has been successfully created.
		/// </summary>
		public event ConnectedEventHandler Connected;
		
		private void ReceiveAnswer(object sender, AnswerEventArgs e)
		{
			Answer p = e.Answer;
			if (p.ErrorCode == 1) {
				// the other will push
				return;
			}
			if (p.ErrorCode != 0)
				return; // TODO: Raise error event
			if (p.Values.Count == 0) {
				if (NoServer != null)
					NoServer(this, EventArgs.Empty);
			} else {
				if (BuildingConnection != null)
					BuildingConnection(this, EventArgs.Empty);
				string endPoint = (string)p.Values.GetValue(0, typeof(string));
				BeginBuildConnection(StringToEndPoint(endPoint));
			}
		}
		
		private delegate void void_EndPoint(EndPoint endPoint);
		
		private void BeginBuildConnection(EndPoint endPoint)
		{
			new void_EndPoint(BuildConnection).BeginInvoke(endPoint, null, null);
		}
		
		private void BuildConnection(EndPoint endPoint)
		{
			try {
				IConnection conn = ConnectionFactory.DefaultFactory.BuildConnection(endPoint);
				if (conn == null) {
					Service.Message.ShowError("ConnectionFactory returned null");
				} else {
					if (Connected != null)
						Connected(this, new ConnectedEventArgs(conn));
				}
			} catch (BuildConnectionException) {
				// TODO raise failed event
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		public static EndPoint StringToEndPoint(string endPoint)
		{
			int pos = endPoint.IndexOf(':');
			int port = int.Parse(endPoint.Substring(pos + 1, endPoint.Length - pos - 1));
			string host = endPoint.Substring(0, pos);
			return new IPEndPoint(IPAddress.Parse(host), port);
		}
		
		public int GetPriority(HttpRequest request)
		{
			if (request.File == code)
				return 10000;
			else
				return int.MinValue;
		}
		
		public void Handle(HttpRequest request)
		{
			try {
				if (Connected != null)
					Connected(this, new ConnectedEventArgs(request.Connection));
			} finally {
				request.Connection.Close();
			}
		}
	}
	
	#region ConnectedEventHandler/ConnectedEventArgs
	/// <summary>
	/// Delegate for the Connector.Connected event.
	/// </summary>
	public delegate void ConnectedEventHandler(object sender, ConnectedEventArgs e);
	
	/// <summary>
	/// Contains the arguments for the Connector.Connected event.
	/// </summary>
	public class ConnectedEventArgs : EventArgs
	{
		IConnection connection;
		
		/// <summary>
		/// The newly built connection.
		/// </summary>
		public IConnection Connection {
			get {
				return connection;
			}
		}
		
		/// <summary>
		/// Creates a new ConnectedEventArgs object.
		/// </summary>
		public ConnectedEventArgs(IConnection connection)
		{
			this.connection = connection;
		}
	}
	#endregion
}
