// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Threading;

using Grunwald.P2PLib;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.DGNetwork;
using Grunwald.Gui.Settings;

using DGNetwork.AddIns;
using DGNetwork.AddIns.Codons;
using DGNetwork.Chat;
using DGNetwork.Gui;
using DGNetwork.Services;
using DGNetwork.Settings;

using ICSharpCode.Core.Services;

namespace DGNetwork.DGN
{
	public class DGNService : AbstractService
	{
		Hashtable networks = new Hashtable();
		DgnRequestHandler requestHandler = new DgnRequestHandler();
		Identity ident;
		ReceivingUser user;
		Timer timer;
		PacketReceiverCollection packetReceivers = new PacketReceiverCollection();
		
		bool initialized = false;
		
		#region Properties
		public bool Initialized {
			get {
				return initialized;
			}
		}
		
		public Network GetNetwork(string identifier)
		{
			return (Network)networks[identifier];
		}
		
		public ICollection Networks {
			get {
				return networks.Values;
			}
		}
		
		public Identity Identity {
			get {
				return ident;
			}
		}
		
		public DgnRequestHandler RequestHandler {
			get {
				return requestHandler;
			}
		}
		
		public ReceivingUser ReceivingUser {
			get {
				return user;
			}
		}
		#endregion
		
		#region Register/Remove Network
		public void RegisterNetwork(Network n)
		{
			if (!initialized)
				throw new InvalidOperationException("Service is not initialized");
			if (networks[n.Identifier] != null)
				throw new ArgumentException("The network is already registered.");
			networks[n.Identifier] = n;
			n.Receive += new PacketEventHandler(NetworkReceive);
			n.Init();
		}
		
		public void RemoveNetwork(Network n)
		{
			n.Shutdown();
			if (networks[n.Identifier] == null)
				throw new ArgumentException("The network is not registered.");
			networks.Remove(n.Identifier);
		}
		#endregion
		
		#region Initialize service
		protected override void OnInitialize(EventArgs e)
		{
			packetReceivers.Add("GetStatus", new PacketReceiver(GetStatus));
			packetReceivers.Add("SE", new PacketReceiver(StatusExchange));
			packetReceivers.Add("Msg", new PacketReceiver(ReceiveMessage));
			packetReceivers.Add("Connector", new PacketReceiver(Connector.MessageHandler));
			packetReceivers.Add("MUCInvitation", new PacketReceiver(MUC.MUCService.ReceiveInvitation));
			packetReceivers.Add("MUC", new PacketReceiver(MUC.MUCService.ReceivePacket));
			initialized = true;
			NetClient.UserAgent = Service.StringParser.Parse("DG Network ${exe:ProductVersion}");
			ident = new Identity(1024);
			user = new ReceivingUser(ident);
			requestHandler.ReceivingUsers.Add(user);
			Service.P2P.Server.RequestHandlers.Add(requestHandler);
			ISettingPage page = Service.Setting.GetPage("NetworkPage");
			foreach (ISettingPage subpage in page.SubPages) {
				AbstractNetworkPage netpage = subpage as AbstractNetworkPage;
				try {
					if (netpage != null)
						netpage.CreateNetwork();
				} catch (Exception ex) {
					Service.Message.ShowError(ex);
				}
			}
			base.OnInitialize(e);
			Service.Chat.StatusChanged += new StatusChangedEventHandler(ChatStatusChanged);
			
			timer = new Timer(new TimerCallback(TimerTick), null, 5000, 10000);
		}
		#endregion
		
		#region Unload service
		protected override void OnUnload(EventArgs e)
		{
			timer.Dispose();
			foreach (Network n in this.Networks) {
				n.Shutdown();
			}
			base.OnUnload(e);
		}
		#endregion
		
		#region Status management
		private void GetStatus(Network n, PacketEventArgs e)
		{
			e.Answer.Values.AddValue(StatusRequest.GetSendStatus(Service.Chat.Status));
		}
		
		private ulong GetCode(Network n, int id)
		{
			unchecked {
				return ((ulong)id << 32) + (ulong)n.GetHashCode();
			}
		}
		
		private void StatusExchange(Network n, PacketEventArgs e)
		{
			Packet p = e.Packet;
			StatusRequest r = GetDGNetworkUser(n, p.Header.Source).StatusRequest;
			r.ReceiveStatus(e);
		}
		
		public void RegisterForStatusExchange(Network network, int userid)
		{
			StatusRequest r = GetDGNetworkUser(network, userid).StatusRequest;
			r.GettingOtherStatus = true;
		}
		
		public void DeregisterFromStatusExchange(Network network, int userid)
		{
			StatusRequest r = GetDGNetworkUser(network, userid).StatusRequest;
			r.GettingOtherStatus = false;
		}
		
		public string GetStatus(Network network, int userid)
		{
			return GetDGNetworkUser(network, userid).Status;
		}
		
		private void ChatStatusChanged(object sender, StatusChangedEventArgs e)
		{
			if (e.OldStatus == "Offline") {
				foreach (Network n in this.Networks) {
					n.Connected = true;
				}
			}
			if (e.NewStatus == "Offline") {
				foreach (Network n in this.Networks) {
					n.Connected = false;
				}
			} else {
				SendStatus();
			}
		}
		
		private void SendStatus()
		{
			lock (chatUsers) {
				foreach (DGNetworkUser r in chatUsers.Values) {
					if (r.statusRequest != null)
						r.statusRequest.SendStatus();
				}
			}
		}
		#endregion
		
		#region Receive Message
		private void NetworkReceive(object sender, PacketEventArgs e)
		{
			if (!packetReceivers.Call((Network)sender, e)) {
				// no receiver found
				e.Answer.ErrorCode = Answer.ErrorCodeTargetNotFound;
			}
		}
		
		Hashtable chatUsers = new Hashtable();
		
		public DGNetworkUser GetDGNetworkUser(Network n, int userid)
		{
			ulong code = GetCode(n, userid);
			DGNetworkUser user;
			lock (chatUsers) {
				user = (DGNetworkUser)chatUsers[code];
				if (user == null) {
					user = new DGNetworkUser(n, userid);
					chatUsers[code] = user;
				}
			}
			return user;
		}
		
		private void ReceiveMessage(Network n, PacketEventArgs e)
		{
			DGNetworkUser user = GetDGNetworkUser(n, e.Packet.Header.Source);
			Message message = user.GetMessageFromPacket(e.Packet);
			if (message == null) {
				e.Answer.ErrorCode = Answer.ErrorCodeInvalidRequest;
				return;
			}
			user.PrivateChatRoom.ReceiveMessage(message);
		}
		#endregion
		
		private void TimerTick(object state)
		{
			foreach (Network n in this.Networks) {
				n.Work();
			}
			SendStatus();
		}
	}
}
