// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using DGNetwork.Chat;
using DGNetwork.FileTransfer;
using DGNetwork.Services;
using Grunwald.P2PLib.DGNetwork;

namespace DGNetwork.DGN
{
	/// <summary>
	/// Chat user implementation for users on the DGNetwork.
	/// </summary>
	public class DGNetworkUser : AbstractUser
	{
		Network network;
		int userID;
		
		/// <summary>
		/// Creates a new DGNetworkUser object. WARNING: normally you should use the
		/// GetDGNetworkUser() method on the DGNService instead to get the already
		/// existing user.
		/// </summary>
		public DGNetworkUser(Network network, int userID)
		{
			this.network = network;
			this.userID = userID;
		}
		
		public Network Network {
			get {
				return network;
			}
		}
		
		public int UserID {
			get {
				return userID;
			}
		}
		
		internal StatusRequest statusRequest;
		
		public StatusRequest StatusRequest {
			get {
				if (statusRequest == null) {
					statusRequest = new StatusRequest(network, userID);
					statusRequest.HisStatusChanged += new StatusChangedEventHandler(OnStatusChanged);
				}
				return statusRequest;
			}
		}
		
		public string Status {
			get {
				if (statusRequest == null)
					return "Unknown";
				else
					return statusRequest.HisStatus;
			}
		}
		
		private void OnStatusChanged(object sender, StatusChangedEventArgs e)
		{
			this.Icon = "Icons.Status." + e.NewStatus;
		}
		
		/// <summary>
		/// Sets the packet's destination to this user and sends the packet to the network.
		/// </summary>
		public SendResult SendPacket(Packet packet)
		{
			packet.Header.Destination = new int[] { this.userID };
			return network.Send(packet);
		}
		
		public override IMessageSendResult SendMessage(Message message)
		{
			Packet p = MessageToPacket(message);
			
			DGNSendResult result = new DGNSendResult(message);
			SendResult res = SendPacket(p);
			res.CallOnAnswer(new AnswerEventHandler(result.ReceiveAnswer));
			return result;
		}
		
		public Message GetMessageFromPacket(Packet packet)
		{
			Message message = PacketToMessage(packet);
			if (message != null)
				SetMessageExtraData(message);
			return message;
		}
		
		/// <summary>
		/// Sets the extra data of a message received from this user.
		/// </summary>
		public void SetMessageExtraData(Message message)
		{
			message.ExtraData.Add("userid", userID);
			message.ExtraData.Add("network", network);
			if (message.Nickname != null)
				this.Name = message.Nickname;
			else
				message.Nickname = this.Name;
		}
		
		/// <summary>
		/// Converts the message into a dg network packet.
		/// The packet header is NOT set, you still have to set source and destination!
		/// </summary>
		public static Packet MessageToPacket(Message message)
		{
			Packet p = new Packet();
			p.Target = "Msg";
			MessageToPacket(message, p.Values, p.NamedValues);
			return p;
		}
		
		/// <summary>
		/// Converts the message into a dg network packet.
		/// The packet data is written to the IndexedValueList and NamedValueList.
		/// </summary>
		public static void MessageToPacket(Message message, IndexedValueList values, NamedValueList namedValues)
		{
			values.AddValue(message.Html);
			values.AddValue(message.Nickname);
			foreach (DictionaryEntry e in message.ExtraData) {
				string key = (string)e.Key;
				if (key.StartsWith("file:")) {
					FileData file = (FileData)e.Value;
					namedValues.AddValue(key, file.Data);
					namedValues.AddValue(key.Substring(0, 4) + "name" + key.Substring(4), file.Name);
				}
			}
		}
		
		public static Message PacketToMessage(Packet packet)
		{
			return PacketToMessage(packet.Values, packet.NamedValues);
		}
		
		public static Message PacketToMessage(IndexedValueList values, NamedValueList namedValues)
		{
			if (values.Count == 0)
				return null;
			Message message = new Message(false);
			message.Html = (string)values.GetValue(0, typeof(string));
			if (values.Count > 1)
				message.Nickname = (string)values.GetValue(1, typeof(string));
			foreach (DictionaryEntry e in namedValues) {
				string key = (string)e.Key;
				if (key.StartsWith("file:")) {
					string name = (string)namedValues.GetValue(key.Substring(0, 4) + "name" + key.Substring(4), typeof(string));
					message.ExtraData.Add(key, new FileData((byte[])e.Value, name));
				}
			}
			return message;
		}
		
		private class DGNSendResult : DefaultMessageSendResult
		{
			public DGNSendResult(Message message) : base(message)
			{
				
			}
			
			public void ReceiveAnswer(object sender, AnswerEventArgs e)
			{
				if (e.Answer.ErrorCode == 0)
					this.SetResult(MessageSendStatus.Success);
				else
					this.SetResult(MessageSendStatus.TargetRejectedMessage);
			}
		}
	}
}
