// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Net;
using System.Xml;
using Grunwald.P2PLib;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;
using DGNetwork.Services;

namespace DGNetwork.DGN
{
	/// <summary>
	/// The default request handler of the DG Network server.
	/// </summary>
	public class DefaultRequestHandler : IRequestHandler
	{
		public void Handle(HttpRequest request)
		{
			string f = request.File.ToLower();
			if (f == "/favicon.ico") {
				using (Stream r = Service.Resource.GetResourceStream("App.ico", "DGNetwork.StartUp")) {
					FileUpload upload = new FileUpload(request);
					upload.Headers["Content-Type"] = "image/x-icon";
					upload.Send(r);
				}
			} else if (f == "/dgn/getnetworkinformation") {
				using (StringWriter w = new StringWriter()) {
					XmlTextWriter writer = new XmlTextWriter(w);
					writer.Formatting = Formatting.Indented;
					writer.WriteStartDocument();
					writer.WriteStartElement("networkInformation");
					string requestedID = request.Request["id"];
					if (requestedID != null && requestedID.Length > 0) {
						WriteNetwork(writer, Service.DGN.GetNetwork(requestedID));
					} else {
						/* Don't publish this information so openly
						foreach (Network network in Service.DGN.Networks) {
							WriteNetwork(writer, network);
						}
						*/
						writer.WriteElementString("error", "You must specify the network id you want information about.");
					}
					writer.WriteEndDocument(); // close all open elements
					SendText(request, w.ToString(), "text/xml");
					writer.Close();
				}
			} else if (f == "/dgn/showmainform") {
				IPEndPoint endPoint = request.Connection.RemoteEndPoint as IPEndPoint;
				if (endPoint == null || !IPAddress.Loopback.Equals(endPoint.Address))
					throw new HttpException(HttpStatusCode.Forbidden);
				HttpResponse response = new HttpResponse(request);
				response.ContentLength = 0;
				response.SendHeaders(request.Connection.Writer);
				Service.Window.ShowMainWindow();
			} else {
				string text = Service.StringParser.Parse("${res:Server.DefaultIndexPage}");
				SendText(request, text, "text/html");
			}
		}
		
		private void WriteNetwork(XmlWriter writer, Network network)
		{
			if (network == null)
				return;
			writer.WriteStartElement("network");
			writer.WriteAttributeString("id", network.Identifier);
			writer.WriteStartElement("connectKey");
			writer.WriteAttributeString("needed", XmlConvert.ToString(network.HasConnectKey));
			writer.WriteEndElement();
			writer.WriteStartElement("users");
			int userID = network.UserID;
			if (userID != 0) {
				writer.WriteStartElement("user");
				writer.WriteAttributeString("id", XmlConvert.ToString(userID));
				writer.WriteElementString("name", Service.Chat.Nickname);
				writer.WriteEndElement();
			}
			writer.WriteEndElement();
			writer.WriteEndElement();
		}
		
		private void SendText(HttpRequest request, string text, string contentType)
		{
			HttpResponse response = new HttpResponse(request);
			response.ContentLength = text.Length;
			response.ContentType = contentType;
			TextWriter w = request.Connection.Writer;
			response.SendHeaders(w, false);
			w.Write(text);
			w.Flush();
		}
		
		public int GetPriority(HttpRequest request)
		{
			string f = request.File.ToLower();
			if (f == "/" || f == "/index.htm" || f == "/index.html" || f == "/defaultindex.html")
				return -1;
			else if (f == "/favicon.ico")
				return 0;
			else if (f == "/dgn/getnetworkinformation")
				return 5000;
			else if (f == "/dgn/showmainform")
				return 0;
			else
				return int.MinValue;
		}
	}
}
