// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Security.Cryptography;
using DGNetwork.Services;
using DGNetwork.Chat;
using Grunwald.P2PLib.DGNetwork;

namespace DGNetwork.DGN.MUC
{
	/// <summary>
	/// A chat room for multi user chat via the DG Network protocol.
	/// </summary>
	public class MUCRoom : AbstractChatRoom
	{
		ArrayList users = new ArrayList();
		ArrayList usersWrapper;
		Network network;
		RijndaelManaged aes = new RijndaelManaged();
		byte[] keyHash;
		DGNetworkUser ownUser;
		
		#region Constructors
		private void Init(Network network)
		{
			this.network = network;
			ownUser = Service.DGN.GetDGNetworkUser(network, network.UserID);
			ownUser.Name = Service.Chat.Nickname;
			users.Add(ownUser);
			aes.IV = Helper.MD5Bytes(aes.Key);
			this.Title = "${res:DGN.MUC.RoomTitle}";
			usersWrapper = ArrayList.Synchronized(ArrayList.ReadOnly(users));
			this.UserListVisible = true;
		}
		
		public MUCRoom(Network network)
		{
			aes.GenerateKey();
			Init(network);
		}
		
		public MUCRoom(Network network, byte[] key)
		{
			aes.Key = key;
			Init(network);
		}
		#endregion
		
		public override bool DisposeOnLeave {
			get {
				return true;
			}
		}
		
		protected override void OnDisposed(EventArgs e)
		{
			SendLeave();
			base.OnDisposed(e);
		}
		
		public byte[] KeyHash {
			get {
				if (keyHash == null)
					keyHash = Helper.SHA1Bytes(aes.Key);
				return keyHash;
			}
		}
		
		#region Invitation
		public void Invite(DGNetworkUser user)
		{
			if (this.Users.Contains(user)) {
				Service.Message.ShowMessageFormatted("${res:DGN.MUC.AlreadyInChat}", new string[] { user.Name });
			} else if (user.Network != network) {
				Service.Message.ShowError("${res:DGN.MUC.DifferentNetwork}");
			} else {
				Packet p = new Packet();
				p.Target = "MUCInvitation";
				p.Values.AddValue(aes.Key);
				p.Values.AddValue(Service.Parse(this.Title));
				p.Values.AddValue(MUCService.UserListToArray(users));
				p.Values.AddValue(ownUser.Name);
				user.SendPacket(p);
				
				Message inviteMessage = new Message(false);
				inviteMessage.IsAction = true;
				inviteMessage.Text = Service.Parse("${res:Chat.IsInvited}");
				inviteMessage.Nickname = user.Name;
				inviteMessage.NotifyUser = false;
				this.ReceiveMessage(inviteMessage);
			}
		}
		#endregion
		
		#region UserManagement
		public override ArrayList Users {
			get {
				return usersWrapper;
			}
		}
		
		public DGNetworkUser GetContainedUser(int userID)
		{
			foreach (DGNetworkUser user in this.Users) {
				if (user.UserID == userID)
					return user;
			}
			return null;
		}
		
		public bool ContainsUser(int userID)
		{
			return GetContainedUser(userID) != null;
		}
		
		int[] GetUnknownUsers(int[] userList)
		{
			if (userList == null)
				throw new ArgumentNullException("userList");
			userList = (int[])userList.Clone(); // clone because we don't want to change the original array
			int count = 0;
			for (int i = 0; i < userList.Length; ++i) {
				if (userList[i] == 0)
					continue;
				if (ContainsUser(userList[i]))
					userList[i] = 0;
				else
					count += 1;
			}
			if (count == 0)
				return null;
			if (count == userList.Length)
				return userList;
			int[] result = new int[count];
			int pos = 0;
			for (int i = 0; i < userList.Length; ++i) {
				if (userList[i] != 0) {
					result[pos++] = userList[i];
				}
			}
			return result;
		}
		#endregion
		
		#region Receive MUC Packet
		public void ReceivePacket(int sourceUser, MUCPacketType packetType, byte[] data, Answer answer)
		{
			IndexedValueList values = new IndexedValueList();
			NamedValueList namedValues = new NamedValueList();
			try {
				using (MemoryStream ms = new MemoryStream(data, false)) {
					using (ICryptoTransform decryptor = aes.CreateDecryptor()) {
						using (CryptoStream s = new CryptoStream(ms, decryptor, CryptoStreamMode.Read)) {
							values.Read(s);
							namedValues.Read(s);
						}
					}
				}
			} catch (CryptographicException) {
				answer.ErrorCode = (byte)MUCResultCode.DecryptionError;
				return;
			}
			PreProcessExtras(sourceUser, namedValues);
			switch (packetType) {
				case MUCPacketType.Joining:
					if (answer != null)
						answer.NamedValues.AddValue("MUC_Userlist", MUCService.UserListToArray(this.Users));
					UserJoined(sourceUser);
					break;
				case MUCPacketType.Leaving:
					UserLeft(sourceUser);
					break;
				case MUCPacketType.Message:
					ReceiveMessage(sourceUser, values, namedValues);
					break;
				default:
					if (answer != null)
						answer.ErrorCode = (byte)MUCResultCode.UnknownMUCType;
					break;
			}
			PostProcessExtras(sourceUser, namedValues);
		}
		
		void PreProcessExtras(int sourceID, NamedValueList namedValues)
		{
			string name = (string)namedValues.GetValue("MUC_Nickname", typeof(string));
			if (name != null) {
				DGNetworkUser user = Service.DGN.GetDGNetworkUser(network, sourceID);
				if (user != null)
					user.Name = name;
			}
		}
		
		void PostProcessExtras(int sourceID, NamedValueList namedValues)
		{
			int[] userList = (int[])namedValues.GetValue("MUC_Userlist", typeof(int[]));
			if (userList != null)
				SendJoin(userList, true);
		}
		#endregion
		
		#region Send MUC Packet
		private SendResult SendMUCPacket(int[] destination, MUCPacketType type, IndexedValueList indexedValues, NamedValueList namedValues)
		{
			Packet p = new Packet();
			p.Header.Destination = destination;
			p.Target = "MUC";
			p.Values.AddValue(this.KeyHash);
			p.Values.AddValue((byte)type);
			
			if (type != MUCPacketType.Message)
				namedValues.AddValue("MUC_Nickname", ownUser.Name);
			
			// when leaving, exclude own user id
			if (type == MUCPacketType.Leaving)
				namedValues.AddValue("MUC_Userlist", destination);
			else
				namedValues.AddValue("MUC_Userlist", MUCService.UserListToArray(this.Users));
			
			using (MemoryStream ms = new MemoryStream()) {
				using (ICryptoTransform encryptor = aes.CreateEncryptor()) {
					using (CryptoStream s = new CryptoStream(ms, encryptor, CryptoStreamMode.Write)) {
						indexedValues.Save(s);
						namedValues.Save(s);
						s.FlushFinalBlock();
						p.Values.AddValue(ms.ToArray());
					}
				}
			}
			return network.Send(p);
		}
		#endregion
		
		#region Joining
		public void InitialJoin(int[] userList)
		{
			SendJoin(userList, true);
		}
		
		bool initialJoin;
		
		private void SendJoin(int[] userList, bool initialJoin)
		{
			this.initialJoin = initialJoin;
			lock (users) {
				userList = GetUnknownUsers(userList);
			}
			if (userList == null)
				return;
			IndexedValueList values = new IndexedValueList();
			NamedValueList namedValues = new NamedValueList();
			SendResult result = SendMUCPacket(userList, MUCPacketType.Joining, values, namedValues);
			result.CallOnAnswer(new AnswerEventHandler(JoinAnswer));
		}
		
		DGNetworkUser UserJoined(int userID)
		{
			DGNetworkUser user = Service.DGN.GetDGNetworkUser(network, userID);
			if (!users.Contains(user)) {
				users.Add(user);
				this.OnUsersChanged(EventArgs.Empty);
				
				if (!initialJoin) {
					Message joinMessage = new Message(false);
					joinMessage.IsAction = true;
					joinMessage.Text = Service.Parse("${res:Chat.HasJoined}");
					joinMessage.Nickname = user.Name;
					joinMessage.NotifyUser = false;
					this.ReceiveMessage(joinMessage);
				}
			}
			return user;
		}
		
		private void JoinAnswer(object sender, AnswerEventArgs e)
		{
			Answer a = e.Answer;
			lock (users) {
				DGNetworkUser user = GetContainedUser(a.FromUser);
				if (user == ownUser)
					return;
				if (a.ErrorCode == (byte)MUCResultCode.Success) {
					PreProcessExtras(a.FromUser, a.NamedValues);
					if (user == null) {
						UserJoined(a.FromUser);
					}
					PostProcessExtras(a.FromUser, a.NamedValues);
				} else {
					if (user != null) {
						UserLeft(a.FromUser);
					}
				}
			}
		}
		#endregion
		
		#region Leaving
		void SendLeave()
		{
			int[] targets = MUCService.UserListToArray(this.Users, ownUser);
			if (targets.Length == 0)
				return;
			IndexedValueList values = new IndexedValueList();
			NamedValueList namedValues = new NamedValueList();
			SendMUCPacket(targets, MUCPacketType.Leaving, values, namedValues);
		}
		
		void UserLeft(int userid)
		{
			DGNetworkUser user = GetContainedUser(userid);
			if (user == null)
				return;
			users.Remove(user);
			this.OnUsersChanged(EventArgs.Empty);
			Message leftMessage = new Message(false);
			leftMessage.IsAction = true;
			leftMessage.Text = Service.Parse("${res:Chat.HasLeft}");
			leftMessage.Nickname = user.Name;
			leftMessage.NotifyUser = false;
			this.ReceiveMessage(leftMessage);
		}
		#endregion
		
		#region SendMessage
		public override IMessageSendResult SendMessage(Message message)
		{
			int[] targets = MUCService.UserListToArray(this.Users, ownUser);
			if (targets.Length == 0)
				return null;
			MUCSendResult res = new MUCSendResult(this, targets.Length, message);
			
			IndexedValueList values = new IndexedValueList();
			NamedValueList namedValues = new NamedValueList();
			DGNetworkUser.MessageToPacket(message, values, namedValues);
			SendResult result = SendMUCPacket(targets, MUCPacketType.Message, values, namedValues);
			
			result.CallOnAnswer(new AnswerEventHandler(res.ReceiveAnswer));
			return res;
		}
		
		private class MUCSendResult : DefaultMessageSendResult
		{
			MUCRoom room;
			int destinationCount;
			bool finished = false;
			
			public MUCSendResult(MUCRoom room, int destinationCount, Message message) : base(message)
			{
				this.room = room;
				this.destinationCount = destinationCount;
			}
			
			public void ReceiveAnswer(object sender, AnswerEventArgs e)
			{
				Answer a = e.Answer;
				if (a.ErrorCode == 0) {
					room.PreProcessExtras(a.FromUser, a.NamedValues);
					if (!finished) {
						finished = true;
						this.SetResult(MessageSendStatus.Success);
					}
					room.PostProcessExtras(a.FromUser, a.NamedValues);
				} else {
					if (a.ErrorCode == (byte)MUCResultCode.UnknownMUCHash)
						room.UserLeft(a.FromUser);
					destinationCount -= 1;
					if (!finished && destinationCount == 0) {
						finished = true;
						this.SetResult(MessageSendStatus.TargetRejectedMessage);
					}
				}
			}
		}
		#endregion
		
		#region Receive Message
		private void ReceiveMessage(int sourceUser, IndexedValueList values, NamedValueList namedValues)
		{
			DGNetworkUser user = GetContainedUser(sourceUser);
			if (user == null) {
				user = UserJoined(sourceUser);
			}
			Message message = DGNetworkUser.PacketToMessage(values, namedValues);
			user.SetMessageExtraData(message);
			message.Room = this;
			this.ReceiveMessage(message);
		}
		#endregion
	}
}
