// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using DGNetwork.Chat;
using DGNetwork.Services;
using Grunwald.P2PLib;
using Grunwald.P2PLib.DGNetwork;

namespace DGNetwork.DGN.MUC
{
	/// <summary>
	/// Static class for multi-user chat helper functions.
	/// </summary>
	public sealed class MUCService
	{
		private MUCService() {}
		
		static Hashtable rooms;
		static Hashtable invitations;
		
		static MUCService()
		{
			ByteArrayProvider p = new ByteArrayProvider();
			rooms = new Hashtable(p, p);
			invitations = new Hashtable(p, p);
		}
		
		#region ByteArrayProvider
		private class ByteArrayProvider : IHashCodeProvider, IComparer
		{
			public int GetHashCode(object obj)
			{
				byte[] d = (byte[])obj;
				int hash = 0;
				unchecked {
					for (int i = 0; i < d.Length; ++i) {
						hash ^= d[i];
						hash *= 13;
					}
				}
				return hash;
			}
			
			public int Compare(object x, object y)
			{
				byte[] a = (byte[])x;
				byte[] b = (byte[])y;
				if (a.Length != b.Length)
					return 1;
				for (int i = 0; i < a.Length; ++i) {
					if (a[i] != b[i])
						return 1;
				}
				return 0;
			}
		}
		#endregion
		
		#region Room management
		public static MUCRoom CreateRoom(DGNetworkUser user1, DGNetworkUser user2)
		{
			if (user1 == user2) {
				Service.Message.ShowError("${res:DGN.MUC.ChooseDifferentUsers}");
			} else if (user1.Network != user2.Network) {
				Service.Message.ShowError("${res:DGN.MUC.DifferentNetwork}");
			} else if (Service.Message.AskQuestionFormatted("${res:DGN.MUC.CreateRoomQuestion}", new string[] { user1.Name, user2.Name })) {
				MUCRoom room = new MUCRoom(user1.Network);
				lock (rooms) {
					RegisterRoom(room);
				}
				room.Invite(user1);
				room.Invite(user2);
				return room;
			}
			return null;
		}
		
		public static MUCRoom CreateRoom(DGNetworkUser user)
		{
			MUCRoom room = new MUCRoom(user.Network);
			lock (rooms) {
				RegisterRoom(room);
			}
			room.Invite(user);
			return room;
		}
		
		/// <summary>
		/// for calling this method, rooms must be locked.
		/// </summary>
		private static void RegisterRoom(MUCRoom room)
		{
			rooms[room.KeyHash] = room;
			room.Disposed += new EventHandler(RoomDisposed);
		}
		
		public static MUCRoom GetRoom(byte[] hash)
		{
			lock (rooms) {
				return (MUCRoom)rooms[hash];
			}
		}
		
		private static void RoomDisposed(object sender, EventArgs e)
		{
			lock (rooms) {
				rooms.Remove(((MUCRoom)sender).KeyHash);
			}
		}
		#endregion
		
		public static int[] UserListToArray(ArrayList users)
		{
			int[] data = new int[users.Count];
			for (int i = 0; i < data.Length; ++i) {
				data[i] = ((DGNetworkUser)users[i]).UserID;
			}
			return data;
		}
		
		public static int[] UserListToArray(ArrayList users, DGNetworkUser exception)
		{
			int[] data = new int[users.Count - 1];
			int pos = 0;
			for (int i = 0; i < users.Count; ++i) {
				DGNetworkUser user = (DGNetworkUser)users[i];
				if (user != exception)
					data[pos++] = user.UserID;
			}
			return data;
		}
		
		#region Invitations
		private class Invitation
		{
			internal byte[] key;
			internal string title;
			internal int[] userList;
			internal Network network;
			
			internal Invitation(IndexedValueList v)
			{
				key = v.GetValue(0);
				title = (string)v.GetValue(1, typeof(string));
				userList = (int[])v.GetValue(2, typeof(int[]));
			}
		}
		
		public static void ReceiveInvitation(Network n, PacketEventArgs e)
		{
			if (e.Packet.Values.Count < 4) {
				if (e.Answer != null)
					e.Answer.ErrorCode = Answer.ErrorCodeInvalidRequest;
				return;
			}
			byte[] key = e.Packet.Values.GetValue(0);
			byte[] keyHash = Helper.SHA1Bytes(key);
			lock (rooms) {
				if (rooms.Contains(keyHash)) {
					if (e.Answer != null)
						e.Answer.ErrorCode = 1;
					return;
				}
			}
			Invitation invitation = new Invitation(e.Packet.Values);
			invitation.network = n;
			lock (invitations) {
				invitations[keyHash] = invitation;
			}
			DGNetworkUser sourceUser = Service.DGN.GetDGNetworkUser(n, e.Packet.Header.Source);
			Message message = new Message(false);
			message.Nickname = (string)e.Packet.Values.GetValue(3, typeof(string));
			message.IsAction = true;
			message.Html = Service.Parse("${res:DGN.MUC.GotInvitation}",
			                             new string[,] {
			                             	{"Invitation", "<" + InvitationTag.TagName + " keyhash=\"" + Convert.ToBase64String(keyHash) + "\">"},
			                             	{"Name", invitation.title}
			                             });
			sourceUser.SetMessageExtraData(message);
			message.Room = sourceUser.PrivateChatRoom;
			sourceUser.PrivateChatRoom.ReceiveMessage(message);
		}
		
		public static void AcceptInvitation(byte[] keyHash)
		{
			MUCRoom room;
			lock (rooms) {
				room = (MUCRoom)rooms[keyHash];
				if (room == null) {
					Invitation invitation;
					lock (invitations) {
						invitation = (Invitation)invitations[keyHash];
					}
					if (invitation == null) {
						Service.Message.ShowError("invitation == null");
						return;
					}
					room = new MUCRoom(invitation.network, invitation.key);
					RegisterRoom(room);
					room.InitialJoin(invitation.userList);
				}
			}
			Service.Chat.OpenChat(room);
		}
		#endregion
		
		#region Receive MUC Packet
		public static void ReceivePacket(Network n, PacketEventArgs e)
		{
			if (e.Packet.Values.Count < 3) {
				if (e.Answer != null)
					e.Answer.ErrorCode = (byte)MUCResultCode.InvalidRequest;
				return;
			}
			byte[] keyHash = e.Packet.Values.GetValue(0);
			MUCRoom room;
			lock (rooms) {
				room = (MUCRoom)rooms[keyHash];
				if (room == null) {
					if (e.Answer != null)
						e.Answer.ErrorCode = (byte)MUCResultCode.UnknownMUCHash;
					return;
				}
			}
			MUCPacketType packetType = (MUCPacketType)e.Packet.Values.GetValue(1)[0];
			room.ReceivePacket(e.Packet.Header.Source, packetType, e.Packet.Values.GetValue(2), e.Answer);
		}
		#endregion
	}
}
