// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Security.Cryptography;
using System.Text;
using System.Xml;
using DGNetwork.Services;
using Grunwald.P2PLib;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.DGNetwork;
using ICSharpCode.Core.Services;
using ICSharpCode.Core.Properties;

namespace DGNetwork.DGN
{
	public class Network
	{
		DgnP2PNetwork network;
		IPServerCollection ipServer = new IPServerCollection();
		string identifier;
		bool active;
		
		#region Properties
		/// <summary>
		/// Gets if the network is currently active.
		/// </summary>
		public bool Active {
			get {
				return active;
			}
		}
		
		public bool Connected {
			get {
				if (network == null)
					return false;
				else
					return !network.StayOffline;
			}
			set {
				if (value)
					network.Connect();
				else
					network.Disconnect();
			}
		}
		
		public string Identifier {
			get { return identifier; }
		}
		
		public IPServerCollection IpServer {
			get { return ipServer; }
		}
		
		public int UserID {
			get {
				if (network == null)
					return 0;
				else
					return network.UserID;
			}
		}
		
		public bool HasConnectKey {
			get {
				if (network == null)
					return false;
				else
					return (network.ConnectKey == null) || (network.ConnectKey.Length == 0);
			}
		}
		
		/// <summary>
		/// Gets the connections of this network.
		/// This can be null if the network is deactivated.
		/// </summary>
		public ConnectionCollection Connections {
			get {
				if (network == null)
					return null;
				return network.Connections;
			}
		}
		
		/// <summary>
		/// Gets the host cache of this network.
		/// This can be null if the network is deactivated.
		/// </summary>
		public AddressCollection HostCache {
			get {
				if (network == null)
					return null;
				return network.HostCache;
			}
		}
		#endregion
		
		#region Constructors
		public Network(string identifier)
		{
			this.identifier = identifier;
		}
		
		public Network(NetworkPage page)
		{
			this.identifier = page.Identifier;
			CreateIPServer(page.IpServer1);
			CreateIPServer(page.IpServer2);
			CreateIPServer(page.IpServer3);
		}
		#endregion
		
		internal void Work()
		{
			if (network != null)
				network.Work();
		}
		
		public void SetIPServers(params string[] urls)
		{
			ipServer.Clear();
			foreach (string url in urls)
				CreateIPServer(url);
			if (network != null) {
				network.IPServers.Clear();
				network.IPServers.AddRange(ipServer);
			}
			if (IPServersChanged != null)
				IPServersChanged(this, EventArgs.Empty);
		}
		
		public void ExecuteRefresh(IPServer server)
		{
			if (network != null)
				server.BeginRefresh(network);
		}
		
		public event EventHandler IPServersChanged;
		
		private void CreateIPServer(string url)
		{
			if (url == null) return;
			if (url.Length == 0) return;
			DgnIPServer ip = new DgnIPServer(url, Service.P2P.Server);
			ip.ReceivingUser = Service.DGN.ReceivingUser;
			ip.RequestHeaders.Add("User-Agent", Service.Parse("${AppIdentification}"));
			
			string prop = GetIPServerPropName(url);
			ip.Password = Service.Property.GetProperty(prop + ".Password", ip.Password);
			ip.UserID = Service.Property.GetProperty(prop + ".UserID", ip.UserID);
			ip.UserIDChanged += new EventHandler(IPServerUserIDChanged);
			ipServer.Add(ip);
		}
		
		private void IPServerUserIDChanged(object sender, EventArgs e)
		{
			DgnIPServer ip = (DgnIPServer)sender;
			Service.Property.SetProperty(GetIPServerPropName(ip.Url) + ".UserID", ip.UserID);
		}
		
		private string GetIPServerPropName(string url)
		{
			byte[] md5 = DGNetwork.Helper.MD5Bytes(Encoding.UTF8.GetBytes(url));
			return "IPServer." + Convert.ToBase64String(md5, 0, 15);
		}
		
		public event EventHandler Initialized;
		
		public void Init()
		{
			if (active)
				throw new InvalidOperationException();
			active = true;
			network = new DgnP2PNetwork(identifier, Service.DGN.ReceivingUser);
			network.IPServers.AddRange(ipServer);
			network.ConnectionAdded   += new ConnectionEventHandler(NetworkConnectionChanged);
			network.ConnectionRemoved += new ConnectionEventHandler(NetworkConnectionChanged);
			network.StayOfflineChanged += new EventHandler(NetworkStayOfflineChanged);
			network.Receive += new PacketEventHandler(NetworkReceive);
			if (Initialized != null)
				Initialized(this, EventArgs.Empty);
		}
		
		public void Shutdown()
		{
			if (!active)
				throw new InvalidOperationException();
			active = false;
			if (network != null) {
				network.Disconnect();
				network = null;
			}
		}
		
		public int GetDistance(int userid)
		{
			if (network == null)
				return -1;
			return network.RoutingStrategy.GetDistance(userid);
		}
		
		public SendResult Send(Packet p)
		{
			if (network == null)
				return null;
			if (p.Header.Source == 0)
				p.Header.Source = network.UserID;
			return network.SendPacket(p);
		}
		
		public DgnP2PNetwork Unwrap()
		{
			return network;
		}
		
		#region Forwarded events
		private void NetworkConnectionChanged(object sender, ConnectionEventArgs e)
		{
			if (ConnectionsChanged != null) {
				ConnectionsChanged(this, e);
			}
		}
		
		public event EventHandler ConnectionsChanged;
		
		private void NetworkStayOfflineChanged(object sender, EventArgs e)
		{
			if (ConnectedChanged != null) {
				ConnectedChanged(this, e);
			}
		}
		
		public event EventHandler ConnectedChanged;
		
		private void NetworkReceive(object sender, PacketEventArgs e)
		{
			if (Receive != null) {
				Receive(this, e);
			}
		}
		
		public event PacketEventHandler Receive;
		#endregion
	}
}
