// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using DGNetwork.Services;
using DGNetwork.Chat;
using Grunwald.P2PLib.DGNetwork;

namespace DGNetwork.DGN
{
	/// <summary>
	/// Description of StatusRequest.
	/// </summary>
	public class StatusRequest
	{
		Network network;
		int userid;
		string hisStatus = "Unknown";
		string ourStatus;
		DateTime lastReceive;
		DateTime lastSend;
		
		public StatusRequest(Network network, int userid)
		{
			this.network = network;
			this.userid = userid;
			this.gettingOtherStatus = gettingOtherStatus;
		}
		
		/// <summary>
		/// Gets the time when the own status was last sent
		/// </summary>
		public Network Network {
			get {
				return network;
			}
		}
		
		public string HisStatus {
			get {
				return hisStatus;
			}
			set {
				if (hisStatus == value)
					return;
				string oldStatus = hisStatus;
				hisStatus = value;
				if (HisStatusChanged != null)
					HisStatusChanged(this, new StatusChangedEventArgs(oldStatus, value));
			}
		}
		
		public event StatusChangedEventHandler HisStatusChanged;
		
		public string OurStatus {
			get {
				return ourStatus;
			}
		}
		
		public int UserID {
			get {
				return userid;
			}
		}
		
		bool gettingOtherStatus;
		
		/// <summary>
		/// Gets/Sets if we want the status of the other user.
		/// </summary>
		public bool GettingOtherStatus {
			get {
				return gettingOtherStatus;
			}
			set {
				if (gettingOtherStatus == value)
					return;
				gettingOtherStatus = value;
				needSend = true;
			}
		}
		
		bool sendingOwnStatus;
		
		/// <summary>
		/// Gets/Sets if the other user wants our status.
		/// </summary>
		public bool SendingOwnStatus {
			get {
				return sendingOwnStatus;
			}
			set {
				if (sendingOwnStatus == value)
					return;
				sendingOwnStatus = value;
				needSend = true;
			}
		}
		
		bool needSend;
		
		public void SendStatus()
		{
			if (needSend) {
				DoSend();
				return;
			}
			if (!sendingOwnStatus && !gettingOtherStatus)
				return;
			double timeDiff = DateTime.Now.Subtract(lastSend).TotalSeconds;
			if (timeDiff < 0 || timeDiff > 90) {
				DoSend();
				return;
			}
			if (sendingOwnStatus) {
				string newStatus = GetSendStatus(Service.Chat.Status);
				if (newStatus != ourStatus && timeDiff > 5) {
					DoSend();
					return;
				}
			}
		}
		
		private void DoSend()
		{
			if (network.GetDistance(userid) < 0)
				return;
			lastSend = DateTime.Now;
			needSend = false;
			Packet p = new Packet();
			p.AnswerId = 0;
			p.Header.PacketID = 0;
			p.Header.Destination = new int[] { userid };
			p.Header.Source = network.UserID;
			p.Target = "SE";
			p.Values.AddValue((byte)((gettingOtherStatus ? 1 : 0) | (sendingOwnStatus ? 2 : 0)));
			if (sendingOwnStatus) {
				ourStatus = GetSendStatus(Service.Chat.Status);
				p.Values.AddValue(ourStatus);
			}
			network.Send(p);
		}
		
		public void ReceiveStatus(PacketEventArgs e)
		{
			Packet p = e.Packet;
			if (p.Values.Count < 1) {
				if (e.Answer != null)
					e.Answer.ErrorCode = Answer.ErrorCodeInvalidRequest;
				return;
			}
			lastReceive = DateTime.Now;
			byte mode = (byte)p.Values.GetValue(0, typeof(byte));
			this.SendingOwnStatus = ((mode & 1) == 1);
			if (this.GettingOtherStatus != ((mode & 2) == 2)) {
				needSend = true;
			}
			if (p.Values.Count > 1) {
				string getStatus = (string)p.Values.GetValue(1, typeof(string));
				this.HisStatus = getStatus;
			} else {
				this.HisStatus = "Unknown";
			}
		}
		
		public static string GetSendStatus(string status)
		{
			if (status == "Invisible")
				return "Offline";
			else
				return status;
		}
	}
}
