// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Collections;
using System.Collections.Specialized;
using System.Text;
using System.Windows.Forms;
using DGNetwork.DGN;
using DGNetwork.Services;
using Grunwald.P2PLib;

namespace DGNetwork.FileTransfer
{
	/// <summary>
	/// The status of the downloads
	/// </summary>
	public enum DownloadState
	{
		/// <summary>
		/// The download is paused.
		/// </summary>
		Paused,
		/// <summary>
		/// The download is active.
		/// </summary>
		Active,
		/// <summary>
		/// The download is completed.
		/// </summary>
		Completed,
		/// <summary>
		/// The download has failed.
		/// </summary>
		Failed
	}
	
	public class Download
	{
		#region Properties
		DownloadState state = DownloadState.Paused;
		
		/// <summary>
		/// Der aktuelle Status des Downloads
		/// </summary>
		public DownloadState State {
			get { return state; }
		}
		
		/// <summary>
		/// Die URL, von der die Datei heruntergeladen wird
		/// </summary>
		public readonly string URL;
		
		/// <summary>
		/// Die Datei, in die der Download gespeichert wird
		/// </summary>
		public readonly string SaveAs;
		
		/// <summary>
		/// Der Dateiname, der angezeigt wird
		/// </summary>
		public string FileName {
			get { return Info.FileName; }
		}
		
		/// <summary>
		/// Weitere Informationen ber die URL
		/// </summary>
		public readonly DGNURL Info;
		
		#endregion
		
		#region Constructor
		FileInfo fileInfo;
		
		public Download(string url, string saveas)
		{
			fileInfo = new FileInfo(saveas + ".part");
			SaveAs = saveas;
			pos = fileInfo.Length;
			speed = new Speed(pos);
			connector = new Connector();
			connector.NoServer += new EventHandler(ConnectorNoServer);
			connector.Connected += new ConnectedEventHandler(ConnectorConnected);
			URL = url;
			Info = new DGNURL(url);
		}
		
		#endregion
		
		/// <summary>
		/// Starts or resumes the download.
		/// </summary>
		public void Start()
		{
			if (State == DownloadState.Paused || State == DownloadState.Failed) {
				state = DownloadState.Active;
				Tick();
			}
		}
		
		/// <summary>
		/// Ticks the download, this makes it reconnect if something went wrong.
		/// </summary>
		public void Tick()
		{
			if (State != DownloadState.Active)
				return;
			Network network = Service.DGN.GetNetwork(Info.Identifier);
			if (network.GetDistance(Info.UserID) >= 0 && !Connected)
				connector.Connect(Info.Identifier, Info.UserID);
		}
		
		/// <summary>
		/// Pauses the download.
		/// </summary>
		public void Pause()
		{
			if (state != DownloadState.Active)
				return;
			state = DownloadState.Paused;
			if (tcp != null)
				tcp.Close();
			tcp = null;
		}
		
		/// <summary>
		/// Aborts the download.
		/// </summary>
		public void Abort()
		{
			if (!Service.Message.AskQuestion("${res:FileTransfer.AbortQuestion}", "${res:FileTransfer.AbortDownload}"))
				return;
			for (int i = 0; i < 10; i++) {
				if (fileInfo.Exists) {
					try {
						fileInfo.Delete();
					} catch (Exception) {
						System.Threading.Thread.Sleep(100);
					}
				}
			}
			Remove();
		}
		
		/// <summary>
		/// Returns the progress of the download in percent.
		/// </summary>
		public double Progress {
			get {
				return pos * 100.0 / Info.FileSize;
			}
		}
		
		private void ConnectorNoServer(object sender, EventArgs e)
		{
			Fail("${res:FileTransfer.NoServer}");
		}
		
		/// <summary>
		/// Returns if the download is currently connected to the server.
		/// </summary>
		public bool Connected {
			get {
				if (tcp == null)
					return false;
				else
					return tcp.Connected;
			}
		}
		
		string failMessage;
		
		/// <summary>
		/// The message that should be shown if the state is failed.
		/// </summary>
		public string FailMessage {
			get {
				return failMessage;
			}
		}
		
		private void Fail(string msg)
		{
			state = DownloadState.Failed;
			failMessage = msg;
		}
		
		long pos;
		IConnection tcp;
		Connector connector;
		
		private void ConnectorConnected(object sender, ConnectedEventArgs e)
		{
			if (Connected)
				return;
			if (state != DownloadState.Active)
				return;
			tcp = e.Connection;
			
			#region HTTP Headers
			tcp.Writer.Write("GET /FileTransfer/" + Info.Hash + "/" + Helper.Escape(Info.FileName) + " HTTP/1.1\r\n" +
			                 Service.Parse("User-Agent: DG Network ${exe:ProductVersion}${AppBuildName}\r\n") +
			                 "Range: bytes=" + pos + "-\r\n" +
			                 "Host: dgn/" + Info.UserID + "\r\n" +
			                 "X-Name: " + Service.Chat.Nickname + "\r\n\r\n");
			tcp.Writer.Flush();
			StringCollection header = Grunwald.P2PLib.Server.HttpLikeServer.GetRequest(tcp);
			if (header == null) {
				Fail("${res:FileTransfer.GotNoResponse}");
				return;
			}
			if (header[0].StartsWith("HTTP/1.1 206 ")) {
				// ok...
			} else if (header[0].StartsWith("HTTP/1.1 200 ")) {
				if (pos != 0) {
					Fail("${res:FileTransfer.ResumeNotSupported}");
					return;
				}
			} else {
				Fail(header[0]);
				return;
			}
			#endregion
			
			int count;
			using (FileStream fs = fileInfo.OpenWrite()) {
				fs.Seek(pos, SeekOrigin.Begin);
				byte[] buffer = new byte[4096];
				while (state == DownloadState.Active) {
					try {
						count = tcp.Stream.Read(buffer, 0, buffer.Length);
					} catch (IOException) {
						count = 0;
					}
					if (state != DownloadState.Active)
						return;
					if (count == 0) {
						tcp = null;
						Tick();
						break;
					} else {
						// Save the data to the file
						fs.Write(buffer, 0, count);
						pos += count;
						if (pos >= Info.FileSize) {
							state = DownloadState.Completed;
							fs.Close();
							fileInfo.MoveTo(SaveAs);
							connector.Dispose();
							connector = null;
							Service.Message.PlaySound("sound/FileDone.wav");
						}
					}
				}
			}
		}
		
		Speed speed;
		
		/// <summary>
		/// Returns the number of bytes transferred since the last call to this method.
		/// </summary>
		public int GetSpeed()
		{
			return speed.GetSpeed(pos);
		}
		
		public void ResetSpeed()
		{
			speed.SetPosition(pos);
		}
		
		private void Remove()
		{
			if (connector != null) {
				connector.Dispose();
				connector = null;
			}
			lock (TransferService.Downloads) {
				TransferService.Downloads.Remove(this);
			}
		}
	}
}
