// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using DGNetwork.Services;
using DGNetwork.Chat.Parser;

namespace DGNetwork.FileTransfer
{
	public class SendFile : Form
	{
		#region WinForms
		private Label label1;
		private ProgressBar progressBar;
		private Label fileName;
		private Button abortButton;
		
		private Container components = null;
		
		protected override void Dispose(bool disposing)
		{
			if (disposing) {
				if (components != null) {
					components.Dispose();
				}
			}
			base.Dispose(disposing);
		}
		
		#region Vom Windows Form-Designer generierter Code
		private void InitializeComponent()
		{
			this.label1 = new Label();
			this.progressBar = new ProgressBar();
			this.fileName = new Label();
			this.abortButton = new Button();
			this.SuspendLayout();
			//
			// label1
			//
			this.label1.Location = new Point(8, 8);
			this.label1.Name = "label1";
			this.label1.Size = new Size(312, 32);
			this.label1.TabIndex = 0;
			//
			// progressBar
			//
			this.progressBar.Location = new Point(8, 40);
			this.progressBar.Name = "progressBar";
			this.progressBar.Size = new Size(304, 16);
			this.progressBar.TabIndex = 1;
			//
			// fileName
			//
			this.fileName.Location = new Point(8, 64);
			this.fileName.Name = "fileName";
			this.fileName.Size = new Size(304, 32);
			this.fileName.TabIndex = 2;
			this.fileName.Text = "label2";
			//
			// abortButton
			//
			this.abortButton.FlatStyle = FlatStyle.System;
			this.abortButton.Location = new Point(232, 88);
			this.abortButton.Name = "abortButton";
			this.abortButton.Size = new Size(80, 24);
			this.abortButton.TabIndex = 3;
			this.abortButton.Text = "Abort";
			this.abortButton.Click += new EventHandler(this.abortButton_Click);
			//
			// SendFile
			//
			this.AutoScaleBaseSize = new Size(5, 13);
			this.ClientSize = new Size(322, 120);
			this.ControlBox = false;
			this.Controls.Add(this.abortButton);
			this.Controls.Add(this.fileName);
			this.Controls.Add(this.progressBar);
			this.Controls.Add(this.label1);
			this.FormBorderStyle = FormBorderStyle.FixedDialog;
			this.Name = "SendFile";
			this.StartPosition = FormStartPosition.CenterParent;
			this.Closed += new EventHandler(this.SendFile_Closed);
			this.ResumeLayout(false);
			
		}
		
		#endregion
		
		#endregion
		
		#region Public Instance
		/// <summary>
		/// Creates a new SendFile-Window
		/// </summary>
		public SendFile(string filename)
		{
			InitializeComponent();
			fileName.Text = filename;
			label1.Text = DGNetwork.Services.Service.Parse("${res:FileTransfer.SendFile.Hashing}");
			this.Text = DGNetwork.Services.Service.Parse("${res:FileTransfer.SendFile.HashingTitle}");
			abortButton.Text = DGNetwork.Services.Service.Parse("${res:Global.Cancel}");
		}
		
		/// <summary>
		/// Sets the ProgressBar auf the window to the specified value.
		/// </summary>
		public void SetProgress(int value)
		{
			if (IsHandleCreated)
				Invoke(new void_int(DoSetProgress), new object[] { value } );
		}
		
		private delegate void void_int(int value);
		
		private void DoSetProgress(int value)
		{
			progressBar.Value = value;
		}
		
		#endregion
		
		#region Send
		
		/// <summary>
		/// Sends the specified file as big file: it is hashed (a progress bar is displayed to the user)
		/// and inserted into the list of shared files.
		/// Returns a chat code tag that can be used to let other users download the file.
		/// Errors are displayed to the user and this method returns null.
		/// </summary>
		public static string Send(string filename)
		{
			FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.Read);
			try {
				foreach (SharedFileHandler f in TransferService.SharedFiles) {
					if (f.FullName.ToLower() == filename.ToLower()) {
						if (f.FileSize != fs.Length) {
							// TODO: Remove shared file!
							break;
						} else {
							// f.LastTransfer = DateTime.Now;
							return f.Tag;
						}
					}
				}
				if (fs.Length > MaxBigFileSize)
					return "<i>Kann Dateien > 1 GB nicht senden!</i>";
				SendFile form = new SendFile(filename);
				form.FileStream = fs;
				form.Start();
				form.ShowDialog();
				if (form.Hash == null)
					return null;
				string tag = string.Format("<file length={0:d} hash={1}>{2}</file>",
				                           HtmlHelper.EncodeAttribute(fs.Length),
				                           HtmlHelper.EncodeAttribute(form.Hash),
				                           HtmlHelper.EncodeEntities(Path.GetFileName(filename)));
				TransferService.ShareFile(filename, fs.Length, form.ShortHash, tag);
				return tag;
			} finally {
				fs.Close();
			}
		}
		
		#endregion
		
		FileStream FileStream;
		string Hash = null;
		string ShortHash;
		
		/// <summary>
		/// The maximum size of that can be sent.
		/// The limit is just so low because I didn't test higher values, in theory
		/// the limit is 100 GB. (a hash block shouldn't be > 1 GB, should it?)
		/// </summary>
		public const long MaxBigFileSize = int.MaxValue;
		
		private void CalcHash()
		{
			try {
				DoCalcHash();
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		bool abort = false;
		
		void DoCalcHash()
		{
			FileStream fs = FileStream;
			StringBuilder hash = new StringBuilder();
			int blocksize = (int) Math.Max(32768/*SendParser.MaxInPacketFileSize*/, fs.Length/100);
			// A file should be made of 100 blocks.
			// -> a 100 MB has 1 MB blocks.
			// A block has a minimum size of 32 KB,
			// so a 500 KB file has only 16 blocks.
			hash.AppendFormat("{0:d}:", blocksize);
			
			int blockcount = (int) Math.Ceiling(fs.Length/(double) blocksize);
			byte[] block = new byte[blocksize];
			StringBuilder blockhash = new StringBuilder();
				for (int i = 1; i <= blockcount; i++) {
					if (i == blockcount && (fs.Length%blocksize) > 0) {
						block = new byte[fs.Length%blocksize];
					}
					if (fs.Read(block, 0, block.Length) != block.Length)
						throw new Exception("fs.Read returned wrong length");
					if (abort)
						return;
					blockhash.Append(Helper.MD5(block), 0, 8);
					if (abort)
						return;
					SetProgress((int) (100.0*i/blockcount));
				}
			string blockHash = blockhash.ToString();
			hash.Append(blockHash);
			hash.Append(':');
			ShortHash = Helper.MD5(blockHash).Substring(0, 16);
			hash.Append(ShortHash);
			Hash = hash.ToString();
			while (!IsHandleCreated)
				Thread.Sleep(100);
			Invoke(new MethodInvoker(Close));
		}
		
		private void Start()
		{
			(new MethodInvoker(CalcHash)).BeginInvoke(null, null);
		}
		
		private void abortButton_Click(object sender, EventArgs e)
		{
			Close();
		}
		
		private void SendFile_Closed(object sender, EventArgs e)
		{
			abort = true;
		}
	}
}
