// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;
using DGNetwork.Chat;
using DGNetwork.Services;
using Grunwald.Gui.Components;

namespace DGNetwork.FileTransfer
{
	/// <summary>
	/// Description of TransferDataObjectConverter.
	/// </summary>
	public class TransferDataObjectConverter : IDataObjectConverter
	{
		/// <summary>
		/// Gets the object out of the IDataObject.
		/// </summary>
		public object GetObject(IDataObject d, object dropTarget)
		{
			if (dropTarget is IEditBox && d.GetDataPresent("FileDrop")) {
				string[] files = (string[]) d.GetData("FileDrop");
				if (files == null) {
					Service.Message.ShowError("${res:Chat.FileDropIsNull}");
				} else {
					return new FileDrop(files);
				}
			}
			return null;
		}
		
		private class FileDrop
		{
			string[] files;
			
			internal FileDrop(string[] files)
			{
				this.files = files;
			}
			
			public override string ToString()
			{
				StringBuilder result = new StringBuilder();
				foreach (string file in files) {
					if (result.Length > 0)
						result.Append("\r\n");
					result.Append("<sendfile=");
					result.Append(DGNetwork.Chat.Parser.HtmlHelper.EncodeAttribute(TransferService.ToShortName(file)));
					result.Append(">");
				}
				return result.ToString();
			}
		}
		
		/// <summary>
		/// Gets the allowed effects when dropping the specified object.
		/// </summary>
		public DragDropEffects GetEffects(IDataObject d, object dropTarget)
		{
			if (dropTarget is IEditBox && (d.GetDataPresent(DataFormats.Bitmap) || d.GetDataPresent("FileDrop")))
				return DragDropEffects.Copy;
			else
				return DragDropEffects.None;
		}
		
		/// <summary>
		/// Gets the allowed effects when dropping the specified object.
		/// </summary>
		public DragDropEffects GetEffects(object o, object dropTarget)
		{
			if (dropTarget is IEditBox && (o is FileDrop || o is Bitmap))
				return DragDropEffects.Copy;
			else
				return DragDropEffects.None;
		}
		
		/// <summary>
		/// Gets the text representing the specified object in the specified DropTarget.
		/// </summary>
		public string GetText(object o, object dropTarget)
		{
			if (!(dropTarget is IEditBox))
				return null;
			if (o is FileDrop)
				return o.ToString();
			Bitmap bmp = o as Bitmap;
			if (bmp != null) {
				bool canDispose = false;
				if (bmp.Width > 300 || bmp.Height > 300 || (bmp.Width > 150 && bmp.Height > 100)) {
					Bitmap n = BitmapSelect.Select(bmp, Service.Parse("${res:Chat.SelectBitmapRange}"));
					if (bmp != n)
						canDispose = true;
					if (n == null)
						return null;
					bmp = n;
				}
				string code = Helper.Random.Next(0, ushort.MaxValue).ToString("x4");
				string name = "clip" + code + ".png";
				string tempPath = Helper.GetTempPath();
				try {
					bmp.Save(tempPath + name, ImageFormat.Png);
				} catch (Exception ex) {
					Service.Message.ShowErrorFormatted("${res:Chat.CantSaveImage}", new string[] { tempPath + name, ex.Message });
					name = "";
				}
				string name2 = "clip" + code + ".jpg";
				try {
					bmp.Save(tempPath + name2, ImageFormat.Jpeg);
				} catch (Exception ex) {
					Service.Message.ShowErrorFormatted("${res:Chat.CantSaveImage}", new string[] { tempPath + name2, ex.Message });
					name2 = "";
				}
				FileInfo inf1 = new FileInfo(tempPath + name);
				FileInfo inf2 = new FileInfo(tempPath + name2);
				if (name == "") {
					name = name2;
				} else {
					if (name2 != "") {
						if (inf1.Length > inf2.Length + 5000)
							name = name2;
						if ((inf1.Length > 32*1024) && (inf2.Length < 32*1024))
							name = name2;
					}
				}
				if (canDispose)
					bmp.Dispose();
				return "<sendfile=\"temp:" + name + "\">";
			}
			return null;
		}
	}
}
