// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Resources;
using System.Windows.Forms;
using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.AddIns.Conditions;
using DGNetwork.AddIns.Codons;
using DGNetwork.Gui;
using DGNetwork.Services;
using Grunwald.Gui.Components;
using Grunwald.P2PLib.Server.Http;

namespace DGNetwork.FileTransfer
{
	public class TransferDialogSink
	{
		Timer timer;
		ListView downloadListView = null;
		ListView uploadListView = null;
		ListView sharedFilesListView = null;
		ImageListManager imageListManager;
		
		private void FormLoad(object sender, EventArgs e)
		{
			timer = new Timer();
			timer.Tick += new EventHandler(TimerTick);
			timer.Interval = 250;
			imageListManager = new ImageListManager();
			Corrections.MakeProportional(sharedFilesListView);
			InitListView(downloadListView);
			InitListView(uploadListView);
			TimerTick(null, null);
			timer.Enabled = true;
		}
		
		private void InitListView(ListView list)
		{
			int width = 4;
			foreach (ColumnHeader h in list.Columns) {
				width += h.Width + 2;
			}
			foreach (ColumnHeader h in list.Columns) {
				h.Width = list.Width * h.Width / width;
			}
			list.SmallImageList = imageListManager.ImageList;
			Corrections.MakeProportional(list);
		}
		
		private void FormClosed(object sender, EventArgs e)
		{
			timer.Dispose();
			imageListManager.Dispose();
		}
		
		private void TimerTick(object sender, EventArgs e)
		{
			try {
				UpdateShared();
				UpdateUploads();
				UpdateDownloads();
				firstTick = false;
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		
		bool firstTick = true;
		bool updating;
		ListView currentLV;
		
		#region Shared files
		private void UpdateShared()
		{
			updating = false;
			ListView leftView = sharedFilesListView;
			ArrayList shared = TransferService.SharedFiles;
			lock (shared) {
				int i;
				currentLV = leftView;
				for (i = 0; i < shared.Count; i++) {
					if (i < leftView.Items.Count) {
						Change(leftView.Items[i], (SharedFileHandler) shared[i]);
					} else {
						ListViewItem item = new ListViewItem();
						Change(item, (SharedFileHandler) shared[i]);
						leftView.Items.Add(item);
					}
				}
				while (i < leftView.Items.Count) {
					leftView.Items.RemoveAt(i);
				}
			}
			if (updating)
				sharedFilesListView.EndUpdate();
		}
		
		private void Change(ListViewItem item, SharedFileHandler upload)
		{
			Change(item, upload.Name);
		}
		#endregion
		
		#region Upload
		private void UpdateUploads()
		{
			updating = false;
			ArrayList shared = TransferService.SharedFiles;
			lock (shared) {
				currentLV = uploadListView;
				int i = 0;
				foreach (SharedFileHandler sfh in shared) {
					lock (sfh.Uploads) {
						foreach (FileUpload upload in sfh.Uploads) {
							if (i < uploadListView.Items.Count) {
								Change(uploadListView.Items[i], sfh, upload);
							} else {
								ListViewItem item = new ListViewItem();
								item.ImageIndex = imageListManager["Icons.ArrowUp"];
								Change(item, sfh, upload);
								uploadListView.Items.Add(item);
							}
							i += 1;
						}
					}
				}
				while (i < uploadListView.Items.Count) {
					uploadListView.Items.RemoveAt(i);
				}
			}
			if (updating)
				uploadListView.EndUpdate();
		}
		
		private void Change(ListViewItem item, SharedFileHandler shared, FileUpload upload)
		{
			while (item.SubItems.Count < 5)
				item.SubItems.Add("");
			item.Tag = upload;
			Change(item.SubItems[0], shared.Name);
			string partnerName = upload.Request.Headers["X-Name"];
			if (partnerName == null || partnerName.Length == 0)
				partnerName = "unknown";
			Change(item.SubItems[1], partnerName);
			Change(item.SubItems[2], SizeToString(shared.FileSize));
			Change(item.SubItems[3], (upload.Position * 100.0 / shared.FileSize).ToString("f1") + " %");
			if (firstTick) {
				//upload.ResetSpeed();
			} else {
				//int speed = upload.GetSpeed();
				//Change(item.SubItems[4], ((speed/102)*0.1).ToString() + " KB/s");
			}
		}
		#endregion
		
		#region Download
		private void UpdateDownloads()
		{
			updating = false;
			ArrayList downloads = TransferService.Downloads;
			lock (downloads) {
				int i;
				currentLV = downloadListView;
				for (i = 0; i < downloads.Count; i++) {
					if (i < downloadListView.Items.Count) {
						Change(downloadListView.Items[i], (Download) downloads[i]);
					} else {
						ListViewItem item = new ListViewItem();
						item.ImageIndex = imageListManager["Icons.ArrowDown"];
						Change(item, (Download) downloads[i]);
						downloadListView.Items.Add(item);
					}
				}
				while (i < downloadListView.Items.Count) {
					downloadListView.Items.RemoveAt(i);
				}
			}
			if (updating)
				downloadListView.EndUpdate();
		}
		
		private void Change(ListViewItem item, Download dl)
		{
			while (item.SubItems.Count < 5)
				item.SubItems.Add("");
			item.Tag = dl;
			Change(item.SubItems[0], dl.FileName);
			Change(item.SubItems[1], dl.Info.UserName);
			Change(item.SubItems[2], SizeToString(dl.Info.FileSize));
			Change(item.SubItems[3], dl.Progress.ToString("f1") + " %");
			
			if (dl.State == DownloadState.Active) {
				if (dl.Connected) {
					if (firstTick) {
						dl.ResetSpeed();
					} else {
						int speed = dl.GetSpeed();
						Change(item.SubItems[4], (speed / 1024.0).ToString("f1") + " KB/s");
					}
				} else {
					Change(item.SubItems[4], "Verbinde...");
				}
			} else if (dl.State == DownloadState.Paused) {
				Change(item.SubItems[4], "Angehalten");
			} else if (dl.State == DownloadState.Completed) {
				Change(item.SubItems[4], "Fertig");
			} else if (dl.State == DownloadState.Failed) {
				Change(item.SubItems[4], dl.FailMessage);
			} else {
				throw new Exception("unknown state");
			}
		}
		
		#endregion
		
		#region Change
		
		private void Change(ListViewItem item, string text)
		{
			if (item.Text != text) {
				BeginUpdate();
				item.Text = text;
			}
		}
		
		private void Change(ListViewItem.ListViewSubItem subitem, string text)
		{
			if (subitem.Text != text) {
				BeginUpdate();
				subitem.Text = text;
			}
		}
		
		private string SizeToString(long filesize)
		{
			if (filesize > 10000000000) {
				// > 10 GB
				return string.Format("{0} GB", (filesize/1024/1024/1024));
			} else if (filesize > 100000000) {
				// > 100 MB
				return string.Format("{0} MB", (filesize/1024/1024));
			} else if (filesize > 1000000) {
				// > 1 MB
				return string.Format("{0:f1} MB", ((filesize/1024)/1024.0));
			} else {
				// < 1 MB
				return string.Format("{0} KB", filesize/1024);
			}
		}
		
		private void BeginUpdate()
		{
			if (!updating) {
				currentLV.BeginUpdate();
				updating = true;
			}
		}
		
		#endregion
		
		private void DownloadMouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button != MouseButtons.Right)
				return;
			ListViewItem item = downloadListView.GetItemAt(e.X, e.Y);
			if (item == null)
				return;
			Service.Menu.ShowContextMenu(item.Tag, "/Context/FileTransfer/Download", downloadListView, e.X, e.Y);
		}
		
		private void DownloadDoubleClick(object sender, EventArgs e)
		{
			if (downloadListView.SelectedItems.Count == 1) {
				OpenDownloadCommand.Open((Download)downloadListView.SelectedItems[0].Tag);
			}
		}
		
		private void UploadMouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button != MouseButtons.Right)
				return;
			ListViewItem item = uploadListView.GetItemAt(e.X, e.Y);
			if (item == null)
				return;
			Service.Menu.ShowContextMenu(item.Tag, "/Context/FileTransfer/Upload", uploadListView, e.X, e.Y);
		}
	}
	
	[Condition]
	public class DownloadStateCondition : AbstractCondition
	{
		[XmlMemberAttribute("downloadstate", IsRequired = true)]
		string downloadstate = null;
		
		public override bool IsValid(object caller)
		{
			Download dl = (Download)caller;
			foreach (string state in downloadstate.Split(',')) {
				if (state.Trim().ToLower() == dl.State.ToString().ToLower())
					return true;
			}
			return false;
		}
	}
	
	public class PauseDownloadCommand : AbstractMenuCommand {
		public override void Run() {
			((Download)this.Owner).Pause();
		}
	}
	
	public class StartDownloadCommand : AbstractMenuCommand {
		public override void Run() {
			((Download)this.Owner).Start();
		}
	}
	
	public class AbortDownloadCommand : AbstractMenuCommand {
		public override void Run() {
			((Download)this.Owner).Pause();
			((Download)this.Owner).Abort();
		}
	}
	
	public class OpenDownloadCommand : AbstractMenuCommand
	{
		public override void Run()
		{
			Open((Download)this.Owner);
		}
		
		public static void Open(Download download)
		{
			if (download.State == DownloadState.Completed) {
				if (MessageBox.Show(Service.Parse("${res:FileTransfer.OpenDownload.Question}", download.FileName),
				                    Service.Parse("${res:FileTransfer.OpenDownload.Title}"),
				                    MessageBoxButtons.YesNo,
				                    MessageBoxIcon.Warning,
				                    MessageBoxDefaultButton.Button2)
				    == DialogResult.Yes) {
					try {
						Process.Start(download.SaveAs);
					} catch (Exception ex) {
						Service.Message.ShowMessageFormatted("${res:FileTransfer.OpenDownload.Error}", new string[] { ex.Message, download.SaveAs });
					}
				}
			}
		}
	}
}
