// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Windows.Forms;
using Grunwald.Gui.Controls;
using Grunwald.Gui.Settings;

using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.AddIns.Codons;
using ICSharpCode.Core.AddIns.Conditions;
using ICSharpCode.Core.Properties;
using DGNetwork.Gui;
using DGNetwork.Gui.Menu;
using DGNetwork.Services;
using TLV = Grunwald.Gui.Controls.TreeListView;

namespace DGNetwork.Gui.Dialogs
{
	#region MonitorSink
	public class MonitorSink
	{
		TLV tree = null;
		ImageListManager imageListManager;
		Timer timer;
		
		private void FormLoad(object sender, EventArgs e)
		{
			imageListManager = new ImageListManager();
			tree.Columns[0].Width = tree.Width * 2 / 5 - 12;
			tree.Columns[1].Width = tree.Width * 3 / 5 - 12;
			tree.SmallImageList = imageListManager.ImageList;
			Grunwald.Gui.Components.Corrections.MakeProportional(tree);
			timer = new Timer();
			// TODO: Optimize the monitor
			// (set interval to 100 and see processor usage go up 10% if monitor is open)
			timer.Interval = 500;
			timer.Tick += new EventHandler(TimerTick);
			CreateMonitorItems("/Monitor");
			Service.Resource.Reloaded += new EventHandler(ResourceReloaded);
			timer.Enabled = true;
		}
		
		private void TreeMouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Right) {
				int count = tree.SelectedNodes.Count;
				if (count == 0)
					return;
				IItemWithContextMenu[] items = new IItemWithContextMenu[count];
				for (int i = 0; i < items.Length; i++) {
					items[i] = tree.SelectedNodes[i].Tag as IItemWithContextMenu;
				}
				Service.Menu.ShowContextMenu(items, tree, e.X, e.Y);
			}
		}
		
		private void FormClosed(object sender, EventArgs e)
		{
			timer.Enabled = false;
			timer.Dispose();
			RemoveNodeTags(tree.Nodes);
			Service.Resource.Reloaded -= new EventHandler(ResourceReloaded);
			imageListManager.Dispose();
		}
		
		private void ResourceReloaded(object sender, EventArgs e)
		{
			// Update all names and values
			tree.BeginUpdate();
			UpdateNodes(tree.Nodes, true);
			tree.EndUpdate();
		}
		
		private void TimerTick(object sender, EventArgs e)
		{
			tree.BeginUpdate();
			UpdateNodes(tree.Nodes, false);
			tree.EndUpdate();
		}
		
		private void UpdateNodes(TreeListNodeCollection col, bool force)
		{
			foreach (TreeListNode node in col) {
				UpdateNodes(node.Nodes, force);
				IMonitorItem item = node.Tag as IMonitorItem;
				if (item != null) {
					item.Update(imageListManager, force);
				}
			}
		}
		
		private void RemoveNodeTags(TreeListNodeCollection col)
		{
			foreach (TreeListNode node in col) {
				RemoveNodeTags(node.Nodes);
				IMonitorItem item = node.Tag as IMonitorItem;
				if (item != null) {
					item.Remove();
				}
			}
		}
		
		public void CreateMonitorItems(string addInTreePath)
		{
			IAddInTreeNode node;
			try {
				node = AddInTreeSingleton.AddInTree.GetTreeNode(addInTreePath);
			} catch (TreePathNotFoundException) {
				Service.Message.ShowErrorFormatted("${res:Global.TreePathNotFound}", addInTreePath);
				return;
			}
			MonitorItemCodon.AddItemsToCollection(node.BuildChildItems(null), tree.Nodes, "Monitor");
		}
	}
	#endregion
	
	[CodonName("MonitorItem")]
	public class MonitorItemCodon : AbstractCodon
	{
		[XmlMemberAttribute("title", IsRequired = true)]
		string title = null;
		
		public string Title { get { return title; } }
		
		/// <summary>
		/// Creates an item with the specified sub items. And the current
		/// Condition status for this item.
		/// </summary>
		public override object BuildItem(object owner, ArrayList subItems, ConditionCollection conditions)
		{
			IMonitorItem m = (IMonitorItem)this.AddIn.CreateObject(this.Class);
			if (m == null) return null;
			TreeListNode node = new TreeListNode();
			AddItemsToCollection(subItems, node.Nodes, this.Name);
			m.Title = this.Title;
			m.Initialize(node);
			return node;
		}
		
		internal static void AddItemsToCollection(ArrayList items, TreeListNodeCollection nodes, string parentName)
		{
			foreach (object o in items) {
				if (o is TreeListNode) {
					nodes.Add((TreeListNode)o);
				} else if (o is IMonitorItem) {
					TreeListNode node = new TreeListNode();
					((IMonitorItem)o).Initialize(node);
					nodes.Add(node);
				} else if (o is IMonitorItemBuilder) {
					((IMonitorItemBuilder)o).BuildItems(nodes);
				} else {
					Service.Message.ShowErrorFormatted("${res:Codon.InvalidSubItem}", parentName, o.GetType().FullName);
				}
			}
		}
	}
	
	public interface IMonitorItemBuilder
	{
		void BuildItems(TreeListNodeCollection targetCollection);
	}
	
	public interface IMonitorItem : IItemWithContextMenu
	{
		void Initialize(TreeListNode node);
		void Update(ImageListManager imageList, bool force);
		void Remove();
		string Title { get; set; }
	}
	
	public class DefaultMonitorItem : IMonitorItem
	{
		TreeListNode node;
		string valueText;
		object formatValue;
		string title;
		object titleFormatValue;
		string icon;
		bool dirty;
		
		object IItemWithContextMenu.Owner {
			get {
				return this;
			}
		}
		
		string contextMenuTreePath;
		
		public string ContextMenuTreePath {
			get {
				return contextMenuTreePath;
			}
			set {
				contextMenuTreePath = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the text for the value column.
		/// This string is translated before it is displayed.
		/// </summary>
		public string ValueText {
			get {
				return valueText;
			}
			set {
				if (valueText == value) return;
				valueText = value;
				dirty = true;
			}
		}
		
		/// <summary>
		/// Gets/Sets the value used as format parameter for ValueText after translation.
		/// If this property is null, no formatting is done after translation.
		/// </summary>
		public object FormatValue {
			get {
				return formatValue;
			}
			set {
				formatValue = value;
				dirty = true;
			}
		}
		
		public string Title {
			get {
				return title;
			}
			set {
				title = value;
				dirty = true;
			}
		}
		
		/// <summary>
		/// Gets/Sets the value used as format parameter for Title after translation.
		/// If this property is null, no formatting is done after translation.
		/// </summary>
		public object TitleFormatValue {
			get {
				return titleFormatValue;
			}
			set {
				titleFormatValue = value;
				dirty = true;
			}
		}
		
		/// <summary>
		/// Gets/Sets the icon of this item.
		/// </summary>
		public string Icon {
			get {
				return icon;
			}
			set {
				icon = value;
				dirty = true;
			}
		}
		
		/// <summary>
		/// Gets the treenode this MonitorItem is representing.
		/// </summary>
		public TreeListNode TreeNode {
			get {
				return node;
			}
		}
		
		public virtual void Initialize(TreeListNode node)
		{
			this.node = node;
			node.Tag = this;
			node.Text = Service.Parse(title);
			node.SubItems.Add(new ContainerSubListViewItem(valueText));
		}
		
		public void Update(ImageListManager imageList, bool force)
		{
			if (!dirty && !force)
				return;
			string s;
			s = Service.StringParser.Parse(title);
			if (titleFormatValue != null)
				s = String.Format(s, titleFormatValue);
			node.Text = s;
			s = Service.StringParser.Parse(valueText);
			if (formatValue != null)
				s = String.Format(s, formatValue);
			node.SubItems[0].Text = s;
			node.ImageIndex = imageList[icon];
		}
		
		public virtual void Remove()
		{
			// do nothing
		}
	}
}
