// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;

using Grunwald.Gui.Controls;
using Grunwald.Gui.Settings;

using ICSharpCode.Core.AddIns;

using DGNetwork.Services;
using DGNetwork.Gui;
using DGNetwork.Search;

namespace DGNetwork.Gui.Dialogs
{
	public class SearchDialogSink
	{
		Form form = null;
		SettingsPanel criteriaPanel = null;
		ComboBox searchTypeCombo = null;
		ListView resultListView = null;
		Button addToListButton = null;
		ImageListManager imageManager;
		
		ArrayList searchTypes = new ArrayList();
		SearchSettingPage page;
		
		#region Search Type Management
		private void FormLoad(object sender, EventArgs e)
		{
			IAddInTreeNode node = AddInTreeSingleton.AddInTree.GetTreeNode("/Search");
			ArrayList subItems = node.BuildChildItems(this);
			imageManager = new ImageListManager(resultListView.SmallImageList);
			foreach (object o in subItems) {
				if (o is ISearchType) {
					searchTypes.Add(o);
				} else {
					Service.Message.ShowErrorFormatted("${res:Codon.InvalidSubItem}",
					                                   "Search",
					                                   o.GetType().FullName);
				}
			}
			foreach (ISearchType search in searchTypes) {
				searchTypeCombo.Items.Add(new SearchTypeBinder(search));
			}
			searchTypeCombo.SelectedIndex = 0;
		}
		
		private class SearchTypeBinder
		{
			ISearchType search;
			public SearchTypeBinder(ISearchType search) {
				this.search = search;
			}
			public override string ToString()
			{
				return Service.Parse(search.Name);
			}
			public ISearchType Search {
				get {
					return search;
				}
			}
		}
		
		private void SearchTypeChanged(object sender, EventArgs e)
		{
			int index = searchTypeCombo.SelectedIndex;
			if (index < 0) {
				page = null;
			} else {
				SearchTypeBinder binder = (SearchTypeBinder)searchTypeCombo.SelectedItem;
				page = new SearchSettingPage(binder.Search);
				page.SearchStarted += new EventHandler(SearchStarted);
				page.SearchEnded += new EventHandler(SearchEnded);
				page.ResultAdded += new ResultEventHandler(SearchResultGot);
			}
			form.AcceptButton = null;
			criteriaPanel.Page = page;
			form.AcceptButton = page.Button;
		}
		#endregion
		
		private void SearchStarted(object sender, EventArgs e)
		{
			resultListView.Clear();
			columnHash.Clear();
			ColumnHeader column = new ColumnHeader();
			column.Text = Service.Parse("${res:SearchDialog.FirstColumn}");
			resultListView.Columns.Add(column);
			searchTypeCombo.Enabled = false;
		}
		
		private void SearchEnded(object sender, EventArgs e)
		{
			searchTypeCombo.Enabled = true;
		}
		
		Hashtable columnHash = new Hashtable(); // string -> ColumnHeader
		
		private void SearchResultGot(object sender, ResultEventArgs e)
		{
			IResult result = e.Result;
			using (Graphics g = resultListView.CreateGraphics()) {
				string[] titles = GetColumnValues(g, result);
				ListViewItem item = new ListViewItem(titles, imageManager[result.Icon]);
				item.Tag = result;
				resultListView.Items.Add(item);
				new SearchResultUpdater(result, item, this);
			}
		}
		
		private class SearchResultUpdater
		{
			IResult result;
			ListViewItem item;
			SearchDialogSink parent;
			
			public SearchResultUpdater(IResult result, ListViewItem item, SearchDialogSink parent)
			{
				this.result = result;
				this.item = item;
				this.parent = parent;
				result.ResultChanged += new EventHandler(ResultChanged);
			}
			
			private void ResultChanged(object sender, EventArgs e)
			{
				if (item.ListView == null)
					return;
				if (item.ListView.InvokeRequired) {
					item.ListView.Invoke(new EventHandler(ResultChanged), new object[] { sender, e });
				} else {
					try {
						item.ImageIndex = parent.imageManager[result.Icon];
						using (Graphics g = item.ListView.CreateGraphics()) {
							string[] titles = parent.GetColumnValues(g, result);
							item.Text = titles[0];
							for (int i = 1; i < titles.Length; ++i) {
								if (item.SubItems.Count == i)
									item.SubItems.Add(new ListViewItem.ListViewSubItem());
								item.SubItems[i].Text = titles[i];
							}
						}
					} catch (Exception ex) {
						Service.Message.ShowError(ex);
					}
				}
			}
		}
		
		private string[] GetColumnValues(Graphics g, IResult result)
		{
			foreach (ResultColumn rc in result.Columns) {
				ColumnHeader column = (ColumnHeader)columnHash[rc.Title];
				if (column == null) {
					column = new ColumnHeader();
					column.Text = Service.Parse(rc.Title);
					columnHash[rc.Title] = column;
					resultListView.Columns.Add(column);
				}
			}
			string[] titles = new string[resultListView.Columns.Count];
			titles[0] = Service.Parse(result.Name);
			foreach (ResultColumn rc in result.Columns) {
				ColumnHeader column = (ColumnHeader)columnHash[rc.Title];
				titles[column.Index] = Service.Parse(rc.Value);
			}
			for (int i = 0; i < titles.Length; i++) {
				CheckWidth(g, resultListView.Columns[i], titles[i]);
			}
			return titles;
		}
		
		private void CheckWidth(Graphics g, ColumnHeader column, string text)
		{
			if (text == null || text.Length == 0)
				return;
			int needed = (int)g.MeasureString(text, resultListView.Font).Width + 7;
			if (column.Index == 0)
				needed += 16;
			if (needed > column.Width)
				column.Width = needed;
		}
		
		private void ResultDoubleClick(object sender, EventArgs e)
		{
			if (resultListView.SelectedItems.Count == 1) {
				ListViewItem item = resultListView.SelectedItems[0];
				IResult result = item.Tag as IResult;
				if (result != null)
					result.DoubleClick();
			}
		}
		
		private void AddToListClick(object sender, EventArgs e)
		{
			foreach (ListViewItem item in resultListView.SelectedItems) {
				IResult result = item.Tag as IResult;
				if (result != null)
					result.AddToList();
			}
		}
		
		private void ListSelectionChanged(object sender, EventArgs e)
		{
			if (resultListView.SelectedItems.Count < 1) {
				addToListButton.Enabled = false;
				return;
			}
			bool ok = true;
			foreach (ListViewItem item in resultListView.SelectedItems) {
				IResult result = item.Tag as IResult;
				if (result != null) {
					if (!result.CanAddToList) {
						ok = false;
						break;
					}
				}
			}
			addToListButton.Enabled = ok;
		}
		
		private void ResultMouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button != MouseButtons.Right)
				return;
			if (resultListView.SelectedItems.Count == 0)
				return;
			ArrayList results = new ArrayList();
			foreach (ListViewItem item in resultListView.SelectedItems) {
				IResult result = item.Tag as IResult;
				if (result != null)
					results.Add(result);
			}
			if (results.Count == 0)
				return;
			IResult[] resultArray = (IResult[])results.ToArray(typeof(IResult));
			Service.Menu.ShowContextMenu(resultArray, resultListView, e.X, e.Y);
		}
		
		private void FormClosed(object sender, EventArgs e)
		{
			if (page != null) {
				ISearch search = page.Search;
				if (search != null)
					search.Abort();
			}
		}
	}
}
