// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;
using DGNetwork.Chat;
using DGNetwork.Contacts;
using DGNetwork.Services;

namespace DGNetwork.Gui
{
	public interface IMainWindowBuilder
	{
		string Name { get; }
		string Title { get; }
		MainWindowCreationEventArgs Create();
		
		/// <summary>
		/// Gets if the main windows build by this window builder can be used as chat windows.
		/// For example, this property is used in the click event handler of the balloon
		/// 'message received' to determine if a chat window or the main window should be opened.
		/// </summary>
		bool UseAsChatWindow { get; }
	}
	
	public abstract class AbstractMainWindowBuilder : IMainWindowBuilder
	{
		string name, title;
		
		public AbstractMainWindowBuilder(string name, string title)
		{
			this.name = name;
			this.title = title;
		}
		
		public string Name {
			get {
				return name;
			}
		}
		
		public string Title {
			get {
				return title;
			}
		}
		
		/// <summary>
		/// Gets if the main windows build by this window builder can be used as chat windows.
		/// For example, this property is used in the click event handler of the balloon
		/// 'message received' to determine if a chat window or the main window should be opened.
		/// </summary>
		public virtual bool UseAsChatWindow {
			get {
				return false;
			}
		}
		
		/// <summary>
		/// Creates the main window by calling <see cref="CreateForm"/>, <see cref="CreateContent"/>
		/// and <see cref="AddMenu"/>.
		/// </summary>
		public MainWindowCreationEventArgs Create()
		{
			Form frm = CreateForm();
			MainWindowCreationEventArgs e = new MainWindowCreationEventArgs(frm);
			AddContent(e);
			AddMenu(e);
			return e;
		}
		
		public virtual Form CreateForm()
		{
			Form frm = new Form();
			frm.Text = "DG Network";
			frm.MinimizeBox = false;
			frm.MaximizeBox = false;
			frm.Icon = Service.Resource.GetIcon("Icons.App");
			frm.MinimumSize = new Size(100, 150);
			Rectangle screenArea = Screen.PrimaryScreen.WorkingArea;
			frm.Bounds = new Rectangle(screenArea.Right - 150, screenArea.Top + (screenArea.Height - 300) / 2,
			                           140, 300);
			Service.Window.ManageWindowSize("MainForm", frm);
			return frm;
		}
		
		public virtual void AddContent(MainWindowCreationEventArgs e)
		{
			Panel panel = CreateContent(e);
			panel.Dock = DockStyle.Fill;
			e.Form.Controls.Add(panel);
		}
		
		public virtual Panel CreateContent(MainWindowCreationEventArgs e)
		{
			Panel panel = new Panel();
			ContactTreeView ctl = new ContactTreeView();
			ctl.Dock = DockStyle.Fill;
			panel.Controls.Add(ctl);
			ButtonRow ctl2 = Service.Menu.CreateButtonRow(this, "/MainForm/ButtonRow");
			ctl2.Dock = DockStyle.Bottom;
			panel.Controls.Add(ctl2);
			e.ContactTreeView = ctl;
			e.ButtonRow = ctl2;
			return panel;
		}
		
		public virtual void AddMenu(MainWindowCreationEventArgs e)
		{
			Service.Menu.CreateMainMenu(e.Form, this, "/MainForm/MainMenu");
		}
	}
	
	public sealed class DefaultMainWindowBuilder : AbstractMainWindowBuilder
	{
		public DefaultMainWindowBuilder() : base("Default", "${res:Settings.Chat.MainWindow.Default}") {}
	}
	
	public class MainWindowCreationEventArgs : EventArgs
	{
		Form form;
		
		public Form Form {
			get {
				return form;
			}
		}
		
		ChatPanel chatPanel;
		
		/// <summary>
		/// Gets/Sets the <see cref="ChatPanel"/> used. Can be null if the main window has no <see cref="ChatPanel"/>.
		/// </summary>
		public ChatPanel ChatPanel {
			get {
				return chatPanel;
			}
			set {
				chatPanel = value;
			}
		}
		
		ContactTreeView contactTreeView;
		
		/// <summary>
		/// Gets/Sets the <see cref="ContactTreeView"/> used. Can be null if the main window has no <see cref="ContactTreeView"/>.
		/// </summary>
		public ContactTreeView ContactTreeView {
			get {
				return contactTreeView;
			}
			set {
				contactTreeView = value;
			}
		}
		
		ButtonRow buttonRow;
		
		/// <summary>
		/// Gets/Sets the <see cref="ButtonRow"/> used. Can be null if the main window has no <see cref="ButtonRow"/>.
		/// </summary>
		public ButtonRow ButtonRow {
			get {
				return buttonRow;
			}
			set {
				buttonRow = value;
			}
		}
		
		public MainWindowCreationEventArgs(Form form)
		{
			if (form == null)
				throw new ArgumentNullException("form");
			this.form = form;
		}
	}
}
