/*
 * Copyright (c) 2004 MyXaml
 * All Rights Reserved
 *
 * Licensed under the terms of the GNU General Public License
 * http://www.gnu.org/licenses/licenses.html#GPL
 */

using System;
using System.Collections;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.Reflection;
using System.Runtime.Serialization;
using System.Windows.Forms;
using System.Xml;

using DGNetwork.Services;
using ICSharpCode.Core.Services;

namespace DGNetwork.Gui.MycroXaml
{
	#region Parser
	public class Parser
	{
		protected Hashtable nsMap;
		protected Hashtable objectCollection;
		ArrayList originalTexts = new ArrayList();
		ArrayList updatableControls = new ArrayList();
		protected object eventSink;
		
		public void UpdateTexts() {
			StringParserService sps = Service.StringParser;
			for (int i = 0; i < originalTexts.Count; i++) {
				object o = updatableControls[i];
				Control ctl = o as Control;
				if (ctl != null) {
					ctl.Text = sps.Parse((string)originalTexts[i]);
				} else {
					PropertyInfo prop = o.GetType().GetProperty("Text");
					prop.SetValue(o, sps.Parse((string)originalTexts[i]), null);
				}
			}
		}
		
		public object Load(XmlDocument doc, object eventSink)
		{
			this.eventSink=eventSink;
			objectCollection=new Hashtable();
			
			object ret = null;
			XmlNode node = doc.DocumentElement;
			if(node.ChildNodes.Count != 1) throw new MycroXamlException("Only one child of the root is allowed.");
			// The last valid node instantiated is returned.
			// The xml root should only have one child.
			ProcessNamespaces(node);
			ret=ProcessNode(node.ChildNodes[0], null);
			return ret;
		}
		
		public object GetInstance(string name)
		{
			return objectCollection[name];
		}
		
		public void AddInstance(string name, object obj)
		{
			// We don't care if we overwrite an existing object.
			objectCollection[name]=obj;
		}
		
		protected void ProcessNamespaces(XmlNode node)
		{
			nsMap=new Hashtable();
			foreach(XmlAttribute attr in node.Attributes) {
				if (attr.Prefix=="xmlns") {
					nsMap[attr.LocalName]=attr.Value;
				}
			}
		}
		
		private Assembly GetAssembly(string assemblyName) {
			foreach (Assembly a in AppDomain.CurrentDomain.GetAssemblies()) {
				string name = StringHelpers.LeftOf(a.FullName, ',');
				if (name == assemblyName)
					return a;
			}
			throw new MycroXamlException("Assembly '" + assemblyName + "' not found!");
		}
		
		protected object ProcessNode(XmlNode node, object parent)
		{
			object ret = null;
			if (node is XmlElement) {
				if (node.Name.ToLower() == "breakpoint") {
					System.Diagnostics.Debugger.Break();
					return null;
				}
				// instantiate the class
				string ns = node.Prefix;
				string cname = node.LocalName;
				if (!nsMap.Contains(ns))
					throw new MycroXamlException("Namespace '"+ns+"' has not been declared.");
				string asyName = (string)nsMap[ns];
				string qname = StringHelpers.LeftOf(asyName, ',') + "." + cname;
				Assembly assembly = GetAssembly(StringHelpers.RightOf(asyName, ',').Trim());
				Type t = assembly.GetType(qname, false);
				if (t == null) {
					Service.Message.ShowErrorFormatted("${res:Global.ClassNotFound}", qname);
					return null;
				}
				try {
					ret = Activator.CreateInstance(t);
				} catch(Exception e) {
					throw new MycroXamlException("Type "+qname+" could not be instantiated", e);
				}
				
				// support the ISupportInitialize interface
				if (ret is ISupportInitialize) {
					((ISupportInitialize)ret).BeginInit();
				}
				
				ProcessAttributes(node, ret, t);
				// implements the class-property-class model
				ProcessChildProperties(node, ret);
				
				// support the ISupportInitialize interface
				if (ret is ISupportInitialize) {
					((ISupportInitialize)ret).EndInit();
				}
			}
			return ret;
		}
		
		protected void ProcessChildProperties(XmlNode node, object parent)
		{
			Type t = parent.GetType();
			
			// children of a class must always be properties
			foreach(XmlNode child in node.ChildNodes) {
				if (child is XmlElement) {
					string pname = child.Name;
					PropertyInfo pi = t.GetProperty(pname);
					
					if (pi == null) {
						Service.Message.ShowErrorFormatted("${res:MycroXaml.UnknownChildProperty}", child.Name, parent.GetType().FullName);
						continue;
					}
					
					// a property can only have one child node unless it's a collection
					foreach(XmlNode grandChild in child.ChildNodes) {
						if (grandChild is XmlElement) {
							object propObject = pi.GetValue(parent, null);
							object obj = ProcessNode(grandChild, propObject);
							
							// A null return is valid in cases where a class implementing the IMicroXaml interface
							// might want to take care of managing the instance it creates itself.  See DataBinding
							if (obj != null) {
								
								// support for ICollection objects
								if (!pi.CanWrite) {
									if (propObject is ICollection) {
										MethodInfo mi = t.GetMethod("Add", new Type[] {obj.GetType()});
										if (mi != null) {
											try {
												mi.Invoke(obj, new object[] {obj});
											} catch(Exception e) {
												throw new MycroXamlException("Adding to collection failed:\r\n"+e.Message);
											}
										} else if (propObject is IList) {
											try {
												((IList)propObject).Add(obj);
											} catch(Exception e) {
												throw new MycroXamlException("List/Collection add failed:\r\n"+e.Message);
											}
										}
									} else {
										throw new MycroXamlException("Unsupported read-only property: "+pname);
									}
								} else {
									// direct assignment if not a collection
									try {
										pi.SetValue(parent, obj, null);
									} catch(Exception e) {
										throw new MycroXamlException("Property setter for "+pname+" failed:\r\n"+e.Message);
									}
								}
							}
						}
					}
				}
			}
		}
		
		protected void ProcessAttributes(XmlNode node, object ret, Type t)
		{
			// process attributes
			foreach (XmlAttribute attr in node.Attributes) {
				string pname = attr.Name;
				string pvalue = attr.Value;
				pvalue = Service.StringParser.Parse(pvalue);
				
				// it's either a property or an event
				PropertyInfo pi = t.GetProperty(pname);
				EventInfo ei = t.GetEvent(pname);
				
				if (pi != null) {
					// it's a property!
					if (pvalue.StartsWith("{") && pvalue.EndsWith("}")) {
						// And the value is a reference to an instance!
						// Get the referenced object.  Late binding is not supported!
						object val = GetInstance(pvalue.Substring(1, pvalue.Length-2));
						try {
							pi.SetValue(ret, val, null);
						} catch(Exception e) {
							throw new MycroXamlException("Couldn't set property "+pname+" to an instance of "+pvalue, e);
						}
					} else {
						object val = null;
						bool ok = false;
						if (pi.PropertyType.IsAssignableFrom(typeof(System.Drawing.Icon))) {
							try {
								val = Service.Resource.GetIcon(pvalue);
								ok = true;
							} catch (ResourceNotFoundException ex) {
								Service.Message.ShowError(ex.Message);
							}
						} else if (pi.PropertyType.IsAssignableFrom(typeof(System.Drawing.Bitmap))) {
							try {
								val = Service.Resource.GetBitmap(pvalue);
								ok = true;
							} catch (ResourceNotFoundException ex) {
								Service.Message.ShowError(ex.Message);
							}
						} else {
							// it's string, so use a type converter.
							TypeConverter tc=TypeDescriptor.GetConverter(pi.PropertyType);
							if (tc.CanConvertFrom(typeof(string))) {
								val = tc.ConvertFrom(null, CultureInfo.InvariantCulture, pvalue);
								ok = true;
							}
						}
						if (ok) {
							try {
								pi.SetValue(ret, val, null);
							} catch(Exception e) {
								throw new MycroXamlException("Property setter for "+pname+" failed", e);
							}
						}
					}
					
					// auto-add to our object collection
					if (pname == "Name") {
						AddInstance(pvalue, ret);
						if (eventSink != null) {
							Type sinkType = eventSink.GetType();
							FieldInfo field = sinkType.GetField(pvalue, BindingFlags.IgnoreCase |
							                                    BindingFlags.Instance | BindingFlags.SetField |
							                                    BindingFlags.Public | BindingFlags.NonPublic);
							if (field != null) {
								if (field.FieldType.IsAssignableFrom(ret.GetType())) {
									try {
										field.SetValue(eventSink, ret);
									} catch (Exception ex) {
										throw new MycroXamlException("Sink field setter for "+pvalue+" failed", ex);
									}
								} else {
									Service.Message.ShowErrorFormatted("${res:MycroXaml.SinkFieldWrongType}",
									                                   eventSink.ToString(),
									                                   field.Name,
									                                   field.FieldType.FullName,
									                                   ret.GetType().FullName);
								}
							}
						}
					}
					if (pname == "Text" && attr.Value.IndexOf("${") >= 0) {
						originalTexts.Add(attr.Value);
						updatableControls.Add(ret);
					}
				} else if (ei != null) {
					Delegate dlgt = null;
					// it's an event!
					if (pvalue.StartsWith("sink.")) {
						// use the event sink
						if (eventSink == null)
							throw new MycroXamlException("An event sink is needed but was not specified");
						string sinkname = pvalue.Substring(5);
						try {
							MethodInfo mi=eventSink.GetType().GetMethod(sinkname, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Static);
							if (mi == null) {
								Service.Message.ShowErrorFormatted("${res:MycroXaml.EventNotInSink}", sinkname, eventSink.ToString());
							} else {
								dlgt=Delegate.CreateDelegate(ei.EventHandlerType, eventSink, mi.Name);
							}
						} catch(Exception e) {
							throw new MycroXamlException("Couldn't create a delegate for the event "+pvalue+":\r\n"+e.Message);
						}
					} else if (pvalue.IndexOf(',') > 0) {
						Assembly assembly = GetAssembly(StringHelpers.RightOf(pvalue, ',').Trim());
						string qname = StringHelpers.LeftOf(pvalue, ',');
						Type qType = assembly.GetType(qname, false);
						if(qType == null) {
							Service.Message.ShowErrorFormatted("${res:MycroXaml.CommandNotFound}", qname);
						} else {
							IXamlCommand command;
							try {
								command=(IXamlCommand)Activator.CreateInstance(qType);
							} catch(Exception e) {
								throw new MycroXamlException("Type "+qname+" could not be instantiated", e);
							}
							command.XmlControl = xmlControl;
							dlgt = new EventHandler(command.Handle);
						}
					} else {
						string qname = "DGNetwork.Commands.Xaml." + pvalue;
						IXamlCommand command = null;
						foreach (Assembly asm in AppDomain.CurrentDomain.GetAssemblies()) {
							Type qType = asm.GetType(qname, false);
							if (qType == null) continue;
							try {
								command=(IXamlCommand)Activator.CreateInstance(qType);
							} catch(Exception e) {
								throw new MycroXamlException("Type "+pvalue+" could not be instantiated", e);
							}
						}
						if (command == null) {
							Service.Message.ShowErrorFormatted("${res:MycroXaml.CommandNotFound}", qname);
						} else {
							command.XmlControl = xmlControl;
							dlgt = new EventHandler(command.Handle);
						}
					}
					
					try {
						if (dlgt != null)
							ei.AddEventHandler(ret, dlgt);
					} catch(Exception e) {
						throw new MycroXamlException("Binding to event "+pname+" failed: "+e.Message);
					}
				} else {
					// who knows what it is???
					Service.Message.ShowErrorFormatted("${res:MycroXaml.UnknownAttribute}", pname, t.FullName);
				}
			}
		}
		
		internal XmlControl xmlControl;
	}
	#endregion
	
	#region StringHelpers
	/// <summary>
	/// Helpers for string manipulation.
	/// </summary>
	internal class StringHelpers
	{
		/// <summary>
		/// Left of the first occurance of c
		/// </summary>
		/// <param name="src"></param>
		/// <param name="c"></param>
		/// <returns></returns>
		public static string LeftOf(string src, char c)
		{
			int idx=src.IndexOf(c);
			if (idx==-1)
			{
				return src;
			}
			
			return src.Substring(0, idx);
		}
		
		/// <summary>
		/// Right of the first occurance of c
		/// </summary>
		/// <param name="src"></param>
		/// <param name="c"></param>
		/// <returns></returns>
		public static string RightOf(string src, char c)
		{
			int idx=src.IndexOf(c);
			if (idx==-1)
			{
				return "";
			}
			
			return src.Substring(idx+1);
		}
	}
	#endregion
	
	#region MycroXamlException
	/// <summary>
	/// This exception is thrown on all MycroXaml errors.
	/// </summary>
	[Serializable()]
	public class MycroXamlException : BaseException
	{
		/// <summary>
		/// Creates a new MycroXamlException.
		/// </summary>
		public MycroXamlException()
			: base()
		{
		}
		
		/// <summary>
		/// Creates a new MycroXamlException.
		/// </summary>
		public MycroXamlException(string message)
			: base(message)
		{
		}
		
		/// <summary>
		/// Creates a new MycroXamlException.
		/// </summary>
		public MycroXamlException(string message, Exception innerException)
			: base(message, innerException)
		{
		}
		
		/// <summary>
		/// Creates a new MycroXamlException.
		/// </summary>
		protected MycroXamlException(SerializationInfo info, StreamingContext context)
			: base(info, context)
		{
		}
	}
	#endregion
}
