// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace DGNetwork.Gui
{
	/// <summary>
	/// Wrapper class around an element in a tree view.
	/// </summary>
	public class TreeElement : ICloneable
	{
		#region Constructor
		public TreeElement() {
			children = new TreeElementCollection(this);
		}
		#endregion
		
		#region Element properties and events
		string text;
		bool canLabelEdit;
		string icon;
		
		public string Text {
			get {
				return text;
			}
			set {
				if (text == value)
					return;
				text = value;
				OnTextChanged(EventArgs.Empty);
			}
		}
		
		public bool CanLabelEdit {
			get {
				return canLabelEdit;
			}
			set {
				canLabelEdit = value;
			}
		}
		
		public string Icon {
			get {
				return icon;
			}
			set {
				if (icon == value)
					return;
				icon = value;
				OnIconChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler TextChanged;
		public event EventHandler IconChanged;
		
		protected virtual void OnTextChanged(EventArgs e) {
			if (TextChanged != null) {
				TextChanged(this, e);
			}
		}
		
		protected virtual void OnIconChanged(EventArgs e) {
			if (IconChanged != null) {
				IconChanged(this, e);
			}
		}
		#endregion
		
		#region Expand/Collapse
		bool expanded;
		
		public bool Expanded {
			get {
				return expanded;
			}
			set {
				if (expanded == value)
					return;
				expanded = value;
				OnExpandedChanged(EventArgs.Empty);
			}
		}
		
		public event EventHandler ExpandedChanged;
		
		protected virtual void OnExpandedChanged(EventArgs e) {
			if (ExpandedChanged != null) {
				ExpandedChanged(this, e);
			}
		}
		#endregion
		
		#region Children
		TreeElementCollection children;
		
		public TreeElementCollection Children {
			get {
				return children;
			}
		}
		
		/// <summary>
		/// Occurs whenever a child has been added.
		/// </summary>
		public event TreeElementEventHandler ChildAdded;
		
		/// <summary>
		/// Occurs whenever a child has been removed.
		/// </summary>
		public event TreeElementEventHandler ChildRemoved;
		
		/// <summary>
		/// Occurs when after the children collection has changed.
		/// </summary>
		public event EventHandler ChildrenChanged;
		
		protected virtual void OnChildAdded(TreeElementEventArgs e) {
			if (ChildAdded != null) {
				ChildAdded(this, e);
			}
		}
		
		protected virtual void OnChildRemoved(TreeElementEventArgs e) {
			if (ChildRemoved != null) {
				ChildRemoved(this, e);
			}
		}
		
		protected virtual void OnChildrenChanged(EventArgs e) {
			if (ChildrenChanged != null) {
				ChildrenChanged(this, e);
			}
		}
		
		bool updating = false;
		bool dirty    = false;
		
		internal void BeginUpdate() {
			if (updating)
				throw new InvalidOperationException();
			updating = true;
			dirty = false;
		}
		
		internal void EndUpdate() {
			if (!updating)
				throw new InvalidOperationException();
			updating = false;
			if (dirty)
				OnChildrenChanged(EventArgs.Empty);
		}
		
		internal void RaiseAdded(TreeElement e, int index) {
			OnChildAdded(new TreeElementEventArgs(e, index));
			if (updating)
				dirty = true;
			else
				OnChildrenChanged(EventArgs.Empty);
		}
		
		internal void RaiseRemoved(TreeElement e) {
			OnChildRemoved(new TreeElementEventArgs(e));
			if (updating)
				dirty = true;
			else
				OnChildrenChanged(EventArgs.Empty);
		}
		#endregion
		
		#region ICloneable interface implementation
		/// <summary>
		/// Creates a new TreeElement copying all values from the specified TreeElement.
		/// </summary>
		public TreeElement(TreeElement e) : this()
		{
			this.canLabelEdit = e.canLabelEdit;
			this.icon = e.icon;
			this.text = e.text;
		}
		
		public virtual TreeElement Clone() {
			return new TreeElement(this);
		}
		
		object ICloneable.Clone() {
			return this.Clone();
		}
		#endregion
	}
	
	#region TreeElementEventHandler + TreeElementEventArgs
	public delegate void TreeElementEventHandler(object sender, TreeElementEventArgs e);
	
	public class TreeElementEventArgs : EventArgs
	{
		TreeElement element;
		
		public TreeElement Element {
			get {
				return element;
			}
		}
		
		int index;
		
		/// <summary>
		/// Gets the index the element was inserted at.
		/// This property is -1 if the element was deleted.
		/// </summary>
		public int Index {
			get {
				return index;
			}
		}
		
		public TreeElementEventArgs(TreeElement element) {
			this.element = element;
			this.index = -1;
		}
		
		public TreeElementEventArgs(TreeElement element, int index) {
			this.element = element;
			this.index = index;
		}
	}
	#endregion
}
