// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace DGNetwork.Gui
{
	/// <summary>
	///   A collection that stores <see cref='TreeElement'/> objects.
	/// </summary>
	[Serializable()]
	public class TreeElementCollection : CollectionBase
	{
		TreeElement parent;
		
		internal TreeElementCollection(TreeElement parent) {
			this.parent = parent;
		}
		
		/// <summary>
		///   Represents the entry at the specified index of the <see cref='TreeElement'/>.
		/// </summary>
		/// <param name='index'>The zero-based index of the entry to locate in the collection.</param>
		/// <value>The entry at the specified index of the collection.</value>
		/// <exception cref='ArgumentOutOfRangeException'><paramref name='index'/> is outside the valid range of indexes for the collection.</exception>
		public TreeElement this[int index] {
			get {
				return ((TreeElement)(List[index]));
			}
			set {
				TreeElement old = (TreeElement)List[index];
				if (old == value) return;
				if (Contains(value))
					throw new ArgumentException("The specified element is already added to the collection.");
				try {
					parent.BeginUpdate();
					parent.RaiseRemoved(old);
					List[index] = value;
					parent.RaiseAdded(value, index);
				} finally {
					parent.EndUpdate();
				}
			}
		}
		
		/// <summary>
		///   Adds a <see cref='TreeElement'/> with the specified value to the
		///   <see cref='TreeElementCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='TreeElement'/> to add.</param>
		/// <returns>The index at which the new element was inserted.</returns>
		/// <seealso cref='TreeElementCollection.AddRange'/>
		public int Add(TreeElement val)
		{
			if (Contains(val))
				throw new ArgumentException("The specified element is already added to the collection.");
			int pos = List.Add(val);
			parent.RaiseAdded(val, pos);
			return pos;
		}
		
		/// <summary>
		///   Copies the elements of an array to the end of the <see cref='TreeElementCollection'/>.
		/// </summary>
		/// <param name='val'>
		///    An array of type <see cref='TreeElement'/> containing the objects to add to the collection.
		/// </param>
		/// <seealso cref='TreeElementCollection.Add'/>
		public void AddRange(TreeElement[] val)
		{
			try {
				parent.BeginUpdate();
				for (int i = 0; i < val.Length; i++) {
					this.Add(val[i]);
				}
			} finally {
				parent.EndUpdate();
			}
		}
		
		/// <summary>
		///   Adds the contents of another <see cref='TreeElementCollection'/> to the end of the collection.
		/// </summary>
		/// <param name='val'>
		///    A <see cref='TreeElementCollection'/> containing the objects to add to the collection.
		/// </param>
		/// <seealso cref='TreeElementCollection.Add'/>
		public void AddRange(TreeElementCollection val)
		{
			try {
				parent.BeginUpdate();
				for (int i = 0; i < val.Count; i++) {
					this.Add(val[i]);
				}
			} finally {
				parent.EndUpdate();
			}
		}
		
		/// <summary>
		///   Gets a value indicating whether the
		///    <see cref='TreeElementCollection'/> contains the specified <see cref='TreeElement'/>.
		/// </summary>
		/// <param name='val'>The <see cref='TreeElement'/> to locate.</param>
		/// <returns>
		/// <see langword='true'/> if the <see cref='TreeElement'/> is contained in the collection;
		///   otherwise, <see langword='false'/>.
		/// </returns>
		/// <seealso cref='TreeElementCollection.IndexOf'/>
		public bool Contains(TreeElement val)
		{
			return List.Contains(val);
		}
		
		/// <summary>
		///   Copies the <see cref='TreeElementCollection'/> values to a one-dimensional <see cref='Array'/> instance at the
		///    specified index.
		/// </summary>
		/// <param name='array'>The one-dimensional <see cref='Array'/> that is the destination of the values copied from <see cref='TreeElementCollection'/>.</param>
		/// <param name='index'>The index in <paramref name='array'/> where copying begins.</param>
		/// <exception cref='ArgumentException'>
		///   <para><paramref name='array'/> is multidimensional.</para>
		///   <para>-or-</para>
		///   <para>The number of elements in the <see cref='TreeElementCollection'/> is greater than
		///         the available space between <paramref name='arrayIndex'/> and the end of
		///         <paramref name='array'/>.</para>
		/// </exception>
		/// <exception cref='ArgumentNullException'><paramref name='array'/> is <see langword='null'/>. </exception>
		/// <exception cref='ArgumentOutOfRangeException'><paramref name='arrayIndex'/> is less than <paramref name='array'/>'s lowbound. </exception>
		/// <seealso cref='Array'/>
		public void CopyTo(TreeElement[] array, int index)
		{
			List.CopyTo(array, index);
		}
		
		/// <summary>
		///    Returns the index of a <see cref='TreeElement'/> in
		///       the <see cref='TreeElementCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='TreeElement'/> to locate.</param>
		/// <returns>
		///   The index of the <see cref='TreeElement'/> of <paramref name='val'/> in the
		///   <see cref='TreeElementCollection'/>, if found; otherwise, -1.
		/// </returns>
		/// <seealso cref='TreeElementCollection.Contains'/>
		public int IndexOf(TreeElement val)
		{
			return List.IndexOf(val);
		}
		
		/// <summary>
		///   Inserts a <see cref='TreeElement'/> into the <see cref='TreeElementCollection'/> at the specified index.
		/// </summary>
		/// <param name='index'>The zero-based index where <paramref name='val'/> should be inserted.</param>
		/// <param name='val'>The <see cref='TreeElement'/> to insert.</param>
		/// <seealso cref='TreeElementCollection.Add'/>
		public void Insert(int index, TreeElement val)
		{
			if (Contains(val))
				throw new ArgumentException("The specified element is already added to the collection.");
			List.Insert(index, val);
			parent.RaiseAdded(val, index);
		}
		
		/// <summary>
		///   Removes a specific <see cref='TreeElement'/> from the <see cref='TreeElementCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='TreeElement'/> to remove from the <see cref='TreeElementCollection'/>.</param>
		/// <exception cref='ArgumentException'><paramref name='val'/> is not found in the Collection.</exception>
		public void Remove(TreeElement val)
		{
			List.Remove(val);
			parent.RaiseRemoved(val);
		}
	}
}
