// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Reflection;
using System.Security.Cryptography;
using System.Text;
using ICSharpCode.Core.AddIns;
using DGNetwork.Services;

namespace DGNetwork
{
	public class Helper
	{
		#region Random
		static Random random;
		/// <summary>
		/// Returns an instance of System.Random. This property is initialized on
		/// first call.<br/>
		/// Warning: This random number generator can not be used for generating
		/// cryptographic keys!
		/// </summary>
		public static Random Random {
			get {
				if (random == null) random = new Random();
				return random;
			}
		}
		#endregion
		
		#region IsCritical(Exception)
		/// <summary>
		/// Checks if the exception is a critical (=non handable) exception that
		/// should not be caught.
		/// </summary>
		public static bool IsCritical(Exception ex)
		{
			if (ex is OutOfMemoryException)
				return true;
			if (ex is AppDomainUnloadedException)
				return true;
			if (ex is BadImageFormatException)
				return true;
			if (ex is CannotUnloadAppDomainException)
				return true;
			if (ex is ExecutionEngineException)
				return true;
			if (ex is InvalidProgramException)
				return true;
			return false;
		}
		#endregion
		
		#region Escape
		
		// NOTE: I know System.Web.HttpUtility, but Microsoft has a bug
		//  regarding the character '#'. Read the RFC and compare it to the
		//  result of this method and Microsoft's method.
		//
		// The character "#" is unsafe and should
		// always be encoded because it is used in World Wide Web and in other
		// systems to delimit a URL from a fragment/anchor identifier that might
		// follow it.
		//
		// WHIDBEY: In .NET 2.0 Beta this bug is fixed so this methods will
		// become obsolete.
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// a part of an URL.
		/// (";", "/", "?", ":", "@", "=" and "&amp;" are encoded)</para>
		/// <para>Non-ascii characters are encoded using UTF-8.</para>
		/// </summary>
		public static string EscapePart(string str)
		{
			return Grunwald.P2PLib.Helper.EscapePart(str);
		}
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// a part of an URL.
		/// (";", "/", "?", ":", "@", "=" and "&amp;" are encoded)</para>
		/// <para>Non-ascii characters are encoded using the specified encoding.</para>
		/// </summary>
		public static string EscapePart(string str, Encoding encoding)
		{
			return Grunwald.P2PLib.Helper.EscapePart(str, encoding);
		}
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// URL. Scheme-specific characters for URLs are not encoded.</para>
		/// <para>Non-ascii characters are encoded using UTF-8.</para>
		/// </summary>
		public static string Escape(string str)
		{
			return Grunwald.P2PLib.Helper.Escape(str);
		}
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// URL. Scheme-specific characters for URLs are not encoded.</para>
		/// <para>Non-ascii characters are encoded using the specified encoding.</para>
		/// </summary>
		public static string Escape(string str, Encoding encoding)
		{
			return Grunwald.P2PLib.Helper.Escape(str, encoding);
		}
		
		/// <summary>
		/// Escapes a (part of an) URL as described in RFC 1738. Special characters
		/// are encoded using UTF-8.
		/// </summary>
		public static string Escape(string str, string allowedChars)
		{
			return Grunwald.P2PLib.Helper.Escape(str, allowedChars);
		}
		
		/// <summary>
		/// Escapes a (part of an) URL as described in RFC 1738. Special characters
		/// are encoded using the specified encoding.
		/// </summary>
		public static string Escape(string str, string allowedChars, Encoding encoding)
		{
			return Grunwald.P2PLib.Helper.Escape(str, allowedChars, encoding);
		}
		
		#endregion
		
		#region Unescape
		
		/// <summary>
		/// Unescapes a (part of an) URL as described in RFC 1738. Special characters
		/// are decoded using UTF8.
		/// </summary>
		public static string Unescape(string str)
		{
			return Unescape(str, Encoding.UTF8);
		}
		
		/// <summary>
		/// Unescapes a (part of an) URL as described in RFC 1738. Special characters
		/// are decoded using the specified encoding.
		/// </summary>
		public static string Unescape(string str, Encoding e)
		{
			return Grunwald.P2PLib.Helper.Unescape(str, e);
		}
		
		#endregion
		
		public static string ToHex(byte[] data)
		{
			StringBuilder b = new StringBuilder();
			for (int i = 0; i < data.Length; ++i) {
				b.Append(data[i].ToString("x2"));
			}
			return b.ToString();
		}
		
		#region MD5
		static MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
		
		public static string MD5(string text)
		{
			return MD5(Encoding.UTF8.GetBytes(text));
		}
		
		public static string MD5(byte[] data)
		{
			return ToHex(MD5Bytes(data));
		}
		
		public static string MD5(Stream stream)
		{
			return ToHex(MD5Bytes(stream));
		}
		
		public static byte[] MD5Bytes(byte[] data)
		{
			return md5.ComputeHash(data);
		}
		
		public static byte[] MD5Bytes(Stream stream)
		{
			return md5.ComputeHash(stream);
		}
		#endregion
		
		#region SHA1
		static SHA1CryptoServiceProvider sha1 = new SHA1CryptoServiceProvider();
		
		public static string SHA1(string text)
		{
			return SHA1(Encoding.UTF8.GetBytes(text));
		}
		
		public static string SHA1(byte[] data)
		{
			return ToHex(SHA1Bytes(data));
		}
		
		public static string SHA1(Stream stream)
		{
			return ToHex(SHA1Bytes(stream));
		}
		
		public static byte[] SHA1Bytes(byte[] data)
		{
			return sha1.ComputeHash(data);
		}
		
		public static byte[] SHA1Bytes(Stream stream)
		{
			return sha1.ComputeHash(stream);
		}
		#endregion
		
		public static Type GetType(string name)
		{
			switch(name.ToLower()) {
				case "string":
					return typeof(string);
				case "long":
					return typeof(long);
				case "int":
					return typeof(int);
				case "byte":
					return typeof(byte);
				case "bool":
					return typeof(bool);
				case "object":
					return typeof(object);
				default:
					foreach (Assembly asm in AppDomain.CurrentDomain.GetAssemblies()) {
						Type qType = asm.GetType(name, false, true);
						if (qType != null)
							return qType;
					}
					return null;
			}
		}
		
		public static object BuildItem(string addInTreePath, object owner)
		{
			IAddInTreeNode node;
			try {
				node = AddInTreeSingleton.AddInTree.GetTreeNode(addInTreePath);
			} catch (TreePathNotFoundException) {
				Service.Message.ShowErrorFormatted("${res:Global.TreePathNotFound}", addInTreePath);
				return null;
			}
			return node.Codon.BuildItem(null, node.BuildChildItems(owner), node.ConditionCollection);
		}
		
		public static ArrayList BuildItems(string addInTreePath, object owner)
		{
			IAddInTreeNode node;
			try {
				node = AddInTreeSingleton.AddInTree.GetTreeNode(addInTreePath);
			} catch (TreePathNotFoundException) {
				Service.Message.ShowErrorFormatted("${res:Global.TreePathNotFound}", addInTreePath);
				return new ArrayList();
			}
			return node.BuildChildItems(owner);
		}
		
		public static string GetTempPath()
		{
			string temp = Path.Combine(Path.GetTempPath(), "DG Network");
			if (!Directory.Exists(temp))
				Directory.CreateDirectory(temp);
			if (temp.EndsWith("\\") || temp.EndsWith("/"))
				return temp;
			else
				return temp + "/";
		}
	}
}
