// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace DGNetwork.Search
{
	/// <summary>
	///   A collection that stores <see cref='ICriteria'/> objects.
	/// </summary>
	[Serializable()]
	public class CriteriaCollection : CollectionBase
	{
		
		/// <summary>
		///   Initializes a new instance of <see cref='CriteriaCollection'/>.
		/// </summary>
		public CriteriaCollection() {
		}
		
		/// <summary>
		///   Initializes a new instance of <see cref='CriteriaCollection'/> based on another <see cref='CriteriaCollection'/>.
		/// </summary>
		/// <param name='val'>
		///   A <see cref='CriteriaCollection'/> from which the contents are copied
		/// </param>
		public CriteriaCollection(CriteriaCollection val) {
			this.AddRange(val);
		}
		
		/// <summary>
		///   Initializes a new instance of <see cref='CriteriaCollection'/> containing any array of <see cref='ICriteria'/> objects.
		/// </summary>
		/// <param name='val'>
		///       A array of <see cref='ICriteria'/> objects with which to intialize the collection
		/// </param>
		public CriteriaCollection(ICriteria[] val)
		{
			this.AddRange(val);
		}
		
		/// <summary>
		///   Represents the entry at the specified index of the <see cref='ICriteria'/>.
		/// </summary>
		/// <param name='index'>The zero-based index of the entry to locate in the collection.</param>
		/// <value>The entry at the specified index of the collection.</value>
		/// <exception cref='ArgumentOutOfRangeException'><paramref name='index'/> is outside the valid range of indexes for the collection.</exception>
		public ICriteria this[int index] {
			get {
				return ((ICriteria)(List[index]));
			}
			set {
				List[index] = value;
			}
		}
		
		/// <summary>
		///   Gets the <see cref='ICriteria'/> with the specified name.
		///   Returns null if the criteria is not found.
		/// </summary>
		public ICriteria this[string name] {
			get {
				foreach (ICriteria c in List) {
					if (c.Name == name)
						return c;
				}
				return null;
			}
		}
		
		/// <summary>
		///   Gets whether this collection contains a <see cref='ICriteria'/> with the specified name.
		/// </summary>
		public bool Contains(string name)
		{
			return this[name] != null;
		}
		
		/// <summary>
		///   Adds a <see cref='ICriteria'/> with the specified value to the
		///   <see cref='CriteriaCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='ICriteria'/> to add.</param>
		/// <returns>The index at which the new element was inserted.</returns>
		/// <seealso cref='CriteriaCollection.AddRange'/>
		public int Add(ICriteria val)
		{
			return List.Add(val);
		}
		
		/// <summary>
		///   Copies the elements of an array to the end of the <see cref='CriteriaCollection'/>.
		/// </summary>
		/// <param name='val'>
		///    An array of type <see cref='ICriteria'/> containing the objects to add to the collection.
		/// </param>
		/// <seealso cref='CriteriaCollection.Add'/>
		public void AddRange(ICriteria[] val)
		{
			for (int i = 0; i < val.Length; i++) {
				this.Add(val[i]);
			}
		}
		
		/// <summary>
		///   Adds the contents of another <see cref='CriteriaCollection'/> to the end of the collection.
		/// </summary>
		/// <param name='val'>
		///    A <see cref='CriteriaCollection'/> containing the objects to add to the collection.
		/// </param>
		/// <seealso cref='CriteriaCollection.Add'/>
		public void AddRange(CriteriaCollection val)
		{
			for (int i = 0; i < val.Count; i++) {
				this.Add(val[i]);
			}
		}
		
		/// <summary>
		///   Gets a value indicating whether the
		///    <see cref='CriteriaCollection'/> contains the specified <see cref='ICriteria'/>.
		/// </summary>
		/// <param name='val'>The <see cref='ICriteria'/> to locate.</param>
		/// <returns>
		/// <see langword='true'/> if the <see cref='ICriteria'/> is contained in the collection;
		///   otherwise, <see langword='false'/>.
		/// </returns>
		/// <seealso cref='CriteriaCollection.IndexOf'/>
		public bool Contains(ICriteria val)
		{
			return List.Contains(val);
		}
		
		/// <summary>
		///   Copies the <see cref='CriteriaCollection'/> values to a one-dimensional <see cref='Array'/> instance at the
		///    specified index.
		/// </summary>
		/// <param name='array'>The one-dimensional <see cref='Array'/> that is the destination of the values copied from <see cref='CriteriaCollection'/>.</param>
		/// <param name='index'>The index in <paramref name='array'/> where copying begins.</param>
		/// <exception cref='ArgumentException'>
		///   <para><paramref name='array'/> is multidimensional.</para>
		///   <para>-or-</para>
		///   <para>The number of elements in the <see cref='CriteriaCollection'/> is greater than
		///         the available space between <paramref name='arrayIndex'/> and the end of
		///         <paramref name='array'/>.</para>
		/// </exception>
		/// <exception cref='ArgumentNullException'><paramref name='array'/> is <see langword='null'/>. </exception>
		/// <exception cref='ArgumentOutOfRangeException'><paramref name='arrayIndex'/> is less than <paramref name='array'/>'s lowbound. </exception>
		/// <seealso cref='Array'/>
		public void CopyTo(ICriteria[] array, int index)
		{
			List.CopyTo(array, index);
		}
		
		/// <summary>
		///    Returns the index of a <see cref='ICriteria'/> in
		///       the <see cref='CriteriaCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='ICriteria'/> to locate.</param>
		/// <returns>
		///   The index of the <see cref='ICriteria'/> of <paramref name='val'/> in the
		///   <see cref='CriteriaCollection'/>, if found; otherwise, -1.
		/// </returns>
		/// <seealso cref='CriteriaCollection.Contains'/>
		public int IndexOf(ICriteria val)
		{
			return List.IndexOf(val);
		}
		
		/// <summary>
		///   Inserts a <see cref='ICriteria'/> into the <see cref='CriteriaCollection'/> at the specified index.
		/// </summary>
		/// <param name='index'>The zero-based index where <paramref name='val'/> should be inserted.</param>
		/// <param name='val'>The <see cref='ICriteria'/> to insert.</param>
		/// <seealso cref='CriteriaCollection.Add'/>
		public void Insert(int index, ICriteria val)
		{
			List.Insert(index, val);
		}
		
		/// <summary>
		///   Removes a specific <see cref='ICriteria'/> from the <see cref='CriteriaCollection'/>.
		/// </summary>
		/// <param name='val'>The <see cref='ICriteria'/> to remove from the <see cref='CriteriaCollection'/>.</param>
		/// <exception cref='ArgumentException'><paramref name='val'/> is not found in the Collection.</exception>
		public void Remove(ICriteria val)
		{
			List.Remove(val);
		}
	}
}
