// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace DGNetwork.Search
{
	/// <summary>
	/// Represents a running search.
	/// </summary>
	public interface ISearch
	{
		void Abort();
		bool IsAborted { get; }
		event EventHandler Aborting;
		event EventHandler Aborted;
		
		ArrayList Results { get; }
		event ResultEventHandler ResultAdded;
		
		bool IsFinished { get; }
		event EventHandler Finished;
	}
	
	#region DefaultSearch
	public class DefaultSearch : ISearch
	{
		bool aborted = false;
		bool finished = false;
		ArrayList results;
		ArrayList resultsWrapper;
		
		public DefaultSearch() {
			results = new ArrayList();
			resultsWrapper = ArrayList.ReadOnly(results);
		}
		
		public bool IsAborted {
			get {
				return aborted;
			}
		}
		
		public bool IsFinished {
			get {
				return finished;
			}
		}
		
		public ArrayList Results {
			get {
				return resultsWrapper;
			}
		}
		
		public virtual void Abort()
		{
			if (aborted || finished) return;
			OnAborting(EventArgs.Empty);
			aborted = true;
			OnAborted(EventArgs.Empty);
		}
		
		public void AddResult(IResult result)
		{
			lock (resultsWrapper) {
				results.Add(result);
			}
			OnResultAdded(new ResultEventArgs(result));
		}
		
		public event ResultEventHandler ResultAdded;
		public event EventHandler Aborting;
		public event EventHandler Aborted;
		public event EventHandler Finished;
		
		protected virtual void OnResultAdded(ResultEventArgs e) {
			if (ResultAdded != null) {
				ResultAdded(this, e);
			}
		}
		
		protected virtual void OnAborting(EventArgs e) {
			if (Aborting != null) {
				Aborting(this, e);
			}
		}
		
		protected virtual void OnAborted(EventArgs e) {
			if (Aborted != null) {
				Aborted(this, e);
			}
		}
		
		protected virtual void OnFinished(EventArgs e) {
			if (Finished != null) {
				Finished(this, e);
			}
		}
		
		public void Finish()
		{
			if (finished || aborted)
				return;
			finished = true;
			OnFinished(EventArgs.Empty);
		}
	}
	#endregion
	
	#region MultiSearch
	/// <summary>
	/// A search that gets its results from a list of searches.
	/// </summary>
	public class MultiSearch : DefaultSearch
	{
		ArrayList clients = new ArrayList();
		
		protected override void OnAborted(EventArgs e)
		{
			foreach (ISearch c in clients) {
				c.Abort();
			}
			base.OnAborted(e);
		}
		
		public void AddSearch(ISearch s)
		{
			if (clients.Contains(s))
				return;
			clients.Add(s);
			s.ResultAdded += new ResultEventHandler(ClientResultAdded);
			s.Finished += new EventHandler(ClientFinished);
			foreach (IResult result in s.Results) {
				this.AddResult(result);
			}
		}
		
		public void RemoveSearch(ISearch s)
		{
			if (!clients.Contains(s))
				return;
			s.ResultAdded -= new ResultEventHandler(ClientResultAdded);
			s.Finished -= new EventHandler(ClientFinished);
			clients.Remove(s);
		}
		
		private void ClientResultAdded(object sender, ResultEventArgs e)
		{
			this.AddResult(e.Result);
		}
		
		private void ClientFinished(object sender, EventArgs e)
		{
			this.FinishIfClientsAreFinished();
		}
		
		public void FinishIfClientsAreFinished()
		{
			foreach (ISearch s in clients) {
				if (!s.IsFinished && !s.IsAborted)
					return;
			}
			this.Finish();
		}
	}
	#endregion
}
