// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.IO;
using System.Windows.Forms;
using System.Collections;
using System.Threading;
using System.Resources;
using System.Drawing;
using System.Diagnostics;
using System.Reflection;
using System.Xml;

using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.AddIns.Codons;
using DGNetwork.AddIns.Codons;
using DGNetwork.Gui;
using DGNetwork.Gui.Menu;
using ICSharpCode.Core.Services;

namespace DGNetwork.Services
{
	public class MenuService : AbstractService
	{
		public ButtonRow CreateButtonRow(object owner, string addInTreePath)
		{
			try {
				return new ButtonRow(AddInTreeSingleton.AddInTree.GetTreeNode(addInTreePath), owner);
			} catch (TreePathNotFoundException) {
				Service.Message.ShowErrorFormatted("${res:Global.TreePathNotFound}", addInTreePath);
				return null;
			}
		}
		
		void ContextMenuPopupHandler(object sender, EventArgs e)
		{
			CommandBarContextMenu contextMenu = (CommandBarContextMenu)sender;
			foreach (object o in contextMenu.Items) {
				if (o is IStatusUpdate) {
					((IStatusUpdate)o).UpdateStatus();
				}
			}
		}
		
		public ContextMenu CreateContextMenu(IItemWithContextMenu[] items)
		{
			if (items.Length == 0)
				return null;
			if (items.Length == 1)
				return CreateContextMenu(items[0].Owner, items[0].ContextMenuTreePath);
			Service.Message.ShowError("ContextMenu for multiple items not implemented!");
			return null;
		}
		
		public void ShowContextMenu(IItemWithContextMenu[] items, Control parent, int x, int y)
		{
			ContextMenu menu = CreateContextMenu(items);
			if (menu != null)
				menu.Show(parent, new Point(x, y));
		}
		
		public ContextMenu CreateContextMenu(object owner, string addInTreePath)
		{
			if (addInTreePath == null || addInTreePath.Length == 0)
				return null;
			try {
				return CreateContextMenu(owner, AddInTreeSingleton.AddInTree.GetTreeNode(addInTreePath).BuildChildItems(owner));
			} catch (TreePathNotFoundException) {
				Service.Message.ShowErrorFormatted("${res:Global.TreePathNotFound}", addInTreePath);
				return null;
			}
		}
		
		public ContextMenu CreateContextMenu(object owner, ArrayList buildItems)
		{
			CommandBarContextMenu contextMenu = new CommandBarContextMenu();
			contextMenu.Popup += new EventHandler(ContextMenuPopupHandler);
			AddItems(owner, buildItems, contextMenu.Items);
			return contextMenu;
		}
		
		public void CreateMainMenu(Control parent, object owner, string addInTreePath)
		{
			try {
				CreateMainMenu(parent, owner, AddInTreeSingleton.AddInTree.GetTreeNode(addInTreePath).BuildChildItems(owner));
			} catch (TreePathNotFoundException) {
				Service.Message.ShowErrorFormatted("${res:Global.TreePathNotFound}", addInTreePath);
			}
		}
		
		public void CreateMainMenu(Control parent, object owner, ArrayList buildItems)
		{
			CommandBar bar = new CommandBar(CommandBarStyle.Menu);
			AddItems(owner, buildItems, bar.Items);
			foreach (CommandBarItem item in bar.Items) {
				if (item is CommandBarMenu && item is IStatusUpdate)
					((CommandBarMenu)item).DropDown += new EventHandler(ItemDropDown);
			}
			bar.Dock = DockStyle.Top;
			parent.Controls.Add(bar);
		}
		
		private void ItemDropDown(object sender, EventArgs e)
		{
			((IStatusUpdate)sender).UpdateStatus();
		}
		
		private void AddItems(object owner, ArrayList buildItems, CommandBarItemCollection items)
		{
			foreach (object item in buildItems) {
				if (item is CommandBarItem) {
					items.Add((CommandBarItem)item);
				} else {
					ISubmenuBuilder submenuBuilder = (ISubmenuBuilder)item;
					items.AddRange(submenuBuilder.BuildSubmenu(null, owner));
				}
			}
		}
		
		public void ShowContextMenu(object owner, string addInTreePath, Control parent, int x, int y)
		{
			ContextMenu menu = CreateContextMenu(owner, addInTreePath);
			if (menu != null)
				menu.Show(parent, new Point(x, y));
		}
		
		class QuickInsertMenuHandler
		{
			TextBoxBase targetControl;
			string      text;
			
			public QuickInsertMenuHandler(TextBoxBase targetControl, string text)
			{
				this.targetControl = targetControl;
				this.text          = text;
			}
			
			public EventHandler EventHandler {
				get {
					return new EventHandler(PopupMenuHandler);
				}
			}
			void PopupMenuHandler(object sender, EventArgs e)
			{
				targetControl.SelectedText += text;
			}
		}
		
		class QuickInsertHandler
		{
			Control               popupControl;
			CommandBarContextMenu quickInsertMenu;
			
			public QuickInsertHandler(Control popupControl, CommandBarContextMenu quickInsertMenu)
			{
				this.popupControl    = popupControl;
				this.quickInsertMenu = quickInsertMenu;
				
				popupControl.Click += new EventHandler(showQuickInsertMenu);
			}
			
			void showQuickInsertMenu(object sender, EventArgs e)
			{
				Point cords = new Point(popupControl.Width, 0);
				quickInsertMenu.Show(popupControl, cords);
			}
		}
		
		public void CreateQuickInsertMenu(TextBoxBase targetControl, Control popupControl, string[,] quickInsertMenuItems)
		{
			CommandBarContextMenu contextMenu = new CommandBarContextMenu();
			for (int i = 0; i < quickInsertMenuItems.GetLength(0); ++i) {
				if (quickInsertMenuItems[i, 0] == "-") {
					contextMenu.Items.Add(new SdMenuSeparator());
				} else {
					SdMenuCommand cmd = new SdMenuCommand(this,
					                                      Service.StringParser.Parse(quickInsertMenuItems[i, 0]),
					                                      new QuickInsertMenuHandler(targetControl, quickInsertMenuItems[i, 1]).EventHandler);
					contextMenu.Items.Add(cmd);
				}
			}
			new QuickInsertHandler(popupControl, contextMenu);
		}
	}
}
