// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.Collections;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using ICSharpCode.Core.Services;

namespace DGNetwork.Services
{
	public sealed class DefaultMessageService : AbstractService, IMessageService
	{
		#region ShowMessageBox
		static ArrayList buffer;
		
		private class PreparedMessageBox
		{
			string text;
			string title;
			MessageBoxButtons buttons;
			MessageBoxIcon icon;
			
			public PreparedMessageBox(string text, string title, MessageBoxButtons buttons, MessageBoxIcon icon)
			{
				this.text = text;
				this.title = title;
				this.buttons = buttons;
				this.icon = icon;
			}
			
			public void Show()
			{
				MessageBox.Show(text, title, buttons, icon);
			}
		}
		
		void ShowMessageBox(string text, string title, MessageBoxButtons buttons, MessageBoxIcon icon)
		{
			if (buffer == null) {
				MessageBox.Show(text, title, buttons, icon);
			} else {
				buffer.Add(new PreparedMessageBox(text, title, buttons, icon));
			}
		}
		
		public static void ActivateBuffer()
		{
			if (buffer == null)
				buffer = new ArrayList();
		}
		
		public static void ReleaseBuffer()
		{
			if (buffer != null) {
				ArrayList oldBuffer = buffer;
				buffer = null;
				foreach (PreparedMessageBox box in oldBuffer) {
					box.Show();
				}
			}
		}
		#endregion
		
		#region ShowError
		public void ShowError(Exception ex)
		{
			Grunwald.Gui.ErrorReporter.ShowErrorReport(ex);
		}
		
		public void ShowError(string message)
		{
			ShowError(null, message);
		}
		
		public void ShowErrorFormatted(string formatstring, params string[] formatitems)
		{
			ShowError(null, Format(formatstring, formatitems));
		}
		
		public void ShowError(Exception ex, string message)
		{
			string msg = String.Empty;
			
			if (message != null) {
				msg += message;
			}
			
			if (message != null && ex != null) {
				msg += "\n\n";
			}
			
			if (ex != null) {
				msg += "${res:Global.ExceptionOccured}: " + ex.ToString();
			}
			
			ShowMessageBox(Service.StringParser.Parse(msg), Service.StringParser.Parse("${res:Global.ErrorTitle}"), MessageBoxButtons.OK, MessageBoxIcon.Error);
		}
		#endregion
		
		#region ShowWarning
		public void ShowWarning(string message)
		{
			ShowMessageBox(Service.StringParser.Parse(message), Service.StringParser.Parse("${res:Global.WarningTitle}"), MessageBoxButtons.OK, MessageBoxIcon.Warning);
		}
		
		public void ShowWarningFormatted(string formatstring, params string[] formatitems)
		{
			ShowWarning(Format(formatstring, formatitems));
		}
		#endregion
		
		#region AskQuestion
		public bool AskQuestion(string question, string caption)
		{
			return MessageBox.Show(Service.StringParser.Parse(question), Service.StringParser.Parse(caption), MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes;
		}
		
		public bool AskQuestionFormatted(string caption, string formatstring, params string[] formatitems)
		{
			return AskQuestion(String.Format(Service.StringParser.Parse(formatstring), formatitems), caption);
		}
		
		public bool AskQuestionFormatted(string formatstring, params string[] formatitems)
		{
			return AskQuestion(String.Format(Service.StringParser.Parse(formatstring), formatitems));
		}
		
		public bool AskQuestion(string question)
		{
			return AskQuestion(Service.StringParser.Parse(question), Service.StringParser.Parse("${res:Global.QuestionTitle}"));
		}
		#endregion
		
		#region ShowCustomDialog
		public int ShowCustomDialog(string caption, string dialogText, int acceptButtonIndex, int cancelButtonIndex, params string[] buttontexts)
		{
			using (CustomDialog messageBox = new CustomDialog(caption, dialogText, acceptButtonIndex, cancelButtonIndex, buttontexts)) {
				messageBox.ShowDialog();
				return messageBox.Result;
			}
		}
		
		public int ShowCustomDialog(string caption, string dialogText, params string[] buttontexts)
		{
			return ShowCustomDialog(caption, dialogText, -1, -1, buttontexts);
		}
		#endregion
		
		#region ShowMessage
		public void ShowMessage(string message)
		{
			ShowMessage(message, "${res:Global.MessageTitle}");
		}
		
		string Format(string formatstring, string[] formatitems)
		{
			try {
				return String.Format(Service.StringParser.Parse(formatstring), formatitems);
			} catch (FormatException) {
				StringBuilder b = new StringBuilder(Service.StringParser.Parse(formatstring));
				foreach(string formatitem in formatitems) {
					b.Append("\nItem: ");
					b.Append(formatitem);
				}
				return b.ToString();
			}
		}
		
		public void ShowMessageFormatted(string formatstring, params string[] formatitems)
		{
			ShowMessage(Format(formatstring, formatitems));
		}
		
		public void ShowMessageFormatted(string caption, string formatstring, params string[] formatitems)
		{
			ShowMessage(Format(formatstring, formatitems), caption);
		}
		
		public void ShowMessage(string message, string caption)
		{
			ShowMessageBox(Service.StringParser.Parse(message), Service.StringParser.Parse(caption), MessageBoxButtons.OK, MessageBoxIcon.Information);
		}
		#endregion
		
		#region ShowInputBox
		public string ShowInputBox(string text, string caption, string defaultValue)
		{
			using (InputBox inputBox = new InputBox(text, caption, defaultValue)) {
				inputBox.ShowDialog();
				return inputBox.Result;
			}
		}
		#endregion
		
		#region Play Sound
		#if MONO
		#region Mono implementation
		public void PlaySound(string file)
		{
			
		}
		#endregion
		#else
		#region Windows implementation
		// PlaySound()
		[DllImport("winmm.dll", CallingConvention = CallingConvention.Winapi)]
		static extern bool PlaySound(string pszSound,
		                             IntPtr hMod,
		                             SoundFlags sf);
		
		// Flags for playing sounds.  For this example, we are reading
		// the sound from a filename, so we need only specify
		// SND_FILENAME | SND_ASYNC
		[Flags]
		public enum SoundFlags
		{
			SND_SYNC = 0x0000, // play synchronously (default)
			SND_ASYNC = 0x0001, // play asynchronously
			SND_NODEFAULT = 0x0002, // silence (!default) if sound not found
			SND_MEMORY = 0x0004, // pszSound points to a memory file
			SND_LOOP = 0x0008, // loop the sound until next sndPlaySound
			SND_NOSTOP = 0x0010, // don't stop any currently playing sound
			SND_NOWAIT = 0x00002000, // don't wait if the driver is busy
			SND_ALIAS = 0x00010000, // name is a registry alias
			SND_ALIAS_ID = 0x00110000, // alias is a predefined ID
			SND_FILENAME = 0x00020000, // name is file name
			SND_RESOURCE = 0x00040004 // name is resource name or atom
		}
		
		public void PlaySound(string file)
		{
			try {
				string fn = System.IO.Path.Combine(Service.Property.DataDirectory, file);
				fn = fn.Replace('/', '\\');
				// play the sound from the selected filename
				if (!PlaySound(fn, IntPtr.Zero,
				               SoundFlags.SND_FILENAME | SoundFlags.SND_ASYNC)) {
					// ignore errors (e.g. when no sound card is installed)
				}
			} catch (System.IO.FileNotFoundException) {
				Service.Message.ShowWarning("Sound file " + file + " not found!");
			} catch (Exception ex) {
				Service.Message.ShowError(ex);
			}
		}
		#endregion
		#endif
		#endregion
	}
}
