// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using DGNetwork.DGN;
using DGNetwork.AddIns;
using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.Services;
using ICSharpCode.Core.Properties;
using Grunwald.P2PLib;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;

namespace DGNetwork.Services
{
	public sealed class P2PService : AbstractService
	{
		HttpServer server;
		PropertyLink link;
		ServerAbilityTester serverAbilityTester;
		PingRequestHandler pingRequestHandler;
		
		public const int DefaultPort = 3336;
		public const string PortPropertyKey = "P2PService.Server.Port";
		
		public override void InitializeService()
		{
			serverAbilityTester = new ServerAbilityTester();
			Service.Timer.Tick += new EventHandler(TimerTick);
			server = new HttpServer();
			server.Port = DefaultPort;
			server.ServerAbilityTester = serverAbilityTester;
			pingRequestHandler = new PingRequestHandler();
			server.RequestHandlers.Add(pingRequestHandler);
			CreateDefaultRequestHandlers();
			link = Service.Property.CreateLink(PortPropertyKey, server, "Port");
			Service.Chat.StatusChanged += new DGNetwork.Chat.StatusChangedEventHandler(ChatStatusChanged);
			base.InitializeService();
		}
		
		void ChatStatusChanged(object sender, DGNetwork.Chat.StatusChangedEventArgs e)
		{
			if (e.NewStatus == "Offline") {
				if (server.Running)
					server.Stop();
			} else {
				StartServer();
			}
		}
		
		void StartServer()
		{
			if (server.Running)
				return;
			try {
				server.Start();
			} catch (System.Net.Sockets.SocketException) {
				int oldPort = server.Port;
				server.Port = Helper.Random.Next(40000) + 5000;
				try {
					server.Start();
					Service.Message.ShowWarningFormatted("${res:P2PService.PortNotAvailable}", oldPort.ToString(), server.Port.ToString());
				} catch (System.Net.Sockets.SocketException) {
					throw;
				}
			}
		}
		
		private void CreateDefaultRequestHandlers()
		{
			IAddInTreeNode node = AddInTreeSingleton.AddInTree.GetTreeNode("/Server");
			ArrayList subItems = node.BuildChildItems(this);
			foreach (object o in subItems) {
				if (o is IRequestHandler) {
					server.RequestHandlers.Add((IRequestHandler)o);
				} else if (o is IConnectionHandler) {
					server.ConnectionHandlers.Add((IConnectionHandler)o);
				} else {
					Service.Message.ShowErrorFormatted("${res:Codon.InvalidSubItem}",
					                                   "Server",
					                                   o.GetType().FullName);
				}
			}
		}
		
		public override void UnloadService()
		{
			link.Dispose();
			base.UnloadService();
			if (server.Running)
				server.Stop();
		}
		
		public HttpServer Server {
			get {
				return server;
			}
		}
		
		public ServerAbilityTester ServerAbilityTester {
			get {
				return serverAbilityTester;
			}
		}
		
		private void TimerTick(object sender, EventArgs e)
		{
			if ((Service.Timer.Ticks % 300) == 15) {
				RefreshServerAbility();
			}
		}
		
		/// <summary>
		/// Begins executing a new server ability test.
		/// </summary>
		public void RefreshServerAbility()
		{
			if (server.IPAddress != null && server.Running) {
				serverAbilityTester.BeginTest(server.InternetEndPoint, pingRequestHandler.ServerId);
			}
		}
	}
}
