// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.IO;
using System.Windows.Forms;
using System.Collections;
using System.Threading;
using System.Globalization;
using System.Resources;
using System.Drawing;
using System.Diagnostics;
using System.Reflection;
using System.Xml;

using ICSharpCode.Core.Services;
using ICSharpCode.Core.Properties;

namespace DGNetwork.Services
{
	/// <summary>
	/// This Class contains two ResourceManagers, which handle string and image resources
	/// for the application. It do handle localization strings on this level.
	/// </summary>
	[Obsolete("You should not directly use this service class, use the interface!")]
	public class DefaultResourceService : AbstractService, IResourceService
	{
		#region Members
		readonly static string uiLanguageProperty = "CoreProperties.UILanguage";
		readonly static string stringResources  = "StringResources";
		readonly static string imageResources   = "BitmapResources";
		
		static string resourceDirectory;
		
		public DefaultResourceService()
		{
			resourceDirectory = Service.Property.DataDirectory + "/resources";
			if (!Directory.Exists(resourceDirectory))
				throw new DirectoryNotFoundException("Resource directory not found: " + resourceDirectory);
			RegisterAssembly(Assembly.GetEntryAssembly());
			LoadLanguageResources();
		}
		
		Hashtable localStrings = new Hashtable();
		Hashtable localIcons   = new Hashtable();
		
		ArrayList strings = new ArrayList();
		ArrayList icon    = new ArrayList();
		
		ArrayList assemblies = new ArrayList();
		#endregion
		
		#region LoadLanguageResources
		void ChangeProperty(object sender, PropertyEventArgs e)
		{
			if (e.Key == uiLanguageProperty && e.OldValue != e.NewValue) {
				LoadLanguageResources();
				OnReloaded(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Occurs when the the resource files were reloaded and the resource strings
		/// (and maybe bitmaps) have to be re-read.
		/// </summary>
		public event EventHandler Reloaded;
		
		protected virtual void OnReloaded(EventArgs e)
		{
			if (Reloaded != null) {
				Reloaded(this, e);
			}
		}
		
		void LoadLanguageResources()
		{
			string language = Service.Property.GetProperty(uiLanguageProperty, Thread.CurrentThread.CurrentUICulture.Name);
			try {
				Thread.CurrentThread.CurrentUICulture = new CultureInfo(language);
			} catch {
				try {
					Thread.CurrentThread.CurrentUICulture = new CultureInfo(language.Split('-')[0]);
				} catch {}
			}
			
			localStrings.Clear();
			localIcons.Clear();
			
			while(true) {
				Load(localStrings, stringResources + "." + language);
				Load(localIcons, imageResources + "." + language);
				int pos = language.LastIndexOf('-');
				if (pos < 0)
					break;
				language = language.Substring(0, pos);
			}
			
			Load(localStrings, stringResources);
			Load(localIcons, imageResources);
		}
		
		void Load(Hashtable resources, string shortName)
		{
			string fileName = resourceDirectory + "/" + shortName + ".resources";
			if (File.Exists(fileName)) {
				ResourceReader rr = new ResourceReader(fileName);
				foreach (DictionaryEntry entry in rr) {
					if (resources.ContainsKey(entry.Key))
						continue;
					resources[entry.Key] = entry.Value;
				}
				rr.Close();
			}
		}
		#endregion
		
		#region Initialize
		/// <summary>
		/// Take resources from an assembly and merge them in the resource service
		/// </summary>
		public void RegisterAssembly(Assembly assembly)
		{
			if (assemblies.Contains(assembly))
				throw new BaseException("This assembly already is registered");
			assemblies.Add(assembly);
			
			if (assembly.GetManifestResourceInfo(stringResources + ".resources") != null) {
				Console.WriteLine("Loading strings from " + assembly.FullName);
				strings.Add(new ResourceManager(stringResources, assembly));
			}
			if (assembly.GetManifestResourceInfo(imageResources + ".resources") != null) {
				Console.WriteLine("Loading images from " + assembly.FullName);
				icon.Add(new ResourceManager(imageResources, assembly));
			}
		}
		
		public override void InitializeService()
		{
			base.InitializeService();
			Service.Property.PropertyChanged += new PropertyEventHandler(ChangeProperty);
		}
		#endregion
		
		#region GetResourceStream
		public Stream GetResourceStream(string name, string nameSpace) {
			string fullName = nameSpace + "." + name;
			Stream s;
			foreach (Assembly assembly in assemblies) {
				s = assembly.GetManifestResourceStream(fullName);
				if (s != null)
					return s;
			}
			foreach (Assembly assembly in assemblies) {
				s = assembly.GetManifestResourceStream(name);
				if (s != null)
					return s;
			}
			throw new ResourceNotFoundException(fullName + " (resource)");
		}
		#endregion
		
		#region GetResource
		/// <summary>
		/// Returns a string from the resource database, it handles localization
		/// transparent for the user.
		/// </summary>
		/// <returns>
		/// The string in the (localized) resource database.
		/// </returns>
		/// <param name="name">
		/// The name of the requested resource.
		/// </param>
		/// <exception cref="ResourceNotFoundException">
		/// Is thrown when the GlobalResource manager can't find a requested resource.
		/// </exception>
		public string GetString(string name)
		{
			return GetString(name, true);
		}
		
		/// <summary>
		/// Returns a string from the resource database, it handles localization
		/// transparent for the user.
		/// </summary>
		public string GetString(string name, bool throwOnError)
		{
			if (localStrings != null && localStrings[name] != null) {
				return localStrings[name].ToString();
			}
			
			string s = null;
			
			if (s == null) {
				foreach (ResourceManager resourceManger in strings) {
					s = resourceManger.GetString(name);
					if (s != null) {
						break;
					}
				}
			}
			if (s == null) {
				if (throwOnError)
					throw new ResourceNotFoundException(name + " (string)");
				else
					return "->" + name + "<-";
			}
			return s;
		}
		
		object GetImageResource(string name)
		{
			name = Service.Parse(name);
			object iconobj = null;
			if (localIcons != null && localIcons[name] != null) {
				iconobj = localIcons[name];
			} else {
				foreach (ResourceManager resourceManger in icon) {
					iconobj = resourceManger.GetObject(name);
					if (iconobj != null) {
						break;
					}
				}
			}
			if (iconobj == null) {
				throw new ResourceNotFoundException(name + " (image)");
			}
			return iconobj;
		}
		
		/// <summary>
		/// Returns a icon from the resource database, it handles localization
		/// transparent for the user. In the resource database can be a bitmap
		/// instead of an icon in the dabase. It is converted automatically.
		/// </summary>
		/// <returns>
		/// The icon in the (localized) resource database.
		/// </returns>
		/// <param name="name">
		/// The name of the requested icon.
		/// </param>
		/// <exception cref="ResourceNotFoundException">
		/// Is thrown when the GlobalResource manager can't find a requested resource.
		/// </exception>
		public Icon GetIcon(string name)
		{
			object iconobj = GetImageResource(name);
			if (iconobj is Icon) {
				return (Icon)iconobj;
			} else {
				return Icon.FromHandle(((Bitmap)iconobj).GetHicon());
			}
		}
		
		/// <summary>
		/// Returns a bitmap from the resource database, it handles localization
		/// transparent for the user.
		/// </summary>
		/// <returns>
		/// The bitmap in the (localized) resource database.
		/// </returns>
		/// <param name="name">
		/// The name of the requested bitmap.
		/// </param>
		/// <exception cref="ResourceNotFoundException">
		/// Is thrown when the GlobalResource manager can't find a requested resource.
		/// </exception>
		public Bitmap GetBitmap(string name)
		{
			object o = GetImageResource(name);
			if (o is Icon)
				return Grunwald.Gui.Components.IconLoader.IconToBitmap((Icon)o);
			else
				return (Bitmap)o;
		}
		#endregion
	}
}
