// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using DGNetwork.Contacts;
using DGNetwork.Chat;
using DGNetwork.Gui;
using ICSharpCode.Core.Services;

namespace DGNetwork.Services
{
	public class WindowService : AbstractService
	{
		Control invokeHelper;
		
		public override void InitializeService()
		{
			invokeHelper = new Control();
			invokeHelper.CreateControl();
			LoadMainWindowBuilders();
			base.InitializeService();
		}
		
		public override void UnloadService()
		{
			hash.Clear();
			base.UnloadService();
		}
		
		#region ManageWindowSize
		Hashtable hash = new Hashtable(); // Form -> name
		
		/// <summary>
		/// Manages the size of the window: Loads the initial size from the property system;
		/// stores any size changes in the property system.
		/// </summary>
		/// <param name="name">The name under which the form bounds are saved.</param>
		/// <param name="form">The form itself</param>
		public void ManageWindowSize(string name, Form form)
		{
			hash[form] = name;
			RestoreWindowSize(name, form);
			form.Closed  += new EventHandler(FormClosed);
			form.Move += new EventHandler(BoundsChanged);
			form.Resize += new EventHandler(BoundsChanged);
		}
		
		/// <summary>
		/// Restores the stored size of the window.
		/// </summary>
		public void RestoreWindowSize(string name, Form form)
		{
			Rectangle def = form.Bounds;
			switch (form.StartPosition) {
				case FormStartPosition.CenterScreen:
					Rectangle s = Screen.FromPoint(Control.MousePosition).WorkingArea;
					def = new Rectangle(s.Left + (s.Width - def.Width) / 2,
					                    s.Top + (s.Height - def.Height) / 2,
					                    def.Width, def.Height);
					break;
			}
			bool defMaximized = form.WindowState == FormWindowState.Maximized;
			form.StartPosition = FormStartPosition.Manual;
			form.WindowState = FormWindowState.Normal;
			def = (Rectangle)Service.Property.GetProperty("Forms." + name + ".Bounds", def);
			form.Bounds = MoveToVisibleScreenArea(def);
			if (Service.Property.GetProperty("Forms." + name + ".Maximized", defMaximized))
				form.WindowState = FormWindowState.Maximized;
		}
		
		/// <summary>
		/// Moves the form back to the visible screen area so that a form cannot remember its position
		/// on the secondary monitor when there isn't a secondary monitor.
		/// </summary>
		public Rectangle MoveToVisibleScreenArea(Rectangle def)
		{
			// Check if form is outside the screen and get it back if necessary.
			Rectangle screen1 = Screen.FromPoint(new Point(def.X, def.Y)).WorkingArea;
			Rectangle screen2 = Screen.FromPoint(new Point(def.X + def.Width, def.Y)).WorkingArea;
			if (def.Y < screen1.Y - 5 && def.Y < screen2.Y - 5)
				def.Y = screen1.Y - 5;
			if (def.X < screen1.X - def.Width / 2)
				def.X = screen1.X - def.Width / 2;
			else if (def.X > screen2.Right - def.Width / 2)
				def.X = screen2.Right - def.Width / 2;
			return def;
		}
		
		private void BoundsChanged(object sender, EventArgs e)
		{
			Form frm = (Form)sender;
			if (frm.WindowState == FormWindowState.Minimized)
				return;
			Service.Property.SetProperty("Forms." + hash[frm] + ".Maximized",
			                             frm.WindowState == FormWindowState.Maximized);
			if (frm.WindowState == FormWindowState.Normal)
				Service.Property.SetProperty("Forms." + hash[frm] + ".Bounds", frm.Bounds);
		}
		
		private void FormClosed(object sender, EventArgs e)
		{
			hash.Remove(sender);
		}
		#endregion
		
		#region Invoke
		public IAsyncResult BeginInvoke(Delegate method)
		{
			return invokeHelper.BeginInvoke(method);
		}
		
		public IAsyncResult BeginInvoke(Delegate method, params object[] args)
		{
			return invokeHelper.BeginInvoke(method, args);
		}
		
		public object EndInvoke(IAsyncResult result)
		{
			return invokeHelper.EndInvoke(result);
		}
		
		public object Invoke(Delegate method)
		{
			return invokeHelper.Invoke(method);
		}
		
		public object Invoke(Delegate method, params object[] args)
		{
			return invokeHelper.Invoke(method, args);
		}
		
		public bool InvokeRequired {
			get {
				return invokeHelper.InvokeRequired;
			}
		}
		#endregion
		
		#region MainWindowBuilder
		ArrayList mainWindowBuilders;
		IMainWindowBuilder activeMainWindowBuilder;
		
		public ArrayList MainWindowBuilders {
			get {
				return mainWindowBuilders;
			}
		}
		
		public event EventHandler ActiveMainWindowBuilderChanged;
		
		public IMainWindowBuilder ActiveMainWindowBuilder {
			get {
				return activeMainWindowBuilder;
			}
			set {
				if (activeMainWindowBuilder == value)
					return;
				activeMainWindowBuilder = value;
				Service.Property.SetProperty("WindowService.ActiveMainWindowBuilder", value.Name);
				if (ActiveMainWindowBuilderChanged != null)
					ActiveMainWindowBuilderChanged(this, EventArgs.Empty);
				if (mainWindow != null) {
					if (InvokeRequired)
						Invoke(new MethodInvoker(mainWindow.Close));
					else
						mainWindow.Close();
					ShowMainWindow();
				}
			}
		}
		
		private void LoadMainWindowBuilders()
		{
			string name;
			mainWindowBuilders = Helper.BuildItems("/MainForm/Builder", this);
			if (mainWindowBuilders.Count == 0)
				mainWindowBuilders.Add(new DefaultMainWindowBuilder());
			activeMainWindowBuilder = (IMainWindowBuilder)mainWindowBuilders[0];
			name = Service.Property.GetProperty("WindowService.ActiveMainWindowBuilder", activeMainWindowBuilder.Name);
			foreach (IMainWindowBuilder builder in mainWindowBuilders) {
				if (builder.Name == name) {
					activeMainWindowBuilder = builder;
					break;
				}
			}
		}
		#endregion
		
		#region Main Window
		Form mainWindow;
		
		/// <summary>
		/// Gets the main window, or null if it is not open.
		/// When the main window is not open, this property can be set.
		/// </summary>
		public Form MainWindow {
			get {
				return mainWindow;
			}
			set {
				if (mainWindow != null)
					throw new InvalidOperationException("MainWindow cannot be set while it is open!");
				mainWindow = value;
			}
		}
		
		/// <summary>
		/// Shows the main window.
		/// </summary>
		public void ShowMainWindow()
		{
			if (this.InvokeRequired) {
				this.BeginInvoke(new MethodInvoker(ShowMainWindow));
				return;
			}
			if (mainWindow != null) {
				mainWindow.Activate();
				return;
			}
			Service.Contact.Root.Update();
			MainWindowCreationEventArgs e = activeMainWindowBuilder.Create();
			Form frm = e.Form;
			frm.Closed += new EventHandler(MainWindowClosed);
			mainWindow = frm;
			frm.Show();
		}
		
		private void MainWindowClosed(object sender, EventArgs e)
		{
			mainWindow = null;
		}
		#endregion
		
		#region Flash
		#if MONO
		#region Mono implementation
		public void Flash(Form form)
		{
			
		}
		#endregion
		#else
		#region Windows implementation
		[StructLayout(LayoutKind.Sequential, Pack = 4)]
		private struct FLASHWINFO
		{
			public int cbSize;
			public IntPtr hwnd;
			public int dwFlags;
			public int uCount;
			public int dwTimeout;
		}
		
		[DllImport("user32")]
		private extern static int FlashWindow (
		                                       IntPtr hWnd,
		                                       ref FLASHWINFO pwfi);
		/// <summary>
		/// Stop flashing. The system restores the window to its original state
		/// </summary>
		private const int FLASHW_STOP = 0;
		/// <summary>
		/// Flash the window caption.
		/// </summary>
		private const int FLASHW_CAPTION = 0x00000001;
		/// <summary>
		/// Flash the taskbar button.
		/// </summary>
		private const int FLASHW_TRAY = 0x00000002;
		/// <summary>
		/// Flash both the window caption and taskbar button.
		/// </summary>
		private const int FLASHW_ALL = (FLASHW_CAPTION | FLASHW_TRAY);
		/// <summary>
		/// Flash continuously, until the FLASHW_STOP flag is set.
		/// </summary>
		private const int FLASHW_TIMER = 0x00000004;
		/// <summary>
		/// Flash continuously until the window comes to the foreground.
		/// </summary>
		private const int FLASHW_TIMERNOFG = 0x0000000C;
		
		public void Flash(Form form)
		{
			FLASHWINFO info = new FLASHWINFO();;
			info.dwFlags = FLASHW_ALL;
			info.dwTimeout = 60;
			info.hwnd = form.Handle;
			info.uCount = 1;
			info.cbSize = Marshal.SizeOf(info);
			FlashWindow(form.Handle, ref info);
		}
		#endregion
		#endif
		#endregion
	}
}
