// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Xml;
using Grunwald.Gui.Settings;
using DGNetwork.Services;
using ICSharpCode.Core.Properties;

namespace DGNetwork.Settings
{
	/// <summary>
	/// This setting is bound to a core property.
	/// </summary>
	public class PropertySetting : DefaultSetting
	{
		string propertyName;
		object defaultValue;
		Type valueType;
		
		public PropertySetting(string name, string displayName, string description, string propertyName, object defaultValue)
			: base(name, displayName, description, Service.Property.GetProperty(propertyName, defaultValue))
		{
			if (defaultValue == null)
				throw new ArgumentNullException("defaultValue");
			this.propertyName = propertyName;
			this.defaultValue = defaultValue;
			this.valueType = defaultValue.GetType();
			Service.Property.PropertyChanged += new PropertyEventHandler(ServicePropertyChanged);
		}
		
		public PropertySetting(string name, string displayName, string description, string propertyName, Type valueType)
			: base(name, displayName, description, Service.Property.GetProperty(propertyName, valueType))
		{
			this.propertyName = propertyName;
			this.valueType = valueType;
			Service.Property.PropertyChanged += new PropertyEventHandler(ServicePropertyChanged);
		}
		
		private void ServicePropertyChanged(object sender, PropertyEventArgs e) {
			if (e.Key == propertyName) {
				this.OnValueChanged(new SettingChangedEventArgs(this, e.OldValue, e.NewValue));
				if (!Object.Equals(e.NewValue, this.InEditValue))
					this.InEditValue = e.NewValue;
			}
		}
		
		public override object Value {
			get {
				return Service.Property.GetProperty(propertyName, defaultValue);
			}
			set {
				try {
					Service.Property.SetProperty(propertyName, value);
				} catch (System.Reflection.TargetInvocationException ex) {
					this.InEditValue = this.Value;
					Service.Message.ShowErrorFormatted("${res:Settings.ErrorOnSave}", this.DisplayName, ex.InnerException.Message);
				}
			}
		}
		
		/// <summary>
		/// The type of the value of this setting.
		/// </summary>
		public Type ValueType {
			get {
				return valueType;
			}
		}
		
		/// <summary>
		/// Creates a new ISettingEditor to edit this setting.
		/// </summary>
		public override ISettingEditor CreateEditor(EditorFactory factory)
		{
			if (factory == null) throw new ArgumentNullException("factory");
			return factory.GetEditor(this, this.ValueType);
		}
		
		public override void Save(XmlWriter writer) {
			// don't need to save because the value is saved with the property system
		}
		
		protected override void SaveValue(XmlWriter writer) {
			// don't need to save because the value is saved with the property system
		}
		
		public override void Load(XmlReader reader) {
			// don't need to load because the value is loaded with the property system
			reader.Skip();
		}
	}
}
