// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.Collections;
using System.ComponentModel;
using System.IO;
using System.Diagnostics;
using System.Xml;
using System.Windows.Forms;
using System.Reflection;

namespace ICSharpCode.Core.Properties
{
	/// <summary>
	/// Default <see cref="IProperties"/> implementation, should
	/// be enough for most cases :)
	/// </summary>
	public class DefaultProperties : IProperties
	{
		public DefaultProperties()
		{
		}
		
		Hashtable properties = new Hashtable();
		
		#region GetProperty
		/// <summary>
		/// Gets a property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <param name="defaultvalue"/>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		public object GetProperty(string key, object defaultvalue)
		{
			object obj = properties[key];
			
			if (obj == null) {
				if (defaultvalue != null) {
					properties[key] = defaultvalue;
					OnPropertyChanged(new PropertyEventArgs(this, key, obj, defaultvalue));
				}
				return defaultvalue;
			}
			
			if (defaultvalue == null)
				return obj;
			
			// stored an XmlElement in properties node ->
			// set a FromXmlElement of the defaultvalue type at this propertyposition.
			if (defaultvalue is IXmlConvertable && obj is XmlElement) {
				obj = ((IXmlConvertable)defaultvalue).FromXmlElement((XmlElement)((XmlElement)obj).FirstChild);
			} else if (defaultvalue is Array && obj is XmlElement) {
				Type elementType = defaultvalue.GetType().GetElementType();
				XmlElement xml = (XmlElement)obj;
				TypeConverter c = TypeDescriptor.GetConverter(elementType);
				Array arr = Array.CreateInstance(elementType, xml.ChildNodes.Count);
				for (int i = 0; i < arr.Length; ++i) {
					arr.SetValue(c.ConvertFromInvariantString(xml.ChildNodes[i].InnerText), i);
				}
				obj = arr;
			} else if (obj is string) {
				if (defaultvalue is string)
					return obj;
				TypeConverter c = TypeDescriptor.GetConverter(defaultvalue.GetType());
				obj = c.ConvertFromInvariantString((string)obj);
			}
			properties[key] = obj;
			return obj;
		}
		
		/// <summary>
		/// Gets a property of the specified type out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <c>null</c>, if the property wasn't found.
		/// </returns>
		public object GetProperty(string key, Type type)
		{
			object obj = properties[key];
			
			if (obj == null)
				return null;
			if (type.IsInstanceOfType(obj))
				return obj;
			
			if (!(obj is string))
				throw new CoreException("Property is not string and not the needed type.");
			
			TypeConverter c = TypeDescriptor.GetConverter(type);
			obj = c.ConvertFromInvariantString((string)obj);
			properties[key] = obj;
			return obj;
		}
		
		/// <summary>
		/// Gets a property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <c>null</c>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		public object GetProperty(string key)
		{
			return GetProperty(key, (object)null);
		}
		
		/// <summary>
		/// Gets a <c>int</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <param name="defaultvalue"/>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		public int GetProperty(string key, int defaultvalue)
		{
			return (int)GetProperty(key, (object)defaultvalue);
		}
		
		/// <summary>
		/// Gets a <c>bool</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <param name="defaultvalue"/>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		public bool GetProperty(string key, bool defaultvalue)
		{
			return (bool)GetProperty(key, (object)defaultvalue);
		}
		
		/// <summary>
		/// Gets a <c>short</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <param name="defaultvalue"/>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		public short GetProperty(string key, short defaultvalue)
		{
			return (short)GetProperty(key, (object)defaultvalue);
		}
		
		/// <summary>
		/// Gets a <c>byte</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <param name="defaultvalue"/>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		public byte GetProperty(string key, byte defaultvalue)
		{
			return (byte)GetProperty(key, (object)defaultvalue);
		}
		
		/// <summary>
		/// Gets a <c>string</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <param name="defaultvalue"/>, if the property wasn't found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		public string GetProperty(string key, string defaultvalue)
		{
			return (string)GetProperty(key, (object)defaultvalue);
		}
		#endregion
		
		#region SetProperty
		/// <summary>
		/// Sets the property <param name="key"/> to the value <param name="val"/>.
		/// If <param name="val"/> is null, the property will be taken out from the
		/// properties.
		/// </summary>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="val">
		/// The value of the property.
		/// </param>
		public void SetProperty(string key, object val)
		{
			object oldValue = properties[key];
			if (!object.Equals(val, oldValue)) {
				properties[key] = val;
				ArrayList l = (ArrayList)links[key];
				if (l != null) {
					foreach (PropertyLink pl in l) {
						pl.SetValue(val);
					}
				}
				OnPropertyChanged(new PropertyEventArgs(this, key, oldValue, val));
			}
		}
		#endregion
		
		#region To/FromXml
		protected void SetValueFromXmlElement(XmlElement element)
		{
			XmlNodeList nodes = element.ChildNodes;
			foreach (XmlElement el in nodes) {
				if (el.Name == "Property") {
					properties[el.Attributes["key"].InnerText] = el.Attributes["value"].InnerText;
				} else if (el.Name == "XmlConvertableProperty") {
					properties[el.Attributes["key"].InnerText] = el;
				} else if (el.Name == "ArrayProperty") {
					properties[el.Attributes["key"].InnerText] = el;
				} else {
					throw new UnknownPropertyNodeException(el.Name);
				}
			}
		}
		
		/// <summary>
		/// Converts a <see cref="XmlElement"/> to a <see cref="DefaultProperties"/> object.
		/// </summary>
		/// <returns>
		/// A new <see cref="DefaultProperties"/> object
		/// </returns>
		public virtual object FromXmlElement(XmlElement element)
		{
			DefaultProperties defaultProperties = new DefaultProperties();
			defaultProperties.SetValueFromXmlElement(element);
			return defaultProperties;
		}
		
		/// <summary>
		/// Converts the <see cref="DefaultProperties"/> object to a <see cref="XmlElement"/>.
		/// </summary>
		/// <returns>
		/// A new <see cref="XmlElement"/> object which represents the state
		/// of the <see cref="DefaultProperties"/> object.
		/// </returns>
		public virtual XmlElement ToXmlElement(XmlDocument doc)
		{
			XmlElement propertiesnode  = doc.CreateElement("Properties");
			
			foreach (DictionaryEntry entry in properties) {
				if (entry.Value != null) {
					if (entry.Value is XmlElement) { // write unchanged XmlElement back
						propertiesnode.AppendChild(doc.ImportNode((XmlElement)entry.Value, true));
					} else if (entry.Value is IXmlConvertable) { // An Xml convertable object
						XmlElement convertableNode = doc.CreateElement("XmlConvertableProperty");
						
						XmlAttribute key = doc.CreateAttribute("key");
						key.InnerText = entry.Key.ToString();
						convertableNode.Attributes.Append(key);
						
						convertableNode.AppendChild(((IXmlConvertable)entry.Value).ToXmlElement(doc));
						
						propertiesnode.AppendChild(convertableNode);
					} else if (entry.Value is Array) {
						XmlElement el = doc.CreateElement("ArrayProperty");
						XmlAttribute key   = doc.CreateAttribute("key");
						key.InnerText      = entry.Key.ToString();
						el.Attributes.Append(key);
						TypeConverter c = TypeDescriptor.GetConverter(entry.Value.GetType().GetElementType());
						foreach (object val in (Array)entry.Value) {
							XmlElement el2 = doc.CreateElement("Element");
							el2.InnerText = c.ConvertToInvariantString(val);
							el.AppendChild(el2);
						}
						propertiesnode.AppendChild(el);
					} else {
						XmlElement el = doc.CreateElement("Property");
						
						XmlAttribute key   = doc.CreateAttribute("key");
						key.InnerText      = entry.Key.ToString();
						el.Attributes.Append(key);
						
						XmlAttribute val   = doc.CreateAttribute("value");
						object value = entry.Value;
						if (value is string) {
							val.InnerText = (string)value;
						} else {
							TypeConverter c = TypeDescriptor.GetConverter(value.GetType());
							val.InnerText = c.ConvertToInvariantString(value);
						}
						el.Attributes.Append(val);
						
						propertiesnode.AppendChild(el);
					}
				}
			}
			return propertiesnode;
		}
		#endregion
		
		/// <summary>
		/// Returns a new instance of <see cref="IProperties"/> which has
		/// the same properties.
		/// </summary>
		public IProperties Clone()
		{
			DefaultProperties df = new DefaultProperties();
			df.properties = (Hashtable)properties.Clone();
			return df;
		}
		
		protected virtual void OnPropertyChanged(PropertyEventArgs e)
		{
			if (PropertyChanged != null) {
				PropertyChanged(this, e);
			}
		}
		
		public event PropertyEventHandler PropertyChanged;
		
		#region Link to object's properties
		Hashtable links = new Hashtable(); // string -> ArrayList(Link)
		
		public PropertyLink CreateLink(string key, object obj, string property) {
			ArrayList l = (ArrayList)links[key];
			if (l == null) {
				l = new ArrayList();
				links[key] = l;
			}
			PropertyLink pl = new PropertyLink(obj, property, this, key);
			l.Add(pl);
			return pl;
		}
		
		public void RemoveLink(PropertyLink link) {
			if (link == null)
				throw new ArgumentNullException("link");
			if (link.Disposed)
				throw new ObjectDisposedException(link.GetType().FullName);
			if (link.Properties != this)
				throw new ArgumentException("The link was not created for this DefaultProperties");
			ArrayList l = (ArrayList)links[link.PropertyKey];
			if (l == null || !l.Contains(link))
				throw new Exception("Internal error: link not found");
			link.InternalRemove();
			l.Remove(link);
		}
		#endregion
	}
}
