// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.Collections;
using System.IO;
using System.Diagnostics;
using System.Xml;
using System.Windows.Forms;
using System.Reflection;

namespace ICSharpCode.Core.Properties
{
	/// <summary>
	/// The <see cref="IProperties"/> interface defines a set of properties
	/// </summary>
	public interface IProperties : IXmlConvertable
	{
		/// <summary>
		/// Gets a property out of the collection. The defaultvalue must either
		/// have a cast to a string (and back) or implement the
		/// <see cref="IXmlConvertable"/> interface.
		/// </summary>
		/// <returns>
		/// The property, or <paramref name="defaultvalue"/>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		object GetProperty(string key, object defaultvalue);
		
		/// <summary>
		/// Gets a property of the specified type out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <c>null</c>, if the property wasn't found.
		/// </returns>
		object GetProperty(string key, Type type);
		
		/// <summary>
		/// Gets a property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <c>null</c>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		object GetProperty(string key);
		
		/// <summary>
		/// Gets an <c>int</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <paramref name="defaultvalue"/>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		int GetProperty(string key, int defaultvalue);
		
		/// <summary>
		/// Gets a <c>bool</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <paramref name="defaultvalue"/>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		bool GetProperty(string key, bool defaultvalue);
		
		/// <summary>
		/// Gets a <c>short</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <paramref name="defaultvalue"/>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		short GetProperty(string key, short defaultvalue);
		
		/// <summary>
		/// Gets a <c>byte</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <paramref name="defaultvalue"/>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		byte GetProperty(string key, byte defaultvalue);
		
		/// <summary>
		/// Gets a <c>string</c> property out of the collection.
		/// </summary>
		/// <returns>
		/// The property, or <paramref name="defaultvalue"/>, if the property wasn't
		/// found.
		/// </returns>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="defaultvalue">
		/// The default value of the property.
		/// </param>
		string GetProperty(string key, string defaultvalue);
		
		/// <summary>
		/// Sets the property <paramref name="key"/> to the value <paramref name="val"/>.
		/// If <paramref name="val"/> is null, the property will be taken out from the
		/// properties.
		/// </summary>
		/// <param name="key">
		/// The name of the property.
		/// </param>
		/// <param name="val">
		/// The value of the property.
		/// </param>
		void SetProperty(string key, object val);
		
		/// <summary>
		/// Returns a new instance of <see cref="IProperties"/> which has
		/// the same properties.
		/// </summary>
		IProperties Clone();
		
		/// <summary>
		/// The property changed event handler, it is called
		/// when a property has changed.
		/// </summary>
		event PropertyEventHandler PropertyChanged;
		
		/// <summary>
		/// <para>
		///   Creates a link between the property <paramref name="key"/> and the property
		///   <paramref name="property"/> on the object <paramref name="obj"/>.
		/// </para>
		/// <para>
		///   This means changes to the property system will also set the
		///   object's property.
		///   For the other way, you have to set the <c>UseEvent</c> property on the returned
		///   <see cref="PropertyLink"/> object.
		/// </para>
		/// <para>
		///   When the link is created, the object's value is set to the stored value
		///   (the old object's value is used as default value for retrieving the property)
		/// </para>
		/// </summary>
		PropertyLink CreateLink(string key, object obj, string property);
		
		/// <summary>
		/// Removes the specified link.
		/// This method is used by <c>PropertyLink.Dispose</c>.
		/// </summary>
		void RemoveLink(PropertyLink link);
	}
}
