// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.Reflection;
using ICSharpCode.Core.Services;

namespace ICSharpCode.Core.Properties
{
	public class PropertyLink : IDisposable
	{
		IProperties properties;
		string propertyKey;
		
		bool isSetting = false;
		object obj;
		PropertyInfo p;
		EventInfo evt;
		string name;
		
		internal PropertyLink(object obj, string name, IProperties properties, string key) {
			this.properties = properties;
			this.propertyKey = key;
			this.obj = obj;
			this.name = name;
			Type t = obj.GetType();
			p = t.GetProperty(name);
			if (p == null)
				throw new MissingMemberException(t.FullName, name);
			object oldValue = p.GetValue(obj, null);
			SetValue(properties.GetProperty(key, oldValue));
		}
		
		private bool CreateMonitor() {
			if (evt != null)
				throw new ArgumentException("Link has already subscribed to the event.");
			evt = obj.GetType().GetEvent(name + "Changed");
			if (evt == null)
				return false;
			evt.AddEventHandler(obj, new EventHandler(ValueChanged));
			return true;
		}
		
		private void ReleaseMonitor() {
			if (evt != null) {
				evt.RemoveEventHandler(obj, new EventHandler(ValueChanged));
				evt = null;
			}
		}
		
		/// <summary>
		/// Gets/Sets if the property link should subscribe to the Changed event on the target object
		/// and save changes back to the property system.
		/// </summary>
		public bool UseEvent {
			get {
				return evt != null;
			}
			set {
				if (disposed)
					throw new ObjectDisposedException(GetType().FullName);
				if ((evt != null) == value)
					return;
				if (value) {
					if (!CreateMonitor())
						throw new MissingMemberException(obj.GetType().FullName, name + "Changed");
				} else {
					ReleaseMonitor();
				}
			}
		}
		
		bool disposed = false;
		
		/// <summary>
		/// Closes this link.
		/// </summary>
		public void Dispose() {
			if (disposed)
				return;
			properties.RemoveLink(this);
		}
		
		internal void InternalRemove() {
			disposed = true;
			ReleaseMonitor();
			obj = null;
		}
		
		/// <summary>
		/// Gets if the link has been closed.
		/// </summary>
		public bool Disposed {
			get {
				return disposed;
			}
		}
		
		/// <summary>
		/// Gets the property key changes are saved at.
		/// </summary>
		public string PropertyKey {
			get {
				return propertyKey;
			}
		}
		
		/// <summary>
		/// Gets the <see cref="IProperties"/> object the changes are saved to.
		/// </summary>
		public IProperties Properties {
			get {
				return properties;
			}
		}
		
		internal void SetValue(object value) {
			if (isSetting) return;
			isSetting = true;
			try {
				p.SetValue(obj, value, null);
			} finally {
				isSetting = false;
			}
		}
		
		private void ValueChanged(object sender, EventArgs e) {
			if (isSetting) return;
			isSetting = true;
			try {
				properties.SetProperty(propertyKey, p.GetValue(obj, null));
			} catch (Exception ex) {
				IMessageService msg = (IMessageService)ServiceManager.Services.GetService(typeof(IMessageService));
				msg.ShowError(ex);
			}
			isSetting = false;
		}
	}
}
