// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.IO;
using System.Diagnostics;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Reflection;
using ICSharpCode.Core.Properties;

namespace ICSharpCode.Core.Services
{
	/// <summary>
	/// This class handles the Global Properties for the IDE, all what can be configured should be
	/// loaded/saved by this class. It is a bit like a Singleton with static delegation instead
	/// of returning a static reference to a <see cref="IProperties"/> object.
	/// </summary>
	public class PropertyService : DefaultProperties, IService
	{
		#region Static Properties
		static bool instanceCreated = false;
		static string propertyFileName = "SharpDevelopProperties.xml";
		static string propertyXmlRootNodeName = "SharpDevelopProperties";
		readonly static string propertyFileVersion = "1.1";
		
		static string configDirectory = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) +
			Path.DirectorySeparatorChar + ".ICSharpCode" +
			Path.DirectorySeparatorChar + "SharpDevelop" +
			Path.DirectorySeparatorChar;
		
		static string dataDirectory = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location) +
			Path.DirectorySeparatorChar + ".." +
			Path.DirectorySeparatorChar + "data";
		
		/// <summary>
		/// <para>
		/// Sets the information the PropertyService should use to save and load the xml file.
		/// Specify <c>null</c> for a parameter to keep the default value.
		/// </para>
		/// <para>
		/// This method can only be called before the instance of the PropertyService is created
		/// (before the ServiceManager is accessed the first time).
		/// </para>
		/// </summary>
		/// <param name="path">The full path of the configDirectory (default: %ApplicationData%\.ICSharpCode\SharpDevelop)</param>
		/// <param name="filename">The filename of the xml file (default: SharpDevelopProperties.xml)</param>
		/// <param name="xmlRootName">The name of the xml root elment (default: SharpDevelopProperties)</param>
		public static void SetInformation(string path, string filename, string xmlRootName) {
			if (instanceCreated)
				throw new CoreException("This method can only be called before the instance of the PropertyService is created" +
				                        " (before the ServiceManager is accessed the first time).");
			
			if (path != null) {
				if (!path.EndsWith(Path.DirectorySeparatorChar.ToString()) && !path.EndsWith(Path.AltDirectorySeparatorChar.ToString()))
					path += Path.DirectorySeparatorChar;
				configDirectory = path;
			}
			if (filename != null)
				propertyFileName = filename;
			if (xmlRootName != null)
				propertyXmlRootNodeName = xmlRootName;
		}
		
		/// <summary>
		/// Sets the data directory of the PropertyService.
		/// </summary>
		public static void SetDataDir(string path) {
			dataDirectory = path;
		}
		#endregion
		
		public PropertyService()
		{
			if (!instanceCreated) { // first instance
				instanceCreated = true;
				string confDataDirectory = System.Configuration.ConfigurationSettings.AppSettings["DataDirectory"];
				
				if (confDataDirectory != null)
					dataDirectory = confDataDirectory;
			}
			try {
				LoadProperties();
			} catch (PropertyFileLoadException) {
				// don't use message service --> cyclic dependency
				System.Windows.Forms.MessageBox.Show("Can't load property file", "Warning");
			}
		}
		
		/// <summary>
		/// returns the path of the default application configuration directory
		/// </summary>
		public string ConfigDirectory {
			get {
				return configDirectory;
			}
		}
		
		public string DataDirectory {
			get {
				return dataDirectory;
			}
		}
		
		void WritePropertiesToFile(string fileName)
		{
			XmlDocument doc = new XmlDocument();
			doc.LoadXml("<?xml version=\"1.0\"?>\n<" + propertyXmlRootNodeName + " fileversion = \"" + propertyFileVersion + "\" />");
			
			doc.DocumentElement.AppendChild(ToXmlElement(doc));
			
			FileUtilityService fileUtilityService = (FileUtilityService)ServiceManager.Services.GetService(typeof(FileUtilityService));
			fileUtilityService.ObservedSave(new NamedFileOperationDelegate(doc.Save), fileName, FileErrorPolicy.ProvideAlternative);
		}
		
		bool LoadPropertiesFromStream(string filename)
		{
			try {
				if (!File.Exists(filename))
					return true;
				XmlDocument doc = new XmlDocument();
				doc.Load(filename);
				
				if (doc.DocumentElement.Attributes["fileversion"].InnerText != propertyFileVersion) {
					return false;
				}
				SetValueFromXmlElement(doc.DocumentElement["Properties"]);
			} catch (Exception e) {
				Console.WriteLine("Exception while load properties from stream: \n" + e.ToString());
				System.Windows.Forms.MessageBox.Show("Exception while load properties from stream: \n" + e.ToString());
				return false;
			}
			return true;
		}
		
		/// <summary>
		/// Loads the global properties from the current users application data folder, or
		/// if it doesn't exists or couldn't read them it reads the default properties out
		/// of the application folder.
		/// </summary>
		/// <exception cref="PropertyFileLoadException">
		/// Is thrown when no property file could be loaded.
		/// </exception>
		void LoadProperties()
		{
			if (!Directory.Exists(configDirectory)) {
				Directory.CreateDirectory(configDirectory);
			}
			
			if (!LoadPropertiesFromStream(configDirectory + propertyFileName)) {
				if (!LoadPropertiesFromStream(DataDirectory + Path.DirectorySeparatorChar + "options" + Path.DirectorySeparatorChar + propertyFileName)) {
					throw new PropertyFileLoadException();
				}
			}
		}
		
		/// <summary>
		/// Saves the current global property state to a file in the users application data folder.
		/// </summary>
		public void SaveProperties()
		{
			WritePropertiesToFile(configDirectory + propertyFileName);
		}
		
		// IService implementation:
		public virtual void InitializeService()
		{
			OnInitialize(EventArgs.Empty);
		}
		
		public virtual void UnloadService()
		{
			// save properties on exit
			SaveProperties();
			OnUnload(EventArgs.Empty);
		}
		
		protected virtual void OnInitialize(EventArgs e)
		{
			if (Initialize != null) {
				Initialize(this, e);
			}
		}
		
		protected virtual void OnUnload(EventArgs e)
		{
			if (Unload != null) {
				Unload(this, e);
			}
		}
		
		public event EventHandler Initialize;
		public event EventHandler Unload;
	}
}
