// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using DGNetwork.Chat;
using DGNetwork.Chat.Parser;
using NUnit.Framework;

namespace DGNetwork.Tests.Chat
{
	[TestFixture]
	public class HelperTest
	{
		[Test]
		public void DecodeEntity()
		{
			Assert.AreEqual("&nonexistantEntity;", HtmlHelper.DecodeEntity("nonexistantEntity"));
			Assert.AreEqual("&", HtmlHelper.DecodeEntity("amp"));
			Assert.AreEqual(" ", HtmlHelper.DecodeEntity("nbsp"));
			Assert.AreEqual("<", HtmlHelper.DecodeEntity("lt"));
			Assert.AreEqual(">", HtmlHelper.DecodeEntity("gt"));
			Assert.AreEqual("@", HtmlHelper.DecodeEntity("#64"));
			Assert.AreEqual("A", HtmlHelper.DecodeEntity("#65"));
		}
		
		[Test]
		public void DecodeEntities()
		{
			Assert.AreEqual("&nonexistantEntity;", HtmlHelper.DecodeEntities("&nonexistantEntity;"));
			Assert.AreEqual("ABC", HtmlHelper.DecodeEntities("A&#66;C"));
			Assert.AreEqual("<b> & </b>", HtmlHelper.DecodeEntities("<b&gt; &amp; &lt;/b>"));
			Assert.AreEqual("<c> & </c>", HtmlHelper.DecodeEntities("<c&gt; & &lt;/c>"));
			Assert.AreEqual("a & <b> < c &lt d", HtmlHelper.DecodeEntities("a &amp; <b> &lt; c &lt d"));
			Assert.AreEqual("", HtmlHelper.DecodeEntities(""));
			Assert.AreEqual("    \r\n\t ", HtmlHelper.DecodeEntities("  &nbsp; \r\n\t "));
			Assert.IsNull(HtmlHelper.DecodeEntities(null));
		}
		
		[Test]
		public void EncodeText()
		{
			Assert.AreEqual("a &amp; b", HtmlHelper.EncodeText("a & b"));
			Assert.AreEqual("a &lt; b", HtmlHelper.EncodeText("a < b"));
			Assert.AreEqual("a&nbsp;&nbsp;b", HtmlHelper.EncodeText("a  b"));
			Assert.AreEqual("&nbsp;&nbsp;", HtmlHelper.EncodeText("  "));
			Assert.AreEqual("&nbsp;&nbsp;a", HtmlHelper.EncodeText("  a"));
			Assert.AreEqual("a&nbsp;&nbsp;", HtmlHelper.EncodeText("a  "));
			Assert.AreEqual("", HtmlHelper.EncodeText(""));
			Assert.AreEqual("a\r\n<br>&nbsp;b", HtmlHelper.EncodeText("a\r\n b"));
			Assert.IsNull(HtmlHelper.EncodeText(null));
		}
		
		[Test]
		public void EncodeEntities()
		{
			Assert.AreEqual("a &amp; b", HtmlHelper.EncodeEntities("a & b"));
			Assert.AreEqual("a &lt; b", HtmlHelper.EncodeEntities("a < b"));
			Assert.AreEqual("a  b", HtmlHelper.EncodeEntities("a  b"));
			Assert.AreEqual("  ", HtmlHelper.EncodeEntities("  "));
			Assert.AreEqual("  a", HtmlHelper.EncodeEntities("  a"));
			Assert.AreEqual("a  ", HtmlHelper.EncodeEntities("a  "));
			Assert.AreEqual("", HtmlHelper.EncodeEntities(""));
			Assert.AreEqual("a\r\n b", HtmlHelper.EncodeEntities("a\r\n b"));
			Assert.IsNull(HtmlHelper.EncodeEntities(null));
		}
		
		[Test]
		public void ParseEntity()
		{
			DoParseEntity("a &amp; b", 2, 6, "&");
			DoParseEntity("a &amp b", 2, 5, "&amp");
			DoParseEntity("a &", 2, 2, "&");
			DoParseEntity("a &amp&nbsp; b", 2, 5, "&amp");
			DoParseEntity("a &amp&nbsp; b", 6, 11, " ");
		}
		
		private void DoParseEntity(string text, int pos, int endPos, string entity)
		{
			Assert.AreEqual(entity, HtmlHelper.ParseEntity(text, ref pos), "entity in " + text);
			Assert.AreEqual(endPos, pos, "pos after " + text);
		}
		
		[Test]
		public void ParseQuotedValue()
		{
			AutoParseQuotedValue("<a description='eine ganz schöne&nbsp;&nbsp;&amp; tolle URL'>",
			                     "eine ganz schöne  & tolle URL", '\'');
			AutoParseQuotedValue("<a description='eine ganz schöne  & tolle URL'>",
			                     "eine ganz schöne & tolle URL", '\'');
			AutoParseQuotedValue("<a description=\"eine ganz schöne&nbsp;&nbsp;&amp; tolle URL\">",
			                     "eine ganz schöne  & tolle URL", '"');
			AutoParseQuotedValue("<a description=\"eine ganz schöne  & tolle URL\">",
			                     "eine ganz schöne & tolle URL", '"');
			DoParseQuotedValue("a 'text", 2, 7, null);
			DoParseQuotedValue("a 'text' b", 2, 7, "text");
			DoParseQuotedValue("a 'c > d' b", 2, 5, null);
			AutoParseQuotedValue("a 'c &gt; d' b", "c > d", '\'');
		}
		
		private void AutoParseQuotedValue(string text, string value, char quote)
		{
			DoParseQuotedValue(text, text.IndexOf(quote), text.LastIndexOf(quote), value);
		}
		
		private void DoParseQuotedValue(string text, int pos, int endPos, string value)
		{
			Assert.AreEqual(value, HtmlHelper.ParseQuotedValue(text, ref pos), "value in " + text);
			Assert.AreEqual(endPos, pos, "pos after " + text);
		}
		
		[Test]
		public void ParseAttributeValue()
		{
			DoParseAttributeValue("[a = b]", 4, 6, "b");
			DoParseAttributeValue("[a = b]", 5, 6, "b");
			DoParseAttributeValue("[a = b ]", 4, 6, "b");
			DoParseAttributeValue("[a = b c = d]", 4, 6, "b");
			DoParseAttributeValue("[a = b c = d]", 10, 12, "d");
			DoParseAttributeValue("[a ='b c =' d]", 4, 11, "b c =");
			DoParseAttributeValue("[a = 'b &amp; c' d]", 4, 16, "b & c");
		}
		
		private void DoParseAttributeValue(string text, int pos, int endPos, string value)
		{
			Assert.AreEqual(value, HtmlHelper.ParseAttributeValue(text, ref pos, '[', ']'), "value in " + text);
			Assert.AreEqual(endPos, pos, "pos after " + text);
		}
		
		[Test]
		public void GetTagFromHtml()
		{
			DoTests(false);
		}
		
		[Test]
		public void GetTagFromChatcode()
		{
			DoTests(true);
		}
		
		private void DoTests(bool ext)
		{
			Assert.IsNull(HtmlHelper.ParseTag("a<b", 1, ext), "a<b");
			Assert.IsNull(HtmlHelper.ParseTag("a< b>c", 1, ext), "a< b>c");
			CheckForSimpleTag("a<b>c", 1, ext, "b", 3);
			CheckForSimpleTag("a<b >c", 1, ext, "b", 4);
		}
		
		private void CheckForSimpleTag(string html, int pos, bool ext, string tagName, int end)
		{
			TagInfo info = HtmlHelper.ParseTag(html, pos, ext);
			Assert.IsNotNull(info, html);
			Assert.AreEqual(tagName, info.Name, html);
			Assert.AreEqual(false, info.IsEndTag, "EndTag: " + tagName);
			Assert.AreEqual(0, info.Attributes.Count, "Attribute count: " + html);
			Assert.IsNull(info.DirectAttribute, "Direct attribute: " + html);
			Assert.AreEqual('<', info.StartChar, html);
			Assert.AreEqual('>', info.EndChar, html);
			Assert.AreEqual(pos, info.StartPos, "StartPos: " + html);
			Assert.AreEqual(end, info.EndPos, "EndPos: " + end);
		}
	}
}
