// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;

namespace Grunwald.Gui.Components
{
	/// <summary>
	/// Selects a cut out of an image.
	/// </summary>
	public class BitmapSelect : System.Windows.Forms.Form
	{
		private System.Windows.Forms.Button buttonAbort;
		private System.Windows.Forms.PictureBox pic;
		private System.Windows.Forms.Button buttonOK;
		private System.Windows.Forms.Panel bottomPanel;
		
		Bitmap original;
		
		#region SelectedRange
		Rectangle sel;
		
		/// <summary>
		/// The selected rectangle.
		/// </summary>
		public Rectangle SelectedRectangle {
			get {
				if (halfSize)
					return new Rectangle(sel.X * 2, sel.Y * 2, sel.Width * 2, sel.Height * 2);
				else
					return sel;
			}
		}
		
		/// <summary>
		/// Gets the bitmap in the selected bitmap.
		/// Can return null if nothing is selected or the original bitmap if everything was selected.
		/// </summary>
		public Bitmap GetSelectedBitmap()
		{
			Rectangle sel = SelectedRectangle;
			if (sel.Width == 0 || sel.Height == 0) return null;
			if (sel.X == 0 && sel.Y == 0 && sel.Width >= original.Width && sel.Height >= original.Height) {
				return original;
			}
			Bitmap bmp = new Bitmap(sel.Width, sel.Height);
			for (int x = 0; x < sel.Width; x++) {
				for (int y = 0; y < sel.Height; y++) {
					bmp.SetPixel(x, y, original.GetPixel(sel.X + x, sel.Y + y));
				}
			}
			return bmp;
		}
		#endregion
		
		#region Halfsize
		bool halfSize = false;
		
		/// <summary>
		/// Specifies if the image is viewn in half-size.
		/// </summary>
		public bool HalfSize {
			get {
				return halfSize;
			}
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Erzeugt einen neuen Ausschnitts-Wähler für das angegebene Bild, der die
		/// Halbierung der Größe erlaubt.
		/// </summary>
		public BitmapSelect(Bitmap bmp) : this(bmp, true)
		{
		}
		
		/// <summary>
		/// Erzeugt einen neuen Ausschnitts-Wähler für das angegebene Bild.
		/// </summary>
		public BitmapSelect(Bitmap bmp, bool allowHalfSize)
		{
			//
			// Erforderlich für die Windows Form-Designerunterstützung
			//
			InitializeComponent();
			
			buttonAbort.Text = Translator.DefaultTranslator.Cancel;
			buttonOK.Text = Translator.DefaultTranslator.OK;
			
			original = bmp;
			
			if (allowHalfSize) {
				Rectangle workingArea = Screen.GetWorkingArea(this);
				if (workingArea.Width - 50 < bmp.Width ||
				    workingArea.Height - 70 - bottomPanel.Height < bmp.Height) {
					halfSize = true;
					bmp = new Bitmap(bmp, bmp.Width / 2, bmp.Height / 2);
				}
			}
			
			ClientSize = new Size(bmp.Width, bmp.Height + bottomPanel.Height);
			base.CenterToScreen();
			CreatePens();
			DialogResult = DialogResult.Cancel;
			pic.BackgroundImage = bmp;
			sel = new Rectangle(0, 0, bmp.Width, bmp.Height);
		}
		#endregion
		
		#region Dispose
		/// <summary>
		/// Die verwendeten Ressourcen bereinigen.
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			if (disposing) {
				DisposePens();
			}
			base.Dispose(disposing);
		}
		#endregion
		
		#region Vom Windows Form-Designer generierter Code
		/// <summary>
		/// Erforderliche Methode für die Designerunterstützung.
		/// Der Inhalt der Methode darf nicht mit dem Code-Editor geändert werden.
		/// </summary>
		private void InitializeComponent() {
			this.bottomPanel = new System.Windows.Forms.Panel();
			this.buttonOK = new System.Windows.Forms.Button();
			this.pic = new System.Windows.Forms.PictureBox();
			this.buttonAbort = new System.Windows.Forms.Button();
			this.bottomPanel.SuspendLayout();
			this.SuspendLayout();
			// 
			// bottomPanel
			// 
			this.bottomPanel.Controls.Add(this.buttonAbort);
			this.bottomPanel.Controls.Add(this.buttonOK);
			this.bottomPanel.Dock = System.Windows.Forms.DockStyle.Bottom;
			this.bottomPanel.Location = new System.Drawing.Point(0, 226);
			this.bottomPanel.Name = "bottomPanel";
			this.bottomPanel.Size = new System.Drawing.Size(292, 40);
			this.bottomPanel.TabIndex = 0;
			// 
			// buttonOK
			// 
			this.buttonOK.Anchor = System.Windows.Forms.AnchorStyles.Right;
			this.buttonOK.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.buttonOK.Location = new System.Drawing.Point(208, 8);
			this.buttonOK.Name = "buttonOK";
			this.buttonOK.Size = new System.Drawing.Size(72, 24);
			this.buttonOK.TabIndex = 0;
			this.buttonOK.Text = "OK";
			this.buttonOK.Click += new System.EventHandler(this.buttonOK_Click);
			// 
			// pic
			// 
			this.pic.Dock = System.Windows.Forms.DockStyle.Fill;
			this.pic.Location = new System.Drawing.Point(0, 0);
			this.pic.Name = "pic";
			this.pic.Size = new System.Drawing.Size(292, 226);
			this.pic.TabIndex = 1;
			this.pic.TabStop = false;
			this.pic.Paint += new System.Windows.Forms.PaintEventHandler(this.pic_Paint);
			this.pic.MouseUp += new System.Windows.Forms.MouseEventHandler(this.pic_MouseUp);
			this.pic.MouseMove += new System.Windows.Forms.MouseEventHandler(this.pic_MouseMove);
			this.pic.MouseDown += new System.Windows.Forms.MouseEventHandler(this.pic_MouseDown);
			// 
			// buttonAbort
			// 
			this.buttonAbort.Anchor = System.Windows.Forms.AnchorStyles.Right;
			this.buttonAbort.DialogResult = System.Windows.Forms.DialogResult.Cancel;
			this.buttonAbort.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.buttonAbort.Location = new System.Drawing.Point(120, 8);
			this.buttonAbort.Name = "buttonAbort";
			this.buttonAbort.Size = new System.Drawing.Size(72, 24);
			this.buttonAbort.TabIndex = 1;
			this.buttonAbort.Text = "Abort";
			this.buttonAbort.Click += new System.EventHandler(this.buttonAbort_Click);
			// 
			// BitmapSelect
			// 
			this.AcceptButton = this.buttonOK;
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.CancelButton = this.buttonAbort;
			this.ClientSize = new System.Drawing.Size(292, 266);
			this.Controls.Add(this.pic);
			this.Controls.Add(this.bottomPanel);
			this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedToolWindow;
			this.KeyPreview = true;
			this.Name = "BitmapSelect";
			this.Text = "Select range in bitmap";
			this.KeyDown += new System.Windows.Forms.KeyEventHandler(this.BitmapSelectKeyDown);
			this.bottomPanel.ResumeLayout(false);
			this.ResumeLayout(false);
		}
		#endregion
		
		#region Buttons
		private void buttonAbort_Click(object sender, System.EventArgs e)
		{
			Close();
		}
		
		private void buttonOK_Click(object sender, System.EventArgs e)
		{
			DialogResult = DialogResult.OK;
			Close();
		}
		#endregion
		
		#region static Select
		/// <summary>Selects a cut out of an image.</summary>
		/// <param name="bmp">The image the cut out should be selected from.</param>
		/// <param name="title">The title of the dialog window.</param>
		/// <returns>The cut out, the original bitmap or null.</returns>
		/// <remarks>This method returns the original bitmap if the user doesn't select a
		/// cut-out. So don't call dispose on the original image if it is returned again!</remarks>
		public static Bitmap Select(Bitmap bmp, string title)
		{
			BitmapSelect sel = new BitmapSelect(bmp);
			sel.Text = title;
			DialogResult res = sel.ShowDialog();
			if (res == DialogResult.OK)
				return sel.GetSelectedBitmap();
			else
				return null;
		}
		#endregion
		
		#region Drawing
		Pen red;
		
		private void CreatePens()
		{
			red = new Pen(Color.Red);
			red.DashStyle = DashStyle.Dash;
		}
		
		private void DisposePens()
		{
			red.Dispose();
		}
		
		Point start;
		bool mouseDown = false;
		
		private void pic_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left) {
				mouseDown = true;
				start = new Point(e.X, e.Y);
				sel = new Rectangle(start, new Size(0, 0));
			}
		}
		
		private void pic_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (!mouseDown) return;
			int x = e.X;
			int y = e.Y;
			if (x < 0) x = 0;
			if (y < 0) y = 0;
			if (x >= pic.Width) x = pic.Width - 1;
			if (y >= pic.Height) y = pic.Height - 1;
			Point lo = new Point(Math.Min(start.X, x), Math.Min(start.Y, y));
			Point ru = new Point(Math.Max(start.X, x), Math.Max(start.Y, y));
			sel = new Rectangle(lo.X, lo.Y, ru.X - lo.X, ru.Y - lo.Y);
			pic.Invalidate();
		}
		
		private void pic_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left) {
				mouseDown = false;
				pic.Invalidate();
			}
		}
		
		private void pic_Paint(object sender, System.Windows.Forms.PaintEventArgs e)
		{
			Graphics g = e.Graphics;
			g.DrawRectangle(red, sel);
		}
		#endregion
		
		#region Keyboard events
		private void BitmapSelectKeyDown(object sender, System.Windows.Forms.KeyEventArgs e)
		{
			if (e.KeyData == (Keys.Control | Keys.C)) {
				Clipboard.SetDataObject(this.GetSelectedBitmap(), true);
			}
		}
		#endregion
	}
}
