// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;
using Grunwald.Gui.Controls;

namespace Grunwald.Gui.Components
{
	/// <summary>
	/// This class is has some static methods to correct the design of Windows.Forms.
	/// </summary>
	public sealed class Corrections
	{
		#region Constructor
		private Corrections() { }
		
		static bool isXP = GetSupportsDesigns();
		
		private static bool GetSupportsDesigns() {
			if (Environment.OSVersion.Platform != PlatformID.Win32NT) return false;
			int major = Environment.OSVersion.Version.Major;
			if (major > 5) return true;
			if (major < 5) return false;
			return Environment.OSVersion.Version.Minor >= 1;
		}
		#endregion
		
		#region CorrectBackground
		/// <summary>
		/// Corrects the background of all TabPages in the specified TabControl
		/// </summary>
		public static void CorrectBackground(TabControl tabControl)
		{
			foreach (TabPage page in tabControl.TabPages) {
				CorrectBackground(page);
			}
		}
		
		/// <summary>
		/// Corrects the background of the specified TabPage.
		/// </summary>
		public static void CorrectBackground(TabPage tabPage)
		{
			if (!isXP) return;
			CorrectBackground((Control)tabPage);
		}
		
		static void CorrectBackground(Control control)
		{
			control.Paint += new PaintEventHandler(PaintBackground);
			foreach (Control ctl in control.Controls) {
				if (ctl is Panel || ctl is UserControl)
					CorrectBackground(ctl);
			}
		}
		
		private static void PaintBackground(object sender, PaintEventArgs e)
		{
			IntPtr hdc = e.Graphics.GetHdc();
			RECT rect = new RECT();
			rect.top = 0;
			rect.left = 0;
			rect.right = (sender as Control).Width;
			rect.bottom = (sender as Control).Height;
			Win32.DrawThemeParentBackground((sender as Control).Handle, hdc, ref rect);
			e.Graphics.ReleaseHdc(hdc);
		}
		#endregion
		
		#region Proportional ListView header
		/// <summary>
		/// Makes the headers resize proportional with the size of the list view.
		/// </summary>
		public static void MakeProportional(ListView listView)
		{
			new ProportionalHelper(listView);
		}
		
		/// <summary>
		/// Makes the headers resize proportional with the size of the list view.
		/// </summary>
		public static void MakeProportional(ContainerListView listView)
		{
			new CLVProportionalHelper(listView);
		}
		
		private class ProportionalHelper
		{
			ListView listView;
			int oldSize;
			Hashtable h = new Hashtable(); // ColumnHeader -> ColEntry
			
			public ProportionalHelper(ListView listView)
			{
				this.listView = listView;
				oldSize = listView.Width;
				listView.SizeChanged += new EventHandler(ListViewSizeChanged);
			}
			
			private void ListViewSizeChanged(object sender, EventArgs e)
			{
				int newSize = listView.Width;
				if (newSize < 50 || newSize == oldSize)
					return;
				if (listView.Columns.Count == 0) {
					h.Clear();
					return;
				}
				// Remove deleted ColumnHeaders
				ArrayList delete = null;
				foreach(ColumnHeader c in h.Keys) {
					if (!listView.Columns.Contains(c)) {
						if (delete == null)
							delete = new ArrayList();
						delete.Add(c);
					}
				}
				if (delete != null) {
					foreach (ColumnHeader c in delete)
						h.Remove(c);
				}
				// Add new ColumnHeaders / Update old ColumnHeaders
				foreach(ColumnHeader c in listView.Columns) {
					ColEntry en = (ColEntry)h[c];
					if (en == null) {
						en = new ColEntry(c.Width, oldSize);
						h.Add(c, en);
					} else {
						if (c.Width != en.CurWidth) {
							en.Width = c.Width;
							en.Total = oldSize;
						}
					}
					en.CurWidth = en.Width * newSize / en.Total;
					if (c.Width != en.CurWidth)
						c.Width = en.CurWidth;
				}
				oldSize = newSize;
			}
		}
		
		private class ColEntry
		{
			public int Width;
			public int Total;
			public int CurWidth;
			
			public ColEntry(int width, int total)
			{
				this.Width = width;
				this.Total = total;
				this.CurWidth = width;
			}
		}
		
		private class CLVProportionalHelper
		{
			ContainerListView listView;
			int oldSize;
			Hashtable h = new Hashtable(); // ToggleColumnHeader -> ColEntry
			
			public CLVProportionalHelper(ContainerListView listView)
			{
				this.listView = listView;
				oldSize = listView.Width;
				listView.SizeChanged += new EventHandler(ListViewSizeChanged);
			}
			
			private void ListViewSizeChanged(object sender, EventArgs e)
			{
				int newSize = listView.Width;
				if (newSize < 50 || newSize == oldSize)
					return;
				if (listView.Columns.Count == 0) {
					h.Clear();
					return;
				}
				// Remove deleted ColumnHeaders
				ArrayList delete = null;
				foreach(ToggleColumnHeader c in h.Keys) {
					bool ok = false;
					foreach(ToggleColumnHeader t in listView.Columns) {
						if (c == t) { ok = true; break; }
					}
					if (!ok) {
						if (delete == null)
							delete = new ArrayList();
						delete.Add(c);
					}
				}
				if (delete != null) {
					foreach (ColumnHeader c in delete)
						h.Remove(c);
				}
				// Add new ColumnHeaders / Update old ColumnHeaders
				foreach(ToggleColumnHeader c in listView.Columns) {
					ColEntry en = (ColEntry)h[c];
					if (en == null) {
						en = new ColEntry(c.Width, oldSize);
						h.Add(c, en);
					} else {
						if (c.Width != en.CurWidth) {
							en.Width = c.Width;
							en.Total = oldSize;
						}
					}
					en.CurWidth = en.Width * newSize / en.Total;
					if (c.Width != en.CurWidth)
						c.Width = en.CurWidth;
				}
				oldSize = newSize;
			}
		}
		#endregion
	}
}
