// This class is composed from the following internet resources:


// http://dotnetrix.co.uk/misc.html - IconToBitmap

using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Grunwald.Gui.Components
{
	/// <summary>
	/// Class for loading icons.
	/// </summary>
	public class IconLoader
	{
		#if !MONO
		[DllImport("shell32.dll")]
		private static extern IntPtr SHGetFileInfo(
		                                           string pszPath,
		                                           uint dwFileAttributes,
		                                           ref SHFILEINFO psfi,
		                                           int cbSizeFileInfo,
		                                           int uFlags);
		
		[StructLayout(LayoutKind.Sequential)]
		private struct SHFILEINFO
		{
			public IntPtr hIcon;
			public IntPtr iIcon;
			public uint dwAttributes;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
			public string szDisplayName;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 80)]
			public string szTypeName;
		}
		#endif
		
		/// <summary>
		/// Gets the icon of the specified file.<br/>
		/// <b>Warning:</b> after you use this icon, don't forget to call
		/// <see cref="Icon.Dispose">Dispose</see> on it!
		/// </summary>
		/// <param name="filename">The name of the file to load the icon from.<br/>
		/// This can be any file type (exe files, word documents etc.)</param>
		/// <param name="largeIcon">Specifies if the large icon or the small icon
		/// should be loaded.</param>
		/// <returns></returns>
		public static Icon GetIcon(string filename, bool largeIcon)
		{
			#if MONO
			return SystemIcons.Error.Clone();
			#else
			SHFILEINFO shinfo = new SHFILEINFO();
			SHGetFileInfo(filename, 0, ref shinfo,
			              Marshal.SizeOf(shinfo),
			              largeIcon ? 0x100 : 0x101);
			return Icon.FromHandle(shinfo.hIcon);
			#endif
		}
		
		#if !MONO
		[DllImport("user32.dll", SetLastError=true)]
		static extern int DestroyIcon(IntPtr hIcon);
		
		[DllImport("user32.dll")]
		static extern bool GetIconInfo(IntPtr hIcon, out ICONINFO piconinfo);
		
		internal struct ICONINFO
		{
			public bool fIcon;
			public int xHotspot;
			public int yHotspot;
			public IntPtr hbmMask;
			public IntPtr hbmColor;
		}
		#endif
		
		/// <summary>
		/// Converts the icon into a bitmap without losing the alpha channel.
		/// </summary>
		public static Bitmap IconToBitmap(Icon ico)
		{
			#if MONO
			return ico.ToBitmap();
			#else
			ICONINFO ii = new ICONINFO();
			GetIconInfo(ico.Handle, out ii);
			Bitmap bmp = Bitmap.FromHbitmap(ii.hbmColor);
			DestroyIcon(ii.hbmColor);
			DestroyIcon(ii.hbmMask);
			
			if (Bitmap.GetPixelFormatSize(bmp.PixelFormat) < 32)
				return ico.ToBitmap();
			
			BitmapData bmData;
			Rectangle bmBounds = new Rectangle(0,0,bmp.Width,bmp.Height);
			
			bmData = bmp.LockBits(bmBounds,ImageLockMode.ReadOnly, bmp.PixelFormat);
			
			Bitmap dstBitmap=new Bitmap(bmData.Width, bmData.Height, bmData.Stride, PixelFormat.Format32bppArgb, bmData.Scan0);
			
			bool IsAlphaBitmap = false;
			
			for (int y=0; y <= bmData.Height-1; y++)
			{
				for (int x=0; x <= bmData.Width-1; x++)
				{
					Color PixelColor = Color.FromArgb(Marshal.ReadInt32(bmData.Scan0, (bmData.Stride * y) + (4 * x)));
					if (PixelColor.A > 0 & PixelColor.A < 255)
					{
						IsAlphaBitmap = true;
						break;
					}
				}
				if (IsAlphaBitmap) break;
			}
			
			bmp.UnlockBits(bmData);
			
			if (IsAlphaBitmap == true)
				return new Bitmap(dstBitmap);
			else
				return new Bitmap(ico.ToBitmap());
			#endif
		}
	}
}


