// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.ComponentModel;
using System.IO;
using System.Data;
using System.Text;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace Grunwald.Gui.Components
{
	/// <summary>
	/// The ScrollSynchronizer synchronizes the scroll position of multiple controls.
	/// This is useful if you have multiple text boxes (or something similar) and
	/// you want to have a scrolling bar only on the rightmost textbox and all other
	/// textboxes should scroll with the last one.<br/>
	/// Of course, scrolling with the mouse wheel should also affect all textboxes.
	/// </summary>
	/// <remarks>
	/// At the moment, this class only works with RichTextBoxes.
	/// I think adding more control types won't be too complicated, so if you
	/// implement another control type, send a mail to daniel AT danielgrunwald DOT de
	/// </remarks>
	public class ScrollSynchronizer
	{
		/// <summary>
		/// Adds the specified control to the list of synchronized controls.
		/// </summary>
		public void Add(RichTextBox ctl)
		{
			Add(ctl as Control);
			ctl.VScroll += new EventHandler(ControlVScroll);
			ctl.HScroll += new EventHandler(ControlHScroll);
		}
		
		/// <summary>
		/// Adds the specified control to the list of synchronized controls.
		/// </summary>
		private void Add(Control ctl)
		{
			if (ctl == null)
				throw new ArgumentNullException();
			controls.Add(ctl);
			if (ctl.Handle != IntPtr.Zero) {
				ControlHandleCreated(ctl, EventArgs.Empty);
			}
			ctl.HandleCreated += new EventHandler(ControlHandleCreated);
			ctl.HandleDestroyed += new EventHandler(ControlHandleDestroyed);
			
		}
		
		/// <summary>
		/// Removes the specified control from the list of synchronized controls.
		/// </summary>
		public void Remove(Control ctl)
		{
			if (ctl == null)
				throw new ArgumentNullException();
			if (!controls.Contains(ctl))
				throw new ArgumentException("control not found");
			ControlHandleDestroyed(ctl, EventArgs.Empty);
		}
		
		#region NativeWindow subclassing
		private delegate void WndProcEventHandler(Message uMsg);
		
		private class Subclass : NativeWindow
		{
			public event WndProcEventHandler WndProcCalled;
			
			Control ctl;
			
			public Control Control {
				get { return ctl; }
			}
			
			public Subclass(Control ctl) : base()
			{
				this.ctl = ctl;
				base.AssignHandle(ctl.Handle);
			}
			
			protected override void WndProc(ref Message uMsg)
			{
				base.WndProc(ref uMsg);
				if (WndProcCalled != null)
					WndProcCalled(uMsg);
			}
			
			public void SendWndProc(Message uMsg)
			{
				base.WndProc(ref uMsg);
			}
		}
		#endregion
		
		ArrayList controls = new ArrayList();
		ArrayList classes = new ArrayList();
		
		bool active = true;
		
		private void SubclassWndProcCalled(Message uMsg)
		{
			if (!active) return;
			active = false;
			switch (uMsg.Msg) {
				case Win32.WM_VSCROLL:
				case Win32.WM_HSCROLL:
				case Win32.WM_MOUSEWHEEL:
					foreach (Subclass r in classes) {
						if (r.Handle != uMsg.HWnd) {
							Win32.SendMessage(r.Handle, uMsg.Msg, uMsg.WParam, uMsg.LParam);
						}
					}
					break;
			}
			active = true;
		}
		
		private void ControlVScroll(object sender, EventArgs e)
		{
			Control ctl = (Control)sender;
			int pos = Win32.GetScrollPos(ctl.Handle, Win32.SBS_VERT);
			foreach (Subclass c in classes) {
				if (c.Control == ctl) {
					Win32.PostMessageA(c.Handle, Win32.WM_VSCROLL,
					                   Win32.SB_THUMBPOSITION + 0x10000 * pos, 0);
				}
			}
		}
		
		private void ControlHScroll(object sender, EventArgs e)
		{
			Control ctl = (Control)sender;
			int pos = Win32.GetScrollPos(ctl.Handle, Win32.SBS_HORZ);
			foreach (Subclass c in classes) {
				if (c.Control == ctl) {
					Win32.PostMessageA(c.Handle, Win32.WM_HSCROLL,
					                   Win32.SB_THUMBPOSITION + 0x10000 * pos, 0);
				}
			}
		}
		
		private void ControlHandleCreated(object sender, EventArgs e)
		{
			Control ctl = (Control)sender;
			Subclass c = new Subclass(ctl);
			c.WndProcCalled += new WndProcEventHandler(SubclassWndProcCalled);
			classes.Add(c);
		}
		
		private void ControlHandleDestroyed(object sender, EventArgs e)
		{
			Control ctl = (Control)sender;
			foreach (Subclass c in classes) {
				if (c.Control == ctl) {
					c.WndProcCalled -= new WndProcEventHandler(SubclassWndProcCalled);
					classes.Remove(c);
					break;
				}
			}
		}
	}
}
