// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace Grunwald.Gui.Controls
{
	#region Row/Column
	/// <summary>
	/// A DockedControlRow is a panel that contains several child controls docked from left to right
	/// that have a size relative to the available size.
	/// Unlike a normal docked panel, not the filling panel gains more size, but all panels
	/// together grow.
	/// </summary>
	public class DockedControlRow : DockedControlList
	{
		/// <summary>
		/// Creates a new DockedControlRow.
		/// </summary>
		public DockedControlRow() : base(false) { }
	}
	
	/// <summary>
	/// A DockedControlColumn is a panel that contains several child controls docked from top to bottom
	/// that have a size relative to the available size.
	/// Unlike a normal docked panel, not the filling panel gains more size, but all panels
	/// together grow.
	/// </summary>
	public class DockedControlColumn : DockedControlList
	{
		/// <summary>
		/// Creates a new DockedControlColumn.
		/// </summary>
		public DockedControlColumn() : base(true) { }
	}
	#endregion
	
	#region DockedControlSetting
	/// <summary>
	/// Contains the settings of a control.
	/// </summary>
	public class DockedControlSetting
	{
		Control control;
		
		/// <summary>
		/// Creates a new setting container for the control.
		/// </summary>
		public DockedControlSetting(Control control)
		{
			this.control = control;
		}
		
		/// <summary>
		/// Gets the control this setting container was created for.
		/// </summary>
		public Control Control {
			get { return control; }
		}
		
		int minSize = 10;
		
		/// <summary>
		/// Gets/Sets the minimum size (in the direction the <see cref="DockedControlList"/> aligns)
		/// the control should have.<br/>
		/// This is used for both automatic sizing and limiting the splitters.
		/// </summary>
		public int MinSize {
			get { return minSize; }
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException("minSize", value, "Value can't be less than zero.");
				minSize = value;
			}
		}
		
		float factor = 1;
		
		/// <summary>
		/// Gets/Sets the factor of the size of this control.
		/// For example if a <see cref="DockedControlRow"/> has 100px width and 2 child controls,
		/// both child controls would get 50px (=50%) if they both have the same factor.
		/// If control 1 has the factor 3 and control 2 has the factor 1 (default),
		/// control 1 will get 75% and control 2 will get 25%.
		/// </summary>
		public float Factor {
			get { return factor; }
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException("factor", value, "Value can't be less than zero.");
				factor = value;
			}
		}
	}
	#endregion
	
	/// <summary>
	/// A DockedControlList is a panel that contains several child controls docked from left to right
	/// or top to bottom (specified by the vertical-parameter in the constructor)
	/// that have a size relative to the available size.
	/// Unlike a normal docked panel, not the filling panel gains more size, but all panels
	/// together grow.
	/// </summary>
	public class DockedControlList : Panel
	{
		#region Load/Save Factors
		/// <summary>
		/// Gets/Sets a byte array containing the list of factors.
		/// This can be used to save and restore the splitter positions.
		/// </summary>
		[Browsable(false)]
		public byte[] Factors {
			get {
				byte[] r = new byte[settings.Count * 4];
				int pos = 0;
				for (int i = 0; i < Controls.Count; ++i) {
					Control ctl = Controls[i];
					if (ctl is Splitter) continue;
					DockedControlSetting set = this[ctl];
					BitConverter.GetBytes(set.Factor).CopyTo(r, pos);
					pos += 8;
				}
				return r;
			}
			set {
				if (value == null)
					throw new ArgumentNullException("factors");
				byte[] r = value;
				if (r.Length != settings.Count * 4)
					throw new ArgumentException("The array must have the length " + (settings.Count * 4).ToString());
				int pos = 0;
				for (int i = 0; i < Controls.Count; ++i) {
					Control ctl = Controls[i];
					if (ctl is Splitter) continue;
					DockedControlSetting set = this[ctl];
					set.Factor = BitConverter.ToSingle(r, pos);
					pos += 8;
				}
				Arrange();
			}
		}
		#endregion
		
		#region Property AddSplitter
		bool addSplitter = false;
		
		/// <summary>
		/// Gets/Sets if the DockedControlList should automatically add splitters between all controls.
		/// Note: if this feature added splitters once, setting this parameter to false won't
		/// remove the existent splitters.<br/>
		/// Setting this property to true will only add splitters on the next resize or call to
		/// <see cref="Refresh"/>.
		/// </summary>
		[Bindable(true), Category("Behaviour"), DefaultValue(false),
		 Description("Specifies if the DockedControlList should automatically add splitters between all controls.")]
		public bool AddSplitter {
			get { return addSplitter; }
			set { addSplitter = value; }
		}
		#endregion
		
		bool vertical;
		
		Hashtable settings = new Hashtable();
		
		#region Setting Indexer
		/// <summary>
		/// Gets the settings of the specified control.
		/// If the control is not a child control of this <see cref="DockedControlList"/>, this method
		/// returns null. If the control is a splitter, it will also return null since splitters
		/// have no settings.
		/// </summary>
		public DockedControlSetting this[Control control] {
			get {
				return (DockedControlSetting)settings[control];
			}
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates a new DockedControlList.
		/// </summary>
		/// <param name="vertical">Specifies if the control list aligns it's children
		/// vertically or horizontally.</param>
		public DockedControlList(bool vertical)
		{
			this.vertical = vertical;
		}
		#endregion
		
		#region Arrange
		bool inArrange = false;
		
		private void Arrange()
		{
			if (DesignMode) return;
			Check();
			if (inArrange) return;
			inArrange = true;
			DoArrange();
			inArrange = false;
		}
		
		private void DoArrange()
		{
			if (Controls.Count == 0) return;
			int clientSize;
			if (vertical)
				clientSize = ClientSize.Height;
			else
				clientSize = ClientSize.Width;
			
			int minSize = 0;
			float combinedFactor = 0;
			foreach (DockedControlSetting set in settings.Values) {
				if (set.Control.Visible) {
					minSize += set.MinSize;
					combinedFactor += set.Factor;
				}
			}
			
			int freeSpace = clientSize;
			
			DockStyle dock = vertical ? DockStyle.Bottom : DockStyle.Right;
			
			bool wasSplitter = true;
			bool first = true;
			
			for (int i = 0; i < Controls.Count; ++i) {
				Control ctl = Controls[i];
				if (ctl is Splitter) {
					ctl.Visible = !wasSplitter;
					Splitter spl = (Splitter)ctl;
					spl.MinExtra = this[Controls[i - 1]].MinSize;
					spl.MinSize = this[Controls[i + 1]].MinSize;
					wasSplitter = true;
					continue;
				}
				if (!ctl.Visible) continue;
				
				if (first)
					ctl.Dock = DockStyle.Fill;
				else if (ctl.Dock != dock)
					ctl.Dock = dock;
				
				DockedControlSetting set = this[ctl];
				
				if (addSplitter && !wasSplitter) {
					// we need a splitter!!!
					Splitter spl = new Splitter();
					Controls.Add(spl);
					Controls.SetChildIndex(spl, i);
					DoArrange();
					return;
				}
				wasSplitter = false;
				
				int size;
				if (combinedFactor > 0) {
					size = (int)((float)freeSpace * set.Factor / combinedFactor);
					int missing = set.MinSize - size;
					if (missing > 0) {
						// the control is too small...
						freeSpace -= missing;
						size += missing;
					}
				} else {
					size = set.MinSize;
				}
				if (!first) {
					if (vertical)
						ctl.Height = size;
					else
						ctl.Width = size;
				}
				first = false;
			}
		}
		#endregion
		
		/// <summary>
		/// Refreshes the arrangement and forces all child controls to repaint.
		/// </summary>
		public override void Refresh()
		{
			Arrange();
			base.Refresh();
		}
		
		#region Event Handler
		private void ControlVisibleChanged(object sender, EventArgs e)
		{
			Control ctl = sender as Control;
			if (ctl is Splitter) return;
			if (ctl.Visible) Arrange();
		}
		
		private void SplitterMoved(object sender, SplitterEventArgs e)
		{
			// save new factor
			float combinedFactor = 0;
			int combinedSize = 0;
			foreach (DockedControlSetting set in settings.Values) {
				if (!set.Control.Visible) continue;
				combinedFactor += set.Factor;
				combinedSize += vertical ? set.Control.Height : set.Control.Width;
			}
			foreach (DockedControlSetting set in settings.Values) {
				if (!set.Control.Visible) continue;
				int size = vertical ? set.Control.Height : set.Control.Width;
				set.Factor = (float)size * combinedFactor / (float)combinedSize;
			}
		}
		#endregion
		
		#region Overridden Events
		/// <summary>
		/// Raises the <see cref="Control.ControlRemoved"/> event and deregisters the control (and it's settings)
		/// from this control list.
		/// </summary>
		protected override void OnControlRemoved(ControlEventArgs e)
		{
			if (e.Control is Splitter) {
				Splitter spl = (Splitter)e.Control;
				spl.SplitterMoved -= new SplitterEventHandler(SplitterMoved);
			} else {
				e.Control.VisibleChanged -= new EventHandler(ControlVisibleChanged);
				settings.Remove(e.Control);
			}
			base.OnControlRemoved(e);
		}
		
		/// <summary>
		/// Raises the <see cref="Control.ControlAdded"/> event and registers the control
		/// for this control list (so a <see cref="DockedControlSetting"/> object is created).
		/// </summary>
		protected override void OnControlAdded(ControlEventArgs e)
		{
			if (DesignMode) {
				e.Control.Dock = vertical ? DockStyle.Bottom : DockStyle.Right;
			} else {
				if (Controls.Count <= 1)
					e.Control.Dock = DockStyle.Fill;
				else
					e.Control.Dock = vertical ? DockStyle.Bottom : DockStyle.Right;
				if (e.Control is Splitter) {
					Splitter spl = (Splitter)e.Control; ;
					spl.SplitterMoved += new SplitterEventHandler(SplitterMoved);
				} else {
					e.Control.VisibleChanged += new EventHandler(ControlVisibleChanged);
					settings.Add(e.Control, new DockedControlSetting(e.Control));
				}
			}
			base.OnControlAdded(e);
		}
		
		/// <summary>
		/// Rearranges the child controls and raises the <see cref="Control.SizeChanged"/> event.
		/// </summary>
		protected override void OnSizeChanged(EventArgs e)
		{
			Arrange();
			base.OnSizeChanged(e);
		}
		#endregion
		
		private void Check()
		{
			if (base.IsDisposed) throw new ObjectDisposedException(GetType().Name);
			if (base.InvokeRequired) throw new InvokeRequiredException();
		}
	}
}
