// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Text;
using System.Reflection;
using System.Globalization;

namespace Grunwald.Gui.Controls
{
	/// <summary>
	/// Description of DragDropTreeView.
	/// </summary>
	public class DragDropTreeView : TreeViewWithPaint
	{
		#region Constructor
		/// <summary>
		/// Creates a new DragDropTreeView.
		/// </summary>
		public DragDropTreeView()
		{
			this.HideSelection = false;
			this.ItemHeight += 3;
			this.Indent += 3;
			this.AllowDrop = true;
		}
		#endregion
		
		#region Move/Copy helpers
		/// <summary>
		/// Gets if the node <c>child</c> is a child node of <c>parent</c>.
		/// This also returns true if child == parent.
		/// </summary>
		public static bool IsChildOf(TreeNode child, TreeNode parent)
		{
			TreeNode n = child;
			while (n != null) {
				if (n == parent)
					return true;
				n = n.Parent;
			}
			return false;
		}
		
		/// <summary>
		/// Gets the expanded map of the specified TreeNode.
		/// </summary>
		public static ExpandedMap GetExpandedMap(TreeNode node)
		{
			if (node.IsExpanded)
				return new ExpandedMap(node.Nodes);
			else
				return null;
		}
		
		/// <summary>
		/// Applies the expanded map to the specified TreeNode.
		/// </summary>
		public static void ApplyExpandedMap(TreeNode node, ExpandedMap map)
		{
			if (map == null) {
				if (node.IsExpanded)
					node.Collapse();
			} else {
				map.ApplyTo(node);
			}
		}
		
		/// <summary>
		/// Container class for the map of 'expanded'-values.
		/// You can use this class only with <see cref="GetExpandedMap"/> and
		/// <see cref="ApplyExpandedMap"/>.
		/// </summary>
		public sealed class ExpandedMap
		{
			ExpandedMap[] m;
			
			internal ExpandedMap(TreeNodeCollection nodes)
			{
				m = new ExpandedMap[nodes.Count];
				for (int i = 0; i < m.Length; i++) {
					m[i] = GetExpandedMap(nodes[i]);
				}
			}
			
			internal void ApplyTo(TreeNode node)
			{
				if (!node.IsExpanded)
					node.Expand();
				if (m.Length != node.Nodes.Count)
					throw new ArgumentException("Expanded Map doesn't fit the TreeNode");
				for (int i = 0; i < m.Length; i++) {
					ApplyExpandedMap(node.Nodes[i], m[i]);
				}
			}
		}
		#endregion
		
		#region Move/Copy
		/// <summary>
		/// Moves the specified node in this tree view.
		/// </summary>
		/// <param name="node">The node to move.</param>
		/// <param name="position">The position to move the node to.</param>
		/// <exception cref="ArgumentNullException">node or position is null.</exception>
		/// <exception cref="ArgumentOutOfRangeException">position.Index is out of the
		/// allowed range.</exception>
		/// <exception cref="ArgumentException">
		/// The position is not a position in this tree view.
		/// </exception>
		public void MoveNode(TreeNode node, TreeViewPosition position)
		{
			if (position == null)
				throw new ArgumentNullException("position");
			MoveNode(node, position.Parent, position.Index);
		}
		
		/// <summary>
		/// Moves the specified node in this tree view.
		/// </summary>
		/// <param name="node">The node to move.</param>
		/// <param name="newParent">The new parent node.
		/// <c>null</c> if the the node should be a root node.</param>
		/// <param name="newIndex">The new index the node should have.</param>
		/// <exception cref="ArgumentNullException">node is null.</exception>
		/// <exception cref="ArgumentOutOfRangeException">newIndex is out of the
		/// allowed range.</exception>
		/// <exception cref="ArgumentException">
		/// newParent is not a node in this tree view.
		/// </exception>
		public void MoveNode(TreeNode node, TreeNode newParent, int newIndex)
		{
			if (node == null)
				throw new ArgumentNullException("node");
			if (IsChildOf(newParent, node))
				throw new ArgumentException("Can't move a node into its own subfolder!");
			if (newParent != null && newParent.TreeView != this)
				throw new ArgumentException("newParent must be child node of this treeview");
			TreeNodeCollection col = (newParent == null) ? this.Nodes : newParent.Nodes;
			if (newIndex < 0 || newIndex > col.Count)
				throw new ArgumentOutOfRangeException("newIndex", newIndex, "Invalid index, index must be between 0 and " + (col.Count - 1).ToString());
			if (node.Parent == newParent && node.Index < newIndex)
				newIndex -= 1;
			node.Remove();
			col.Insert(newIndex, node);
			if (newParent != null)
				newParent.Expand();
			this.SelectedNode = node;
		}
		
		/// <summary>
		/// Copies the specified node in this tree view.
		/// The node's tag value is cloned, too, if it implements <see cref="ICloneable"/>.
		/// </summary>
		/// <param name="node">The node to copy.</param>
		/// <param name="position">The position to copy the node to.</param>
		/// <exception cref="ArgumentNullException">node or position is null.</exception>
		/// <exception cref="ArgumentOutOfRangeException">position.Index is out of the
		/// allowed range.</exception>
		/// <exception cref="ArgumentException">The position is not a position in this
		/// tree view.</exception>
		/// <returns>Returns the new tree node.</returns>
		public TreeNode CopyNode(TreeNode node, TreeViewPosition position)
		{
			if (position == null)
				throw new ArgumentNullException("position");
			return CopyNode(node, position.Parent, position.Index);
		}
		
		/// <summary>
		/// Copies the specified node in this tree view.
		/// The node's tag value is cloned, too, if it implements <see cref="ICloneable"/>.
		/// </summary>
		/// <param name="node">The node to copy.</param>
		/// <param name="newParent">The new parent node.
		/// <c>null</c> if the the node should be a root node.</param>
		/// <param name="newIndex">The new index the node should have.</param>
		/// <exception cref="ArgumentNullException">node is null.</exception>
		/// <exception cref="ArgumentOutOfRangeException">newIndex is out of the
		/// allowed range.</exception>
		/// <exception cref="ArgumentException">newParent is not a node in this
		/// tree view.</exception>
		/// <returns>Returns the new tree node.</returns>
		public TreeNode CopyNode(TreeNode node, TreeNode newParent, int newIndex)
		{
			if (node == null)
				throw new ArgumentNullException("node");
			if (newParent != null && newParent.TreeView != this)
				throw new ArgumentException("newParent must be child node of this treeview");
			TreeNodeCollection col = (newParent == null) ? this.Nodes : newParent.Nodes;
			if (newIndex < 0 || newIndex > col.Count)
				throw new ArgumentOutOfRangeException("newIndex", newIndex, "Invalid index, index must be between 0 and " + (col.Count - 1).ToString());
			ExpandedMap expanded = GetExpandedMap(node);
			TreeNode copy = (TreeNode)node.Clone();
			col.Insert(newIndex, copy);
			ApplyExpandedMap(copy, expanded);
			if (newParent != null)
				newParent.Expand();
			return copy;
		}
		#endregion
		
		#region Overridden events
		/// <summary>
		/// Raises the <see cref="Control.MouseDown">MouseDown event</see>.
		/// </summary>
		protected override void OnMouseDown(MouseEventArgs e)
		{
			this.SelectedNode = this.GetNodeAt(e.X, e.Y);
			base.OnMouseDown(e);
		}
		
		/// <summary>
		/// Raises the <see cref="Control.Paint">Paint event</see>.
		/// </summary>
		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e);
			OnPaintMiddleLayer(e);
			DrawPlaceholder(e.Graphics);
		}
		
		public event PaintEventHandler PaintMiddleLayer;
		
		protected virtual void OnPaintMiddleLayer(PaintEventArgs e)
		{
			if (PaintMiddleLayer != null) {
				PaintMiddleLayer(this, e);
			}
		}
		#endregion
		
		#region GetNodePosition
		/// <summary>
		/// Gets the position where a node would be be moved to if the user would drop
		/// on the specified mouse position.
		/// </summary>
		/// <param name="mousePos">The mouse position in client coordinates.<br/>
		/// Use PointToClient to convert screen coordinates to client coordinates.</param>
		/// <param name="allowMoveIn">Specifies if it is possible to a node into other nodes,
		/// so creating new folders.</param>
		/// <returns>Returns the position, or null if the method failed.</returns>
		public TreeViewPosition GetNodePosition(Point mousePos, bool allowMoveIn)
		{
			TreeNode over = this.GetNodeAt(mousePos);
			if (over == null) {
				if (this.Nodes.Count == 0)
					return new TreeViewPosition(0, null);
				if (mousePos.Y <= this.Nodes[0].Bounds.Top)
					return NodeBefore(mousePos, this.Nodes[0]);
				TreeNode lastNode = this.Nodes[this.Nodes.Count - 1];
				while (lastNode.IsExpanded && lastNode.Nodes.Count > 0) {
					lastNode = lastNode.Nodes[lastNode.Nodes.Count - 1];
				}
				return NodeAfter(mousePos, lastNode);
			}
			int offsetY = mousePos.Y - over.Bounds.Top;
			int height = over.Bounds.Height;
			if (allowMoveIn && offsetY < height / 4) {
				return NodeBefore(mousePos, over);
			} else if (allowMoveIn && offsetY < height *  3 / 4) {
				return new TreeViewPosition(over);
			} else if (offsetY < height / 2) {
				return NodeBefore(mousePos, over);
			} else {
				if (over.IsExpanded)
					return new TreeViewPosition(0, over);
				return NodeAfter(mousePos, over);
			}
		}
		
		private TreeViewPosition NodeBefore(Point mousePos, TreeNode over)
		{
			// over is first node?
			if (over.Index == 0)
				return new TreeViewPosition(0, over.Parent);
			// it's possible that the item above has children, so get the
			// item above the current and call NodeAfter on it to support
			// level-inserting based on the x coordinate of the mouse
			return NodeAfter(mousePos, over.PrevVisibleNode);
		}
		
		private TreeViewPosition NodeAfter(Point mousePos, TreeNode over)
		{
			int x = mousePos.X - over.Bounds.Left + placeholderLeftSize;
			while (x < 0) {
				if (over.Parent == null)
					break;
				over = over.Parent;
				x = mousePos.X - over.Bounds.Left + placeholderLeftSize;
			}
			return new TreeViewPosition(over.Index + 1, over.Parent);
		}
		
		/// <summary>
		/// Gets the position where a node would be be moved to if the user would drop
		/// on the specified mouse position.
		/// </summary>
		/// <param name="mousePos">The mouse position in client coordinates.<br/>
		/// Use PointToClient to convert screen coordinates to client coordinates.</param>
		/// <returns>Returns the position, or null if the method failed.</returns>
		public TreeViewPosition GetNodePosition(Point mousePos)
		{
			return GetNodePosition(mousePos, false);
		}
		#endregion
		
		#region Placeholder properties
		TreeViewPosition placeholder;
		
		/// <summary>
		/// Gets/Sets the position where the placeholder should be drawn.
		/// Set this to <c>null</c> if no placeholder should be drawn.
		/// </summary>
		[Browsable(false)]
		public TreeViewPosition Placeholder {
			get {
				return placeholder;
			}
			set {
				if (placeholder == value)
					return;
				placeholder = value;
				this.Invalidate();
			}
		}
		
		int placeholderLeftSize = 22;
		
		/// <summary>
		/// Gets/Sets the size the placeholder is longer than normal on the left side.
		/// The default value is 22, decrease it if you don't use icons in your tree view.
		/// </summary>
		[Description("the size the placeholder is longer than normal on the left side." +
		             "The default value is 22, decrease it if you don't use icons in your tree view."),
		 DefaultValue(22)]
		public int PlaceholderLeftSize {
			get {
				return placeholderLeftSize;
			}
			set {
				if (placeholderLeftSize == value)
					return;
				placeholderLeftSize = value;
				if (placeholder != null)
					this.Invalidate();
			}
		}
		#endregion
		
		#region Draw placeholder
		private void DrawPlaceholder(Graphics g)
		{
			if (placeholder == null)
				return;
			if (placeholder.IsMoveIn) {
				int x = placeholder.Parent.Bounds.Right + 6;
				int y = placeholder.Parent.Bounds.Top + placeholder.Parent.Bounds.Height / 2;
				Point[] rightTriangle = new Point[] {
					new Point(x,     y + 4),
					new Point(x,     y + 4),
					new Point(x - 4, y    ),
					new Point(x - 4, y - 1),
					new Point(x,     y - 5)};
				g.FillPolygon(Brushes.Black, rightTriangle);
			} else if (this.Nodes.Count == 0) {
				DrawPlaceholder(g, 5, 5, this.Width - 10);
			} else {
				int y;
				TreeNodeCollection nodes;
				if (placeholder.Parent == null)
					nodes = this.Nodes;
				else
					nodes = placeholder.Parent.Nodes;
				TreeNode node;
				if (placeholder.Index >= nodes.Count) {
					node = nodes[nodes.Count - 1];
					TreeNode n = node;
					while(n.Nodes.Count > 0 && n.IsExpanded) {
						n = n.LastNode;
					}
					y = n.Bounds.Bottom;
				} else {
					node = nodes[placeholder.Index];
					y = node.Bounds.Top;
				}
				int x = node.Bounds.Left - placeholderLeftSize;
				int x2 = node.Bounds.Right;
				DrawPlaceholder(g, x, y, x2);
			}
		}
		
		private void DrawPlaceholder(Graphics g, int x, int y, int x2)
		{
			Point[] leftTriangle = new Point[] {
				new Point(x,     y - 4),
				new Point(x,     y + 4),
				new Point(x + 4, y    ),
				new Point(x + 4, y - 1),
				new Point(x,     y - 5)};
			
			Point[] rightTriangle = new Point[5] {
				new Point(x2,     y - 4),
				new Point(x2,     y + 4),
				new Point(x2 - 4, y    ),
				new Point(x2 - 4, y - 1),
				new Point(x2,     y - 5)};
			
			g.FillPolygon(Brushes.Black, leftTriangle);
			g.FillPolygon(Brushes.Black, rightTriangle);
			using (Pen pen = new Pen(Color.Black, 2)) {
				g.DrawLine(pen, x, y, x2, y);
			}
		}
		#endregion
	}
	
	/// <summary>
	/// Class for the position of a tree view item.
	/// </summary>
	public sealed class TreeViewPosition
	{
		bool moveIn;
		int index;
		TreeNode parent;
		
		/// <summary>
		/// Creates a new position object with the specified index and parent.
		/// </summary>
		public TreeViewPosition(int index, TreeNode parent)
		{
			this.index = index;
			this.parent = parent;
			this.moveIn = false;
		}
		
		/// <summary>
		/// Creates a position object with the index 0 and the
		/// <see cref="IsMoveIn">MoveIn</see> parameter set.
		/// </summary>
		public TreeViewPosition(TreeNode parent)
		{
			this.index = 0;
			this.parent = parent;
			this.moveIn = true;
		}
		
		/// <summary>
		/// Gets the index of the item location.
		/// </summary>
		public int Index {
			get {
				return index;
			}
		}
		
		/// <summary>
		/// Gets the parent of the item location.
		/// </summary>
		public TreeNode Parent {
			get {
				return parent;
			}
		}
		
		/// <summary>
		/// Gets if this location is a 'move in' location (moving a node into a
		/// not-expanded folder)
		/// </summary>
		public bool IsMoveIn {
			get {
				return moveIn;
			}
		}
		
		/// <summary>
		/// Tests two objects for equality.
		/// </summary>
		public override bool Equals(object obj)
		{
			TreeViewPosition b = obj as TreeViewPosition;
			if (b == null) return false;
			return (this.Index == b.Index && this.Parent == b.Parent && this.IsMoveIn == b.IsMoveIn);
		}
		
		/// <summary>
		/// Gets the hash code of this TreeViewPosition.
		/// </summary>
		public override int GetHashCode()
		{
			unchecked {
				return 27 * this.Index.GetHashCode() + this.Parent.GetHashCode();
			}
		}
		
		/// <summary>
		/// Tests two objects for equality.
		/// </summary>
		public static bool operator ==(TreeViewPosition a, TreeViewPosition b)
		{
			if ((object)a == null && (object)b == null) return true;
			if ((object)a == null || (object)b == null) return false;
			return a.Equals(b);
		}
		
		/// <summary>
		/// Tests two objects for inequality.
		/// </summary>
		public static bool operator !=(TreeViewPosition a, TreeViewPosition b)
		{
			if ((object)a == null && (object)b == null) return false;
			if ((object)a == null || (object)b == null) return true;
			return !a.Equals(b);
		}
		
		/// <summary>
		/// Gets the string representation for debugging purposes.
		/// </summary>
		public override string ToString()
		{
			if (moveIn) {
				return String.Format("[TreeViewPosition: MoveIn Parent = {0}]",
				                     parent == null ? "null" : parent.ToString());
			} else {
				return String.Format("[TreeViewPosition: Parent = {0}, Index = {1}]",
				                     parent == null ? "null" : parent.ToString(),
				                     index);
			}
		}
	}
}
