// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.ComponentModel;
using System.Text;
using System.Windows.Forms;
using Grunwald.Gui.Components;

namespace Grunwald.Gui.Controls
{
	/// <summary>
	/// Description of HexViewer.
	/// </summary>
	public class HexViewer : System.Windows.Forms.UserControl
	{
		#region Constructor
		RichTextBox textView = new RichTextBox();
		RichTextBox hexView = new RichTextBox();
		RichTextBox addrView = new RichTextBox();
		
		MenuItem menuItem_copy = new MenuItem("Copy");
		ContextMenu context_menu = new ContextMenu();
		RichTextBox rtb_last_clicked;
		
		/// <summary>
		/// Creates a new HexViewer
		/// </summary>
		public HexViewer()
		{
			context_menu.MenuItems.Add(menuItem_copy);
			menuItem_copy.Click += new System.EventHandler(this.menuItem_copy_Click);
			rtb_last_clicked = hexView;
			
			Font f = new Font("Courier New", 8.25F, FontStyle.Regular);
			hexView.Dock = System.Windows.Forms.DockStyle.Left;
			hexView.ScrollBars = RichTextBoxScrollBars.None;
			hexView.TabIndex = 1;
			hexView.WordWrap = false;
			hexView.Font = f;
			
			hexView.AutoWordSelection = false;
			textView.AutoWordSelection = false;
			
			textView.WordWrap = false;
			textView.Dock = System.Windows.Forms.DockStyle.Fill;
			textView.ScrollBars = RichTextBoxScrollBars.ForcedVertical;
			textView.TabIndex = 2;
			textView.Font = f;
			textView.DetectUrls = false;
			
			hexView.HideSelection = false;
			textView.HideSelection = false;
			
			addrView.Dock = DockStyle.Left;
			addrView.HideSelection = false;
			addrView.ScrollBars = RichTextBoxScrollBars.None;
			addrView.WordWrap = false;
			addrView.Font = f;
			addrView.Width = 65;// depending of format
			
			
			hexView.SelectionChanged  += new EventHandler(HexSelectionChanged);
			textView.SelectionChanged += new EventHandler(TextSelectionChanged);
			addrView.SelectionChanged += new EventHandler(addrView_SelectionChanged);
			
			hexView.MouseUp += new MouseEventHandler(RichTextBox_MouseUp);
			textView.MouseUp += new MouseEventHandler(RichTextBox_MouseUp);
			
			ScrollSynchronizer sync = new ScrollSynchronizer();
			sync.Add(hexView);
			sync.Add(addrView);
			sync.Add(textView);
			
			
			addrView.ReadOnly = true;
			hexView.ReadOnly = true;
			textView.ReadOnly = true;
			
			this.Controls.Add(this.textView);
			this.Controls.Add(this.hexView);
			this.Controls.Add(this.addrView);
			this.Size = new Size(424, 280);
		}
		
		float fontWidth = 7;
		
		void Check()
		{
			if (IsDisposed)
				throw new ObjectDisposedException("HexViewer");
			if (InvokeRequired)
				throw new InvokeRequiredException();
		}
		#endregion
		
		#region Overridden Events
		/// <summary>
		/// Raises the <see cref="Control.SizeChanged">SizeChanged</see> event.
		/// </summary>
		protected override void OnSizeChanged(System.EventArgs e)
		{
			base.OnSizeChanged(e);
			if (visible_addr)
				hexView.Width = (int)((ClientSize.Width - addrView.Width) * 0.65);
			else
				hexView.Width = (int)(ClientSize.Width * 0.65);
			UpdateView(false);
		}
		
		bool ignoreSelChange;
		
		private void TextSelectionChanged(object sender, EventArgs e)
		{
			if (ignoreSelChange) return;
			int pos, row, col;
			pos = textView.SelectionStart;
			row = pos / (byteCount + 1);
			col = pos % (byteCount + 1);
			int startByte = row * byteCount + col;
			
			pos += textView.SelectionLength;
			row = pos / (byteCount + 1);
			col = pos % (byteCount + 1);
			int endByte = row * byteCount + col;
			Select(startByte, endByte - startByte, 1);
		}
		
		private void HexSelectionChanged(object sender, EventArgs e)
		{
			if (ignoreSelChange) return;
			int pos, row;
			pos = hexView.SelectionStart;
			row = pos / (lineChars + 1);
			int startByte = row * byteCount + calcByteCol(pos);
			
			pos += hexView.SelectionLength;
			row = pos / (lineChars + 1);
			int endByte = row * byteCount + calcByteCol(pos);
			Select(startByte, endByte - startByte, 2);
		}
		
		private void addrView_SelectionChanged(object sender, EventArgs e)
		{
			if (ignoreSelChange) return;
			int pos, row;
			pos = addrView.SelectionStart;
			row=pos/9;//8+\n
			Select(row * byteCount, 0, 0);
		}
		
		int calcByteCol(int pos) {
			int col = pos % (lineChars + 1);
			int bcol = 0;
			while(col > 0) {
				bcol += 1;
				col -= 2;
				if (bcol % byteCount == 0)
					col -= 1;
				else if (blockLength > 0 && bcol % blockLength == 0)
					col -= 1;
			}
			return bcol;
		}
		
		private void RichTextBox_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Right) {
				this.rtb_last_clicked = (RichTextBox)sender;
				this.context_menu.Show((Control)sender,new Point(e.X,e.Y));
			}
		}
		
		private void menuItem_copy_Click(object sender, System.EventArgs e)
		{
			Clipboard.SetDataObject(this.rtb_last_clicked.SelectedText, true);
		}
		#endregion
		
		#region Properties
		byte[] data = null;
		
		/// <summary>
		/// Gets/Sets the data to display.
		/// </summary>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		public byte[] Data {
			get {
				return data;
			}
			set {
				if (data == value) return;
				Check();
				data = value;
				Select(0, 0);
				UpdateView();
				OnDataChanged(EventArgs.Empty);
			}
		}
		
		int blockLength = 2;
		
		/// <summary>
		/// Gets/Sets the length of a block of connected bytes. After each block,
		/// a additional space is inserted in the hex view. Line breaks are only
		/// possible after a full block.<br/>
		/// If this property is changed to 0 or a negative value, byte blocks are
		/// disabled.
		/// </summary>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		public int BlockLength {
			get {
				return blockLength;
			}
			set {
				if (value == blockLength) return;
				Check();
				blockLength = value;
				UpdateView();
			}
		}
		
		bool visible_addr = true;
		
		/// <summary>
		/// Gets/Sets a value indicating whether the address RichTextBox is displayed.
		/// </summary>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		public bool AddressVisible {
			get {
				return visible_addr;
			}
			set {
				if (value == visible_addr) return;
				Check();
				visible_addr = value;
				// compute new size for RichTextBox
				OnSizeChanged(EventArgs.Empty);
				UpdateView();
			}
		}

		bool uppercaseHex = true;
		
		/// <summary>
		/// Gets/Sets whether the hexadicimal letters are uppercase.
		/// </summary>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		public bool UppercaseHex {
			get {
				return uppercaseHex;
			}
			set {
				if (value == uppercaseHex) return;
				Check();
				uppercaseHex = value;
				UpdateView();
			}
		}
		#endregion
		
		#region Selection
		int selectionStart;
		int selectionLength;
		
		/// <summary>
		/// Selects the specified range.
		/// </summary>
		/// <param name="start">The start index of the selection.</param>
		/// <param name="length">The length of the selection.</param>
		/// <exception cref="ArgumentOutOfRangeException">
		/// The specified range is not a valid range for the current data block.
		/// </exception>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		public void Select(int start, int length)
		{
			Check();
			int len = data==null ? 0 : data.Length;
			if (start < 0 || start > len)
				throw new ArgumentOutOfRangeException("start", start, "start must be between 0 and " + len.ToString());
			if (length < 0 || start + length > len)
				throw new ArgumentOutOfRangeException("length", length, "length must be between 0 and " + (len - start).ToString());
			Select(start, length, 0);
		}
		
		void Select(int start, int length, int mode)
		{
			bool dirty;
			if (start != selectionStart || length != selectionLength)
				dirty = true;
			else
				dirty = false;
			selectionStart = start;
			selectionLength = length;
			
			int row1 = start / byteCount;
			int col1 = start % byteCount;
			int row2 = (start + length) / byteCount;
			int col2 = (start + length) % byteCount;
			
			int textStart = row1 * (byteCount + 1) + col1;
			int textEnd   = row2 * (byteCount + 1) + col2;
			
			
			int hexStart, hexEnd;
			if (blockLength > 0) {
				int group1 = col1 / blockLength;
				int group2 = col2 / blockLength;
				hexStart = row1 * (lineChars + 1) + col1 * 2 + group1;
				hexEnd   = row2 * (lineChars + 1) + col2 * 2 + group2;
			} else {
				hexStart = row1 * (lineChars + 1) + col1 * 2;
				hexEnd   = row2 * (lineChars + 1) + col2 * 2;
			}
			
			ignoreSelChange = true;
			if (mode != 1) textView.Select(textStart, textEnd - textStart);
			if (mode != 2) hexView.Select(hexStart,   hexEnd -  hexStart);
			addrView.SelectionStart=row2*9;//8+\n
			ignoreSelChange = false;
			if (dirty)
				OnSelectionChanged(EventArgs.Empty);
		}
		
		/// <summary>
		/// Gets/Sets the start position of the selection.
		/// </summary>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// The specified start position is outside the current data block.
		/// </exception>
		public int SelectionStart {
			get {
				return selectionStart;
			}
			set {
				int len = data==null?0:data.Length;
				if (value + selectionLength > len)
					Select(value, len - value);
				else
					Select(value, selectionLength);
			}
		}
		
		/// <summary>
		/// Gets/Sets the length of the selection.
		/// </summary>
		/// <exception cref="InvokeRequiredException">This method is called from an other thread than the thread that created this control.</exception>
		/// <exception cref="ObjectDisposedException">This control was already disposed.</exception>
		/// <exception cref="ArgumentOutOfRangeException">
		/// The specified range is not a valid range for the current data block.
		/// </exception>
		public int SelectionLength {
			get {
				return selectionLength;
			}
			set {
				Select(selectionStart, value);
			}
		}
		#endregion
		
		#region UpdateView
		int byteCount = 1; // number of bytes in one line
		int lineChars = 1; // number of hex characters in each line
		
		void UpdateView() { UpdateView(true); }
		
		Encoding enc = Encoding.GetEncoding("ISO8859-1");
		
		void UpdateView(bool force)
		{
			if (data == null) {
				hexView.Text = "";
				textView.Text = "";
				addrView.Text = "";
				return;
			}
			
			addrView.Visible = visible_addr;
			
			float avaSize = hexView.Width - 5;
			
			int charCount = (int)(avaSize / fontWidth);
			int oldByteCount = byteCount;
			if (blockLength > 0) {
				byteCount = (charCount + 1) / (blockLength * 2 + 1) * blockLength;
				if (byteCount == 0)
					byteCount = blockLength;
				lineChars = byteCount * 2 + byteCount / blockLength - 1;
			} else {
				byteCount = charCount / 2;
				if (byteCount == 0)
					byteCount = 1;
				lineChars = byteCount * 2;
			}
			
			if (!force && oldByteCount == byteCount)
				return;
			
			StringBuilder b = new StringBuilder();
			StringBuilder b2 = new StringBuilder();
			
			for (int i = 0; i < data.Length; i++) {
				if (i > 0) {
					if (i % byteCount == 0) {
						b.Append('\n');
						b2.Append('\n');
					} else if (blockLength > 0 && i % blockLength == 0) {
						b.Append(' ');
					}
				}
				byte by = data[i];
				b.Append(by.ToString(uppercaseHex ? "X2" : "x2"));
				if (by >= 32 && by < 127)
					b2.Append((char)by);
				else if (by > 160)
					b2.Append(enc.GetChars(new byte[] {by})[0]);
				else
					b2.Append('.');
			}
			hexView.Text = b.ToString();
			textView.Text = b2.ToString();
			
			StringBuilder baddr = new StringBuilder();
			for (int i = 0; i < data.Length; i += byteCount) {
				if (baddr.Length > 0)
					baddr.Append('\n');
				baddr.Append(i.ToString(uppercaseHex ? "X8" : "x8"));
			}
			addrView.Text = baddr.ToString();
			Select(selectionStart, selectionLength);
		}
		#endregion
		
		#region Events
		/// <summary>
		/// Is raised whenever the <see cref="Data">data property</see> is changed.
		/// </summary>
		public event EventHandler DataChanged;
		
		/// <summary>
		/// Is raised whenever the selection is changed.
		/// </summary>
		public event EventHandler SelectionChanged;
		
		/// <summary>
		/// Raises the <see cref="DataChanged">DataChanged event</see>.
		/// </summary>
		protected virtual void OnDataChanged(EventArgs e) {
			if (DataChanged != null) {
				DataChanged(this, e);
			}
		}
		
		/// <summary>
		/// Raises the <see cref="SelectionChanged">SelectionChanged event</see>.
		/// </summary>
		protected virtual void OnSelectionChanged(EventArgs e) {
			if (SelectionChanged != null) {
				SelectionChanged(this, e);
			}
		}
		#endregion
	}
}
