// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace Grunwald.Gui.Controls
{
	/// <summary>
	/// A drag'n'drop tree view that uses the
	/// <see cref="ITreeTag">ITreeTag interface</see>.
	/// </summary>
	public class RearrangeTreeView : DragDropTreeView
	{
		#region Delete
		bool canDelete = false;
		
		/// <summary>
		/// Gets/Sets whether the tree nodes can be deleted using the 'del' key.
		/// </summary>
		[Description("Specifies whether the tree nodes can be deleted using the 'del' key."),
		 DefaultValue(false)]
		public bool CanDelete {
			get {
				return canDelete;
			}
			set {
				canDelete = value;
			}
		}
		
		/// <summary>
		/// Raises the <see cref="Control.KeyUp">KeyUp event</see>.
		/// </summary>
		/// <remarks>
		/// If <see cref="CanDelete"/> is <c>true</c> and the 'del' key was pressed,
		/// the currently selected node is removed.
		/// </remarks>
		protected override void OnKeyUp(KeyEventArgs e)
		{
			base.OnKeyUp(e);
			if (e.KeyData == Keys.Delete && this.SelectedNode != null)
				Delete(this.SelectedNode);
		}
		
		/// <summary>
		/// Deletes the specified node is its tag is valid and the delete operation is allowed.
		/// </summary>
		/// <returns>
		/// Returns <c>true</c> if the delete operation was allowed and successful.
		/// </returns>
		public virtual bool Delete(TreeNode node)
		{
			if (node == null)
				throw new ArgumentNullException("node");
			ITreeTag tag = node.Tag as ITreeTag;
			if (tag != null && tag.CanDelete) {
				if (autoTreeNodeManagement)
					node.Remove();
				tag.Delete();
				return true;
			} else {
				return false;
			}
		}
		#endregion
		
		bool canRearrange = true;
		
		/// <summary>
		/// Gets/Sets whether the tree nodes can be rearranged.
		/// </summary>
		[Description("Specifies whether the tree nodes can be rearranged."),
		 DefaultValue(true)]
		public bool CanRearrange {
			get {
				return canRearrange;
			}
			set {
				canRearrange = value;
			}
		}
		
		/// <summary>
		/// Raises the <see cref="TreeView.ItemDrag">ItemDrag event</see>.
		/// </summary>
		/// <remarks>This method starts dragging is the item is a draggable node.</remarks>
		protected override void OnItemDrag(ItemDragEventArgs e)
		{
			base.OnItemDrag(e);
			if (canRearrange && !DesignMode) {
				TreeNode node = e.Item as TreeNode;
				if (node == null) return;
				ITreeTag tag = node.Tag as ITreeTag;
				if (tag == null) return;
				if (tag.CanDrag)
					StartDrag(this, node);
			}
		}
		
		static DataWrapper lastDataWrapper;
		
		public static void StartDrag(Control control, TreeNode node)
		{
			ITreeTag tag = (ITreeTag)node.Tag;
			DragDropEffects effects = DragDropEffects.Move | DragDropEffects.Link;
			if (tag.CanClone)
				effects |= DragDropEffects.Copy;
			StartDrag(control, node, effects);
		}
		
		public static void StartDrag(Control control, TreeNode node, DragDropEffects effects)
		{
			DataWrapper w = new DataWrapper(node);
			lastDataWrapper = w;
			control.DoDragDrop(w, effects);
		}
		
		/// <summary>
		/// Raises the <see cref="Control.DragEnter">DragEnter event</see>.
		/// </summary>
		/// <remarks>This method sets the effect if the contained data is a tree node
		/// with an <see cref="ITreeTag"/>.</remarks>
		protected override void OnDragEnter(DragEventArgs e)
		{
			base.OnDragEnter(e);
			if (canRearrange && !DesignMode) {
				if (e.Data.GetDataPresent(typeof(DataWrapper)))
					SetEffect(e);
			}
		}
		
		/// <summary>
		/// Wrapper around the tree node, used as data object.
		/// </summary>
		[Serializable]
		public class DataWrapper
		{
			TreeNode node;
			
			public TreeNode Node {
				get {
					return node;
				}
			}
			
			public ITreeTag Tag {
				get {
					try {
						return (ITreeTag)node.Tag;
					} catch (System.Runtime.Remoting.RemotingException) {
						return null;
					}
				}
			}
			
			/// <summary>
			/// Try to get the original node instance back (by default .NET uses a deserialized version).
			/// The Node property is changed to the original node.
			/// </summary>
			public bool GetOriginalNode()
			{
				if (lastDataWrapper == null)
					return false;
				if (lastDataWrapper.internalID != internalID)
					return false;
				this.node = lastDataWrapper.node;
				return true;
			}
			
			int internalID;
			static Random rnd = new Random();
			
			public DataWrapper(TreeNode node)
			{
				this.node = node;
				internalID = rnd.Next();
			}
		}
		
		private void SetEffect(DragEventArgs e)
		{
			DragDropEffects moveCopy = DragDropEffects.Copy | DragDropEffects.Move;
			DragDropEffects possible = e.AllowedEffect & moveCopy;
			if (possible != moveCopy) {
				e.Effect = possible;
			} else {
				if ((Control.ModifierKeys & Keys.Control) == Keys.Control)
					e.Effect = DragDropEffects.Copy;
				else
					e.Effect = DragDropEffects.Move;
			}
		}
		
		/// <summary>
		/// Raises the <see cref="Control.DragDrop">DragDrop event</see>.
		/// </summary>
		/// <remarks>This method accepts the dragged item if it is a valid item.</remarks>
		protected override void OnDragDrop(DragEventArgs e)
		{
			isPerformingDrop = true;
			try {
				base.OnDragDrop(e);
				this.Placeholder = null;
				if (canRearrange && !DesignMode)
					DoDrop(e);
			} catch (Exception ex) {
				ErrorReporter.ShowErrorReport(ex);
			}
			isPerformingDrop = false;
		}
		
		ITreeTag rootTag;
		
		/// <summary>
		/// Gets/Sets the tag to use as 'root tag' (invisible tag that
		/// contains the tags of the root nodes).
		/// </summary>
		public ITreeTag RootTag {
			get {
				return rootTag;
			}
			set {
				rootTag = value;
			}
		}
		
		bool autoTreeNodeManagement = true;
		bool autoMoveInsideParent = true;
		
		/// <summary>
		/// Gets/Sets whether the RearrangeTreeView automatically moves/copies/deletes
		/// the tree nodes if the move is NOT local (parent was changed).
		/// If this is set to false, the TreeTag is responsible for changing the tree nodes.
		/// </summary>
		public bool AutoTreeNodeManagement {
			get {
				return autoTreeNodeManagement;
			}
			set {
				autoTreeNodeManagement = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets whether the RearrangeTreeView automatically moves the tree nodes
		/// if the move is local (parent not changed)
		/// </summary>
		public bool AutoMoveInsideParent {
			get {
				return autoMoveInsideParent;
			}
			set {
				autoMoveInsideParent = value;
			}
		}
		
		bool isPerformingDrop;
		
		/// <summary>
		/// Gets if a drop operation is currently being performed.
		/// </summary>
		public bool IsPerformingDrop {
			get {
				return isPerformingDrop;
			}
		}
		
		ITreeTag GetTag(TreeNode node)
		{
			if (node == null)
				return rootTag;
			else
				return node.Tag as ITreeTag;
		}
		
		void DoDrop(DragEventArgs e)
		{
			SetEffect(e);
			DataWrapper dataWrapper = (DataWrapper)e.Data.GetData(typeof(DataWrapper));
			if (dataWrapper == null)
				return;
			dataWrapper.GetOriginalNode();
			TreeNode node = dataWrapper.Node;
			ITreeTag tag;
			Point mouse = this.PointToClient(new Point(e.X, e.Y));
			TreeNode over = this.GetNodeAt(mouse);
			ITreeTag nodeTag = dataWrapper.Tag;
			bool folder = false;
			if (over != null && over.Tag != null && over.Tag is ITreeTag && nodeTag != null) {
				folder = (over.Tag as ITreeTag).AcceptChild(nodeTag, 0);
			}
			TreeViewPosition pos = this.GetNodePosition(mouse, folder);
			if (pos == null)
				return;
			ITreeTag pTag = GetTag(pos.Parent);
			if (pTag == null || nodeTag == null || !pTag.AcceptChild(nodeTag, pos.Index))
				return;
			ITreeTag nTag = GetTag(node.Parent);
			
			if (e.Effect == DragDropEffects.Copy) {
				if (autoTreeNodeManagement) {
					TreeNode copy = this.CopyNode(node, pos);
					tag = CloneTag(copy);
					if (tag == null)
						throw new GuiException("Error: cloned copy of ITreeTag is no ITreeTag");
					//OnNodeCopyCreated(new NodeCopyEventArgs(node, copy));
					pTag.AddChild(tag, pos.Index);
					this.SelectedNode = copy;
				} else {
					pTag.AddChild((ITreeTag)nodeTag.Clone(), pos.Index);
				}
			} else if (DragDropTreeView.IsChildOf(pos.Parent, node)) {
				return;
			} else {
				if (node.Parent == pos.Parent) {
					// simple move without parent change
					if (autoMoveInsideParent)
						this.MoveNode(node, pos);
					if (nTag != null)
						nTag.MoveChild(nodeTag, pos.Index);
				} else {
					// move with changing parent
					if (nTag != null)
						nTag.RemoveChild(nodeTag);
					if (autoTreeNodeManagement)
						this.MoveNode(node, pos);
					if (pTag != null)
						pTag.AddChild(nodeTag, pos.Index);
				}
			}
		}
		
		/// <summary>
		/// Raises the <see cref="Control.DragOver">DragOver event</see>.
		/// </summary>
		/// <remarks>This method draws the placeholder if the dragged item is a valid item.</remarks>
		protected override void OnDragOver(DragEventArgs e)
		{
			try {
				base.OnDragOver(e);
				if (canRearrange && !DesignMode)
					this.Placeholder = DoDragOver(e);
			} catch (Exception ex) {
				ErrorReporter.ShowErrorReport(ex);
			}
		}
		
		TreeViewPosition DoDragOver(DragEventArgs e)
		{
			SetEffect(e);
			DataWrapper dataWrapper = (DataWrapper)e.Data.GetData(typeof(DataWrapper));
			if (dataWrapper == null)
				return null;
			dataWrapper.GetOriginalNode();
			TreeNode node = dataWrapper.Node;
			Point mouse = this.PointToClient(new Point(e.X, e.Y));
			TreeNode over = this.GetNodeAt(mouse);
			bool folder = false;
			ITreeTag nodeTag = dataWrapper.Tag;
			if (over != null && over.Tag != null && over.Tag is ITreeTag && nodeTag != null)
				folder = (over.Tag as ITreeTag).AcceptChild(nodeTag, 0);
			TreeViewPosition pos = this.GetNodePosition(mouse, folder);
			if (pos == null)
				return null;
			ITreeTag ptag = GetTag(pos.Parent);
			if (ptag == null || nodeTag == null || !ptag.AcceptChild(nodeTag, pos.Index))
				return null;
			
			if (e.Effect == DragDropEffects.Copy) {
				
			} else if (DragDropTreeView.IsChildOf(pos.Parent, node)) {
				return null;
			} else if (node.Parent != pos.Parent) {
				ptag = GetTag(node.Parent);
				if (ptag == null)
					return null;
			}
			return pos;
		}
		
		/// <summary>
		/// Raises the <see cref="Control.DragLeave">DragLeave event</see>.
		/// </summary>
		/// <remarks>This method clears the placeholder.</remarks>
		protected override void OnDragLeave(EventArgs e)
		{
			base.OnDragLeave(e);
			this.Placeholder = null;
		}
		
		private ITreeTag CloneTag(TreeNode node)
		{
			if (node.Tag is ICloneable)
				node.Tag = ((ICloneable)node.Tag).Clone();
			ITreeTag tag = node.Tag as ITreeTag;
			foreach(TreeNode subnode in node.Nodes) {
				ITreeTag subclone = CloneTag(subnode);
				if (subclone != null && tag != null)
					subclone.Parent = tag;
			}
			return tag;
		}
	}
	
	/// <summary>
	/// Interface used by the <see cref="RearrangeTreeView"/> as Tag element for
	/// <see cref="TreeNode">TreeNodes</see>.
	/// </summary>
	public interface ITreeTag : ICloneable
	{
		/// <summary>
		/// Gets if this tree node can accept the specified node as a child.
		/// </summary>
		bool AcceptChild(ITreeTag child, int index);
		
		/// <summary>
		/// Gets if the object can be cloned.
		/// </summary>
		bool CanClone { get; }
		
		/// <summary>
		/// Gets if this node can be dragged around.
		/// </summary>
		bool CanDrag { get; }
		
		/// <summary>
		/// Gets if this node can be delete using the
		/// <see cref="RearrangeTreeView.Delete">Delete method</see>.
		/// </summary>
		bool CanDelete { get; }
		
		/// <summary>
		/// Deletes this tree node.<br/>
		/// This method is responsible for calling <see cref="RemoveChild"/> on its
		/// parent!
		/// </summary>
		void Delete();
		
		/// <summary>
		/// Adds a child to this tree node.
		/// </summary>
		void AddChild(ITreeTag node, int index);
		
		/// <summary>
		/// Moves a child inside this tree node.
		/// </summary>
		void MoveChild(ITreeTag node, int index);
		
		/// <summary>
		/// Removes a child from this tree node.
		/// </summary>
		void RemoveChild(ITreeTag node);
		
		/// <summary>
		/// Gets the parent page of the wrapped page.
		/// </summary>
		ITreeTag Parent {
			get;
			set;
		}
	}
}
