////////////////////////////////////////////////
////          Tree View with Paint          ////
////  Version 1.0.0               09.10.04  ////
////////////////////////////////////////////////
// This class is based on http://www.codeproject.com/cs/miscctrl/genmissingpaintevent.asp
//  by J. Young

using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace Grunwald.Gui.Controls
{
	/// <summary>
	/// Displays a hierarchical collection of labeled items, each represented by a System.Windows.Forms.TreeNode.
	/// </summary>
	[
	 //Use the same attibutes as TreeView
	 DesignerAttribute("System.Windows.Forms.Design.TreeViewDesigner, System.Design, Version=1.0.3300.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a"),
	 DefaultPropertyAttribute("Nodes"),
	 DefaultEventAttribute("AfterSelect")
	]
	public class TreeViewWithPaint : TreeView
	{
		#if !MONO
		Bitmap   internalBitmap   = null;
		Graphics internalGraphics = null;
		
		private void DisposeInternal()
		{
			if (internalGraphics != null)
				internalGraphics.Dispose();
			if (internalBitmap != null)
				internalBitmap.Dispose();
		}
		
		/// <summary>Releases resources.</summary>
		/// <param name="disposing">true = Both managed and unmanaged, false = Unmanaged only.</param>
		protected override void Dispose(bool disposing)
		{
			if (disposing)
				DisposeInternal();
			base.Dispose(disposing);
		}
		
		/// <summary>Occurs when window is resized.</summary>
		/// <remarks>Recreates internal Graphics object. </remarks>
		protected override void OnResize(System.EventArgs e)
		{
			if (!DesignMode) {
				if (internalBitmap == null ||
				    internalBitmap.Width != Width || internalBitmap.Height != Height) {
					
					if (Width > 0 && Height > 0) {
						DisposeInternal();
						internalBitmap = new Bitmap (Width, Height);
						internalGraphics = Graphics.FromImage(internalBitmap);
					}
				}
			}
			base.OnResize(e);
		}
		
		/// <summary>Occurs when a Windows message is dispatched.</summary>
		/// <param name="message">Message to process.</param>
		/// <remarks>Overrides WM_PAINT, WM_ERASEBKGND.</remarks>
		protected override void WndProc(ref Message message)
		{
			if (DesignMode) {
				base.WndProc(ref message);
				return;
			}
			const int WM_PAINT = 0x000F;
			const int WM_PRINTCLIENT = 0x0318;
			const int WM_ERASEBKGND = 0x0014;
			
			switch (message.Msg) {
				case WM_ERASEBKGND:
					//removes flicker
					break;
					
				case WM_PAINT:
					try {
						// The designer host does not call OnResize()
						if (internalGraphics == null)
							OnResize(EventArgs.Empty);
						
						//Set up
						RECT updateRect = new RECT();
						if (Win32.GetUpdateRect(message.HWnd, ref updateRect, false) == 0)
							break;
						
						PAINTSTRUCT paintStruct = new PAINTSTRUCT();
						IntPtr screenHdc = Win32.BeginPaint(message.HWnd, ref paintStruct);
						using (Graphics screenGraphics = Graphics.FromHdc(screenHdc)) {
							
							//Draw Internal Graphics
							IntPtr hdc = internalGraphics.GetHdc();
							Message printClientMessage = Message.Create(Handle, WM_PRINTCLIENT, hdc, IntPtr.Zero);
							DefWndProc(ref printClientMessage);
							internalGraphics.ReleaseHdc(hdc);
							
							//Add the missing OnPaint() call
							Rectangle rect = Rectangle.FromLTRB(updateRect.left,
							                                    updateRect.top,
							                                    updateRect.right,
							                                    updateRect.bottom);
							
							OnPaint(new PaintEventArgs(internalGraphics, rect));
							
							//Draw Screen Graphics
							screenGraphics.DrawImage(internalBitmap, 0, 0);
						}
						//Tear down
						Win32.EndPaint(message.HWnd, ref paintStruct);
					} catch (Exception ex) {
						ErrorReporter.ShowErrorReport(ex);
					}
					break;
					
				default:
					base.WndProc(ref message);
					break;
			}
		}
		#endif
	}
}
