// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Runtime.CompilerServices;
using System.Windows.Forms;
using System.Web;
using Grunwald.Gui.Controls;

namespace Grunwald.Gui
{
	/// <summary>
	/// Description of ErrorReporter.
	/// </summary>
	public class ErrorReporter
	{
		#region Properties
		Translator translator;
		string infoText = "An error occured.";
		string reportButtonText = "Send Report";
		string ignoreButtonText = "Ignore";
		string killAppButtonText = null;
		string title = "Report error";
		
		/// <summary>
		/// Gets/Sets the translator to use.
		/// When showing the reporting dialog, all strings are translated.
		/// If this value is null, the default translator will be used.
		/// </summary>
		public Translator Translator {
			get { return translator; }
			set { translator = value; }
		}
		
		/// <summary>
		/// Gets/Sets the info text (text at the top of the dialog).
		/// If this value is null, the info text will not be shown.
		/// </summary>
		public string InfoText {
			get { return infoText; }
			set { infoText = value; }
		}
		
		/// <summary>
		/// Gets/Sets the text on the report button.
		/// If this value is null, the button will not be shown.
		/// </summary>
		public string ReportButtonText {
			get { return reportButtonText; }
			set { reportButtonText = value; }
		}
		
		/// <summary>
		/// Gets/Sets the text on the ignore button.
		/// If this value is null, the button will not be shown.
		/// </summary>
		public string IgnoreButtonText {
			get { return ignoreButtonText; }
			set { ignoreButtonText = value; }
		}
		
		/// <summary>
		/// Gets/Sets the text on the kill application button.
		/// If this value is null, the button will not be shown.
		/// </summary>
		public string KillAppButtonText {
			get { return killAppButtonText; }
			set { killAppButtonText = value; }
		}
		
		/// <summary>
		/// Gets/Sets the title of the reporting dialog.
		/// </summary>
		public string Title {
			get { return title; }
			set {
				if (value == null)
					throw new ArgumentNullException("title");
				title = value;
			}
		}
		
		bool errorDetailsVisible = true;
		bool reportTextVisible = true;
		
		/// <summary>
		/// Gets/Sets if the error details are visible.
		/// </summary>
		public bool ErrorDetailsVisible {
			get { return errorDetailsVisible; }
			set { errorDetailsVisible = value; }
		}
		
		/// <summary>
		/// Gets/Sets if the user can enter report text.
		/// </summary>
		public bool ReportTextVisible {
			get { return reportTextVisible; }
			set { reportTextVisible = value; }
		}
		#endregion
		
		#region IsCritical(Exception)
		
		/// <summary>
		/// Checks if the exception is a critical (=non handable) exception that
		/// should not be caught.
		/// </summary>
		public static bool IsCritical(Exception ex)
		{
			if (ex is OutOfMemoryException)
				return true;
			if (ex is AppDomainUnloadedException)
				return true;
			if (ex is BadImageFormatException)
				return true;
			if (ex is CannotUnloadAppDomainException)
				return true;
			if (ex is ExecutionEngineException)
				return true;
			if (ex is InvalidProgramException)
				return true;
			return false;
		}
		
		#endregion
		
		ErrorReporterForm form;
		
		#region ReportError
		Exception exception;
		
		/// <summary>
		/// Reports the specified error.
		/// </summary>
		public virtual void ReportError(Exception ex)
		{
			try {
				Console.WriteLine(ex.ToString());
				
				lock (this) {
					if (form != null) return;
					form = new ErrorReporterForm();
				}
				
				exception = ex;
				Translator t = translator;
				if (t == null) t = Translator.DefaultTranslator;
				form.Text = t.Translate(title);
				if (this.infoText == null)
					form.ErrorDescriptionVisible = false;
				else
					form.ErrorDescriptionText = t.Translate(this.infoText);
				form.ErrorDetailsText = ex.ToString();
				form.ReportTextVisible = this.ReportTextVisible;
				Button btn = null;
				
				if (this.killAppButtonText != null) {
					btn = new AutoSizedButton();
					btn.Text = t.Translate(this.killAppButtonText);
					btn.Click += new EventHandler(KillAppClick);
					form.Buttons.Add(btn);
				}
				
				if (this.reportButtonText != null) {
					btn = new AutoSizedButton();
					btn.Text = t.Translate(this.reportButtonText);
					btn.Click += new EventHandler(SendReportClick);
					form.Buttons.Add(btn);
				}
				
				if (this.ignoreButtonText != null) {
					btn = new AutoSizedButton();
					btn.Text = t.Translate(this.ignoreButtonText);
					btn.Click += new EventHandler(IgnoreClick);
					form.Buttons.Add(btn);
				}
				
				if (btn != null)
					btn.TabIndex = 0;
				for(int i = 0; i < form.Buttons.Count - 1; i++) {
					Control ctl = form.Buttons[i];
					ctl.TabIndex = i + 1;
				}
				
				form.ShowDialog();
				form = null;
			} catch (Exception except) {
				MessageBox.Show(except.ToString());
			}
		}
		#endregion
		
		#region Button Clicks
		private void IgnoreClick(object sender, EventArgs e)
		{
			form.Close();
		}
		
		private void KillAppClick(object sender, EventArgs e)
		{
			Application.Exit();
		}
		
		private void SendReportClick(object sender, EventArgs e)
		{
			form.Enabled = false;
			try {
				OnSendReport(new ReportEventArgs(exception, form.ReportText));
			} catch (Exception ex) {
				Console.WriteLine(ex);
				MessageBox.Show(ex.ToString());
			} finally {
				form.Close();
			}
		}
		#endregion
		
		#region SendReport-Event
		/// <summary>
		/// Raises the SendReport-event.
		/// </summary>
		protected virtual void OnSendReport(ReportEventArgs e)
		{
			if (SendReport != null) {
				SendReport(this, e);
			}
		}
		
		/// <summary>
		/// Occurs when a error report should be sent.
		/// This event is <b>not</b> called in the thread of the error reporter, so
		/// blocking this event does not block the dialog.
		/// (though the dialog is enabled = false while this event is executed)
		/// </summary>
		public event ReportEventHandler SendReport;
		#endregion
		
		#region Default Error Reporter
		private static ErrorReporter defaultErrorReporter;
		
		/// <summary>
		/// Gets/Sets the default error reporter.
		/// </summary>
		public static ErrorReporter DefaultErrorReporter {
			get {
				if (defaultErrorReporter == null)
					ErrorReporter.DefaultErrorReporter = null;
				return defaultErrorReporter;
			}
			set {
				if (value == null) {
					defaultErrorReporter = new ErrorReporter();
					defaultErrorReporter.ReportTextVisible = false;
					defaultErrorReporter.ReportButtonText = null;
					defaultErrorReporter.KillAppButtonText = "Exit application";
				} else {
					defaultErrorReporter = value;
				}
			}
		}
		
		/// <summary>
		/// Shows the default error reporter.
		/// </summary>
		public static void ShowErrorReport(Exception ex)
		{
			try {
				DefaultErrorReporter.ReportError(ex);
			} catch (Exception e) {
				Console.WriteLine(e);
				Console.WriteLine(ex);
				MessageBox.Show(e.ToString());
				MessageBox.Show(ex.ToString());
			}
		}
		#endregion
	}
	
	#region SendReport-Event Delegate + EventArgs definition
	/// <summary>
	/// EventHandler for the SendReport-event.
	/// </summary>
	public delegate void ReportEventHandler(object sender, ReportEventArgs e);
	
	/// <summary>
	/// EventArgs for the SendReport-event.
	/// </summary>
	public class ReportEventArgs : EventArgs
	{
		Exception exception;
		string userText;
		
		/// <summary>
		/// Gets the exception that was thrown.
		/// </summary>
		public Exception Exception {
			get { return exception; }
		}
		
		/// <summary>
		/// Gets the text the user has entered.
		/// </summary>
		public string UserText {
			get { return userText; }
		}
		
		/// <summary>
		/// Creates a new ReportEventArgs-object.
		/// </summary>
		/// <param name="exception">The exception that was thrown.</param>
		/// <param name="userText">The text the user has entered.</param>
		public ReportEventArgs(Exception exception, string userText)
		{
			this.exception = exception;
			this.userText = userText;
		}
	}
	#endregion
}
