// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;
using System.Xml;
using Grunwald.Gui.Settings.Editors;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// The ChooseOptionSetting is used for multiple choice options, where the options are dynamic
	/// so that an enum cannot be used.
	/// </summary>
	public class ChooseOptionSetting : DefaultSetting
	{
		/// <summary>
		/// Creates a new ChooseOptionSetting object using name, displayedName and description
		/// from the <see cref="SettingAttribute"/>.
		/// </summary>
		/// <param name="attribute">The <see cref="SettingAttribute"/> used to get name, displayedName and description.</param>
		public ChooseOptionSetting(SettingAttribute attribute)
			: base(attribute, null)
		{
		}
		
		/// <summary>
		/// This constructor is needed for using the ChooseOptionSetting as a field in a class.
		/// <code>[Setting(...)]
		/// ChooseOptionSetting optionSetting;</code>
		/// </summary>
		public ChooseOptionSetting(SettingAttribute attribute, FieldInfo field)
			: base(attribute, null)
		{
		}
		
		/// <summary>
		/// Creates a new ChooseOptionSetting object using the specified properties.
		/// </summary>
		/// <param name="name">The name of the setting (used for storing the setting).</param>
		/// <param name="displayName">The displayed name of this setting. This string is passed to the <see cref="Translator"/>.</param>
		/// <param name="description">The description of the setting. This string is passed to the <see cref="Translator"/>.</param>
		public ChooseOptionSetting(string name, string displayName, string description)
			: base(name, displayName, description, null)
		{
		}
		
		string[] texts;
		string[] options;
		
		/// <summary>
		/// Gets/Sets the array of option value strings to use.
		/// </summary>
		public string[] Options {
			get {
				return options;
			}
			set {
				options = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the array of option display texts.
		/// </summary>
		public string[] Texts {
			get {
				return texts;
			}
			set {
				texts = value;
			}
		}
		
		string currentOption;
		
		/// <summary>
		/// Gets/Sets the value string of the currently selected option.
		/// </summary>
		public override object Value {
			get {
				return currentOption;
			}
			set {
				string oldValue = currentOption;
				string newValue = (string)value;
				currentOption = newValue;
				if (oldValue != newValue)
					this.OnValueChanged(new SettingChangedEventArgs(this, oldValue, newValue));
				if (newValue != (string)this.InEditValue)
					this.InEditValue = newValue;
			}
		}
		
		/// <summary>
		/// Saves the setting's value into the specified xml writer.
		/// </summary>
		protected override void SaveValue(XmlWriter writer)
		{
			object value = this.Value;
			new ValueWriter().Write(value, writer);
		}
		
		/// <summary>
		/// Restores the setting page from the specified xml reader.
		/// </summary>
		public override void Load(XmlReader reader)
		{
			object value = new ValueWriter().Read(reader);
			try {
				CheckValue(value);
			} catch (SettingEditException ex) {
				throw new SettingLoadException("The saved value '" + value.ToString() + "' is invalid for the setting " + this.Name, ex);
			}
			this.Value = value;
		}
		
		/// <summary>
		/// Creates a new ISettingEditor to edit this setting.
		/// </summary>
		public override ISettingEditor CreateEditor(EditorFactory factory)
		{
			return new ChooseOptionEditor(options, texts, this);
		}
	}
}
