// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// This is the default implementation of <see cref="ISetting"/>.
	/// It checks for valid names etc., but the access to the value itself is abstract.
	/// You can extend this class and easily create your own settings.
	/// </summary>
	public abstract class DefaultSetting : ISetting
	{
		string name;
		string displayName;
		string description;
		
		#region Constructor
		/// <summary>
		/// Creates a new DefaultSetting object using name, displayedName and description
		/// from the <see cref="SettingAttribute"/> and the specified start value.
		/// </summary>
		/// <param name="attribute">The <see cref="SettingAttribute"/> used to get name, displayedName and description.</param>
		/// <param name="inEditValue">The start value of the <see cref="InEditValue"/> property.</param>
		protected DefaultSetting(SettingAttribute attribute, object inEditValue)
		{
			this.name = attribute.Name;
			this.displayName = attribute.DisplayName;
			this.description = attribute.Description;
			this.inEditValue = inEditValue;
			if (displayName == null)
				displayName = name;
			if (description == null)
				description = "";
		}
		
		/// <summary>
		/// Creates a new DefaultSetting object using the specified properties.
		/// </summary>
		/// <param name="name">The name of the setting (used for storing the setting).</param>
		/// <param name="displayName">The displayed name of this setting. This string is passed to the <see cref="Translator"/>.</param>
		/// <param name="description">The description of the setting. This string is passed to the <see cref="Translator"/>.</param>
		/// <param name="inEditValue">The start value of the <see cref="InEditValue"/> property.</param>
		protected DefaultSetting(string name, string displayName, string description, object inEditValue)
		{
			CheckName(name);
			if (displayName == null)
				displayName = name;
			if (description == null)
				description = "";
			this.name = name;
			this.displayName = displayName;
			this.description = description;
			this.inEditValue = inEditValue;
		}
		
		internal static void CheckName(string name)
		{
			if (name == null) throw new ArgumentNullException("name");
			for (int i = 1; i < name.Length; ++i) {
				if (!Char.IsLetterOrDigit(name[i]))
					throw new ArgumentException("name contains invalid character '" + name[i] + "'");
			}
		}
		#endregion
		
		#region Getter
		/// <summary>
		/// The name of the setting (used for storing the setting).
		/// </summary>
		public string Name {
			get { return name; }
		}
		
		/// <summary>
		/// The displayed name of this setting. This string is passed to the <see cref="Translator"/>.
		/// </summary>
		public string DisplayName {
			get { return displayName; }
		}
		
		/// <summary>
		/// The description of the setting. This string is passed to the <see cref="Translator"/>.
		/// </summary>
		public string Description {
			get { return description; }
		}
		
		/// <summary>
		/// Creates a new ISettingEditor to edit this setting.
		/// </summary>
		public abstract ISettingEditor CreateEditor(EditorFactory factory);
		
		/// <summary>Gets the string representation of this object.</summary>
		public override string ToString()
		{
			return base.ToString() + ": name = " + name;
		}
		#endregion
		
		#region Value + events
		/// <summary>
		/// Gets/Sets the active value of this setting.
		/// </summary>
		public abstract object Value { get; set; }
		
		object inEditValue;
		
		/// <summary>
		/// Gets/Sets the value in the edit dialog. This value gets active when the user
		/// clicks on 'Apply'.
		/// </summary>
		/// <exception cref="SettingEditException">Occurs if this property is set to
		/// an invalid value.</exception>
		public object InEditValue {
			get { return inEditValue; }
			set {
				if (inEditValue == value) return;
				if (inEditValue != null && inEditValue.Equals(value)) return;
				CheckValue(value);
				object oldValue = inEditValue;
				inEditValue = value;
				OnInEditValueChanged(new SettingChangedEventArgs(this, oldValue, value));
			}
		}
		
		/// <summary>
		/// Checks if the value if acceptable for this setting.
		/// If the value is not acceptable, a <see cref="SettingEditException"/> is thrown.
		/// </summary>
		public virtual void CheckValue(object value)
		{
			// override this to check the value.
		}
		
		/// <summary>
		/// This event occurs whenever the value of <see cref="Value"/> is changed.
		/// </summary>
		public event SettingChangedEventHandler ValueChanged;
		
		/// <summary>
		/// This event occurs whenever the value of <see cref="InEditValue"/> is changed.
		/// </summary>
		public event SettingChangedEventHandler InEditValueChanged;
		
		/// <summary>
		/// Raises the <see cref="ValueChanged"/> event.
		/// </summary>
		protected virtual void OnValueChanged(SettingChangedEventArgs e)
		{
			if (ValueChanged != null) {
				ValueChanged(this, e);
			}
		}
		
		/// <summary>
		/// Raises the <see cref="InEditValueChanged"/> event.
		/// </summary>
		protected virtual void OnInEditValueChanged(SettingChangedEventArgs e)
		{
			if (InEditValueChanged != null) {
				InEditValueChanged(this, e);
			}
		}
		#endregion
		
		/// <summary>
		/// Saves the setting into the specified xml writer.
		/// </summary>
		public virtual void Save(XmlWriter writer)
		{
			writer.WriteStartElement("setting");
			writer.WriteAttributeString("name", name);
			SaveValue(writer);
			writer.WriteEndElement();
		}
		
		/// <summary>
		/// Saves the setting's value into the specified xml writer.
		/// </summary>
		protected abstract void SaveValue(XmlWriter writer);
		
		/// <summary>
		/// Restores the setting page from the specified xml reader.
		/// </summary>
		public abstract void Load(XmlReader reader);
	}
}
