// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Reflection;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// Default implementation of <see cref="ISettingPage"/>
	/// </summary>
	public abstract class DefaultSettingPage : ISettingPage
	{
		#region Constructor
		/// <summary>
		/// Creates new DefaultSettingPage.
		/// </summary>
		protected DefaultSettingPage() {}
		
		/// <summary>
		/// Creates a copy of the specified DefaultSettingPage.
		/// </summary>
		protected DefaultSettingPage(DefaultSettingPage page)
		{
			this.allowAdd    = page.allowAdd;
			this.allowDelete = page.allowDelete;
			this.allowMove   = page.allowMove;
			this.title       = page.title;
			this.image       = page.image;
			foreach (ISettingPage subpage in page.subPages) {
				this.subPages.Add(subpage.Clone());
			}
		}
		#endregion
		
		#region Settings
		/// <summary>
		/// Gets the settings of this object as array.
		/// </summary>
		public abstract ISetting[] Settings { get; }
		#endregion
		
		#region Allow-Properties
		bool allowAdd = false;
		
		/// <summary>
		/// Gets/Sets if it is allowed to add new sub pages to this page.
		/// </summary>
		public bool AllowAddNew {
			get { return allowAdd; }
			set { allowAdd = value; }
		}
		
		bool allowDelete = false;
		
		/// <summary>
		/// Gets/Sets if it is allowed to delete sub pages from this page.
		/// </summary>
		public bool AllowDelete {
			get { return allowDelete; }
			set { allowDelete = value; }
		}
		
		bool allowMove = false;
		
		/// <summary>
		/// Gets/Sets if it is allowed to rearrange the sub pages of this page.
		/// </summary>
		public bool AllowMove {
			get { return allowMove; }
			set { allowAdd = value; }
		}
		
		/// <summary>
		/// This whole setting system is crap and crying out: I NEED REFACTORING!
		/// Needed as workaround for DGNetwork's ExtensibleSettingPage.
		/// Returns true if not overridden.
		/// </summary>
		public virtual bool CanLoadWithOtherTitle {
			get {
				return true;
			}
		}
		#endregion
		
		#region Sub Pages
		ArrayList subPages = new ArrayList();
		
		readonly static ISettingPage[] emptySubPages = new ISettingPage[0];
		
		/// <summary>
		/// Get the sub pages of this page.
		/// </summary>
		public ISettingPage[] SubPages {
			get {
				if (subPages.Count == 0)
					return emptySubPages;
				return (ISettingPage[])subPages.ToArray(typeof(ISettingPage));
			}
		}
		
		/// <summary>
		/// Occurs whenever the list of sub pages is changed.
		/// </summary>
		public event EventHandler SubPagesChanged;
		
		/// <summary>
		/// Raises the SubPagesChanged event.
		/// </summary>
		protected virtual void OnSubPagesChanged(EventArgs e)
		{
			if (SubPagesChanged != null) {
				SubPagesChanged(this, e);
			}
		}
		
		/// <summary>
		/// Gets the first sub page with the specified title.
		/// If the sub page is not found, this method returns null.
		/// </summary>
		public ISettingPage GetPage(string title)
		{
			foreach (ISettingPage page in subPages) {
				if (page.Title == title)
					return page;
			}
			return null;
		}
		#endregion
		
		#region Moving
		/// <summary>
		/// Gets if rearranging the specified page in general is allowed.
		/// </summary>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		public virtual bool CanMovePage(ISettingPage page)
		{
			if (page == null)
				throw new ArgumentNullException();
			if (!subPages.Contains(page))
				throw new ArgumentException("The page is not a sub page of this setting page", "page");
			return allowMove;
		}
		
		/// <summary>
		/// Moves the specified page to the specified target index and raises the
		/// OnSubPagesChanged event.
		/// </summary>
		/// <exception cref="NotSupportedException">The move operation is not allowed.</exception>
		/// <exception cref="ArgumentOutOfRangeException">The index is out of range.</exception>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		public virtual void MovePage(ISettingPage page, int index)
		{
			if (!CanMovePage(page, index))
				throw new NotSupportedException("The operation is not supported by this SettingPage");
			int oldIndex = subPages.IndexOf(page);
			if (oldIndex == index) return;
			subPages.RemoveAt(oldIndex);
			if (oldIndex < index) --index;
			subPages.Insert(index, page);
			OnSubPagesChanged(EventArgs.Empty);
		}
		
		/// <summary>
		/// Returns if moving the specified page to the specified target index is possible.
		/// </summary>
		/// <exception cref="ArgumentOutOfRangeException">The index is out of range.</exception>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		public virtual bool CanMovePage(ISettingPage page, int index)
		{
			if (index < 0 || index >= subPages.Count)
				throw new ArgumentOutOfRangeException("index", index, "0 <= index < " + subPages.Count.ToString());
			return CanMovePage(page);
		}
		#endregion
		
		#region Add / Delete
		/// <summary>
		/// Gets if adding new sub-pages is allowed.
		/// </summary>
		bool ISettingPage.CanAddNewPage {
			get { return allowAdd; }
		}
		
		/// <summary>
		/// Creates a new sub page and adds it to this page.
		/// </summary>
		/// <exception cref="NotSupportedException">The add operation is not allowed.</exception>
		public virtual void AddPage()
		{
			if (allowAdd)
				throw new NotImplementedException("The class " + this.GetType().FullName + " failed to override AddPage or made a call to base.AddPage()!");
			else
				throw new NotSupportedException("The operation is not supported by this SettingPage");
		}
		
		/// <summary>
		/// Gets if adding the specified page is allowed.
		/// </summary>
		public virtual bool CanAddPage(ISettingPage page, int index)
		{
			if (page == null)
				throw new ArgumentNullException("page");
			return false;
		}
		
		/// <summary>
		/// Adds the specified page to the end of the list of sub pages
		/// by calling AddPage(ISettingPage, int)
		/// </summary>
		public void AddPage(ISettingPage page)
		{
			AddPage(page, subPages.Count);
		}
		
		/// <summary>
		/// Adds the specified page to the list of sub pages and raises the
		/// OnSubPagesChanged event.
		/// </summary>
		public virtual void AddPage(ISettingPage page, int index)
		{
			if (subPages.Contains(page))
				throw new ArgumentException("The page is already added to this setting page!");
			subPages.Insert(index, page);
			OnSubPagesChanged(EventArgs.Empty);
		}
		
		/// <summary>
		/// Gets if deleting the specified page is allowed.
		/// </summary>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		public virtual bool CanDeletePage(ISettingPage page)
		{
			if (page == null)
				throw new ArgumentNullException("page");
			if (!subPages.Contains(page))
				throw new ArgumentException("The page is not a sub page of this setting page", "page");
			if (!page.DynamicPage)
				return false;
			return allowDelete;
		}
		
		/// <summary>
		/// Deletes the specified page from the list of sub pages and
		/// raises the OnSubPagesChanged event.
		/// </summary>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		/// <exception cref="NotSupportedException">The delete operation is not allowed.</exception>
		public void DeletePage(ISettingPage page)
		{
			if (!CanDeletePage(page))
				throw new NotSupportedException("The operation is not supported by this SettingPage");
			subPages.Remove(page);
			OnSubPagesChanged(EventArgs.Empty);
		}
		#endregion
		
		#region Title
		string title;
		
		/// <summary>
		/// Gets/Sets the title of this setting page.
		/// </summary>
		public string Title {
			get { return title; }
			set {
				if (title == value)
					return;
				title = value;
				OnTitleChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Occurs whenever the title property is changed.
		/// </summary>
		public event EventHandler TitleChanged;
		
		/// <summary>
		/// Raises the <see cref="TitleChanged">TitleChanged event</see>.
		/// </summary>
		protected virtual void OnTitleChanged(EventArgs e) {
			if (TitleChanged != null) {
				TitleChanged(this, e);
			}
		}
		#endregion
		
		#region Image
		System.Drawing.Image image;
		
		/// <summary>
		/// Gets/Sets the image associated with this setting page.
		/// </summary>
		public System.Drawing.Image Image {
			get { return image; }
			set {
				if (image == value)
					return;
				image = value;
				OnImageChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Occurs whenever the image property is changed.
		/// </summary>
		public event EventHandler ImageChanged;
		
		/// <summary>
		/// Raises the <see cref="ImageChanged">ImageChanged event</see>.
		/// </summary>
		protected virtual void OnImageChanged(EventArgs e) {
			if (ImageChanged != null) {
				ImageChanged(this, e);
			}
		}
		#endregion
		
		#region Save/Load
		/// <summary>
		/// Saves the setting page into the specified xml writer.<br/>
		/// If you only want to add additional elements to the setting page,
		/// override <see cref="SaveContent"/> instead!
		/// </summary>
		public virtual void Save(XmlWriter writer)
		{
			writer.WriteStartElement("page");
			if (this.Title != null)
				writer.WriteAttributeString("title", this.Title);
			writer.WriteAttributeString("dynamic", XmlConvert.ToString(this.DynamicPage));
			writer.WriteAttributeString("type", this.GetType().FullName);
			this.SaveContent(writer);
			writer.WriteEndElement();
		}
		
		/// <summary>
		/// Saves the content of the setting page (settings and sub pages)
		/// </summary>
		protected virtual void SaveContent(XmlWriter writer)
		{
			foreach (ISetting setting in this.Settings) {
				setting.Save(writer);
			}
			if (this.SubPages != null) {
				foreach (ISettingPage page in this.SubPages) {
					page.Save(writer);
				}
			}
		}
		
		/// <summary>
		/// Restores the setting page from the specified xml reader.<br/>
		/// If you only want to load additional elements from the setting page,
		/// override <see cref="LoadContentElement"/> instead!
		/// </summary>
		public virtual void Load(XmlReader reader)
		{
			if (reader.Name != "page")
				throw new SettingLoadException("Reader is not positioned on a page!");
			string dynamic = reader.GetAttribute("dynamic");
			if (dynamic == null)
				throw new SettingLoadException("<page> must contain 'dynamic' attribute!");
			this.dynamicPage = XmlConvert.ToBoolean(dynamic);
			this.Title = reader.GetAttribute("title");
			if (this.SubPages != null) {
				for (int i = 0; i < subPages.Count; i++) {
					ISettingPage page = (ISettingPage)subPages[i];
					if (page.DynamicPage)
						subPages.RemoveAt(i--);
				}
			}
			if (reader.IsEmptyElement) {
				reader.Read();
				return;
			}
			ArrayList loadedElements = new ArrayList();
			reader.Read();
			while (true) {
				if (reader.NodeType == XmlNodeType.EndElement) {
					if (reader.Name != "page")
						throw new SettingLoadException("Invalid end element </" + reader.Name + ">");
					reader.ReadEndElement();
					break;
				}
				if (reader.NodeType != XmlNodeType.Element) {
					throw new SettingLoadException("Invalid node type: " + reader.NodeType.ToString() + " in <page> body.");
				}
				if (!LoadContentElement(reader, loadedElements))
					throw new SettingLoadException("Invalid element <" + reader.Name + ">");
			}
		}
		
		/// <summary>
		/// Loads the content element the reader is currently positioned on.<br/>
		/// The array list <paramref name="loadedElements"/> contains a list of all
		/// elements loaded so far, use it to prevemt double loading and add your new
		/// elements to it!
		/// </summary>
		/// <returns>
		/// Returns <c>true</c> if the specified element was loaded, <c>false</c> if the
		/// element was an unknown element.<br/>
		/// </returns>
		protected virtual bool LoadContentElement(XmlReader reader, ArrayList loadedElements)
		{
			switch (reader.Name) {
				case "page":
					LoadSubPage(reader, loadedElements);
					return true;
				case "setting":
					LoadSetting(reader, loadedElements);
					return true;
				default:
					return false;
			}
		}
		
		/// <summary>
		/// Loads a setting from the XmlReader. This method is used by Load(XmlReader).
		/// </summary>
		private void LoadSetting(XmlReader reader, ArrayList loadedElements)
		{
			//System.Windows.Forms.MessageBox.Show("Loading setting " + reader.GetAttribute("name") + " on " + Title);
			string name = reader.GetAttribute("name");
			foreach (ISetting setting in this.Settings) {
				if (setting.Name == name) {
					setting.Load(reader);
					loadedElements.Add(setting);
					return;
				}
			}
			reader.Skip();
		}
		
		/// <summary>
		/// Loads a sub page from the XmlReader. This method is used by Load(XmlReader).
		/// </summary>
		private void LoadSubPage(XmlReader reader, ArrayList loadedPages)
		{
			if (subPages == null)
				throw new SettingLoadException("This page cannot have subpages, but the xml file specified subpages!");
			string title = reader.GetAttribute("title");
			string type = reader.GetAttribute("type");
			if (type == null)
				throw new SettingLoadException("<page> must contain 'type' attribute!");
			string dynamicString = reader.GetAttribute("dynamic");
			if (dynamicString == null)
				throw new SettingLoadException("<page> must contain 'dynamic' attribute!");
			bool dynamic = XmlConvert.ToBoolean(dynamicString);
			// try to find the page
			ISettingPage page = FindPage(title, type, dynamic, loadedPages);
			if (page == null && dynamic) {
				page = CreatePage(type);
				subPages.Add(page);
			}
			loadedPages.Add(page);
			if (page == null) {
				// ignore this old static page
				reader.Skip();
			} else {
				page.Load(reader);
			}
		}
		
		
		/// <summary>
		/// This method is used by Load(XmlReader) and must be overridden if this page
		/// can have dynamic sub pages.
		/// </summary>
		protected virtual ISettingPage CreatePage(string type)
		{
			throw new NotImplementedException("CreatePage must be overridden to allow creation of dynamic pages!");
		}
		
		private ISettingPage FindPage(string title, string type, bool dynamic, ArrayList loadedPages)
		{
			foreach(ISettingPage page in subPages) {
				if (page.Title == title && page.DynamicPage == dynamic) {
					if (loadedPages.Contains(page)) continue;
					if (page.GetType().FullName == type) {
						return page;
					}
				}
			}
			foreach(ISettingPage page in subPages) {
				if (page.DynamicPage == dynamic) {
					if (loadedPages.Contains(page)) continue;
					if (page.GetType().FullName == type && page.CanLoadWithOtherTitle) {
						return page;
					}
				}
			}
			return null;
		}
		
		/// <summary>
		/// Saves the InEditValues in this and all subpages to the Values.
		/// </summary>
		public void Save()
		{
			foreach(ISettingPage subPage in subPages) {
				subPage.Save();
			}
			foreach(ISetting setting in this.Settings) {
				setting.Value = setting.InEditValue;
			}
		}
		
		/// <summary>
		/// Loads all InEditValues in this and all subpages from the Values.
		/// </summary>
		public void Load()
		{
			foreach(ISettingPage subPage in subPages) {
				subPage.Load();
			}
			foreach(ISetting setting in this.Settings) {
				setting.InEditValue = setting.Value;
			}
		}
		#endregion
		
		#region Event SettingChanged
		/// <summary>
		/// Occurs when a setting on this setting page is changed.
		/// </summary>
		public event SettingChangedEventHandler SettingChanged;
		
		/// <summary>
		/// Raises the <see cref="SettingChanged">SettingChanged event</see>.
		/// </summary>
		/// <remarks>
		/// The class deriving from DefaultSettingPage is responsible for calling this
		/// method when a setting has changed!
		/// </remarks>
		protected virtual void OnSettingChanged(SettingChangedEventArgs e) {
			if (SettingChanged != null) {
				SettingChanged(this, e);
			}
		}
		#endregion
		
		bool dynamicPage = true;
		
		/// <summary>
		/// Gets if this page is dynamic (return value <b>true</b>) or static (return value
		/// <c>false</c>).<br/>
		/// This influences how this page is loaded from XML:<br/>
		/// A dynamic page is recreated if it didn't exist before load or is dropped if
		/// it don't exist in the XML data, but a static can't get recreated or dropped.
		/// </summary>
		public bool DynamicPage {
			get { return dynamicPage; }
			set { dynamicPage = value; }
		}
		
		/// <summary>
		/// Gets a copy of this SettingPage.
		/// </summary>
		public abstract object Clone();
	}
}
