// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using Grunwald.Gui.Settings.Editors;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// The EditorFactory is used to get default editors for settings that don't have their
	/// own editor defined.<br/>
	/// You can extend this class to add more default editors.
	/// </summary>
	public class EditorFactory
	{
		/// <summary>
		/// Gets the editor for the specified <see cref="Type"/>.
		/// </summary>
		public virtual ISettingEditor GetEditor(ISetting setting, Type type)
		{
			if (type == typeof(string)) {
				if (setting.Name.StartsWith("file")) {
					return new FileNameEditor(setting);
				} else if (setting.Name.EndsWith("Path")) {
					return new DirectoryEditor(setting);
				} else {
					StringEditor editor = new StringEditor(setting);
					if (setting.Name == "Password")
						editor.TextBox.PasswordChar = '*';
					return editor;
				}
			}
			if (type == typeof(int))
				return new IntEditor(setting);
			if (type == typeof(bool))
				return new BoolEditor(setting);
			if (type.BaseType == typeof(Enum))
				return new EnumEditor(type, setting);
			throw new ArgumentException("The EditorFactory could not find a default editor for the type '" +
			                            type.FullName + "'!");
		}
		
		static EditorFactory defaultFactory = new EditorFactory();
		
		/// <summary>
		/// Gets/Sets the default translator to use.
		/// </summary>
		public static EditorFactory DefaultFactory {
			get { return defaultFactory; }
			set {
				if (value == null)
					defaultFactory = new EditorFactory();
				else
					defaultFactory = value;
			}
		}
	}
}
