// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;
using System.Xml;
using Grunwald.Gui.Controls;
using Grunwald.Gui.Settings;
using Grunwald.Gui.Settings.Editors;

namespace Grunwald.Gui.Settings.Editors
{
	/// <summary>
	/// Setting editor that can be used to display a button.
	/// Useful in combination with <see cref="FixedEditorSetting"/> or <see cref="SpecialEditorSetting"/>.
	/// </summary>
	public class ButtonEditor : SingleControlEditor
	{
		AutoSizedButton btn;
		
		/// <summary>
		/// Creates a new ButtonEditor for the specified setting.
		/// </summary>
		public ButtonEditor(ISetting setting) : base(setting, new AutoSizedButton())
		{
			btn = (AutoSizedButton)Control;
			text = setting.DisplayName;
		}
		
		Translator translator;
		
		string text;
		
		/// <summary>
		/// Gets/Sets the text of the button. By default, the DisplayName of the setting is used.
		/// </summary>
		public string ButtonText {
			get {
				return text;
			}
			set {
				text = value;
				if (translator != null)
					btn.Text = translator.Translate(value);
			}
		}
		
		bool useFullWidth = false;
		
		/// <summary>
		/// Gets/Sets if the button should use the full with available.
		/// </summary>
		public bool UseFullWidth {
			get {
				return useFullWidth;
			}
			set {
				useFullWidth = value;
				btn.AutoSize = !value;
			}
		}
		
		/// <summary>
		/// This event is fired when the button was clicked.
		/// </summary>
		public event EventHandler ButtonClick {
			add    { btn.Click += value; }
			remove { btn.Click -= value; }
		}
		
		/// <summary>
		/// Do nothing because a button has no value.
		/// </summary>
		protected override void LoadEditValue() { }
		
		/// <summary>
		/// Creates the controls needed for editing the <see cref="ISetting"/> and
		/// adds them to the specified parent control.
		/// </summary>
		public override void CreateControls(Rectangle bounds, EditorProperties properties)
		{
			translator = properties.Translator;
			btn.Text = translator.Translate(text);
			if (useFullWidth)
				bounds.Height = btn.Height;
			else
				bounds.Size = btn.Size;
			base.CreateControls(bounds, properties);
			if (!useFullWidth)
				btn.Anchor = AnchorStyles.Top | AnchorStyles.Left;
		}
		
		/// <summary>
		/// Gets the label mode of the button.
		/// This method always returns <see cref="LabelMode">LabelMode.NoLabel</see>.
		/// </summary>
		public override LabelMode LabelMode {
			get {
				return LabelMode.NoLabel;
			}
		}
	}
}
