// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;

namespace Grunwald.Gui.Settings.Editors
{
	/// <summary>
	/// Description of DefaultEditor.
	/// </summary>
	public abstract class DefaultEditor : ISettingEditor
	{
		ISetting setting;
		
		/// <summary>
		/// Creates a new DefaultEditor
		/// </summary>
		protected DefaultEditor(ISetting setting)
		{
			if (setting == null)
				throw new ArgumentNullException("setting");
			this.setting = setting;
			setting.InEditValueChanged += new SettingChangedEventHandler(SettingInEditValueChanged);
		}
		
		private void SettingInEditValueChanged(object sender, SettingChangedEventArgs e)
		{
			InEditValueChanged();
		}
		
		/// <summary>
		/// Gets called when the edit value is changed and should be loaded.
		/// If not overridden, this calls <see cref="LoadEditValue"/>.
		/// </summary>
		protected virtual void InEditValueChanged()
		{
			LoadEditValue();
		}
		
		/// <summary>
		/// Gets called when the edit value is changed and should be loaded.
		/// </summary>
		protected abstract void LoadEditValue();
		
		/// <summary>
		/// Gets the <see cref="ISetting">Setting</see> this Editor was created for.
		/// </summary>
		public ISetting Setting {
			get { return setting; }
		}
		
		/// <summary>
		/// Gets the size the edit field needs.
		/// </summary>
		public abstract Size GetSize();
		
		/// <summary>
		/// Gets if the edit field can be bigger than the minimal size.
		/// </summary>
		public virtual bool CanBeBigger {
			get { return false; }
		}
		
		/// <summary>
		/// Gets/Sets whether the editor is enabled.
		/// </summary>
		public abstract bool Enabled { get; set; }
		
		/// <summary>
		/// Gets/Sets the anchorstyle of the edit control.
		/// </summary>
		public abstract AnchorStyles Anchor { get; set; }
		
		/// <summary>
		/// Closes the editor.
		/// </summary>
		public virtual void Close()
		{
			setting.InEditValueChanged -= new SettingChangedEventHandler(SettingInEditValueChanged);
		}
		
		/// <summary>
		/// Creates the controls needed for editing the <see cref="Setting"/> and
		/// adds them to the specified parent control.
		/// </summary>
		public abstract void CreateControls(Rectangle bounds, EditorProperties properties);
		
		/// <summary>
		/// Occurs if the editing control is "activated" by getting focus or moving the mouse over it.
		/// </summary>
		public event EventHandler Activated;
		
		/// <summary>
		/// Raises the <see cref="Activated"/> event.
		/// </summary>
		protected virtual void OnActivated(EventArgs e)
		{
			if (Activated != null) {
				Activated(this, e);
			}
		}
		
		/// <summary>
		/// Activates this editor (sets the focus to the input field).
		/// </summary>
		public abstract void Activate();
		
		/// <summary>
		/// Gets the <see cref="LabelMode"/> this setting editor uses.<br/>
		/// If not overridden, this property always returns
		/// <see cref="LabelMode">LabelMode.SameRow</see>.
		/// </summary>
		public virtual LabelMode LabelMode {
			get { return LabelMode.SameRow; }
		}
	}
}
