// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;

namespace Grunwald.Gui.Settings.Editors
{
	/// <summary>
	/// The enum editor can edit enumeration values.
	/// </summary>
	public class EnumEditor : SingleControlEditor
	{
		/// <summary>
		/// Creates a new EnumEditor.
		/// </summary>
		public EnumEditor(Type type, ISetting setting) : base(setting, MakeControl(setting))
		{
			control = base.Control;
			int pos = 0;
			int x = 4;
			if (control is GroupBox) {
				pos += 12;
				x += 4;
			}
			foreach(FieldInfo info in type.GetFields()) {
				if (!info.IsStatic)
					continue;
				RadioButton radio = new RadioButton();
				radio.FlatStyle = FlatStyle.System;
				object[] descriptions = info.GetCustomAttributes(typeof(DescriptionAttribute), false);
				if (descriptions != null && descriptions.Length == 1)
					radio.Text = ((DescriptionAttribute)descriptions[0]).Description;
				else
					radio.Text = info.Name;
				radio.Tag = info.GetValue(null);
				radio.Location = new Point(x, pos);
				radio.Width = 60 - 2 * x;
				radio.CheckedChanged += new EventHandler(RadioCheckedChanged);
				control.Controls.Add(radio);
				this.RegisterForOnActivated(radio);
				pos += radio.Height - 2;
			}
			if (control is GroupBox)
				pos += 4;
			control.Size = new Size(60, pos);
		}
		
		Control control;
		
		private static Control MakeControl(ISetting setting)
		{
			if (setting.DisplayName == null || setting.DisplayName == "")
				return new Panel();
			else
				return new GroupBox();
		}
		
		/// <summary>
		/// Creates the controls needed for editing the <see cref="ISetting"/> and
		/// adds them to the specified parent control.
		/// </summary>
		public override void CreateControls(Rectangle bounds, EditorProperties properties)
		{
			control.Text = properties.Translator.Translate(Setting.DisplayName);
			foreach(Control ctl in control.Controls) {
				ctl.Anchor = AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right;
				ctl.Text = properties.Translator.Translate(ctl.Text);
			}
			base.CreateControls(bounds, properties);
		}
		
		bool working = false;
		
		/// <summary>
		/// Loads the edit value into the editor.
		/// </summary>
		protected override void LoadEditValue()
		{
			working = true;
			foreach(RadioButton radio in control.Controls) {
				if (this.Setting.InEditValue != null &&
				    this.Setting.InEditValue.Equals(radio.Tag)) {
					if (!radio.Checked)
						radio.Checked = true;
				} else {
					if (radio.Checked)
						radio.Checked = false;
				}
			}
			working = false;
		}
		
		private void RadioCheckedChanged(object sender, EventArgs e)
		{
			if (working)
				return;
			RadioButton radio = (RadioButton)sender;
			this.Setting.InEditValue = radio.Tag;
		}
		
		/// <summary>
		/// Gets the <see cref="LabelMode"/> this setting editor uses.<br/>
		/// If not overridden, this property always returns
		/// <see cref="LabelMode">LabelMode.NoLabel</see>.
		/// </summary>
		public override LabelMode LabelMode {
			get {
				return LabelMode.NoLabel;
			}
		}
	}
}
