// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;

namespace Grunwald.Gui.Settings.Editors
{
	/// <summary>
	/// Description of SingleControlEditor.
	/// </summary>
	public abstract class SingleControlEditor : DefaultEditor
	{
		Control control;
		
		/// <summary>
		/// Gets the editing control.
		/// </summary>
		public Control Control {
			get { return control; }
		}
		
		/// <summary>
		/// Creates a new SingleControlEditor.
		/// </summary>
		protected SingleControlEditor(ISetting setting, Control control) : base(setting)
		{
			if (control == null) throw new ArgumentNullException("control");
			this.control = control;
		}
		
		/// <summary>
		/// Gets called when the edit value is changed and should be loaded.
		/// This calls <see cref="DefaultEditor.LoadEditValue"/> using Control.Invoke.
		/// </summary>
		protected override void InEditValueChanged()
		{
			if (control.InvokeRequired)
				control.Invoke(new MethodInvoker(LoadEditValue));
			else
				LoadEditValue();
		}
		
		/// <summary>
		/// Gets the size the edit field needs.
		/// </summary>
		public override Size GetSize()
		{
			return control.Size;
		}
		
		/// <summary>
		/// Gets/Sets whether the control in the SingleControlEditor is enabled or disabled.
		/// </summary>
		public override bool Enabled {
			get {
				return control.Enabled;
			}
			set {
				control.Enabled = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the anshorstyle of the edit control.
		/// </summary>
		public override AnchorStyles Anchor {
			get { return control.Anchor; }
			set { control.Anchor = value; }
		}
		
		int rightSpace = 0;
		
		/// <summary>
		/// Gets/Sets the amount of additional space to leave on the right side.
		/// This property must be set before the call to CreateControls, so best is
		/// to set it in the constructor.
		/// </summary>
		public int RightSpace {
			get { return rightSpace; }
			set { rightSpace = value; }
		}
		
		EditorProperties properties;
		
		/// <summary>
		/// Gets the properties to access the container element.
		/// </summary>
		public EditorProperties Properties {
			get { return properties; }
		}
		
		string currentError = null;
		
		/// <summary>
		/// Sets the displayed error to the specified error text.
		/// If the text is null, the currently displayed error will be removed.<br/>
		/// Calling this will automatically resize the control to have space for the
		/// exclamation mark, so there is no need to use <see cref="RightSpace"/> for errors.
		/// </summary>
		public void SetError(string text)
		{
			if (control.InvokeRequired) throw new InvokeRequiredException();
			if (currentError == text) return;
			if (properties == null)
				throw new InvalidOperationException("Can't call SetError before CreateControls was called. Make sure you call base.CreateControls!");
			if (text == null) {
				properties.ErrorProvider.SetError(control, null);
				control.Width += 16;
			} else {
				if (currentError == null)
					control.Width -= 16;
				Translator t = properties.Translator;
				properties.ErrorProvider.SetError(control, t.Translate(text));
			}
			currentError = text;
		}
		
		/// <summary>
		/// Creates the controls needed for editing the <see cref="ISetting"/> and
		/// adds them to the specified parent control.
		/// </summary>
		public override void CreateControls(Rectangle bounds, EditorProperties properties)
		{
			Control parent = properties.Parent;
			if (control.InvokeRequired || parent.InvokeRequired)
				throw new InvokeRequiredException();
			this.properties = properties;
			bounds.Width -= rightSpace;
			control.Bounds = bounds;
			control.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
			RegisterForOnActivated(control);
			LoadEditValue();
			parent.Controls.Add(control);
		}
		
		/// <summary>
		/// Registers the control's MouseEnter and Enter events to raise the
		/// activated event of this control editor.
		/// </summary>
		protected void RegisterForOnActivated(Control control)
		{
			control.MouseEnter += new EventHandler(ControlEnter);
			control.Enter += new EventHandler(ControlEnter);
		}
		
		private void ControlEnter(object sender, EventArgs e)
		{
			OnActivated(e);
		}
		
		/// <summary>
		/// Activates this editor (sets the focus to the input field).
		/// </summary>
		public override void Activate()
		{
			if (control.InvokeRequired) throw new InvokeRequiredException();
			control.Focus();
		}
		
		/// <summary>
		/// Closes the editor.
		/// </summary>
		/// <remarks>This method disposes the control.</remarks>
		public override void Close()
		{
			base.Close();
			control.Dispose();
		}
	}
}
