// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// The FixedEditorSetting is used to display a special element in a setting page, for example
	/// a button.
	/// A FixedEditorSetting has no value.
	/// Unlike the SpecialEditorSetting, the FixedEditorSetting does not use the type of the editor to create,
	/// but it's CreateEditor method always returns the same editor instance.
	/// This is only useful for *very special* uses of the setting system.
	/// </summary>
	public class FixedEditorSetting : SpecialEditorSetting
	{
		/// <summary>
		/// Creates a new FixedEditorSetting object using name, displayedName and description
		/// from the <see cref="SettingAttribute"/>.
		/// </summary>
		/// <param name="attribute">The <see cref="SettingAttribute"/> used to get name, displayedName and description.</param>
		public FixedEditorSetting(SettingAttribute attribute)
			: base(attribute)
		{
		}
		
		/// <summary>
		/// This constructor is needed for using the FixedEditorSetting as a field in a class.
		/// <code>[Setting(...)]
		/// SpecialEditorSetting ses;</code>
		/// </summary>
		public FixedEditorSetting(SettingAttribute attribute, FieldInfo field)
			: base(attribute, field)
		{
		}
		
		/// <summary>
		/// Creates a new FixedEditorSetting object using the specified properties.
		/// </summary>
		/// <param name="name">The name of the setting (used for storing the setting).</param>
		/// <param name="displayName">The displayed name of this setting. This string is passed to the <see cref="Translator"/>.</param>
		/// <param name="description">The description of the setting. This string is passed to the <see cref="Translator"/>.</param>
		public FixedEditorSetting(string name, string displayName, string description)
			: base(name, displayName, description)
		{
		}
		
		/// <summary>
		/// Creates a new FixedEditorSetting object using the specified properties.
		/// </summary>
		/// <param name="name">The name of the setting (used for storing the setting).</param>
		/// <param name="displayName">The displayed name of this setting. This string is passed to the <see cref="Translator"/>.</param>
		public FixedEditorSetting(string name, string displayName)
			: base(name, displayName, "")
		{
		}
		
		ISettingEditor editor;
		
		/// <summary>
		/// Gets/Sets the editor to return when CreateEditor is called.
		/// </summary>
		public ISettingEditor Editor {
			get { return editor; }
			set { editor = value; }
		}
		
		/// <summary>
		/// Accessing this property results in a NotSupportedException because you have to use
		/// the <see cref="Editor">Editor property</see> instead of EditorType in the FixedEditorSetting.
		/// </summary>
		public override Type EditorType {
			get { throw new NotSupportedException(); }
			set { throw new NotSupportedException(); }
		}
		
		
		/// <summary>
		/// Gets the value of the <see cref="Editor">Editor property</see>.
		/// </summary>
		public override ISettingEditor CreateEditor(EditorFactory factory)
		{
			return editor;
		}
	}
}
