// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.ComponentModel;
using System.Windows.Forms;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// The HelpingSettingsPanel is a <see cref="SettingsPanel"/> with a small box
	/// showing help for the currently edited field.
	/// </summary>
	public class HelpingSettingsPanel : UserControl
	{
		Panel helpPanel = new Panel();
		Label helpLabel = new Label();
		SettingsPanel settingsPanel = new SettingsPanel();
		
		/// <summary>
		/// Gets/Sets the translator to use.
		/// </summary>
		public Translator Translator {
			get { return settingsPanel.Translator; }
			set { settingsPanel.Translator = value; }
		}
		
		/// <summary>
		/// Gets/Sets the <see cref="EditorFactory">editor factory</see> to use.
		/// </summary>
		public EditorFactory EditorFactory {
			get { return settingsPanel.EditorFactory; }
			set { settingsPanel.EditorFactory = value; }
		}
		
		/// <summary>
		/// Gets/Sets if the value property is directly set or the edit values
		/// are just stored in <see cref="ISetting.InEditValue">InEditValue</see>.
		/// </summary>
		[Description("Specifies if the value property is directly set or the edit values " +
		             "are just stored in InEditValue."),
		 DefaultValue(false)]
		public bool DirectEdit {
			get { return settingsPanel.DirectEdit; }
			set { settingsPanel.DirectEdit = value; }
		}
		
		/// <summary>
		/// Gets/Sets the <see cref="ISettingPage"/> currently edited in this SettingsPanel.
		/// </summary>
		public ISettingPage Page {
			get { return settingsPanel.Page; }
			set { settingsPanel.Page = value; }
		}
		
		/// <summary>
		/// Gets the active editor (last mouse move)
		/// </summary>
		public ISettingEditor ActiveEditor {
			get { return settingsPanel.ActiveEditor; }
			set { settingsPanel.ActiveEditor = value; }
		}
		
		/// <summary>
		/// Occurs when the "ActiveEditor" property is changed.
		/// </summary>
		public event EventHandler ActiveEditorChanged {
			add    { settingsPanel.ActiveEditorChanged += value; }
			remove { settingsPanel.ActiveEditorChanged -= value; }
		}
		
		/// <summary>
		/// This event occurs whenever the value in the edit field is changed to a valid
		/// value.
		/// </summary>
		public event SettingChangedEventHandler SettingChanged {
			add    { settingsPanel.SettingChanged += value; }
			remove { settingsPanel.SettingChanged -= value; }
		}
		
		/// <summary>
		/// Creates a new HelpingSettingsPanel
		/// </summary>
		public HelpingSettingsPanel()
		{
			settingsPanel.Dock = System.Windows.Forms.DockStyle.Fill;
			helpLabel.Dock = System.Windows.Forms.DockStyle.Fill;
			helpPanel.Controls.Add(this.helpLabel);
			helpPanel.Dock = System.Windows.Forms.DockStyle.Bottom;
			helpPanel.DockPadding.All = 4;
			helpPanel.Height = 50;
			Controls.Add(settingsPanel);
			Controls.Add(helpPanel);
			
			settingsPanel.ActiveEditorChanged += new EventHandler(OnActiveEditorChanged);
		}
		
		private void OnActiveEditorChanged(object sender, EventArgs e)
		{
			ISettingEditor editor = settingsPanel.ActiveEditor;
			if (editor == null) {
				helpPanel.Visible = false;
			} else {
				Translator t = this.Translator;
				if (t == null) t = Translator.DefaultTranslator;
				helpLabel.Text = t.Translate(editor.Setting.Description);
				helpPanel.Visible = true;
			}
		}
	}
}
