// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// Interface for a single setting that can be edited.
	/// </summary>
	public interface ISetting
	{
		/// <summary>
		/// The name of the setting (used for storing the setting).
		/// </summary>
		string Name { get; }
		
		/// <summary>
		/// The displayed name of this setting. This string is passed to the <see cref="Translator"/>.
		/// </summary>
		string DisplayName { get; }
		
		/// <summary>
		/// The description of the setting. This string is passed to the <see cref="Translator"/>.
		/// </summary>
		string Description { get; }
		
		/// <summary>
		/// Gets/Sets the active value of this setting.
		/// </summary>
		/// <exception cref="SettingEditException">Occurs if this property is set to
		/// an invalid value.</exception>
		object Value { get; set; }
		
		/// <summary>
		/// Gets/Sets the value in the edit dialog. This value gets active when the user
		/// clicks on 'Apply'.
		/// </summary>
		/// <exception cref="SettingEditException">Occurs if this property is set to
		/// an invalid value.</exception>
		object InEditValue { get; set; }
		
		/// <summary>
		/// This event occurs whenever the value of <see cref="Value"/> is changed.
		/// </summary>
		event SettingChangedEventHandler ValueChanged;
		
		/// <summary>
		/// This event occurs whenever the value of <see cref="InEditValue"/> is changed.
		/// </summary>
		event SettingChangedEventHandler InEditValueChanged;
		
		/// <summary>
		/// Creates a new ISettingEditor to edit this setting.
		/// </summary>
		ISettingEditor CreateEditor(EditorFactory factory);
		
		/// <summary>
		/// Saves the setting page into the specified xml writer.
		/// </summary>
		void Save(XmlWriter writer);
		
		/// <summary>
		/// Restores the setting page from the specified xml reader.
		/// </summary>
		void Load(XmlReader reader);
	}
	
	/// <summary>
	/// Event handler used for the <see cref="ISetting.ValueChanged">Setting.ValueChanged</see>
	/// events.
	/// </summary>
	public delegate void SettingChangedEventHandler(object sender, SettingChangedEventArgs e);
	
	/// <summary>
	/// Event arguments used for the <see cref="ISetting.ValueChanged">Setting.ValueChanged</see>
	/// events.
	/// </summary>
	public class SettingChangedEventArgs : EventArgs
	{
		ISetting setting;
		object oldValue;
		object newValue;
		
		/// <summary>
		/// Creates a new SettingChangedEventArgs object.
		/// </summary>
		/// <param name="setting">The changed setting.</param>
		/// <param name="oldValue">The old value of the setting.</param>
		/// <param name="newValue">The new value of the setting.</param>
		public SettingChangedEventArgs(ISetting setting, object oldValue, object newValue)
		{
			this.newValue = newValue;
			this.oldValue = oldValue;
			this.setting = setting;
		}
		
		/// <summary>
		/// The new value of the setting.
		/// </summary>
		public object NewValue {
			get {
				return newValue;
			}
		}
		
		/// <summary>
		/// The old value of the setting.
		/// </summary>
		public object OldValue {
			get {
				return oldValue;
			}
		}
		
		/// <summary>
		/// The changed setting.
		/// </summary>
		public ISetting Setting {
			get {
				return setting;
			}
		}
	}
}
