// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Drawing;
using System.Windows.Forms;
using Grunwald.Gui.Settings.Editors;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// This interface is used for classes that can display an edit field for
	/// <see cref="ISetting">Settings</see> on a windows form.
	/// </summary>
	public interface ISettingEditor
	{
		/// <summary>
		/// Gets the size the edit field needs.
		/// </summary>
		Size GetSize();
		
		/// <summary>
		/// Gets if the edit field can be bigger than the minimal size.
		/// </summary>
		bool CanBeBigger { get; }
		
		/// <summary>
		/// Gets/Sets if the editor is enabled.
		/// </summary>
		bool Enabled { get; set; }
		
		/// <summary>
		/// Gets/Sets the anchorstyle of the edit control.
		/// </summary>
		AnchorStyles Anchor { get; set; }
		
		/// <summary>
		/// Gets the <see cref="ISetting">Setting</see> this Editor was created for.
		/// </summary>
		ISetting Setting { get; }
		
		/// <summary>
		/// "Closes" the Editor, so the editor should deregister all events.
		/// </summary>
		void Close();
		
		/// <summary>
		/// Creates the controls needed for editing the <see cref="Setting"/> and
		/// adds them to the specified parent control.
		/// </summary>
		void CreateControls(Rectangle bounds, EditorProperties properties);
		
		/// <summary>
		/// Occurs if the editing control is "activated" by getting focus or moving the mouse over it.
		/// </summary>
		event EventHandler Activated;
		
		/// <summary>
		/// Activates this editor (sets the focus to the input field).
		/// </summary>
		void Activate();
		
		/// <summary>
		/// Gets the <see cref="LabelMode"/> this setting editor uses.
		/// </summary>
		LabelMode LabelMode { get; }
	}
	
	/// <summary>
	/// The label mode describes how the editor control is aligned after the label.
	/// </summary>
	public enum LabelMode
	{
		/// <summary>
		/// No label is used, the editor control should contain the setting's display name.
		/// </summary>
		NoLabel,
		/// <summary>
		/// The editor control is on the same row as the label
		/// </summary>
		SameRow,
		/// <summary>
		/// The editor control is on the next row under the label.
		/// </summary>
		NextRow
	}
}
