// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// This interface is used for objects that have settings.
	/// </summary>
	public interface ISettingPage : ICloneable
	{
		// WHIDBEY .NET 2.0 - use generic collections insted of arrays.
		
		/// <summary>
		/// Gets the settings of this object as array.
		/// </summary>
		ISetting[] Settings { get; }
		
		/// <summary>
		/// Get the sub pages of this page.
		/// Null means this page has no sub pages and can't have any sub pages.
		/// </summary>
		ISettingPage[] SubPages { get; }
		
		/// <summary>
		/// Gets the first sub page with the specified title.
		/// If the sub page is not found, this method returns null.
		/// </summary>
		ISettingPage GetPage(string title);
		
		/// <summary>
		/// Occurs whenever the list of sub pages is changed.
		/// </summary>
		event EventHandler SubPagesChanged;
		
		/// <summary>
		/// Gets/Sets the title of this setting page.
		/// </summary>
		string Title { get; set; }
		
		/// <summary>
		/// Occurs whenever the title property is changed.
		/// </summary>
		event EventHandler TitleChanged;
		
		/// <summary>
		/// Gets/Sets the image associated with this setting page.
		/// </summary>
		System.Drawing.Image Image { get; set; }
		
		/// <summary>
		/// Occurs whenever the image property is changed.
		/// </summary>
		event EventHandler ImageChanged;
		
		/// <summary>
		/// Gets if adding new sub-pages is allowed.
		/// </summary>
		bool CanAddNewPage { get; }
		
		/// <summary>
		/// Gets if this page is dynamic (return value <b>true</b>) or static (return value
		/// <c>false</c>).<br/>
		/// This influences how this page is loaded from XML:<br/>
		/// A dynamic page is recreated if it didn't exist before load or is dropped if
		/// it don't exist in the XML data, but a static can't get recreated or dropped.
		/// </summary>
		bool DynamicPage { get; }
		
		/// <summary>
		/// Creates a new sub page and adds it to this page.
		/// </summary>
		/// <exception cref="NotSupportedException">The add operation is not allowed.</exception>
		void AddPage();
		
		/// <summary>
		/// Gets if adding the specified page is allowed.
		/// </summary>
		bool CanAddPage(ISettingPage page, int index);
		
		/// <summary>
		/// Adds the specified page to this setting page.
		/// </summary>
		/// <exception cref="NotSupportedException">The add operation is not allowed.</exception>
		void AddPage(ISettingPage page);
		
		/// <summary>
		/// Adds the specified page to this setting page.
		/// </summary>
		/// <exception cref="NotSupportedException">The add operation is not allowed.</exception>
		void AddPage(ISettingPage page, int index);
		
		/// <summary>
		/// Gets if deleting the specified page is allowed.
		/// </summary>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		bool CanDeletePage(ISettingPage page);
		
		/// <summary>
		/// Deletes the specified page from the list of sub pages.
		/// </summary>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		/// <exception cref="NotSupportedException">The delete operation is not allowed.</exception>
		void DeletePage(ISettingPage page);
		
		/// <summary>
		/// Gets if rearranging the specified page in general is allowed.
		/// </summary>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		bool CanMovePage(ISettingPage page);
		
		/// <summary>
		/// Moves the specified page to the specified target index.
		/// </summary>
		/// <exception cref="NotSupportedException">The move operation is not allowed.</exception>
		/// <exception cref="ArgumentOutOfRangeException">The index is out of range.</exception>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		void MovePage(ISettingPage page, int index);
		
		/// <summary>
		/// Returns if moving the specified page to the specified target index is possible.
		/// </summary>
		/// <exception cref="ArgumentOutOfRangeException">The index is out of range.</exception>
		/// <exception cref="ArgumentException">The page is not a sub page of this page.</exception>
		bool CanMovePage(ISettingPage page, int index);
		
		/// <summary>
		/// Saves the InEditValues in this and all subpages to the Values.
		/// </summary>
		void Save();
		
		/// <summary>
		/// Loads all InEditValues in this and all subpages from the Values.
		/// </summary>
		void Load();
		
		/// <summary>
		/// Saves the setting page into the specified xml writer.
		/// </summary>
		void Save(XmlWriter writer);
		
		/// <summary>
		/// Restores the setting page from the specified xml reader.
		/// </summary>
		void Load(XmlReader reader);
		
		/// <summary>
		/// Occurs when a setting on this setting page is changed.
		/// </summary>
		event SettingChangedEventHandler SettingChanged;
		
		/// <summary>
		/// This whole setting system is crap and crying out: I NEED REFACTORING!
		/// Needed as workaround for DGNetwork's ExtensibleSettingPage.
		/// </summary>
		bool CanLoadWithOtherTitle { get; }
	}
}
