// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// The ValidatorAttribute can be used to specify a <see cref="ISettingValidator"/>
	/// for a setting.
	/// </summary>
	[AttributeUsage(AttributeTargets.Field | AttributeTargets.Property,
	                Inherited=true, AllowMultiple = false)]
	public class ValidatorAttribute : Attribute
	{
		ConstructorInfo c;
		
		/// <summary>
		/// Gets the constructor of the class implementing <see cref="ISettingValidator"/>
		/// that needs no arguments.
		/// </summary>
		public ConstructorInfo Constructor {
			get {
				return c;
			}
		}
		
		/// <summary>
		/// Creates a new ValidatorAttribute for validating a setting.
		/// </summary>
		/// <param name="validatorType">The type of the <see cref="ISettingValidator"/>
		/// used for validation.</param>
		public ValidatorAttribute(Type validatorType)
		{
			if (!typeof(ISettingValidator).IsAssignableFrom(validatorType))
				throw new ArgumentException("The type is not a type of a validator.");
			c = validatorType.GetConstructor(new Type[0]);
			if (c == null)
				throw new ArgumentException("The validor must have a constructor without parameters!");
		}
	}
	
	/// <summary>
	/// Interface for classes that are able to validate the value of a
	/// <see cref="ISetting">setting</see>.
	/// </summary>
	public interface ISettingValidator
	{
		/// <summary>
		/// Check if the specified value is valid for the setting.
		/// This method throws a <see cref="SettingEditException"/> if the value is
		/// not valid.
		/// </summary>
		/// <exception cref="SettingEditException">The value is invalid.</exception>
		void CheckValue(ISetting setting, object value);
	}
}
