// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Windows.Forms;
using Grunwald.Gui.Controls;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// Wrapper between SettingPage and RearrangeTreeView
	/// </summary>
	public class PageTreeTag : ITreeTag
	{
		ISettingPage page;
		PageTreeTag parent;
		
		/// <summary>
		/// Creates a new PageTreeTag object for the specified SettingPage
		/// </summary>
		public PageTreeTag(ISettingPage page, PageTreeTag parent)
		{
			if (page == null)
				throw new ArgumentNullException("page");
			this.page = page;
			this.parent = parent;
		}
		
		/// <summary>
		/// Gets the page this tree tag is wrapping.
		/// </summary>
		public ISettingPage Page {
			get {
				return page;
			}
		}
		
		/// <summary>
		/// Gets the parent page of the wrapped page.
		/// </summary>
		public ITreeTag Parent {
			get {
				return parent;
			}
			set {
				parent = (PageTreeTag)value;
			}
		}
		
		/// <summary>
		/// Gets if this node can be a root node.
		/// </summary>
		public bool CanBeRoot {
			get {
				return false;
			}
		}
		
		/// <summary>
		/// Gets if this node can be dragged around.
		/// </summary>
		public bool CanDrag {
			get {
				if (parent == null) return false;
				return parent.Page.CanMovePage(page) || parent.Page.CanDeletePage(page);
			}
		}
		
		/// <summary>
		/// Gets if this tree node can accept the specified node as a child.
		/// </summary>
		public bool AcceptChild(ITreeTag child, int index)
		{
			PageTreeTag tag = child as PageTreeTag;
			if (tag == null)
				return false;
			if (tag.Parent is PageTreeTag && ((PageTreeTag)tag.Parent).Page == this.Page) {
				return page.CanMovePage(tag.Page, index);
			} else {
				if (tag.Parent != null && !tag.CanDelete) return false;
				return page.CanAddPage(tag.Page, index);
			}
		}
		
		/// <summary>
		/// Adds a child to this tree node.
		/// </summary>
		public void AddChild(ITreeTag node, int index)
		{
			PageTreeTag tag = (PageTreeTag)node;
			tag.parent = this;
			page.AddPage(tag.Page, index);
		}
		
		/// <summary>
		/// Removes a child from this tree node.
		/// </summary>
		public void RemoveChild(ITreeTag node)
		{
			PageTreeTag tag = (PageTreeTag)node;
			tag.parent = null;
			page.DeletePage(tag.Page);
		}
		
		/// <summary>
		/// Does nothing because the visual move is done automatically by the <see cref="RearrangeTreeView"/>
		/// and the actual order in the SubPages collection is not saved.
		/// </summary>
		public void MoveChild(ITreeTag node, int index)
		{
			// visual move is automatical, real move not necessary.
		}
		
		/// <summary>
		/// Gets a copy of this object.
		/// </summary>
		public object Clone()
		{
			return new PageTreeTag((ISettingPage)page.Clone(), null);
		}
		
		/// <summary>
		/// Gets if the object can be cloned.
		/// </summary>
		public bool CanClone {
			get {
				return true;
			}
		}
		
		/// <summary>
		/// Gets if this node can be delete using the
		/// <see cref="RearrangeTreeView.Delete">Delete method</see>.
		/// </summary>
		public bool CanDelete {
			get {
				if (parent == null) return false;
				return parent.Page.CanDeletePage(page);
			}
		}
		
		/// <summary>
		/// Deletes this tree node.<br/>
		/// This method is responsible for calling <see cref="RemoveChild"/> on its
		/// parent!
		/// </summary>
		public void Delete()
		{
			if (!CanDelete)
				throw new InvalidOperationException("delete not allowed");
			parent.RemoveChild(this);
			if (page is IDisposable)
				((IDisposable)page).Dispose();
		}
	}
}
