// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// This is an implementation of <see cref="ISetting"/> based on <see cref="DefaultSetting"/>.
	/// It uses Reflection to access to value of the field or property storing the setting.
	/// </summary>
	public class ReflectionSetting : DefaultSetting
	{
		object obj;
		FieldInfo field;
		PropertyInfo property;
		
		/// <summary>
		/// Creates a new ReflectionSetting using the specified parameters.
		/// </summary>
		/// <param name="attribute">The <see cref="SettingAttribute"/> used to get the
		/// name etc.</param>
		/// <param name="obj">The object (normally a <see cref="ISettingPage"/>) that has
		/// the setting stored.</param>
		/// <param name="field">The field on the object that stores the value of the
		/// setting.</param>
		public ReflectionSetting(SettingAttribute attribute, object obj, FieldInfo field)
			: base(attribute, field.GetValue(obj))
		{
			this.obj = obj;
			this.field = field;
			CheckAttributes(field);
		}
		
		/// <summary>
		/// Creates a new ReflectionSetting using the specified parameters.
		/// </summary>
		/// <param name="attribute">The <see cref="SettingAttribute"/> used to get the
		/// name etc.</param>
		/// <param name="obj">The object (normally a <see cref="ISettingPage"/>) that has
		/// the setting stored.</param>
		/// <param name="property">The property on the object that stores the value of the
		/// setting.</param>
		public ReflectionSetting(SettingAttribute attribute, object obj, PropertyInfo property)
			: base(attribute, property.GetValue(obj, null))
		{
			this.obj = obj;
			this.property = property;
			CheckAttributes(property);
		}
		
		ISettingValidator validator;
		
		private void CheckAttributes(MemberInfo member)
		{
			foreach (Attribute attr in member.GetCustomAttributes(true)) {
				if (attr is ValidatorAttribute) {
					ValidatorAttribute v = (ValidatorAttribute)attr;
					validator = (ISettingValidator)v.Constructor.Invoke(null);
				}
			}
		}
		
		/// <summary>
		/// Check if the specified value is valid for this setting.
		/// This method throws a <see cref="SettingEditException"/> if the value is
		/// not valid.
		/// </summary>
		/// <exception cref="SettingEditException">The valid is invalid.</exception>
		public override void CheckValue(object value) {
			if (validator != null)
				validator.CheckValue(this, value);
		}
		
		/// <summary>
		/// Gets/Sets the value of the field using reflection.
		/// </summary>
		public override object Value {
			get {
				if (field != null)
					return field.GetValue(obj);
				else
					return property.GetValue(obj, null);
			}
			set {
				object oldValue = this.Value;
				if (Object.Equals(oldValue, value)) return;
				if (field != null) {
					field.SetValue(obj, value);
					this.OnValueChanged(new SettingChangedEventArgs(this, oldValue, value));
					if (!Object.Equals(value, this.InEditValue))
						this.InEditValue = value;
				} else {
					property.SetValue(obj, value, null);
					object newValue = property.GetValue(obj, null);
					if (!Object.Equals(oldValue, newValue))
						this.OnValueChanged(new SettingChangedEventArgs(this, oldValue, newValue));
					if (!Object.Equals(newValue, this.InEditValue))
						this.InEditValue = newValue;
				}
			}
		}
		
		/// <summary>
		/// The type of the value of this setting.
		/// </summary>
		public Type ValueType {
			get {
				if (field != null)
					return field.FieldType;
				else
					return property.PropertyType;
			}
		}
		
		/// <summary>
		/// Creates a new ISettingEditor to edit this setting.
		/// </summary>
		public override ISettingEditor CreateEditor(EditorFactory factory)
		{
			if (factory == null) throw new ArgumentNullException("factory");
			return factory.GetEditor(this, this.ValueType);
		}
		
		/// <summary>Gets the string representation of this object.</summary>
		public override string ToString()
		{
			if (field != null)
				return base.ToString() + "; field = " + field.Name;
			else
				return base.ToString() + "; property = " + property.Name;
		}
		
		/// <summary>
		/// Saves the setting's value into the specified xml writer.
		/// </summary>
		protected override void SaveValue(XmlWriter writer)
		{
			object value = this.Value;
			new ValueWriter().Write(value, writer);
		}
		
		/// <summary>
		/// Restores the setting page from the specified xml reader.
		/// </summary>
		public override void Load(XmlReader reader)
		{
			object value = new ValueWriter().Read(reader);
			try {
				CheckValue(value);
			} catch (SettingEditException ex) {
				throw new SettingLoadException("The saved value '" + value.ToString() + "' is invalid for the setting " + this.Name, ex);
			}
			this.Value = value;
		}
	}
}
