// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Reflection;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// This implementation of <see cref="ISettingPage"/> uses reflection to get the fields of
	/// it's own instance.
	/// For all fields with a <see cref="SettingAttribute"/> a <see cref="ReflectionSetting"/>
	/// is created.
	/// </summary>
	public abstract class ReflectionSettingPage : DefaultSettingPage
	{
		ISetting[] settings;
		
		/// <summary>
		/// Creates a new ReflectionSettingPage
		/// </summary>
		protected ReflectionSettingPage()
		{
			CreateSettings();
		}
		
		/// <summary>
		/// Creates a copy of the specified ReflectionSettingPage.
		/// </summary>
		protected ReflectionSettingPage(ReflectionSettingPage page) : base(page)
		{
			CreateSettings();
			if (this.settings.Length != page.settings.Length)
				throw new ArgumentException("Can't create the copy, mismatch of setting count");
			for (int i = 0; i < settings.Length; i++) {
				settings[i].Value = page.settings[i].Value;
				settings[i].InEditValue = page.settings[i].InEditValue;
			}
		}
		
		private void CreateSettings()
		{
			ArrayList settingList = new ArrayList();
			Type c = this.GetType();
			if (this.Title == null)
				this.Title = c.Name;
			BindingFlags flags = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;
			foreach (FieldInfo field in c.GetFields(flags)) {
				foreach (Attribute attr in field.GetCustomAttributes(true)) {
					if (attr is SettingAttribute) {
						if (typeof(ISetting).IsAssignableFrom(field.FieldType)) {
							if (field.GetValue(this) == null) {
								ISetting set = (ISetting)Activator.CreateInstance(field.FieldType, new object[] {attr,field});
								field.SetValue(this, set);
							}
							settingList.Add(field.GetValue(this));
						} else {
							settingList.Add(new ReflectionSetting((SettingAttribute)attr, this, field));
						}
						break;
					}
				}
			}
			foreach (PropertyInfo property in c.GetProperties(flags)) {
				foreach (Attribute attr in property.GetCustomAttributes(true)) {
					if (attr is SettingAttribute) {
						settingList.Add(new ReflectionSetting((SettingAttribute)attr, this, property));
						break;
					}
				}
			}
			settings = (ISetting[])settingList.ToArray(typeof(ISetting));
			foreach (ISetting setting in settings) {
				setting.ValueChanged += new SettingChangedEventHandler(SettingValueChanged);
			}
		}
		
		private void SettingValueChanged(object sender, SettingChangedEventArgs e)
		{
			this.OnSettingChanged(e);
		}
		
		/// <summary>
		/// Gets the settings of this object as array.
		/// </summary>
		public override ISetting[] Settings {
			get { return settings; }
		}
		
		/// <summary>
		/// Gets a copy of this SettingPage.<br/>
		/// </summary>
		/// <remarks>
		/// This method uses reflection to get the copy constructor of this instance.
		/// </remarks>
		public override object Clone()
		{
			Type t = this.GetType();
			ConstructorInfo c = t.GetConstructor(new Type[] {t});
			if (c == null)
				throw new NotSupportedException(t.FullName + " doesn't support cloning!");
			return c.Invoke(new object[] {this});
		}
		
		/// <summary>
		/// This method is used by Load(XmlReader) and must be overridden if this page
		/// can have dynamic sub pages.
		/// </summary>
		protected override ISettingPage CreatePage(string typeName)
		{
			foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies()) {
				Type type = assembly.GetType(typeName);
				if (type == null)
					continue;
				
				ConstructorInfo c = type.GetConstructor(Type.EmptyTypes);
				if (c == null)
					throw new SettingLoadException("The type '" + typeName + "' has no parameterless constructor.");
				return (ISettingPage)c.Invoke(null);
			}
			throw new SettingLoadException("Could not find page type '" + typeName + "'.");
		}
	}
}
