// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using Grunwald.Gui.Controls;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// Description of SettingTreeView.
	/// </summary>
	public class SettingTreeView : RearrangeTreeView
	{
		/// <summary>
		/// Creates a new SettingTreeView
		/// </summary>
		public SettingTreeView()
		{
			this.AutoTreeNodeManagement = false;
			Translator.DefaultTranslator.Changed += new EventHandler(TranslatorChanged);
		}
		
		#region Dispose
		/// <summary>
		/// Disposes the control
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			if (isChangeRegistered) {
				isChangeRegistered = false;
				Translator.DefaultTranslator.Changed -= new EventHandler(TranslatorChanged);
			}
			if (disposing)
				ClearMonitoredPages();
			base.Dispose(disposing);
		}
		#endregion
		
		#region TranslatorChanged
		bool isChangeRegistered = true;
		// is true while this tree view is registered on the Translator.Changed event.
		
		private void TranslatorChanged(object sender, EventArgs e) {
			UpdateText(this.Nodes);
		}
		
		private void UpdateText(TreeNodeCollection col) {
			foreach (TreeNode node in col) {
				PageTreeTag tag = node.Tag as PageTreeTag;
				if (tag != null) {
					node.Text = Translator.DefaultTranslator.Translate(tag.Page.Title);
				}
				UpdateText(node.Nodes);
			}
		}
		#endregion
		
		#region SelectedPage
		ISettingPage selectedPage;
		
		/// <summary>
		/// Gets/Sets the currently selected SettingsPage.
		/// </summary>
		public ISettingPage SelectedPage {
			get {
				return selectedPage;
			}
			set {
				if (selectedPage == value) return;
				selectedPage = value;
				TreeNode node = GetNode(selectedPage);
				this.SelectedNode = node;
				node.Expand();
				OnSelectedPageChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Is called whenever the <see cref="SelectedPage">SelectedPage property</see>
		/// is changed.
		/// </summary>
		public event EventHandler SelectedPageChanged;
		
		/// <summary>
		/// Raises the <see cref="SelectedPageChanged">SelectedPageChanged event</see>.
		/// </summary>
		protected virtual void OnSelectedPageChanged(EventArgs e)
		{
			if (SelectedPageChanged != null) {
				SelectedPageChanged(this, e);
			}
		}
		#endregion
		
		#region RootPage
		ISettingPage rootPage;
		
		/// <summary>
		/// Gets/Sets the root page.
		/// </summary>
		public ISettingPage RootPage {
			get {
				return rootPage;
			}
			set {
				if (rootPage == value) return;
				rootPage = value;
				ClearMonitoredPages();
				this.Nodes.Clear();
				if (value != null) {
					this.Nodes.Add(CreateNodeFor(value, null));
					this.ExpandAll();
					MonitorPage(value);
				}
			}
		}
		
		/// <summary>
		/// Is called whenever the <see cref="RootPage">RootPage property</see>
		/// is changed.
		/// </summary>
		public event EventHandler RootPageChanged;
		
		/// <summary>
		/// Raises the <see cref="RootPageChanged">RootPageChanged event</see>.
		/// </summary>
		protected virtual void OnRootPageChanged(EventArgs e)
		{
			if (RootPageChanged != null) {
				RootPageChanged(this, e);
			}
		}
		#endregion
		
		#region CreateNodeFor
		private TreeNode CreateNodeFor(ISettingPage page, PageTreeTag parent)
		{
			if (page == null)
				throw new ArgumentNullException("page");
			PageTreeTag tag = new PageTreeTag(page, parent);
			ISettingPage[] sub = page.SubPages;
			string title = Translator.DefaultTranslator.Translate(page.Title);
			TreeNode node;
			if (sub == null || sub.Length == 0) {
				node = new TreeNode(title);
			} else {
				TreeNode[] nodes = new TreeNode[sub.Length];
				for (int i = 0; i < sub.Length; i++) {
					nodes[i] = CreateNodeFor(sub[i], tag);
				}
				node = new TreeNode(title, nodes);
			}
			UpdateIcon(page, node);
			node.Tag = tag;
			return node;
		}
		#endregion
		
		#region Keep title up to date
		ArrayList monitoredPages = new ArrayList();
		
		private void ClearMonitoredPages()
		{
			foreach(ISettingPage page in monitoredPages) {
				page.SubPagesChanged -= new EventHandler(PagesChanged);
				page.TitleChanged    -= new EventHandler(TitleChanged);
				page.ImageChanged    -= new EventHandler(ImageChanged);
			}
			monitoredPages.Clear();
		}
		
		private void MonitorPage(ISettingPage page)
		{
			monitoredPages.Add(page);
			page.SubPagesChanged += new EventHandler(PagesChanged);
			page.TitleChanged    += new EventHandler(TitleChanged);
			page.ImageChanged    += new EventHandler(ImageChanged);
			ISettingPage[] subPages = page.SubPages;
			if (subPages != null) {
				foreach(ISettingPage sub in subPages) {
					MonitorPage(sub);
				}
			}
		}
		
		private void PagesChanged(object sender, EventArgs e)
		{
			if (this.InvokeRequired) {
				this.Invoke(new EventHandler(PagesChanged), new object[] { sender, e });
			} else {
				ClearMonitoredPages();
				MonitorPage(rootPage);
				ISettingPage page = (ISettingPage)sender;
				TreeNode node = GetNode(page);
				if (node == null) return;
				PageTreeTag tag = (PageTreeTag)node.Tag;
				ISettingPage[] sub = page.SubPages;
				node.Nodes.Clear();
				TreeNode[] nodes = new TreeNode[sub.Length];
				for (int i = 0; i < sub.Length; i++) {
					nodes[i] = CreateNodeFor(sub[i], tag);
				}
				node.Nodes.AddRange(nodes);
				if (this.IsPerformingDrop)
					node.Expand();
			}
		}
		
		private void TitleChanged(object sender, EventArgs e)
		{
			if (this.InvokeRequired) {
				this.Invoke(new EventHandler(TitleChanged), new object[] { sender, e });
			} else {
				ISettingPage page = (ISettingPage)sender;
				TreeNode node = GetNode(page);
				if (node != null)
					node.Text = page.Title;
			}
		}
		
		private void ImageChanged(object sender, EventArgs e)
		{
			if (this.InvokeRequired) {
				this.Invoke(new EventHandler(ImageChanged), new object[] { sender, e });
			} else {
				ISettingPage page = (ISettingPage)sender;
				TreeNode node = GetNode(page);
				if (node != null) {
					UpdateIcon(page, node);
				}
			}
		}
		#endregion
		
		#region GetNode
		/// <summary>
		/// Gets the node representing the specified page.
		/// </summary>
		/// <returns>Returns the node or <c>null</c> if the page was not found.</returns>
		public TreeNode GetNode(ISettingPage page)
		{
			return FindNode(this.Nodes, page);
		}
		
		private TreeNode FindNode(TreeNodeCollection col, ISettingPage page)
		{
			foreach(TreeNode node in col) {
				PageTreeTag tag = node.Tag as PageTreeTag;
				if (tag != null && tag.Page == page)
					return node;
				TreeNode n = FindNode(node.Nodes, page);
				if (n != null)
					return n;
			}
			return null;
		}
		#endregion
		
		#region Icon
		Hashtable imageIndexCache = new Hashtable(); // Image -> int
		
		private void UpdateIcon(ISettingPage page, TreeNode node)
		{
			ImageList imageList = this.ImageList;
			if (usePageImages && imageList != null) {
				Image image = page.Image;
				int imageIndex;
				if (image == null) {
					imageIndex = -1;
				} else {
					if (imageIndexCache.ContainsKey(image)) {
						imageIndex = (int)imageIndexCache[image];
					} else {
						imageIndex = imageList.Images.Count;
						imageList.Images.Add(image);
						imageIndexCache[image] = imageIndex;
					}
				}
				node.ImageIndex = imageIndex;
				node.SelectedImageIndex = imageIndex;
			}
		}
		
		bool usePageImages = true;
		
		/// <summary>
		/// Gets/Sets whether the tree view should automatically use the images
		/// from the setting pages.
		/// </summary>
		/// <remarks>
		/// Setting this property to <c>true</c> requires that the
		/// <see cref="TreeView.ImageList">ImageList property</see> is set.
		/// The SettingTreeView then takes over control over the ImageList, meaning
		/// that it addes images if neccessary. You may not remove images from the ImageList,
		/// because then the internally stored image indexes would not match.
		/// </remarks>
		[Description("Specified whether the tree view should automatically use the images" +
		             " from the setting pages"),
		 DefaultValue(true)]
		public bool UsePageImages {
			get {
				return usePageImages;
			}
			set {
				usePageImages = value;
			}
		}
		#endregion
		
		#region Overrides
		/// <summary>
		/// Raises the <see cref="TreeView.AfterSelect">AfterSelect event</see> and
		/// sets <see cref="SelectedPage"/> to the page of the specified tree node.
		/// </summary>
		protected override void OnAfterSelect(TreeViewEventArgs e)
		{
			if (e.Node != null) {
				PageTreeTag tag = e.Node.Tag as PageTreeTag;
				if (tag != null && tag.Page != selectedPage) {
					selectedPage = tag.Page;
					OnSelectedPageChanged(EventArgs.Empty);
				}
			}
			base.OnAfterSelect(e);
		}
		#endregion
	}
}
