// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using Grunwald.Gui.Settings.Editors;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// Description of SettingsPanel.
	/// </summary>
	public class SettingsPanel : UserControl
	{
		ISetting[] settings;
		ISettingPage page;
		Panel innerPanel;
		
		#region Translator / EditorFactory
		Translator translator;
		
		/// <summary>
		/// Gets/Sets the <see cref="Translator">translator</see> to use.
		/// </summary>
		public Translator Translator {
			get { return translator; }
			set {
				if (translator == value) return;
				translator = value;
				// reset the view:
				ISettingPage oldPage = Page;
				Page = null;
				Page = oldPage;
			}
		}
		
		EditorFactory editorFactory;
		
		/// <summary>
		/// Gets/Sets the <see cref="EditorFactory">editor factory</see> to use.
		/// </summary>
		public EditorFactory EditorFactory {
			get {
				return editorFactory;
			}
			set {
				editorFactory = value;
			}
		}
		#endregion
		
		ErrorProvider errorProvider;
		
		#region Constructor
		/// <summary>
		/// Creates a new SettingsPanel.
		/// </summary>
		public SettingsPanel()
		{
			errorProvider = new ErrorProvider(this);
			AutoScroll = true;
			innerPanel = new Panel();
			innerPanel.Dock = DockStyle.Top;
			Controls.Add(innerPanel);
		}
		#endregion
		
		#region Changing Page
		/// <summary>
		/// Gets/Sets the <see cref="ISettingPage"/> currently edited in this SettingsPanel.
		/// </summary>
		public ISettingPage Page {
			get { return page; }
			set {
				Check();
				if (page == value) return;
				DestroyView();
				page = value;
				if (page == null) {
					settings = null;
				} else {
					settings = page.Settings;
				}
				try {
					CreateView();
				} catch (Exception ex) {
					if (ErrorReporter.IsCritical(ex)) throw;
					Page = null;
					throw;
				}
			}
		}
		#endregion
		
		#region CreateView
		ISettingEditor[] editors;
		int minWidth, minHeight;
		bool hasBigger;
		
		private void CreateView()
		{
			if (settings == null) return;
			if (settings.Length == 0) return;
			
			Translator t = translator;
			if (t == null) t = Translator.DefaultTranslator;
			
			editors = new ISettingEditor[settings.Length];
			string[] texts = new string[settings.Length];
			int[] heights = new int[settings.Length];
			
			// measure needed width
			int freeWidth = 0;
			int height = 4;
			int longestName = 15;
			hasBigger = false;
			EditorFactory factory = editorFactory;
			if (factory == null) factory = EditorFactory.DefaultFactory;
			using (Graphics g = this.CreateGraphics()) {
				for (int i = 0; i < settings.Length; ++i) {
					ISetting set = settings[i];
					editors[i] = set.CreateEditor(factory);
					editors[i].Activated += new EventHandler(EditorActivated);
					set.InEditValueChanged += new SettingChangedEventHandler(SettingInEditValueChanged);
					Size size = editors[i].GetSize();
					freeWidth = Math.Max(freeWidth, size.Width);
					heights[i] = size.Height;
					texts[i] = t.Translate(set.DisplayName + ":");
					LabelMode labelMode = editors[i].LabelMode;
					if (labelMode == LabelMode.SameRow) {
						longestName = Math.Max(longestName, (int)g.MeasureString(texts[i], Font).Width);
						height += 4 + Math.Max(size.Height, 10);
					} else if (labelMode == LabelMode.NoLabel) {
						height += 4 + size.Height;
					} else {
						throw new NotImplementedException("The LabelMode " + labelMode.ToString() + " is not implemented yet in SettingsPanel.cs!");
					}
					if (!hasBigger && editors[i].CanBeBigger)
						hasBigger = true;
				}
			}
			
			minWidth = freeWidth + longestName + 4;
			minHeight = height;
			CalcPanelSize();
			
			EditorProperties properties = new EditorProperties(errorProvider, t, innerPanel);
			
			bool afterBigger = false;
			int pos = 4;
			for (int i = 0; i < editors.Length; ++i) {
				ISettingEditor editor = editors[i];
				bool isBigger = false;
				if (hasBigger && !afterBigger && editor.CanBeBigger) {
					afterBigger = true;
					isBigger = true;
					heights[i] += this.ClientSize.Height - height;
				}
				LabelMode labelMode = editors[i].LabelMode;
				if (labelMode != LabelMode.NoLabel) {
					Label label = new Label();
					label.TextAlign = ContentAlignment.MiddleRight;
					label.Location = new Point(2, pos + 4);
					label.Size = new Size(longestName + 2, 10);
					label.Text = texts[i];
					innerPanel.Controls.Add(label);
					editor.CreateControls(new Rectangle(longestName + 4, pos,
					                                    ClientSize.Width - longestName - 8,
					                                    heights[i]), properties);
					label.Tag = editor;
					RegisterForOnActivated(label);
					pos += 4 + Math.Max(heights[i], 10);
				} else if (labelMode == LabelMode.NoLabel) {
					editor.CreateControls(new Rectangle(4, pos,
					                                    ClientSize.Width - 4,
					                                    heights[i]), properties);
					pos += 4 + heights[i];
				} else {
					throw new NotImplementedException("The LabelMode " + labelMode.ToString() + " is not implemented yet in SettingsPanel.cs!");
				}
				if (isBigger) {
					editor.Anchor |= AnchorStyles.Top | AnchorStyles.Bottom;
				} else if (afterBigger) {
					editor.Anchor = (editor.Anchor & ~AnchorStyles.Top) | AnchorStyles.Bottom;
				} else {
					editor.Anchor = (editor.Anchor & ~AnchorStyles.Bottom) | AnchorStyles.Top;
				}
			}
		}
		#endregion
		
		#region DestroyView
		private void DestroyView()
		{
			if (editors == null) return;
			innerPanel.Controls.Clear();
			foreach (ISettingEditor editor in editors) {
				if (editor == null)
					continue;
				editor.Setting.InEditValueChanged -= new SettingChangedEventHandler(SettingInEditValueChanged);
				editor.Activated -= new EventHandler(EditorActivated);
				editor.Close();
			}
			editors = null;
			SetActiveEditor(null);
		}
		#endregion
		
		#region ActiveEditor
		ISettingEditor activeEditor;
		
		/// <summary>
		/// Gets the active editor (last mouse move)
		/// </summary>
		public ISettingEditor ActiveEditor {
			get { return activeEditor; }
			set {
				if (value == null) throw new ArgumentNullException();
				foreach (ISettingEditor e in editors) {
					if (e == value) {
						e.Activate();
						return;
					}
				}
				throw new ArgumentException("The specified value is not an editor of this SettingsPanel.");
			}
		}
		
		private void EditorActivated(object sender, EventArgs e)
		{
			SetActiveEditor((ISettingEditor)sender);
		}
		
		/// <summary>
		/// Occurs when the "ActiveEditor" property is changed.
		/// </summary>
		public event EventHandler ActiveEditorChanged;
		
		/// <summary>
		/// Raises the <see cref="ActiveEditorChanged"/> event.
		/// </summary>
		protected virtual void OnActiveEditorChanged(EventArgs e)
		{
			if (ActiveEditorChanged != null) {
				ActiveEditorChanged(this, e);
			}
		}
		
		private void SetActiveEditor(ISettingEditor editor)
		{
			if (activeEditor == editor) return;
			activeEditor = editor;
			OnActiveEditorChanged(EventArgs.Empty);
		}
		
		private void RegisterForOnActivated(Control control)
		{
			control.MouseEnter += new EventHandler(ControlEnter);
			control.Enter += new EventHandler(ControlEnter);
		}
		
		private void ControlEnter(object sender, EventArgs e)
		{
			SetActiveEditor((ISettingEditor)((Control)sender).Tag);
		}
		#endregion
		
		#region Dispose/Check
		/// <summary>
		/// Removes all event handlers and calls dispose on the <see cref="UserControl"/>.
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			try {
				if (disposing && !base.IsDisposed){
					DestroyView();
					errorProvider.Dispose();
				}
			} finally {
				base.Dispose(disposing);
			}
		}
		
		private void Check()
		{
			if (base.IsDisposed) throw new ObjectDisposedException("SettingsPanel");
			if (base.InvokeRequired) throw new InvokeRequiredException();
		}
		#endregion
		
		/// <summary>
		/// Raises the <see cref="Control.SizeChanged"/> event.
		/// </summary>
		protected override void OnSizeChanged(System.EventArgs e)
		{
			base.OnSizeChanged(e);
			CalcPanelSize();
		}
		
		private void CalcPanelSize()
		{
			if (hasBigger && ClientSize.Height >= minHeight) {
				if (ClientSize.Width >= minWidth) {
					innerPanel.Dock = DockStyle.Fill;
				} else {
					innerPanel.Dock = DockStyle.Left;
					innerPanel.Width = minWidth;
				}
			} else {
				if (ClientSize.Width >= minWidth) {
					innerPanel.Dock = DockStyle.Top;
					innerPanel.Height = minHeight;
				} else {
					innerPanel.Dock = DockStyle.None;
					innerPanel.Size = new Size(minWidth, minHeight);
				}
			}
		}
		
		bool directEdit = false;
		
		/// <summary>
		/// Gets/Sets if the value property is directly set or the edit values
		/// are just stored in <see cref="ISetting.InEditValue">InEditValue</see>.
		/// </summary>
		[Description("Specifies if the value property is directly set or the edit values " +
		             "are just stored in InEditValue."),
		 DefaultValue(false)]
		public bool DirectEdit {
			get { return directEdit; }
			set { directEdit = value; }
		}
		
		private void SettingInEditValueChanged(object sender, SettingChangedEventArgs e)
		{
			if (directEdit) {
				((ISetting)sender).Value = ((ISetting)sender).InEditValue;
			}
			OnSettingChanged(e);
		}
		
		/// <summary>
		/// This event occurs whenever the value in the edit field is changed to a valid
		/// value.
		/// </summary>
		public event SettingChangedEventHandler SettingChanged;
		
		/// <summary>
		/// Raises the <see cref="SettingChanged">SettingChanged event</see>.
		/// </summary>
		protected virtual void OnSettingChanged(SettingChangedEventArgs e) {
			if (SettingChanged != null) {
				SettingChanged(this, e);
			}
		}
	}
}
