// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;
using System.Xml;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// The SpecialEditorSetting is used to display a special element in a setting page, for example
	/// a button.
	/// A SpecialEditorSetting has no value.
	/// </summary>
	public class SpecialEditorSetting : DefaultSetting
	{
		/// <summary>
		/// Creates a new SpecialEditorSetting object using name, displayedName and description
		/// from the <see cref="SettingAttribute"/>.
		/// </summary>
		/// <param name="attribute">The <see cref="SettingAttribute"/> used to get name, displayedName and description.</param>
		public SpecialEditorSetting(SettingAttribute attribute)
			: base(attribute, null)
		{
		}
		
		/// <summary>
		/// This constructor is needed for using the SpecialEditorSetting as a field in a class.
		/// <code>[Setting(...)]
		/// SpecialEditorSetting ses;</code>
		/// </summary>
		public SpecialEditorSetting(SettingAttribute attribute, FieldInfo field)
			: base(attribute, null)
		{
		}
		
		/// <summary>
		/// Creates a new SpecialEditorSetting object using the specified properties.
		/// </summary>
		/// <param name="name">The name of the setting (used for storing the setting).</param>
		/// <param name="displayName">The displayed name of this setting. This string is passed to the <see cref="Translator"/>.</param>
		/// <param name="description">The description of the setting. This string is passed to the <see cref="Translator"/>.</param>
		public SpecialEditorSetting(string name, string displayName, string description)
			: base(name, displayName, description, null)
		{
		}
		
		/// <summary>
		/// Overridden method for the value.
		/// Always returns null, calling the setter has no effect.
		/// </summary>
		public override object Value {
			get { return null; }
			set { }
		}
		
		Type editorType;
		
		/// <summary>
		/// Gets/Sets the type to create an instance of when CreateEditor is called.
		/// </summary>
		public virtual Type EditorType {
			get { return editorType; }
			set { editorType = value; }
		}
		
		/// <summary>
		/// The factory is not used, this method returns a new instance of the class
		/// specified by the <see cref="EditorType">EditorType property</see>.
		/// </summary>
		public override ISettingEditor CreateEditor(EditorFactory factory)
		{
			return (ISettingEditor)Activator.CreateInstance(editorType, new object[] {this});
		}
		
		/// <summary>
		/// Overridden SaveValue method, does nothing because no value means no saving.
		/// </summary>
		protected override void SaveValue(XmlWriter writer)
		{
			// ignore this setting
		}
		
		/// <summary>
		/// Overridden Load method, because the SpecialEditorSetting can't have a value this
		/// method just skips the current xml element.
		/// </summary>
		public override void Load(XmlReader reader)
		{
			reader.Skip();
		}
	}
}
