// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Reflection;
using System.Xml;
using System.Runtime.Serialization;
using System.Xml.Serialization;

namespace Grunwald.Gui.Settings
{
	/// <summary>
	/// Interface for objects capable of writing values to xml.
	/// </summary>
	public class ValueWriter
	{
		#region Write
		/// <summary>
		/// Writes the specifed object to the XmlWriter.
		/// </summary>
		public void Write(object value, XmlWriter writer)
		{
			if (value == null) {
				writer.WriteString("null");
			} else {
				bool ok = DoWrite(value, writer);
				if (!ok) {
					Type type = value.GetType();
					writer.WriteAttributeString("type", type.FullName);
					XmlSerializer s = new XmlSerializer(type);
					s.Serialize(writer, value);
				}
			}
		}
		
		/// <summary>
		/// Writes the specifed object to the XmlWriter.
		/// </summary>
		protected virtual bool DoWrite(object value, XmlWriter writer)
		{
			if (value is string) {
				writer.WriteAttributeString("type", "string");
				writer.WriteString((string)value);
			} else if (value is bool) {
				writer.WriteAttributeString("type", "bool");
				writer.WriteString(XmlConvert.ToString((bool)value));
			} else if (value is Enum) {
				writer.WriteAttributeString("type", "enum");
				writer.WriteAttributeString("enum", value.GetType().FullName);
				writer.WriteString(value.ToString());
			} else if (value is sbyte) {
				writer.WriteAttributeString("type", "sbyte");
				writer.WriteString(XmlConvert.ToString((sbyte)value));
			} else if (value is short) {
				writer.WriteAttributeString("type", "short");
				writer.WriteString(XmlConvert.ToString((short)value));
			} else if (value is int) {
				writer.WriteAttributeString("type", "int");
				writer.WriteString(XmlConvert.ToString((int)value));
			} else if (value is long) {
				writer.WriteAttributeString("type", "long");
				writer.WriteString(XmlConvert.ToString((long)value));
			} else if (value is byte) {
				writer.WriteAttributeString("type", "byte");
				writer.WriteString(XmlConvert.ToString((byte)value));
			} else if (value is ushort) {
				writer.WriteAttributeString("type", "ushort");
				writer.WriteString(XmlConvert.ToString((ushort)value));
			} else if (value is uint) {
				writer.WriteAttributeString("type", "uint");
				writer.WriteString(XmlConvert.ToString((uint)value));
			} else if (value is ulong) {
				writer.WriteAttributeString("type", "ulong");
				writer.WriteString(XmlConvert.ToString((ulong)value));
			} else if (value is float) {
				writer.WriteAttributeString("type", "float");
				writer.WriteString(XmlConvert.ToString((float)value));
			} else if (value is double) {
				writer.WriteAttributeString("type", "double");
				writer.WriteString(XmlConvert.ToString((double)value));
			} else if (value is decimal) {
				writer.WriteAttributeString("type", "decimal");
				writer.WriteString(XmlConvert.ToString((decimal)value));
			} else if (value is char) {
				writer.WriteAttributeString("type", "char");
				writer.WriteString(XmlConvert.ToString((char)value));
			} else if (value is TimeSpan) {
				writer.WriteAttributeString("type", "TimeSpan");
				writer.WriteString(XmlConvert.ToString((TimeSpan)value));
			} else if (value is DateTime) {
				writer.WriteAttributeString("type", "DateTime");
				writer.WriteString(XmlConvert.ToString((DateTime)value));
			} else if (value is Guid) {
				writer.WriteAttributeString("type", "Guid");
				writer.WriteString(XmlConvert.ToString((Guid)value));
			} else {
				return false;
			}
			return true;
		}
		#endregion
		
		#region Read
		/// <summary>
		/// Reads an object from the XmlReader.
		/// </summary>
		public object Read(XmlReader reader)
		{
			string typeName = reader.GetAttribute("type");
			if (typeName == null) {
				if (reader.ReadInnerXml() != "null") {
					throw new SettingLoadException("A setting without 'type'-attribute can only have the value null!");
				}
				return null;
			} else {
				object value = DoRead(reader, typeName);
				if (value == null) {
					foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies()) {
						Type type = assembly.GetType(typeName);
						if (type == null)
							continue;
						XmlSerializer s = new XmlSerializer(type);
						string text = reader.ReadInnerXml();
						using (System.IO.StringReader r = new System.IO.StringReader(text)) {
							value = s.Deserialize(r);
						}
					}
				}
				if (value == null)
					throw new SettingLoadException("The unknown type '" + typeName + "' could not be loaded.");
				return value;
			}
		}
		
		/// <summary>
		/// Reads an object from the XmlReader.
		/// </summary>
		protected virtual object DoRead(XmlReader reader, string type)
		{
			switch (type) {
				case "string":
					return ReadString(reader);
				case "bool":
					return XmlConvert.ToBoolean(ReadString(reader));
				case "enum":
					return ReadEnum(reader);
				case "sbyte":
					return XmlConvert.ToSByte(ReadString(reader));
				case "short":
					return XmlConvert.ToInt16(ReadString(reader));
				case "int":
					return XmlConvert.ToInt32(ReadString(reader));
				case "long":
					return XmlConvert.ToInt64(ReadString(reader));
				case "byte":
					return XmlConvert.ToByte(ReadString(reader));
				case "ushort":
					return XmlConvert.ToUInt16(ReadString(reader));
				case "uint":
					return XmlConvert.ToUInt32(ReadString(reader));
				case "ulong":
					return XmlConvert.ToUInt64(ReadString(reader));
				case "float":
					return XmlConvert.ToSingle(ReadString(reader));
				case "double":
					return XmlConvert.ToDouble(ReadString(reader));
				case "decimal":
					return XmlConvert.ToDecimal(ReadString(reader));
				case "char":
					return XmlConvert.ToChar(ReadString(reader));
				case "TimeSpan":
					return XmlConvert.ToTimeSpan(ReadString(reader));
				case "DateTime":
					return XmlConvert.ToDateTime(ReadString(reader));
				case "Guid":
					return XmlConvert.ToGuid(ReadString(reader));
				default:
					return null;
			}
		}
		
		private string ReadString(XmlReader reader)
		{
			if (reader.IsEmptyElement) {
				reader.Read();
				return "";
			}
			if (!reader.Read())
				throw new SettingLoadException("Could not read expected text value.");
			string text = reader.Value;
			reader.Read();
			reader.ReadEndElement();
			return text;
		}
		
		private object ReadEnum(XmlReader reader)
		{
			string enumName = reader.GetAttribute("enum");
			if (enumName == null)
				throw new SettingLoadException("A setting of the type 'enum' must have the 'enum' attribute!");
			foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies()) {
				Type enumType = assembly.GetType(enumName);
				if (enumType == null)
					continue;
				return Enum.Parse(enumType, ReadString(reader));
			}
			throw new SettingLoadException("Could not find enum type.");
		}
		#endregion
	}
}
