// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Text;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// A answer transmitted via DG Network.
	/// </summary>
	public class Answer : ICloneable
	{
		#region Time
		DateTime time;
		
		/// <summary>
		/// Gets/Sets the time field. The time field is <b>not</b> saved to byte[], it
		/// can just be used to control the time answers are cached.
		/// </summary>
		public DateTime Time {
			get {
				return time;
			}
			set {
				time = value;
			}
		}
		#endregion
		
		#region Fields
		int fromUser, answerId, packetId;
		byte errorCode = Answer.ErrorCodeSuccess;
		int[] hops;
		
		IndexedValueList values;
		NamedValueList namedValues;
		byte[] restData; // values/namedValues are only created when accessed
		#endregion
		
		#region Constants
		/// <summary>
		/// Error code signalling success.
		/// </summary>
		public const byte ErrorCodeSuccess = 0;
		
		/// <summary>
		/// The request values were invalid.
		/// </summary>
		public const byte ErrorCodeInvalidRequest = 253;
		
		/// <summary>
		/// The target (Packet.Target, not the target user) of the packet was not found.
		/// </summary>
		public const byte ErrorCodeTargetNotFound = 254;
		
		/// <summary>
		/// Error code signalling an internal error while processing the request.
		/// </summary>
		public const byte ErrorCodeInternalError = 255;
		#endregion
		
		#region Properties
		/// <summary>
		/// Gets/Sets the answer id.
		/// The answer id is the same as in the encrypted part of the original answer.
		/// </summary>
		public int AnswerId {
			get {
				return answerId;
			}
			set {
				answerId = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the user who received the answer and has sent the answer.
		/// </summary>
		public int FromUser {
			get {
				return fromUser;
			}
			set {
				fromUser = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets a byte that can determine the status of the answer.
		/// 0 means success, every other number is some kind of error.
		/// Answers with error codes always come from the user who sent the packet,
		/// so this cannot be used as a kind of 'target not found' message.
		/// </summary>
		public byte ErrorCode {
			get {
				return errorCode;
			}
			set {
				errorCode = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the hop array.
		/// </summary>
		public int[] Hops {
			get {
				return hops;
			}
			set {
				hops = value;
			}
		}
		
		/// <summary>
		/// Removes the last hop from the array and returns it.
		/// When the hops array is empty, this method returns 0.
		/// </summary>
		public int DropHop()
		{
			if (hops.Length == 0)
				return 0;
			int val = hops[hops.Length - 1];
			int[] nhops = new int[hops.Length - 1];
			Array.Copy(hops, nhops, nhops.Length);
			hops = nhops;
			return val;
		}
		
		/// <summary>
		/// Gets the last hop of the answer.
		/// </summary>
		public int GetLastHop()
		{
			if (hops.Length == 0)
				return 0;
			return hops[hops.Length - 1];
		}
		
		/// <summary>
		/// Gets/Sets the id of the answer that was answered.
		/// </summary>
		public int PacketId {
			get {
				return packetId;
			}
			set {
				packetId = value;
			}
		}
		
		/// <summary>
		/// Gets the list of named values.
		/// </summary>
		public NamedValueList NamedValues {
			get {
				LoadValues();
				return namedValues;
			}
		}
		
		/// <summary>
		/// Gets the list of indexed values.
		/// </summary>
		public IndexedValueList Values {
			get {
				LoadValues();
				return values;
			}
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates an empty answer.
		/// </summary>
		public Answer()
		{
			hops = new int[0];
			values = new IndexedValueList();
			namedValues = new NamedValueList();
		}
		
		/// <summary>
		/// Creates an answer to the specified packet.
		/// </summary>
		public Answer(Packet packet, int fromUser)
		{
			this.fromUser = fromUser;
			this.packetId = packet.Header.PacketID;
			this.answerId = packet.AnswerId;
			int[] phops = packet.Header.Hops;
			int count = 0;
			for (int i = 0; i < phops.Length; ++i) {
				if (phops[i] == 0)
					break;
				count += 1;
			}
			hops = new int[count];
			Array.Copy(phops, hops, count);
			values = new IndexedValueList();
			namedValues = new NamedValueList();
		}
		#endregion
		
		#region ToByte
		/// <summary>
		/// Gets the length of this answer in bytes.
		/// </summary>
		public int Length {
			get {
				int count = 14 + 4 * hops.Length;
				if (values == null)
					return count + restData.Length;
				else
					return count + values.GetByteLength() + namedValues.GetByteLength();
			}
		}
		
		/// <summary>
		/// Writes the byte representation of this answer to the specified byte array.
		/// </summary>
		public void ToByte(byte[] data, int index)
		{
			int pos = index;
			NetworkBitConverter.PutBytes(data, pos, fromUser);
			pos += 4;
			NetworkBitConverter.PutBytes(data, pos, packetId);
			pos += 4;
			NetworkBitConverter.PutBytes(data, pos, answerId);
			pos += 4;
			data[pos++] = errorCode;
			data[pos++] = (byte)hops.Length;
			for (int i = 0; i < hops.Length; ++i) {
				NetworkBitConverter.PutBytes(data, pos, hops[i]);
				pos += 4;
			}
			if (values == null) {
				restData.CopyTo(data, pos);
			} else {
				using (Stream s = new MemoryStream(data, true)) {
					s.Position = pos;
					values.Save(s);
					namedValues.Save(s);
				}
			}
		}
		
		/// <summary>
		/// Gets a byte array with the byte representation of the answer.
		/// </summary>
		public byte[] ToByte()
		{
			byte[] data = new byte[Length];
			ToByte(data, 0);
			return data;
		}
		#endregion
		
		#region FromByte
		/// <summary>
		/// Restores the answer from the byte array.
		/// </summary>
		public Answer(byte[] data)
			: this(data, 0, data.Length)
		{
		}
		
		/// <summary>
		/// Restores the answer from the byte array.
		/// </summary>
		public Answer(byte[] data, int index, int length)
		{
			int pos = index;
			fromUser = NetworkBitConverter.ToInt32(data, pos);
			pos += 4;
			packetId = NetworkBitConverter.ToInt32(data, pos);
			pos += 4;
			answerId = NetworkBitConverter.ToInt32(data, pos);
			pos += 4;
			errorCode = data[pos++];
			hops = GetArray(data, ref pos);
			restData = new byte[index + length - pos];
			Array.Copy(data, pos, restData, 0, restData.Length);
		}
		
		private int[] GetArray(byte[] data, ref int pos)
		{
			int length = data[pos];
			pos += 1;
			int[] arr = new int[length];
			for (int i = 0; i < length; ++i) {
				arr[i] = NetworkBitConverter.ToInt32(data, pos);
				pos += 4;
			}
			return arr;
		}
		
		private void LoadValues()
		{
			if (values != null)
				return;
			values = new IndexedValueList();
			namedValues = new NamedValueList();
			using (Stream s = new MemoryStream(restData)) {
				values.Read(s);
				namedValues.Read(s);
			}
		}
		#endregion
		
		#region System.ICloneable interface implementation
		object ICloneable.Clone() {
			return this.Clone();
		}
		
		/// <summary>
		/// Clones the answer.
		/// </summary>
		public Answer Clone() {
			Answer n = new Answer();
			n.answerId = this.answerId;
			n.fromUser = this.fromUser;
			n.packetId = this.packetId;
			n.errorCode   = this.errorCode;
			n.hops        = this.hops.Clone() as int[];
			n.values      = this.values.Clone();
			n.namedValues = this.namedValues.Clone();
			n.time = this.time;
			return n;
		}
		#endregion
		
		#region ToString
		/// <summary>
		/// Gets the string representation of this answer.
		/// </summary>
		public override string ToString() {
			return String.Format("[{0} AnswerID={1} FromUser={2} PacketID={3} ErrorCode={4}]",
			                     this.GetType().Name, answerId, fromUser, packetId, errorCode);
		}
		#endregion
	}
}
