// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.ComponentModel;
using System.IO;
using System.Text;

namespace Grunwald.P2PLib.DGNetwork
{
	#region ValueHolder
	internal class ValueHolder : ICloneable
	{
		byte[] data;
		object val;
		
		public byte[] Data {
			get {
				if (data == null) {
					if (val is byte[]) {
						data = (byte[])val;
					} else if (val is byte) {
						data = new byte[] { (byte)val };
					} else if (val is string) {
						data = Encoding.UTF8.GetBytes((string)val);
					} else if (val is short) {
						data = NetworkBitConverter.GetBytes((short)val);
					} else if (val is ushort) {
						data = NetworkBitConverter.GetBytes((ushort)val);
					} else if (val is int) {
						data = NetworkBitConverter.GetBytes((int)val);
					} else if (val is uint) {
						data = NetworkBitConverter.GetBytes((uint)val);
					} else if (val is long) {
						data = NetworkBitConverter.GetBytes((long)val);
					} else if (val is ulong) {
						data = NetworkBitConverter.GetBytes((ulong)val);
					} else if (val is int[]) {
						data = ConvertArray((int[])val);
					} else {
						TypeConverter conv = TypeDescriptor.GetConverter(val.GetType());
						if (conv.CanConvertTo(typeof(byte[]))) {
							data = (byte[])conv.ConvertTo(val, typeof(byte[]));
						} else {
							data = Encoding.UTF8.GetBytes(conv.ConvertToInvariantString(val));
						}
					}
				}
				return data;
			}
			set {
				data = value;
				val = null;
			}
		}
		
		// WHIDBEY GENERIC: Make able to convert all primitive number arrays
		private byte[] ConvertArray(int[] arr)
		{
			byte[] data = new byte[arr.Length * 4];
			int pos = 0;
			for (int i = 0; i < arr.Length; ++i) {
				NetworkBitConverter.PutBytes(data, pos, arr[i]);
				pos += 4;
			}
			return data;
		}
		
		private int[] ConvertArrayBack(byte[] data)
		{
			int[] arr = new int[data.Length / 4];
			int pos = 0;
			for (int i = 0; i < arr.Length; ++i) {
				arr[i] = NetworkBitConverter.ToInt32(data, pos);
				pos += 4;
			}
			return arr;
		}
		
		public void SetValue(object val)
		{
			data = null;
			this.val = val;
		}
		
		public object GetValue(Type type)
		{
			if (val == null || !type.IsInstanceOfType(val)) {
				byte[] data = this.Data;
				if (type == typeof(byte[])) {
					val = data;
				} else if (type == typeof(byte)) {
					val = data[0];
				} else if (type == typeof(string)) {
					val = Encoding.UTF8.GetString(data);
				} else if (type == typeof(short)) {
					val = NetworkBitConverter.ToInt16(data, 0);
				} else if (type == typeof(ushort)) {
					val = NetworkBitConverter.ToUInt16(data, 0);
				} else if (type == typeof(int)) {
					val = NetworkBitConverter.ToInt32(data, 0);
				} else if (type == typeof(uint)) {
					val = NetworkBitConverter.ToUInt32(data, 0);
				} else if (type == typeof(long)) {
					val = NetworkBitConverter.ToInt64(data, 0);
				} else if (type == typeof(ulong)) {
					val = NetworkBitConverter.ToUInt64(data, 0);
				} else if (type == typeof(int[])) {
					val = ConvertArrayBack(data);
				} else {
					TypeConverter conv = TypeDescriptor.GetConverter(type);
					if (conv.CanConvertFrom(typeof(byte[]))) {
						val = conv.ConvertFrom(data);
					} else {
						val = conv.ConvertFromInvariantString(Encoding.UTF8.GetString(data));
					}
				}
			}
			return val;
		}
		
		public ValueHolder(object val)
		{
			if (val == null)
				throw new ArgumentNullException("val");
			this.val = val;
		}
		
		public ValueHolder(byte[] data)
		{
			if (data == null)
				throw new ArgumentNullException("data");
			this.data = data;
		}
		
		public object Clone()
		{
			return new ValueHolder(this.Data);
		}
	}
	#endregion
	
	#region DataHelper
	internal class DataHelper
	{
		/// <summary>Writes a string and it's length.</summary>
		public static void Write(Stream s, string str)
		{
			if (str != null) {
				byte[] data = Encoding.UTF8.GetBytes(str);
				s.WriteByte((byte)data.Length);
				Write(s, data);
			} else {
				s.WriteByte(0);
			}
		}
		
		public static void Write(Stream s, byte[] val)
		{
			s.Write(val, 0, val.Length);
		}
		
		public static string ReadString(Stream s)
		{
			int val = s.ReadByte();
			if (val == -1)
				return null;
			byte length = (byte)val;
			if (length == 0)
				return "";
			byte[] buffer = ReadBytes(s, length);
			return Encoding.UTF8.GetString(buffer);
		}
		
		public static ushort ReadShort(Stream s)
		{
			return NetworkBitConverter.ToUInt16(ReadBytes(s, 2), 0);
		}
		
		public static int ReadInt(Stream s)
		{
			return NetworkBitConverter.ToInt32(ReadBytes(s, 4), 0);
		}
		
		public static byte[] ReadBytes(Stream s, int count)
		{
			if (count < 0)
				throw new ProtocolViolationException("Negative data length is illegal.");
			byte[] b = new byte[count];
			int pos = 0;
			int c;
			while (pos < count) {
				c = s.Read(b, pos, count - pos);
				if (c == 0)
					throw new ProtocolViolationException("Stream ends unexpectedly (only read " + pos + " of " + count + " bytes)");
				pos += c;
			}
			return b;
		}
	}
	#endregion
	
	#region IndexedValueList
	/// <summary>
	/// A list of indexed values that can be converted to byte[].
	/// </summary>
	public class IndexedValueList : ICloneable, IEnumerable
	{
		ArrayList namelessValues = new ArrayList();
		
		/// <summary>
		/// Gets the count of the values accessed by index.
		/// </summary>
		public int Count {
			get {
				return namelessValues.Count;
			}
		}
		
		/// <summary>
		/// Gets an indexed value as the specified type.
		/// </summary>
		public object GetValue(int index, Type type)
		{
			ValueHolder h = (ValueHolder)namelessValues[index];
			if (h == null)
				return null;
			else
				return h.GetValue(type);
		}
		
		/// <summary>
		/// Gets an indexed value as byte array.
		/// </summary>
		public byte[] GetValue(int index)
		{
			ValueHolder h = (ValueHolder)namelessValues[index];
			if (h == null)
				return null;
			else
				return h.Data;
		}
		
		/// <summary>
		/// Sets the indexed value.
		/// </summary>
		public void SetValue(int index, object val)
		{
			if (val == null)
				throw new ArgumentNullException("val");
			ValueHolder h = (ValueHolder)namelessValues[index];
			if (h == null)
				namelessValues[index] = new ValueHolder(val);
			else
				h.SetValue(val);
		}
		
		/// <summary>
		/// Adds an indexed value.
		/// </summary>
		public int AddValue(object val)
		{
			if (val == null)
				throw new ArgumentNullException("val");
			return namelessValues.Add(new ValueHolder(val));
		}
		
		/// <summary>
		/// Saves the values to the stream.
		/// </summary>
		public void Save(Stream s)
		{
			DataHelper.Write(s, NetworkBitConverter.GetBytes((ushort)namelessValues.Count));
			foreach (ValueHolder h in namelessValues) {
				DataHelper.Write(s, NetworkBitConverter.GetBytes(h.Data.Length));
				DataHelper.Write(s, h.Data);
			}
		}
		
		/// <summary>
		/// Gets the number of bytes the list would use when saved.
		/// </summary>
		public int GetByteLength()
		{
			int count = 2;
			foreach (ValueHolder h in namelessValues) {
				count += 4 + h.Data.Length;
			}
			return count;
		}
		
		/// <summary>
		/// Reads the values from the stream.
		/// </summary>
		public void Read(Stream s)
		{
			int namelessCount = DataHelper.ReadShort(s);
			namelessValues.Clear();
			int datalength;
			byte[] valuedata;
			for (int i = 0; i < namelessCount; ++i) {
				datalength = DataHelper.ReadInt(s);
				valuedata = DataHelper.ReadBytes(s, datalength);
				namelessValues.Add(new ValueHolder(valuedata));
			}
		}
		
		/// <summary>
		/// Gets a copy of this data container.
		/// </summary>
		public IndexedValueList Clone()
		{
			IndexedValueList n = new IndexedValueList();
			foreach (ValueHolder h in namelessValues) {
				n.namelessValues.Add(h.Clone());
			}
			return n;
		}
		
		object ICloneable.Clone()
		{
			return this.Clone();
		}
		
		/// <summary>
		/// Gets an enumerator over the byte arrays of all the values.
		/// </summary>
		public IEnumerator GetEnumerator()
		{
			return new IndexedValueEnumerator(namelessValues.GetEnumerator());
		}
		
		private class IndexedValueEnumerator : IEnumerator
		{
			IEnumerator e;
			
			public IndexedValueEnumerator(IEnumerator e)
			{
				this.e = e;
			}
			
			public object Current {
				get {
					return ((ValueHolder)e.Current).Data;
				}
			}
			
			public void Reset()
			{
				
			}
			
			public bool MoveNext()
			{
				return false;
			}
		}
	}
	#endregion
	
	#region NamedValueList
	/// <summary>
	/// A list of named values that can be converted to byte[].
	/// </summary>
	public class NamedValueList : ICloneable, IEnumerable
	{
		ListDictionary namedValues = new ListDictionary();
		
		/// <summary>
		/// Gets the count of the values accessed by key.
		/// </summary>
		public int Count {
			get {
				return namedValues.Count;
			}
		}
		
		/// <summary>
		/// Gets an indexed value as byte array.
		/// </summary>
		public byte[] GetValue(string key)
		{
			ValueHolder h = (ValueHolder)namedValues[key];
			if (h == null)
				return null;
			else
				return h.Data;
		}
		
		/// <summary>
		/// Gets an indexed value as the specified type.
		/// </summary>
		public object GetValue(string key, Type type)
		{
			ValueHolder h = (ValueHolder)namedValues[key];
			if (h == null)
				return null;
			else
				return h.GetValue(type);
		}
		
		/// <summary>
		/// Sets a named value.
		/// </summary>
		public void SetValue(string key, object val)
		{
			if (val == null) {
				namedValues.Remove(key);
			} else if (namedValues.Contains(key)) {
				((ValueHolder)namedValues[key]).SetValue(val);
			} else {
				namedValues[key] = new ValueHolder(val);
			}
		}
		
		/// <summary>
		/// Adds a named value.
		/// </summary>
		public void AddValue(string key, object val)
		{
			SetValue(key, val);
		}
		
		/// <summary>
		/// Saves the values to the stream.
		/// </summary>
		public void Save(Stream s)
		{
			DataHelper.Write(s, NetworkBitConverter.GetBytes((ushort)namedValues.Count));
			foreach (DictionaryEntry e in namedValues) {
				if (e.Value == null)
					continue;
				ValueHolder h = (ValueHolder)e.Value;
				DataHelper.Write(s, (string)e.Key);
				DataHelper.Write(s, NetworkBitConverter.GetBytes(h.Data.Length));
				DataHelper.Write(s, h.Data);
			}
		}
		
		/// <summary>
		/// Gets the number of bytes the list would use when saved.
		/// </summary>
		public int GetByteLength()
		{
			int count = 2;
			foreach (DictionaryEntry e in namedValues) {
				count += Encoding.UTF8.GetByteCount((string)e.Key);
				count += 5; // length(key) + length(value)
				if (e.Value != null)
					count += ((ValueHolder)e.Value).Data.Length;
			}
			return count;
		}
		
		/// <summary>
		/// Reads the values from the stream.
		/// </summary>
		public void Read(Stream s)
		{
			int namelessCount = DataHelper.ReadShort(s);
			namedValues.Clear();
			for (int i = 0; i < namelessCount; ++i) {
				string key = DataHelper.ReadString(s);
				int datalength = DataHelper.ReadInt(s);
				if (datalength == 0) {
					namedValues.Add(key, null);
				} else {
					byte[] valuedata = DataHelper.ReadBytes(s, datalength);
					namedValues.Add(key, new ValueHolder(valuedata));
				}
			}
		}
		
		/// <summary>
		/// Gets a copy of this data container.
		/// </summary>
		public NamedValueList Clone()
		{
			NamedValueList n = new NamedValueList();
			foreach (DictionaryEntry h in namedValues) {
				if (h.Value == null)
					n.namedValues.Add(h.Key, null);
				else
					n.namedValues.Add(h.Key, ((ValueHolder)h.Value).Clone());
			}
			return n;
		}
		
		object ICloneable.Clone()
		{
			return this.Clone();
		}
		
		/// <summary>
		/// Gets an enumerator over the dictionary entries.
		/// </summary>
		public IDictionaryEnumerator GetEnumerator()
		{
			return new NamedValueListEnumerator(namedValues.GetEnumerator());
		}
		
		IEnumerator IEnumerable.GetEnumerator()
		{
			return this.GetEnumerator();
		}
		
		private class NamedValueListEnumerator : IDictionaryEnumerator
		{
			IDictionaryEnumerator e;
			
			public NamedValueListEnumerator(IDictionaryEnumerator e)
			{
				this.e = e;
			}
			
			public object Key {
				get {
					return e.Key;
				}
			}
			
			public object Value {
				get {
					return ((ValueHolder)e.Value).Data;
				}
			}
			
			public DictionaryEntry Entry {
				get {
					return new DictionaryEntry(this.Key, this.Value);
				}
			}
			
			public object Current {
				get {
					return this.Entry;
				}
			}
			
			public void Reset()
			{
				e.Reset();
			}
			
			public bool MoveNext()
			{
				return e.MoveNext();
			}
		}
	}
	#endregion
}
