// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Net;
using System.Text;
using System.Xml;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.IPDetectors;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Implementation of the DG Network IP-Server.
	/// </summary>
	public class DgnIPServer : IPServer
	{
		#region Constructor
		/// <summary>
		/// Creates a new DGNIPServer.
		/// </summary>
		public DgnIPServer(string url, AbstractServer server)
		{
			this.Server = server;
			this.Url = url;
			ipDetector = new XmlBasedDetector(url, "/dgnipserver/status/ip");
			ipDetector.IPAddressChanged += new EventHandler(DetectorIPAddressChanged);
			StringBuilder b = new StringBuilder();
			for (int i = 0; i < 16; ++i) {
				int num = Helper.Random.Next((int)'A', (int)'Z');
				if (Helper.Random.NextDouble() > 0.5) num += 32;
				b.Append((char)num);
			}
			password = b.ToString();
		}
		
		#endregion
		
		#region Properties
		/// <summary>
		/// The collection of request headers that are sent to the server.
		/// </summary>
		public WebHeaderCollection RequestHeaders {
			get {
				return ipDetector.RequestHeaders;
			}
		}
		
		Encoding encoding = Encoding.UTF8;
		
		/// <summary>
		/// Gets/Sets the encoding to use to encode the password and network identifier in the URL.
		/// </summary>
		public Encoding Encoding {
			get {
				return encoding;
			}
			set {
				if (value == null)
					throw new ArgumentNullException("encoding");
				encoding = value;
			}
		}
		
		DgnIPServerError lastError = DgnIPServerError.DetectorError;
		
		/// <summary>
		/// Gets the description for the last error of this ip server.
		/// </summary>
		public DgnIPServerError LastError {
			get { return lastError; }
		}
		
		string lastServerError = null;
		
		/// <summary>
		/// Gets the last error message returned from the server.
		/// This property should only be used when LastError is DgnIPServerError.ServerError.
		/// </summary>
		public string LastServerError {
			get { return lastServerError; }
		}
		
		int lastHostCount;
		
		/// <summary>
		/// Gets the number of hosts added to the cache in the last successful request.
		/// </summary>
		public int LastHostCount {
			get { return lastHostCount; }
		}
		
		AbstractServer server;
		
		/// <summary>
		/// Gets/Sets the server used for incoming connections.
		/// </summary>
		public AbstractServer Server {
			get {
				return server;
			}
			set {
				if (value == null)
					throw new ArgumentNullException("server");
				server = value;
			}
		}
		
		string url;
		
		/// <summary>
		/// The URL where the IP-Server is located.
		/// </summary>
		public string Url {
			get { return url; }
			set { url = value; }
		}
		
		XmlBasedDetector ipDetector;
		
		/// <summary>
		/// The DGNIPServer uses the <see cref="XmlBasedDetector"/> for contacting
		/// the server.
		/// </summary>
		public XmlBasedDetector IPDetector {
			get { return ipDetector; }
		}
		
		/// <summary>
		/// The IPAddress that was detected last.
		/// </summary>
		public IPAddress IPAddress {
			get { return IPDetector.IPAddress; }
		}
		
		#region Time
		int time;
		
		/// <summary>
		/// Gets the time on the server (unix time stamp) when the last request was
		/// sent. To get the current time, add <see cref="SecondsSinceLastRequest"/>
		/// to this value.
		/// </summary>
		public int ServerTime {
			get { return time; }
		}
		
		/// <summary>
		/// Gets the number of seconds that have passed since the last request was sent.
		/// </summary>
		public int SecondsSinceLastRequest {
			get {
				return (int)DateTime.Now.Subtract(IPDetector.LastRefresh).TotalSeconds;
			}
		}
		
		int nextRefresh;
		/// <summary>
		/// Gets the time in seconds when the next request should be sent.
		/// So you should send the next request when
		/// <see cref="SecondsSinceLastRequest"/> &gt; NextRefresh.
		/// </summary>
		public int NextRefresh {
			get { return nextRefresh; }
		}
		
		/// <summary>
		/// Gets if a refresh is needed.
		/// </summary>
		public override bool RefreshNeeded {
			get {
				int passedSecs = SecondsSinceLastRequest;
				if (passedSecs < 0) return true;
				return passedSecs > NextRefresh;
			}
		}
		#endregion
		
		#region UserID
		int userID;
		
		/// <summary>
		/// Gets/Sets the user number that is used to make the requests.
		/// This can change if the server assigns a different user id.
		/// Is 0 if no user id is assigned.
		/// </summary>
		public int UserID {
			get { return userID; }
			set {
				if (value < 0) throw new ArgumentOutOfRangeException();
				if (userID == value) return;
				userID = value;
				OnUserIDChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Occurs whenever the value of UserID is changed
		/// (either by another object or by the ip server)
		/// </summary>
		public event EventHandler UserIDChanged;
		
		/// <summary>
		/// Occurs whenever the value of UserID is changed
		/// (either by another object or by the ip server)
		/// </summary>
		protected virtual void OnUserIDChanged(EventArgs e)
		{
			if (UserIDChanged != null) {
				UserIDChanged(this, e);
			}
		}
		#endregion
		
		string password;
		
		/// <summary>
		/// Gets/Sets the password to use to connect to the ip server.
		/// </summary>
		public string Password {
			get { return password; }
			set { password = value; }
		}
		
		#region ReceivingUser
		ReceivingUser receivingUser;
		UserAddress userAddress;
		
		/// <summary>
		/// Gets/Sets the receiving user that should automatically be assigned the addresses
		/// got by this IP Server.
		/// </summary>
		public ReceivingUser ReceivingUser {
			get {
				return receivingUser;
			}
			set {
				userAddress = null;
				receivingUser = value;
			}
		}
		
		/// <summary>
		/// Updates the address this ip server has automatically assigned to the receiving user.
		/// This method does nothing if no receiving user was set.
		/// </summary>
		protected void UpdateUserAddress(DgnP2PNetwork network) {
			if (receivingUser == null)
				return;
			if (userAddress != null) {
				int pos = receivingUser.Addresses.IndexOf(userAddress);
				userAddress = new UserAddress(network, UserID, receivingUser.Identity);
				receivingUser.Addresses[pos] = userAddress;
			} else {
				userAddress = new UserAddress(network, UserID, receivingUser.Identity);
				receivingUser.Addresses.Add(userAddress);
			}
		}
		#endregion
		#endregion
		
		#region Refresh
		/// <summary>
		/// Refreshes the IP-Server.
		/// </summary>
		/// <param name="network">The <see cref="DgnP2PNetwork"/> that should
		/// be filled with new addresses.</param>
		/// <returns>Returns true if the refresh was successfull.</returns>
		protected override bool DoRefresh(P2PNetwork network)
		{
			if (!(network is DgnP2PNetwork))
				throw new ArgumentException("Network must be a P2PNetwork!");
			
			StringBuilder b = new StringBuilder(this.Url);
			
			if (this.Url.IndexOf('?') >= 0)
				b.Append("&pwd=");
			else
				b.Append("?pwd=");
			b.Append(Helper.EscapePart(password, encoding));
			
			// only report that we have no server ability when
			// a test has already run and it returned that we cannot
			// accept incoming connections.
			b.Append("&port=");
			if (server.ServerAbilityTester == null ||
			    server.ServerAbilityTester.LastResult ||
			    server.ServerAbilityTester.LastFinishedTest == DateTime.MinValue)
				b.Append(server.Port);
			else
				b.Append('0');
			
			if (userID > 0) {
				b.Append("&uid=");
				b.Append(userID);
			}
			b.Append("&identifier=");
			b.Append(Helper.EscapePart(network.Identifier, encoding));
			IPDetector.Url = b.ToString();
			
			bool success = this.IPDetector.Refresh();
			if (success) {
				success = ProcessXml(this.IPDetector.LastXmlDocument, (DgnP2PNetwork)network);
			} else {
				nextRefresh = 120;
				if (this.IPDetector.LastError == IPDetectorError.AddressNotFoundInResponse) {
					XmlElement root = this.IPDetector.LastXmlDocument.DocumentElement;
					if (root["error"] != null) {
						this.lastError = DgnIPServerError.ServerError;
						lastServerError = root["error"].InnerText;
					} else {
						this.lastError = DgnIPServerError.DetectorError;
					}
				} else {
					this.lastError = DgnIPServerError.DetectorError;
				}
			}
			return success;
		}
		
		#endregion
		
		#region Process XML
		
		/// <summary>
		/// Parses the XMLDocument and adds all found host entries to the P2PNetwork.
		/// </summary>
		public bool ProcessXml(XmlDocument xmlDocument, DgnP2PNetwork network)
		{
			XmlElement root = xmlDocument.DocumentElement;
			if (root["error"] != null) {
				lastServerError = root["error"].InnerText;
				lastError = DgnIPServerError.ServerError;
				return false;
			}
			XmlElement status = root["status"];
			try {
				nextRefresh = int.Parse(status["nextrefresh"].InnerText);
				time = int.Parse(status["time"].InnerText);
				UserID = int.Parse(status["yourid"].InnerText);
				UpdateUserAddress(network);
				network.ConnectKey = status["connectkey"].InnerText;
				XmlElement hosts = root["hosts"];
				XmlElement host;
				int count = 0;
				foreach (XmlNode n in hosts.ChildNodes) {
					if (!(n is XmlElement)) continue;
					host = (XmlElement)n;
					if (host.Name != "host") continue;
					count += 1;
					int userid = int.Parse(host.GetAttribute("userid"));
					IPEndPoint endPoint = GetEndPoint(host.InnerText);
					UserAddress adr = new UserAddress(network, userid);
					network.HostCache.Remove(adr);
					network.HostCache.Add(new AddressEntry(adr, endPoint));
				}
				this.lastHostCount = count;
			} catch (Exception ex) {
				this.lastError = DgnIPServerError.UnhandledException;
				Helper.LogError(ex);
				if (Helper.IsCritical(ex)) throw;
				return false;
			}
			return true;
		}
		
		static IPEndPoint GetEndPoint(string text)
		{
			int pos = text.IndexOf(':');
			if (pos < 0) throw new ArgumentException();
			IPAddress ip = IPAddress.Parse(text.Substring(0, pos));
			int port = int.Parse(text.Substring(pos + 1));
			return new IPEndPoint(ip, port);
		}
		
		#endregion
		
		#region IPAddress Changed
		private void DetectorIPAddressChanged(object sender, EventArgs e)
		{
			if (server != null && ipDetector.LastRefreshWasSuccessful)
				server.IPAddress = ipDetector.IPAddress;
			OnIPAddressChanged(e);
		}
		
		/// <summary>
		/// Occurs when the ip address changes.
		/// </summary>
		public event EventHandler IPAddressChanged;
		
		/// <summary>
		/// Raises the IPAddressChanged event.
		/// </summary>
		protected virtual void OnIPAddressChanged(EventArgs e)
		{
			if (IPAddressChanged != null) {
				IPAddressChanged(this, e);
			}
		}
		#endregion
	}
	
	/// <summary>
	/// Enumeration containing the possible errors in the IP server
	/// </summary>
	public enum DgnIPServerError {
		/// <summary>
		/// See the error in the detector
		/// </summary>
		DetectorError,
		/// <summary>
		/// The server returned an error tag.
		/// </summary>
		ServerError,
		/// <summary>
		/// Unknown exception error
		/// </summary>
		UnhandledException,
	}
}
