// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.Server;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Description of DGNetwork.
	/// </summary>
	public class DgnP2PNetwork : P2PNetwork
	{
		#region Properties
		string identifier;
		/// <summary>
		/// The identifier of this network.
		/// </summary>
		public override string Identifier {
			get { return identifier; }
		}
		
		string connectKey;
		/// <summary>
		/// The key to connect to this network.
		/// </summary>
		public string ConnectKey {
			get { return connectKey; }
			set { connectKey = value; }
		}
		
		/// <summary>
		/// Gets the first user id of the first receiving user.
		/// </summary>
		public int UserID {
			get {
				ReceivingUser user = this.ReceivingUsers[0];
				UserAddress adr = (UserAddress)user.Addresses.GetAddressByNetwork(this);
				return adr.UserID;
			}
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates a new DGNetwork with the specified identifier.
		/// </summary>
		public DgnP2PNetwork(string identifier)
		{
			this.identifier = identifier;
			this.routingStrategy = new PeerRoutingStrategy(this);
			this.encryptionStrategy.Network = this;
		}
		
		/// <summary>
		/// Creates a new DGNetwork with the specified identifier.
		/// </summary>
		public DgnP2PNetwork(string identifier, ReceivingUser mainUser) : this(identifier)
		{
			this.ReceivingUsers.Add(mainUser);
		}
		#endregion
		
		#region AcceptConnection
		/// <summary>
		/// Checks if a connection with the same partner already exists.
		/// If a connection already exists, this method returns <b>false</b>.
		/// If the connection does not exist yet, the connection is added
		/// to the <see cref="ConnectionCollection"/> and this method returns
		/// <b>true</b>.
		/// </summary>
		/// <exception cref="ArgumentNullException">An argument is null.</exception>
		/// <exception cref="ArgumentException">The connection is no NetClient.</exception>
		public override bool AcceptConnection(Connection connection)
		{
			lock (this) {
				if (connection == null) throw new ArgumentNullException("connection");
				if (!(connection is NetClient)) throw new ArgumentException("Connection must be a NetClient");
				return base.AcceptConnection(connection);
			}
		}
		#endregion
		
		#region BuildConnection
		/// <summary>
		/// Initiates a new network connection using the existing <see cref="IConnection"/>.
		/// </summary>
		/// <param name="connection">The TCP/IP connection to use.</param>
		/// <param name="addressEntry">The address of the network node connecting to</param>
		public override void BuildConnection(IConnection connection, AddressEntry addressEntry)
		{
			UserAddress adr = addressEntry.Address as UserAddress;
			if (adr == null)
				throw new ArgumentException("Can only create connections for UserAddresses!");
			Console.WriteLine("Build it...");
			// receivingUsers[0] is a bit hacky...
			NetClient nc = BuildConnection(connection, this.ReceivingUsers[0], adr.UserID);
			if (nc == null) return;
			nc.ReceiveServerResponse();
			if (nc.Status == NetClientStatus.ReadyForLoop)
				nc.ReceiveLoop();
		}
		
		/// <summary>
		/// Creates a new <see cref="NetClient"/>, registers the NetClient for this network,
		/// lets the NetClient send the client request and returns it.
		/// </summary>
		/// <param name="connection">The connection the new netclient should use.</param>
		/// <param name="owner">The ReceivingUser who owns the connection.
		/// The user must have an <see cref="DgnAddress"/> for this network, the id of
		/// that address is used as source user id.</param>
		/// <param name="partnerID">The user id of the user to connect to.
		/// The connection can't be established if this id is wrong.
		/// This id is used for the address <see cref="Connection.Partner">NetClient.Partner</see></param>
		/// <exception cref="ArgumentException">
		/// The owner has no address for this network.<br />
		/// -or-<br />
		/// The owner's address for this network is not an instance of DGNAddress.
		/// </exception>
		/// <exception cref="ArgumentNullException">connection or owner is null</exception>
		/// <returns>Returns the new NetClient object, or null if some error occured.</returns>
		/// <remarks>
		/// After this operation the connection is managed by the NetClient, so you shouldn't
		/// call anything on the connection again but use the NetClient.
		/// If the connection cannot be build, it is closed and this method returns null.
		/// </remarks>
		public virtual NetClient BuildConnection(IConnection connection, ReceivingUser owner, int partnerID)
		{
			if (connection == null) throw new ArgumentNullException("connection");
			if (owner == null) throw new ArgumentNullException("owner");
			NameValueCollection headers;
			lock (this) {
				P2PAddress adr = owner.Addresses.GetAddressByNetwork(this);
				if (adr == null)
					throw new ArgumentException("The owner has no address for this network.", "owner");
				if (!(adr is DgnAddress))
					throw new ArgumentException("The owner's address for this network is no DGNAddress.", "owner");
				DgnAddress a = adr as DgnAddress;
				headers = new NameValueCollection();
				headers.Add("Me", a.UserID.ToString("d"));
				headers.Add("MyIdentity", owner.Identity.PublicKey);
			}
			NetClient c = new NetClient(connection, new UserAddress(this, partnerID), headers);
			if (!AcceptConnection(c)) {
				connection.Close();
				return null;
			} else {
				return c;
			}
		}
		#endregion
		
		#region Work
		DateTime lastRTSendTime = DateTime.Now;
		
		/// <summary>
		/// This method removes dead connections from the collection and
		/// tries to build new connections if neccessary.
		/// </summary>
		protected override void OnWorking(EventArgs e)
		{
			lock (this) {
				ConnectionCollection col = base.Connections;
				// Remove dead connections
				for (int i = 0; i < col.Count; ++i) {
					Connection c = col[i];
					if (!c.Connected) {
						col.RemoveAt(i--);
						this.OnConnectionRemoved(new ConnectionEventArgs(c));
					}
				}
				DateTime now = DateTime.Now;
				if (lastRTSendTime.AddSeconds(10) < now || lastRTSendTime > now) {
					foreach (Connection c in col) {
						if (c is NetClient)
							routingStrategy.SendRoutingTable(c as NetClient, false);
					}
					lastRTSendTime = now;
				}
			}
			base.OnWorking(e);
		}
		#endregion
		
		#region Routing
		IRoutingStrategy routingStrategy;
		
		/// <summary>
		/// Gets/Sets the RoutingStrategy to use.
		/// Note: The RoutingStragegy can't be changed while the network is
		/// connected, so changing this will disconnect and reconnect.
		/// </summary>
		public IRoutingStrategy RoutingStrategy {
			get {
				return routingStrategy;
			}
			set {
				if (value == null)
					throw new ArgumentNullException();
				bool connected = this.Connected;
				if (connected)
					Disconnect();
				routingStrategy = value;
				if (connected)
					Connect();
			}
		}
		
		/// <summary>
		/// Is raised whenever a packet is received.
		/// </summary>
		public event PacketEventHandler Receive;
		
		/// <summary>
		/// Raises the Receive event.
		/// </summary>
		protected virtual void OnReceive(PacketEventArgs e)
		{
			if (Receive != null) {
				Receive(this, e);
			}
		}
		
		/// <summary>
		/// Raises the Receive event for the specified packet.
		/// </summary>
		/// <param name="p">The packet that has been received.</param>
		/// <param name="userId">The id of the user who has received the packet.</param>
		public void ReceivePacket(Packet p, int userId)
		{
			Answer answer;
			if (p.Header.PacketID == 0)
				answer = null;
			else
				answer = new Answer(p, userId);
			try {
				ReceivePacketInternal(p, answer);
			} catch (Exception ex) {
				if (answer != null)
					answer.ErrorCode = Answer.ErrorCodeInternalError;
				Helper.LogError(ex);
			}
			if (answer != null)
				ReceiveAnswer(answer);
		}
		
		private void ReceivePacketInternal(Packet p, Answer answer)
		{
			switch (p.Target) {
				case "RoutedAnswer":
					ReceiveAnswer(new Answer(p.Values.GetValue(0)));
					break;
				case "Ping":
					break;
				case "Trace":
					if (answer != null)
						answer.Values.AddValue(p.Header.Hops);
					break;
				case "GetIdentity":
					if (answer != null)
						encryptionStrategy.RespondToGetIdentity(p, answer);
					break;
				default:
					OnReceive(new PacketEventArgs(p, answer));
					break;
			}
		}
		
		/// <summary>
		/// Decrypts the packet and raises the receive event for it.
		/// </summary>
		/// <param name="header">The header of the packet.</param>
		/// <param name="data">The array containing the packet body.</param>
		/// <param name="offset">The offset where in the array the packet body starts.</param>
		/// <param name="length">The length of the packet body.</param>
		/// <param name="userId">The user receiving the packet.</param>
		public void ReceivePacket(PacketHeader header, byte[] data, int offset, int length, int userId)
		{
			try {
				ReceivePacket(new Packet(header, data, offset, length, this.EncryptionStrategy), userId);
			} catch (Exception ex) {
				if (Helper.IsCritical(ex))
					throw;
				Helper.LogError(ex);
			}
		}
		
		/// <summary>
		/// Sends the specified packet over the network.
		/// </summary>
		public SendResult SendPacket(Packet p)
		{
			SendResult result = null;
			byte[] data = p.ToByte(this.EncryptionStrategy);
			int headerLength = p.Header.Length;
			if (p.AnswerId != 0) {
				result = new SendResult(p);
				RegisterSendResult(result);
			}
			routingStrategy.SendPacket(p.Header, data, headerLength, data.Length - headerLength);
			return result;
		}
		#endregion
		
		#region Answer management
		// TODO: Clear these tables after time limit is exceeded.
		Hashtable sendResults = new Hashtable(); // ulong -> SendResult
		Hashtable answers = new Hashtable(); // ulong -> Answer
		
		private void RegisterSendResult(SendResult result)
		{
			lock (sendResults) {
				sendResults[result.UniqueKey] = result;
			}
		}
		
		/// <summary>
		/// Gets the send result for the specified source user / packet id.
		/// Only send results that don't have an answer yet can be accessed.
		/// </summary>
		public SendResult GetSendResult(int packetId, int sourceUser)
		{
			lock (sendResults) {
				return (SendResult)sendResults[SendResult.GetUniqueIdentifier(packetId, sourceUser)];
			}
		}
		
		/// <summary>
		/// Gets the answer to the packet <paramref name="packetId"/>
		/// from the user <paramref name="from"/> to the user <paramref name="to"/>.
		/// </summary>
		public Answer GetAnswer(int packetId, int from, int to)
		{
			lock (answers) {
				return (Answer)answers[GetCode(packetId, from, to)];
			}
		}
		
		private ulong GetCode(Answer answer)
		{
			return GetCode(answer.PacketId, answer.Hops[0], answer.FromUser);
		}
		
		private ulong GetCode(int packetId, int from, int to)
		{
			unchecked {
				return ((ulong)packetId << 32) + ((ulong)from << 16) + (ulong)to;
			}
		}
		
		/// <summary>
		/// Receives the answer packet. It is cache and forwarded if necessary.
		/// </summary>
		public void ReceiveAnswer(Answer answer)
		{
			int targetUser = answer.Hops[0];
			ulong code = GetCode(answer);
			Answer old;
			lock (answers) {
				old = (Answer)answers[code];
				if (old != null) {
					old.Time = DateTime.Now;
				} else {
					answer.Time = DateTime.Now;
					answers[code] = answer;
				}
			}
			if (routingStrategy.SendAnswer(answer)) {
				// Answer is for me
				if (old != null)
					return; // we already processed the answer.
				ulong srcode = SendResult.GetUniqueIdentifier(answer.PacketId, targetUser);
				SendResult res;
				lock (sendResults) {
					res = (SendResult)sendResults[srcode];
					if (res != null)
						sendResults.Remove(srcode);
				}
				if (res != null)
					res.ReceiveAnswer(answer);
			}
		}
		#endregion
		
		#region EncryptionStrategy
		EncryptionStrategy encryptionStrategy = new EncryptionStrategy();
		
		/// <summary>
		/// Gets the encryption strategy used by the network.
		/// </summary>
		public EncryptionStrategy EncryptionStrategy {
			get {
				return encryptionStrategy;
			}
		}
		#endregion
	}
}
