// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Globalization;
using System.Net;
using System.IO;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.Server.Http;
using log4net;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Handles HTTP requests to /system/network/dgnetwork/.
	/// </summary>
	public class DgnRequestHandler : IRequestHandler
	{
		static readonly ILog log = LogManager.GetLogger(typeof(DgnRequestHandler));
		
		ReceivingUserCollection receivingUsers = new ReceivingUserCollection();
		
		/// <summary>
		/// Gets the <see cref="ReceivingUserCollection"/> containing the
		/// <see cref="ReceivingUser"/>s that can accept connections via this
		/// RequestHandler.
		/// </summary>
		public ReceivingUserCollection ReceivingUsers {
			get { return receivingUsers; }
		}
		
		NetClient lastNC;
		
		/// <summary>
		/// Gets the NetClient created recently. This method should only be used
		/// for special purposes such as unit tests, not for real code where net clients
		/// are accepted in non-deterministic order.
		/// </summary>
		public NetClient LastNetClient {
			get { return lastNC; }
		}
		
		/// <summary>
		/// Handles the specified request.
		/// </summary>
		public void Handle(HttpRequest request)
		{
			string file = request.File.Substring(DgnPath.Length);
			if (file == "connect") {
				string host = request.Headers["Host"];
				int pos = host.IndexOf('@');
				if (pos <= 0)
					throw new HttpException(HttpStatusCode.BadRequest);
				int userid = GetInt(host.Substring(0, pos));
				string network = Helper.Unescape(host.Substring(pos + 1));
				DgnAddress address = GetAddress(userid, network);
				if (address == null)
					throw new HttpException(HttpStatusCode.NotFound, "The requested user/network was not found.");
				log.Debug("Accepting DGN connection...");
				NetClient nc = new NetClient(address, request);
				lastNC = nc;
				if (nc.Status != NetClientStatus.Dead) {
					if (address.Network.AcceptConnection(nc)) {
						nc.ReceiveClientResponse();
						if (nc.Status != NetClientStatus.Dead) {
							nc.ReceiveLoop();
						}
					}
				}
				request.Connection.Close();
			} else {
				throw new HttpException(HttpStatusCode.NotFound);
			}
		}
		
		DgnAddress GetAddress(int userid, string network)
		{
			DgnAddress a;
			foreach (ReceivingUser r in ReceivingUsers) {
				foreach (P2PAddress adr in r.Addresses) {
					if (!(adr is DgnAddress)) continue;
					a = (DgnAddress)adr;
					if (a.Network.Identifier == network && a.UserID == userid)
						return a;
				}
			}
			return null;
		}
		
		internal static int GetInt(string text)
		{
			try {
				return int.Parse(text, CultureInfo.InvariantCulture);
			} catch (Exception e) {
				if (Helper.IsCritical(e)) throw;
				throw new HttpException(HttpStatusCode.BadRequest, e);
			}
		}
		
		/// <summary>
		/// The Path the DG Network requests go to.
		/// This is <see cref="Connection.RequestPath"/> + "dgnetwork/";
		/// </summary>
		public const string DgnPath = Connection.RequestPath + "dgnetwork/";
		
		#region Priority
		/// <summary>
		/// Gets the priority this ConnectionHandler has for the given request.
		/// The ConnectionHandler that returns the highest priority will handle the
		/// request. You can set DGNRequestHandler's priority using <see cref="Priority"/>
		/// GetPriority will return that priority if the request goes to
		/// <see cref="DgnPath"/>, otherwise it returns Int32.MinValue.
		/// </summary>
		public int GetPriority(HttpRequest request)
		{
			if (request.File.StartsWith(DgnPath))
				return 1;
			else
				return Int32.MinValue;
		}
		
		int priority = 1;
		
		/// <summary>
		/// Gets/Sets the priority the connection handler should return if the
		/// given request is an DGN request
		/// The default value is 1.
		/// </summary>
		public int Priority {
			get { return priority; }
			set { priority = value; }
		}
		#endregion
	}
}
