// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Security.Cryptography;
using Grunwald.P2PLib.Network;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Helper class for packet encryption/decryption
	/// </summary>
	public class EncryptionStrategy
	{
		DgnP2PNetwork network;
		
		/// <summary>
		/// Gets/Sets the network used to send the key exchange packets.
		/// </summary>
		public DgnP2PNetwork Network {
			get {
				return network;
			}
			set {
				network = value;
			}
		}
		
		#region IdentityCache
		Hashtable identityCache = new Hashtable();
		
		/// <summary>
		/// Gets the identity of the user with the specified id.
		/// Return null if the identity is unknown.
		/// </summary>
		public virtual Identity GetIdentity(int userid)
		{
			lock (identityCache) {
				return (Identity)identityCache[userid];
			}
		}
		
		/// <summary>
		/// Gets the identity of the user with the specified id.
		/// If the identity does not exist, a new identity is created using the specified
		/// publiy key.
		/// </summary>
		public virtual Identity GetIdentity(int userid, string publicKey)
		{
			if (publicKey == null)
				throw new ArgumentNullException("publicKey");
			lock (identityCache) {
				Identity iden = (Identity)identityCache[userid];
				if (iden == null) {
					iden = new Identity(publicKey);
					identityCache.Add(userid, iden);
				}
				return iden;
			}
		}
		#endregion
		
		#region Decryption
		/// <summary>
		/// Decrypts the packet body.
		/// </summary>
		/// <param name="data">The byte array containing the encrypted data.</param>
		/// <param name="offset">The offset where the encrypted data starts.</param>
		/// <param name="length">The length of the encrypted data.</param>
		/// <param name="header">The header of the packet.</param>
		public virtual Stream Decrypt(byte[] data, int offset, int length, PacketHeader header)
		{
			MemoryStream ms = new MemoryStream(data, offset, length, false);
			ms.Position = 0;
			switch (header.EncryptionType) {
				case 0:
					return ms;
				case 1:
					return CreateAESDecryptionStream(ms, header);
				default:
					ms.Close();
					throw new ProtocolViolationException("Invalid encryption type.");
			}
		}
		
		private Stream CreateAESDecryptionStream(Stream targetStream, PacketHeader header)
		{
			int uid = header.Source;
			Identity ident = GetIdentity(uid);
			if (ident == null) {
				throw new P2PLibException("Identity not found, make sure encryption has been set up before using it -- TODO: auto get Identity");
			}
			byte[] keyHash = DataHelper.ReadBytes(targetStream, ident.KeyHashLength);
			byte[] fullkey = ident.GetSymmetricKey(keyHash);
			if (fullkey == null) {
				throw new P2PLibException("Unknown key -- TODO: auto get encryption key");
			}
			RijndaelManaged algo = new RijndaelManaged();
			byte[] key = new byte[32];
			byte[] iv = new byte[fullkey.Length - 32];
			Array.Copy(fullkey, key, 32);
			Array.Copy(fullkey, 32, iv, 0, iv.Length);
			ICryptoTransform decryptor = algo.CreateDecryptor(key, iv);
			return new CryptoStream(targetStream, decryptor, CryptoStreamMode.Read);
		}
		#endregion
		
		#region Encryption
		/// <summary>
		/// Creates a CryptoStream that writes to the target stream.
		/// </summary>
		public virtual Stream CreateEncryptionStream(Stream targetStream, PacketHeader header)
		{
			switch (header.EncryptionType) {
				case 0:
					return targetStream;
				case 1:
					return CreateAESStream(targetStream, header);
				default:
					throw new ProtocolViolationException("Invalid encryption type.");
			}
		}
		
		private Stream CreateAESStream(Stream targetStream, PacketHeader header)
		{
			if (header.Destination.Length != 1)
				throw new P2PLibException("Encryption can only be used when there is just one destination user.");
			int uid = header.Destination[0];
			Identity ident = GetIdentity(uid);
			if (ident == null)
				throw new P2PLibException("Identity not found, make sure encryption has been set up before using it");
			byte[] fullkey = ident.GetSymmetricKey();
			RijndaelManaged algo = new RijndaelManaged();
			byte[] key = new byte[32];
			byte[] iv = new byte[fullkey.Length - 32];
			Array.Copy(fullkey, key, 32);
			Array.Copy(fullkey, 32, iv, 0, iv.Length);
			ICryptoTransform encryptor = algo.CreateEncryptor(key, iv);
			byte[] keyHash = ident.GetSymmetricKeyHash(fullkey);
			targetStream.Write(keyHash, 0, keyHash.Length);
			return new CryptoStream(targetStream, encryptor, CryptoStreamMode.Write);
		}
		
		/// <summary>
		/// Finishes encrypting data into the cryptoStream.
		/// </summary>
		public virtual void FinishEncryption(Stream cryptoStream, PacketHeader header)
		{
			CryptoStream cs = cryptoStream as CryptoStream;
			if (cs != null)
				cs.FlushFinalBlock();
		}
		#endregion
		
		#region Key exchanging
		/// <summary>
		/// Gets whether packets to the specified user can be encrypted.
		/// </summary>
		public virtual bool CanEncryptPackets(int userid)
		{
			Identity ident = this.GetIdentity(userid);
			if (ident == null)
				return false;
			else
				return ident.KeyCount > 0;
		}
		
		/*
		/// <summary>
		/// Exchanges an encryption key with the specified user so packets can be
		/// encrypted.
		/// </summary>
		public virtual void SendEncryptionKey(Identity myIdentity, int myUserId, Identity targetIdentity, int targetUserId)
		{
			if (myIdentity == null)
				throw new ArgumentNullException("myIdentity");
			if (targetIdentity == null)
				throw new ArgumentNullException("targetIdentity");
			if (targetIdentity.KeyCount > 0)
				return;
			byte[] key = targetIdentity.GetSymmetricKey(); // create an encryption key
			Packet packet = new Packet();
			packet.Header.Source = myUserId;
			packet.Header.Destination = new int[] { targetUserId };
			packet.Target = "CreateEncryptionKey";
			packet.Values.AddValue(targetIdentity.Encrypt(key));
			packet.Values.AddValue(myIdentity.Sign(key));
			packet.Header.PacketID = 0;
			packet.AnswerId = 0;
			network.SendPacket(packet);
		}
		 */
		
		ListDictionary requestingIdentities = new ListDictionary(); // int -> DateTime
		
		/// <summary>
		/// Gets the identity from cache if possible, otherwise it starts a request,
		/// but waits a maximum of <paramref name="timeout"/> milliseconds for the result.
		/// In case of any error the method returns null.
		/// </summary>
		public Identity RequestIdentity(int userid, int timeout)
		{
			Identity iden;
			lock (requestingIdentities) {
				iden = this.GetIdentity(userid);
				if (iden != null)
					return iden;
				if (requestingIdentities.Contains(userid)) {
					DateTime lastTry = (DateTime)requestingIdentities[userid];
					if (DateTime.Now.Subtract(lastTry).TotalSeconds < 30)
						return null;
					else
						requestingIdentities[userid] = DateTime.Now;
				} else {
					requestingIdentities.Add(userid, DateTime.Now);
				}
			}
			Packet packet = new Packet();
			packet.Header.Source = network.UserID;
			packet.Header.Destination = new int[] { userid };
			packet.Target = "GetIdentity";
			if (network == null)
				throw new InvalidOperationException("Sending identity request is only possible when network-property is set!");
			SendResult res = network.SendPacket(packet);
			if (res == null)
				return null;
			res.CallOnAnswer(new AnswerEventHandler(IdentityResponse));
			if (res.WaitForAnswer(timeout))
				return GetIdentity(userid);
			else
				return null;
		}
		
		private void IdentityResponse(object sender, AnswerEventArgs e)
		{
			SendResult res = (SendResult)sender;
			if (e.Answer.ErrorCode == 0 && e.Answer.Values.Count > 0) {
				lock (requestingIdentities) {
					string key = (string)e.Answer.Values.GetValue(0, typeof(string));
					GetIdentity(res.Source, key);
					requestingIdentities.Remove(res.Source);
				}
			}
		}
		
		/// <summary>
		/// Responds to the GetIdentity request <paramref name="p"/> by setting the properties
		/// of <paramref name="answer"/>.
		/// </summary>
		public virtual void RespondToGetIdentity(Packet p, Answer answer)
		{
			if (network == null)
				throw new InvalidOperationException("Calling RespondToGetIdentity is only possible when network-property is set!");
			UserAddress adr = new UserAddress(network, p.Header.Source);
			foreach (ReceivingUser user in network.ReceivingUsers) {
				if (user.Addresses.Contains(adr)) {
					if (user.Identity == null)
						answer.ErrorCode = 1;
					else
						answer.Values.AddValue(user.Identity.PublicKey);
					return;
				}
			}
			answer.ErrorCode = Answer.ErrorCodeInternalError;
		}
		#endregion
	}
}
