// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Interface for the routing strategy.
	/// </summary>
	public interface IRoutingStrategy : IDisposable
	{
		/// <summary>
		/// Adds a new NetClient to this <see cref="IRoutingStrategy">RoutingStrategy</see>.
		/// The NetClient is automatically removed when it's connection is closed.
		/// </summary>
		void AddNetClient(NetClient netClient);
		
		/// <summary>
		/// Receives a routing table from the specified <see cref="NetClient"/>.
		/// The entries in the table are merged into the main table, and the changelog
		/// is written so that the entries can be forwarded to other partners.
		/// </summary>
		/// <param name="netClient">The <see cref="NetClient"/> that has received the
		/// routing table.</param>
		/// <param name="data">The byte array containing the data of the routing table.</param>
		/// <exception cref="ProtocolViolationException">The routing table was invalid.</exception>
		void Receive(NetClient netClient, byte[] data);
		
		/// <summary>
		/// Sends the RoutingTable to the specified NetClient.
		/// </summary>
		/// <param name="netClient">The NetClient to send the routing table to.</param>
		/// <param name="sendFullTable">
		/// Specifies if the full table should be resent or just the changes should be sent.
		/// </param>
		void SendRoutingTable(NetClient netClient, bool sendFullTable);
		
		/// <summary>
		/// Gets the distance from this computer to the specified user id.
		/// If the user is not found, the method returns -1.
		/// </summary>
		int GetDistance(int userid);
		
		/// <summary>
		/// Find best route and send packet.
		/// </summary>
		/// <param name="header">The header containing the information where to send the packet to.</param>
		/// <param name="data">The array containing the packet body.</param>
		/// <param name="offset">The offset where in the array the packet body starts.</param>
		/// <param name="length">The length of the packet body.</param>
		void SendPacket(PacketHeader header, byte[] data, int offset, int length);
		
		/// <summary>
		/// Find next node and send the answer. The method returns true if the current
		/// computer is the target for the answer.
		/// </summary>
		bool SendAnswer(Answer answer);
	}
}
