// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections.Specialized;
using System.IO;
using System.Net;
using System.Text;
using System.Security.Cryptography;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;
using log4net;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Manages a connection to the network.
	/// </summary>
	public class NetClient : Connection, IBufferedOutput
	{
		#region Status
		NetClientStatus status = NetClientStatus.Connecting;
		
		/// <summary>
		/// Gets if this connection is ready.
		/// </summary>
		public override bool Ready {
			get {
				return status == NetClientStatus.Ready;
			}
		}
		
		/// <summary>
		/// Gets the current state of this NetClient.
		/// </summary>
		public NetClientStatus Status {
			get {
				return status;
			}
		}
		
		/// <summary>
		/// Is raised whenever the value of the <see cref="Status">Status property</see> is changed.
		/// </summary>
		public event EventHandler StatusChanged;
		
		/// <summary>
		/// Sets the status property.
		/// </summary>
		protected void SetStatus(NetClientStatus status) {
			if (this.status == status)
				return;
			this.status = status;
			if (StatusChanged != null)
				StatusChanged(this, EventArgs.Empty);
		}
		#endregion
		
		#region Properties
		static readonly ILog log = LogManager.GetLogger(typeof (NetClient));
		
		IConnection connection;
		
		/// <summary>
		/// Gets if this NetClient is still connected with a socket.
		/// </summary>
		public override bool Connected {
			get { return connection.Connected; }
		}
		
		bool outgoing;
		
		/// <summary>
		/// Gets if this connection is outgoing (true) or incoming (false).
		/// </summary>
		public bool Outgoing {
			get { return outgoing; }
		}
		
		/// <summary>
		/// Gets the network this NetClient is connected with.
		/// </summary>
		public DgnP2PNetwork Network {
			get { return (DgnP2PNetwork)Partner.Network; }
		}
		
		/// <summary>
		/// Gets the address of the partner of this connection.
		/// </summary>
		public new DgnAddress Partner {
			get { return base.Partner as DgnAddress; }
		}
		
		static string userAgent;
		
		/// <summary>
		/// Gets/Sets the user agent to use.
		/// </summary>
		public static string UserAgent {
			get { return userAgent; }
			set { userAgent = value; }
		}
		#endregion
		
		#region ConnectKey and Seed
		string mySeed = Helper.Random.Next().ToString("d");
		
		string GetConnectKey(string seed) {
			string data = Network.ConnectKey + seed;
			byte[] md5;
			using (MD5CryptoServiceProvider md5Provider = new MD5CryptoServiceProvider()) {
				md5 = md5Provider.ComputeHash(Encoding.UTF8.GetBytes(data));
			}
			return Convert.ToBase64String(md5);
		}
		#endregion
		
		#region Constructor
		#region New Accepted Connection
		/// <summary>
		/// Creates a netclient using a <see cref="HttpRequest"/>.
		/// </summary>
		/// <param name="ownAddress">The <see cref="DgnAddress">address</see>
		/// of the owner of this <see cref="Connection"/></param>
		/// <param name="request">The <see cref="HttpRequest"/> containing
		/// a DGN connection request.</param>
		public NetClient(DgnAddress ownAddress, HttpRequest request)
			: base(CreatePartner(ownAddress, request))
		{
			outgoing = false;
			connection = request.Connection;
			StreamWriter w = connection.Writer;
			DgnP2PNetwork network = ownAddress.Network;
			try {
				w.Write("DGN/0.1 200 OK\r\n");
				w.Write("MyIdentity: {0}\r\n", ownAddress.Identity.PublicKey);
				w.Write("ConnectKey: {0}\r\n", GetConnectKey(request.Headers["Seed"]));
				w.Write("Seed: {0}\r\n", mySeed);
				if (userAgent != null)
					w.Write("User-Agent: {0}\r\n", userAgent);
				w.Write("\r\n");
				w.Flush();
			} catch (ObjectDisposedException) {
				DoClose(new ClosedEventArgs(true, true, "Connection closed unexpectedly."));
			} catch (IOException) {
				DoClose(new ClosedEventArgs(true, true, "Connection closed unexpectedly."));
			}
		}
		
		static UserAddress CreatePartner(DgnAddress ownAddress, HttpRequest request)
		{
			if (ownAddress == null) throw new ArgumentNullException("ownAddress");
			if (request == null) throw new ArgumentNullException("request");
			int partnerID = DgnRequestHandler.GetInt(request.Headers["Me"]);
			if (partnerID == ownAddress.UserID)
				throw new HttpException();
			Identity identity = new Identity(request.Headers["MyIdentity"]);
			return new UserAddress(ownAddress.Network, partnerID, identity);
		}
		#endregion
		
		#region Create new connection
		/// <summary>
		/// Creates a new NetClient using an outgoing connection to send a request.
		/// The connection request is send to the connections's
		/// <see cref="IConnection.Writer"/> and <see cref="StreamWriter.Flush"/>()
		/// is called.
		/// </summary>
		/// <param name="connection">The connection that should be used by this NetClient.</param>
		/// <param name="partner">The address of the partner of this connection.</param>
		/// <param name="headers">The headers of the connect request that should be sent.</param>
		public NetClient(IConnection connection, DgnAddress partner, NameValueCollection headers)
			: base(partner)
		{
			if (connection == null)
				throw new ArgumentNullException("connection");
			if (headers == null)
				throw new ArgumentNullException("headers");
			if (headers["Host"] != null)
				throw new ArgumentException("headers must not contain the host-field, it is added automatically.");
			if (headers["Seed"] != null)
				throw new ArgumentException("headers must not contain the seed-field, it is added automatically.");
			if (!connection.Connected)
				throw new ArgumentException("The connection is not connected.");
			outgoing = true;
			this.connection = connection;
			StreamWriter w = connection.Writer;
			try {
				w.Write("GET {0}connect HTTP/1.1\r\n", DgnRequestHandler.DgnPath);
				w.Write("Host: {0:d}@{1}\r\n", partner.UserID, partner.Network.Identifier);
				w.Write("Seed: {0}\r\n", mySeed);
				if (userAgent != null)
					w.Write("User-Agent: {0}\r\n", userAgent);
				for (int i = 0; i < headers.Count; i ++) {
					w.Write("{0}: {1}\r\n", headers.GetKey(i), headers[i]);
				}
				w.Write("\r\n");
				w.Flush();
			} catch (ObjectDisposedException) {
				DoClose(new ClosedEventArgs(true, true, "Connection closed unexpectedly."));
			} catch (IOException) {
				DoClose(new ClosedEventArgs(true, true, "Connection closed unexpectedly."));
			}
		}
		#endregion
		#endregion
		
		#region Work on server response
		/// <summary>
		/// Receives the server's response and send the client's response.
		/// </summary>
		/// <exception cref="HttpException">Is thrown if the invalid response was invalid.</exception>
		public void ReceiveServerResponse()
		{
			if (!Outgoing)
				throw new InvalidOperationException("ReceiveServerResponse is invalid for incoming connections.");
			if (status != NetClientStatus.Connecting)
				throw new InvalidOperationException("ReceiveServerResponse was already called.");
			SetStatus(NetClientStatus.WaitForServerResponse);
			try {
				StringCollection response = HttpLikeServer.GetRequest(connection);
				if (response == null)
					throw new HttpException();
				WorkOnServerResponse(response);
			} catch (HttpException ex) {
				log.Debug(ex);
				StreamWriter w = connection.Writer;
				try {
					w.Write("DGN/0.1 ");
					w.Write(ex.StatusCode.ToString("d"));
					w.Write(ex.Message);
					w.Write("\r\n\r\n");
					w.Flush();
				} catch (ObjectDisposedException) {
				} catch (IOException) {
					// ignore
				}
				DoClose(new ClosedEventArgs(false, true, ex.Message));
			} catch (Exception ex) {
				if (Helper.IsCritical(ex))
					throw;
				log.Fatal(ex);
				StreamWriter w = connection.Writer;
				try {
					w.Write("DGN/0.1 500 Internal server error\r\n\r\n");
					w.Flush();
				} catch (IOException) {
					// ignore
				}
				DoClose(new ClosedEventArgs(false, true, ex.Message));
			}
		}
		
		private void WorkOnServerResponse(StringCollection response)
		{
			StreamWriter w = connection.Writer;
			string head = response[0];
			if (head != "DGN/0.1 200 OK") {
				StringBuilder b = new StringBuilder();
				foreach(string line in response) {
					b.Append(line);
					b.Append('\n');
				}
				log.Debug(b.ToString());
				if (!head.StartsWith("DGN/0.1 5")) {
					DgnP2PNetwork network = Network;
					lock (network) {
						network.HostCache.Remove(Partner);
					}
				}
				DoClose(new ClosedEventArgs(true, true, b.ToString()));
				return;
			}
			WebHeaderCollection header = new WebHeaderCollection();
			try {
				for (int i = 1; i < response.Count; ++i) {
					header.Add(response[i]);
				}
			} catch (Exception ex) {
				if (Helper.IsCritical(ex)) throw;
				throw new HttpException();
			}
			if (header["ConnectKey"] != GetConnectKey(mySeed))
				throw new HttpException(HttpStatusCode.Forbidden, "ConnectKey is invalid.");
			try {
				w.Write("DGN/0.1 200 Client OK\r\n");
				w.Write("ConnectKey: {0}\r\n", GetConnectKey(header["Seed"]));
				w.Write("\r\n");
				w.Flush();
			} catch (ObjectDisposedException) {
				DoClose(new ClosedEventArgs(true, true, "Connection closed unexpectedly."));
				return;
			} catch (IOException) {
				DoClose(new ClosedEventArgs(true, true, "Connection closed unexpectedly."));
				return;
			}
			SetStatus(NetClientStatus.ReadyForLoop);
		}
		#endregion
		
		#region Work on client response
		/// <summary>
		/// Receives the client's response.
		/// </summary>
		public void ReceiveClientResponse()
		{
			if (Outgoing)
				throw new InvalidOperationException("ReceiveClientResponse is invalid for outgoing connections.");
			if (status != NetClientStatus.Connecting)
				throw new InvalidOperationException("ReceiveClientResponse was already called.");
			SetStatus(NetClientStatus.WaitForClientResponse);
			StringCollection request = HttpLikeServer.GetRequest(connection);
			if (request == null || request.Count == 0) {
				DoClose(new ClosedEventArgs(true, true, "Cannot read response."));
				return;
			}
			string head = request[0];
			if (head != "DGN/0.1 200 Client OK") {
				DoClose(new ClosedEventArgs(true, true, "The client didn't return a valid header (" + head + ")."));
				return;
			}
			WebHeaderCollection header = new WebHeaderCollection();
			for (int i = 1; i < request.Count; ++i) {
				header.Add(request[i]);
			}
			SetStatus(NetClientStatus.ReadyForLoop);
		}
		#endregion
		
		#region Start Receiving
		bool running = false;
		/// <summary>
		/// Executes the receiving loop of this NetClient.
		/// </summary>
		public void ReceiveLoop()
		{
			if (running)
				throw new InvalidOperationException("ReceiveLoop is already running");
			if (status != NetClientStatus.ReadyForLoop)
				throw new InvalidOperationException("The connection is not ready.");
			try {
				running = true;
				SetStatus(NetClientStatus.Ready);
				try {
					this.Network.RoutingStrategy.AddNetClient(this);
				} catch (IOException) {
					if (Connected)
						throw;
				}
				Receive();
				DoClose(new ClosedEventArgs(true, false, "Connection was lost"));
			} catch (ProtocolViolationException ex) {
				log.Debug(ex);
				Close(ex.Message);
				return;
			} catch (Exception ex) {
				log.Fatal(ex);
				DoClose(new ClosedEventArgs(false, false, "Exception occured"));
				throw;
			} finally {
				running = false;
			}
		}
		#endregion
		
		#region ReceiveLoop
		/// <summary>
		/// Occurs when the NetClient receives a data block.
		/// If no listener sets the
		/// <see cref="ReceivedEventArgs.Handled">Handled property</see>,
		/// the connection will be closed.
		/// </summary>
		public event ReceivedEventHandler Received;
		
		private void Receive()
		{
			while (Connected) {
				try {
					byte[] fieldLength = ReceiveBytes(8);
					if (fieldLength == null) break;
					DataBlockType type = (DataBlockType)NetworkBitConverter.ToInt32(fieldLength, 0);
					int length = NetworkBitConverter.ToInt32(fieldLength, 4);
					
					byte[] data = ReceiveBytes(length);
					if (data == null) break;
					OnReceived(new ReceivedEventArgs(type, data));
				} catch (IOException) {
					if (Connected)
						throw;
					else
						break;
				}
			}
		}
		
		private byte[] ReceiveBytes(int count)
		{
			if (count > 102400)
				throw new ProtocolViolationException("Exceeded 100 KB memory limit.");
			if (count < 0)
				throw new ProtocolViolationException("Negative data length is illegal.");
			byte[] b = new byte[count];
			// TODO: Warning! Read is pinning newly created object for long time!
			int pos = 0;
			int c;
			Stream s = connection.Stream;
			while (pos < count) {
				c = s.Read(b, pos, count - pos);
				if (c == 0) return null;
				pos += c;
			}
			return b;
		}
		
		/// <summary>
		/// <para>Raises the <see cref="Received">Received event</see> for the data
		/// block in the specified <see cref="ReceivedEventArgs"/>.
		/// If the data block is a <see cref="DataBlockType.RoutingTable">routing table</see>,
		/// it will be loaded.</para>
		/// <para>If no listener of the <see cref="Received">Received event</see>
		/// sets the <see cref="ReceivedEventArgs.Handled">Handled property</see>,
		/// the connection will be closed.</para>
		/// </summary>
		protected virtual void OnReceived(ReceivedEventArgs e)
		{
			if (Received != null)
				Received(this, e);
			if (e.Handled)
				return;
			switch (e.DataBlockType) {
				case DataBlockType.RoutingTable:
					this.Network.RoutingStrategy.Receive(this, e.Data);
					//rt.Receive(e.Data);
					break;
				case DataBlockType.Quit:
					e.Handled = true;
					int size = e.Data.Length;
					DoClose(new ClosedEventArgs(true, true, Encoding.UTF8.GetString(e.Data)));
					break;
				case DataBlockType.SendPacket:
					ReceivePacket(e.Data);
					break;
				case DataBlockType.Answer:
					this.Network.ReceiveAnswer(new Answer(e.Data));
					break;
				default:
					this.Close("Unknown data block type");
					break;
			}
		}
		
		private void ReceivePacket(byte[] data)
		{
			int index = 0;
			PacketHeader header = new PacketHeader(data, ref index);
			this.Network.RoutingStrategy.SendPacket(header, data, index, data.Length - index);
		}
		#endregion
		
		#region Close
		/// <summary>
		/// Closes the connection without specifying a reason.
		/// </summary>
		public override void Close()
		{
			Close("Connection closed.");
		}
		
		/// <summary>
		/// Closes the connection with specifying a reason.
		/// </summary>
		public override void Close(string message)
		{
			allowClose = false;
			try {
				if (this.Ready)
					Send(DataBlockType.Quit, Encoding.UTF8.GetBytes(message));
			} catch {
				// ignore errors sending the quit block
			}
			allowClose = true;
			DoClose(new ClosedEventArgs(false, true, message));
		}
		
		bool allowClose = true;
		
		private void DoClose(ClosedEventArgs e)
		{
			SetStatus(NetClientStatus.Dead);
			if (!allowClose) return;
			allowClose = false;
			if (connection.Connected)
				connection.Close();
			OnClosed(e);
		}
		#endregion
		
		#region Asynchronous send
		private delegate void void_DataBlockType_byteArray(DataBlockType dataBlockType, byte[] data);
		
		/// <summary>Sends the specified data block over the connection.</summary>
		/// <param name="dataBlockType">The type of the data block to send.</param>
		/// <param name="data">The actual data.</param>
		public void BeginSend(DataBlockType dataBlockType, byte[] data)
		{
			lock(sendLock) {
				if (buffers > 0) {
					DoSend(dataBlockType, data);
					return;
				}
			}
			new void_DataBlockType_byteArray(StartSend).BeginInvoke(dataBlockType, data, null, null);
		}
		
		private void StartSend(DataBlockType dataBlockType, byte[] data)
		{
			try {
				Send(dataBlockType, data);
			} catch (ObjectDisposedException) {
				// Tcp connection is already closed, so fail quietly.
			} catch (IOException) {
				// Tcp connection is already closed, so fail quietly.
			} catch (Exception ex) {
				log.Fatal(ex);
			}
		}
		#endregion
		
		#region Send
		object sendLock = new object();
		
		/// <summary>Sends the specified data block over the connection.</summary>
		/// <param name="dataBlockType">The type of the data block to send.</param>
		/// <param name="data">The actual data.</param>
		public void Send(DataBlockType dataBlockType, byte[] data)
		{
			lock(sendLock) {
				DoSend(dataBlockType, data);
			}
		}
		
		private void DoSend(DataBlockType dataBlockType, byte[] data)
		{
			byte[] type = NetworkBitConverter.GetBytes((Int32)dataBlockType);
			byte[] length;
			if (data != null)
				length = NetworkBitConverter.GetBytes((Int32)data.Length);
			else
				length = null;
			if (buffers > 0) {
				sendBuffer.Write(type,   0, type.Length);
				if (data != null) {
					sendBuffer.Write(length, 0, length.Length);
					sendBuffer.Write(data  , 0, data.Length);
				}
			} else {
				byte[] sendData = new byte[((data == null) ? 0 : data.Length) + 8];
				type.CopyTo(sendData, 0);
				if (data != null) {
					length.CopyTo(sendData, 4);
					data.CopyTo(sendData, 8);
				}
				connection.Stream.Write(sendData, 0, sendData.Length);
				// Flush is necessary for the PartnerStream and has no effect on NetworkStream.
				connection.Stream.Flush();
			}
		}
		#endregion
		
		#region Buffered Send
		int buffers;
		
		MemoryStream sendBuffer;
		
		/// <summary>
		/// <para>Increases the count of send buffers by one.</para>
		/// <para>This makes the send method redirect it's output into the buffer
		/// instead of the stream.</para>
		/// <para>Use this method to send multiple data blocks in one TCP packet.
		/// When you are finished sending the blocks, call EndBuffer to send
		/// them all at once.</para>
		/// <para>If you fail to call EndBuffer (e.g. exception), this NetClient
		/// won't ever send any data again but cache everything!</para>
		/// </summary>
		public void StartBuffer()
		{
			lock(sendLock) {
				if (sendBuffer == null)
					sendBuffer = new MemoryStream();
				buffers += 1;
			}
		}
		
		/// <summary>
		/// <para>Decreaes the count of send buffers by one.</para>
		/// <para>If the count reaches zero, the current buffer content is flushed
		/// to the network stream.</para>
		/// </summary>
		public void EndBuffer()
		{
			lock(sendLock) {
				if (buffers == 0)
					throw new InvalidOperationException("There is no open buffer.");
				buffers -= 1;
				if (sendBuffer.Length > 0) {
					sendBuffer.WriteTo(connection.Stream);
				}
				sendBuffer.Close();
				sendBuffer = null;
			}
		}
		
		/// <summary>
		/// Flushes the current buffer content without closing it.
		/// </summary>
		public void FlushBuffer()
		{
			lock(sendLock) {
				if (buffers > 0) {
					if (sendBuffer.Length > 0) {
						sendBuffer.WriteTo(connection.Stream);
						sendBuffer.Close();
						sendBuffer = new MemoryStream();
					}
				}
			}
		}
		#endregion
		
		/// <summary>
		/// Gets the string representation of this connection.
		/// </summary>
		public override string ToString()
		{
			return String.Format("[{0}: {3}, MyUser = {4}, Partner = {1}, Status = {2}]",
			                     GetType().Name, Partner, this.Status,
			                     this.outgoing ? "Outgoing" : "Incoming", Network.UserID);
		}
	}
}
