// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Text;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// A packet transmitted via DG Network.
	/// </summary>
	public class Packet : ICloneable
	{
		#region Fields
		PacketHeader header;
		
		int answerId;
		string target;
		
		IndexedValueList values = new IndexedValueList();
		NamedValueList namedValues = new NamedValueList();
		#endregion
		
		#region Properties
		/// <summary>
		/// Gets/Sets the header of this packet.
		/// </summary>
		public PacketHeader Header {
			get {
				return header;
			}
			set {
				if (value == null)
					throw new ArgumentNullException();
				header = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the answer id needed to acknowlegde receipt of this packet.
		/// </summary>
		public int AnswerId {
			get {
				return answerId;
			}
			set {
				answerId = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the string specifying what target on the receiving computer should
		/// process the packet.
		/// This can be used as some kind of port information.
		/// </summary>
		public string Target {
			get {
				return target;
			}
			set {
				target = value;
			}
		}
		
		/// <summary>
		/// Gets the list of named values.
		/// </summary>
		public NamedValueList NamedValues {
			get {
				return namedValues;
			}
		}
		
		/// <summary>
		/// Gets the list of indexed values.
		/// </summary>
		public IndexedValueList Values {
			get {
				return values;
			}
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates a new empty packet.
		/// </summary>
		public Packet()
		{
			header = new PacketHeader();
			answerId = Helper.GetRandomInt();
		}
		#endregion
		
		#region ToByte
		/// <summary>
		/// Writes the byte representation of this packet to the specified stream.
		/// </summary>
		public void WriteTo(Stream stream, EncryptionStrategy encryption)
		{
			byte[] headerBytes = header.ToByte();
			stream.Write(headerBytes, 0, headerBytes.Length);
			Stream s = encryption.CreateEncryptionStream(stream, header);
			DataHelper.Write(s, NetworkBitConverter.GetBytes(answerId));
			DataHelper.Write(s, target);
			values.Save(s);
			namedValues.Save(s);
			s.Close();
			encryption.FinishEncryption(s, header);
		}
		
		/// <summary>
		/// Gets a byte array with the byte representation of the packet.
		/// </summary>
		public byte[] ToByte(EncryptionStrategy encryption)
		{
			MemoryStream ms = new MemoryStream();
			WriteTo(ms, encryption);
			return ms.ToArray();
		}
		
		/// <summary>
		/// Gets the data bytes of the packet that should be encrypted.
		/// </summary>
		private int GetDataByteCount()
		{
			int count = 5; // answerID + length(target)
			if (target != null)
				count += Encoding.UTF8.GetByteCount(target);
			count += values.GetByteLength();
			count += namedValues.GetByteLength();
			return count;
		}
		#endregion
		
		#region FromByte
		/// <summary>
		/// Restores the packet from the byte array.
		/// </summary>
		public Packet(byte[] data, EncryptionStrategy encryption)
			: this(data, 0, data.Length, encryption)
		{
		}
		
		/// <summary>
		/// Restores the packet from the byte array.
		/// </summary>
		public Packet(byte[] data, int index, int length, EncryptionStrategy encryption)
		{
			int pos = index;
			header = new PacketHeader(data, ref pos);
			FromByte(data, pos, length - pos + index, encryption);
		}
		
		/// <summary>
		/// Restores the packet body from the byte array while using the existing header.
		/// </summary>
		public Packet(PacketHeader header, byte[] data, int index, int length, EncryptionStrategy encryption)
		{
			this.header = header;
			FromByte(data, index, length, encryption);
		}
		
		private void FromByte(byte[] data, int index, int length, EncryptionStrategy encryption)
		{
			int pos = index;
			Stream s = encryption.Decrypt(data, pos, length, header);
			answerId = DataHelper.ReadInt(s);
			target   = DataHelper.ReadString(s);
			values.Read(s);
			namedValues.Read(s);
			if (s.ReadByte() != -1)
				throw new ProtocolViolationException("Packets must not have dummy data at the end.");
			s.Close();
		}
		#endregion
		
		#region System.ICloneable interface implementation
		object ICloneable.Clone() {
			return this.Clone();
		}
		
		/// <summary>
		/// Clones the packet.
		/// </summary>
		public Packet Clone() {
			Packet n = new Packet();
			
			n.header       = this.header.Clone();
			
			n.answerId = this.answerId;
			n.target   = this.target;
			n.values   = this.values.Clone();
			n.namedValues = this.namedValues.Clone();
			return n;
		}
		#endregion
		
		#region ToString
		/// <summary>
		/// Gets the string representation of this packet.
		/// </summary>
		public override string ToString() {
			return String.Format("[{0} Header={1} AnswerID={2} Target={3}]",
			                     this.GetType().Name, header, answerId, target);
		}
		#endregion
	}
}
