// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// The header of a packet transmitted via DG Network.
	/// </summary>
	public class PacketHeader : ICloneable
	{
		#region Fields
		int source;
		int[] destination;
		
		byte encryptionType;
		int packetID;
		int dependsOn;
		int[] hops;
		#endregion
		
		#region Properties
		/// <summary>
		/// Gets/Sets the source user id of this packet.
		/// </summary>
		public int Source {
			get {
				return source;
			}
			set {
				source = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the destination user ids of this packet.
		/// </summary>
		public int[] Destination {
			get {
				return destination;
			}
			set {
				destination = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the encryption type.
		/// </summary>
		public byte EncryptionType {
			get {
				return encryptionType;
			}
			set {
				encryptionType = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the packet id.
		/// </summary>
		public int PacketID {
			get {
				return packetID;
			}
			set {
				// Packet id can (and must) be 0 for unsafe packets.
				//if (value == 0)
				//	throw new ArgumentOutOfRangeException("packetID", "The packet ID can't be 0");
				packetID = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the ID of the packet this packet depends on.
		/// </summary>
		public int DependsOn {
			get {
				return dependsOn;
			}
			set {
				dependsOn = value;
			}
		}
		
		/// <summary>
		/// Gets/Sets the Hops array.
		/// </summary>
		public int[] Hops {
			get {
				return hops;
			}
			set {
				if (value == null)
					throw new ArgumentNullException();
				hops = value;
			}
		}
		
		/// <summary>
		/// Inserts the specified user id in the first free field in the hops array.
		/// If the array is full, the method returns false.
		/// </summary>
		public bool AddHop(int userid)
		{
			for (int i = 0; i < hops.Length; ++i) {
				if (hops[i] == 0) {
					hops[i] = userid;
					return true;
				}
			}
			return false;
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates a new packet header.
		/// </summary>
		public PacketHeader()
		{
			packetID = Helper.GetRandomInt();
			hops = new int[10];
			destination = new int[0];
		}
		#endregion
		
		#region Length
		/// <summary>
		/// Gets the length of this packet header in bytes.
		/// </summary>
		public int Length {
			get {
				return 1 + 4 + 1 + FilledLength(Destination) + 1 + 4 + 4 + 1 + hops.Length * 4;
			}
		}
		
		private int FilledLength(int[] array)
		{
			int c = 0;
			for (int i = 0; i < array.Length; ++i) {
				if (array[i] != 0)
					c += 4;
			}
			return c;
		}
		#endregion
		
		#region ToByte
		/// <summary>
		/// Writes the byte representation of this packet header to the specified byte array.
		/// </summary>
		public void ToByte(byte[] data, int index)
		{
			int pos = index;
			data[pos++] = 80; // identification byte
			NetworkBitConverter.PutBytes(data, pos, source);
			pos = ArrayToByte(destination, data, pos + 4);
			data[pos++] = encryptionType;
			NetworkBitConverter.PutBytes(data, pos, packetID);
			pos += 4;
			NetworkBitConverter.PutBytes(data, pos, dependsOn);
			pos += 4;
			data[pos++] = (byte)hops.Length;
			for (int i = 0; i < hops.Length; ++i) {
				NetworkBitConverter.PutBytes(data, pos, hops[i]);
				pos += 4;
			}
		}
		
		private int ArrayToByte(int[] array, byte[] data, int index)
		{
			int pos = index + 1;
			byte c = 0;
			for (int i = 0; i < array.Length; ++i) {
				if (array[i] != 0) {
					NetworkBitConverter.PutBytes(data, pos, array[i]);
					pos += 4;
					c += 1;
				}
			}
			data[index] = c;
			return pos;
		}
		
		/// <summary>
		/// Gets a byte array with the byte representation of the packet header.
		/// </summary>
		public byte[] ToByte()
		{
			byte[] data = new byte[Length];
			ToByte(data, 0);
			return data;
		}
		#endregion
		
		#region FromByte
		/// <summary>
		/// Restores the packet header from the byte array.
		/// </summary>
		public PacketHeader(byte[] data) : this(data, 0)
		{
		}
		
		/// <summary>
		/// Restores the packet header from the byte array.
		/// </summary>
		public PacketHeader(byte[] data, int index) : this(data, ref index)
		{
		}
		
		internal PacketHeader(byte[] data, ref int index)
		{
			int pos = index;
			if (data[pos++] != 80)
				throw new ProtocolViolationException("There is no packet header at the specified index.");
			source = NetworkBitConverter.ToInt32(data, pos);
			pos += 4;
			destination = GetArray(data, ref pos);
			encryptionType = data[pos++];
			packetID = NetworkBitConverter.ToInt32(data, pos);
			pos += 4;
			dependsOn = NetworkBitConverter.ToInt32(data, pos);
			pos += 4;
			hops = GetArray(data, ref pos);
			index = pos;
		}
		
		private int[] GetArray(byte[] data, ref int index)
		{
			int pos = index;
			int length = data[pos];
			pos += 1;
			int[] arr = new int[length];
			for (int i = 0; i < length; ++i) {
				arr[i] = NetworkBitConverter.ToInt32(data, pos);
				pos += 4;
			}
			index = pos;
			return arr;
		}
		#endregion
		
		#region System.ICloneable interface implementation
		object ICloneable.Clone() {
			return this.Clone();
		}
		
		/// <summary>
		/// Clones the packet header.
		/// </summary>
		public PacketHeader Clone() {
			PacketHeader n = new PacketHeader();
			n.source         = this.source;
			n.destination    = this.destination.Clone() as int[];
			n.encryptionType = this.encryptionType;
			n.packetID       = this.packetID;
			n.dependsOn      = this.dependsOn;
			n.hops           = this.hops.Clone() as int[];
			return n;
		}
		#endregion
		
		#region ToString
		/// <summary>
		/// Gets the string representation of this packet.
		/// </summary>
		public override string ToString() {
			return String.Format("[{5} Source={0} Destination={1} " +
			                     "EncryptionType={2} PacketID={3} " +
			                     "DependsOn={4}]",
			                     source,
			                     ToString(destination),
			                     encryptionType,
			                     packetID,
			                     dependsOn,
			                     //HopsToString(hops),
			                     GetType().Name);
		}
		
		private string ToString(int[] array) {
			string res = "{";
			// This is just for debugging purposes, performance doesn't matter
			for (int i = 0; i < array.Length; ++i) {
				if (i > 0) res += ", ";
				res += array[i].ToString();
			}
			return res + "}";
		}
		
		private string HopsToString(int[] array) {
			string res = "{";
			// This is just for debugging purposes, performance doesn't matter
			for (int i = 0; i < array.Length; ++i) {
				if (i > 0) res += ", ";
				res += array[i].ToString();
				if (array[i] == 0) break;
			}
			return res + "...}";
		}
		#endregion
	}
}
