// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Text;
using Grunwald.P2PLib.Network;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// Routing Strategy, where all clients are peers, no supernodes exist.
	/// </summary>
	public class PeerRoutingStrategy : IRoutingStrategy
	{
		// WHIDBEY GENERIC : This class really needs generics!
		Hashtable routingTables = new Hashtable(); // NetClient -> RoutingTable
		Hashtable mainTable = new Hashtable(); // uid -> RoutingTable
		
		DgnP2PNetwork network;
		object routingLock = new object();
		// All public methods have to lock the routingLock because the
		// private methods expect to be the only thread accessing the routing tables!
		
		static void Log(string text)
		{
			#if DEBUG
			DateTime now = DateTime.Now;
			Console.WriteLine("{0:hh:mm:ss}:{1:d3} {2}", now, now.Millisecond, text);
			#endif
		}
		
		static void Log(string format, params object[] args)
		{
			#if DEBUG
			Log(string.Format(format, args));
			#endif
		}
		
		#region Constructor / Dispose
		/// <summary>
		/// Registers the userid as own user id (distance and time set to zero).
		/// </summary>
		public PeerRoutingStrategy(DgnP2PNetwork network)
		{
			if (network == null)
				throw new ArgumentNullException("network");
			this.network = network;
			network.ReceivingUserAdded += new ReceivingUserEventHandler(ReceivingUserAdded);
			network.ReceivingUserRemoved += new ReceivingUserEventHandler(ReceivingUserRemoved);
			foreach (ReceivingUser user in network.ReceivingUsers) {
				AddReceivingUser(user);
			}
		}
		
		/// <summary>
		/// Clears all information in the routing strategy and deregisters events wired up.
		/// </summary>
		public void Dispose()
		{
			lock(routingLock) {
				if (network != null) {
					foreach (ReceivingUser user in network.ReceivingUsers) {
						RemoveReceivingUser(user);
					}
					network.ReceivingUserAdded -= new ReceivingUserEventHandler(ReceivingUserAdded);
					network.ReceivingUserRemoved -= new ReceivingUserEventHandler(ReceivingUserRemoved);
				}
				network = null;
				mainTable = null;
				routingTables = null;
			}
		}
		#endregion
		
		#region Manage ReceivingUsers
		private void ReceivingUserAdded(object sender, ReceivingUserEventArgs e)
		{
			lock (routingLock) {
				AddReceivingUser(e.User);
			}
		}
		
		private void ReceivingUserRemoved(object sender, ReceivingUserEventArgs e)
		{
			lock (routingLock) {
				RemoveReceivingUser(e.User);
			}
		}
		
		private void AddReceivingUser(ReceivingUser u)
		{
			foreach (P2PAddress adr in u.Addresses) {
				AddReceivingUserAddress(adr);
			}
			u.AddressAdded += new P2PAddressEventHandler(ReceivingUserAddressAdded);
		}
		
		private void RemoveReceivingUser(ReceivingUser u)
		{
			foreach (P2PAddress adr in u.Addresses) {
				RemoveReceivingUserAddress(adr);
			}
			u.AddressAdded -= new P2PAddressEventHandler(ReceivingUserAddressAdded);
		}
		
		private void ReceivingUserAddressAdded(object sender, P2PAddressEventArgs e)
		{
			lock (routingLock) {
				AddReceivingUserAddress(e.Address);
			}
		}
		
		private void ReceivingUserAddressRemoved(object sender, P2PAddressEventArgs e)
		{
			lock (routingLock) {
				RemoveReceivingUserAddress(e.Address);
			}
		}
		
		private void AddReceivingUserAddress(P2PAddress adr)
		{
			if (adr.Network != network)
				return;
			UserAddress a = (UserAddress)adr;
			mainTable[a.UserID] = adr;
			MergeToAll(a.UserID, new RoutingEntry(a.UserID, 0, 0, a.UserID));
		}
		
		private void RemoveReceivingUserAddress(P2PAddress adr)
		{
			if (adr.Network != network)
				return;
			UserAddress a = (UserAddress)adr;
			mainTable.Remove(a.UserID);
			MergeToAll(a.UserID, null);
		}
		#endregion
		
		#region Manage Routing Tables
		#region Add/Remove NetClients
		/// <summary>
		/// Adds a new NetClient to this <see cref="IRoutingStrategy">RoutingStrategy</see>.
		/// The NetClient is automatically removed when it's connection is closed.
		/// </summary>
		public void AddNetClient(NetClient netClient)
		{
			lock (routingLock) {
				netClient.Closed += new ClosedEventHandler(NetClientClosed);
				RoutingTable rt = new RoutingTable(netClient);
				routingTables[netClient] = rt;
			}
			SendRoutingTable(netClient, true);
		}
		
		private void NetClientClosed(object sender, ClosedEventArgs e)
		{
			lock (routingLock) {
				NetClient nc = (NetClient)sender;
				RoutingTable table = (RoutingTable)routingTables[nc];
				routingTables.Remove(nc);
				Log("NetClient closed: " + nc.Partner.UserID);
				// WHIDBEY GENERIC : use <RoutingEntry>.
				ArrayList deadEntries = new ArrayList();
				foreach (DictionaryEntry de in mainTable) {
					if (de.Value == table) {
						deadEntries.Add(de.Key);
					}
				}
				foreach (int uid in deadEntries) {
					SetBestBySearch(uid);
				}
			}
		}
		#endregion
		
		#region Receive
		/// <summary>
		/// Receives a routing table from the specified <see cref="NetClient"/>.
		/// The entries in the table are merged into the main table, and the changelog
		/// is written so that the entries can be forwarded to other partners.
		/// </summary>
		/// <param name="netClient">The <see cref="NetClient"/> that has received the
		/// routing table.</param>
		/// <param name="data">The byte array containing the data of the routing table.</param>
		/// <exception cref="ProtocolViolationException">The routing table was invalid.</exception>
		public void Receive(NetClient netClient, byte[] data)
		{
			lock(routingLock) {
				RoutingTable newTable = (RoutingTable)routingTables[netClient];
				if (newTable == null)
					throw new InvalidOperationException("The netclient was not added to this RoutingStrategy!");
				#if DEBUG
				StringBuilder builder = new StringBuilder();
				for (int pos = 0; pos < data.Length; pos += RoutingEntry.Length) {
					RoutingEntry debugEntry = new RoutingEntry(data, pos);
					builder.Append("<");
					builder.Append(debugEntry.ToString());
					builder.Append(">");
				}
				Log("Receiving {0} entries from {1}: {2}", data.Length / RoutingEntry.Length, netClient.Partner.UserID, builder.ToString());
				#endif
				int networkUserID = network.UserID;
				
				// newTable is the table of the receiving net client
				// now we have to parse the values, save them in newTable and
				// we have to merge the new values into the main table and into the
				// other netclients' changelogs
				
				for (int pos = 0; pos < data.Length; pos += RoutingEntry.Length) {
					RoutingEntry newEntry = new RoutingEntry(data, pos);
					int uid = newEntry.UserID;
					if (newEntry.Server == 0) {
						newEntry = null;
					} else {
						newEntry.Distance += 1;
						newEntry.Time += 2;
						
						if (newEntry.Distance == 1) {
							newEntry.Server = networkUserID;
						} else {
							// Validating check against circular passing of entries
							// This is the reason we are using the "Server" field.
							int serverDistance = this.GetDistance(newEntry.Server, data, pos + RoutingEntry.Length);
							if (serverDistance + 1 < newEntry.Distance) {
								newEntry = null;
							}
						}
					}
					
					object oldTableObject = mainTable[uid];
					RoutingTable oldTable = oldTableObject as RoutingTable;
					if (oldTable == null) {
						if (oldTableObject != null) {
							// there is an entry in mainTable, but it isn't a RoutingTable:
							// this is a user on our own computer, so don't accept the received value!
							continue;
						}
						
						// there is no entry in mainTable -> there is no old value,
						// so we accept the new value in any case
						
						newTable[uid] = newEntry; // save new entry in receiving table
						
						if (newEntry != null) { // only save when there IS a new entry
							SetBest(newTable, newEntry);
						}
					} else {
						// there is an entry in mainTable -> there is an old value
						RoutingEntry oldEntry = (RoutingEntry)oldTable[uid];
						
						// we could not save the newEntry in newTable before because we had to
						// be able to get the oldEntry when oldTable == newTable
						newTable[uid] = newEntry; // save new entry in receiving table
						
						if (oldTable == newTable) {
							// the new entry comes from the same source as the old entry
							if (newEntry != null && newEntry.Time <= oldEntry.Time) {
								// the entry is better than the last time (so better than the best):
								SetBest(newTable, newEntry, oldEntry);
							} else {
								// the entry has got worse since the last time:
								// find the best entry in all routing tables
								SetBestBySearch(uid, oldEntry);
							}
						} else {
							// the new entry comes from a different net client than the old entry
							if (newEntry != null && newEntry.Time < oldEntry.Time) {
								SetBest(newTable, newEntry, oldEntry);
							} else if (newEntry != null) {
								// Maybe this entry is now the best-but-one entry and should be sent
								// to the source of the best entry.
								if (SearchBest(newEntry.UserID, oldTable) == newEntry) {
									MergeToTable(newEntry.UserID, newEntry, oldTable);
								}
							}
						}
					}
				}
			}
		}
		
		/// <summary>
		/// Sets the best entry in the main table. The other netclients are notified about the change.
		/// </summary>
		private void SetBest(RoutingTable newTable, RoutingEntry newEntry)
		{
			SetBest(newTable, newEntry, null);
		}
		
		/// <summary>
		/// Sets the best entry in the main table. The other netclients are notified about the change.
		/// </summary>
		private void SetBest(RoutingTable newTable, RoutingEntry newEntry, RoutingEntry oldEntry)
		{
			#if DEBUG
			if (newTable == null)
				throw new ArgumentNullException("newTable");
			if (newEntry == null)
				throw new ArgumentNullException("newEntry");
			Log("SetBest: #{0} -> {1}", newEntry.UserID, newTable.NetClient.Partner.UserID);
			#endif
			mainTable[newEntry.UserID] = newTable;
			if (!newEntry.Equals(oldEntry)) {
				MergeToOthers(newTable, newEntry);
			}
		}
		
		/// <summary>
		/// Removes the entry with the specified user id from the main table.
		/// The other netclients are notified about the change.
		/// </summary>
		private void RemoveEntryFromMainTable(int userid)
		{
			#if DEBUG
			Log("Remove {0} from main table", userid);
			#endif
			mainTable.Remove(userid);
			MergeToAll(userid, null);
		}
		
		/// <summary>
		/// Searches in all routing tables for the best entry for userid and calls SetBest for
		/// the result.
		/// </summary>
		private void SetBestBySearch(int userid)
		{
			SetBestBySearch(userid, null);
		}
		
		/// <summary>
		/// Searches in all routing tables for the best entry for userid and calls SetBest for
		/// the result.
		/// </summary>
		private void SetBestBySearch(int userid, RoutingEntry oldEntry)
		{
			RoutingTable bestTable;
			RoutingEntry bestEntry = SearchBest(userid, null, out bestTable);
			if (bestEntry == null) {
				RemoveEntryFromMainTable(userid);
			} else {
				SetBest(bestTable, bestEntry, oldEntry);
			}
		}
		
		/// <summary>
		/// Searches all routing tables except table '<paramref name="except"/>' for the best way
		/// to <paramref name="userid"/>.
		/// </summary>
		private RoutingEntry SearchBest(int userid, RoutingTable except)
		{
			RoutingTable tmp;
			return SearchBest(userid, except, out tmp);
		}
		
		/// <summary>
		/// Searches all routing tables except table '<paramref name="except"/>' for the best way
		/// to <paramref name="userid"/>. <paramref name="table"/> is set to the table in which
		/// the best entry is found.
		/// </summary>
		private RoutingEntry SearchBest(int userid, RoutingTable except, out RoutingTable table)
		{
			RoutingEntry bestEntry = null;
			RoutingTable bestTable = null;
			foreach (RoutingTable rt in routingTables.Values) {
				if (rt == except)
					continue;
				RoutingEntry entry = (RoutingEntry)rt[userid];
				if (entry != null && (bestEntry == null || bestEntry.Time > entry.Time)) {
					bestEntry = entry;
					bestTable = rt;
				}
			}
			table = bestTable;
			return bestEntry;
		}
		
		/// <summary>
		/// Merges the entry with the specified id from the source netclient into the
		/// changelog of all other netclients.
		/// </summary>
		private void MergeToOthers(RoutingTable source, RoutingEntry entry)
		{
			#if DEBUG
			if (source == null)
				throw new ArgumentNullException("source");
			if (entry == null)
				throw new ArgumentNullException("entry");
			#endif
			foreach (RoutingTable rt in routingTables.Values) {
				if (rt == source) {
					MergeToTable(entry.UserID, SearchBest(entry.UserID, rt), rt);
				} else {
					MergeToTable(entry.UserID, entry, rt);
				}
			}
		}
		
		/// <summary>
		/// Merges the entry into the changelog of all netclients.
		/// </summary>
		private void MergeToAll(int userid, RoutingEntry entry)
		{
			foreach (RoutingTable rt in routingTables.Values) {
				MergeToTable(userid, entry, rt);
			}
		}
		
		/// <summary>
		/// Merges the specified routing entry into the target's routing table's changelog if necessary.
		/// </summary>
		private void MergeToTable(int id, RoutingEntry entry, RoutingTable target)
		{
			if (entry == null)
				entry = new RoutingEntry(id, 0, 0, 0);
			ArrayList changeLog = target.ChangeLog;
			
			// prevent sending entry about user #x to user #x
			RoutingEntry targetEntry = (RoutingEntry)target[id];
			if (targetEntry != null && targetEntry.Distance == 1)
				return;
			
			for (int i = 0; i < changeLog.Count; ++i) {
				if (((RoutingEntry)changeLog[i]).UserID == id) {
					changeLog[i] = entry;
					return;
				}
			}
			changeLog.Add(entry);
		}
		#endregion
		
		#region Send
		private bool ChangeLogContainsUser(ArrayList changeLog, int userid)
		{
			foreach (RoutingEntry e in changeLog) {
				if (e.UserID == userid)
					return true;
			}
			return false;
		}
		
		/// <summary>
		/// Sends the RoutingTable to the specified NetClient.
		/// </summary>
		/// <param name="netClient">The NetClient to send the routing table to.</param>
		/// <param name="sendFullTable">
		/// Specifies if the full table should be resent or just the changes should be sent.
		/// </param>
		public void SendRoutingTable(NetClient netClient, bool sendFullTable)
		{
			byte[] data;
			lock(routingLock) {
				RoutingTable rt = (RoutingTable)routingTables[netClient];
				if (rt == null)
					return;
				
				ArrayList changeLog = rt.ChangeLog;
				int count = changeLog.Count;
				if (sendFullTable) {
					int count2 = 0;
					foreach (DictionaryEntry de in mainTable) {
						int userID = (int)de.Key;
						if (ChangeLogContainsUser(changeLog, userID))
							continue;
						
						// prevent sending entry about user #x to user #x
						RoutingEntry entryFromTarget = rt[userID];
						if (entryFromTarget != null && entryFromTarget.Distance == 1)
							continue;
						
						RoutingTable e = de.Value as RoutingTable;
						if (e != null && e.NetClient == netClient) {
							if (SearchBest(userID, e) == null)
								continue;
						}
						count2 += 1;
					}
					count2 += count;
					Log("Sending {0} fresh routing entries on {1}.", count2, netClient.Partner.UserID);
					data = new byte[count2 * RoutingEntry.Length];
				} else {
					if (count == 0) return;
					Log("Sending {0} changed routing entries on {1}.", count, netClient.Partner.UserID);
					data = new byte[count * RoutingEntry.Length];
				}
				int pos = 0;
				foreach (RoutingEntry re in changeLog) {
					re.ToByte(data, pos);
					pos += RoutingEntry.Length;
				}
				
				if (sendFullTable) {
					foreach (DictionaryEntry de in mainTable) {
						RoutingTable e = de.Value as RoutingTable;
						int uid = (int)de.Key;
						if (ChangeLogContainsUser(changeLog, uid))
							continue;
						RoutingEntry entryFromTarget = rt[uid];
						if (entryFromTarget != null && entryFromTarget.Distance == 1)
							continue;
						RoutingEntry re;
						if (e != null && e.NetClient == netClient)
							re = SearchBest(uid, e);
						else
							re = null;
						if (e == null)
							re = new RoutingEntry(uid, 0, 0, uid); // Address for own user
						else
							re = (RoutingEntry)e[uid];
						re.ToByte(data, pos);
						pos += RoutingEntry.Length;
					}
				}
				changeLog.Clear();
			}
			if (data.Length > 0)
				netClient.BeginSend(DataBlockType.RoutingTable, data);
		}
		#endregion
		#endregion
		
		#region Provide information
		/// <summary>
		/// Gets the distance from this computer to the specified user id.
		/// If the user is not found, the method returns -1.
		/// </summary>
		public int GetDistance(int userid)
		{
			lock (routingLock) {
				return GetDistanceInternal(userid);
			}
		}
		
		private int GetDistanceInternal(int userid)
		{
			object o = mainTable[userid];
			RoutingTable t = o as RoutingTable;
			if (t != null)
				return ((RoutingEntry)t[userid]).Distance;
			else if (o != null)
				return 0;
			else
				return -1;
		}
		
		/// <summary>
		/// Gets the distance from this computer to the specified user id.
		/// If the user is not found, the method returns -1.
		/// </summary>
		private int GetDistance(int userid, byte[] data, int dataStart)
		{
			int storedDistance = GetDistanceInternal(userid);
			// Do a quick find through the data array, looking for user id
			byte[] lookFor = NetworkBitConverter.GetBytes(userid);
			byte b1 = lookFor[0];
			byte b2 = lookFor[1];
			byte b3 = lookFor[2];
			byte b4 = lookFor[3];
			int foundDistance = -1;
			for (int i = dataStart; i < data.Length; i += RoutingEntry.Length) {
				if (data[i] == b1 && data[i + 1] == b2 && data[i + 2] == b3 && data[i + 3] == b4) {
					foundDistance = new RoutingEntry(data, i).Distance + 1;
					break;
				}
			}
			if (foundDistance == -1)
				return storedDistance;
			else if (storedDistance == -1)
				return foundDistance;
			else
				return Math.Min(storedDistance, foundDistance);
		}
		
		/// <summary>
		/// Gets the string representation of all routing entries in all routing tables.
		/// </summary>
		public override string ToString()
		{
			lock (routingLock) {
				StringBuilder b = new StringBuilder("[PeerRoutingStrategy:");
				foreach (RoutingTable rt in routingTables.Values) {
					b.Append(rt.ToString());
				}
				b.Append("\nMain table:\n");
				foreach (DictionaryEntry e in mainTable) {
					if (e.Value is RoutingTable)
						b.AppendFormat("    {0} -> #{1}\n", e.Key, ((RoutingTable)e.Value).NetClient.Partner.UserID);
					else
						b.AppendFormat("    {0} -> {1}\n", e.Key, e.Value);
				}
				b.Append("]");
				return b.ToString();
			}
		}
		#endregion
		
		#region Route packets
		/// <summary>
		/// Find best route and send packet.
		/// </summary>
		/// <param name="header">The header containing the information where to send the packet to.</param>
		/// <param name="data">The array containing the packet body.</param>
		/// <param name="offset">The offset where in the array the packet body starts.</param>
		/// <param name="length">The length of the packet body.</param>
		public void SendPacket(PacketHeader header, byte[] data, int offset, int length)
		{
			// TODO: translate comments
			
			// So, wir haben hier 1 Postkarte mit n Zieladressen.
			// Dabei haben wir m Postboten zur Verfügung.
			// Wir wollen Papier und Weg sparen, also kopieren wir
			// das ganze nur so wenig wie möglich. Sollen die
			// anderen Poststellen dann weiter kopieren.
			// Aber das der gesamte Weg möglichst kurz ist, ist das wichtigste!
			
			// Eigenen Poststempel draufpacken
			if (!header.AddHop(network.UserID)) {
				// TODO: Log that limit is reached.
				return;
			}
			
			// Erst mal vom Paket den Adresszettel abtrennen.
			int[] dest = header.Destination;
			header.Destination = null;
			NetClient[] sendWith = new NetClient[dest.Length];
			
			ArrayList receiveAddresses = null; // contains UserAddress of local users who receive the packet
			
			// Jetzt wird für jede Adresse das optimale Ziel errechnet
			lock (routingLock) {
				for (int i = 0; i < dest.Length; ++i) {
					int u = dest[i];
					sendWith[i] = null;
					if (u == 0)
						continue; // 0 = kein Ziel (da einige Funktionen das Array-Verkleinern sparen)
					object routObj = mainTable[u];
					if (routObj == null) // TODO: Check packet flags and send failure messag
						continue;
					UserAddress address = routObj as UserAddress;
					if (address != null) {
						if (receiveAddresses == null)
							receiveAddresses = new ArrayList();
						receiveAddresses.Add(address);
						// address.Network.ReceivePacket cannot be called directly because we are inside the
						// routingLock
					} else {
						sendWith[i] = ((RoutingTable)routObj).NetClient;
					}
				}
				
				// Jetzt ist hinter jeder Adresse vermerkt, mit welcher Verbindung das Paket
				// weiter verschickt werden soll.
				// Also alle Verbindungen durchsuchen und an passend gekennzeichneten
				// Adressen versenden!
				
				foreach (NetClient conn in routingTables.Keys) {
					if (conn.Status != NetClientStatus.Ready)
						continue;
					// Jetzt erst mal zählen wie groß der neue Adresszettel sein muss
					int count = 0;
					for (int i = 0; i < sendWith.Length; ++i) {
						if (sendWith[i] == conn) {
							count += 1; // Und ein Eintrag mehr
						}
					}
					if (count == 0)
						continue; // Nichts für diese Verbindung, weiter mit der nächsten
					header.Destination = new int[count];
					// Und jetzt auf die Liste eintragen
					int j = 0; // counter over header.Destination
					for (int i = 0; i < sendWith.Length; ++i) {
						if (sendWith[i] == conn) {
							header.Destination[j] = dest[i]; // eintragen
							j += 1;
						}
					}
					// So, wir haben jetzt ein neues Paket mit dem passenden Adresszettel
					// Dann mal abschicken!
					int headerLength = header.Length;
					byte[] sendData = new byte[headerLength + length];
					header.ToByte(sendData, 0);
					Array.Copy(data, offset, sendData, headerLength, length);
					conn.BeginSend(DataBlockType.SendPacket, sendData);
				}
			}
			header.Destination = dest;
			if (receiveAddresses != null) {
				foreach (UserAddress address in receiveAddresses) {
					address.Network.ReceivePacket(header, data, offset, length, address.UserID);
				}
			}
		}
		#endregion
		
		#region Send answer
		/// <summary>
		/// Find next node and send the answer. The method returns true if the current
		/// computer is the target for the answer.
		/// </summary>
		public bool SendAnswer(Answer answer)
		{
			// TODO: skip hops if possible
			int endTarget = answer.Hops[0];
			int nextHop = answer.GetLastHop();
			object tableObject;
			RoutingTable table;
			RoutingEntry e;
			lock (routingLock) {
				tableObject = mainTable[nextHop];
				table = tableObject as RoutingTable;
				if (table == null)
					e = null;
				else
					e = table[nextHop];
			}
			if (tableObject == null || (e != null && e.Distance > 1)) {
				// can't follow back, so route the packet
				Packet p = new Packet();
				p.Header.Source = network.UserID;
				p.Header.PacketID = 0;
				p.AnswerId = 0;
				p.Target = "RoutedAnswer";
				answer.Hops = new int[] { endTarget };
				p.Header.Destination = answer.Hops;
				p.Values.AddValue(answer.ToByte());
				network.SendPacket(p);
			} else {
				if (table == null) {
					if (answer.Hops.Length > 1) {
						answer.DropHop();
						return SendAnswer(answer);
					} else {
						return true;
					}
				} else {
					table.NetClient.BeginSend(DataBlockType.Answer, answer.ToByte());
				}
			}
			return false;
		}
		#endregion
	}
}
