// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

// WHIDBEY: Use generic hashtable and make this a struct
namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// An entry in a <see cref="RoutingTable"/>.
	/// </summary>
	public class RoutingEntry
	{
		/// <summary>
		/// The length of an entry in byte.
		/// </summary>
		public const int Length = 12;
		
		int userId;
		/// <summary>
		/// The id of the user this entry describes.
		/// </summary>
		public int UserID { get { return userId; } }
		
		short distance;
		/// <summary>
		/// The distance (number of computers) between us and the user.
		/// </summary>
		public short Distance {
			get { return distance; }
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException("distance");
				distance = value;
			}
		}
		
		short time;
		/// <summary>
		/// The time (unit not specified yet) packets need for this route.
		/// </summary>
		/// <remarks>
		/// This field is the sum of the "connection quality" of all connections on this route.
		/// </remarks>
		public short Time {
			get { return time; }
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException("time");
				time = value;
			}
		}
		
		int server;
		/// <summary>
		/// The "server" the user is connected to.
		/// This is the last user on the route.
		/// </summary>
		public int Server {
			get { return server; }
			set {
				server = value;
			}
		}
		
		/// <summary>
		/// Gets if this routing entry equals exactly the specified entry.
		/// Note: RoutingEntry does NOT override Object.Equals!!!
		/// </summary>
		public bool IsEqual(RoutingEntry e)
		{
			if (e == null) return false;
			return e.userId == this.userId && e.server == this.server && e.time == this.time && e.distance == this.distance;
		}
		
		/// <summary>
		/// Creates a new RoutingEntry with the specified parameters.
		/// </summary>
		public RoutingEntry(int userId, short distance, short time, int server)
		{
			//if (server == 0)   Erlaubt für user-Lösch-Eintrag
			//	throw new ArgumentException("server");
			if (userId == 0)
				throw new ProtocolViolationException("RoutingEntry.UserID cannot be zero.");
			if (distance < 0)
				throw new ProtocolViolationException("RoutingEntry.Distance cannot be negative.");
			if (time < 0)
				throw new ProtocolViolationException("RoutingEntry.Time cannot be negative.");
			
			this.distance = distance;
			this.time = time;
			this.server = server;
			this.userId = userId;
		}
		
		/// <summary>
		/// Creates a new RoutingEntry getting its parameters from a byte array.
		/// </summary>
		/// <param name="data">The byte array containing the data.</param>
		/// <param name="index">The index in the byte array where the entry starts at.</param>
		public RoutingEntry(byte[] data, int index)
			: this(
			       NetworkBitConverter.ToInt32(data, index    ),
			       NetworkBitConverter.ToInt16(data, index + 4), // distance
			       NetworkBitConverter.ToInt16(data, index + 6), // time
			       NetworkBitConverter.ToInt32(data, index + 8)  // server
			      )
		{
		}
		
		/// <summary>
		/// Gets a byte array containing the data of this RoutingEntry.
		/// </summary>
		public byte[] ToByte()
		{
			byte[] b = new byte[Length];
			ToByte(b, 0);
			return b;
		}
		
		/// <summary>
		/// Copies the data of this RoutingEntry to the specified byte array.
		/// </summary>
		/// <param name="b">The byte array to copy the data to.</param>
		/// <param name="index">The index in the byte array where the entry starts at.</param>
		public void ToByte(byte[] b, int index)
		{
			NetworkBitConverter.PutBytes(b, index    , userId);
			NetworkBitConverter.PutBytes(b, index + 4, distance);
			NetworkBitConverter.PutBytes(b, index + 6, time);
			NetworkBitConverter.PutBytes(b, index + 8, server);
		}
		
		/// <summary>
		/// Converts this RoutingEntry into a string.
		/// </summary>
		public override string ToString() {
			return String.Format("#{0}, {1:d02} {2:d02}, #{3}",
			                     userId,
			                     distance,
			                     time,
			                     server);
		}
	}
}
