// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Threading;

namespace Grunwald.P2PLib.DGNetwork
{
	/// <summary>
	/// The SendResult class is used to receive answers to your request packets.
	/// </summary>
	public sealed class SendResult
	{
		int packetId;
		int answerId;
		int source;
		DateTime sendTime;
		Answer answer;
		ArrayList answers;
		
		internal SendResult(Packet p)
		{
			this.packetId = p.Header.PacketID;
			this.source = p.Header.Source;
			this.answerId = p.AnswerId;
			this.sendTime = DateTime.Now;
		}
		
		#region Getter
		/// <summary>
		/// Gets the answer id of the packet.
		/// </summary>
		public int AnswerId {
			get {
				return answerId;
			}
		}
		
		/// <summary>
		/// Gets the first answer. This property is null until the answer is received.
		/// </summary>
		public Answer Answer {
			get {
				return answer;
			}
		}
		
		/// <summary>
		/// Gets the list of all answers.
		/// </summary>
		public ArrayList Answers {
			get {
				if (answers == null) {
					answers = new ArrayList();
					if (answer != null)
						answers.Add(answer);
				}
				return answers;
			}
		}
		
		/// <summary>
		/// Gets the packet id of the packet.
		/// </summary>
		public int PacketId {
			get {
				return packetId;
			}
		}
		
		/// <summary>
		/// Gets the source user id of the packet.
		/// </summary>
		public int Source {
			get {
				return source;
			}
		}
		
		/// <summary>
		/// Gets the time the packet was sent.
		/// </summary>
		public DateTime SendTime {
			get {
				return sendTime;
			}
		}
		#endregion
		
		#region Unique identifier
		internal ulong UniqueKey {
			get {
				unchecked {
					return ((ulong)source << 32) + (ulong)packetId;
				}
			}
		}
		
		internal static ulong GetUniqueIdentifier(int packetId, int source)
		{
			unchecked {
				return ((ulong)source << 32) + (ulong)packetId;
			}
		}
		#endregion
		
		#region Receive Answer
		/// <summary>
		/// Occurs when an answer to the packet has been received.
		/// WARNING: when the answer already has been received, this method isn't called!
		/// Use CallOnAnswer instead!
		/// </summary>
		private event AnswerEventHandler AnswerReceived;
		
		object lockObject = new object();
		
		internal void ReceiveAnswer(Answer answer)
		{
			lock (lockObject) {
				if (answer == null) {
					this.answer = answer;
				} else {
					this.Answers.Add(answer);
				}
				if (AnswerReceived != null)
					AnswerReceived(this, new AnswerEventArgs(answer));
			}
		}
		
		/// <summary>
		/// Calls the method when an answer was received.
		/// This means the method is called immediately if the answer has already been received.
		/// </summary>
		public void CallOnAnswer(AnswerEventHandler method)
		{
			lock (lockObject) {
				this.AnswerReceived += method;
				if (answers != null) {
					foreach (Answer ans in answers) {
						method(this, new AnswerEventArgs(ans));
					}
				} else if (answer != null) {
					method(this, new AnswerEventArgs(answer));
				}
			}
		}
		
		/// <summary>
		/// Waits until an answer has been received, but at maximum <paramref name="timeout"/>
		/// milliseconds.
		/// </summary>
		/// <returns>Returns true when the answer has received, or false when the timeout
		/// was reached.</returns>
		public bool WaitForAnswer(int timeout)
		{
			if (timeout <= 0)
				throw new ArgumentOutOfRangeException("timeout", "timeout must be a positive number");
			while (true) {
				lock (lockObject) {
					if (answer != null)
						return true;
				}
				if (timeout > 50) {
					Thread.Sleep(50);
					timeout -= 50;
				} else {
					Thread.Sleep(timeout);
					break;
				}
			}
			lock (lockObject) {
				if (answer != null)
					return true;
			}
			return false;
		}
		#endregion
	}
}
