// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Text;
using NUnit.Framework;
using Grunwald.P2PLib.DGNetwork;

namespace Grunwald.P2PLib.Tests.DGNetwork
{
	[TestFixture]
	public class AnswerTest
	{
		static string CorrectAnswer =
			"000000c8" + // FromUID
			"00000190" + // PacketID
			"00000064" + // AnswerID
			"c5" + // ErrorCode
			"05" + // Number of hops
			"0000000a000000140000001e0000002800000032" + // {10, 20, 30, 40, 50}
			
			"0004" +  // count of nameless values
			"0000000a74657374737472696e67" + // 10 byte length: "teststring"
			"00000004ffffff9c" +             // 4 byte length: -100 (integer)
			"0000000831353020e282ac2e" +     // 8 byte length: UTF8("150 €.")
			"00000004000000c8" +             // 4 byte length: 200 (integer)
			"0003" +  // count of named values
			"046b657931" +           // name: "key1"
			"0000000676616c756531" + // 6 byte length: "value1"
			"046b657932" +       // name: "key2"
			"00000004f1f2f3f4" + // 4 byte length: xF1F2F3F4 (uint)
			"0464617461" +       // name: data
			"0000000444556677";  // 4 byte length: {0x44, 0x55, 0x66, 0x77}
		
		[Test]
		public void SetAndCheck()
		{
			Answer answer = new Answer();
			SetInfo(answer);
			CheckValues(answer);
		}
		
		[Test]
		public void Clone()
		{
			Answer answer = new Answer();
			SetInfo(answer);
			Answer clone = answer.Clone();
			if (answer == clone)
				Assert.Fail("Clone returned same reference");
			if (answer.Hops == clone.Hops)
				Assert.Fail("Clone returned same hops-array reference");
			CheckValues(clone);
		}
		
		[Test]
		public void ToByte()
		{
			Answer answer = new Answer();
			SetInfo(answer);
			byte[] data = answer.ToByte();
			Assert.AreEqual(CorrectAnswer, Dump(data), "binary representation");
		}
		
		[Test]
		public void FromByte()
		{
			byte[] data = FromString(CorrectAnswer, 3);
			Answer answer = new Answer(data, 3, data.Length - 3);
			CheckValues(answer);
		}
		
		private void SetInfo(Answer answer)
		{
			answer.AnswerId = 100;
			answer.FromUser = 200;
			answer.Hops = new int[] {10, 20, 30, 40, 50};
			answer.PacketId = 400;
			answer.ErrorCode = 197;
			Answer p = answer;
			p.Values.AddValue("teststring");
			p.Values.AddValue(-100);
			p.Values.AddValue(150);
			p.Values.AddValue((uint)200);
			p.Values.SetValue(2, "150 €.");
			p.NamedValues.AddValue("key1", "value1");
			p.NamedValues.AddValue("key2", 0xF1F2F3F4); // uint!!
			p.NamedValues.AddValue("data", new byte[] {0x44, 0x55, 0x66, 0x77});
		}
		
		private void CheckValues(Answer answer)
		{
			Assert.AreEqual(100, answer.AnswerId, "AnswerId");
			Assert.AreEqual(200, answer.FromUser, "FromUser");
			Assert.AreEqual(5, answer.Hops.Length, "Hops.Length");
			Assert.AreEqual(197, answer.ErrorCode, "ErrorCode");
			for (int i = 0; i < 5; ++i) {
				Assert.AreEqual((i + 1) * 10, answer.Hops[i], "Hops[" + i + "]");
			}
			Assert.AreEqual(400, answer.PacketId);
			
			Answer p = answer;
			Assert.AreEqual(4, p.Values.Count, "IndexedValueCount");
			Assert.AreEqual("teststring", p.Values.GetValue(0, typeof(string)), "teststring");
			Assert.AreEqual(-100, p.Values.GetValue(1, typeof(int)), "int -100");
			Assert.AreEqual(unchecked((uint)-100), p.Values.GetValue(1, typeof(uint)), "uint -100");
			Assert.AreEqual("150 €.", p.Values.GetValue(2, typeof(string)), "150 €");
			Assert.AreEqual(200, p.Values.GetValue(3, typeof(int)), "int 200");
			Assert.AreEqual(200, p.Values.GetValue(3, typeof(uint)), "uint 200");
			
			Assert.AreEqual(3, p.NamedValues.Count, "NamedValueCount");
			Assert.AreEqual("value1", p.NamedValues.GetValue("key1", typeof(string)), "key1");
			Assert.AreEqual(0xF1F2F3F4, p.NamedValues.GetValue("key2", typeof(uint)), "key2");
			Assert.AreEqual("44556677", Dump((byte[])p.NamedValues.GetValue("data", typeof(byte[]))), "data (key3)");
		}
		
		private byte[] FromString(string str, int offset)
		{
			byte[] data = new byte[str.Length / 2 + offset];
			for (int i = 0; i < (str.Length / 2); ++i) {
				data[i + offset] = Convert.ToByte(str.Substring(i * 2, 2), 16);
			}
			return data;
		}
		
		private string Dump(byte[] data)
		{
			StringBuilder b = new StringBuilder();
			foreach(byte by in data) {
				b.Append(by.ToString("x2"));
			}
			return b.ToString();
		}
	}
}
