// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Text;
using NUnit.Framework;
using Grunwald.P2PLib.DGNetwork;

namespace Grunwald.P2PLib.Tests.DGNetwork
{
	[TestFixture]
	public class DataContainerTest
	{
		#region Indexed
		string correctIndexed =
			"0004" +  // count of nameless values
			"0000000a74657374737472696e67" + // 10 byte length: "teststring"
			"00000004ffffff9c" +             // 4 byte length: -100 (integer)
			"0000000831353020e282ac2e" +     // 8 byte length: UTF8("150 €.")
			"00000004000000c8";              // 4 byte length: 200 (integer)
		
		[Test]
		public void SaveIndexed()
		{
			IndexedValueList values = new IndexedValueList();
			values.AddValue("teststring");
			values.AddValue(-100);
			values.AddValue(150);
			values.AddValue((uint)200);
			values.SetValue(2, "150 €.");
			using (MemoryStream ms = new MemoryStream()) {
				values.Save(ms);
				Assert.AreEqual(correctIndexed, Dump(ms.ToArray()));
			}
		}
		
		[Test]
		public void ReadIndexed()
		{
			IndexedValueList values = new IndexedValueList();
			using (MemoryStream ms = new MemoryStream(FromString(correctIndexed), false)) {
				values.Read(ms);
			}
			Assert.AreEqual(4, values.Count, "IndexedValueCount");
			Assert.AreEqual("teststring", values.GetValue(0, typeof(string)), "teststring");
			Assert.AreEqual(-100, values.GetValue(1, typeof(int)), "int -100");
			Assert.AreEqual(unchecked((uint)-100), values.GetValue(1, typeof(uint)), "uint -100");
			Assert.AreEqual("150 €.", values.GetValue(2, typeof(string)), "150 €");
			Assert.AreEqual(200, values.GetValue(3, typeof(int)), "int 200");
			Assert.AreEqual(200, values.GetValue(3, typeof(uint)), "uint 200");
		}
		#endregion
		
		#region Named
		string correctNamed =
			"0003" +  // count of named values
			"046b657931" +           // name: "key1"
			"0000000676616c756531" + // 6 byte length: "value1"
			"046b657932" +       // name: "key2"
			"00000004f1f2f3f4" + // 4 byte length: xF1F2F3F4 (uint)
			"0464617461" +       // name: data
			"0000000444556677";  // 4 byte length: {0x44, 0x55, 0x66, 0x77}
		
		[Test]
		public void SaveNamed()
		{
			NamedValueList values = new NamedValueList();
			values.AddValue("key1", "value1");
			values.AddValue("key2", 0xF1F2F3F4); // uint!!
			values.AddValue("data", new byte[] {0x44, 0x55, 0x66, 0x77});
			using (MemoryStream ms = new MemoryStream()) {
				values.Save(ms);
				Assert.AreEqual(correctNamed, Dump(ms.ToArray()));
			}
		}
		
		[Test]
		public void ReadNamed()
		{
			NamedValueList values = new NamedValueList();
			using (MemoryStream ms = new MemoryStream(FromString(correctNamed), false)) {
				values.Read(ms);
			}
			Assert.AreEqual(3, values.Count, "NamedValueCount");
			Assert.AreEqual("value1", values.GetValue("key1", typeof(string)), "key1");
			Assert.AreEqual(0xF1F2F3F4, values.GetValue("key2", typeof(uint)), "key2");
			Assert.AreEqual("44556677", Dump((byte[])values.GetValue("data", typeof(byte[]))), "data (key3)");
		}
		#endregion
		
		private byte[] FromString(string str)
		{
			byte[] data = new byte[str.Length / 2];
			for (int i = 0; i < (str.Length / 2); ++i) {
				data[i] = Convert.ToByte(str.Substring(i * 2, 2), 16);
			}
			return data;
		}
		
		private string Dump(byte[] data)
		{
			StringBuilder b = new StringBuilder();
			foreach(byte by in data) {
				b.Append(by.ToString("x2"));
			}
			return b.ToString();
		}
	}
}
