// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections.Specialized;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using NUnit.Framework;
using Grunwald.P2PLib;
using Grunwald.P2PLib.Network;
using Grunwald.P2PLib.DGNetwork;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;
using Grunwald.P2PLib.Tests.Server;

namespace Grunwald.P2PLib.Tests.DGNetwork
{
	/// <summary>
	/// This class is one big test seperated in 4 steps.
	/// It builds a (virtual) connection between two (unrouted) NetClients using two threads
	/// and then disconnects again.
	/// NetClient1 is controlled directly by the test while NetClient2 is controlled by
	/// the DGNRequestHandler started in the other thread.
	/// </summary>
	[TestFixture]
	public class NetClientTest
	{
		#region Set Up
		Identity ident1 = new Identity(1024); // 1024 bit encryption
		Identity ident2 = new Identity(2048); // 2048 bit encryption
		
		DgnP2PNetwork net1;
		DgnP2PNetwork net2;
		
		ReceivingUser user1, user2;
		
		public NetClientTest()
		{
			// init user 1
			user1 = new ReceivingUser(ident1);
			net1 = new DgnP2PNetwork("bla", user1);
			user1.Addresses.Add(new UserAddress(net1, 1, ident1));
			net1.ConnectKey = "TopSecret";
			
			// init user 2
			user2 = new ReceivingUser(ident2);
			net2 = new DgnP2PNetwork("bla", user2);
			user2.Addresses.Add(new UserAddress(net2, 2, ident2));
			net2.ConnectKey = "TopSecret";
		}
		#endregion
		
		#region Step 1 - send request
		HttpRequest request = null;
		LocalConnection c1, c2;
		NetClient nc1, nc2;
		
		[Test]
		public void Connect1()
		{
			// Test the first connection request
			// user 1 builds a connection to user2.
			
			// Create "socket" connection
			LocalConnection.Create(out c1, out c2);
			
			nc1 = net1.BuildConnection(c1, user1, 2);
			StringCollection request1 = HttpLikeServer.GetRequest(c2);
			request = new HttpRequest(c2, request1);
			Assert.AreEqual("/system/network/dgnetwork/connect", request.File, "URL wrong");
			Assert.AreEqual("2@bla", request.Headers["Host"], "Connect Request: wrong host header.");
			Assert.AreEqual("1", request.Headers["Me"], "Connect Request: wrong me header.");
			Assert.AreEqual(ident1.PublicKey, request.Headers["MyIdentity"], "Connect Request: wrong identity header.");
			Assert.IsNotNull(request.Headers["Seed"], "Connect Request: Missing seed header.");
		}
		#endregion
		
		#region Step 2 - Work on request and send response
		DgnRequestHandler handler = new DgnRequestHandler();
		Thread thread;
		Exception threadException = null;
		bool threadFinished = false;
		
		void Check(StringCollection c, int i, string name, string expected)
		{
			Assert.AreEqual(expected, c[i], "Wrong " + name + " header");
		}
		
		StringCollection response;
		
		[Test]
		public void Connect2()
		{
			if (request == null)
				Assert.Fail("Connect2 can only be tested after Connect1 was successfull.");
			
			// init request handler
			handler.ReceivingUsers.Add(user2);
			
			// Start a thread that receives the request and sends a response
			thread = new Thread(new ThreadStart(StartThread));
			thread.IsBackground = true;
			thread.Name = "Connection request handling.";
			thread.Start();
			// wait for the thread to execute
			Thread.Sleep(10);
			// the NetClient has to run in that thread
			if (threadFinished)
				throw new AssertionException("Thread finished too quick", threadException);
			try {
				// get the response
				response = HttpLikeServer.GetRequest(c1);
				if (threadException != null)
					throw new AssertionException("Exception occured in helper thread.", threadException);
				nc2 = handler.LastNetClient;
				Assert.IsNotNull(nc2, "Could not get created netclient");
				// check if the response is valid.
				Check(response, 0, "response", "DGN/0.1 200 OK");
				Check(response, 1, "response identity", "MyIdentity: " + ident2.PublicKey);
				MD5CryptoServiceProvider md5Provider = new MD5CryptoServiceProvider();
				byte[] data = Encoding.UTF8.GetBytes(net1.ConnectKey + request.Headers["Seed"]);
				string connectKey = Convert.ToBase64String(md5Provider.ComputeHash(data));
				Check(response, 2, "response connectkey", "ConnectKey: " + connectKey);
				if (!response[3].StartsWith("Seed: "))
					Assert.Fail("Wrong response seed header: " + response[3]);
				if (threadException != null)
					throw new AssertionException("Exception occured in helper thread.", threadException);
			} catch (Exception) {
				response = null;
				if (!threadFinished)
					thread.Abort();
				throw;
			}
		}
		
		void StartThread() {
			try {
				handler.Handle(request);
			} catch (Exception ex) {
				Console.WriteLine("Connect2 Thread:\r\n {0}", ex);
				request.Connection.Close();
				threadException = ex;
			}
			threadFinished = true;
		}
		#endregion
		
		#region Step 3 - work on server response and send client response
		[Test]
		public void Connect3()
		{
			if (response == null)
				Assert.Fail("Connect3 can only be tested after Connect2 was successfull.");
			
			// resend the response so that the NetClient1 can receive it.
			StreamWriter w = c2.Writer;
			foreach (string line in response) {
				w.WriteLine(line);
			}
			w.WriteLine();
			w.Flush();
			
			bool ok = false;
			for (int i = 0; i < 20; ++i) {
				ok = nc1.Status == NetClientStatus.Connecting &&
					nc2.Status == NetClientStatus.WaitForClientResponse;
				if (ok) break;
				Thread.Sleep(5);
			}
			Assert.IsTrue(ok, "NetClient status1 is wrong: " + nc1.Status.ToString() + " / " + nc2.Status.ToString());
			
			nc1.ReceiveServerResponse();
			
			// Wait for both sides to process the messages
			for (int i = 0; i < 20; ++i) {
				ok = nc1.Status == NetClientStatus.ReadyForLoop &&
					nc2.Status == NetClientStatus.Ready;
				if (ok) break;
				Thread.Sleep(5);
			}
			Assert.IsTrue(ok, "NetClient status2 is wrong: " + nc1.Status.ToString() + " / " + nc2.Status.ToString());
			step3success = true;
		}
		#endregion
		
		#region Step 4 - test connection shutdown
		bool step3success = false;
		[Test]
		public void Connect4()
		{
			if (!step3success)
				Assert.Fail("Connect4 can only be tested after Connect3 was successfull.");
			
			nc1.Close();
			for (int i = 0; i < 15; ++i) {
				if (threadFinished) break;
				Thread.Sleep(5);
			}
			if (!threadFinished)
				Assert.Fail("Thread did not finished though connection was closed.");
			if (threadException != null)
				throw new AssertionException("Exception occured in helper thread.", threadException);
		}
		#endregion
	}
}
