// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Text;
using NUnit.Framework;
using Grunwald.P2PLib.DGNetwork;

namespace Grunwald.P2PLib.Tests.DGNetwork
{
	[TestFixture]
	public class PacketTest
	{
		static string correctHeader =
			"50" +       // Identification : is always 0x50 = 'P'
			"00000001" + // FromUID        : The user id of the source user of this packet.
			"03" +       // Targetcount    : Number of destination users
			"000000050000000600000007" + // the three destination users {5,6,7}
			"05" +       // EncryptionType : The type of the encryption used
			"00000014" + // Packet ID      : ID used to track delivery of packet and answer
			"0000000a" + // DependsOn      : ID of a packet this packet depends on.
			"0a" +       // HopsAllowed    : The maximum number of hops
			"00000002000000030000000000000000" + // the ten possible hops {2,3,...}
			"00000000000000000000000000000000" +
			"0000000000000000";
		
		static string correctPacket =
			"50000000010300000005000000060000" + // packet header, see above
			"000700000000140000000a0a00000002" + // (difference: encryption turned off)
			"00000003000000000000000000000000" +
			"00000000000000000000000000000000" +
			"00000000" + // end of packet header
			
			"00bc71b4" + // AnswerID
			"0e6d79437573746f6d546172676574" + // Target (string): one byte string length, then string as UTF-8
			"0004" +  // count of nameless values
			"0000000a74657374737472696e67" + // 10 byte length: "teststring"
			"00000004ffffff9c" +             // 4 byte length: -100 (integer)
			"0000000831353020e282ac2e" +     // 8 byte length: UTF8("150 €.")
			"00000004000000c8" +             // 4 byte length: 200 (integer)
			"0003" +  // count of named values
			"046b657931" +           // name: "key1"
			"0000000676616c756531" + // 6 byte length: "value1"
			"046b657932" +       // name: "key2"
			"00000004f1f2f3f4" + // 4 byte length: xF1F2F3F4 (uint)
			"0464617461" +       // name: data
			"0000000444556677";  // 4 byte length: {0x44, 0x55, 0x66, 0x77}
		
		[Test]
		public void SetHeaderAndCheck()
		{
			PacketHeader h = new PacketHeader();
			SetInfo(h);
			CheckValues(h);
		}
		
		[Test]
		public void SetPacketAndCheck()
		{
			Packet p = new Packet();
			SetInfo(p);
			CheckValues(p);
		}
		
		[Test]
		public void HeaderToByte()
		{
			PacketHeader h = new PacketHeader();
			SetInfo(h);
			byte[] data = h.ToByte();
			Assert.AreEqual(correctHeader, Dump(data), "byte representation wrong");
		}
		
		[Test]
		public void CloneHeader()
		{
			PacketHeader h = new PacketHeader();
			SetInfo(h);
			PacketHeader clone = h.Clone();
			if (h == clone)
				Assert.Fail("Clone returned same reference");
			if (h.Hops == clone.Hops)
				Assert.Fail("Clone returned same hops-array reference");
			CheckValues(clone);
		}
		
		[Test]
		public void HeaderFromByte()
		{
			int offset = 3;
			byte[] data = FromString(correctHeader, offset);
			PacketHeader h = new PacketHeader(data, offset);
			CheckValues(h);
		}
		
		[Test]
		public void PacketToByte()
		{
			Packet p = new Packet();
			SetInfo(p);
			byte[] data = p.ToByte(new EncryptionStrategy());
			//Console.WriteLine(Helper.HexDump(data));
			Assert.AreEqual(correctPacket, Dump(data), "byte representation wrong");
		}
		
		[Test]
		public void ClonePacket()
		{
			Packet p = new Packet();
			SetInfo(p);
			Packet clone = p.Clone();
			if (p == clone)
				Assert.Fail("Clone returned same reference");
			if (p.Header == clone.Header)
				Assert.Fail("Clone returned same header reference");
			CheckValues(clone);
		}
		
		[Test]
		public void PacketFromByte()
		{
			int offset = 3;
			byte[] data = FromString(correctPacket, offset);
			Packet p = new Packet(data, offset, data.Length - offset, new EncryptionStrategy());
			CheckValues(p);
		}
		
		private byte[] FromString(string str, int offset)
		{
			byte[] data = new byte[str.Length / 2 + offset];
			for (int i = 0; i < (str.Length / 2); ++i) {
				data[i + offset] = Convert.ToByte(str.Substring(i * 2, 2), 16);
			}
			return data;
		}
		
		private string Dump(byte[] data)
		{
			StringBuilder b = new StringBuilder();
			foreach(byte by in data) {
				b.Append(by.ToString("x2"));
			}
			return b.ToString();
		}
		
		private void SetInfo(Packet p)
		{
			SetInfo(p.Header);
			p.Header.EncryptionType = 0;
			p.AnswerId = 12349876;
			p.Target = "myCustomTarget";
			
			p.Values.AddValue("teststring");
			p.Values.AddValue(-100);
			p.Values.AddValue(150);
			p.Values.AddValue((uint)200);
			p.Values.SetValue(2, "150 €.");
			p.NamedValues.AddValue("key1", "value1");
			p.NamedValues.AddValue("key2", 0xF1F2F3F4); // uint!!
			p.NamedValues.AddValue("data", new byte[] {0x44, 0x55, 0x66, 0x77});
		}
		
		private void SetInfo(PacketHeader h)
		{
			h.PacketID = 20;
			h.DependsOn = 10;
			h.Destination = new int[] {5,6,7};
			h.EncryptionType = 5;
			h.Hops = new int[10];
			h.Hops[0] = 2;
			h.Hops[1] = 3;
			h.Source = 1;
		}
		
		private void CheckValues(Packet p)
		{
			Assert.AreEqual(0, p.Header.EncryptionType, "EncryptionType");
			p.Header.EncryptionType = 5;
			CheckValues(p.Header);
			Assert.AreEqual(12349876, p.AnswerId, "AnswerID");
			Assert.AreEqual("myCustomTarget", p.Target, "Target");
			
			Assert.AreEqual(4, p.Values.Count, "IndexedValueCount");
			Assert.AreEqual("teststring", p.Values.GetValue(0, typeof(string)), "teststring");
			Assert.AreEqual(-100, p.Values.GetValue(1, typeof(int)), "int -100");
			Assert.AreEqual(unchecked((uint)-100), p.Values.GetValue(1, typeof(uint)), "uint -100");
			Assert.AreEqual("150 €.", p.Values.GetValue(2, typeof(string)), "150 €");
			Assert.AreEqual(200, p.Values.GetValue(3, typeof(int)), "int 200");
			Assert.AreEqual(200, p.Values.GetValue(3, typeof(uint)), "uint 200");
			
			Assert.AreEqual(3, p.NamedValues.Count, "NamedValueCount");
			Assert.AreEqual("value1", p.NamedValues.GetValue("key1", typeof(string)), "key1");
			Assert.AreEqual(0xF1F2F3F4, p.NamedValues.GetValue("key2", typeof(uint)), "key2");
			Assert.AreEqual("44556677", Dump((byte[])p.NamedValues.GetValue("data", typeof(byte[]))), "data (key3)");
		}
		
		private void CheckValues(PacketHeader h)
		{
			Assert.AreEqual(10, h.DependsOn, "DependsOn");
			Assert.AreEqual(3, h.Destination.Length, "Destination.Length");
			Assert.AreEqual(5, h.Destination[0], "Destination[0]");
			Assert.AreEqual(6, h.Destination[1], "Destination[1]");
			Assert.AreEqual(7, h.Destination[2], "Destination[2]");
			Assert.AreEqual(5, h.EncryptionType, "EncryptionType");
			Assert.AreEqual(10, h.Hops.Length, "Hops.Length");
			Assert.AreEqual(1, h.Source, "Source");
			Assert.AreEqual(2, h.Hops[0], "Hops[0]");
			Assert.AreEqual(3, h.Hops[1], "Hops[1]");
			for(int i = 2; i < 10; ++i) {
				Assert.AreEqual(0, h.Hops[i], "Hops[" + i.ToString() + "]");
			}
		}
	}
}
