// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Net;
using System.Text;
using NUnit.Framework;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;

namespace Grunwald.P2PLib.Tests.Server
{
	[TestFixture]
	public class HttpTest
	{
		#region SetUp
		HttpConnectionHandler handler;
		string dir;
		
		[SetUp]
		public void Init() {
			handler = new HttpConnectionHandler();
			handler.RequestHandlers.Add(new SmallFileHandler());
			handler.RequestHandlers.Add(new PrimeHandler());
			dir = MainClass.TestDataDir + "http/";
		}
		#endregion
		
		#region Tests
		/*
		[Test]
		public void TestAll() {
			string
			int count = 0;
			foreach(string file in Directory.GetFiles(dir, "*.txt")) {
				ExecuteTestFile(file);
				count += 1;
			}
			if (count == 0)
				Assert.Fail("No test files found.");
		}
		 */
		
		[Test, Ignore("Accept not programmed yet")]
		public void AcceptTest() {
			ExecuteTestFile(dir + "Accept.txt");
		}
		
		[Test]
		public void BasicTest() {
			ExecuteTestFile(dir + "BasicTest.txt");
		}
		
		[Test]
		public void ContentMD5Test() {
			ExecuteTestFile(dir + "Content-MD5.txt");
		}
		
		[Test]
		public void ETAGTest() {
			ExecuteTestFile(dir + "ETAG.txt");
		}
		
		[Test]
		public void RangeTest() {
			ExecuteTestFile(dir + "RangeTest.txt");
		}
		#endregion
		
		#region ParseTestFile
		void ExecuteTestFile(string file) {
			try {
				ParseTestFile(file);
			} catch (Exception ex) {
				throw new AssertionException("Test file " + Path.GetFileNameWithoutExtension(file) + " failed:", ex);
			}
		}
		
		void ParseTestFile(string file) {
			using (StreamReader r = File.OpenText(file)) {
				string title = null;
				StringBuilder request = null;
				StringBuilder response = null;
				string line;
				int mode = 0;
				int emptyResponseLines = 0;
				// 1 = request
				// 2 = response
				while ((line = r.ReadLine()) != null) {
					if (line.StartsWith("==")) {
						if (title != null)
							ExecuteTest(title, request.ToString(), response.ToString());
						title = line;
						mode = 1;
						emptyResponseLines = 0;
						request = new StringBuilder();
						response = new StringBuilder();
					} else if (mode == 1) {
						if (line == "") {
							mode = 2;
						} else {
							request.Append(line);
							request.Append("\r\n");
						}
					} else if (mode == 2) {
						if (line == "") {
							emptyResponseLines += 1;
						} else {
							for (; emptyResponseLines > 0; --emptyResponseLines) {
								response.Append("\r\n");
							}
							response.Append(line);
							response.Append("\r\n");
						}
					} else {
						Assert.Fail("Invalid test file");
					}
				}
				ExecuteTest(title, request.ToString(), response.ToString());
			}
		}
		#endregion
		
		#region ExecuteTest
		void ExecuteTest(string title, string request, string response) {
			string responseGot = "";
			try {
				LocalConnection client, server;
				LocalConnection.Create(out client, out server);
				client.Writer.WriteLine(request);
				client.Writer.Flush();
				StringCollection requestCol = HttpServer.GetRequest(server);
				handler.Handle(server, requestCol);
				server.Close();
				responseGot = client.Reader.ReadToEnd();
				client.Close();
				// compare response got to response expected
				string[] gotLines = responseGot.Replace("\r\n", "\n").Split('\n');
				string[] expectLines = response.Replace("\r\n", "\n").Split('\n');
				if (response.StartsWith("Status: ")) {
					string expectStart = "HTTP/1.1 " + expectLines[0].Substring(8) + " ";
					if (!responseGot.StartsWith(expectStart))
						Assert.Fail("First line does not start with '" + expectStart + "', but is " + gotLines[0]);
					CheckHeaders(gotLines, expectLines);
				} else if (response.StartsWith("HTTP/1.1 ")) {
					Assert.AreEqual(expectLines[0], gotLines[0], "First line is wrong");
					int pos = CheckHeaders(gotLines, expectLines);
					bool headers = true;
					for (int i = 1; i < expectLines.Length; ++i) {
						if (expectLines[i] == "")
							headers = false;
						else if (headers)
							continue;
						pos += 1;
						if (pos >= gotLines.Length) {
							if (expectLines[i] != "")
								Assert.Fail("actual response is not long enough for checking '" + expectLines[i] + "'");
							else
								break;
						}
						Assert.AreEqual(expectLines[i], gotLines[pos], "data line is wrong");
					}
					for (int i = pos; i < gotLines.Length; ++i) {
						Assert.AreEqual("", gotLines[i], "there was a line too much returned");
					}
				} else {
					Assert.Fail("Invalid expected response: " + response);
				}
			} catch (Exception ex) {
				throw new AssertionException("Error executing test " + title + "\nRequest was:\n" + request + "\n\nResponse was:\n" + responseGot, ex);
			}
		}
		
		// checks the headers, returns the number of got header lines
		int CheckHeaders(string[] gotLines, string[] expectLines) {
			WebHeaderCollection h = new WebHeaderCollection();
			int count = 0;
			for (int i = 1; i < gotLines.Length; ++i) {
				string line = gotLines[i];
				if (line == "")
					break;
				h.Add(line);
				count += 1;
			}
			for (int i = 1; i < expectLines.Length; ++i) {
				string line = expectLines[i];
				if (line == "")
					break;
				int pos = line.IndexOf(':');
				if (pos < 0)
					Assert.Fail("No ':' in expected line " + line);
				string name = line.Substring(0, pos);
				Assert.AreEqual(' ', line[pos+1], "char after : in expected line is no space");
				string expectVal = line.Substring(pos + 2);
				if (expectVal == "missing") {
					if (h[name] != null)
						Assert.Fail("Header " + name + " does exist, but should be missing");
				} else {
					if (h[name] == null)
						Assert.Fail("Header " + name + " does not exist, but should be there");
					if (expectVal != "*")
						Assert.AreEqual(expectVal, h[name], "Header " + name + " has wrong value");
				}
			}
			return count;
		}
		#endregion
	}
	
	#region SmallFileHandler
	public class SmallFileHandler : IRequestHandler
	{
		public void Handle(HttpRequest request) {
			string text = "This is the content of the small text file.\r\n";
			MemoryStream ms = new MemoryStream(Encoding.UTF8.GetBytes(text));
			FileUpload upload = new FileUpload(request);
			upload.Headers["Content-Type"] = "text/plain; charset=UTF-8";
			upload.Send(ms);
		}
		
		public int GetPriority(HttpRequest request) {
			if (request.File == "/small.txt")
				return 1;
			else
				return Int32.MinValue;
		}
	}
	#endregion
	
	#region PrimeHandler
	public class PrimeHandler : IRequestHandler
	{
		public void Handle(HttpRequest request) {
			StringBuilder builder = new StringBuilder();
			int max = int.Parse(request.Request["to"]);
			builder.Append("2\r\n");
			for (int i = 3; i < max; i += 2) {
				int sqrMax = (int)Math.Sqrt(i);
				bool ok = true;
				for (int j = 3; j < sqrMax; j += 2) {
					if (i % j == 0) {
						ok = false;
						break;
					}
				}
				if (ok) {
					builder.Append(i.ToString(System.Globalization.CultureInfo.InvariantCulture));
					builder.Append("\r\n");
				}
			}
			MemoryStream ms = new MemoryStream(Encoding.UTF8.GetBytes(builder.ToString()));
			FileUpload upload = new FileUpload(request);
			upload.Headers["Content-Type"] = "text/plain; charset=UTF-8";
			upload.Send(ms);
		}
		
		public int GetPriority(HttpRequest request) {
			if (request.File == "/primes")
				return 1;
			else
				return Int32.MinValue;
		}
	}
	#endregion
}
