// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using Grunwald.P2PLib.Server;
using Grunwald.P2PLib.Server.Http;
using log4net;
using NUnit.Framework;

namespace Grunwald.P2PLib.Tests.Server
{
	[TestFixture]
	public class ServerTest
	{
		public static int TestPort = 11241;
		
		#region SetUp
		HttpLikeServer server = new HttpLikeServer();
		
		internal static readonly ILog log = LogManager.GetLogger(typeof (ServerTest));
		
		[SetUp]
		public void Init()
		{
			MainClass.Init();
			server.Port = TestPort;
			server.Start();
		}
		#endregion
		
		#region Test HTTP
		[Test]
		public void TestHTTPServer()
		{
			server.ConnectionHandlers.Clear();
			HttpConnectionHandler handler = new HttpConnectionHandler();
			TestRequestHandler test = new TestRequestHandler();
			handler.RequestHandlers.Add(test);
			server.ConnectionHandlers.Add(handler);
			WebClient wc = new WebClient();
			UriBuilder b = new UriBuilder("http", "localhost", server.Port, "/test");
			b.Query = "vorname=Daniel%20%26%20Tim&nachname=Grünwald%21";
			string url = b.ToString();
			byte[] by = wc.DownloadData(url);
			wc.Dispose();
			string answer = Encoding.UTF8.GetString(by);
			Assert.AreEqual("Daniel & Tim Grünwald!", test.Name, "wrong name");
			Assert.AreEqual("Answer", answer, "wrong answer");
			handler.RequestHandlers.Remove(test);
			server.ConnectionHandlers.Remove(handler);
		}
		#endregion
		
		#region Portchange-Test
		bool eventFired = false;
		
		[Test]
		public void TestPortChange()
		{
			if (!server.Running)
				Assert.Fail("Server is not running");
			server.PortChanged += new EventHandler(Server_PortChanged);
			server.Port += 1;
			if (!eventFired)
				Assert.Fail("Event didn't fire");
			if (!server.Running)
				Assert.Fail("Server is not running");
		}
		
		private void Server_PortChanged(object sender, EventArgs e)
		{
			eventFired = true;
		}
		#endregion
		
		#region Test double port use
		[Test]
		[ExpectedException(typeof (SocketException))]
		public void TestDoublePort()
		{
			HttpLikeServer s = new HttpLikeServer();
			s.Port = server.Port;
			s.Start();
			Assert.Fail("This should not be reached.");
		}
		#endregion
		
		#region Test proper stopping of server
		[Test]
		public void TestServerStop()
		{
			HttpLikeServer s = new HttpLikeServer();
			s.Port = server.Port;
			// This should give no port conflict:
			server.Stop();
			s.Start();
			s.Stop();
			server.Start();
		}
		#endregion
		
		#region Test ServerAbilityTester
		#region Test Successfull Run
		[Test]
		public void TestServerAbility()
		{
			server.ConnectionHandlers.Clear();
			HttpConnectionHandler handler = new HttpConnectionHandler();
			PingRequestHandler h = new PingRequestHandler();
			handler.RequestHandlers.Add(h);
			server.ConnectionHandlers.Add(handler);
			ServerAbilityTester t = new ServerAbilityTester();
			IPEndPoint endPoint = new IPEndPoint(IPAddress.Loopback, server.Port);
			if (!t.Test(endPoint, h.ServerId))
				Assert.Fail("Test returned wrong value");
			
			if (!t.Test(endPoint, 0))
				Assert.Fail("Test returned wrong value");
			
			if (t.Test(endPoint, h.ServerId + 1))
				Assert.Fail("Test returned true though the ServerID was wrong");
			server.ConnectionHandlers.Clear();
		}
		#endregion
		
		#region Test Failing Run
		[Test]
		public void TestFailServerAbility()
		{
			server.ConnectionHandlers.Clear();
			HttpConnectionHandler handler = new HttpConnectionHandler();
			server.ConnectionHandlers.Add(handler);
			ServerAbilityTester t = new ServerAbilityTester();
			IPEndPoint endPoint = new IPEndPoint(IPAddress.Loopback, server.Port);
			if (t.Test(endPoint, 0))
				Assert.Fail("Test 1 returned wrong value");
			server.ConnectionHandlers.Remove(handler);
			if (t.Test(endPoint, 0))
				Assert.Fail("Test 2 returned wrong value");
			server.Stop();
			if (t.Test(endPoint, 0))
				Assert.Fail("Test 3 returned wrong value");
			server.ConnectionHandlers.Clear();
			server.Start();
		}
		#endregion
		#endregion
		
		#region TearDown
		[TearDown()]
		public void Shutdown()
		{
			server.Stop();
		}
		#endregion
	}
	
	#region TestRequestHandler
	public class TestRequestHandler : IRequestHandler
	{
		public string Name = "";
		
		public void Handle(HttpRequest request)
		{
			HttpResponse response = new HttpResponse(request);
			response.Server = "P2PLib Test";
			response.ContentLength = 6;
			response.Headers.Set("Testheader", "Wert");
			StreamWriter w = request.Connection.Writer;
			response.SendHeaders(w);
			Name = request.Request["vorname"] + " " + request.Request["nachname"];
			//HTTPServerTest.log.Info(request.ToString());
			w.Write("Answer");
			w.Flush();
			w.Close();
		}
		
		public int GetPriority(HttpRequest request)
		{
			if (request.File == "/test")
				return 1;
			else
				return Int32.MinValue;
		}
	}
	#endregion
}
