// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Globalization;
using System.IO;
using System.Text;
using Microsoft.Win32;
using log4net;

namespace Grunwald.P2PLib
{
	/// <summary>
	/// Class for static helping functions
	/// </summary>
	public sealed class Helper
	{
		private Helper() { }
		
		#region Logging
		static readonly ILog log = LogManager.GetLogger(typeof(Helper));
		
		/// <summary>
		/// Uses log4net to log the specified exceptions.
		/// ThreadAbortExceptions are ignored.
		/// </summary>
		public static void LogError(Exception ex) {
			if (IsCritical(ex)) return;
			if (ex is System.Threading.ThreadAbortException) return;
			log.Error(ex);
		}
		
		/// <summary>
		/// Uses log4net to log the specified object.
		/// </summary>
		public static void Info(object o)
		{
			log.Debug(o);
		}
		#endregion
		
		#region Random
		static Random random;
		/// <summary>
		/// <para>Returns an instance of System.Random. This property is initialized on
		/// first call.</para>
		/// <para>Warning: This random number generator can not be used for generating
		/// cryptographic keys!</para>
		/// </summary>
		public static Random Random {
			get {
				if (random == null) random = new Random();
				return random;
			}
		}
		
		/// <summary>
		/// Gets a random integer value. The value can be negative or positive, but not zero.
		/// </summary>
		public static int GetRandomInt()
		{
			int val = Helper.Random.Next(int.MinValue, int.MaxValue);
			if (val == 0)
				val = int.MaxValue;
			return val;
		}
		#endregion
		
		#region IsCritical(Exception)
		
		/// <summary>
		/// Checks if the exception is a critical (=non handable) exception that
		/// should not be caught.
		/// </summary>
		public static bool IsCritical(Exception ex)
		{
			if (ex is OutOfMemoryException)
				return true;
			if (ex is AppDomainUnloadedException)
				return true;
			if (ex is BadImageFormatException)
				return true;
			if (ex is CannotUnloadAppDomainException)
				return true;
			if (ex is ExecutionEngineException)
				return true;
			if (ex is InvalidProgramException)
				return true;
			return false;
		}
		
		#endregion
		
		#region Escape
		
		// NOTE: I know System.Web.HttpUtility, but Microsoft has a bug
		//  regarding the character '#'. Read the RFC and compare it to the
		//  result of this method and Microsoft's method.
		//
		// The character "#" is unsafe and should
		// always be encoded because it is used in World Wide Web and in other
		// systems to delimit a URL from a fragment/anchor identifier that might
		// follow it.
		//
		// WHIDBEY: In .NET 2.0 Beta this bug is fixed so this methods will
		// become obsolete.
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// a part of an URL.
		/// (";", "/", "?", ":", "@", "=" and "&amp;" are encoded)</para>
		/// <para>Non-ascii characters are encoded using UTF-8.</para>
		/// </summary>
		public static string EscapePart(string str)
		{
			return Escape(str, "$-_.!*'(),");
		}
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// a part of an URL.
		/// (";", "/", "?", ":", "@", "=" and "&amp;" are encoded)</para>
		/// <para>Non-ascii characters are encoded using the specified encoding.</para>
		/// </summary>
		public static string EscapePart(string str, Encoding encoding)
		{
			return Escape(str, "$-_.!*'(),", encoding);
		}
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// URL. Scheme-specific characters for URLs are not encoded.</para>
		/// <para>Non-ascii characters are encoded using UTF-8.</para>
		/// </summary>
		public static string Escape(string str)
		{
			return Escape(str, "$-_.!*'(),;/?:@=&");
		}
		
		/// <summary>
		/// <para>Escapes (as defined in RFC 1738) the string so that it can be used as
		/// URL. Scheme-specific characters for URLs are not encoded.</para>
		/// <para>Non-ascii characters are encoded using the specified encoding.</para>
		/// </summary>
		public static string Escape(string str, Encoding encoding)
		{
			return Escape(str, "$-_.!*'(),;/?:@=&", encoding);
		}
		
		/// <summary>
		/// Escapes a (part of an) URL as described in RFC 1738. Special characters
		/// are encoded using UTF-8.
		/// </summary>
		public static string Escape(string str, string allowedChars)
		{
			return Escape(str, allowedChars, Encoding.UTF8);
		}
		
		/// <summary>
		/// Escapes a (part of an) URL as described in RFC 1738. Special characters
		/// are encoded using the specified encoding.
		/// </summary>
		public static string Escape(string str, string allowedChars, Encoding encoding)
		{
			if (allowedChars == null)
				throw new ArgumentNullException("allowedChars");
			if (str == null)
				return null;
			if (str.Length == 0)
				return String.Empty;
			byte[] data = encoding.GetBytes(str);
			StringBuilder b = new StringBuilder(data.Length + 10);
			char c;
			for (int i = 0; i < data.Length; ++i) {
				if (data[i] >= 128) {
					b.AppendFormat("%{0:x02}", data[i]);
				} else {
					c = (char)data[i];
					if (char.IsLetterOrDigit(c)) {
						b.Append(c);
					} else if (allowedChars.IndexOf(c) >= 0) {
						b.Append(c);
					} else {
						b.AppendFormat("%{0:x02}", data[i]);
					}
				}
			}
			return b.ToString();
		}
		
		#endregion
		
		#region Unescape
		
		/// <summary>
		/// Unescapes a (part of an) URL as described in RFC 1738. Special characters
		/// are decoded using UTF8.
		/// </summary>
		public static string Unescape(string str)
		{
			return Unescape(str, Encoding.UTF8);
		}
		
		/// <summary>
		/// Unescapes a (part of an) URL as described in RFC 1738. Special characters
		/// are decoded using the specified encoding.
		/// </summary>
		public static string Unescape(string str, Encoding e)
		{
			if (str == null)
				return null;
			if (str.Length == 0)
				return String.Empty;
			int length = str.Length;
			for (int i = 0; i < str.Length; ++i) {
				if (str[i] == '%')
					length -= 2;
			}
			byte[] data = new byte[length];
			char c;
			int pos = 0;
			for (int i = 0; i < str.Length; ++i) {
				c = str[i];
				if (c == '+') {
					data[pos++] = 32;
				} else if (c == '%') {
					try {
						string hex = str.Substring(i + 1, 2);
						int num = Int32.Parse(hex, NumberStyles.HexNumber, CultureInfo.InvariantCulture);
						data[pos++] = (byte) num;
						i += 2;
					} catch (FormatException) {
					}
				} else {
					data[pos++] = (byte) c;
				}
			}
			return e.GetString(data);
		}
		
		#endregion
		
		#region Hexdump
		/// <summary>
		/// Converts the specified byte array into a displayable string for debugging
		/// purposes.
		/// </summary>
		public static string HexDump(byte[] data)
		{
			StringBuilder b = new StringBuilder();
			for(int i = 0; i < data.Length; ++i) {
				if (i > 0) {
					if (i % 4 == 0) b.Append(" ");
					if (i % 16 == 0) {
						for (int j = i - 16; j < i; ++j) {
							byte by = data[j];
							if (by < 32) {
								b.Append(".");
							} else {
								char c = (char)by;
								if (Char.IsControl(c))
									b.Append('.');
								else if (Char.IsWhiteSpace(c))
									b.Append(' ');
								else
									b.Append(c);
							}
						}
						b.Append("\n");
					}
				}
				b.Append(data[i].ToString("x2"));
			}
			return b.ToString();
		}
		#endregion
		
		/// <summary>
		/// Gets the content type of the specified file.
		/// </summary>
		public static string GetContentType(string file)
		{
			if (!Path.HasExtension(file))
				return "text/plain";
			string ext = Path.GetExtension(file).ToLower();
			switch (ext) {
				case ".htm":
				case ".html":
					return "text/html";
				case ".txt":
					return "text/plain";
				case ".gif":
					return "image/gif";
				case ".jpg":
				case ".jpeg":
					return "image/jpeg";
				case ".png":
					return "image/png";
				case ".js":
					return "text/javascript";
				case ".css":
					return "text/css";
				case ".ico":
					return "image/x-icon";
				default:
					RegistryKey k = Registry.ClassesRoot.OpenSubKey(ext);
					if (k == null)
						return "application/octet-stream";
					else
						return (string)k.GetValue("Content Type", "application/octet-stream");
			}
		}
	}
}
