// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Collections.Specialized;
using System.Text;
using System.Security.Cryptography;
using Grunwald.P2PLib.Network;

namespace Grunwald.P2PLib
{
	/// <summary>
	/// Describes the unique identity of an user.
	/// Contains the user's keys and the collection of ReceivingUsers.
	/// </summary>
	public class Identity
	{
		bool isOwnIdentity;
		
		/// <summary>
		/// Gets if this identity is controlled on this computer (private key is known) (=true)
		/// or if this is the identity of someone else (only public key known) (=false)
		/// </summary>
		public bool IsOwnIdentity {
			get {
				return isOwnIdentity;
			}
		}
		
		#region Constructors
		/// <summary>
		/// Creates a new Identity.
		/// </summary>
		/// <param name="keySize">The key size to use (in bit)</param>
		public Identity(int keySize)
		{
			rsaProvider = new RSACryptoServiceProvider(keySize);
			isOwnIdentity = true;
		}
		
		/// <summary>
		/// Creates a new Identity using the specified public key.
		/// </summary>
		public Identity(string publicKey)
		{
			if (publicKey == null)
				throw new ArgumentNullException("publicKey");
			isOwnIdentity = false;
			int pos = publicKey.IndexOf(",");
			if (pos <= 0)
				throw new ArgumentException("The specified key is no valid public key.", "publicKey");
			string exponent = publicKey.Substring(0, pos);
			string modulus = publicKey.Substring(pos + 1);
			byte[] exp = Convert.FromBase64String(exponent);
			byte[] mod = Convert.FromBase64String(modulus);
			int keySize = mod.Length * 8;
			rsaProvider = new RSACryptoServiceProvider();
			RSAParameters para;
			para.D = null;
			para.DP = null;
			para.DQ = null;
			para.InverseQ = null;
			para.P = null;
			para.Q = null;
			para.Exponent = exp;
			para.Modulus = mod;
			rsaProvider.ImportParameters(para);
		}
		#endregion
		
		#region RSAProvider / PublicKey
		RSACryptoServiceProvider rsaProvider;
		
		/// <summary>
		/// Gets the base64 representation of the user's public key.
		/// (first 4 characters/3 byte is the exponent, the rest is the modulus)
		/// </summary>
		public virtual string PublicKey {
			get {
				RSAParameters para = rsaProvider.ExportParameters(false);
				string exp = Convert.ToBase64String(para.Exponent);
				return exp + "," + Convert.ToBase64String(para.Modulus);
			}
		}
		
		/// <summary>
		/// Gets the signature for the specified data.
		/// </summary>
		public byte[] Sign(byte[] data)
		{
			if (!this.IsOwnIdentity)
				throw new InvalidOperationException("Only own identities can sign.");
			return rsaProvider.SignData(data, typeof(SHA1CryptoServiceProvider));
		}
		
		/// <summary>
		/// Checks the signature.
		/// </summary>
		public bool CheckSignature(byte[] data, byte[] signature)
		{
			return rsaProvider.VerifyData(data, typeof(SHA1CryptoServiceProvider), signature);
		}
		
		/// <summary>
		/// Encrypt data so that only the owner of this identity can decrypt it.
		/// </summary>
		public byte[] Encrypt(byte[] data)
		{
			return rsaProvider.Encrypt(data, false);
		}
		
		/// <summary>
		/// Decrypt data.
		/// </summary>
		public byte[] Decrypt(byte[] data)
		{
			if (!this.IsOwnIdentity)
				throw new InvalidOperationException("Only own identities can decrypt.");
			return rsaProvider.Decrypt(data, false);
		}
		
		/// <summary>
		/// Gets the string representation of this Identity.
		/// </summary>
		public override string ToString()
		{
			return "[" + GetType().FullName + " Identity: " + PublicKey + "]";
		}
		#endregion
		
		#region Symmetric key cache
		// MD5(key) -> Key
		ListDictionary symmetricKeyCache = new ListDictionary(); // byte[] -> byte[]
		
		/// <summary>
		/// Gets a symmetric key that can be used to communicate with this user.
		/// </summary>
		public byte[] GetSymmetricKey()
		{
			lock (symmetricKeyCache) {
				if (symmetricKeyCache.Count == 0) {
					byte[] key;
					using (RijndaelManaged aes = new RijndaelManaged()) {
						aes.KeySize = 256;
						aes.GenerateKey();
						aes.GenerateIV();
						key = new byte[aes.Key.Length + aes.IV.Length];
						symmetricKeyCache.Add(GetSymmetricKeyHash(key), key);
					}
					return key;
				} else {
					foreach (byte[] key in symmetricKeyCache.Values) {
						return key;
					}
					throw new ApplicationException();
				}
			}
		}
		
		/// <summary>
		/// Gets the number of keys registered for this identity.
		/// </summary>
		public int KeyCount {
			get {
				lock (symmetricKeyCache) {
					return symmetricKeyCache.Count;
				}
			}
		}
		
		/// <summary>
		/// Gets the symmetric key with the specified hash.
		/// null if the key is not found.
		/// </summary>
		public byte[] GetSymmetricKey(byte[] keyHash)
		{
			lock (symmetricKeyCache) {
				return (byte[])symmetricKeyCache[keyHash];
			}
		}
		
		/// <summary>
		/// Adds the specified symmetric key to the cache so it can be read using its hash
		/// later.
		/// </summary>
		public void AddSymmetricKey(byte[] key)
		{
			lock (symmetricKeyCache) {
				symmetricKeyCache.Add(GetSymmetricKeyHash(key), key);
			}
		}
		
		/// <summary>
		/// Gets the hash of the specified symmetric key.
		/// </summary>
		public byte[] GetSymmetricKeyHash(byte[] key)
		{
			using (MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider()) {
				return md5.ComputeHash(key);
			}
		}
		
		/// <summary>
		/// Gets the length of the key hashes in bytes.
		/// </summary>
		public int KeyHashLength {
			get {
				return 16;
			}
		}
		#endregion
	}
}
