// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Net;
using System.Text;

namespace Grunwald.P2PLib.Network
{
	/// <summary>
	/// Contains a collection of <see cref="AddressEntry">AddressEntries</see>.
	/// </summary>
	public class AddressCollection : IEnumerable
	{
		ArrayList adr = new ArrayList();
		
		#region Add
		
		/// <summary>
		/// Add the entry to the AddressCollection.
		/// If an entry with the same <see cref="Identity"/> or
		/// <see cref="P2PAddress"/> already exists, it will be overwritten.
		/// </summary>
		public void Add(AddressEntry e)
		{
			if (e.Identity != null) {
				for (int i = 0; i < adr.Count; ++i) {
					AddressEntry a = (AddressEntry) adr[i];
					if (a.Identity == e.Identity) {
						adr[i] = e;
						return;
					}
				}
			} else {
				for (int i = 0; i < adr.Count; ++i) {
					AddressEntry a = (AddressEntry) adr[i];
					if (a.Address == e.Address) {
						adr[i] = e;
						return;
					}
				}
			}
			adr.Add(e);
		}
		
		/// <summary>
		/// Sets the <see cref="AddressEntry.EndPoint"/> property of the entry with
		/// the specified identity. If the entry does not exist, it will be created.
		/// </summary>
		public void Set(Identity identity, EndPoint e)
		{
			AddressEntry a = this[identity];
			if (a == null)
				Add(new AddressEntry(identity, e));
			else
				a.EndPoint = e;
		}
		
		#endregion
		
		#region Remove
		
		/// <summary>Removes the given entry from the collection</summary>
		public void Remove(AddressEntry e)
		{
			if (e == null)
				throw new ArgumentNullException("e");
			adr.Remove(e);
		}
		
		/// <summary>Removes the entry of the user if it exists.</summary>
		public void Remove(Identity identity)
		{
			AddressEntry e = this[identity];
			if (e != null)
				Remove(e);
		}
		
		/// <summary>Removes all entries with the specified P2PAddress.</summary>
		public void Remove(P2PAddress address)
		{
			for (int i = 0; i < adr.Count; ++i) {
				AddressEntry e = (AddressEntry)adr[i];
				if (e.Address.Equals(address))
					adr.RemoveAt(i--);
			}
		}
		
		#endregion
		
		#region Indexer
		
		/// <summary>Return the <see cref="AddressEntry"/> of the given EndPoint.
		/// If the entry does not exist, null will be returned.</summary>
		public AddressEntry this[EndPoint e] {
			get {
				foreach (AddressEntry a in adr) {
					if (a.EndPoint == e)
						return a;
				}
				return null;
			}
		}
		
		/// <summary>Return the <see cref="AddressEntry"/> of the specified user.
		/// If the entry does not exist, null will be returned.</summary>
		public AddressEntry this[Identity identity] {
			get {
				if (identity == null) throw new ArgumentNullException("identity");
				foreach (AddressEntry e in adr) {
					if (e.Identity == identity)
						return e;
				}
				return null;
			}
		}
		
		#endregion
		
		#region Contains
		
		/// <summary>Returns true if this address cache contains the specified entry.</summary>
		public bool Contains(AddressEntry e)
		{
			return adr.Contains(e);
		}
		
		/// <summary>Returns true if this address cache contains an entry for this user.</summary>
		public bool Contains(Identity identity)
		{
			return this[identity] != null;
		}
		
		/// <summary>Returns true if this address cache contains an entry for this EndPoint.</summary>
		public bool Contains(EndPoint e)
		{
			return this[e] != null;
		}
		
		#endregion
		
		#region IEnumerable Member
		
		/// <summary>
		/// Gets the enumerator used to enumerate over this collection.
		/// </summary>
		public IEnumerator GetEnumerator()
		{
			return adr.GetEnumerator();
		}
		
		/// <summary>Converts all AddressEntries to strings and returns the strings.</summary>
		public override string ToString()
		{
			StringBuilder b = new StringBuilder();
			b.Append("[AddressCollection: {\n");
			foreach (AddressEntry address in this) {
				b.Append(address);
				b.Append('\n');
			}
			b.Append("}]\n");
			return b.ToString();
		}
		
		#endregion
		
		/// <summary>
		/// Gets the number of entries in the <see cref="AddressCollection"/>
		/// </summary>
		public int Count {
			get {
				return adr.Count;
			}
		}
	}
}
