// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using log4net;

namespace Grunwald.P2PLib.Network
{
	/// <summary>
	/// The ConnectionManager analyses the behaviour of a network's connections and
	/// builds new connections if neccessary.
	/// </summary>
	public class ConnectionManager
	{
		#region Properties
		static readonly ILog log = LogManager.GetLogger(typeof(ConnectionManager));
		
		ConnectionFactory connectionFactory = ConnectionFactory.DefaultFactory;
		
		/// <summary>
		/// <para>Gets/Sets the <see cref="ConnectionFactory"/> this ConnectionManager uses
		/// to create new connections.</para>
		/// <para>The default value is the value of ConnectionFactory.DefaultFactory
		/// at the time when this ConnectionManager was instanciated.</para>
		/// </summary>
		public ConnectionFactory ConnectionFactory {
			get { return connectionFactory; }
			set {
				if (value == null) throw new ArgumentNullException();
				connectionFactory = value;
			}
		}
		
		int minimalConnections = 2;
		int maximalConnections = 50;
		
		/// <summary>
		/// <para>Gets/Sets the minimal count of connections the network should have.
		/// If the real connection count is below this value, the manager tries to
		/// build additional connections.</para>
		/// <para>The default value is 2.</para>
		/// </summary>
		public int MinimalConnections {
			get { return minimalConnections; }
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException("minimalConnections", value, "The value can't be negative.");
				if (value > maximalConnections)
					maximalConnections = value;
				minimalConnections = value;
			}
		}
		
		/// <summary>
		/// <para>Gets/Sets the maximal count of connections the network should have.
		/// If the real connection count is above this value, the manager tries to
		/// close unused connections.</para>
		/// <para>The default value is 50.</para>
		/// </summary>
		public int MaximalConnections {
			get { return maximalConnections; }
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException("maximalConnections", value, "The value can't be negative.");
				if (value < minimalConnections)
					minimalConnections = value;
				maximalConnections = value;
			}
		}
		#endregion
		
		P2PNetwork network;
		
		/// <summary>
		/// Creates a new ConnectionManager for the specified network.
		/// </summary>
		public ConnectionManager(P2PNetwork network)
		{
			this.network = network;
		}
		
		/// <summary>Gets the network this ConnectionManager was created for.</summary>
		public P2PNetwork Network { get { return network; } }
		
		#region Work
		/// <summary>
		/// This method is called by the network's work method.
		/// </summary>
		public virtual void Work()
		{
			bool ok;
			lock (network) {
				ConnectionCollection col = network.Connections;
				if (col.Count < MinimalConnections) {
					ArrayList possible = new ArrayList();
					ArrayList deleted = new ArrayList();
					// build more connections
					foreach (AddressEntry e in network.HostCache) {
						ok = true;
						if (e.Tries > maximumTries) {
							deleted.Add(e);
							continue;
						}
						foreach (Connection c in col) {
							if (c.Partner.Equals(e.Address)) { ok = false; break; }
						}
						if (ok) possible.Add(e);
					}
					if (possible.Count > 0) {
						int nr = Helper.Random.Next(possible.Count);
						AddressEntry adr = (AddressEntry)possible[nr];
						BuildDelegate bc = new BuildDelegate(BuildConnection);
						bc.BeginInvoke(adr, null, null);
					}
					foreach (AddressEntry e in deleted) {
						network.HostCache.Remove(e);
					}
				} else if (col.Count > MaximalConnections) {
					// TODO: drop unused connections
				}
				
				DateTime now = DateTime.Now;
				foreach (IPServer ipServer in network.IPServers) {
					if (ipServer.RefreshNeeded)
						ipServer.BeginRefresh(network);
				}
			}
		}
		#endregion
		
		#region Failures/Maximum tries
		int maximumTries = 10;
		
		/// <summary>
		/// Gets/Sets the maximum count of failures after which the address is ignored.
		/// </summary>
		public int MaximumTries {
			get {
				return maximumTries;
			}
			set {
				if (value <= 0)
					throw new ArgumentOutOfRangeException("maximumTries", value, "The value must be positive.");
				maximumTries = value;
			}
		}
		#endregion
		
		private delegate void BuildDelegate(AddressEntry adr);
		
		/// <summary>
		/// Builds a connection to the specified address.
		/// </summary>
		public virtual void BuildConnection(AddressEntry adr)
		{
			try {
				adr.Tries += 1;
				// log.Debug("Trying to connect to {0} the {1}. time", adr.EndPoint, adr.Tries);
				IConnection connection = null;
				try {
					connection = connectionFactory.BuildConnection(adr.EndPoint);
				} catch (BuildConnectionException) {
					// ignore build errors
					log.Debug(String.Format("Failed to connect to {0}.", adr.EndPoint));
					return;
				}
				if (connection == null)
					throw new P2PLibException("Could not build the connection to " + adr.EndPoint.ToString());
				network.BuildConnection(connection, adr);
			} catch (Exception ex) {
				log.Fatal(ex);
				throw;
			}
		}
	}
}
