// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace Grunwald.P2PLib.Network
{
	/// <summary>
	/// Abstract class for accessing a source for getting the IP-addresses of
	/// partners in a network.
	/// </summary>
	public abstract class IPServer
	{
		#region Constructor
		
		/// <summary>
		/// Creates a new IPDetector
		/// </summary>
		protected IPServer()
		{
			delegateRefresh = new DelegateRefresh(Refresh);
		}
		
		#endregion
		
		#region LastRefresh
		DateTime lastRefresh;
		
		/// <summary>
		/// Returns the time when the last refresh was started.
		/// </summary>
		public virtual DateTime LastRefresh {
			get { return lastRefresh; }
		}
		
		#endregion
		
		#region LastSuccess
		DateTime lastSuccess;
		
		/// <summary>
		/// Returns the time when the last refresh ended successfull.
		/// </summary>
		public virtual DateTime LastSuccess {
			get { return lastSuccess; }
		}
		
		#endregion
		
		#region Refresh
		
		delegate bool DelegateRefresh(P2PNetwork network);
		
		DelegateRefresh delegateRefresh;
		
		/// <summary>
		/// Starts an asynchron refresh
		/// </summary>
		/// <param name="network">The <see cref="P2PNetwork"/> whose
		/// <see cref="P2PNetwork.HostCache"/> should be filled
		/// with new addresses.</param>
		public virtual IAsyncResult BeginRefresh(P2PNetwork network)
		{
			return BeginRefresh(network, null, null);
		}
		
		/// <summary>
		/// Starts an asynchron refresh
		/// </summary>
		public virtual IAsyncResult BeginRefresh(P2PNetwork network, AsyncCallback asyncCallback, object state)
		{
			return delegateRefresh.BeginInvoke(network, asyncCallback, state);
		}
		
		/// <summary>
		/// Waits for the end of an asynchron refresh.
		/// </summary>
		/// <returns>Returns true if the refresh was successfull.</returns>
		public virtual bool EndRefresh(IAsyncResult result)
		{
			return delegateRefresh.EndInvoke(result);
		}
		
		/// <summary>
		/// Refreshes the IP-Server. This fills the P2PNetwork with the addresses
		/// returned by the server.
		/// </summary>
		/// <param name="network">The <see cref="P2PNetwork"/> that should
		/// be filled with new addresses.</param>
		/// <returns>Returns true if the refresh was successfull.</returns>
		public bool Refresh(P2PNetwork network)
		{
			try {
				lastRefresh = DateTime.Now;
				bool success = DoRefresh(network);
				lastRefreshWasSuccessful = success;
				if (success)
					lastSuccess = DateTime.Now;
				OnRefreshFinished(EventArgs.Empty);
				return success;
			} catch (Exception ex) {
				Helper.LogError(ex);
				if (Helper.IsCritical(ex))
					throw;
				return false;
			}
		}
		
		/// <summary>
		/// Refreshes the IP-Server. Override this method to implement the IPServer.
		/// </summary>
		/// <param name="network">The <see cref="P2PNetwork"/> that should
		/// be filled with new addresses.</param>
		/// <returns>Returns true if the refresh was successfull.</returns>
		protected abstract bool DoRefresh(P2PNetwork network);
		
		#endregion
		
		#region LastRefreshWasSuccessfull
		bool lastRefreshWasSuccessful = false;
		
		/// <summary>
		/// Returns true if the last refresh was successful.
		/// </summary>
		public virtual bool LastRefreshWasSuccessful {
			get { return lastRefreshWasSuccessful; }
		}
		
		#endregion
		
		/// <summary>
		/// Gets if a refresh is needed.
		/// </summary>
		public abstract bool RefreshNeeded { get; }
		
		/// <summary>
		/// Is raised when a refresh has finished (either successful or unsuccessful)
		/// </summary>
		public event EventHandler RefreshFinished;
		
		/// <summary>
		/// Raises the RefreshFinished event.
		/// </summary>
		protected virtual void OnRefreshFinished(EventArgs e) {
			if (RefreshFinished != null) {
				RefreshFinished(this, e);
			}
		}
	}
}
