// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace Grunwald.P2PLib.Network
{
	/// <summary>
	/// Represents a collection of <see cref="IPServer"/> objects.
	/// </summary>
	public class IPServerCollection : CollectionBase
	{
		#region Constructor
		
		/// <summary>
		/// Initializes a new instance of the <see cref="IPServerCollection"/> class.
		/// </summary>
		public IPServerCollection()
		{
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="IPServerCollection"/> class containing the elements of the specified source collection.
		/// </summary>
		/// <param name="value">A <see cref="IPServerCollection"/> with which to initialize the collection.</param>
		public IPServerCollection(IPServerCollection value)
		{
			this.AddRange(value);
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="IPServerCollection"/> class containing the specified array of <see cref="IPServer"/> objects.
		/// </summary>
		/// <param name="value">An array of <see cref="IPServer"/> objects with which to initialize the collection. </param>
		public IPServerCollection(IPServer[] value)
		{
			this.AddRange(value);
		}
		
		#endregion
		
		#region Indexer
		
		/// <summary>
		/// Gets the <see cref="IPServer"/> at the specified index in the collection.
		/// <para>
		/// In C#, this property is the indexer for the <see cref="IPServerCollection"/> class.
		/// </para>
		/// </summary>
		public IPServer this[int index] {
			get { return (IPServer) (this.List[index]); }
		}
		
		#endregion
		
		#region Add
		
		/// <summary>
		/// Adds an item to the end of the collection
		/// </summary>
		/// <param name="value">Item to add</param>
		/// <returns>The positition where the new element has been added.</returns>
		public int Add(IPServer value)
		{
			return this.List.Add(value);
		}
		
		/// <summary>
		/// Copies the elements of the specified <see cref="IPServer"/> array to the end of the collection.
		/// </summary>
		/// <param name="value">An array of type <see cref="IPServer"/> containing the objects to add to the collection.</param>
		public void AddRange(IPServer[] value)
		{
			for (int i = 0; (i < value.Length); i = (i + 1)) {
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		/// Adds the contents of another <see cref="IPServerCollection"/> to the end of the collection.
		/// </summary>
		/// <param name="value">A <see cref="IPServerCollection"/> containing the objects to add to the collection. </param>
		public void AddRange(IPServerCollection value)
		{
			for (int i = 0; (i < value.Count); i = (i + 1)) {
				this.Add((IPServer) value.List[i]);
			}
		}
		
		#endregion
		
		#region Contains
		
		/// <summary>
		/// Gets a value indicating whether the collection contains the specified <see cref="IPServer"/>.
		/// </summary>
		/// <param name="value">The <see cref="IPServer"/> to search for in the collection.</param>
		/// <returns><b>true</b> if the collection contains the specified object; otherwise, <b>false</b>.</returns>
		public bool Contains(IPServer value)
		{
			return this.List.Contains(value);
		}
		
		#endregion
		
		#region ToArray
		
		/// <summary>
		/// Copies the collection objects to a one-dimensional <see cref="System.Array">Array</see> instance beginning at the specified index.
		/// </summary>
		/// <param name="array">The one-dimensional <see cref="System.Array">Array</see> that is the destination of the values copied from the collection.</param>
		/// <param name="index">The index of the array at which to begin inserting.</param>
		public void CopyTo(IPServer[] array, int index)
		{
			this.List.CopyTo(array, index);
		}
		
		/// <summary>
		/// Creates a one-dimensional <see cref="System.Array">Array</see> instance containing the collection items.
		/// </summary>
		/// <returns>Array of type IPServer</returns>
		public IPServer[] ToArray()
		{
			IPServer[] array = new IPServer[this.Count];
			this.CopyTo(array, 0);
			
			return array;
		}
		
		#endregion
		
		#region IndexOf
		
		/// <summary>
		/// Gets the index in the collection of the specified <see cref="IPServer"/>, if it exists in the collection.
		/// </summary>
		/// <param name="value">The <see cref="IPServer"/> to locate in the collection.</param>
		/// <returns>The index in the collection of the specified object, if found; otherwise, -1.</returns>
		public int IndexOf(IPServer value)
		{
			return this.List.IndexOf(value);
		}
		
		#endregion
		
		#region Insert
		
		/// <summary>
		/// Inserts a new item into the collection.
		/// </summary>
		/// <param name="index">Position, whrere the item should be added</param>
		/// <param name="value">Item to add</param>
		public void Insert(int index, IPServer value)
		{
			List.Insert(index, value);
		}
		
		#endregion
		
		#region Remove
		
		/// <summary>
		/// Removes the specified item from the collection
		/// </summary>
		/// <param name="value">Item to remove</param>
		public void Remove(IPServer value)
		{
			List.Remove(value);
		}
		
		/// <summary>
		/// Removes the item at the given position
		/// </summary>
		/// <param name="index">Position of item</param>
		public new void RemoveAt(int index)
		{
			base.RemoveAt(index);
		}
		
		/// <summary>
		/// Clears all items from the collection
		/// </summary>
		public new void Clear()
		{
			base.Clear();
		}
		
		#endregion
		
		#region Enumerator
		
		/// <summary>
		/// Returns an enumerator that can iterate through the <see cref="IPServerCollection"/> instance.
		/// </summary>
		/// <returns>An <see cref="IPServerCollectionEnumerator">IPServerCollectionEnumerator</see> for the <see cref="IPServerCollection"/> instance.</returns>
		public new IPServerCollectionEnumerator GetEnumerator()
		{
			return new IPServerCollectionEnumerator(this);
		}
		
		/// <summary>
		/// Supports a simple iteration over a <see cref="IPServerCollection"/>.
		/// </summary>
		public class IPServerCollectionEnumerator : IEnumerator
		{
			private IEnumerator _enumerator;
			private IEnumerable _temp;
			
			/// <summary>
			/// Initializes a new instance of the <see cref="IPServerCollectionEnumerator">IPServerCollectionEnumerator</see> class referencing the specified <see cref="IPServerCollection"/> object.
			/// </summary>
			/// <param name="mappings">The <see cref="IPServerCollection"/> to enumerate.</param>
			public IPServerCollectionEnumerator(IPServerCollection mappings)
			{
				_temp = mappings;
				_enumerator = _temp.GetEnumerator();
			}
			
			/// <summary>
			/// Gets the current element in the collection.
			/// </summary>
			public IPServer Current {
				get { return (IPServer) (_enumerator.Current); }
			}
			
			object IEnumerator.Current {
				get { return _enumerator.Current; }
			}
			
			/// <summary>
			/// Advances the enumerator to the next element of the collection.
			/// </summary>
			/// <returns><b>true</b> if the enumerator was successfully advanced to the next element; <b>false</b> if the enumerator has passed the end of the collection.</returns>
			public bool MoveNext()
			{
				return _enumerator.MoveNext();
			}
			
			bool IEnumerator.MoveNext()
			{
				return _enumerator.MoveNext();
			}
			
			/// <summary>
			/// Sets the enumerator to its initial position, which is before the first element in the collection.
			/// </summary>
			public void Reset()
			{
				_enumerator.Reset();
			}
			
			void IEnumerator.Reset()
			{
				_enumerator.Reset();
			}
		}
		
		#endregion
	}
}
