// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;

namespace Grunwald.P2PLib.Network
{
	/// <summary>
	/// Base class for all peer to peer networks.
	/// </summary>
	public abstract class P2PNetwork
	{
		#region Constructor
		/// <summary>
		/// Creates a new instance of P2PNetwork.
		/// </summary>
		protected P2PNetwork()
		{
			connectionManager = new ConnectionManager(this);
		}
		#endregion
		
		#region HostCache / IPServer
		AddressCollection hostCache = new AddressCollection();
		
		/// <summary>
		/// The <see cref="AddressCollection"/> this network uses to build new
		/// connections.
		/// </summary>
		public AddressCollection HostCache {
			get { return hostCache; }
		}
		
		IPServerCollection ipServers = new IPServerCollection();
		
		/// <summary>
		/// Gets the collection of IPServers used to get the IP addresses of the
		/// network partners.
		/// </summary>
		public IPServerCollection IPServers {
			get {
				return ipServers;
			}
		}
		#endregion
		
		#region Connections
		ConnectionCollection connections = new ConnectionCollection();
		
		/// <summary>
		/// The <see cref="ConnectionCollection"/> containing the current connections
		/// with this network.
		/// </summary>
		public ConnectionCollection Connections {
			get { return connections; }
		}
		
		/// <summary>
		/// Initiates a new network connection using the existing <see cref="IConnection"/>.
		/// </summary>
		/// <param name="connection">The TCP/IP connection to use.</param>
		/// <param name="addressEntry">The address of the network node connecting to</param>
		public abstract void BuildConnection(IConnection connection, AddressEntry addressEntry);
		
		/// <summary>
		/// Is raised when a new connection is added to the <see cref="Connections"/> collection.
		/// </summary>
		public event ConnectionEventHandler ConnectionAdded;
		
		/// <summary>
		/// Is raised when a connection is removed from the <see cref="Connections"/> collection.
		/// </summary>
		public event ConnectionEventHandler ConnectionRemoved;
		
		/// <summary>
		/// Raises the <see cref="ConnectionAdded">ConnectionAdded event</see>.
		/// </summary>
		protected virtual void OnConnectionAdded(ConnectionEventArgs e) {
			if (ConnectionAdded != null) {
				ConnectionAdded(this, e);
			}
		}
		
		/// <summary>
		/// Raises the <see cref="ConnectionAdded">ConnectionRemoved event</see>.
		/// </summary>
		protected virtual void OnConnectionRemoved(ConnectionEventArgs e) {
			if (ConnectionRemoved != null) {
				ConnectionRemoved(this, e);
			}
		}
		#endregion
		
		#region ReceivingUsers
		ReceivingUserCollection receivingUsers = new ReceivingUserCollection();
		
		/// <summary>
		/// Gets the collection of receiving users.
		/// </summary>
		public ReceivingUserCollection ReceivingUsers {
			get {
				return receivingUsers;
			}
		}
		
		/// <summary>
		/// Is raised when a receiving user is added to the collection.
		/// </summary>
		public event ReceivingUserEventHandler ReceivingUserAdded {
			add    { receivingUsers.ReceivingUserAdded += value; }
			remove { receivingUsers.ReceivingUserAdded -= value; }
		}
		
		/// <summary>
		/// Is raised when a receiving user is removed from the collection.
		/// </summary>
		public event ReceivingUserEventHandler ReceivingUserRemoved {
			add    { receivingUsers.ReceivingUserRemoved += value; }
			remove { receivingUsers.ReceivingUserRemoved -= value; }
		}
		#endregion
		
		#region Work
		/// <summary>
		/// <para>Lets the network do some "work".</para>
		/// <para>You can't override this method directly, override either
		/// OnWorking or OnWorked.</para>
		/// <para>This method raises the <see cref="Working">Working event</see>,
		/// calls Work on the <see cref="ConnectionManager"/>
		/// and raises the <see cref="Worked">Worked event</see>.</para>
		/// </summary>
		/// <remarks>
		/// So the network should (re)connect, build more connection or something else,
		/// depending on the network. This method is called regularily, so your network
		/// doesn't need to create it's own timer but fully rely on this method.<br />
		/// The interval is not specified, but it should be something between one second
		/// and ten seconds. So if your code is time critical, you should still use a timer,
		/// but remember that this little work method is much easier to test and debug than
		/// a timer.
		/// </remarks>
		public void Work()
		{
			try {
				OnWorking(EventArgs.Empty);
				if (!this.stayOffline)
					connectionManager.Work();
				OnWorked(EventArgs.Empty);
			} catch (Exception ex) {
				Helper.LogError(ex);
				throw;
			}
		}
		
		/// <summary>
		/// This event occurs before the work method is executed.
		/// </summary>
		public event EventHandler Working;
		
		/// <summary>
		/// Raises the Working event.
		/// </summary>
		protected virtual void OnWorking(EventArgs e) {
			if (Working != null) {
				Working(this, e);
			}
		}
		
		/// <summary>
		/// This event occurs after the Work method has been executed.
		/// </summary>
		public event EventHandler Worked;
		
		/// <summary>
		/// Raises the Worked event.
		/// </summary>
		protected virtual void OnWorked(EventArgs e) {
			if (Worked != null) {
				Worked(this, e);
			}
		}
		#endregion
		
		#region ConnectionManager
		ConnectionManager connectionManager;
		
		/// <summary>
		/// Gets/Sets the <see cref="ConnectionManager"/> used to build new connections
		/// and close old connections.<br />
		/// The default value is an instance of <see cref="ConnectionManager"/>, you
		/// can change this to an instance of a derived class.
		/// </summary>
		/// <exception cref="ArgumentNullException">The specified value is null</exception>
		/// <exception cref="ArgumentException">The specified connection manager does
		/// not manage this network.</exception>
		public ConnectionManager ConnectionManager {
			get { return connectionManager; }
			set {
				if (value == null)
					throw new ArgumentNullException();
				if (value.Network != this)
					throw new ArgumentException("The connection manager must use this network.");
				connectionManager = value;
			}
		}
		#endregion
		
		#region AcceptConnection
		/// <summary>
		/// Checks if a connection with the same partner already exists.
		/// If a connection already exists, this method returns <b>false</b>.
		/// If the connection does not exist yet, the connection is added
		/// to the <see cref="ConnectionCollection"/>,
		/// the <see cref="ConnectionAdded">ConnectionAdded event</see> is fired
		/// and this method returns <c>true</c>.
		/// </summary>
		/// <exception cref="ArgumentNullException">An argument is null.</exception>
		public virtual bool AcceptConnection(Connection connection)
		{
			if (connection == null) throw new ArgumentNullException("connection");
			P2PAddress adr = connection.Partner;
			lock (this) {
				foreach (Connection c in Connections) {
					if (adr.Equals(c.Partner)) {
						if (c.Connected)
							return false;
					}
				}
				Connections.Add(connection);
			}
			OnConnectionAdded(new ConnectionEventArgs(connection));
			return true;
		}
		#endregion
		
		#region Connect / Disconnect
		/// <summary>
		/// Gets if we are currently connected with this network.
		/// </summary>
		public virtual bool Connected {
			get {
				lock (this) {
					foreach (Connection c in this.Connections) {
						if (c.Ready)
							return true;
					}
				}
				return false;
			}
		}
		
		bool stayOffline = true;
		
		/// <summary>
		/// <para>Gets/Sets if the P2PNetwork should stay offline (so not try to build new
		/// connections of it's own).</para>
		/// <para>Existing connections are not closed if this property is set to true;
		/// if this property is set to false, the network won't try to find new
		/// connections immediately but wait some time.</para>
		/// <para>The default value of this property is <b>true</b>.</para>
		/// </summary>
		public bool StayOffline {
			get {
				return stayOffline;
			}
			set {
				if (stayOffline == value)
					return;
				stayOffline = value;
				OnStayOfflineChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Occurs when the StayOffline property is changed.
		/// </summary>
		public event EventHandler StayOfflineChanged;
		
		/// <summary>
		/// Raises the StayOfflineChanged event.
		/// </summary>
		protected virtual void OnStayOfflineChanged(EventArgs e) {
			if (StayOfflineChanged != null) {
				StayOfflineChanged(this, e);
			}
		}
		
		/// <summary>
		/// Deactivates the <see cref="StayOffline">offline mode</see> and tries to
		/// connect immediately.
		/// </summary>
		public virtual void Connect()
		{
			this.StayOffline = false;
			connectionManager.Work();
		}
		
		/// <summary>
		/// Closes all active connections and activates the offline mode.
		/// </summary>
		public virtual void Disconnect()
		{
			this.StayOffline = true;
			lock (this) {
				foreach (Connection c in this.Connections) {
					if (c.Connected)
						c.Close();
				}
			}
		}
		#endregion
		
		/// <summary>
		/// Gets an unique identifier for this network (e.g. a hash of the url of the primary ip server)
		/// </summary>
		public abstract string Identifier { get; }
		
		/// <summary>
		/// Gets the string representation of this P2PNetwork.
		/// </summary>
		public override string ToString()
		{
			return "[" + GetType().Name + ": Identifier = " + Identifier + "]";
		}
	}
}
