// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace Grunwald.P2PLib.Network
{
	/// <summary>
	/// Represents a collection of <see cref="ReceivingUser"/> objects.
	/// </summary>
	public class ReceivingUserCollection : CollectionBase
	{
		#region Events
		/// <summary>
		/// Is raised when a new ReceivingUser is added to the collection.
		/// </summary>
		public event ReceivingUserEventHandler ReceivingUserAdded;
		
		/// <summary>
		/// Is raised when a ReceivingUser is removed from the collection.
		/// </summary>
		public event ReceivingUserEventHandler ReceivingUserRemoved;
		
		/// <summary>
		/// Raises the ReceivingUserAdded event.
		/// </summary>
		protected virtual void OnReceivingUserAdded(ReceivingUserEventArgs e)
		{
			if (ReceivingUserAdded != null) {
				ReceivingUserAdded(this, e);
			}
		}
		
		/// <summary>
		/// Raises the ReceivingUserRemoved event.
		/// </summary>
		protected virtual void OnReceivingUserRemoved(ReceivingUserEventArgs e)
		{
			if (ReceivingUserRemoved != null) {
				ReceivingUserRemoved(this, e);
			}
		}
		#endregion
		
		#region Constructor
		
		/// <summary>
		/// Initializes a new instance of the <see cref="ReceivingUserCollection"/> class.
		/// </summary>
		public ReceivingUserCollection()
		{
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="ReceivingUserCollection"/> class containing the elements of the specified source collection.
		/// </summary>
		/// <param name="value">A <see cref="ReceivingUserCollection"/> with which to initialize the collection.</param>
		public ReceivingUserCollection(ReceivingUserCollection value)
		{
			this.AddRange(value);
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="ReceivingUserCollection"/> class containing the specified array of <see cref="ReceivingUser"/> objects.
		/// </summary>
		/// <param name="value">An array of <see cref="ReceivingUser"/> objects with which to initialize the collection. </param>
		public ReceivingUserCollection(ReceivingUser[] value)
		{
			this.AddRange(value);
		}
		
		#endregion
		
		#region Indexer
		
		/// <summary>
		/// Gets/Sets the <see cref="ReceivingUser"/> at the specified index in the collection.
		/// <para>
		/// In C#, this property is the indexer for the <see cref="ReceivingUserCollection"/> class.
		/// </para>
		/// </summary>
		public ReceivingUser this[int index] {
			get {
				return (ReceivingUser) (this.List[index]);
			}
			set {
				if (value == null)
					throw new ArgumentNullException();
				ReceivingUser old = (ReceivingUser)this.List[index];
				this.List[index] = value;
				OnReceivingUserRemoved(new ReceivingUserEventArgs(old));
				OnReceivingUserAdded(new ReceivingUserEventArgs(value));
			}
		}
		
		#endregion
		
		#region Add
		
		/// <summary>
		/// Adds an item to the end of the collection
		/// </summary>
		/// <param name="value">Item to add</param>
		/// <returns>The positition where the new element has been added.</returns>
		public int Add(ReceivingUser value)
		{
			if (value == null)
				throw new ArgumentNullException();
			int pos = this.List.Add(value);
			OnReceivingUserAdded(new ReceivingUserEventArgs(value));
			return pos;
		}
		
		/// <summary>
		/// Copies the elements of the specified <see cref="ReceivingUser"/> array to the end of the collection.
		/// </summary>
		/// <param name="value">An array of type <see cref="ReceivingUser"/> containing the objects to add to the collection.</param>
		public void AddRange(ReceivingUser[] value)
		{
			for (int i = 0; (i < value.Length); i = (i + 1)) {
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		/// Adds the contents of another <see cref="ReceivingUserCollection"/> to the end of the collection.
		/// </summary>
		/// <param name="value">A <see cref="ReceivingUserCollection"/> containing the objects to add to the collection. </param>
		public void AddRange(ReceivingUserCollection value)
		{
			for (int i = 0; (i < value.Count); i = (i + 1)) {
				this.Add((ReceivingUser) value.List[i]);
			}
		}
		
		#endregion
		
		#region Contains
		
		/// <summary>
		/// Gets a value indicating whether the collection contains the specified <see cref="ReceivingUser"/>.
		/// </summary>
		/// <param name="value">The <see cref="ReceivingUser"/> to search for in the collection.</param>
		/// <returns><b>true</b> if the collection contains the specified object; otherwise, <b>false</b>.</returns>
		public bool Contains(ReceivingUser value)
		{
			return this.List.Contains(value);
		}
		
		#endregion
		
		#region ToArray
		
		/// <summary>
		/// Copies the collection objects to a one-dimensional <see cref="System.Array">Array</see> instance beginning at the specified index.
		/// </summary>
		/// <param name="array">The one-dimensional <see cref="System.Array">Array</see> that is the destination of the values copied from the collection.</param>
		/// <param name="index">The index of the array at which to begin inserting.</param>
		public void CopyTo(ReceivingUser[] array, int index)
		{
			this.List.CopyTo(array, index);
		}
		
		/// <summary>
		/// Creates a one-dimensional <see cref="System.Array">Array</see> instance containing the collection items.
		/// </summary>
		/// <returns>Array of type ReceivingUser</returns>
		public ReceivingUser[] ToArray()
		{
			ReceivingUser[] array = new ReceivingUser[this.Count];
			this.CopyTo(array, 0);
			return array;
		}
		
		#endregion
		
		#region IndexOf
		
		/// <summary>
		/// Gets the index in the collection of the specified <see cref="ReceivingUser"/>, if it exists in the collection.
		/// </summary>
		/// <param name="value">The <see cref="ReceivingUser"/> to locate in the collection.</param>
		/// <returns>The index in the collection of the specified object, if found; otherwise, -1.</returns>
		public int IndexOf(ReceivingUser value)
		{
			return this.List.IndexOf(value);
		}
		
		#endregion
		
		#region Insert
		
		/// <summary>
		/// Inserts a new item into the collection.
		/// </summary>
		/// <param name="index">Position, whrere the item should be added</param>
		/// <param name="value">Item to add</param>
		public void Insert(int index, ReceivingUser value)
		{
			List.Insert(index, value);
			OnReceivingUserAdded(new ReceivingUserEventArgs(value));
		}
		
		#endregion
		
		#region Remove
		
		/// <summary>
		/// Removes the specified item from the collection
		/// </summary>
		/// <param name="value">Item to remove</param>
		public void Remove(ReceivingUser value)
		{
			List.Remove(value);
			OnReceivingUserRemoved(new ReceivingUserEventArgs(value));
		}
		
		/// <summary>
		/// Removes the item at the given position
		/// </summary>
		/// <param name="index">Position of item</param>
		public new void RemoveAt(int index)
		{
			ReceivingUser old = (ReceivingUser)this.List[index];
			base.RemoveAt(index);
			OnReceivingUserRemoved(new ReceivingUserEventArgs(old));
		}
		
		/// <summary>
		/// Clears all items from the collection
		/// </summary>
		public new void Clear()
		{
			while (this.Count > 0) {
				this.RemoveAt(0);
			}
		}
		
		#endregion
	}
}
