// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using Grunwald.P2PLib.Network;

namespace Grunwald.P2PLib
{
	/// <summary>
	/// Contains an address to an element in a peer to peer network.
	/// </summary>
	public abstract class P2PAddress
	{
		#region Network
		P2PNetwork network;
		/// <summary>
		/// Gets/Sets the network this address points to.
		/// </summary>
		public P2PNetwork Network {
			get { return network; }
			set {
				if (value == null)
					throw new ArgumentNullException("network");
				network = value;
			}
		}
		#endregion
		
		#region Target (Type)
		AddressTarget target;
		/// <summary>
		/// Gets the type of the target.
		/// </summary>
		public AddressTarget Target {
			get { return target; }
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates a new P2PAddress pointing to the specified network with the
		/// specified target type.
		/// </summary>
		protected P2PAddress(P2PNetwork network, AddressTarget target)
		{
			this.Network = network;
			this.target = target;
		}
		#endregion
		
		/// <summary>
		/// Determines whether the specified P2PAddress is equal to the current P2PAddress.
		/// </summary>
		/// <returns>
		/// Returns true if both addresses are equal, returns false if the addresses are
		/// different or the object is not an address.
		/// </returns>
		public override abstract bool Equals(object obj);
		
		/// <summary>
		/// Serves as a hash function for the object.
		/// </summary>
		public override int GetHashCode()
		{
			int result;
			unchecked {
				result = (Network != null ? Network.GetHashCode() : 0);
				result = 29 * result + (int)Target;
			}
			return result;
		}
		
		/// <summary>
		/// Gets the string representation of this P2PAddress.
		/// </summary>
		public override string ToString()
		{
			return String.Format(System.Globalization.CultureInfo.InvariantCulture,
			                     "[{0}: Network = {1}, Target = {2}]",
			                     GetType().FullName,
			                     Network,
			                     Target);
		}
	}
	
	/// <summary>
	/// Delegate for the P2PAddressAdded/P2PAddressRemoved events.
	/// </summary>
	public delegate void P2PAddressEventHandler(object sender, P2PAddressEventArgs e);
	
	/// <summary>
	/// Event args for the P2PAddressAdded/P2PAddressRemoved events.
	/// </summary>
	public class P2PAddressEventArgs : EventArgs
	{
		P2PAddress address;
		
		/// <summary>
		/// The affected P2PAddress.
		/// </summary>
		public P2PAddress Address {
			get {
				return address;
			}
		}
		
		/// <summary>
		/// Creates a new P2PAddressEventArgs object.
		/// </summary>
		/// <param name="address">The affected P2PAddress.</param>
		public P2PAddressEventArgs(P2PAddress address)
		{
			this.address = address;
		}
	}
}
