// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using Grunwald.P2PLib.Network;

namespace Grunwald.P2PLib
{
	/// <summary>
	/// Represents a collection of <see cref="P2PAddress"/> objects.
	/// </summary>
	public class P2PAddressCollection : CollectionBase
	{
		/// <summary>
		/// Gets the first <see cref="P2PAddress"/> that is found for the specified
		/// <see cref="P2PNetwork"/>.
		/// If no address for the network is found, null is returned.
		/// </summary>
		public P2PAddress GetAddressByNetwork(P2PNetwork network) {
			foreach (P2PAddress a in this) {
				if (a.Network == network)
					return a;
			}
			return null;
		}
		
		#region Events
		/// <summary>
		/// Is raised when a new P2PAddress is added to the collection.
		/// </summary>
		public event P2PAddressEventHandler AddressAdded;
		
		/// <summary>
		/// Is raised when a P2PAddress is removed from the collection.
		/// </summary>
		public event P2PAddressEventHandler AddressRemoved;
		
		/// <summary>
		/// Raises the P2PAddressAdded event.
		/// </summary>
		protected virtual void OnP2PAddressAdded(P2PAddressEventArgs e)
		{
			if (AddressAdded != null) {
				AddressAdded(this, e);
			}
		}
		
		/// <summary>
		/// Raises the P2PAddressRemoved event.
		/// </summary>
		protected virtual void OnP2PAddressRemoved(P2PAddressEventArgs e)
		{
			if (AddressRemoved != null) {
				AddressRemoved(this, e);
			}
		}
		#endregion
		
		#region Constructor
		
		/// <summary>
		/// Initializes a new instance of the <see cref="P2PAddressCollection"/> class.
		/// </summary>
		public P2PAddressCollection()
		{
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="P2PAddressCollection"/> class containing the elements of the specified source collection.
		/// </summary>
		/// <param name="value">A <see cref="P2PAddressCollection"/> with which to initialize the collection.</param>
		public P2PAddressCollection(P2PAddressCollection value)
		{
			this.AddRange(value);
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="P2PAddressCollection"/> class containing the specified array of <see cref="P2PAddress"/> objects.
		/// </summary>
		/// <param name="value">An array of <see cref="P2PAddress"/> objects with which to initialize the collection. </param>
		public P2PAddressCollection(P2PAddress[] value)
		{
			this.AddRange(value);
		}
		
		#endregion
		
		#region Indexer
		
		/// <summary>
		/// Gets/Sets the <see cref="P2PAddress"/> at the specified index in the collection.
		/// <para>
		/// In C#, this property is the indexer for the <see cref="P2PAddressCollection"/> class.
		/// </para>
		/// </summary>
		public P2PAddress this[int index] {
			get {
				return (P2PAddress) (this.List[index]);
			}
			set {
				if (value == null)
					throw new ArgumentNullException();
				P2PAddress old = (P2PAddress)this.List[index];
				this.List[index] = value;
				OnP2PAddressRemoved(new P2PAddressEventArgs(old));
				OnP2PAddressAdded(new P2PAddressEventArgs(value));
			}
		}
		
		#endregion
		
		#region Add
		
		/// <summary>
		/// Adds an item to the end of the collection
		/// </summary>
		/// <param name="value">Item to add</param>
		/// <returns>The positition where the new element has been added.</returns>
		public int Add(P2PAddress value)
		{
			if (value == null)
				throw new ArgumentNullException();
			int pos = this.List.Add(value);
			OnP2PAddressAdded(new P2PAddressEventArgs(value));
			return pos;
		}
		
		/// <summary>
		/// Copies the elements of the specified <see cref="P2PAddress"/> array to the end of the collection.
		/// </summary>
		/// <param name="value">An array of type <see cref="P2PAddress"/> containing the objects to add to the collection.</param>
		public void AddRange(P2PAddress[] value)
		{
			for (int i = 0; (i < value.Length); i = (i + 1)) {
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		/// Adds the contents of another <see cref="P2PAddressCollection"/> to the end of the collection.
		/// </summary>
		/// <param name="value">A <see cref="P2PAddressCollection"/> containing the objects to add to the collection. </param>
		public void AddRange(P2PAddressCollection value)
		{
			for (int i = 0; (i < value.Count); i = (i + 1)) {
				this.Add((P2PAddress) value.List[i]);
			}
		}
		
		#endregion
		
		#region Contains
		
		/// <summary>
		/// Gets a value indicating whether the collection contains the specified <see cref="P2PAddress"/>.
		/// </summary>
		/// <param name="value">The <see cref="P2PAddress"/> to search for in the collection.</param>
		/// <returns><b>true</b> if the collection contains the specified object; otherwise, <b>false</b>.</returns>
		public bool Contains(P2PAddress value)
		{
			return this.List.Contains(value);
		}
		
		#endregion
		
		#region ToArray
		
		/// <summary>
		/// Copies the collection objects to a one-dimensional <see cref="System.Array">Array</see> instance beginning at the specified index.
		/// </summary>
		/// <param name="array">The one-dimensional <see cref="System.Array">Array</see> that is the destination of the values copied from the collection.</param>
		/// <param name="index">The index of the array at which to begin inserting.</param>
		public void CopyTo(P2PAddress[] array, int index)
		{
			this.List.CopyTo(array, index);
		}
		
		/// <summary>
		/// Creates a one-dimensional <see cref="System.Array">Array</see> instance containing the collection items.
		/// </summary>
		/// <returns>Array of type P2PAddress</returns>
		public P2PAddress[] ToArray()
		{
			P2PAddress[] array = new P2PAddress[this.Count];
			this.CopyTo(array, 0);
			return array;
		}
		
		#endregion
		
		#region IndexOf
		
		/// <summary>
		/// Gets the index in the collection of the specified <see cref="P2PAddress"/>, if it exists in the collection.
		/// </summary>
		/// <param name="value">The <see cref="P2PAddress"/> to locate in the collection.</param>
		/// <returns>The index in the collection of the specified object, if found; otherwise, -1.</returns>
		public int IndexOf(P2PAddress value)
		{
			return this.List.IndexOf(value);
		}
		
		#endregion
		
		#region Insert
		
		/// <summary>
		/// Inserts a new item into the collection.
		/// </summary>
		/// <param name="index">Position, whrere the item should be added</param>
		/// <param name="value">Item to add</param>
		public void Insert(int index, P2PAddress value)
		{
			List.Insert(index, value);
			OnP2PAddressAdded(new P2PAddressEventArgs(value));
		}
		
		#endregion
		
		#region Remove
		
		/// <summary>
		/// Removes the specified item from the collection
		/// </summary>
		/// <param name="value">Item to remove</param>
		public void Remove(P2PAddress value)
		{
			List.Remove(value);
			OnP2PAddressRemoved(new P2PAddressEventArgs(value));
		}
		
		/// <summary>
		/// Removes the item at the given position
		/// </summary>
		/// <param name="index">Position of item</param>
		public new void RemoveAt(int index)
		{
			P2PAddress old = (P2PAddress)this.List[index];
			base.RemoveAt(index);
			OnP2PAddressRemoved(new P2PAddressEventArgs(old));
		}
		
		/// <summary>
		/// Clears all items from the collection
		/// </summary>
		public new void Clear()
		{
			while (this.Count > 0) {
				this.RemoveAt(0);
			}
		}
		
		#endregion
	}
}
