// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Net;

namespace Grunwald.P2PLib.Server
{
	/// <summary>
	/// Defines abstract methods for all supported servers.
	/// </summary>
	public abstract class AbstractServer
	{
		#region IP/Port Detection
		IPAddress ipAddress;
		
		/// <summary>
		/// Gets/Sets the IP Address of this server. This property is also used for the
		/// InternetEndPoint property.
		/// </summary>
		public IPAddress IPAddress {
			get {
				return ipAddress;
			}
			set {
				if (Object.Equals(ipAddress, value))
					return;
				ipAddress = value;
				OnIPAddressChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Gets called whenever the IPAddress property is changed.
		/// </summary>
		public event EventHandler IPAddressChanged;
		
		/// <summary>
		/// Raises the IPAddressChanged event.
		/// </summary>
		protected virtual void OnIPAddressChanged(EventArgs e)
		{
			if (IPAddressChanged != null) {
				IPAddressChanged(this, e);
			}
		}
		
		/// <summary>
		/// Returns the IPEndPoint over which this server <b>could</b> be reached from the
		/// internet, if it <b>would</b> have server ability. This returns null if the
		/// current IP is unknown. This does <b>not</b> return null if the IP is known
		/// but this server can't be reached from the Internet.
		/// </summary>
		public IPEndPoint InternetEndPoint {
			get {
				if (this.IPAddress == null)
					return null;
				return new IPEndPoint(this.IPAddress, Port);
			}
		}
		
		ServerAbilityTester serverAbilityTester;
		
		/// <summary>
		/// The ServerAbilityTester used for detecting this server's ability to be reached from the internet.
		/// </summary>
		public ServerAbilityTester ServerAbilityTester {
			get { return serverAbilityTester; }
			set { serverAbilityTester = value; }
		}
		#endregion
		
		#region Port
		int port = 0;
		
		/// <summary>
		/// Gets/Sets the port on which this server is running
		/// </summary>
		public virtual int Port {
			get { return port; }
			set {
				if (value <= 0 || value >= UInt16.MaxValue)
					throw new ArgumentOutOfRangeException("value", value, "Port is out of valid range (ushort)");
				if (value == port)
					return;
				int oldport = port;
				port = value;
				if (Running) {
					Stop();
					try {
						Start();
					} catch (Exception) {
						if (!Running) {
							port = oldport;
							Start();
						}
						throw;
					}
				}
				if (oldport != 0)
					OnPortChanged(EventArgs.Empty);
			}
		}
		
		/// <summary>
		/// Is called when the port of this server is changed
		/// </summary>
		protected virtual void OnPortChanged(EventArgs e)
		{
			if (PortChanged != null) {
				PortChanged(this, e);
			}
		}
		
		/// <summary>
		/// Is called when the port of this server is changed
		/// </summary>
		public event EventHandler PortChanged;
		
		#endregion
		
		/// <summary>
		/// Gets if the server is currently listening to new connections
		/// </summary>
		public abstract bool Running { get; }
		
		/// <summary>
		/// Starts the server
		/// </summary>
		public abstract void Start();
		
		/// <summary>
		/// Stops the server
		/// </summary>
		public abstract void Stop();
	}
}
