// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Globalization;
using System.IO;
using System.Net;
using System.Security.Cryptography;
using System.Text;

namespace Grunwald.P2PLib.Server.Http
{
	/// <summary>
	/// <para>The FileUpload class can be used to answer a <see cref="HttpRequest"/>
	/// with the upload of a file.</para>
	/// <para>This class supports resuming downloads using the Content-Range header,
	/// conditionals (If-Match, If-None-Match, If-Range) on small files/streams and
	/// time conditionals (If-Modified-Since) on files.</para>
	/// </summary>
	public class FileUpload
	{
		HttpRequest request;
		static int smallFileMaxSize = 32768;
		WebHeaderCollection extraHeaders;
		
		#region Constructor
		/// <summary>Creates a new FileUpload object serving the specified <paramref name="request"/>.</summary>
		public FileUpload(HttpRequest request)
		{
			this.request = request;
			this.extraHeaders = new WebHeaderCollection();
		}
		#endregion
		
		#region Properties
		/// <summary>
		/// Gets the <see cref="HttpRequest"/> this FileUpload was created for.
		/// </summary>
		public HttpRequest Request {
			get {
				return request;
			}
		}
		
		/// <summary>
		/// Gets a collection that can be used to specify additional headers for the
		/// response.
		/// </summary>
		public WebHeaderCollection Headers {
			get {
				return extraHeaders;
			}
		}
		
		/// <summary>
		/// Gets/Sets the number of bytes a file may have to be considered as small file.
		/// For small files an MD5 sum is calculated and sent, therefor the whole file
		/// has to be cached in a buffer; so this should only be used for small files.
		/// The default value for this property is 32 KB.
		/// </summary>
		public static int SmallFileMaxSize {
			get {
				return smallFileMaxSize;
			}
			set {
				if (value < 0)
					throw new ArgumentOutOfRangeException();
				smallFileMaxSize = value;
			}
		}
		#endregion
		
		/// <summary>
		/// Sends the specified file as answer to the request used to create this FileUpload.
		/// </summary>
		/// <exception cref="HttpException">
		/// Is thrown when the request was invalid and the server should return a HTTP error
		/// instead of delivering the data.
		/// </exception>
		public void Send(string file)
		{
			if (!File.Exists(file))
				throw new HttpException(HttpStatusCode.NotFound);
			if (extraHeaders["Content-Type"] == null)
				extraHeaders["Content-Type"] = Helper.GetContentType(file);
			// TODO: Provide LastChangedDate if possible
			using (FileStream fs = new FileStream(file, FileMode.Open, FileAccess.Read, FileShare.Read)) {
				Send(fs);
			}
		}
		
		/// <summary>
		/// Sends the data from the specified stream as answer to the request used to create this FileUpload.
		/// </summary>
		/// <exception cref="HttpException">
		/// Is thrown when the request was invalid and the server should return a HTTP error
		/// instead of delivering the data.
		/// </exception>
		public void Send(Stream stream)
		{
			HttpResponse response = new HttpResponse(request);
			SendInternal(response, stream);
			try {
				request.Connection.Stream.Flush();
				if (!response.KeepAlive)
					request.Connection.Close();
			} catch (IOException) {
				// ignore IO errors when sending
			}
		}
		
		long startPosition, endPosition, position;
		
		/// <summary>
		/// Gets the end position of the transfer.
		/// </summary>
		public long EndPosition {
			get {
				return endPosition;
			}
		}
		
		/// <summary>
		/// Gets the start position of the transfer.
		/// </summary>
		public long StartPosition {
			get {
				return startPosition;
			}
		}
		
		/// <summary>
		/// Gets the current position of the upload.
		/// </summary>
		public long Position {
			get {
				return position;
			}
		}
		
		private void SendInternal(HttpResponse response, Stream fileStream)
		{
			Stream w = request.Connection.Stream;
			if (fileStream.Length == 0) {
				response.StatusCode = HttpStatusCode.NoContent;
				response.MergeHeaders(extraHeaders);
				response.SendHeaders(request.Connection.Writer, true);
				return;
			}
			if (fileStream.CanSeek)
				response.Headers.Set("Accept-Ranges", "bytes");
			int count;
			byte[] buffer;
			string range = request.Headers["Range"];
			// since we don't send ETags with partial content, the condition always fails
			if (request.Headers["If-Range"] != null)
				range = null;
			long startPos, endPos;
			CalcRange(range, fileStream.Length, out startPos, out endPos);
			startPosition = startPos;
			endPosition = endPos;
			response.ContentLength = endPos - startPos + 1;
			if (range != null) {
				string responseRange = string.Format("bytes {0}-{1}/{2}",
				                                     startPos, endPos, fileStream.Length,
				                                     CultureInfo.InvariantCulture);
				response.Headers["Content-Range"] = responseRange;
				response.StatusCode = HttpStatusCode.PartialContent;
			}
			if (range == null && fileStream.Length < SmallFileMaxSize) {
				buffer = new byte[fileStream.Length];
				int offset = 0;
				do {
					count = fileStream.Read(buffer, offset, buffer.Length - offset);
					offset += count;
				} while (count > 0);
				byte[] md5;
				using (MD5CryptoServiceProvider md5Provider = new MD5CryptoServiceProvider()) {
					md5 = md5Provider.ComputeHash(buffer);
				}
				string md5String = Convert.ToBase64String(md5);
				if (extraHeaders["ETAG"] == null)
					extraHeaders["ETAG"] = "\"" + md5String.TrimEnd('=') + "\"";
				string etag = extraHeaders["ETAG"];
				if (request.Headers["If-Match"] != null) {
					if (!Match(request.Headers["If-Match"], etag))
						throw new HttpException(HttpStatusCode.PreconditionFailed, extraHeaders);
				}
				if (request.Headers["If-None-Match"] != null) {
					if (Match(request.Headers["If-None-Match"], etag))
						throw new HttpException(HttpStatusCode.NotModified, extraHeaders);
				}
				response.MergeHeaders(extraHeaders);
				response.Headers["Content-MD5"] = md5String;
				try {
					response.SendHeaders(request.Connection.Writer, true);
					w.Write(buffer, 0, buffer.Length);
				} catch (IOException) {
					return;
				}
			} else {
				if (request.Headers["If-Match"] != null)
					throw new HttpException(HttpStatusCode.PreconditionFailed, extraHeaders);
				response.MergeHeaders(extraHeaders);
				try {
					response.SendHeaders(request.Connection.Writer, true);
				} catch (IOException) {
					return;
				}
				buffer = new byte[4096];
				long pos = startPos;
				if (pos > 0)
					fileStream.Seek(pos, SeekOrigin.Begin);
				do {
					position = pos;
					if (pos + buffer.Length > endPos)
						count = fileStream.Read(buffer, 0, (int)(endPos - pos) + 1);
					else
						count = fileStream.Read(buffer, 0, buffer.Length);
					if (count > 0) {
						try {
							w.Write(buffer, 0, count);
						} catch (IOException) {
							return;
						}
					}
					pos += count;
				} while (count > 0 && pos <= endPos);
			}
		}
		
		private bool Match(string condition, string tag)
		{
			if (condition == "*")
				return true;
			tag = tag.Trim();
			foreach (string c in condition.Split(',')) {
				if (c.Trim() == tag)
					return true;
			}
			return false;
		}
		
		private void CalcRange(string range, long fileSize, out long startPos, out long endPos)
		{
			startPos = 0;
			endPos = fileSize - 1;
			if (range == null)
				return;
			if (!range.StartsWith("bytes="))
				throw new HttpException();
			range = range.Substring(6);
			if (range.IndexOf(',') >= 0) // bytes=100-200,300-400
				throw new HttpException(HttpStatusCode.RequestedRangeNotSatisfiable);
			
			if (range.StartsWith("-")) {
				// bytes=-500
				try {
					long length = long.Parse(range.Substring(1), CultureInfo.InvariantCulture);
					startPos = fileSize - length;
					endPos = fileSize - 1;
				} catch (FormatException ex) {
					throw new HttpException(HttpStatusCode.BadRequest, ex);
				}
			} else {
				int pos = range.IndexOf('-');
				if (pos < 1) // es muss ein - vorkommen
					throw new HttpException();
				try {
					startPos = long.Parse(range.Substring(0, pos), CultureInfo.InvariantCulture);
					if (range.EndsWith("-"))
						endPos = fileSize - 1;
					else
						endPos = long.Parse(range.Substring(pos + 1), CultureInfo.InvariantCulture);
				} catch (FormatException ex) {
					throw new HttpException(HttpStatusCode.BadRequest, ex);
				}
			}
			if (startPos > endPos || startPos < 0 || endPos >= fileSize)
				throw new HttpException(HttpStatusCode.RequestedRangeNotSatisfiable);
		}
	}
}
