// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections.Specialized;
using System.IO;
using System.Net;

namespace Grunwald.P2PLib.Server.Http
{
	/// <summary>
	/// Handles a http connection
	/// </summary>
	public class HttpConnectionHandler : IConnectionHandler
	{
		RequestHandlerCollection requestHandlers = new RequestHandlerCollection();
		
		/// <summary>
		/// The collection of the request handlers.
		/// </summary>
		public RequestHandlerCollection RequestHandlers {
			get { return requestHandlers; }
		}
		
		#region Handle
		
		/// <summary>
		/// Handles the HTTP Request.
		/// </summary>
		public virtual void Handle(IConnection connection, StringCollection request)
		{
			if (GetPriority(request) == FailedPriority) {
				FallbackError(new HttpException(), connection);
				return;
			}
			try {
				HttpRequest r;
				try {
					r = new HttpRequest(connection, request);
				} catch (HttpException ex) {
					FallbackError(ex, connection);
					return;
				}
				IRequestHandler h = RequestHandlers.GetHandler(r);
				if (h == null) {
					HandleError(new HttpException(HttpStatusCode.NotFound), r);
				} else {
					try {
						h.Handle(r);
					} catch (HttpException ex) {
						HandleError(ex, r);
					} catch (Exception ex) {
						HandleError(new HttpException(HttpStatusCode.InternalServerError, ex), r);
						Helper.LogError(ex);
					}
				}
			} catch (Exception ex) {
				FallbackError(new HttpException(HttpStatusCode.InternalServerError, ex), connection);
				throw;
			}
		}
		
		private void HandleError(HttpException ex, HttpRequest request)
		{
			// TODO: Try to get managed error handler
			int status = (int)ex.StatusCode;
			if (status / 100 == 3) {
				HttpResponse r = new HttpResponse(request);
				r.StatusCode = ex.StatusCode;
				r.StatusMessage = ex.Message;
				r.Server = "P2PLib Error Handler";
				r.MergeHeaders(ex.Headers);
				try {
					r.SendHeaders(request.Connection.Writer, true);
				} catch (IOException) {
					// ignore error
				}
				if (!r.KeepAlive)
					request.Connection.Close();
			} else {
				FallbackError(ex, request.Connection);
			}
		}
		
		private void FallbackError(HttpException ex, IConnection connection)
		{
			HttpResponse r = new HttpResponse();
			r.StatusCode = ex.StatusCode;
			r.StatusMessage = ex.Message;
			r.ContentType = "text/plain; charset=UTF-8";
			r.Server = "P2PLib Error-Fallback";
			r.MergeHeaders(ex.Headers);
			try {
				r.SendHeaders(connection.Writer, false);
				connection.Writer.WriteLine("{0:d} {1}", ex.StatusCode, ex, System.Globalization.CultureInfo.InvariantCulture);
				connection.Writer.Flush();
				connection.Close();
			} catch (ObjectDisposedException) {
			} catch (IOException) {
				// Ignore such errors...
			}
		}
		
		#endregion
		
		#region Priority
		
		/// <summary>
		/// Gets the priority of the HTTPConnectionHandler for the given request.
		/// If the request is an HTTP-Request, this returns <see cref="Priority"/>,
		/// else Int32.MinValue is returned
		/// </summary>
		public virtual int GetPriority(StringCollection request)
		{
			if (request == null)
				throw new ArgumentNullException("request");
			if (request.Count == 0)
				return int.MinValue;
			string l = request[0];
			int pos = l.IndexOf(' ');
			if (pos <= 0)
				return FailedPriority;
			string w = l.Substring(0, pos);
			if (w != "GET" && w != "HEAD" && w != "POST")
				return FailedPriority;
			pos = l.LastIndexOf(' ');
			if (pos < 0 || pos >= l.Length)
				return FailedPriority;
			w = l.Substring(pos + 1);
			if (w.Length != 8 || !w.StartsWith("HTTP/"))
				return FailedPriority;
			return Priority;
		}
		
		int priority = 1;
		
		/// <summary>
		/// Gets/Sets the priority the connection handler should return if the
		/// given request is an HTTP request.
		/// The default value is 1.
		/// </summary>
		public int Priority {
			get { return priority; }
			set { priority = value; }
		}
		
		int failedPriority = -10000;
		
		/// <summary>
		/// <para>Gets/Sets the priority the connection handler should return if the
		/// given request is <b>NOT</b> an HTTP request.
		/// The default value is -10000.</para>
		/// <para>If the HttpConnectionHandler handles an connection that is no http
		/// connection, it sends 400 Invalid Request and closes the connection.</para>
		/// </summary>
		public int FailedPriority {
			get { return failedPriority; }
			set { failedPriority = value; }
		}
		#endregion
	}
}
