// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Net;
using System.Runtime.Serialization;

namespace Grunwald.P2PLib.Server.Http
{
	/// <summary>
	/// An exception that correspondents to a http error.
	/// </summary>
	[Serializable()]
	public class HttpException : P2PLibException
	{
		#region Properties
		HttpStatusCode statusCode;
		WebHeaderCollection headers;
		
		/// <summary>
		/// Return the HTTP error code.
		/// </summary>
		public HttpStatusCode StatusCode {
			get {
				return statusCode;
			}
		}
		
		/// <summary>
		/// Gets/Sets a collection of http headers that should be sent with the error message.
		/// </summary>
		public WebHeaderCollection Headers {
			get {
				return headers;
			}
			set {
				headers = value;
			}
		}
		#endregion
		
		#region Constructor
		/// <summary>
		/// Creates a new HTTP exception with the status code 400 (Bad Request).
		/// </summary>
		public HttpException() : this(HttpStatusCode.BadRequest)
		{
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the status code 400 (Bad Request) and the specified text.
		/// </summary>
		public HttpException(string message) : this(HttpStatusCode.BadRequest, message)
		{
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the status code 400 (Bad Request) and the specified text.
		/// </summary>
		public HttpException(string message, Exception innerException)
			: this(HttpStatusCode.BadRequest, message, innerException)
		{
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the status code and the equivalent status message.
		/// </summary>
		public HttpException(HttpStatusCode statusCode)
			: this(statusCode, HttpResponse.GetMessage(statusCode))
		{
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the specified status code and
		/// the specified additional headers.
		/// </summary>
		public HttpException(HttpStatusCode statusCode, WebHeaderCollection headers)
			: this(statusCode, HttpResponse.GetMessage(statusCode))
		{
			this.headers = headers;
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the status code and the status message.
		/// </summary>
		public HttpException(HttpStatusCode statusCode, string message)
			: base(message)
		{
			this.statusCode = statusCode;
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the status code and the equivalent status message.
		/// </summary>
		public HttpException(HttpStatusCode statusCode, Exception innerException)
			: this(statusCode, HttpResponse.GetMessage(statusCode), innerException)
		{
		}
		
		/// <summary>
		/// Creates a new HTTP exception with the status code and the status message.
		/// </summary>
		public HttpException(HttpStatusCode statusCode, string message, Exception innerException)
			: base(message, innerException)
		{
			this.statusCode = statusCode;
		}
		
		/// <summary>
		/// Creates a new HTTP exception using serialization.
		/// </summary>
		protected HttpException(SerializationInfo info, StreamingContext context)
			: base(info, context)
		{
			
		}
		#endregion
	}
}
