// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections.Specialized;
using System.IO;
using System.Net;
using log4net;

namespace Grunwald.P2PLib.Server.Http
{
	/// <summary>
	/// This class is a container for the properties of an HTTP request.
	/// </summary>
	public class HttpRequest
	{
		#region Fields
		IConnection connection;
		/// <summary>
		/// The Connection on which the request was sent.
		/// </summary>
		public IConnection Connection {
			get { return connection; }
		}
		
		string method;
		/// <summary>
		/// The method the browser sent.
		/// </summary>
		public string Method {
			get { return method; }
		}
		
		string uri;
		/// <summary>
		/// The URL as it was sent by the browser (e.g. "/cgi-bin/get%20profile.exe?name=Daniel%20Grunwald%21"),
		/// but <b>without</b> the host name!
		/// </summary>
		public string FullRequest {
			get { return uri; }
		}
		
		string file;
		/// <summary>
		/// The (decoded) part of the URI before the '?' (e.g. "/cgi-bin/get profile.exe")
		/// </summary>
		public string File {
			get { return file; }
		}
		
		string requestString;
		/// <summary>
		/// The (encoded) part of the URI before the '?' (e.g. "name=Daniel%20Grunwald%21")
		/// </summary>
		public string RequestString {
			get { return requestString; }
		}
		
		WebHeaderCollection headers;
		/// <summary>
		/// The Request Headers sent by the browser
		/// </summary>
		public WebHeaderCollection Headers {
			get { return headers; }
		}
		
		static readonly ILog log = LogManager.GetLogger(typeof (HttpRequest));
		
		#endregion
		
		#region Constructor
		
		/// <summary>
		/// Create a new HTTP-Request container that reads it's values from the request.
		/// </summary>
		public HttpRequest(IConnection connection, StringCollection request)
		{
			if (connection == null)
				throw new ArgumentNullException("connection");
			if (request == null)
				throw new ArgumentNullException("request");
			if (request.Count == 0)
				throw new ArgumentException("collection is empty", "request");
			
			this.connection = connection;
			string h = request[0];
			log.Debug("Creating HTTPRequest for " + h);
			int pos = h.IndexOf(" ");
			if (pos < 0)
				throw new HttpException(HttpStatusCode.BadRequest);
			this.method = h.Substring(0, pos);
			string para = h.Substring(pos + 1);
			pos = para.LastIndexOf(" ");
			if (pos >= 0)
				this.uri = para.Substring(0, pos);
			else
				this.uri = "/";
			try {
				this.headers = new WebHeaderCollection();
				for (int i = 1; i < request.Count; ++i) {
					Headers.Add(request[i]);
				}
			} catch (Exception ex) {
				if (Helper.IsCritical(ex))
					throw;
				throw new HttpException(HttpStatusCode.BadRequest, ex);
			}
			if (!uri.StartsWith("/"))
				throw new HttpException(HttpStatusCode.BadRequest);
			pos = uri.IndexOf('?');
			if (pos >= 0) {
				this.file = Helper.Unescape(uri.Substring(0, pos));
				this.requestString = uri.Substring(pos + 1);
			} else {
				this.requestString = "";
				this.file = Helper.Unescape(uri);
			}
			// RFC 2068, 14.23 requires servers to send HTTP 400 if no Host-field is present.
			if (Headers["Host"] == null)
				throw new HttpException(HttpStatusCode.BadRequest);
			if (File.IndexOf("/..") >= 0) {
				log.Warn("File name contains /.. - possible hacking attempt");
				throw new HttpException(HttpStatusCode.BadRequest);
			}
			if (File.IndexOf('\\') >= 0)
				throw new HttpException(HttpStatusCode.BadRequest);
			if (File.IndexOfAny(Path.InvalidPathChars) >= 0)
				throw new HttpException(HttpStatusCode.BadRequest);
		}
		
		#endregion
		
		#region RequestNameValue
		NameValueCollection requestCollection;
		
		/// <summary>
		/// Gets a collection containing the decoded values of RequestString.
		/// (e.g. "name" = "Daniel Grunwald!")
		/// </summary>
		public NameValueCollection Request {
			get {
				if (requestCollection == null) {
					requestCollection = new NameValueCollection();
					string s = RequestString + "&";
					string e, name, value;
					int startpos = 0;
					int pos = s.IndexOf('&');
					while (pos >= 0) {
						e = s.Substring(startpos, pos - startpos);
						int pos2 = e.IndexOf('=');
						if (pos2 < 0) {
							log.Debug("Missing '=' in RequestString: " + RequestString);
						} else {
							name = Helper.Unescape(e.Substring(0, pos2));
							value = Helper.Unescape(e.Substring(pos2 + 1));
							requestCollection.Add(name, value);
						}
						startpos = pos + 1;
						pos = s.IndexOf('&', startpos);
					}
				}
				return requestCollection;
			}
		}
		
		#endregion
		
		/// <summary>
		/// Gets the string representation of this HTTPRequest.
		/// </summary>
		public override string ToString()
		{
			System.Text.StringBuilder b = new System.Text.StringBuilder("[HTTPRequest: ");
			b.AppendFormat("{0} {1}\n", method, uri);
			foreach (string k in headers.Keys) {
				b.AppendFormat("{0}: {1}\n", k, headers[k]);
			}
			b.Append("]");
			return b.ToString();
		}
	}
}
