// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections.Specialized;
using System.Globalization;
using System.IO;
using System.Net;
using System.Text;

namespace Grunwald.P2PLib.Server.Http
{
	/// <summary>
	/// This class helps sending an http response header.
	/// </summary>
	public class HttpResponse
	{
		#region Constructor
		
		/// <summary>
		/// Creates a new HTTPResponse.
		/// </summary>
		public HttpResponse()
		{
			Date = DateTime.UtcNow.ToString("r", CultureInfo.InvariantCulture);
			Server = "P2PLib Server";
			KeepAlive = false;
		}
		
		/// <summary>
		/// Creates a new HTTPResponse taking some values (e.g. KeepAlive) from what
		/// the client has sent in the request.
		/// </summary>
		public HttpResponse(HttpRequest request) : this()
		{
			string connectionHeader = request.Headers.Get("Connection");
			if (connectionHeader == null)
				KeepAlive = false;
			else
				KeepAlive = connectionHeader.ToLower() == "keep-alive";
		}
		
		#endregion
		
		#region HTTP Headers
		bool keepAlive;
		
		/// <summary>
		/// Gets/Sets if the connection should be kept open after the response was sent.
		/// </summary>
		public bool KeepAlive {
			get { return keepAlive; }
			set {
				keepAlive = value;
				Headers.Set("Connection", value ? "Keep-Alive" : "close");
				if (value)
					Headers.Set("Keep-Alive", "timeout=15, max=99");
				else
					Headers.Remove("Keep-Alive");
			}
		}
		
		long contentLength = -1;
		
		/// <summary>
		/// Gets/Sets the Length of the Content that should be sent.
		/// Set this to -1 if you don't want to send a Content-Length header.
		/// </summary>
		public long ContentLength {
			get { return contentLength; }
			set {
				if (value < -1)
					throw new ArgumentOutOfRangeException();
				contentLength = value;
				if (value < 0)
					Headers.Remove("Content-Length");
				else
					Headers.Set("Content-Length", value.ToString("d", CultureInfo.InvariantCulture));
			}
		}
		
		/// <summary>
		/// Gets/Sets the server name;
		/// </summary>
		public string Server {
			get { return Headers.Get("Server"); }
			set {
				if (value == null)
					Headers.Remove("Server");
				else
					Headers.Set("Server", value);
			}
		}
		
		/// <summary>
		/// Gets/Sets the content type;
		/// </summary>
		public string ContentType {
			get { return Headers.Get("Content-Type"); }
			set {
				if (value == null)
					Headers.Remove("Content-Type");
				else
					Headers.Set("Content-Type", value);
			}
		}
		
		/// <summary>
		/// Gets/Sets the date header;
		/// </summary>
		public string Date {
			get { return Headers.Get("Date"); }
			set {
				if (value == null)
					Headers.Remove("Date");
				else
					Headers.Set("Date", value);
			}
		}
		
		#endregion
		
		#region Status
		
		string statusMessage;
		/// <summary>
		/// The status message that will be sent to the browser. If this is set to null,
		/// <see cref="GetMessage"/> will be used for getting the message.
		/// </summary>
		public string StatusMessage {
			get { return statusMessage; }
			set { statusMessage = value; }
		}
		
		HttpStatusCode statusCode = HttpStatusCode.OK;
		/// <summary>
		/// The status code that will be sent to the browser. The default value is OK.
		/// </summary>
		public HttpStatusCode StatusCode {
			get { return statusCode; }
			set { statusCode = value; }
		}
		
		#endregion
		
		#region Headers
		
		WebHeaderCollection headers = new WebHeaderCollection();
		/// <summary>
		/// The Headers that should be sent to the browser
		/// </summary>
		public WebHeaderCollection Headers {
			get { return headers; }
		}
		
		/// <summary>
		/// <para>Merges the headers from the specified collection to the WebHeaderCollection,
		/// overwriting the existent headers.</para>
		/// <para>If <paramref name="col"/> is null, no operation is performed.</para>
		/// </summary>
		public void MergeHeaders(NameValueCollection col) {
			if (col == null)
				return;
			foreach (string key in col.Keys) {
				headers.Set(key, col[key]);
			}
		}
		
		#endregion
		
		#region SendHeaders
		
		/// <summary>
		/// Sends the HTTP-headers to the text writer and flushes the TextWriter.
		/// </summary>
		public void SendHeaders(TextWriter w)
		{
			SendHeaders(w, true);
		}
		
		/// <summary>
		/// Sends the HTTP-headers to the stream writer. If flush is false, the headers
		/// are not sent immediately but buffered to the TextWriter.
		/// </summary>
		/// <exception cref="ArgumentNullException"><paramref name="writer"/> is null.</exception>
		/// <exception cref="IOException">It is not possible to write to the <paramref name="writer"/>.</exception>
		public void SendHeaders(TextWriter writer, bool flush)
		{
			if (writer == null)
				throw new ArgumentNullException("writer");
			if (ContentLength == -1 && (int)StatusCode / 100 != 3)
				KeepAlive = false;
			StringBuilder b = new StringBuilder();
			string message = StatusMessage;
			if (message == null)
				message = GetMessage(StatusCode);
			b.AppendFormat("HTTP/1.1 {0:d} {1}\r\n", StatusCode, message);
			for (int i = 0; i < Headers.Count; ++i) {
				b.Append(Headers.GetKey(i));
				b.Append(": ");
				b.Append(Headers.Get(i));
				b.Append("\r\n");
			}
			b.Append("\r\n");
			try {
				writer.Write(b.ToString());
				if (flush)
					writer.Flush();
			} catch (ObjectDisposedException ex) {
				throw new IOException("Error sending headers.", ex);
			}
		}
		
		#endregion
		
		#region GetMessage
		
		/// <summary>
		/// Gets the HTTP status message for the status code.
		/// </summary>
		public static string GetMessage(HttpStatusCode statusCode)
		{
			switch ((int) statusCode) {
				case 100:
					return "Continue";
				case 101:
					return "Switching Protocols";
				case 200:
					return "OK";
				case 201:
					return "Created";
				case 202:
					return "Accepted";
				case 203:
					return "Non-Authoritative Information";
				case 204:
					return "No Content";
				case 205:
					return "Reset Content";
				case 206:
					return "Partial Content";
				case 300:
					return "Multiple Choices";
				case 301:
					return "Moved Permanently";
				case 302:
					return "Found";
				case 303:
					return "See Other";
				case 304:
					return "Not Modified";
				case 305:
					return "Use Proxy";
				case 306:
					return "(Unused)";
				case 307:
					return "Temporary Redirect";
				case 400:
					return "Bad Request";
				case 401:
					return "Unauthorized";
				case 402:
					return "Payment Required";
				case 403:
					return "Forbidden";
				case 404:
					return "Not Found";
				case 405:
					return "Method Not Allowed";
				case 406:
					return "Not Acceptable";
				case 407:
					return "Proxy Authentication Required";
				case 408:
					return "Request Timeout";
				case 409:
					return "Conflict";
				case 410:
					return "Gone";
				case 411:
					return "Length Required";
				case 412:
					return "Precondition Failed";
				case 413:
					return "Request Entity Too Large";
				case 414:
					return "Request-URI Too Long";
				case 415:
					return "Unsupported Media Type";
				case 416:
					return "Requested Range Not Satisfiable";
				case 417:
					return "Expectation Failed";
				case 500:
					return "Internal Server Error";
				case 501:
					return "Not Implemented";
				case 502:
					return "Bad Gateway";
				case 503:
					return "Service Unavailable";
				case 504:
					return "Gateway Timeout";
				case 505:
					return "HTTP Version Not Supported";
				default:
					return "Unknown";
			}
		}
		
		#endregion
		
		/// <summary>
		/// Gets the string representation of this HTTPResponse.
		/// </summary>
		public override string ToString()
		{
			return String.Format(CultureInfo.InvariantCulture,
			                     "[HTTPResponse: statusCode = {1}, statusMessage = {0}]",
			                     statusMessage,
			                     statusCode);
		}
	}
}
