// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;

namespace Grunwald.P2PLib.Server.Http
{
	/// <summary>
	/// Represents a collection of <see cref="IRequestHandler">IRequestHandler</see> objects.
	/// </summary>
	public class RequestHandlerCollection : CollectionBase
	{
		/// <summary>
		/// Gets the <see cref="IRequestHandler"/> with the highest priority for
		/// the request.
		/// </summary>
		public IRequestHandler GetHandler(HttpRequest request)
		{
			int priority = Int32.MinValue;
			IRequestHandler handler = null;
			foreach (IRequestHandler h in this) {
				int p = h.GetPriority(request);
				if (p > priority) {
					priority = p;
					handler = h;
				}
			}
			return handler;
		}
		
		#region Constructor
		
		/// <summary>
		/// Initializes a new instance of the <see cref="RequestHandlerCollection">RequestHandlerCollection</see> class.
		/// </summary>
		public RequestHandlerCollection()
		{
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="RequestHandlerCollection">RequestHandlerCollection</see> class containing the elements of the specified source collection.
		/// </summary>
		/// <param name="value">A <see cref="RequestHandlerCollection">RequestHandlerCollection</see> with which to initialize the collection.</param>
		public RequestHandlerCollection(RequestHandlerCollection value)
		{
			this.AddRange(value);
		}
		
		/// <summary>
		/// Initializes a new instance of the <see cref="RequestHandlerCollection">RequestHandlerCollection</see> class containing the specified array of <see cref="IRequestHandler">IRequestHandler</see> objects.
		/// </summary>
		/// <param name="value">An array of <see cref="IRequestHandler">IRequestHandler</see> objects with which to initialize the collection. </param>
		public RequestHandlerCollection(IRequestHandler[] value)
		{
			this.AddRange(value);
		}
		
		#endregion
		
		#region Indexer
		
		/// <summary>
		/// Gets the <see cref="IRequestHandler">IRequestHandler</see> at the specified index in the collection.
		/// <para>
		/// In C#, this property is the indexer for the <see cref="RequestHandlerCollection">RequestHandlerCollection</see> class.
		/// </para>
		/// </summary>
		public IRequestHandler this[int index] {
			get { return (IRequestHandler) (this.List[index]); }
		}
		
		#endregion
		
		#region Add
		
		/// <summary>
		/// Adds an item to the end of the collection
		/// </summary>
		/// <param name="value">Item to add</param>
		/// <returns>The positition where the new element has been added.</returns>
		public int Add(IRequestHandler value)
		{
			return this.List.Add(value);
		}
		
		/// <summary>
		/// Copies the elements of the specified <see cref="IRequestHandler">IRequestHandler</see> array to the end of the collection.
		/// </summary>
		/// <param name="value">An array of type <see cref="IRequestHandler">IRequestHandler</see> containing the objects to add to the collection.</param>
		public void AddRange(IRequestHandler[] value)
		{
			for (int i = 0; (i < value.Length); i = (i + 1)) {
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		/// Adds the contents of another <see cref="RequestHandlerCollection">RequestHandlerCollection</see> to the end of the collection.
		/// </summary>
		/// <param name="value">A <see cref="RequestHandlerCollection">RequestHandlerCollection</see> containing the objects to add to the collection. </param>
		public void AddRange(RequestHandlerCollection value)
		{
			for (int i = 0; (i < value.Count); i = (i + 1)) {
				this.Add((IRequestHandler) value.List[i]);
			}
		}
		
		#endregion
		
		#region Contains
		
		/// <summary>
		/// Gets a value indicating whether the collection contains the specified <see cref="IRequestHandler">IRequestHandler</see>.
		/// </summary>
		/// <param name="value">The <see cref="IRequestHandler">IRequestHandler</see> to search for in the collection.</param>
		/// <returns><b>true</b> if the collection contains the specified object; otherwise, <b>false</b>.</returns>
		public bool Contains(IRequestHandler value)
		{
			return this.List.Contains(value);
		}
		
		#endregion
		
		#region ToArray
		
		/// <summary>
		/// Copies the collection objects to a one-dimensional <see cref="System.Array">Array</see> instance beginning at the specified index.
		/// </summary>
		/// <param name="array">The one-dimensional <see cref="System.Array">Array</see> that is the destination of the values copied from the collection.</param>
		/// <param name="index">The index of the array at which to begin inserting.</param>
		public void CopyTo(IRequestHandler[] array, int index)
		{
			this.List.CopyTo(array, index);
		}
		
		/// <summary>
		/// Creates a one-dimensional <see cref="System.Array">Array</see> instance containing the collection items.
		/// </summary>
		/// <returns>Array of type IRequestHandler</returns>
		public IRequestHandler[] ToArray()
		{
			IRequestHandler[] array = new IRequestHandler[this.Count];
			this.CopyTo(array, 0);
			
			return array;
		}
		
		#endregion
		
		#region IndexOf
		
		/// <summary>
		/// Gets the index in the collection of the specified <see cref="IRequestHandler">IRequestHandler</see>, if it exists in the collection.
		/// </summary>
		/// <param name="value">The <see cref="IRequestHandler">IRequestHandler</see> to locate in the collection.</param>
		/// <returns>The index in the collection of the specified object, if found; otherwise, -1.</returns>
		public int IndexOf(IRequestHandler value)
		{
			return this.List.IndexOf(value);
		}
		
		#endregion
		
		#region Insert
		
		/// <summary>
		/// Inserts a new item into the collection.
		/// </summary>
		/// <param name="index">Position, whrere the item should be added</param>
		/// <param name="value">Item to add</param>
		public void Insert(int index, IRequestHandler value)
		{
			List.Insert(index, value);
		}
		
		#endregion
		
		#region Remove
		
		/// <summary>
		/// Removes the specified item from the collection
		/// </summary>
		/// <param name="value">Item to remove</param>
		public void Remove(IRequestHandler value)
		{
			List.Remove(value);
		}
		
		/// <summary>
		/// Removes the item at the given position
		/// </summary>
		/// <param name="index">Position of item</param>
		public new void RemoveAt(int index)
		{
			base.RemoveAt(index);
		}
		
		/// <summary>
		/// Clears all items from the collection
		/// </summary>
		public new void Clear()
		{
			base.Clear();
		}
		
		#endregion
		
		#region Enumerator
		
		/// <summary>
		/// Returns an enumerator that can iterate through the <see cref="RequestHandlerCollection">RequestHandlerCollection</see> instance.
		/// </summary>
		/// <returns>An <see cref="RequestHandlerCollectionEnumerator">RequestHandlerCollectionEnumerator</see> for the <see cref="RequestHandlerCollection">RequestHandlerCollection</see> instance.</returns>
		public new RequestHandlerCollectionEnumerator GetEnumerator()
		{
			return new RequestHandlerCollectionEnumerator(this);
		}
		
		/// <summary>
		/// Supports a simple iteration over a <see cref="RequestHandlerCollection">RequestHandlerCollection</see>.
		/// </summary>
		public class RequestHandlerCollectionEnumerator : IEnumerator
		{
			private IEnumerator _enumerator;
			private IEnumerable _temp;
			
			/// <summary>
			/// Initializes a new instance of the <see cref="RequestHandlerCollectionEnumerator">RequestHandlerCollectionEnumerator</see> class referencing the specified <see cref="RequestHandlerCollection">RequestHandlerCollection</see> object.
			/// </summary>
			/// <param name="mappings">The <see cref="RequestHandlerCollection">RequestHandlerCollection</see> to enumerate.</param>
			public RequestHandlerCollectionEnumerator(RequestHandlerCollection mappings)
			{
				_temp = mappings;
				_enumerator = _temp.GetEnumerator();
			}
			
			/// <summary>
			/// Gets the current element in the collection.
			/// </summary>
			public IRequestHandler Current {
				get { return (IRequestHandler) (_enumerator.Current); }
			}
			
			object IEnumerator.Current {
				get { return _enumerator.Current; }
			}
			
			/// <summary>
			/// Advances the enumerator to the next element of the collection.
			/// </summary>
			/// <returns><b>true</b> if the enumerator was successfully advanced to the next element; <b>false</b> if the enumerator has passed the end of the collection.</returns>
			public bool MoveNext()
			{
				return _enumerator.MoveNext();
			}
			
			bool IEnumerator.MoveNext()
			{
				return _enumerator.MoveNext();
			}
			
			/// <summary>
			/// Sets the enumerator to its initial position, which is before the first element in the collection.
			/// </summary>
			public void Reset()
			{
				_enumerator.Reset();
			}
			
			void IEnumerator.Reset()
			{
				_enumerator.Reset();
			}
		}
		
		#endregion
	}
}
