// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections.Specialized;
using System.IO;
using System.Net.Sockets;
using log4net;

namespace Grunwald.P2PLib.Server
{
	/// <summary>
	/// This server accepts connections and waits for the client to send an
	/// HTTP-like-request (multiple CRLF-lines terminated with an empty line).
	/// </summary>
	public class HttpLikeServer : TcpServer
	{
		static readonly ILog log = LogManager.GetLogger(typeof (HttpLikeServer));
		
		ConnectionHandlerCollection connectionHandlers = new ConnectionHandlerCollection();
		
		/// <summary>
		/// This collection contains the <see cref="IConnectionHandler">IConnectionHandlers</see>
		/// this server uses.
		/// </summary>
		public ConnectionHandlerCollection ConnectionHandlers {
			get { return connectionHandlers; }
		}
		
		/// <summary>
		/// Accepts the connection with the socket
		/// </summary>
		protected override void OnAccept(Socket socket)
		{
			DelRequest d = new DelRequest(GetRequest);
			d.BeginInvoke(socket, null, null);
		}
		
		delegate void DelRequest(Socket socket);
		
		/// <summary>
		/// Gets a StringCollection containing the lines of the request on the connection
		/// until the first empty line.
		/// </summary>
		public static StringCollection GetRequest(IConnection c)
		{
			try {
				StreamReader r = c.Reader;
				StringCollection request = new StringCollection();
				string line = r.ReadLine();
				if (line == null) {
					c.Close(); // set Connection.Connected to false, really necessary for sockets!!
					return null;
				}
				while (line.Length > 0) {
					request.Add(line);
					line = r.ReadLine();
					if (line == null) {
						c.Close();
						return null;
					}
				}
				return request;
			} catch (ObjectDisposedException) {
				return null;
			} catch (IOException) {
				return null;
			}
		}
		
		private void GetRequest(Socket socket)
		{
			try {
				AcceptRequests(new TcpConnection(socket));
			} catch (Exception ex) {
				if (Helper.IsCritical(ex))
					throw;
				socket.Close();
				log.Error(ex);
			}
		}
		
		/// <summary>
		/// Runs a loop on this connection accepting all requests.
		/// </summary>
		public void AcceptRequests(IConnection connection)
		{
			while (connection.Connected) {
				AcceptOneRequest(connection);
			}
		}
		
		/// <summary>
		/// Accepts the next request on the connection and handles it.
		/// </summary>
		public void AcceptOneRequest(IConnection connection)
		{
			StringCollection request = GetRequest(connection);
			if (request == null) return;
			OnRequest(connection, request);
		}
		
		/// <summary>
		/// Processes the request by calling the request-event.
		/// </summary>
		protected virtual void OnRequest(IConnection connection, StringCollection request)
		{
			IConnectionHandler h = ConnectionHandlers.GetHandler(request);
			if (h == null) {
				connection.Close();
				return;
			}
			h.Handle(connection, request);
		}
	}
}
