// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Net;

namespace Grunwald.P2PLib.Server
{
	/// <summary>
	/// The IPDetector provides a mechanism to detect the current internet ip.
	/// </summary>
	public abstract class IPDetector
	{
		#region Constructor
		
		/// <summary>
		/// Creates a new IPDetector
		/// </summary>
		protected IPDetector()
		{
			delegateRefresh = new DelegateRefresh(Refresh);
		}
		
		#endregion
		
		#region LastRefresh
		DateTime lastRefresh;
		
		/// <summary>
		/// Returns the time when the last refresh was started.
		/// </summary>
		public virtual DateTime LastRefresh {
			get { return lastRefresh; }
		}
		
		#endregion
		
		#region LastSuccess
		DateTime lastSuccess;
		
		/// <summary>
		/// Returns the time when the last refresh ended successfull.
		/// </summary>
		public virtual DateTime LastSuccess {
			get { return lastSuccess; }
		}
		
		#endregion
		
		#region RequestHeaders
		WebHeaderCollection requestHeaders = new WebHeaderCollection();
		
		/// <summary>
		/// The collection of request headers that are sent to the server.
		/// </summary>
		public WebHeaderCollection RequestHeaders {
			get {
				return requestHeaders;
			}
		}
		#endregion
		
		#region Error
		IPDetectorError lastError = IPDetectorError.NeverTested;
		
		/// <summary>
		/// Gets the reason for the last error.
		/// </summary>
		public IPDetectorError LastError {
			get {
				return lastError;
			}
			set {
				lastError = value;
			}
		}
		#endregion
		
		#region Refresh
		
		delegate bool DelegateRefresh();
		
		DelegateRefresh delegateRefresh;
		
		/// <summary>
		/// Starts an asynchron refresh
		/// </summary>
		public virtual IAsyncResult BeginRefresh()
		{
			return BeginRefresh(null, null);
		}
		
		/// <summary>
		/// Starts an asynchron refresh
		/// </summary>
		public virtual IAsyncResult BeginRefresh(AsyncCallback asyncCallback, object state)
		{
			return delegateRefresh.BeginInvoke(asyncCallback, state);
		}
		
		/// <summary>
		/// Waits for the end of an asynchron refresh.
		/// </summary>
		/// <returns>Returns true if the refresh was successfull.</returns>
		public virtual bool EndRefresh(IAsyncResult result)
		{
			return delegateRefresh.EndInvoke(result);
		}
		
		/// <summary>
		/// Re-Detects the IP.
		/// </summary>
		/// <returns>Returns true if the refresh was successfull.</returns>
		public virtual bool Refresh()
		{
			lastRefresh = DateTime.Now;
			bool success = DoRefresh();
			lastRefreshWasSuccessful = success;
			if (success) {
				lastSuccess = DateTime.Now;
				IPAddress newIP = this.IPAddress;
				if (newIP == null)
					throw new ApplicationException("If IPDetector.DoRefresh() returns true, IPDetector.IPAddress may not return null.");
				if (IPAddressChanged != null) {
					if (oldIP == null) {
						IPAddressChanged(this, EventArgs.Empty);
					} else if (!oldIP.Equals(newIP)) {
						IPAddressChanged(this, EventArgs.Empty);
					}
				}
				oldIP = newIP;
			}
			return success;
		}
		
		/// <summary>
		/// Re-Detects the IP without setting the time properties.
		/// </summary>
		/// <returns>Returns true if the refresh was successfull.</returns>
		protected abstract bool DoRefresh();
		
		#endregion
		
		#region IPAddress
		IPAddress oldIP;
		
		/// <summary>
		/// Gets called whenever a new IPAddress is detected.
		/// </summary>
		public event EventHandler IPAddressChanged;
		
		/// <summary>
		/// Gets the IPAddress that was detected last.
		/// </summary>
		public abstract IPAddress IPAddress { get; }
		
		#endregion
		
		#region LastRefreshWasSuccessful
		bool lastRefreshWasSuccessful = false;
		
		/// <summary>
		/// Returns true if the last refresh was successful.
		/// </summary>
		public virtual bool LastRefreshWasSuccessful {
			get { return lastRefreshWasSuccessful; }
		}
		
		#endregion
	}
	
	/// <summary>
	/// Enumeration containing the possible errors in the IP Detectors
	/// </summary>
	public enum IPDetectorError
	{
		/// <summary>
		/// The IP Detector never ran.
		/// </summary>
		NeverTested,
		/// <summary>
		/// Error accessing the url.
		/// </summary>
		WebException,
		/// <summary>
		/// Server returned HTTP error.
		/// </summary>
		ServerError,
		/// <summary>
		/// Unknown exception error
		/// </summary>
		UnhandledException,
		/// <summary>
		/// Invalid server response.
		/// </summary>
		InvalidResponse,
		/// <summary>
		/// Valid server response that didn't contain the address.
		/// </summary>
		AddressNotFoundInResponse
	}
}
