// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Net;
using System.Text.RegularExpressions;
using log4net;

namespace Grunwald.P2PLib.Server.IPDetectors
{
	/// <summary>
	/// This Detector uses a Text page as source.
	/// </summary>
	public class TextBasedDetector : IPDetector
	{
		static readonly ILog log = LogManager.GetLogger(typeof (TextBasedDetector));
		
		string url;
		/// <summary>
		/// The URL used for getting the detector
		/// </summary>
		public string Url {
			get { return url; }
			set {
				if (value == null) throw new ArgumentNullException();
				url = value;
			}
		}
		
		/// <summary>
		/// Creates a new TextBasedDetector using showmyip.com
		/// </summary>
		public TextBasedDetector()
		{
			this.Url = "http://www.showmyip.com/simple/";
		}
		
		/// <summary>
		/// Creates a new TextBasedDetector.
		/// </summary>
		public TextBasedDetector(string url)
		{
			this.Url = url;
		}
		
		#region Refresh
		
		Regex regularExpression = new Regex(@"\b(\d{1,3}\.){3}\d{1,3}\b");
		/// <summary>
		/// Gets/Sets the regular expression used for finding the IP in the text.
		/// The default value is @"\b(\d{1,3}\.){3}\d{1,3}\b".
		/// </summary>
		public Regex RegularExpression {
			get { return regularExpression; }
			set {
				if (value == null) throw new ArgumentNullException();
				regularExpression = value;
			}
		}
		
		/// <summary>
		/// Re-Detects the IP without setting the time properties.
		/// </summary>
		/// <returns>Returns true if the refresh was successfull.</returns>
		protected override bool DoRefresh()
		{
			try {
				using (WebClient wc = new WebClient()) {
					wc.Headers.Add(this.RequestHeaders);
					Stream s;
					if (Url.StartsWith("http"))
						s = wc.OpenRead(Url);
					else
						s = File.OpenRead(Url);
					StreamReader r = new StreamReader(s);
					string txt = r.ReadToEnd();
					r.Close();
					s.Close();
					lastTextDocument = txt;
					Match m = RegularExpression.Match(txt);
					if (!m.Success) {
						this.LastError = IPDetectorError.AddressNotFoundInResponse;
						return false;
					}
					lastIP = IPAddress.Parse(m.Value);
					return true;
				}
			} catch (WebException) {
				this.LastError = IPDetectorError.WebException;
				return false;
			} catch (Exception ex) {
				this.LastError = IPDetectorError.UnhandledException;
				log.Error("Error in DoRefresh", ex);
				throw;
			}
		}
		
		#endregion
		
		#region IP-Address
		string lastTextDocument;
		
		/// <summary>
		/// The full response from the last successfull request.
		/// </summary>
		public string LastTextDocument {
			get { return lastTextDocument; }
		}
		
		private IPAddress lastIP;
		
		/// <summary>
		/// Gets the IPAddress that was detected last.
		/// </summary>
		public override IPAddress IPAddress {
			get { return lastIP; }
		}
		
		#endregion
		
		/// <summary>
		/// Gets the string representation of this TextBasedDetector.
		/// </summary>
		public override string ToString()
		{
			return String.Format("[TextBasedDetector: Url = {0}, RegularExpression = {1}, IPAddress = {2}]",
			                     Url,
			                     RegularExpression,
			                     IPAddress);
		}
	}
}
