// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Net;
using System.Xml;
using log4net;

namespace Grunwald.P2PLib.Server.IPDetectors
{
	/// <summary>
	/// This Detector uses a XML web page as source.
	/// </summary>
	public class XmlBasedDetector : IPDetector
	{
		#region Fields
		static readonly ILog log = LogManager.GetLogger(typeof (XmlBasedDetector));
		
		string url;
		/// <summary>
		/// Gets/Sets the URL of the XML page.
		/// </summary>
		public string Url {
			get { return url; }
			set {
				if (value == null) throw new ArgumentNullException();
				url = value;
			}
		}
		
		XmlAddress addressOfIP;
		/// <summary>
		/// Gets/Sets the address inside the XML document where the ip address can be found.
		/// </summary>
		public XmlAddress AddressOfIP {
			get { return addressOfIP; }
			set {
				if (value == null) throw new ArgumentNullException();
				addressOfIP = value;
			}
		}
		
		#endregion
		
		#region Contructors
		
		/// <summary>
		/// Creates a new detector using www.showmyip.com as source.
		/// </summary>
		public XmlBasedDetector()
		{
			this.Url = "http://www.showmyip.com/xml/";
			AddressOfIP = new XmlAddress("/ip_address/ip");
		}
		
		/// <summary>
		/// Creates a new detector using the url and xmlPath as source.
		/// </summary>
		public XmlBasedDetector(string url, string xmlPath)
		{
			this.Url = url;
			AddressOfIP = new XmlAddress(xmlPath);
		}
		
		/// <summary>
		/// Creates a new detector using the url and xmlPath as source.
		/// </summary>
		public XmlBasedDetector(string url, XmlAddress ip)
		{
			this.Url = url;
			AddressOfIP = ip;
		}
		
		#endregion
		
		#region Refresh
		
		/// <summary>
		/// Re-Detects the IP without setting the time properties.
		/// </summary>
		/// <returns>Returns true if the refresh was successfull.</returns>
		protected override bool DoRefresh()
		{
			try {
				XmlDocument xml = new XmlDocument();
				if (Url.StartsWith("http://")) {
					using (WebClient wc = new WebClient()) {
						wc.Headers.Add(this.RequestHeaders);
						using (Stream s = wc.OpenRead(Url)) {
							xml.Load(s);
						}
					}
				} else {
					xml.Load(Url);
				}
				string ip = AddressOfIP.GetValue(xml);
				lastXmlDocument = xml;
				if (ip == null) {
					this.LastError = IPDetectorError.AddressNotFoundInResponse;
					return false;
				}
				lastIP = IPAddress.Parse(ip);
				return true;
			} catch (WebException ex) {
				if (ex.Status == WebExceptionStatus.ProtocolError)
					this.LastError = IPDetectorError.ServerError;
				else
					this.LastError = IPDetectorError.WebException;
				return false;
			} catch (XmlException) {
				this.LastError = IPDetectorError.InvalidResponse;
				return false;
			} catch (Exception ex) {
				this.LastError = IPDetectorError.UnhandledException;
				log.Error("Error in DoRefresh", ex);
				throw;
			}
		}
		
		#endregion
		
		#region IP-Address
		XmlDocument lastXmlDocument;
		
		/// <summary>
		/// The full XmlDocument from the last successfull request.
		/// </summary>
		public XmlDocument LastXmlDocument {
			get { return lastXmlDocument; }
		}
		
		private IPAddress lastIP;
		
		/// <summary>
		/// Gets the IPAddress that was detected last.
		/// </summary>
		public override IPAddress IPAddress {
			get { return lastIP; }
		}
		
		#endregion
		
		/// <summary>
		/// Gets the string representation of this TextBasedDetector.
		/// </summary>
		public override string ToString()
		{
			return String.Format("[XmlBasedDetector: URL = {0}, AddressOfIP = {1}, IPAddress = {2}]",
			                     url,
			                     addressOfIP,
			                     lastIP);
		}
	}
}
