// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Globalization;
using System.IO;
using System.Net;
using log4net;

namespace Grunwald.P2PLib.Server
{
	/// <summary>
	/// Tests if the server can be reached from the internet.
	/// This class uses a http "ping" request
	/// </summary>
	public class ServerAbilityTester
	{
		// We have to totally make a new test concept.
		// We can't test ourselves directly.
		// This method is only correct if you are behind a NAT.
		// Every *direct* internet access (corporate network, dial-in without router, ...)
		// detects server ability every time even if we are behind a firewall.
		
		#region Constructor
		
		static readonly ILog log = LogManager.GetLogger(typeof (ServerAbilityTester));
		
		/// <summary>
		/// Creates a new ServerAbilityTester testing if the server can be reached from
		/// the internet.
		/// </summary>
		public ServerAbilityTester()
		{
			delegateTest = new DelegateTest(Test);
		}
		
		#endregion
		
		#region Last Test
		DateTime lastTestStart = DateTime.MinValue;
		
		/// <summary>
		/// Gets the time when the last test was started.
		/// </summary>
		public DateTime LastTestStart {
			get { return lastTestStart; }
		}
		
		DateTime lastFinishedTest = DateTime.MinValue;
		
		/// <summary>
		/// Gets the time when the last test has ended.
		/// </summary>
		public DateTime LastFinishedTest {
			get { return lastFinishedTest; }
		}
		
		bool lastResult = false;
		
		/// <summary>
		/// Gets the last result.
		/// </summary>
		public bool LastResult {
			get { return lastResult; }
		}
		#endregion
		
		#region Events
		/// <summary>
		/// Is raised when a test is started.
		/// </summary>
		public event EventHandler TestStarted;
		
		/// <summary>
		/// Is raised when a test has finished.
		/// </summary>
		public event EventHandler TestFinished;
		
		/// <summary>Raises the TestFinished event.</summary>
		protected virtual void OnTestFinished(EventArgs e)
		{
			if (TestFinished != null) {
				TestFinished(this, e);
			}
		}
		
		/// <summary>Raises the TestStarted event.</summary>
		protected virtual void OnTestStarted(EventArgs e)
		{
			if (TestStarted != null) {
				TestStarted(this, e);
			}
		}
		#endregion
		
		#region Begin/End Test
		
		delegate bool DelegateTest(IPEndPoint ipEndPoint, int controlId);
		
		DelegateTest delegateTest;
		
		/// <summary>
		/// Starts an asynchron test
		/// </summary>
		public virtual IAsyncResult BeginTest(IPEndPoint ipEndPoint, int controlId)
		{
			return BeginTest(ipEndPoint, controlId, null, null);
		}
		
		/// <summary>
		/// Starts an asynchron test
		/// </summary>
		public virtual IAsyncResult BeginTest(IPEndPoint ipEndPoint, int controlId, AsyncCallback asyncCallback, object state)
		{
			return delegateTest.BeginInvoke(ipEndPoint, controlId, asyncCallback, state);
		}
		
		/// <summary>
		/// Waits for the end of an asynchron test.
		/// </summary>
		/// <returns>Returns true if the server port can be reached from the internet.</returns>
		public virtual bool EndTest(IAsyncResult result)
		{
			return delegateTest.EndInvoke(result);
		}
		
		#endregion
		
		#region Test
		/// <summary>
		/// The name of the service
		/// </summary>
		public const string PingPath = "/_special/p2plib_ping";
		
		/// <summary>
		/// Runs the test by executing ExecuteTest and then sets the LastResult and LastFinishedTest
		/// properties.
		/// </summary>
		/// <param name="ipEndPoint">the IPEndPoint on which the server listens</param>
		/// <param name="controlId">a (possibly unique) id of the server. If the server
		/// doesn't return this id in his answer, maybe another server listens on the
		/// port. If you use 0 for this, every answer from the server will be accepted.</param>
		/// <returns>Returns true if the server port can be reached from the internet.</returns>
		public bool Test(IPEndPoint ipEndPoint, int controlId)
		{
			lastTestStart = DateTime.Now;
			OnTestStarted(EventArgs.Empty);
			try {
				bool result = ExecuteTest(ipEndPoint, controlId);
				lastResult = result;
				lastFinishedTest = DateTime.Now;
				try {
					OnTestFinished(EventArgs.Empty);
				} catch (Exception ex) {
					if (Helper.IsCritical(ex))
						throw;
					log.Error(ex);
				}
				return result;
			} catch (Exception ex) {
				if (Helper.IsCritical(ex))
					throw;
				lastResult = false;
				lastFinishedTest = DateTime.Now;
				log.Error(ex);
				throw;
			}
		}
		
		/// <summary>
		/// Protected method to run the test. Is executed by Test().
		/// </summary>
		/// <param name="ipEndPoint">the IPEndPoint on which the server listens</param>
		/// <param name="controlId">a (possibly unique) id of the server. If the server
		/// doesn't return this id in his answer, maybe another server listens on the
		/// port. If you use 0 for this, every answer from the server will be accepted.</param>
		/// <returns>Returns true if the server port can be reached from the internet.</returns>
		protected virtual bool ExecuteTest(IPEndPoint ipEndPoint, int controlId)
		{
			string url = "http://" + ipEndPoint.ToString() + PingPath;
			string correctAnswer = "[[P2PLib:" + controlId.ToString("d", CultureInfo.InvariantCulture) + "]]";
			try {
				HttpWebRequest request = (HttpWebRequest) HttpWebRequest.Create(url);
				request.UserAgent = "P2PLib Mastertest";
				request.Timeout = 10000;
				request.AllowAutoRedirect = false;
				WebResponse res = request.GetResponse();
				StreamReader sr = new StreamReader(res.GetResponseStream());
				string data = sr.ReadToEnd();
				sr.Close();
				res.Close();
				if (controlId == 0 || data == correctAnswer)
					return true;
				else
					return false;
			} catch (WebException) {
				return false;
			}
		}
		
		#endregion
	}
}
