// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Globalization;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using log4net;

namespace Grunwald.P2PLib.Server
{
	/// <summary>
	/// Implements the AbstractServer as an TCP/IP-Server
	/// </summary>
	public abstract class TcpServer : AbstractServer
	{
		Thread thread;
		
		static readonly ILog log = LogManager.GetLogger(typeof(TcpServer));
		
		#region Running/Start/Stop
		
		/// <summary>
		/// Gets if the server is currently running
		/// </summary>
		public override bool Running {
			get { return thread != null; }
		}
		
		/// <summary>
		/// Starts the server
		/// </summary>
		public override void Start()
		{
			if (Port == 0)
				throw new InvalidOperationException("You have to initialize Port before calling Start()");
			if (Running)
				throw new InvalidOperationException("Server is already running.");
			
			Bind();
			
			log.Debug("Server started on port " + Port.ToString(CultureInfo.CurrentCulture));
			thread = new Thread(new ThreadStart(Run));
			thread.IsBackground = true;
			thread.Name = "TCPServer on port " + Port.ToString(CultureInfo.CurrentCulture);
			thread.Start();
		}
		
		/// <summary>
		/// Stops the server
		/// </summary>
		public override void Stop()
		{
			if (!Running)
				throw new InvalidOperationException("Server is not running.");
			log.Debug("Stopping server...");
			thread = null;
			socket.Close();
			// Wait until socket is closed (so the thread has stopped)
			// If we don't wait, Stop();Start(); will result in an error because
			// the old socket was not closed yet.
			while (socket != null)
				Thread.Sleep(5);
		}
		
		#endregion
		
		#region CreateSocket
		Socket socket;
		
		void Bind()
		{
			socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
			socket.Bind(new IPEndPoint(IPAddress.Any, Port));
			socket.Listen(10);
		}
		
		#endregion
		
		#region Run
		
		private void Run()
		{
			try {
				Socket nextSocket;
				while (thread != null) {
					try {
						nextSocket = socket.Accept();
					} catch (SocketException) {
						if (thread != null)
							throw;
						else
							break;
					}
					try {
						OnAccept(nextSocket);
					} catch (Exception ex) {
						if (Helper.IsCritical(ex))
							throw;
						log.Error(ex);
						break;
					}
				}
			} catch (Exception ex) {
				if (Helper.IsCritical(ex))
					throw;
				log.Fatal(ex);
			} finally {
				socket = null;
			}
		}
		
		#endregion
		
		/// <summary>
		/// Accepts the connection with the socket
		/// </summary>
		protected abstract void OnAccept(Socket socket);
	}
}
