// P2PLib - Copyright (C) 2004-2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Sockets;

namespace Grunwald.P2PLib
{
	// TODO: Think about renaming this as it isn't restricted to TCP.
	
	/// <summary>
	/// Contains some objects to access the socket.
	/// </summary>
	public class TcpConnection : IConnection
	{
		#region Properties
		Socket socket;
		
		Stream stream;
		/// <summary>
		/// The stream for accessing this connection.
		/// </summary>
		public Stream Stream {
			get { return stream; }
		}
		
		StreamReader reader;
		
		/// <summary>
		/// The reader for reading data from this connection.
		/// </summary>
		public StreamReader Reader {
			get { return reader; }
		}
		
		StreamWriter writer;
		/// <summary>
		/// The writer for sending data on this connection.
		/// </summary>
		public StreamWriter Writer {
			get { return writer; }
		}
		
		/// <summary>
		/// Returns if the connection is still alive
		/// </summary>
		public virtual bool Connected {
			get {
				if (socket == null)
					throw new InvalidOperationException("When using a custom stream for" +
					                                    "TCPConnection, you have to override" +
					                                    "Close() and Connected");
				return socket.Connected;
			}
		}
		#endregion
		
		#region Constructors
		/// <summary>
		/// Creates a new TcpConnection for a socket.
		/// </summary>
		public TcpConnection(Socket socket)
		{
			if (socket == null)
				throw new ArgumentNullException("socket");
			this.socket = socket;
			stream = new NetworkStream(socket);
			reader = new StreamReader(Stream);
			writer = new StreamWriter(Stream);
		}
		
		/// <summary>
		/// Creates a new TcpConnection for any stream.
		/// </summary>
		protected TcpConnection(Stream stream)
		{
			if (stream == null)
				throw new ArgumentNullException("stream");
			this.stream = stream;
			reader = new StreamReader(stream);
			writer = new StreamWriter(stream);
		}
		#endregion
		
		#region Close
		/// <summary>
		/// Closes the connection
		/// </summary>
		public virtual void Close()
		{
			Writer.Close();
			Reader.Close();
			Stream.Close();
			if (socket == null)
				throw new InvalidOperationException("When using a custom stream for" +
				                                    "TCPConnection, you have to override" +
				                                    "Close() and Connected");
			socket.Close();
		}
		#endregion
		
		#region ToString
		/// <summary>
		/// Gets the string representation of this TCPConnection.
		/// </summary>
		public override string ToString()
		{
			if (socket == null)
				return String.Format(CultureInfo.InvariantCulture, "[{0}: Stream = {1}]", GetType().FullName, Stream);
			else
				return String.Format(CultureInfo.InvariantCulture, "[{0}: RemoteEndPoint = {1}]", GetType().FullName, socket.RemoteEndPoint);
		}
		#endregion
		
		/// <summary>
		/// Gets the remote end point this connection is connected with.
		/// Can be null if the end point is unknown.
		/// </summary>
		public EndPoint RemoteEndPoint {
			get {
				if (socket != null)
					return socket.RemoteEndPoint;
				else
					return null;
			}
		}
	}
}
