// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Net;
using System.Windows.Forms;
using System.Xml;

using Grunwald.Gui;

using DGNetwork.Services;
using DGNetwork.AddIns;

using ICSharpCode.Core.AddIns;
using ICSharpCode.Core.AddIns.Codons;
using ICSharpCode.Core.Services;

namespace DGNetwork.StartUp
{
	public class Startup
	{
		public const string BuildName = ""
			#if MONO
			+ " Mono"
			#endif
			#if DEBUG
			+ " Debug"
			#endif
			;
		
		[STAThread]
		public static int Main(params string[] args)
		{
			return new Startup().Start(args);
		}
		
		public int Start(string[] args)
		{
			bool ok = true;
			try {
				Application.ThreadException += new ThreadExceptionEventHandler(ThreadException);
				AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(UnhandledException);
				log4net.Config.XmlConfigurator.Configure(new FileInfo(AppDomain.CurrentDomain.SetupInformation.ConfigurationFile));
				CommandLine cmd = new CommandLine(args);
				CommandLine.Instance = cmd;
				if (cmd.HasFlag("help") || cmd.HasFlag("?")) {
					MessageBox.Show("Sorry, no command line help available.", "DG Network");
					return 0;
				}
				if (cmd.HasFlag("postUpdate")) {
					cmd.SetFlag("show");
				}
				if (cmd.HasFlag("test")) {
					if (!cmd.HasFlag("propertyFile"))
						cmd["propertyFile"] = "TestModeProperties.xml";
				}
				
				DefaultMessageService.ActivateBuffer();
				InitializeCore();
				
				string mutexName = cmd.HasFlag("test") ? "DGNetworkSingleInstanceTestMutex" : "DGNetworkSingleInstanceMutex";
				bool firstInstance;
				using (Mutex mtx = new Mutex(true, mutexName, out firstInstance)) {
					if (firstInstance) {
						InitializeAddins();
						RunApplication();
					} else {
						NotifyOldInstance();
					}
				}
				return 0;
			} catch (Exception ex) {
				try {
					DefaultMessageService.ReleaseBuffer();
				} catch (Exception) { }
				ok = false;
				ErrorReporter.ShowErrorReport(ex);
				return 1;
			} finally {
				try {
					ServiceManager.Services.UnloadAllServices();
				} catch (Exception ex) {
					// ignore shutdown errors if there were runtime errors
					if (ok)
						ErrorReporter.ShowErrorReport(ex);
				}
			}
		}
		
		void NotifyOldInstance()
		{
			int port = Service.Property.GetProperty(P2PService.PortPropertyKey, P2PService.DefaultPort);
			try {
				using (WebClient wc = new WebClient()) {
					wc.DownloadData("http://localhost:" + port + "/dgn/showmainform");
				}
			} catch (WebException ex) {
				Service.Message.ShowErrorFormatted("${res:Global.NotifyOldInstanceFailed}", ex.Message);
			}
		}
		
		void RunApplication()
		{
			if (CommandLine.Instance.HasFlag("show"))
				Service.Window.ShowMainWindow();
			Service.Window.BeginInvoke(new MethodInvoker(DefaultMessageService.ReleaseBuffer));
			Application.Run();
		}
		
		class StartupException : ApplicationException {
			public StartupException(Exception innerException)
				: base(GetText(), innerException)
			{
			}
			
			static string GetText() {
				IResourceService res = (IResourceService)ServiceManager.Services.GetService(typeof(IResourceService));
				if (res == null)
					return "";
				else
					return res.GetString("Global.StartupError", false);
			}
		}
		
		string defaultCoreDirectory = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location) +
			Path.DirectorySeparatorChar + "AddIns";
		string defaultDataDirectory = Path.GetDirectoryName(Assembly.GetEntryAssembly().Location) +
			Path.DirectorySeparatorChar + "data";
		
		void InitializeCore()
		{
			bool ignoreDefaultPath = true;
			string[] addInDirs = { defaultCoreDirectory };
			AddInTreeSingleton.SetAddInDirectories(addInDirs, ignoreDefaultPath);
			
			string path = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "/Grunwald/DG Network";
			string propertyFile = CommandLine.Instance["propertyFile"];
			if (propertyFile == null)
				propertyFile = "DGNetworkProperties.xml";
			PropertyService.SetInformation(path, propertyFile, "DGNetworkProperties");
			PropertyService.SetDataDir(defaultDataDirectory);
			
			ServiceManager.Services.AddService(new PropertyService());
			ServiceManager.Services.AddService(new DefaultMessageService());
			ServiceManager.Services.AddService(new DefaultResourceService());
			
			Service.StringParser.Properties.Add("AppBuildName", BuildName);
			Service.StringParser.Properties.Add("AppIdentification", Service.Parse("DG Network ${exe:ProductVersion}${AppBuildName}"));
		}
		
		void InitializeAddins()
		{
			try {
				Translator.DefaultTranslator = new DGNetwork.Gui.DGNTranslator();
				ErrorReporter error = ErrorReporter.DefaultErrorReporter;
				error.KillAppButtonText = "${res:ErrorReporter.ExitApplication}";
				error.ReportButtonText = "${res:ErrorReporter.ReportError}";
				error.IgnoreButtonText = "${res:ErrorReporter.Ignore}";
				error.InfoText = "${res:ErrorReporter.InfoText}";
				error.Title = "${res:ErrorReporter.Title}";
				error.SendReport += new ReportEventHandler(ErrorReportEventHandler);
				
				//AddInTreeSingleton.CreateAddInTree(new DGNAddInTree());
				ServiceManager.Services.InitializeServicesSubsystem("/Services");
				
				ArrayList commands;
				commands = AddInTreeSingleton.AddInTree.GetTreeNode("/Autostart").BuildChildItems(null);
				for (int i = 0; i < commands.Count; ++i) {
					((ICommand)commands[i]).Run();
				}
			} catch (XmlException ex) {
				// maybe display another dialog for malformed XML files?
				throw new StartupException(ex);
			} catch (Exception ex) {
				throw new StartupException(ex);
			}
		}
		
		void ErrorReportEventHandler(object sender, ReportEventArgs e)
		{
			WindowService windowService = (WindowService)ServiceManager.Services.GetService(typeof(WindowService));
			if (windowService != null && windowService.InvokeRequired) {
				windowService.Invoke(new ReportEventHandler(ErrorReportEventHandler), sender, e);
				return;
			}
			string ver = Service.Parse("${AppIdentification}");
			string txt = ver + "\r\n" +
				".NET-Version: " + Environment.Version + "\r\n" +
				"Betriebsystem: " + Environment.OSVersion.Platform.ToString() + "  " + Environment.OSVersion.Version.ToString() + "\r\n" +
				"\r\n\r\n" + e.Exception.ToString();
			Clipboard.SetDataObject(txt);
			Service.Browser.OpenWindow("http://www.danielgrunwald.de/network/bugreport.php?version=" + Helper.EscapePart(ver));
		}
		
		void UnhandledException(object sender, UnhandledExceptionEventArgs e)
		{
			ErrorReporter.ShowErrorReport((Exception)e.ExceptionObject);
		}
		
		void ThreadException(object sender, ThreadExceptionEventArgs e)
		{
			ErrorReporter.ShowErrorReport(e.Exception);
		}
	}
}
