// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using Grunwald.Gui.Components;
using Grunwald.Gui.Settings;
using Grunwald.Gui.Settings.Editors;

namespace TCPSniffer
{
	public enum AcceptMode {
		ByFilename,
		ByProcessId
	}
	
	public class ApplicationFilter : Filter
	{
		public ApplicationFilter() { }
		public ApplicationFilter(ApplicationFilter f) : base(f) {}
		
		public override string Icon { get { return "Software"; } }
		
		[Setting("acceptMode", "Accept mode", "Sets the mode how the processes are" +
		         "accepted.")]
		public AcceptMode acceptMode;
		
		[Setting("process", "Process", "The process that should be accepted.")]
		public ProcessInfo info;
		
		protected override bool AcceptPacket(IPPacket p) {
			if (p is TcpPacket) {
				return AcceptProcess(((TcpPacket)p).Process);
			} else if (p is UdpPacket) {
				return AcceptProcess(((UdpPacket)p).Process);
			} else {
				return false;
			}
		}
		
		private bool AcceptProcess(ProcessInfo p)
		{
			if (info == null)
				return true;
			if (p == null)
				return false;
			if (acceptMode == AcceptMode.ByProcessId)
				return p.Id == info.Id;
			return ProcessCache.NormalizePath(p.Filename) == ProcessCache.NormalizePath(info.Filename);
		}
		
		public override string ToString() {
			if (info == null)
				return "ApplicationFilter: " + NegString + "[no process specified]";
			if (acceptMode == AcceptMode.ByProcessId)
				return String.Format("ApplicationFilter: {2}[{0} ({1})]", info.Id, info.Name, NegString);
			else
				return String.Format("ApplicationFilter: {1}[{0}]", info.Name, NegString);
		}
	}
	
	public class ApplicationEditor : SingleControlEditor
	{
		/// <summary>
		/// Creates a new TextBoxBasedEditor.
		/// </summary>
		public ApplicationEditor(ISetting setting) : base(setting, new ListView())
		{
			list = (ListView)base.Control;
			list.View = View.Details;
			list.SmallImageList = ProcessCache.Instance.ImageList;
			ColumnHeader h;
			h = new ColumnHeader();
			h.Text = "Application";
			h.Width = list.Width * 2 / 3 - 5;
			list.Columns.Add(h);
			h = new ColumnHeader();
			h.Text = "ID";
			h.Width = list.Width / 3 - 10;
			list.Columns.Add(h);
			list.MultiSelect = false;
			list.FullRowSelect = true;
			list.HideSelection = false;
			list.Height = 100;
			Corrections.MakeProportional(list);
			list.KeyUp += new KeyEventHandler(ListKeyUp);
			list.SelectedIndexChanged += new EventHandler(ListSelectionChanged);
		}
		
		ListView list;
		
		public override bool CanBeBigger {
			get { return true; }
		}
		
		/// <summary>
		/// Creates the controls needed for editing the <see cref="ISetting"/> and
		/// adds them to the specified parent control.
		/// </summary>
		public override void CreateControls(Rectangle bounds, EditorProperties properties)
		{
			CreateItems();
			base.CreateControls(bounds, properties);
			if (list.SelectedItems.Count > 0)
				list.SelectedItems[0].EnsureVisible();
		}
		
		private void CreateItems()
		{
			Process[] processes = Process.GetProcesses();
			ProcessInfo[] info = new ProcessInfo[processes.Length];
			string[] names = new string[processes.Length];
			for (int i = 0; i < processes.Length; i++) {
				info[i] = new ProcessInfo(processes[i]);
				ProcessCache.Instance.SetProcessInfo(info[i]);
				names[i] = info[i].Name;
			}
			Array.Sort(names, info);
			for (int i = 0; i < info.Length; i++) {
				int image = ProcessCache.Instance.GetIcon(info[i].Filename);
				ListViewItem item = new ListViewItem(new string[] {
				                                     	names[i],
				                                     	info[i].Id.ToString()
				                                     }, image);
				item.Tag = info[i];
				list.Items.Add(item);
			}
		}
		
		/// <summary>
		/// Loads the edit value into the editor.
		/// </summary>
		protected override void LoadEditValue()
		{
			if (saving)
				return;
			try {
				saving = true;
				ProcessInfo val = (ProcessInfo)this.Setting.Value;
					foreach (ListViewItem item in list.Items) {
						item.Selected = false;
					}
				if (val == null)
					return;
				foreach (ListViewItem item in list.Items) {
					ProcessInfo tag = (ProcessInfo)item.Tag;
					if (tag.Id == val.Id) {
						item.Selected = true;
						return;
					}
				}
				foreach (ListViewItem item in list.Items) {
					ProcessInfo tag = (ProcessInfo)item.Tag;
					if (ProcessCache.NormalizePath(tag.Filename) == ProcessCache.NormalizePath(val.Filename)) {
						item.Selected = true;
						item.EnsureVisible();
						return;
					}
				}
			} finally {
				saving = false;
			}
		}
		
		private void ListKeyUp(object sender, KeyEventArgs e)
		{
			if (e.KeyData == Keys.F5) {
				list.Items.Clear();
				CreateItems();
			}
		}
		
		bool saving = false;
		
		private void ListSelectionChanged(object sender, EventArgs e)
		{
			try {
				if (saving)
					return;
				saving = true;
				if (list.SelectedItems.Count == 0) {
					this.Setting.InEditValue = null;
				} else {
					this.Setting.InEditValue = list.SelectedItems[0].Tag;
				}
			} finally {
				saving = false;
			}
		}
		
		/// <summary>
		/// Gets the <see cref="LabelMode"/> this setting editor uses.<br/>
		/// If not overridden, this property always returns
		/// <see cref="LabelMode">LabelMode.NoLabel</see>.
		/// </summary>
		public override LabelMode LabelMode {
			get { return LabelMode.NoLabel; }
		}
	}
	
	public class SnifferEditorFactory : EditorFactory
	{
		public override ISettingEditor GetEditor(ISetting setting, Type type) {
			if (type == typeof(ProcessInfo))
				return new ApplicationEditor(setting);
			return base.GetEditor(setting, type);
		}
	}
}
