// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Net;
using System.Text;
using System.Windows.Forms;
using Grunwald.Gui.Controls;

namespace TCPSniffer
{
	public class Connection
	{
		#region Basic connection view
		public readonly IPAddress SourceIP;
		public readonly IPAddress DestinationIP;
		public readonly int SourcePort;
		public readonly int DestinationPort;
		
		public readonly ListViewItem Item;
		public ConnectionAnalyzer Analyzer;
		
		Status status;
		
		public Status Status {
			get {
				return status;
			}
			set {
				if (status == value) return;
				status = value;
				this.Item.SubItems[3].Text = status.ToString();
			}
		}
		
		public Connection(TcpPacket p, ListViewItem item)
		{
			this.DestinationIP = p.DestinationAddress;
			this.DestinationPort = p.DestinationPort;
			this.SourceIP = p.SourceAddress;
			this.SourcePort = p.SourcePort;
			this.Item = item;
			this.Status = Status.Unknown;
		}
		
		public bool MatchesPacket(TcpPacket p)
		{
			if (p.SourcePort == this.SourcePort &&
			    p.DestinationPort == this.DestinationPort &&
			    this.SourceIP.Equals(p.SourceAddress) &&
			    this.DestinationIP.Equals(p.DestinationAddress))
				return true;
			if (p.SourcePort == this.DestinationPort &&
			    p.DestinationPort == this.SourcePort &&
			    this.DestinationIP.Equals(p.SourceAddress) &&
			    this.SourceIP.Equals(p.DestinationAddress))
				return true;
			return false;
		}
		
		public TcpPacket LastPacket;
		
		ArrayList packets = new ArrayList();
		public ArrayList Packets { get { return packets; } }
		
		public bool IsOutgoing(TcpPacket p)
		{
			return (this.SourcePort == p.SourcePort &&
			        this.SourceIP.Equals(p.SourceAddress));
		}
		
		public void Receive(TcpPacket p)
		{
			this.LastPacket = p;
			if (status == Status.Inactive)
				this.Status = Status.Active;
			bool outgoing = IsOutgoing(p);
			// Parse data...
			int length = p.Raw.Length - p.TotalHeaderLength;
			if (length > 0) {
				packets.Add(p);
				this.Item.SubItems[4].Text = packets.Count.ToString();
				if (status < Status.Closed)
					this.Status = Status.Active;
			}
			byte flag = p.Flag;
			bool FIN = (flag & 0x01) == 0x01;
			bool SYN = (flag & 0x02) == 0x02;
			bool RST = (flag & 0x04) == 0x04;
			bool PSH = (flag & 0x08) == 0x08;
			bool ACK = (flag & 0x10) == 0x10;
			bool URG = (flag & 0x20) == 0x20;
			if (FIN)
				this.Status = Status.Closed;
			else if (RST)
				this.Status = Status.Cancelled;
			else if (SYN)
				this.Status = Status.Connecting;
		}
		
		public void Append(TcpPacket p, ExtendedRichTextBox box, bool freshBox)
		{
			bool outgoing = (this.SourcePort == p.SourcePort &&
			                 this.SourceIP.Equals(p.SourceAddress));
			string txt = ToString(p.Raw, p.TotalHeaderLength);
			try {
				if (!freshBox) {
					box.BeginUpdate();
					box.Select(box.TextLength, 0);
				}
				box.ReadOnly = false;
				box.SelectionColor = outgoing ? Color.Red : Color.Blue;
				box.AppendText(txt);
				box.ReadOnly = true;
			} finally {
				if (!freshBox)
					box.EndUpdate();
			}
		}
		
		static Encoding enc = Encoding.GetEncoding("ISO8859-1");
		public static bool ShowCarriageReturn = false;
		
		public string ToString(byte[] data, int start)
		{
			string text = enc.GetString(data, start, data.Length - start);
			StringBuilder b = new StringBuilder();
			bool schwarzenegger = true; // the terminator
			int lastTerminator = 0;
			for (int i = 0; i < text.Length; i++) {
				char c = text[i];
				schwarzenegger = false;
				if (c == '^') {
					b.Append("^^");
				} else if (c == '\n') {
					b.Append('\n');
					schwarzenegger = true;
					lastTerminator = i;
				} else if (c == '\r' && !ShowCarriageReturn) {
					// ignore ^M
				} else if (c < ' ' && c != '\t') {
					b.Append('^');
					b.Append((char)(c + 64));
				} else if (c == 128) {
					b.Append('€');
				} else {
					b.Append(c);
				}
				if (i - lastTerminator > 255) {
					lastTerminator = i;
					b.Append('\n');
				}
			}
			if (!schwarzenegger) {
				b.Append('\n');
			}
			return b.ToString();
		}
		#endregion
	}
}
