// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

#if WithoutDGN
#else
using System;
using System.Collections;
using System.ComponentModel;
using System.Text;
using System.Windows.Forms;
using Grunwald.Gui.Controls;
using Grunwald.P2PLib;
using Grunwald.P2PLib.DGNetwork;

namespace TCPSniffer
{
	public class DGNetworkAnalyzer : PacketBasedAnalyzer
	{
		#region Members
		// info list
		ListView infoList;
		ColumnHeader nameHeader = new ColumnHeader();
		ColumnHeader valueHeader = new ColumnHeader();
		
		HexViewer hex;
		Splitter splitter, splitter2;
		#endregion
		
		#region Constructor
		public DGNetworkAnalyzer()
		{
		}
		
		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);
			
			splitter = new Splitter();
			splitter.Dock = DockStyle.Bottom;
			this.Controls.Add(splitter);
			
			hex = new HexViewer();
			infoList = new ListView();
			
			ColumnHeader h;
			h = new ColumnHeader();
			h.Text = "Size";
			this.List.Columns.Add(h);
			
			h = new ColumnHeader();
			h.Text = "Type";
			h.Width *= 2;
			this.List.Columns.Add(h);
			
			h = new ColumnHeader();
			h.Text = "From";
			this.List.Columns.Add(h);
			
			Panel panel = new Panel();
			panel.Dock = splitter.Dock;
			
			// hex viewer
			hex.Visible = false;
			hex.Dock = DockStyle.Fill;
			panel.Controls.Add(hex);
			
			splitter2 = new Splitter();
			splitter2.Dock = DockStyle.Top;
			panel.Controls.Add(splitter2);
			
			// list
			nameHeader.Text = "Name";
			nameHeader.Width = 120;
			valueHeader.Text = "Value";
			infoList.View = View.Details;
			infoList.FullRowSelect = true;
			infoList.HideSelection = false;
			infoList.GridLines = true;
			infoList.Dock = DockStyle.Top;
			infoList.Height = 40;
			infoList.HeaderStyle = ColumnHeaderStyle.None;
			infoList.Columns.Add(nameHeader);
			infoList.Columns.Add(valueHeader);
			infoList.SelectedIndexChanged += new EventHandler(InfoListSelectionChanged);
			panel.Controls.Add(infoList);
			
			this.Controls.Add(panel);
			
			infoList.Height = this.ClientSize.Height / 2;
		}
		
		
		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged(e);
			int width = this.ClientSize.Width;
			if (width < 140) return;
			valueHeader.Width = width - 132;
		}
		#endregion
		
		#region CanAnalyze
		const string requestText = "GET /system/network/dgnetwork/connect HTTP/1.1\r\nHost: ";
		byte[] correctRequest  = Encoding.ASCII.GetBytes(requestText);
		const string responseText = "DGN/0.1 200 OK\r\nMyIdentity: ";
		byte[] correctResponse = Encoding.ASCII.GetBytes(responseText);
		const string response2Text = "DGN/0.1 200 Client OK\r\nConnectKey: ";
		byte[] correctResponse2 = Encoding.ASCII.GetBytes(response2Text);
		
		public override bool CanAnalyze(Connection c)
		{
			ArrayList packets = c.Packets;
			if (packets.Count <= 3) // too few packets to analyze this
				return false;
			TcpPacket req = (TcpPacket)packets[0];
			TcpPacket res1 = (TcpPacket)packets[1];
			TcpPacket res2 = (TcpPacket)packets[2];
			bool outgoing = c.IsOutgoing(req);
			if (c.IsOutgoing(res1) == outgoing)
				return false;
			if (c.IsOutgoing(res2) != outgoing)
				return false;
			return Match(req, correctRequest)
				&& Match(res1, correctResponse)
				&& Match(res2, correctResponse2);
		}
		#endregion
		
		#region Analyze
		protected override AnalyzedPacket AnalyzePacket(TcpPacket p, byte[] data, ref int pos)
		{
			int lengthAva = data.Length - pos;
			if (lengthAva < 8)
				return null;
			DataBlockType type = (DataBlockType)NetworkBitConverter.ToInt32(data, pos);
			int length = NetworkBitConverter.ToInt32(data, pos + 4);
			int oldLength = length;
			if (length < 0) length = 0;
			if (length > lengthAva - 8) length = lengthAva - 8;
			AnalyzedPacket ap = null;
			string typeString, fromUser = "";
			int displayLength = lengthAva;
			ArrayList packets = this.Connection.Packets;
			if (p == packets[0]) {
				typeString = "Client request";
			} else if (p == packets[1]) {
				typeString = "Server response";
			} else if (p == packets[2]) {
				typeString = "Client response";
			} else {
				typeString = type.ToString();
				displayLength = oldLength;
				if (type == DataBlockType.SendPacket) {
					try {
						DGNPacket np = new DGNPacket(p);
						ap = np;
						PacketHeader h = new PacketHeader(data, pos + 8);
						fromUser = h.Source.ToString();
						np.Add("PacketID", h.PacketID);
						np.Add("Source", h.Source);
						np.Add("Destination", h.Destination);
						np.Add("Encryption", h.EncryptionType);
						np.Add("Hops", h.Hops);
						if (h.EncryptionType == 0) {
							Packet dgnp = new Packet(h, data, pos + 8 + h.Length, length - h.Length, new EncryptionStrategy());
							np.Add("AnswerID", dgnp.AnswerId);
							np.Add("Target", dgnp.Target);
							for (int i = 0; i < dgnp.Values.Count; ++i) {
								np.Add("Value " + i.ToString(), dgnp.Values.GetValue(i));
							}
							foreach (DictionaryEntry e in dgnp.NamedValues) {
								np.Add("'" + e.Key.ToString() + "'", (byte[])e.Value);
							}
						}
					} catch (ProtocolViolationException) {
						// ignore this error (=show hex dump)
					} catch (Exception ex) {
						Grunwald.Gui.ErrorReporter.ShowErrorReport(ex);
					}
				}
			}
			if (ap == null)
				ap = new AnalyzedPacket(p, data, pos, lengthAva);
			ap.Items = new string[] {
				null, // time
				displayLength.ToString(),
				typeString,
				fromUser
			};
			pos += 8 + length;
			return ap;
		}
		#endregion
		
		protected override void PacketSelected(AnalyzedPacket p)
		{
			base.PacketSelected(p);
			infoList.Items.Clear();
			if (p == null) {
				hex.Data = null;
				return;
			}
			byte[] raw = p.Data;
			hex.Data = raw;
			if (p is DGNPacket) {
				infoList.Items.Clear();
				if (!infoList.Visible) {
					infoList.Visible = true;
				}
				DGNPacket d = (DGNPacket)p;
				for (int i = 0; i < d.Names.Count; ++i) {
					ListViewItem item = new ListViewItem(new string[] {
					                                     	(string)d.Names[i],
					                                     	(string)d.Values[i]
					                                     });
					item.Tag = d.Tags[i];
					infoList.Items.Add(item);
				}
			} else {
				if (infoList.Visible) {
					infoList.Visible = false;
					hex.Visible = true;
				}
			}
		}
		
		private void InfoListSelectionChanged(object sender, EventArgs e)
		{
			if (infoList.SelectedItems.Count == 0)
				hex.Data = null;
			else
				hex.Data = infoList.SelectedItems[0].Tag as byte[];
		}
		
		private class DGNPacket : AnalyzedPacket
		{
			public DGNPacket(TcpPacket p) : base(p, null) { }
			
			public readonly ArrayList Names = new ArrayList();
			public readonly ArrayList Values = new ArrayList();
			public readonly ArrayList Tags = new ArrayList();
			
			public void Add(string name, string value)
			{
				Add(name, value, null);
			}
			
			public void Add(string name, string value, object tag)
			{
				Names.Add(name);
				Values.Add(value);
				Tags.Add(tag);
			}
			
			public void Add(string name, int value)
			{
				Add(name, value.ToString() + " = 0x" + value.ToString("x"));
			}
			
			public void Add(string name, byte[] data)
			{
				Add(name, data.Length.ToString() + " bytes", data);
			}
			
			public void Add(string name, int[] arr)
			{
				StringBuilder b = new StringBuilder();
				foreach (int i in arr) {
					if (b.Length > 0)
						b.Append(", ");
					b.Append(i);
				}
				Add(name, b.ToString());
			}
		}
	}
}
#endif
