// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Net;

namespace TCPSniffer
{
	public delegate void ResolveCallback(object tag, IPAddress address, string name);
	
	public class DnsCache
	{
		public static readonly DnsCache Instance = new DnsCache();
		
		Hashtable cache = new Hashtable();
		
		/// <summary>
		/// Gets the name matching the specified ip address.
		/// Returns null if the name is currently not available.
		/// </summary>
		public string GetName(IPAddress address)
		{
			lock (cache) {
				return (string)cache[address];
			}
		}
		
		private delegate void ResolveCall(object tag, IPAddress address, ResolveCallback callback);
		
		public void BeginResolve(IPAddress address, ResolveCallback callback, object tag)
		{
			ResolveCall c = new ResolveCall(DoResolve);
			c.BeginInvoke(tag, address, callback, null, null);
		}
		
		ArrayList resolving = new ArrayList(); // contains the IPAddress currently being resolved
		
		private void DoResolve(object tag, IPAddress address, ResolveCallback callback)
		{
			try {
				string name;
				bool r;
				lock (resolving) {
					r = resolving.Contains(address);
					if (!r) {
						lock (cache) {
							r = cache.ContainsKey(address);
						}
						if (!r)
							resolving.Add(address);
					}
				}
				if (r) {
					while (r) {
						System.Threading.Thread.Sleep(100);
						lock (resolving) {
							r = resolving.Contains(address);
						}
					}
					lock (cache) {
						name = (string)cache[address];
					}
				} else {
					try {
						try {
							IPHostEntry entry = Dns.GetHostByAddress(address);
							name = entry.HostName;
						} catch (System.Net.Sockets.SocketException) {
							name = null;
						}
						
						if (name == null || name == "")
							name = address.ToString();
						lock (cache) {
							cache[address] = name;
						}
					} finally {
						lock (resolving) {
							resolving.Remove(address);
						}
					}
				}
				callback(tag, address, name);
			} catch (Exception ex) {
				Grunwald.Gui.ErrorReporter.ShowErrorReport(ex);
			}
		}
	}
}
