// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Net;
using System.Text;
using Grunwald.Gui.Settings;

namespace TCPSniffer
{
	public abstract class Filter : ReflectionSettingPage
	{
		protected Filter() {
			this.Title = ToString();
			this.Image = FilterIcons.GetIcon(this.Icon);
		}
		protected Filter(Filter f) : base(f) { }
		
		public static readonly Filter RootFilter = new RootFilter();
		
		[Setting("negate", "Negate result", "If this property is set, the result is negated.")]
		public bool Negate;
		
		public bool Accept(IPPacket p) {
			return Negate ^ AcceptPacket(p);
		}
		
		protected abstract bool AcceptPacket(IPPacket p);
		
		protected override void OnSettingChanged(SettingChangedEventArgs e) {
			this.Title = ToString();
		}
		
		public string NegString {
			get {
				return Negate ? "Not " : "";
			}
		}
		
		public virtual string Icon { get { return "Questionmark"; } }
	}
	
	public enum IPTestWhat {
		Any,
		Source,
		Destination
	}
	
	public class ProtocolFilter : Filter
	{
		public ProtocolFilter() { }
		public ProtocolFilter(ProtocolFilter f) : base(f) {}
		
		[Setting("protocol", "Protocol", "Select what protocol is accepted")]
		public Protocol Protocol = Protocol.Tcp;
		
		protected override bool AcceptPacket(IPPacket p) {
			return p.Protocol == this.Protocol;
		}
		
		public override string ToString() {
			return "ProtocolFilter: " + this.NegString + this.Protocol.ToString();
		}
		
		public override string Icon { get { return "Plug"; } }
	}
	
	public abstract class ContainerFilter : Filter
	{
		protected ContainerFilter(ContainerFilter f) : base(f) {}
		
		protected ContainerFilter() {
			AllowDelete = true;
			AllowMove = true;
		}
		
		public override bool CanAddPage(ISettingPage page, int index) {
			return true;
		}
		
		public override bool CanMovePage(ISettingPage page, int index) {
			return true;
		}
		
		public override string Icon { get { return "Folder"; } }
	}
	
	public class LogicalFilter : ContainerFilter
	{
		public LogicalFilter() { }
		public LogicalFilter(LogicalFilter f) : base(f) {}
		
		public enum LogicalOperator {
			And,
			Or
		}
		
		[Setting("operator", "Operator", "Sets the logical operator to use.")]
		public LogicalOperator Operator;
		
		protected override bool AcceptPacket(IPPacket p) {
			if (this.SubPages.Length == 0)
				return true;
			if (Operator == LogicalOperator.And)
				return AndFilter(p);
			else
				return OrFilter(p);
		}
		
		private bool AndFilter(IPPacket p)
		{
			foreach(Filter f in this.SubPages) {
				if (!f.Accept(p)) return false;
			}
			return true;
		}
		
		private bool OrFilter(IPPacket p)
		{
			foreach(Filter f in this.SubPages) {
				if (f.Accept(p)) return true;
			}
			return false;
		}
		
		public override string ToString() {
			StringBuilder b = new StringBuilder("Logical filter: ");
			b.Append(this.NegString);
			b.Append(Operator);
			return b.ToString();
		}
	}
	
	public class RootFilter : LogicalFilter
	{
		public RootFilter()
		{
			IPFilter filter = new IPFilter();
			this.AddPage(filter, 0);
		}
	}
}
