// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;

namespace TCPSniffer
{
	public class IPPacket
	{
		private byte[] raw;
		private DateTime time;
		private int version;
		private int headerLength;
		private IPAddress sourceAddress;
		private IPAddress destinationAddress;
		private int sourcePort;
		private int destinationPort;
		
		public IPPacket(byte[] raw) : this(raw, DateTime.Now) {}
		
		public IPPacket(byte[] raw, DateTime time)
		{
			if (raw == null)
				throw new ArgumentNullException();
			if (raw.Length < 20)
				throw new ArgumentException(); // invalid IP packet
			this.raw = raw;
			this.time = time;
			this.version = (raw[0] & 0xF0) >> 4;
			this.headerLength = (raw[0] & 0x0F) * 4; /* sizeof(int) */
			this.sourceAddress = new IPAddress(BitConverter.ToUInt32(raw, 12));
			this.destinationAddress = new IPAddress(BitConverter.ToUInt32(raw, 16));
			if (this.Protocol == Protocol.Tcp || this.Protocol == Protocol.Udp) {
				sourcePort = raw[headerLength] * 256 + raw[headerLength + 1];
				destinationPort = raw[headerLength + 2] * 256 + raw[headerLength + 3];
			} else {
				sourcePort = -1;
				destinationPort = -1;
			}
		}
		
		public byte[] Raw {
			get {
				return raw;
			}
		}
		
		public DateTime Time {
			get {
				return time;
			}
		}
		
		public int Version {
			get {
				return version;
			}
		}
		
		public int HeaderLength {
			get {
				return headerLength;
			}
		}
		
		public virtual int TotalHeaderLength {
			get {
				return headerLength;
			}
		}
		
		public Precedence Precedence {
			get {
				return (Precedence)((raw[1] & 0xE0) >> 5);
			}
		}
		
		public byte TypeOfService {
			get {
				return raw[1];
			}
		}
		
		public int TotalLength {
			get {
				return raw[2] * 256 + raw[3];
			}
		}
		
		public int Identification {
			get {
				return raw[4] * 256 + raw[5];
			}
		}
		
		public int Offset {
			get {
				return (raw[6] & 0x1F) * 256 + raw[7];
			}
		}
		
		public int TimeToLive {
			get {
				return raw[8];
			}
		}
		
		public Protocol Protocol {
			get {
				return (Protocol)raw[9];
			}
		}
		
		public int Checksum {
			get {
				return raw[10] * 256 + raw[11];
			}
		}
		
		public IPAddress SourceAddress {
			get {
				return sourceAddress;
			}
		}
		
		public IPAddress DestinationAddress {
			get {
				return destinationAddress;
			}
		}
		
		public virtual int SourcePort {
			get {
				return sourcePort;
			}
		}
		
		public virtual int DestinationPort {
			get {
				return destinationPort;
			}
		}
		
		internal static string Format(string title, int value, int length)
		{
			return title + ": " + value.ToString() + " (0x" +
				value.ToString("x" + length.ToString()) + ")";
		}
		
		internal static string Format(string title, uint value, int length)
		{
			return title + ": " + value.ToString() + " (0x" +
				value.ToString("x" + length.ToString()) + ")";
		}
		
		public virtual void CreateInfoNodes(TreeNodeCollection c, TreeNodeFactory f)
		{
			TreeNode node = f.Create("IP Header", 0, this.HeaderLength);
			c.Add(node);
			TreeNodeCollection a = node.Nodes;
			string txt;
			
			txt = Format("Version", this.Version, 1);
			a.Add(f.Create(txt, 0, 1));
			
			txt = string.Format("Header Length: {0} (0x{1:x})", this.HeaderLength, this.HeaderLength / 4);
			a.Add(f.Create(txt, 0, 1));
			
			if (this.Precedence != Precedence.Routine) {
				txt = string.Format("Precedence: {0} (0x{1:x})", this.Precedence, (int)this.Precedence);
				a.Add(f.Create(txt, 1, 1));
			}
			
			if ((raw[1] & 0x10) == 0x10)
				a.Add(f.Create("Low Delay", 1, 1));
			if ((raw[1] & 0x08) == 0x08)
				a.Add(f.Create("High Throughput", 1, 1));
			if ((raw[1] & 0x04) == 0x04)
				a.Add(f.Create("High Reliability", 1, 1));
			if ((raw[1] & 0x02) == 0x02)
				a.Add(f.Create("Minimize Cost", 1, 1));
			if ((raw[1] & 0x01) == 0x01)
				a.Add(f.Create("Last TOS bit set!", 1, 1));
			a.Add(f.Create(Format("Total Length", this.TotalLength, 4), 2, 2));
			a.Add(f.Create(Format("Identification", this.Identification, 4), 4, 2));
			
			if ((raw[6] & 0x80) == 0x80)
				a.Add(f.Create("First fragment bit set!", 6, 1));
			if ((raw[6] & 0x40) == 0x40)
				a.Add(f.Create("Don't fragment", 6, 1));
			if ((raw[6] & 0x20) == 0x20)
				a.Add(f.Create("More fragments", 6, 1));
			if (this.Offset != 0)
				a.Add(f.Create(Format("Offset", this.Offset, 4), 6, 2));
			
			a.Add(f.Create(Format("Time To Live", this.TimeToLive, 2), 8, 1));
			
			txt = string.Format("Protocol: {0} (0x{1:x})", this.Protocol, (int)this.Protocol);
			a.Add(f.Create(txt, 9, 1));
			
			a.Add(f.Create(Format("Checksum", this.Checksum, 4), 10, 2));
			a.Add(f.Create("Source IP: " + SourceAddress.ToString(), 12, 4));
			a.Add(f.Create("Dest IP: " + DestinationAddress.ToString(), 16, 4));
		}
		
		public virtual string TypeText {
			get {
				return this.Protocol.ToString() + " packet";
			}
		}
		
		public virtual int ImageIndex {
			get {
				return -1;
			}
		}
	}
}
