// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Net;
using System.Text;
using Grunwald.Gui.Settings;

namespace TCPSniffer
{
	public class IPFilter : Filter
	{
		public IPFilter() { }
		public IPFilter(IPFilter f) : base(f) {}
		
		public override string Icon { get { return "IP"; } }
		
		[Setting("testwhat", "What to test", "Select what ip address must match the filter")]
		public IPTestWhat TestWhat;
		
		[Setting("ip", "IP Address", "Enter the IP Mask (e.g. 192.168.0.*)."),
		 Validator(typeof(IPMaskValidator))]
		public string IPMask = "*.*.*.*";
		
		[Setting("port", "Port", "Enter the port range (e.g. 20-100)."),
		 Validator(typeof(RangeValidator))]
		public string PortMask = "*";
		
		protected override bool AcceptPacket(IPPacket p) {
			if (TestWhat != IPTestWhat.Destination) {
				if (TestIP(p.SourceAddress) && TestPort(p.SourcePort))
					return true;
			}
			if (TestWhat != IPTestWhat.Source) {
				if (TestIP(p.DestinationAddress) && TestPort(p.DestinationPort))
					return true;
			}
			return false;
		}
		
		private bool TestIP(IPAddress ip) {
			string[] blocks = IPMask.Split('.');
			byte[] ipAddress = ip.GetAddressBytes();
			for (int i = 0; i < blocks.Length; i++) {
				blocks[i] = blocks[i].Trim();
				if (blocks[i] == "*")
					continue;
				int num = int.Parse(blocks[i]);
				if (num != ipAddress[i])
					return false;
			}
			return true;
		}
		
		private bool TestPort(int port) {
			string s = PortMask.Trim();
			if (s == "*")
				return true;
			int pos = s.IndexOf('-');
			if (pos > 0) {
				int v1 = int.Parse(s.Substring(0, pos));
				int v2 = int.Parse(s.Substring(pos + 1));
				return v1 <= port && port <= v2;
			} else {
				return port == int.Parse(s);
			}
		}
		
		public override string ToString() {
			return String.Format("IP/Port Filter: {0} {1} {2} : {3}",
			                     TestWhat, Negate?"!=":"=", IPMask, PortMask);
		}
	}
	
	public class IPMaskValidator : ISettingValidator
	{
		public void CheckValue(ISetting setting, object value) {
			string s = (string)value;
			if (s == null)
				throw new SettingEditException("String is null");
			string[] blocks = s.Split('.');
			if (blocks.Length > 4)
				throw new SettingEditException("Invalid IP-Mask: too many blocks.");
			bool wasStar = false;
			for (int i = 0; i < blocks.Length; i++) {
				blocks[i] = blocks[i].Trim();
				if (blocks[i] == "")
					throw new SettingEditException("Invalid IP-Mask: empty block");
				wasStar = blocks[i] == "*";
				if (wasStar)
					continue;
				try {
					int num = int.Parse(blocks[i]);
					if (num < 0 || num > 255)
						throw new SettingEditException("Invalid IP-Mask: number is out of range");
				} catch (FormatException) {
					throw new SettingEditException("Invalid IP-Mask: block does not contain number");
				}
			}
			if (!wasStar && blocks.Length < 4)
				throw new SettingEditException("Invalid IP-Mask: if you use less than 4 blocks, the last block must be *!");
		}
	}
	
	public class RangeValidator : ISettingValidator
	{
		public void CheckValue(ISetting setting, object value) {
			string s = (string)value;
			if (s == null)
				throw new SettingEditException("String is null");
			s = s.Trim();
			if (s == "")
				throw new SettingEditException("Invalid port range");
			if (s == "*")
				return;
			int pos = s.IndexOf('-');
			if (pos == 0)
				throw new SettingEditException("Invalid port range");
			if (pos > 0) {
				try {
					int v1 = int.Parse(s.Substring(0, pos));
					int v2 = int.Parse(s.Substring(pos + 1));
					CheckPort(v1);
					CheckPort(v2);
					if (v2 < v1)
						throw new SettingEditException("Invalid port range");
				} catch (FormatException ex) {
					throw new SettingEditException("Invalid port range", ex);
				}
			} else {
				try {
					CheckPort(int.Parse(s));
				} catch (FormatException ex) {
					throw new SettingEditException("Invalid port", ex);
				}
			}
		}
		
		void CheckPort(int portnum) {
			if (portnum < 0 || portnum > ushort.MaxValue)
				throw new SettingEditException("Invalid port");
		}
	}
}
