// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Net;
using System.Reflection;
using System.Windows.Forms;
using System.Xml;
using Grunwald.Gui;

namespace TCPSniffer
{
	public class MainForm : System.Windows.Forms.Form
	{
		private System.ComponentModel.IContainer components;
		private System.Windows.Forms.MenuItem menuSep2;
		private System.Windows.Forms.Label startInfoLabel;
		private System.Windows.Forms.ToolBarButton clearButton;
		private System.Windows.Forms.ToolBar toolBar;
		private TCPSniffer.TcpTable tcpTable;
		private System.Windows.Forms.ToolBarButton stopButton;
		private System.Windows.Forms.MenuItem menuMonitor;
		private System.Windows.Forms.ToolBarButton loadButton;
		private System.Windows.Forms.MenuItem menuSave;
		private System.Windows.Forms.TabPage tabIPTable;
		private System.Windows.Forms.MenuItem menuExit;
		private System.Windows.Forms.MenuItem menuOpen;
		private TCPSniffer.FilterEditor filterEditor;
		private System.Windows.Forms.MenuItem menuClear;
		private System.Windows.Forms.TabPage tabFilter;
		private System.Windows.Forms.MenuItem menuStop;
		private System.Windows.Forms.ToolBarButton aboutButton;
		private System.Windows.Forms.ContextMenu hostsMenu;
		private System.Windows.Forms.MenuItem menuHelp;
		private System.Windows.Forms.TabPage tabConnections;
		private System.Windows.Forms.MainMenu mainMenu;
		private System.Windows.Forms.ToolBarButton saveButton;
		private System.Windows.Forms.MenuItem menuFile;
		private System.Windows.Forms.TabControl tabControl;
		private System.Windows.Forms.StatusBar statusBar;
		private System.Windows.Forms.ToolBarButton connectButton;
		private System.Windows.Forms.MenuItem menuInfo;
		private System.Windows.Forms.TabPage tabPackets;
		private System.Windows.Forms.ImageList imageList;
		private System.Windows.Forms.MenuItem menuSep1;
		
		
		PacketMonitor[] packetMonitors;
		IPacketListener[] listeners;
		ConnectionView conView;
		PacketList packetList;
		
		public MainForm()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			
			tabControl.Dock = DockStyle.Fill;
			
			packetList = new PacketList();
			packetList.Dock = DockStyle.Fill;
			tabPackets.Controls.Add(packetList);
			conView = new ConnectionView();
			listeners = new IPacketListener[] {packetList, conView};
			
			IPAddress[] hosts = Dns.Resolve(Dns.GetHostName()).AddressList;
			if (hosts.Length == 0)
				MessageBox.Show("This computer does not have non-loopback interfaces installed!");
			for(int i = 0; i < hosts.Length; i++) {
				menuMonitor.MenuItems.Add(hosts[i].ToString(), new EventHandler(this.OnHostsClick));
				hostsMenu.MenuItems.Add(hosts[i].ToString(), new EventHandler(this.OnHostsClick));
			}
			packetMonitors = new PacketMonitor[hostsMenu.MenuItems.Count];
			for(int i = 0; i <  packetMonitors.Length; i++) {
				packetMonitors[i] = new PacketMonitor(hosts[i]);
				packetMonitors[i].NewPacket += new NewPacketEventHandler(this.OnNewPacket);
			}
			connectButton.ImageIndex = LoadImage("ConnectedGlobe");
			connectButton.DropDownMenu = hostsMenu;
			stopButton.ImageIndex = LoadImage("OffButton");
			clearButton.ImageIndex = LoadImage("XPEmptyTrash");
			aboutButton.ImageIndex = LoadImage("Questionmark");
			loadButton.ImageIndex = LoadImage("Open");
			saveButton.ImageIndex = LoadImage("Save");
			Stream s = Assembly.GetExecutingAssembly().GetManifestResourceStream("App.ico");
			this.Icon = new Icon(s, 16, 16);
			s.Close();
			
			startInfoLabel.Text = "Please press the 'start monitor' button in the toolbar " +
				"to start capturing!";
			
			//Grunwald.Gui.Components.Corrections.CorrectBackground(tabConnections);
		}
		
		private int LoadImage(string name)
		{
			using (Stream s = Assembly.GetExecutingAssembly().GetManifestResourceStream(name + ".ico")) {
				using (Icon ico = new Icon(s, 16, 16)) {
					imageList.Images.Add(ico);
				}
			}
			return imageList.Images.Count - 1;
		}
		
		public void OnHostsClick(object sender, EventArgs e)
		{
			CheckFirewall();
			// start or stop listening on the specified interface
			int index = ((MenuItem)sender).Index;
			hostsMenu.MenuItems[index].Checked = !hostsMenu.MenuItems[index].Checked;
			menuMonitor.MenuItems[index].Checked = hostsMenu.MenuItems[index].Checked;
			if (hostsMenu.MenuItems[index].Checked) {
				packetMonitors[index].Start();
				statusBar.Text = "Monitoring " + packetMonitors[index].IP.ToString();
			} else {
				packetMonitors[index].Stop();
				statusBar.Text = "Stopped monitoring " + packetMonitors[index].IP.ToString();
			}
		}
		
		
		bool checkedFirewall = false;
		private void CheckFirewall()
		{
			if (checkedFirewall) return;
			checkedFirewall = true;
			tabConnections.Controls.Remove(startInfoLabel);
			startInfoLabel.Dispose();
			startInfoLabel = null;
			
			if (Process.GetProcessesByName("kpf4ss").Length > 0) {
				MessageBox.Show("You have Kerio Firewall installed.\n" +
				                "Please disable the firewall while you are capturing " +
				                "packets.\nIf you leave Kerio running, you will see " +
				                "all incoming packets, but no outgoing TCP packets.",
				                "Kerio Firewall");
			}
			
			conView.Dock = DockStyle.Fill;
			tabConnections.Controls.Add(conView);
		}
		
		#region Main
		public static bool Is2000;
		
		[STAThread]
		public static void Main(string[] args)
		{
			try {
				Application.ThreadException += new System.Threading.ThreadExceptionEventHandler(ThreadException);
				AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(UnhandledException);
				try {
					if (Environment.OSVersion.Platform != PlatformID.Win32NT ||
					    Environment.OSVersion.Version.Major < 5) {
						MessageBox.Show("This program only runs on Windoes 2000 or higher.");
						return;
					}
					Is2000 = Environment.OSVersion.Version.Minor == 0 && Environment.OSVersion.Version.Major == 5;
					Application.Run(new MainForm());
				} catch (Exception ex) {
					ErrorReporter.ShowErrorReport(ex);
				}
			} catch (Exception ex) {
				MessageBox.Show(ex.ToString());
			}
		}
		
		static void UnhandledException(object sender, UnhandledExceptionEventArgs e)
		{
			ErrorReporter.ShowErrorReport((Exception)e.ExceptionObject);
		}
		
		static void ThreadException(object sender, System.Threading.ThreadExceptionEventArgs e)
		{
			ErrorReporter.ShowErrorReport(e.Exception);
		}
		#endregion
		
		#region Windows Forms Designer generated code
		/// <summary>
		/// This method is required for Windows Forms designer support.
		/// Do not change the method contents inside the source code editor. The Forms designer might
		/// not be able to load this method if it was changed manually.
		/// </summary>
		private void InitializeComponent() {
			this.components = new System.ComponentModel.Container();
			this.menuSep1 = new System.Windows.Forms.MenuItem();
			this.imageList = new System.Windows.Forms.ImageList(this.components);
			this.tabPackets = new System.Windows.Forms.TabPage();
			this.menuInfo = new System.Windows.Forms.MenuItem();
			this.connectButton = new System.Windows.Forms.ToolBarButton();
			this.statusBar = new System.Windows.Forms.StatusBar();
			this.tabControl = new System.Windows.Forms.TabControl();
			this.menuFile = new System.Windows.Forms.MenuItem();
			this.saveButton = new System.Windows.Forms.ToolBarButton();
			this.mainMenu = new System.Windows.Forms.MainMenu();
			this.tabConnections = new System.Windows.Forms.TabPage();
			this.menuHelp = new System.Windows.Forms.MenuItem();
			this.hostsMenu = new System.Windows.Forms.ContextMenu();
			this.aboutButton = new System.Windows.Forms.ToolBarButton();
			this.menuStop = new System.Windows.Forms.MenuItem();
			this.tabFilter = new System.Windows.Forms.TabPage();
			this.menuClear = new System.Windows.Forms.MenuItem();
			this.filterEditor = new TCPSniffer.FilterEditor();
			this.menuOpen = new System.Windows.Forms.MenuItem();
			this.menuExit = new System.Windows.Forms.MenuItem();
			this.tabIPTable = new System.Windows.Forms.TabPage();
			this.menuSave = new System.Windows.Forms.MenuItem();
			this.loadButton = new System.Windows.Forms.ToolBarButton();
			this.menuMonitor = new System.Windows.Forms.MenuItem();
			this.stopButton = new System.Windows.Forms.ToolBarButton();
			this.tcpTable = new TCPSniffer.TcpTable();
			this.toolBar = new System.Windows.Forms.ToolBar();
			this.clearButton = new System.Windows.Forms.ToolBarButton();
			this.startInfoLabel = new System.Windows.Forms.Label();
			this.menuSep2 = new System.Windows.Forms.MenuItem();
			this.tabControl.SuspendLayout();
			this.tabConnections.SuspendLayout();
			this.tabFilter.SuspendLayout();
			this.tabIPTable.SuspendLayout();
			this.SuspendLayout();
			// 
			// menuSep1
			// 
			this.menuSep1.Index = 2;
			this.menuSep1.Text = "-";
			// 
			// imageList
			// 
			this.imageList.ImageSize = new System.Drawing.Size(16, 16);
			this.imageList.TransparentColor = System.Drawing.Color.Magenta;
			// 
			// tabPackets
			// 
			this.tabPackets.Location = new System.Drawing.Point(4, 22);
			this.tabPackets.Name = "tabPackets";
			this.tabPackets.Size = new System.Drawing.Size(664, 310);
			this.tabPackets.TabIndex = 1;
			this.tabPackets.Text = "IP Packets";
			this.tabPackets.Visible = false;
			// 
			// menuInfo
			// 
			this.menuInfo.Index = 0;
			this.menuInfo.Text = "About";
			this.menuInfo.Click += new System.EventHandler(this.MenuInfoClick);
			// 
			// connectButton
			// 
			this.connectButton.Style = System.Windows.Forms.ToolBarButtonStyle.DropDownButton;
			this.connectButton.ToolTipText = "Start monitoring";
			// 
			// statusBar
			// 
			this.statusBar.Location = new System.Drawing.Point(0, 368);
			this.statusBar.Name = "statusBar";
			this.statusBar.Size = new System.Drawing.Size(672, 22);
			this.statusBar.TabIndex = 3;
			this.statusBar.Text = "TCP Sniffer ready";
			// 
			// tabControl
			// 
			this.tabControl.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
			                                                                | System.Windows.Forms.AnchorStyles.Left)
			                                                               | System.Windows.Forms.AnchorStyles.Right)));
			this.tabControl.Controls.Add(this.tabConnections);
			this.tabControl.Controls.Add(this.tabPackets);
			this.tabControl.Controls.Add(this.tabIPTable);
			this.tabControl.Controls.Add(this.tabFilter);
			this.tabControl.Location = new System.Drawing.Point(0, 31);
			this.tabControl.Name = "tabControl";
			this.tabControl.SelectedIndex = 0;
			this.tabControl.Size = new System.Drawing.Size(672, 336);
			this.tabControl.TabIndex = 4;
			// 
			// menuFile
			// 
			this.menuFile.Index = 0;
			this.menuFile.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
			                                 	this.menuMonitor,
			                                 	this.menuStop,
			                                 	this.menuSep1,
			                                 	this.menuClear,
			                                 	this.menuOpen,
			                                 	this.menuSave,
			                                 	this.menuSep2,
			                                 	this.menuExit});
			this.menuFile.Text = "File";
			// 
			// saveButton
			// 
			this.saveButton.ToolTipText = "Saves the current packets and filters";
			// 
			// mainMenu
			// 
			this.mainMenu.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
			                                 	this.menuFile,
			                                 	this.menuHelp});
			// 
			// tabConnections
			// 
			this.tabConnections.Controls.Add(this.startInfoLabel);
			this.tabConnections.Location = new System.Drawing.Point(4, 22);
			this.tabConnections.Name = "tabConnections";
			this.tabConnections.Size = new System.Drawing.Size(664, 310);
			this.tabConnections.TabIndex = 0;
			this.tabConnections.Text = "Connections";
			// 
			// menuHelp
			// 
			this.menuHelp.Index = 1;
			this.menuHelp.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
			                                 	this.menuInfo});
			this.menuHelp.Text = "Help";
			// 
			// aboutButton
			// 
			this.aboutButton.ToolTipText = "Shows the about dialog";
			// 
			// menuStop
			// 
			this.menuStop.Index = 1;
			this.menuStop.Text = "&Stop monitoring";
			this.menuStop.Click += new System.EventHandler(this.MenuStopClick);
			// 
			// tabFilter
			// 
			this.tabFilter.Controls.Add(this.filterEditor);
			this.tabFilter.Location = new System.Drawing.Point(4, 22);
			this.tabFilter.Name = "tabFilter";
			this.tabFilter.Size = new System.Drawing.Size(664, 310);
			this.tabFilter.TabIndex = 3;
			this.tabFilter.Text = "Filter";
			// 
			// menuClear
			// 
			this.menuClear.Index = 3;
			this.menuClear.Text = "&Clear packet list";
			this.menuClear.Click += new System.EventHandler(this.MenuClearClick);
			// 
			// filterEditor
			// 
			this.filterEditor.Dock = System.Windows.Forms.DockStyle.Fill;
			this.filterEditor.Location = new System.Drawing.Point(0, 0);
			this.filterEditor.Name = "filterEditor";
			this.filterEditor.Size = new System.Drawing.Size(664, 310);
			this.filterEditor.TabIndex = 0;
			// 
			// menuOpen
			// 
			this.menuOpen.Index = 4;
			this.menuOpen.Text = "&Open";
			this.menuOpen.Click += new System.EventHandler(this.MenuOpenClick);
			// 
			// menuExit
			// 
			this.menuExit.Index = 7;
			this.menuExit.Text = "E&xit";
			this.menuExit.Click += new System.EventHandler(this.MenuExitClick);
			// 
			// tabIPTable
			// 
			this.tabIPTable.Controls.Add(this.tcpTable);
			this.tabIPTable.Location = new System.Drawing.Point(4, 22);
			this.tabIPTable.Name = "tabIPTable";
			this.tabIPTable.Size = new System.Drawing.Size(664, 310);
			this.tabIPTable.TabIndex = 2;
			this.tabIPTable.Text = "TCP Table";
			// 
			// menuSave
			// 
			this.menuSave.Index = 5;
			this.menuSave.Text = "&Save";
			this.menuSave.Click += new System.EventHandler(this.MenuSaveClick);
			// 
			// loadButton
			// 
			this.loadButton.ToolTipText = "Loads saved packets and filters";
			// 
			// menuMonitor
			// 
			this.menuMonitor.Index = 0;
			this.menuMonitor.Text = "&Monitor";
			// 
			// stopButton
			// 
			this.stopButton.ToolTipText = "Stop monitoring";
			// 
			// tcpTable
			// 
			this.tcpTable.Dock = System.Windows.Forms.DockStyle.Fill;
			this.tcpTable.Location = new System.Drawing.Point(0, 0);
			this.tcpTable.Name = "tcpTable";
			this.tcpTable.Size = new System.Drawing.Size(664, 310);
			this.tcpTable.TabIndex = 1;
			// 
			// toolBar
			// 
			this.toolBar.Appearance = System.Windows.Forms.ToolBarAppearance.Flat;
			this.toolBar.Buttons.AddRange(new System.Windows.Forms.ToolBarButton[] {
			                              	this.connectButton,
			                              	this.stopButton,
			                              	this.clearButton,
			                              	this.loadButton,
			                              	this.saveButton,
			                              	this.aboutButton});
			this.toolBar.ButtonSize = new System.Drawing.Size(23, 22);
			this.toolBar.DropDownArrows = true;
			this.toolBar.ImageList = this.imageList;
			this.toolBar.Location = new System.Drawing.Point(0, 0);
			this.toolBar.Name = "toolBar";
			this.toolBar.ShowToolTips = true;
			this.toolBar.Size = new System.Drawing.Size(672, 28);
			this.toolBar.TabIndex = 1;
			this.toolBar.ButtonClick += new System.Windows.Forms.ToolBarButtonClickEventHandler(this.ToolBarButtonClick);
			// 
			// clearButton
			// 
			this.clearButton.ToolTipText = "Clears the packet list";
			// 
			// startInfoLabel
			// 
			this.startInfoLabel.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left)
			                                                                   | System.Windows.Forms.AnchorStyles.Right)));
			this.startInfoLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.startInfoLabel.Location = new System.Drawing.Point(32, 24);
			this.startInfoLabel.Name = "startInfoLabel";
			this.startInfoLabel.Size = new System.Drawing.Size(592, 128);
			this.startInfoLabel.TabIndex = 0;
			this.startInfoLabel.Text = "startInfoLabel";
			// 
			// menuSep2
			// 
			this.menuSep2.Index = 6;
			this.menuSep2.Text = "-";
			// 
			// MainForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(672, 390);
			this.Controls.Add(this.tabControl);
			this.Controls.Add(this.statusBar);
			this.Controls.Add(this.toolBar);
			this.Menu = this.mainMenu;
			this.MinimumSize = new System.Drawing.Size(448, 312);
			this.Name = "MainForm";
			this.Text = "TCP Sniffer";
			this.Closed += new System.EventHandler(this.MainFormClosed);
			this.tabControl.ResumeLayout(false);
			this.tabConnections.ResumeLayout(false);
			this.tabFilter.ResumeLayout(false);
			this.tabIPTable.ResumeLayout(false);
			this.ResumeLayout(false);
		}
		#endregion
		
		void ToolBarButtonClick(object sender, System.Windows.Forms.ToolBarButtonClickEventArgs e)
		{
			if (e.Button == stopButton) { // stop listening on all interfaces
				menuStop.PerformClick();
			} else if (e.Button == connectButton) { // start listening on all interfaces
				CheckFirewall();
				for(int i = 0; i < packetMonitors.Length; i++) {
					try {
						packetMonitors[i].Start();
					} catch (Exception ex) {
						MessageBox.Show(this, "There was a problem starting the packet monitor for interface " + packetMonitors[i].IP.ToString() + "\r\n\r\n[" + ex.Message + "]", "TCP Sniffer", MessageBoxButtons.OK, MessageBoxIcon.Error);
					}
					hostsMenu.MenuItems[i].Checked = true;
					menuMonitor.MenuItems[i].Checked = true;
				}
				statusBar.Text = "Monitoring all interfaces";
			} else if (e.Button == clearButton) { // clear the packet list
				menuClear.PerformClick();
			} else if (e.Button == aboutButton) {
				menuInfo.PerformClick();
			} else if (e.Button == loadButton) {
				menuOpen.PerformClick();
			} else if (e.Button == saveButton) {
				menuSave.PerformClick();
			} else {
				MessageBox.Show("Unknown button clicked.");
			}
		}
		
		void MenuStopClick(object sender, System.EventArgs e)
		{
			for(int i = 0; i < packetMonitors.Length; i++) {
				packetMonitors[i].Stop();
				hostsMenu.MenuItems[i].Checked = false;
				menuMonitor.MenuItems[i].Checked = false;
			}
			statusBar.Text = "Stopped monitoring";
		}
		
		void MenuInfoClick(object sender, System.EventArgs e)
		{
			InfoForm af = new InfoForm();
			af.ShowDialog(this);
		}
		
		void MenuExitClick(object sender, System.EventArgs e)
		{
			Close();
		}
		
		void MenuClearClick(object sender, System.EventArgs e)
		{
			foreach (IPacketListener l in listeners) {
				l.Clear();
			}
			count = 0;
			size = 0;
			GC.Collect();
			statusBar.Text = "Cleared packet list";
		}
		
		void MainFormClosed(object sender, System.EventArgs e)
		{
			for(int i = 0; i < packetMonitors.Length; i++) {
				packetMonitors[i].Stop();
			}
		}
		
		int size = 0;
		int count = 0;
		
		bool filterActive = true;
		
		private void OnNewPacket(PacketMonitor pm, IPPacket p)
		{
			if (InvokeRequired) {
				Invoke(new NewPacketEventHandler(OnNewPacket), new object[] {pm, p});
			} else {
				try {
					if (p.Protocol == Protocol.Udp)
						p = new UdpPacket(p);
					else if (p.Protocol == Protocol.Tcp)
						p = new TcpPacket(p);
					if (filterActive && !Filter.RootFilter.Accept(p))
						return;
					// add the new packet to the list
					count += 1;
					size += p.TotalLength;
					foreach (IPacketListener l in listeners) {
						l.Receive(p);
					}
					statusBar.Text = string.Format("Intercepted {0} packet(s) [{1} bytes]", count, size);
				} catch (Exception ex) {
					ErrorReporter.ShowErrorReport(ex);
				}
			}
		}
		
		void MenuOpenClick(object sender, System.EventArgs e)
		{
			using (OpenFileDialog dlg = new OpenFileDialog()) {
				dlg.Filter = "XML files|*.xml|All files|*.*";
				dlg.CheckFileExists = true;
				dlg.CheckPathExists = true;
				dlg.DefaultExt = ".xml";
				dlg.AddExtension = true;
				if (dlg.ShowDialog() != DialogResult.OK)
					return;
				using (Stream s = dlg.OpenFile()) {
					menuClear.PerformClick();
					filterEditor.TreeView.RootPage = null;
					XmlTextReader reader = new XmlTextReader(s);
					reader.WhitespaceHandling = WhitespaceHandling.None;
					reader.ReadStartElement("TCPSniffer");
					reader.ReadStartElement("filter");
					Filter.RootFilter.Load(reader);
					if (reader.Name != "filter")
						MessageBox.Show("Internal error in setting loader: finished but didn't point to root element.");
					reader.ReadEndElement();
					int count = 0;
					bool hasPackets = !reader.IsEmptyElement;
					reader.ReadStartElement("packets");
					if (hasPackets) {
						try {
							filterActive = false;
							while (reader.NodeType != XmlNodeType.EndElement) {
								if (reader.NodeType == XmlNodeType.Text) {
									reader.Read();
									continue;
								}
								string time = reader.GetAttribute("time");
								if (time == null)
									throw new ApplicationException("Missing 'time' attribute in <packet>.");
								reader.ReadStartElement("packet");
								byte[] data = Convert.FromBase64String(reader.Value);
								OnNewPacket(null, new IPPacket(data, Convert.ToDateTime(time)));
								reader.Read();
								reader.ReadEndElement();
								count += 1;
							}
						} finally {
							filterActive = true;
						}
						reader.ReadEndElement();
					}
					reader.ReadEndElement();
					reader.Close();
					filterEditor.TreeView.RootPage = Filter.RootFilter;
					filterEditor.ClearRightSide();
					CheckFirewall(); // activate connection view
					statusBar.Text = count.ToString() + " packets loaded from file.";
				}
			}
		}
		
		void MenuSaveClick(object sender, System.EventArgs e)
		{
			using (SaveFileDialog dlg = new SaveFileDialog()) {
				dlg.Filter = "XML files|*.xml|All files|*.*";
				dlg.OverwritePrompt = true;
				dlg.CheckPathExists = true;
				dlg.DefaultExt = ".xml";
				dlg.AddExtension = true;
				if (dlg.ShowDialog() != DialogResult.OK)
					return;
				using (Stream s = dlg.OpenFile()) {
					XmlTextWriter writer = new XmlTextWriter(s, System.Text.Encoding.UTF8);
					writer.Formatting = Formatting.Indented;
					writer.WriteStartDocument();
					writer.WriteStartElement("TCPSniffer");
					writer.WriteStartElement("filter");
					Filter.RootFilter.Save(writer);
					writer.WriteEndElement();
					writer.WriteStartElement("packets");
					int count = 0;
					foreach (IPPacket p in packetList.Packets) {
						writer.WriteStartElement("packet");
						writer.WriteAttributeString("time", XmlConvert.ToString(p.Time));
						writer.WriteString(Convert.ToBase64String(p.Raw));
						writer.WriteEndElement();
						count += 1;
					}
					writer.WriteEndElement();
					writer.WriteEndElement();
					writer.WriteEndDocument();
					writer.Close();
					statusBar.Text = count.ToString() + " packets saved to file.";
				}
			}
		}
	}
}
