// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

#if WithoutOscar
#else
using System;
using System.Collections;
using System.ComponentModel;
using System.Text;
using System.Windows.Forms;
using Grunwald.Gui.Controls;
using Grunwald.P2PLib.DGNetwork;
using Grunwald.NOscar;

namespace TCPSniffer
{
	public class OscarAnalyzer : PacketBasedAnalyzer
	{
		#region Members
		// info list
		ListView infoList;
		ColumnHeader nameHeader = new ColumnHeader();
		ColumnHeader valueHeader = new ColumnHeader();
		
		HexViewer hex;
		Splitter splitter;
		#endregion
		
		#region Constructor
		protected override void OnLoad(EventArgs e) {
			base.OnLoad(e);
			
			splitter = new Splitter();
			splitter.Dock = DockStyle.Bottom;
			this.Controls.Add(splitter);
			
			hex = new HexViewer();
			infoList = new ListView();
			
			ColumnHeader h;
			h = new ColumnHeader();
			h.Text = "Channel";
			this.List.Columns.Add(h);
			
			h = new ColumnHeader();
			h.Text = "Length";
			this.List.Columns.Add(h);
			
			h = new ColumnHeader();
			h.Text = "Flags";
			h.Width /= 2;
			this.List.Columns.Add(h);
			
			h = new ColumnHeader();
			h.Text = "RequestID";
			this.List.Columns.Add(h);
			
			Panel panel = new Panel();
			panel.Dock = splitter.Dock;
			
			// list
			nameHeader.Text = "Name";
			nameHeader.Width = 120;
			valueHeader.Text = "Value";
			infoList.View = View.Details;
			infoList.FullRowSelect = true;
			infoList.HideSelection = false;
			infoList.GridLines = true;
			infoList.Dock = DockStyle.Fill;
			infoList.HeaderStyle = ColumnHeaderStyle.None;
			infoList.Columns.Add(nameHeader);
			infoList.Columns.Add(valueHeader);
			panel.Controls.Add(infoList);
			
			// hex viewer
			hex.Visible = false;
			hex.Dock = DockStyle.Fill;
			panel.Controls.Add(hex);
			
			this.Controls.Add(panel);
			
			infoList.Height = this.ClientSize.Height / 2;
		}
		
		
		protected override void OnSizeChanged(EventArgs e) {
			base.OnSizeChanged(e);
			int width = this.ClientSize.Width;
			if (width < 140) return;
			valueHeader.Width = width - 132;
		}
		#endregion
		
		#region CanAnalyze
		public override bool CanAnalyze(Connection c)
		{
			if (c.Packets.Count == 0) // too few packets to analyze this
				return false;
			TcpPacket p = (TcpPacket)c.Packets[0];
			int pos = p.TotalHeaderLength;
			byte[] raw = p.Raw;
			if (raw.Length < pos + 10)
				return false;
			if (raw[pos    ] != 0x2a) return false;
			if (raw[pos + 1] != 0x01) return false;
			if (raw[pos + 6] != 0x00) return false;
			if (raw[pos + 7] != 0x00) return false;
			if (raw[pos + 8] != 0x00) return false;
			if (raw[pos + 9] != 0x01) return false;
			return true;
		}
		#endregion
		
		#region Analyze
		protected override AnalyzedPacket AnalyzePacket(TcpPacket p, byte[] data, ref int pos)
		{
			int lengthAva = data.Length - pos;
			if (lengthAva < 6)
				return null;
			AnalyzedPacket ap;
			if (data[pos] != 0x2a) {
				ap = new AnalyzedPacket(p, data, pos, lengthAva);
				ap.Items = new string[] { null, "invalid", lengthAva.ToString() };
				pos = data.Length;
				return ap;
			}
			byte flapChannel = data[pos + 1];
			int length = NetworkBitConverter.ToUInt16(data, pos + 4);
			if (length > lengthAva)
				return null;
			if (flapChannel == 2) {
				DataSnac snac = new DataSnac(data, pos + 6, length);
				ap = new AnalyzedPacket(p, snac.Data);
				ap.Items = new string[] {
					null,
					snac.Family.ToString("x4") + " / " + snac.SubType.ToString("x4"),
					snac.Data.Length.ToString(),
					snac.Flags.ToString("x"),
					snac.RequestID.ToString()
				};
			} else {
				ap = new AnalyzedPacket(p, data, pos, length + 6);
				ap.Items = new string[] {
					null,
					flapChannel.ToString(),
					length.ToString()
				};
			}
			pos += 6 + length;
			return ap;
		}
		#endregion
		
		protected override void PacketSelected(AnalyzedPacket p)
		{
			base.PacketSelected(p);
			infoList.Items.Clear();
			if (p == null) {
				hex.Data = null;
				return;
			}
			byte[] raw = p.Data;
			if (raw != null) {
				if (!hex.Visible) {
					infoList.Visible = false;
					hex.Height = infoList.Height;
					hex.Visible = true;
				}
				hex.Data = raw;
			} else {
				if (hex.Visible) {
					hex.Visible = false;
					infoList.Height = hex.Height;
					infoList.Visible = true;
				}
				
			}
		}
	}
}
#endif
