// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Windows.Forms;
using Grunwald.Gui.Controls;

namespace TCPSniffer
{
	public abstract class PacketBasedAnalyzer : ConnectionAnalyzer
	{
		#region static members
		static ImageList imageList;
		
		private static void Init()
		{
			if (imageList != null)
				return;
			imageList = new ImageList();
			LoadImage("ArrowUp");
			LoadImage("ArrowDown");
		}
		
		private static int LoadImage(string name)
		{
			using (Stream s = Assembly.GetExecutingAssembly().GetManifestResourceStream(name + ".ico")) {
				using (Icon ico = new Icon(s, 16, 16)) {
					imageList.Images.Add(ico);
				}
			}
			return imageList.Images.Count - 1;
		}
		#endregion
		
		#region Constructor
		//DockedControlColumn controlColumn;
		ListView list = new ListView();
		ColumnHeader timeHeader = new ColumnHeader();
		
		protected PacketBasedAnalyzer()
		{
			Init();
		}
		
		protected override void OnLoad(EventArgs e)
		{
			timeHeader.Text = "Time";
			timeHeader.Width *= 2;
			list.SmallImageList = imageList;
			list.View = View.Details;
			list.FullRowSelect = true;
			list.HideSelection = false;
			list.MultiSelect = false;
			list.HeaderStyle = ColumnHeaderStyle.Nonclickable;
			list.Dock = DockStyle.Fill;
			list.Columns.Add(timeHeader);
			list.SelectedIndexChanged += new EventHandler(SelectionChanged);
			this.Controls.Add(list);
			Grunwald.Gui.Components.Corrections.MakeProportional(list);
		}
		#endregion
		
		#region ListView
		public ListView List { get { return list; } }
		
		AnalyzedPacket selectedPacket;
		public AnalyzedPacket SelectedPacket { get { return selectedPacket; } }
		
		public void Clear()
		{
			if (list == null) return;
			SelectPacket(null);
			list.Items.Clear();
		}
		
		private void SelectionChanged(object sender, EventArgs e)
		{
			if (list.SelectedItems.Count == 0)
				SelectPacket(null);
			else
				SelectPacket((AnalyzedPacket)list.SelectedItems[0].Tag);
		}
		
		private void SelectPacket(AnalyzedPacket p)
		{
			if (p == selectedPacket) return;
			selectedPacket = p;
			PacketSelected(p);
		}
		#endregion
		
		#region CanAnalyze
		/// <summary>checks if the beginning of the raw data in p matches data</summary>
		protected bool Match(IPPacket p, byte[] data)
		{
			int start = p.TotalHeaderLength;
			byte[] raw = p.Raw;
			int packetLength = raw.Length - start;
			if (packetLength <= data.Length)
				return false;
			for (int i = 0; i < data.Length; i++) {
				if (raw[start + i] != data[i])
					return false;
			}
			return true;
		}
		#endregion
		
		#region Analyze
		byte[] outgoingStore;
		byte[] incomingStore;
		Connection connection;
		
		public Connection Connection {
			get {
				return connection;
			}
		}
		
		public override void BeginAnalyze(Connection c)
		{
			Clear();
			connection = c;
			foreach (TcpPacket p in c.Packets) {
				AnalyzePacket(p);
			}
		}
		
		public override void AnalyzePacket(TcpPacket p)
		{
			bool outgoing = connection.IsOutgoing(p);
			byte[] raw = p.Raw;
			int length = raw.Length - p.TotalHeaderLength;
			byte[] store = outgoing ? outgoingStore : incomingStore;
			int storelength = 0;
			if (store != null) {
				storelength = store.Length;
			}
			byte[] data = new byte[length + storelength];
			if (store != null) {
				store.CopyTo(data, 0);
			}
			Array.Copy(raw, p.TotalHeaderLength, data, storelength, length);
			int i = 0;
			AnalyzedPacket ap = AnalyzePacket(p, data, ref i);
			if (ap == null) {
				store = data;
			} else {
				do {
					if (ap.Items == null)
						throw new ApplicationException("analyzer must set Items");
					ap.Items[0] = p.Time.ToString();
					ListViewItem item = new ListViewItem(ap.Items, outgoing?0:1);
					item.Tag = ap;
					list.Items.Add(item);
					if (i >= data.Length)
						break;
					ap = AnalyzePacket(p, data, ref i);
				} while (ap != null);
				if (i > data.Length || i < 0)
					throw new ApplicationException("ref i was set outside the valid range!");
				if (i == data.Length) {
					store = null;
				} else {
					store = new byte[data.Length - i];
					Array.Copy(data, i, store, 0, store.Length);
				}
			}
			if (outgoing)
				outgoingStore = store;
			else
				incomingStore = store;
		}
		#endregion
		
		protected abstract AnalyzedPacket AnalyzePacket(TcpPacket p, byte[] data, ref int pos);
		
		protected virtual void PacketSelected(AnalyzedPacket p)
		{
			// do nothing
		}
	}
	
	public class AnalyzedPacket
	{
		TcpPacket basePacket;
		byte[] data;
		public TcpPacket BasePacket { get { return basePacket; } }
		public byte[] Data { get { return data; } }
		public object Tag;
		public string[] Items;
		
		public AnalyzedPacket(TcpPacket basePacket, byte[] data)
		{
			this.basePacket = basePacket;
			this.data = data;
		}
		
		public AnalyzedPacket(TcpPacket basePacket, byte[] data, int index, int length)
		{
			this.basePacket = basePacket;
			if (index == 0 && length == data.Length) {
				this.data = data;
			} else {
				this.data = new byte[length];
				Array.Copy(data, index, this.data, 0, length);
			}
		}
	}
}
