// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;
using Grunwald.Gui.Components;
using Grunwald.Gui.Controls;

namespace TCPSniffer
{
	public class PacketList : UserControl, IPacketListener
	{
		#region Member variables
		ColumnHeader TimeHeader = new ColumnHeader();
		ColumnHeader ProtocolHeader = new ColumnHeader();
		ColumnHeader SourceHeader = new ColumnHeader();
		ColumnHeader DestinationHeader = new ColumnHeader();
		ColumnHeader LengthHeader = new ColumnHeader();
		ColumnHeader TypeHeader = new ColumnHeader();
		ListView list = new ListView();
		DockedControlColumn column = new DockedControlColumn();
		
		
		DockedControlRow row = new DockedControlRow();
		TreeView InfoList = new TreeView();
		HexViewer RawData = new HexViewer();
		
		Timer timer = new Timer();
		
		protected override void Dispose(bool disposing) {
			if (disposing)
				timer.Dispose();
			base.Dispose(disposing);
		}
		#endregion
		
		#region Constructor
		public PacketList()
		{
			if (base.DesignMode)
				return;
			column.Dock = DockStyle.Fill;
			
			timer.Interval = 100;
			timer.Enabled = true;
			timer.Tick += new EventHandler(TimerTick);
			
			// 
			// TimeHeader
			// 
			this.TimeHeader.Text = "Time";
			this.TimeHeader.Width = 120;
			// 
			// ProtocolHeader
			// 
			this.ProtocolHeader.Text = "Protocol";
			// 
			// SourceHeader
			// 
			this.SourceHeader.Text = "Source";
			this.SourceHeader.Width = 130;
			// 
			// DestinationHeader
			// 
			this.DestinationHeader.Text = "Destination";
			this.DestinationHeader.Width = 130;
			// 
			// LengthHeader
			// 
			this.LengthHeader.Text = "Length";
			// 
			// TypeHeader
			// 
			this.TypeHeader.Text = "Type";
			this.TypeHeader.Width = 130;
			
			this.list.Columns.AddRange(new ColumnHeader[] {
			                           	this.TimeHeader,
			                           	this.ProtocolHeader,
			                           	this.SourceHeader,
			                           	this.DestinationHeader,
			                           	this.LengthHeader,
			                           	this.TypeHeader});
			this.list.FullRowSelect = true;
			this.list.HeaderStyle = ColumnHeaderStyle.Nonclickable;
			this.list.MultiSelect = false;
			this.list.HideSelection = false;
			this.list.View = View.Details;
			this.list.Size = new Size(640, 100);
			if (!MainForm.Is2000)
				this.list.SmallImageList = ProcessCache.Instance.ImageList;
			
			// 
			// InfoList
			// 
			this.InfoList.Dock = System.Windows.Forms.DockStyle.Fill;
			this.InfoList.FullRowSelect = true;
			this.InfoList.Name = "InfoList";
			this.InfoList.Size = new System.Drawing.Size(480, 208);
			this.InfoList.TabIndex = 0;
			
			Corrections.MakeProportional(list);
			
			column.AddSplitter = true;
			column.Controls.Add(list);
			
			row.AddSplitter = true;
			row.Controls.Add(InfoList);
			row.Controls.Add(RawData);
			column.Controls.Add(row);
			row[InfoList].MinSize = 270;
			row[InfoList].Factor = 0.1f;
			this.Controls.Add(column);
			
			column[row].MinSize = 50;
			column[list].MinSize = 50;
			list.SelectedIndexChanged += new EventHandler(ListIndexChanged);
			InfoList.AfterSelect += new TreeViewEventHandler(InfoListClick);
		}
		#endregion
		
		#region ListIndexChanged
		bool changed;
		
		void ListIndexChanged(object sender, EventArgs e) {
			changed = true;
		}
		
		void TimerTick(object sender, EventArgs e) {
			if (!changed) return;
			changed = false;
			InfoList.Nodes.Clear();
			if (list.SelectedIndices.Count != 1) {
				RawData.Data = null;
				return;
			}
			IPPacket packet = (IPPacket)packets[list.SelectedIndices[0]];
			InfoList.Nodes.Add("Time: " + packet.Time.ToString());
			if (packet.SourcePort >= 0) {
				InfoList.Nodes.Add(String.Format("Source: {0}:{1}",
				                                 packet.SourceAddress,
				                                 packet.SourcePort));
			} else {
				InfoList.Nodes.Add(String.Format("Source: {0}", packet.SourceAddress));
			}
			if (packet.DestinationPort >= 0) {
				InfoList.Nodes.Add(String.Format("Destination: {0}:{1}",
				                                 packet.DestinationAddress,
				                                 packet.DestinationPort));
			} else {
				InfoList.Nodes.Add(String.Format("Destination: {0}", packet.DestinationAddress));
			}
			packet.CreateInfoNodes(InfoList.Nodes, new TreeNodeFactory());
			InfoList.ExpandAll();
			InfoList.Nodes[0].EnsureVisible();
			RawData.Data = packet.Raw;
		}
		#endregion
		
		#region InfoListClick
		void InfoListClick(object sender, TreeViewEventArgs e)
		{
			TreeNode n = InfoList.SelectedNode;
			if (n == null) return;
			Range r = n.Tag as Range;
			if (r == null) return;
			RawData.Select(r.Start, r.Length);
		}
		#endregion
		
		#region Clear / Receive
		ArrayList packets = new ArrayList();
		
		public ArrayList Packets {
			get {
				return packets;
			}
		}
		
		public void Clear()
		{
			changed = false;
			packets.Clear();
			list.Items.Clear();
			InfoList.Nodes.Clear();
			RawData.Data = null;
		}
		
		public void Receive(IPPacket p)
		{
			packets.Add(p);
			ResolveTag r1 = null;
			ResolveTag r2 = null;
			string source = DnsCache.Instance.GetName(p.SourceAddress);
			if (source == null) {
				source = p.SourceAddress.ToString();
				r1 = new ResolveTag(true, p);
			}
			if (p.SourcePort >= 0)
				source += ":" + p.SourcePort.ToString();
			string destination = DnsCache.Instance.GetName(p.DestinationAddress);
			if (destination == null) {
				destination = p.DestinationAddress.ToString();
				r2 = new ResolveTag(false, p);
			}
			if (p.DestinationPort >= 0)
				destination += ":" + p.DestinationPort.ToString();
			string[] items = new string[] { p.Time.ToString(), p.Protocol.ToString(), source, destination, p.TotalLength.ToString(), p.TypeText };
			ListViewItem item = new ListViewItem(items, p.ImageIndex);
			list.Items.Add(item);
			ResolveCallback callback = new ResolveCallback(Resolved);
			if (r1 != null) {
				r1.Item = item;
				DnsCache.Instance.BeginResolve(p.SourceAddress, callback, r1);
			}
			if (r2 != null) {
				r2.Item = item;
				DnsCache.Instance.BeginResolve(p.DestinationAddress, callback, r2);
			}
			if (list.SelectedIndices.Count == 1) {
				if (list.SelectedIndices[0] == list.Items.Count - 2) {
					list.Items[list.Items.Count - 1].Selected = true;
					list.Items[list.Items.Count - 1].EnsureVisible();
				}
			}
		}
		#endregion
		
		#region Dns Resolving
		private class ResolveTag {
			public bool Source;
			public IPPacket Packet;
			public ListViewItem Item;
			
			public ResolveTag(bool source, IPPacket packet)
			{
				this.Packet = packet;
				this.Source = source;
			}
		}
		
		private void Resolved(object tag, System.Net.IPAddress address, string name)
		{
			if (InvokeRequired) {
				Invoke(new ResolveCallback(Resolved),
				       new object[] { tag, address, name });
				return;
			}
			ResolveTag r = (ResolveTag)tag;
			ListViewItem item = r.Item;
			if (item.ListView == null)
				return;
			if (r.Source) {
				if (r.Packet.SourcePort >= 0)
					name += ":" + r.Packet.SourcePort.ToString();
				item.SubItems[2].Text = name;
			} else {
				if (r.Packet.DestinationPort >= 0)
					name += ":" + r.Packet.DestinationPort.ToString();
				item.SubItems[3].Text = name;
			}
		}
		#endregion
	}
}
