// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Windows.Forms;

namespace TCPSniffer
{
	public class ProcessCache
	{
		public static readonly ProcessCache Instance = new ProcessCache();
		
		ImageList imageList = new ImageList();
		
		public ImageList ImageList {
			get {
				return imageList;
			}
		}
		
		private ProcessCache()
		{
			this.imageList.ImageSize = new System.Drawing.Size(16, 16);
			this.imageList.TransparentColor = System.Drawing.Color.Transparent;
		}
		
		Hashtable processHash = new Hashtable();
		
		public ProcessInfo GetProcessInfo(int pid)
		{
			lock (processHash) {
				ProcessInfo info = (ProcessInfo)processHash[pid];
				if (info != null) return info;
				try {
					info = new ProcessInfo(Process.GetProcessById(pid));
				} catch (ArgumentException) {
					info = new ProcessInfo(pid, null, "Error");
				}
				processHash.Add(pid, info);
				return info;
			}
		}
		
		public void SetProcessInfo(ProcessInfo info)
		{
			lock (processHash) {
				processHash[info.Id] = info;
			}
		}
		
		Hashtable iconHash = new Hashtable();
		
		public int GetIcon(string filename)
		{
			if (filename == "") return -1;
			filename = NormalizePath(filename);
			if (iconHash.ContainsKey(filename)) {
				return (int)iconHash[filename];
			} else {
				try {
					using (Icon ico = Grunwald.Gui.Components.IconLoader.GetIcon(filename, false)) {
						imageList.Images.Add(ico);
						int index = imageList.Images.Count - 1;
						iconHash.Add(filename, index);
						return index;
					}
				} catch (ArgumentException) {
					iconHash.Add(filename, -1);
					return -1;
				}
			}
		}
		
		public static string NormalizePath(string path)
		{
			if (path == null) return null;
			if (path == "") return "";
			if (path.IndexOf(":\\") > 1) {
				path = path.Substring(path.IndexOf(":\\") - 1);
			}
			try {
				return Path.GetFullPath(path).ToLower();
			} catch (Exception ex) {
				throw new Exception(path, ex);
			}
		}
	}
	
	[Serializable]
	public class ProcessInfo
	{
		public int Id;
		public string Filename;
		public string Name;
		
		public ProcessInfo() {} // for serialization
		
		public ProcessInfo(int pid, string filename, string name)
		{
			this.Id = pid;
			this.Filename = filename;
			this.Name = name;
		}
		
		public ProcessInfo(Process p)
		{
			this.Id = p.Id;
			this.Name = p.ProcessName;
			this.Filename = "";
			if (Id == 0 || Id == 4) // system processes without path
				return;
			try {
				ProcessModule mod = p.MainModule;
				if (mod != null)
					this.Filename = mod.FileName;
			} catch {}
		}
	}
}
