// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Net;
using System.Windows.Forms;

namespace TCPSniffer
{
	public class TcpPacket : IPPacket
	{
		int pos;
		
		public TcpPacket(IPPacket basePacket) : base(basePacket.Raw, basePacket.Time)
		{
			pos = base.HeaderLength;
		}
		
		public TcpPacket(byte[] raw) : base(raw)
		{
			pos = base.HeaderLength;
		}
		
		public uint Sequence {
			get {
				unchecked {
					return (uint)(Raw[pos + 4] << 24 | Raw[pos + 5] << 16 |
					              Raw[pos + 6] << 8 | Raw[pos + 7]);
				}
			}
		}
		
		public uint Acknowledgement {
			get {
				unchecked {
					return (uint)(Raw[pos + 8] << 24 | Raw[pos + 9] << 16 |
					              Raw[pos + 10] << 8 | Raw[pos + 11]);
				}
			}
		}
		
		public int TcpOffset {
			get {
				return Raw[pos + 12] >> 4;
			}
		}
		
		public override int TotalHeaderLength {
			get {
				return base.TotalHeaderLength + 4 * (Raw[pos + 12] >> 4);
			}
		}
		
		public byte Flag {
			get {
				return Raw[pos + 13];
			}
		}
		
		public int Window {
			get {
				return Raw[pos + 14] * 256 + Raw[pos + 15];
			}
		}
		
		public int TcpChecksum {
			get {
				return Raw[pos + 16] * 256 + Raw[pos + 17];
			}
		}
		
		public int UrgentPointer {
			get {
				return Raw[pos + 18] * 256 + Raw[pos + 19];
			}
		}
		
		public override void CreateInfoNodes(TreeNodeCollection c, TreeNodeFactory f)
		{
			base.CreateInfoNodes(c, f);
			TreeNode node = f.Create("TCP header", pos, this.TcpOffset * 4);
			c.Add(node);
			TreeNodeCollection a = node.Nodes;
			string txt;
			txt = Format("Source port", this.SourcePort, 4);
			a.Add(f.Create(txt, pos, 2));
			txt = Format("Destination port", this.DestinationPort, 4);
			a.Add(f.Create(txt, pos + 2, 2));
			txt = Format("Sequence", this.Sequence, 8);
			a.Add(f.Create(txt, pos + 4, 4));
			txt = Format("Acknowledgement", this.Acknowledgement, 8);
			a.Add(f.Create(txt, pos + 8, 4));
			txt = string.Format("Offset: {0} (0x{1:x})", this.TcpOffset * 4, this.TcpOffset);
			a.Add(f.Create(txt, pos + 12, 1));
			txt = string.Format("Flag: 0x{0:x2} ({1}{2}{3}{4}{5}{6})", this.Flag,
			                    GetFlag(0x01, "F"),
			                    GetFlag(0x02, "S"),
			                    GetFlag(0x04, "R"),
			                    GetFlag(0x08, "P"),
			                    GetFlag(0x10, "A"),
			                    GetFlag(0x20, "U"));
			a.Add(f.Create(txt, pos + 13, 1));
			txt = Format("Window", this.Window, 4);
			a.Add(f.Create(txt, pos + 14, 2));
			txt = Format("Checksum", this.TcpChecksum, 4);
			a.Add(f.Create(txt, pos + 16, 2));
			txt = Format("Urgent Pointer", this.UrgentPointer, 4);
			a.Add(f.Create(txt, pos + 18, 2));
		}
		
		string GetFlag(byte test, string name)
		{
			if ((test & Flag) == 0)
				return ".";
			else
				return name;
		}
		
		public override string TypeText {
			get {
				if ((Flag & 2) == 2) {
					if ((Flag & 0x10) == 0x10)
						return "Connection accept";
					else
						return "Connection request";
				} else if ((Flag & 1) == 1) {
					return "Connection shutdown";
				} else if ((Flag & 4) == 4) {
					return "Connection reset";
				} else if ((Flag & 0x10) == 0x10) {
					int len = Raw.Length - 20 - HeaderLength;
					if (len == 0)
						return "Tcp acknowledgement";
					else
						return "Tcp data (" + len.ToString() + " byte)";
				} else {
					return "Unknown Tcp packet";
				}
			}
		}
		
		public override int ImageIndex {
			get {
				if (MainForm.Is2000)
					return base.ImageIndex;
				ProcessInfo p = this.Process;
				if (p != null)
					return ProcessCache.Instance.GetIcon(p.Filename);
				else
					return -1;
			}
		}
		
		ProcessInfo process;
		bool gotProcess = false;
		public ProcessInfo Process {
			get {
				if (gotProcess) return process;
				gotProcess = true;
				if (MainForm.Is2000)
					return null;
				CMIB_TCPEXTABLE table = iphelper.GetTcpExTable();
				int sourcePort = this.SourcePort;
				int destPort = this.DestinationPort;
				IPAddress sourceAddress = this.SourceAddress;
				IPAddress destAddress = this.DestinationAddress;
				for (int i = 0; i < table.dwNumEntries; i++) {
					CMIB_TCPEXROW row = table.table[i];
					//MessageBox.Show(row.dwLocalPort.ToString() + "   "+ sourcePort.ToString());
					uint lPort = row.get_local_port();
					uint rPort = row.get_remote_port();
					IPAddress lIP = row.get_localip();
					IPAddress rIP = row.get_remoteip();
					if ((sourcePort == lPort &&
					     destPort == rPort &&
					     sourceAddress.Equals(lIP) &&
					     destAddress.Equals(rIP)) ||
					    (sourcePort == rPort &&
					     destPort == lPort &&
					     sourceAddress.Equals(rIP) &&
					     destAddress.Equals(lIP)))
					{
						process = ProcessCache.Instance.GetProcessInfo(row.dwProcessId);
						return process;
					}
				}
				return null;
			}
		}
	}
}
