// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using ICSharpCode.SharpZipLib.Zip;

namespace ApplyUpdate
{
	/// <summary>
	/// Description of MainForm.
	/// </summary>
	public class MainForm : System.Windows.Forms.Form
	{
		private System.Windows.Forms.ProgressBar progressBar;
		private System.Windows.Forms.TextBox logTextBox;
		
		static MainForm instance;
		
		public MainForm()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			
			instance = this;
		}
		
		static string[] cmd_args;
		
		[STAThread]
		public static void Main(string[] args)
		{
			if (args.Length == 0 || (args.Length % 2) > 0) {
				MessageBox.Show("This is the DG Network update program. It should only be called by the AutoUpdate plugin.", "DG Network Update");
				return;
			}
			cmd_args = args;
			Application.Run(new MainForm());
		}
		
		#region Windows Forms Designer generated code
		/// <summary>
		/// This method is required for Windows Forms designer support.
		/// Do not change the method contents inside the source code editor. The Forms designer might
		/// not be able to load this method if it was changed manually.
		/// </summary>
		private void InitializeComponent() {
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(MainForm));
			this.logTextBox = new System.Windows.Forms.TextBox();
			this.progressBar = new System.Windows.Forms.ProgressBar();
			this.SuspendLayout();
			// 
			// logTextBox
			// 
			this.logTextBox.BorderStyle = System.Windows.Forms.BorderStyle.None;
			this.logTextBox.Dock = System.Windows.Forms.DockStyle.Fill;
			this.logTextBox.Location = new System.Drawing.Point(8, 8);
			this.logTextBox.Multiline = true;
			this.logTextBox.Name = "logTextBox";
			this.logTextBox.ReadOnly = true;
			this.logTextBox.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
			this.logTextBox.Size = new System.Drawing.Size(328, 78);
			this.logTextBox.TabIndex = 1;
			this.logTextBox.Text = "Running DG Network update...";
			this.logTextBox.WordWrap = false;
			// 
			// progressBar
			// 
			this.progressBar.Dock = System.Windows.Forms.DockStyle.Bottom;
			this.progressBar.Location = new System.Drawing.Point(8, 86);
			this.progressBar.Name = "progressBar";
			this.progressBar.Size = new System.Drawing.Size(328, 24);
			this.progressBar.TabIndex = 0;
			// 
			// MainForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(344, 118);
			this.Controls.Add(this.logTextBox);
			this.Controls.Add(this.progressBar);
			this.DockPadding.All = 8;
			this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedSingle;
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.MaximizeBox = false;
			this.MinimizeBox = false;
			this.Name = "MainForm";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
			this.Text = "DG Network Update";
			this.Closing += new System.ComponentModel.CancelEventHandler(this.MainFormClosing);
			this.Load += new System.EventHandler(this.MainFormLoad);
			this.ResumeLayout(false);
		}
		#endregion
		
		
		static volatile bool updateAbort, updatePause, updateFinished;
		
		/// <summary>
		/// Notifies that the update is currently in a safe state (e.g. between two
		/// archives). This method could pause the update or abort it by throwing
		/// an AbortException.
		/// </summary>
		public static void SafeStep()
		{
			if (updateAbort)
				throw new AbortException();
			while (updatePause) {
				Thread.Sleep(100);
				if (updateAbort)
					throw new AbortException();
			}
		}
		
		delegate void void_string(string text);
		
		public static void AppendToLog(string text)
		{
			instance.BeginInvoke(new void_string(instance.DoAppendToLog), new object[] {text});
		}
		
		void DoAppendToLog(string text)
		{
			logTextBox.AppendText(text);
			logTextBox.Select(logTextBox.TextLength, 0);
			logTextBox.ScrollToCaret();
		}
		
		void MainFormClosing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			if (updateFinished)
				return;
			updatePause = true;
			if (MessageBox.Show("Cancelling the update could corrupt your DG Network installation.\nDo you really want to cancel?", "DG Network Update", MessageBoxButtons.YesNo) == DialogResult.Yes) {
				updateAbort = true;
				thread.Join(1000);
			} else {
				e.Cancel = true;
			}
			updatePause = false;
		}
		
		Thread thread;
		
		void MainFormLoad(object sender, EventArgs e)
		{
			thread = new Thread(new ThreadStart(RunThread));
			thread.Name = "";
			thread.Start();
		}
		
		void RunThread()
		{
			try {
				string dgndir = Path.GetFullPath(".").ToLower();
				AppendToLog("\r\nWaiting for DG Network to exit...");
				while (IsDGNetworkRunning(dgndir)) {
					SafeStep();
					Thread.Sleep(500);
				}
				for (int i = 0; i < cmd_args.Length; i += 2) {
					Extract(cmd_args[i], cmd_args[i + 1]);
				}
				SafeStep();
				AppendToLog("\r\nStarting DG Network...");
				Process.Start(Path.GetFullPath("DG Network.exe"), "/postUpdate");
				SafeStep();
			} catch (AbortException) {
				// ignore
			} catch (ThreadAbortException) {
				// ignore (but you should not use this!)
			} catch (Exception ex) {
				Console.WriteLine(ex);
				MessageBox.Show(ex.ToString(), "Exception");
			} finally {
				updateFinished = true;
				instance.BeginInvoke(new MethodInvoker(instance.Close));
			}
		}
		
		bool IsDGNetworkRunning(string dgndir)
		{
			Process[] processList = Process.GetProcessesByName("DG Network");
			foreach (Process p in processList) {
				string filename = p.MainModule.FileName;
				string path = Path.GetFullPath(Path.GetDirectoryName(filename)).ToLower();
				if (path == dgndir)
					return true;
			}
			return false;
		}
		
		void Extract(string archive, string to)
		{
			SafeStep();
			if (to.StartsWith("/"))
				to = to.Substring(1);
			
			ArrayList deleteFiles = new ArrayList();
			if (File.Exists(Path.Combine(to, "filelist.txt"))) {
				using (StreamReader reader = new StreamReader(Path.Combine(to, "filelist.txt"))) {
					AppendToLog("\r\nExtracting " + reader.ReadLine() + "...");
					string line;
					while ((line = reader.ReadLine()) != null) {
						line = line.Trim();
						if (line.Length == 0 || line[0] == '#')
							continue;
						CheckFilename(line);
						deleteFiles.Add(line);
					}
				}
				
			} else {
				AppendToLog("\r\nExtracting without filelist: " + Path.GetFileName(archive));
			}
			
			using (ZipInputStream input = new ZipInputStream(new FileStream(archive, FileMode.Open, FileAccess.Read))) {
				ZipEntry objEntry;
				while ((objEntry = input.GetNextEntry()) != null) {
					CheckFilename(objEntry.Name);
					if (objEntry.Name == "ApplyUpdate.new.exe") {
						RemoveFromCollection("ApplyUpdate.exe", deleteFiles);
						continue;
					}
					RemoveFromCollection(objEntry.Name, deleteFiles);
					string file = Path.Combine(to, objEntry.Name);
					if (objEntry.IsDirectory) {
						if (!Directory.Exists(file))
							Directory.CreateDirectory(file);
					} else {
						ExtractFile(file, input);
						try {
							File.SetLastWriteTime(file, objEntry.DateTime);
						} catch { }
					}
				}
			}
			// delete files in directory before trying to delete the directory
			deleteFiles.Sort(new StringLengthComparer());
			foreach (string file in deleteFiles) {
				DeleteFile(Path.Combine(to, file));
			}
		}
		
		class StringLengthComparer : IComparer
		{
			public int Compare(object x, object y)
			{
				return y.ToString().Length - x.ToString().Length;
			}
		}
		
		void RemoveFromCollection(string name, ArrayList col)
		{
			for (int i = 0; i < col.Count; ++i) {
				if (string.Compare(name, (string)col[i], true) == 0) {
					col.RemoveAt(i);
					break;
				}
			}
		}
		
		void CheckFilename(string text)
		{
			if (text.IndexOf("..") >= 0)
				throw new ApplicationException("'..' forbidden in file name.");
			if (text.IndexOfAny(Path.InvalidPathChars) >= 0)
				throw new ApplicationException("Invalid file name: " + text);
		}
		
		void ExtractFile(string file, ZipInputStream input)
		{
			try {
				using (FileStream fs = new FileStream(file, FileMode.Create, FileAccess.Write)) {
					byte[] abyData = new byte[4096];
					int nSize;
					do {
						nSize = input.Read(abyData, 0, abyData.Length);
						if (nSize > 0)
							fs.Write(abyData, 0, nSize);
					} while (nSize > 0);
				}
			} catch (IOException ex) {
				DialogResult res = MessageBox.Show("Error extracting '" + file + "':\n" + ex.Message + "\nMake sure no other process is accessing the file.", "Update Error", MessageBoxButtons.AbortRetryIgnore);
				if (res == DialogResult.Retry) {
					ExtractFile(file, input);
				} else if (res == DialogResult.Ignore) {
					// well, ignore the file...
				} else {
					throw new AbortException(); // abort everything
				}
			}
		}
		
		void DeleteFile(string file)
		{
			try {
				if (file.EndsWith("/")) {
					if (Directory.Exists(file)) {
						if (Directory.GetFiles(file).Length == 0 && Directory.GetDirectories(file).Length == 0) {
							Directory.Delete(file, false);
						}
					}
				} else {
					if (File.Exists(file))
						File.Delete(file);
				}
			} catch (IOException ex) {
				DialogResult res = MessageBox.Show("Error deleting '" + file + "':\n" + ex.Message + "\nMake sure no other process is accessing the file.", "Update Error", MessageBoxButtons.AbortRetryIgnore);
				if (res == DialogResult.Retry) {
					DeleteFile(file);
				} else if (res == DialogResult.Ignore) {
					// well, ignore the file...
				} else {
					throw new AbortException(); // abort everything
				}
			}
		}
	}
}
