// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.IO;
using System.Text;
using System.Xml;
using System.Security.Cryptography;
using ICSharpCode.SharpZipLib.Zip;

namespace CreateUpdate
{
	public interface IArchiveInfo
	{
		string Path { get; }
		long GetFileSize();
		byte[] GetChecksum();
		string ArchiveFilename { get; }
	}
	
	public class OldArchiveInfo : IArchiveInfo
	{
		string path;
		long fileSize;
		byte[] checksum;
		string archiveFilename;
		
		public OldArchiveInfo(XmlElement element)
		{
			path = element.GetAttribute("extractTo");
			fileSize = XmlConvert.ToInt64(element.GetAttribute("filesize"));
			checksum = Convert.FromBase64String(element.GetAttribute("checksum"));
			archiveFilename = element.InnerText;
		}
		
		public string ArchiveFilename {
			get {
				return archiveFilename;
			}
		}
		public string Path {
			get {
				return path;
			}
		}
		public long GetFileSize()
		{
			return fileSize;
		}
		public byte[] GetChecksum()
		{
			return checksum;
		}
	}
	
	public class ArchiveInfo : IArchiveInfo
	{
		string path;
		
		public ArchiveInfo(string path, string friendlyName)
		{
			this.path = path;
			this.friendlyName = friendlyName;
		}
		
		public string Path {
			get {
				return path;
			}
		}
		
		string archiveFilename;
		
		public string ArchiveFilename {
			get {
				return archiveFilename;
			}
			set {
				archiveFilename = value;
			}
		}
		
		string friendlyName;
		
		public string FriendlyName {
			get {
				return friendlyName;
			}
		}
		
		public long GetFileSize()
		{
			return new FileInfo(archiveFilename).Length;
		}
		
		public byte[] GetChecksum()
		{
			using (FileStream fs = new FileStream(archiveFilename, FileMode.Open, FileAccess.Read)) {
				using (SHA1CryptoServiceProvider sha1 = new SHA1CryptoServiceProvider()) {
					return sha1.ComputeHash(fs);
				}
			}
		}
		
		ZipOutputStream zipStream;
		StringBuilder filenames;
		
		public void CreateZIP(string path)
		{
			archiveFilename = path;
			Console.WriteLine("Creating {0}...", path);
			zipStream = new ZipOutputStream(new FileStream(path, FileMode.Create));
			filenames = new StringBuilder();
		}
		
		public void AddFolder(string realPath, string pathInZIP, bool recursive)
		{
			AddFolderToZIP(realPath, zipStream, recursive, pathInZIP);
		}
		
		public void AddFile(string realPath)
		{
			AddFile(realPath, System.IO.Path.GetFileName(realPath));
		}
		
		public void AddFile(string realPath, string pathInZIP)
		{
			AddFileToZip(zipStream, realPath, pathInZIP);
		}
		
		public void AddEmptyFolder(string path)
		{
			zipStream.PutNextEntry(new ZipEntry(path + "/"));
		}
		
		public void FinishZIP()
		{
			zipStream.PutNextEntry(new ZipEntry("filelist.txt"));
			using (StreamWriter w = new StreamWriter(zipStream, Encoding.UTF8)) {
				w.WriteLine(friendlyName);
				w.WriteLine("# The filelist is used by the updater to know what");
				w.WriteLine("# files can be deleted when they are no longer");
				w.WriteLine("# contained in new releases.");
				w.WriteLine("# The first line of the file MUST be the display name.");
				w.WriteLine();
				w.WriteLine("# DO NOT MODIFY THIS FILE!!!");
				w.WriteLine();
				w.Write(filenames.ToString());
			} // probably already closes zipStream
			zipStream.Close();
		}
		
		void AddFolderToZIP(string path, ZipOutputStream z, bool rekursiv, string BasePath)
		{
			if (!BasePath.EndsWith("/"))
				throw new ArgumentException("BasePath muss ein Verzeichnis sein", "BasePath");
			if (BasePath == "/")
				BasePath = "";
			string[] a = Directory.GetFiles(path);
			if (BasePath.Length > 0) {
				filenames.Append(BasePath);
				filenames.Append("\r\n");
				z.PutNextEntry(new ZipEntry(BasePath));
			}
			for(int i = 0; i < a.Length; i++) {
				string filename = System.IO.Path.GetFileName(a[i]);
				if (filename.ToLower() == MainClass.ApplyUpdate.ToLower()) {
					filename = MainClass.ApplyUpdateTemp;
				}
				AddFileToZip(z, a[i], BasePath + filename);
			}
			if (rekursiv) {
				a = Directory.GetDirectories(path);
				for(int i = 0; i < a.Length; i++) {
					if (a[i].EndsWith("\\.svn")) continue;
					AddFolderToZIP(a[i], z, true, BasePath + (new DirectoryInfo(a[i]).Name) + "/");
				}
			}
		}
		
		void AddFileToZip(ZipOutputStream z, string realFile, string virtualName)
		{
			AddFileToZip(z, realFile, new ZipEntry(virtualName));
		}
		
		void AddFileToZip(ZipOutputStream z, string realFile, ZipEntry virtualFile)
		{
			if (virtualFile.Name == MainClass.ApplyUpdateTemp)
				filenames.Append(MainClass.ApplyUpdate);
			else
				filenames.Append(virtualFile.Name);
			filenames.Append("\r\n");
			virtualFile.DateTime = File.GetLastWriteTime(realFile);
			z.PutNextEntry(virtualFile);
			FileStream c = new FileStream(realFile, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
			byte[] b = new byte[4096];
			int l=1;
			while (l > 0) {
				l = c.Read(b, 0, 4096);
				if (l > 0)
					z.Write(b, 0, l);
			}
			c.Close();
			z.CloseEntry();
		}
	}
}
