// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Xml;

namespace CreateUpdate
{
	class MainClass
	{
		const string dgnPath = "../../../bin/release"; // use release directory
		static string updateInfoPath;
		public const string ApplyUpdate = "ApplyUpdate.exe";
		public const string ApplyUpdateTemp = "ApplyUpdate.new.exe";
		
		public static int Main(string[] args)
		{
			try {
				if (args.Length == 0) {
					Console.WriteLine(@"Missing argument: directory update information (e.g. 'CreateUpdate.exe d:\homepage\network\update')");
					return 1;
				}
				updateInfoPath = args[0];
				if (!Directory.Exists(updateInfoPath)) {
					Console.WriteLine("Directory not found: " + updateInfoPath);
					return 1;
				}
				return new MainClass().Run();
			} catch (Exception ex) {
				Console.WriteLine(ex);
				return 2;
			}
		}
		
		XmlWriter combinedWriter;
		
		int Run()
		{
			if (!File.Exists(Path.Combine(dgnPath, "DG Network.exe"))) {
				Console.WriteLine("DG Network not found in " + Path.GetFullPath(dgnPath));
				return 1;
			}
			XmlTextWriter writer = new XmlTextWriter("all.xml", Encoding.UTF8);
			writer.Formatting = Formatting.Indented;
			writer.WriteStartDocument();
			writer.WriteStartElement("updateInfoCombination");
			combinedWriter = writer;
			
			GenerateDGNUpdate();
			GeneratePluginUpdate("MSIE", "Browser/MSIE", "Browser/MSIE");
			GeneratePluginUpdate("Mozilla", "Browser/Mozilla", "Browser/Mozilla");
			GeneratePluginUpdate("AutoUpdate");
            GeneratePluginUpdate("History");
			GeneratePluginUpdate("Discovery", "Discovery", "DiscoveryAddin");
			GeneratePluginUpdate("IRC", "IRCPlugin");
			GeneratePluginUpdate("ICQ", "ICQPlugin");
			
			writer.WriteEndDocument();
			writer.Close();
			return 0;
		}
		
		void GeneratePluginUpdate(string name)
		{
			GeneratePluginUpdate(name, name, name, name + ".dll");
		}
		
		void GeneratePluginUpdate(string name, string libname)
		{
			GeneratePluginUpdate(name, name, name, libname + ".dll");
		}
		
		void GeneratePluginUpdate(string name, string binName, string srcName)
		{
			GeneratePluginUpdate(name, binName, srcName, name + ".dll");
		}
		
		void GeneratePluginUpdate(string name, string binName, string srcName, string libname)
		{
			if (File.Exists(name + ".zip"))
				File.Delete(name + ".zip");
			if (File.Exists(name + ".xml"))
				File.Delete(name + ".xml");
			Console.WriteLine();
			XmlElement e;
			int oldRevision = GetOldRevision(name, out e);
			if (!IsChangedSince(oldRevision, "Addins/" + srcName)) {
				Console.WriteLine("{0}: not changed since r{1}", name, oldRevision);
				if (combinedWriter != null)
					e.WriteTo(combinedWriter);
				return;
			}
			FileVersionInfo version = FileVersionInfo.GetVersionInfo(Path.Combine(dgnPath, "Addins/" + binName + "/" + libname));
			Console.WriteLine("{0}: creating update (r{1} -> r{2})", name, oldRevision, version.FilePrivatePart);
			ArchiveInfo a = new ArchiveInfo("~", name);
			a.CreateZIP(name + ".zip");
			a.AddFolder(Path.Combine(dgnPath, "Addins/" + binName), "/", true);
			a.FinishZIP();
			GenerateXML(name, version.FileVersion, a);
		}
		
		int GetOldRevision(string xmlName, out XmlElement e)
		{
			e = null;
			string fileName = Path.Combine(updateInfoPath, xmlName + ".xml");
			if (!File.Exists(fileName))
				return 0;
			XmlDocument document = new XmlDocument();
			document.Load(fileName);
			e = document.DocumentElement;
			string version = e.GetAttribute("version");
			string[] numbers = version.Split('.');
			return int.Parse(numbers[numbers.Length - 1]);
		}
		
		string Shell(string app, string args)
		{
			ProcessStartInfo info = new ProcessStartInfo(app, args);
			info.UseShellExecute = false;
			info.RedirectStandardOutput = true;
			return Process.Start(info).StandardOutput.ReadToEnd();
		}
		
		bool IsChangedSince(int baseRevision, string srcDir)
		{
			int rev = baseRevision + 1;
			string cmd = "log http://localhost/DGN/trunk/" + srcDir + " -r " + rev + ":HEAD --xml";
			string result = Shell("svn", cmd);
			XmlDocument doc = new XmlDocument();
			try {
				doc.LoadXml(result);
			} catch (XmlException) {
				Console.WriteLine("$svn " + cmd);
				Console.WriteLine(result);
				throw;
			}
			return doc.DocumentElement.ChildNodes.Count > 0;
		}
		
		void GenerateDGNUpdate()
		{
			if (File.Exists("core.xml"))
				File.Delete("core.xml");
			if (File.Exists("core.zip"))
				File.Delete("core.zip");
			if (File.Exists("data.zip"))
				File.Delete("data.zip");
			
			XmlElement e;
			int oldRevision = GetOldRevision("core", out e);
			
			Console.WriteLine("Generating main DG Network update... (from r{0})", oldRevision);
			ArchiveInfo main = new ArchiveInfo("/", "DG Network program");
			main.CreateZIP("core.zip");
			main.AddFile(Path.Combine(dgnPath, "DG Network.exe"));
			main.AddFile(Path.Combine(dgnPath, "Core.dll"));
			main.AddFile(Path.Combine(dgnPath, "Base.dll"));
			main.AddFile(Path.Combine(dgnPath, "log4net.dll"));
			main.AddFile(Path.Combine(dgnPath, "Grunwald.Gui.dll"));
			main.AddFile(Path.Combine(dgnPath, "Grunwald.P2PLib.dll"));
			main.AddFile(Path.Combine(dgnPath, "Grunwald.P2PLib.DGNetwork.dll"));
			main.AddEmptyFolder("Addins");
			main.AddFile(Path.Combine(dgnPath, "Addins/Core.addin"), "Addins/Core.addin");
			main.FinishZIP();
			
			IArchiveInfo data;
			if (IsChangedSince(oldRevision, "data")) {
				ArchiveInfo data2 = new ArchiveInfo("/data", "DG Network data");
				data2.CreateZIP("data.zip");
				data2.AddFolder(Path.Combine(dgnPath, "data"), "/", true);
				data2.FinishZIP();
				data = data2;
			} else {
				data = GetArchive(e, "/data");
			}
			
			FileVersionInfo version = FileVersionInfo.GetVersionInfo(Path.Combine(dgnPath, "DG Network.exe"));
			GenerateXML("core", version.FileVersion, main, data);
		}
		
		OldArchiveInfo GetArchive(XmlElement updateInfo, string extractTo)
		{
			foreach (XmlElement archive in updateInfo) {
				if (archive.GetAttribute("extractTo") == extractTo)
					return new OldArchiveInfo(archive);
			}
			throw new Exception("Archive '" + extractTo + "' not found!");
		}
		
		void GenerateXML(string xmlName, string version, params IArchiveInfo[] archives)
		{
			XmlTextWriter writer = new XmlTextWriter(xmlName + ".xml", Encoding.UTF8);
			writer.Formatting = Formatting.Indented;
			writer.WriteStartDocument();
			WriteXML(writer, xmlName, version, archives);
			writer.WriteEndDocument();
			writer.Close();
			#if !DEBUG
			File.Copy(xmlName + ".xml", Path.Combine(updateInfoPath, xmlName + ".xml"), true);
			#endif
			if (combinedWriter != null)
				WriteXML(combinedWriter, xmlName, version, archives);
		}
		
		void WriteXML(XmlWriter writer, string xmlName, string version, IArchiveInfo[] archives)
		{
			writer.WriteStartElement("updateInfo");
			writer.WriteAttributeString("name", xmlName);
			writer.WriteAttributeString("version", version);
			foreach (IArchiveInfo archive in archives) {
				writer.WriteStartElement("archive");
				writer.WriteAttributeString("checksum", Convert.ToBase64String(archive.GetChecksum()));
				writer.WriteAttributeString("filesize", XmlConvert.ToString(archive.GetFileSize()));
				writer.WriteAttributeString("extractTo", archive.Path);
				writer.WriteString("http://www.danielgrunwald.de/network/update/" + Path.GetFileName(archive.ArchiveFilename));
				writer.WriteEndElement();
			}
			writer.WriteEndElement();
		}
	}
}
