// Copyright (C) 2005  Daniel Grunwald
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// For more information read the file "LICENSE.txt".

using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Grunwald.Gui.Controls;
using Grunwald.Gui.Components;
using ListViewSubItem = System.Windows.Forms.ListViewItem.ListViewSubItem;

namespace ResourceStringEditor
{
	/// <summary>
	/// Description of MainForm.
	/// </summary>
	public class MainForm : System.Windows.Forms.Form
	{
		private System.Windows.Forms.TabControl tabControl;
		private System.Windows.Forms.Splitter splitter;
		private System.Windows.Forms.ListView listView;
		
		#region Constructor
		public MainForm(string[] files)
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			this.Text = "DGN Resource String Editor";
			tabControl.Visible = false;
			AddCol("ID");
			foreach (string file in files) {
				string s = Path.GetFileNameWithoutExtension(file);
				if (s.ToLower() == "stringresources") {
					AddLang(file, "");
				} else if (s.ToLower().StartsWith("stringresources.")) {
					AddLang(file, s.Substring(("stringresources.").Length));
				} else {
					MessageBox.Show(file + " is not an string resource file!");
				}
			}
			mainHash = (Hashtable)hash[""];
			bool direct;
			defaultFont = listView.Font;
			italicFont = new Font(defaultFont, FontStyle.Italic);
			foreach (DictionaryEntry entry in mainHash) {
				string id = (string)entry.Key;
				ListViewItem item = new ListViewItem(id);
				item.UseItemStyleForSubItems = false;
				foreach (string language in languages) {
					string val = GetString(id, language, out direct);
					ListViewSubItem subItem = new ListViewSubItem(item, val);
					if (!direct) {
						subItem.ForeColor = Color.Gray;
						subItem.Font = italicFont;
					}
					item.SubItems.Add(subItem);
				}
				listView.Items.Add(item);
			}
			foreach (ColumnHeader h in listView.Columns) {
				h.Width = (listView.ClientSize.Width - 30) / (languages.Count + 1);
			}
			listView.SelectedIndexChanged += new EventHandler(ListViewIndexChanged);
			listView.AfterLabelEdit += new LabelEditEventHandler(ListViewLabelEdit);
			Corrections.MakeProportional(listView);
			ControlRow row = new ControlRow();
			row.Dock = DockStyle.Top;
			row.Height = 32;
			row.Direction = Direction.LeftToRight;
			row.ContentAlignment = ContentAlignment.MiddleCenter;
			this.Controls.Add(row);
			AutoSizedButton btn;
			btn = new AutoSizedButton("New");
			btn.Click += new EventHandler(BtnNewClick);
			row.Controls.Add(btn);
			btn = new AutoSizedButton("Save");
			btn.Click += new EventHandler(BtnSaveClick);
			row.Controls.Add(btn);
			row.ArrangeActive = true;
		}
		#endregion
		
		private void BtnNewClick(object sender, EventArgs e) {
			if (mainHash[""] != null) {
				MessageBox.Show("There is already an element without name - fix that first!");
			} else {
				ListViewItem item = new ListViewItem("");
				item.UseItemStyleForSubItems = false;
				foreach (string language in languages) {
					ListViewSubItem subItem = new ListViewSubItem(item, "new string (" + language + ")");
					subItem.ForeColor = Color.Gray;
					subItem.Font = italicFont;
					item.SubItems.Add(subItem);
				}
				listView.Items.Add(item);
				item.BeginEdit();
			}
		}
		
		private void BtnSaveClick(object sender, EventArgs e) {
			((Button)sender).Enabled = false;
			Application.DoEvents();
			for (int i = 0; i < languages.Count; i++) {
				Save((string)languages[i], (string)files[i]);
			}
			string firstfile = (string)files[0];
			string buildfile = Path.Combine(Path.GetDirectoryName(firstfile), "../../lib/NAnt/NAnt.exe");
			if (File.Exists(buildfile)) {
				ProcessStartInfo info = new ProcessStartInfo(buildfile);
				info.CreateNoWindow = true;
				info.UseShellExecute = false;
				info.RedirectStandardError = true;
				info.RedirectStandardOutput = true;
				info.WorkingDirectory = Path.GetDirectoryName(firstfile);
				Process p = Process.Start(info);
				if (!p.WaitForExit(10000)) {
					MessageBox.Show("build.bat didn't finish in time!");
				}
				if (p.ExitCode != 0) {
					MessageBox.Show("ExitCode = " + p.ExitCode.ToString() + ":\n" +
					                p.StandardOutput.ReadToEnd() + "\n\n\n" +
					                p.StandardError.ReadToEnd());
				}
			}
			((Button)sender).Enabled = true;
		}
		private void Save(string language, string file) {
			Hashtable l = (Hashtable)hash[language];
			ArrayList entries = new ArrayList();
			foreach (DictionaryEntry e in l) {
				if (e.Value == null) continue;
				entries.Add(e);
			}
			entries.Sort(new KeyStringComparer());
			using (StreamWriter writer = new StreamWriter(file)) {
				writer.WriteLine("# This file was edited by ResourceStringEditor");
				writer.WriteLine("# You may edit it by hand, but it *will* be painful ;-)");
				writer.WriteLine();
				foreach (DictionaryEntry e in entries) {
					string key = (string)e.Key;
					writer.Write(key);
					if (key.Length < 30)
						writer.Write(new string(' ', 30 - key.Length));
					writer.Write(" = \"");
					string val = (string)e.Value;
					val = val.Replace("\r", "");
					val = val.Replace("\\", "\\\\");
					val = val.Replace("\n", "\\n");
					val = val.Replace("\"", "\\\"");
					writer.Write(val);
					writer.WriteLine("\"");
				}
				writer.WriteLine();
				writer.WriteLine("# End of language resource file");
			}
		}
		
		private class KeyStringComparer : IComparer {
			public int Compare(object x, object y) {
				return ((string)((DictionaryEntry)x).Key).CompareTo((string)((DictionaryEntry)y).Key);
			}
		}
		
		Font defaultFont;
		Font italicFont;
		Hashtable mainHash;
		Hashtable hash = new Hashtable(); // string->Hashtable
		ArrayList edit = new ArrayList(); // TextBox
		ArrayList languages = new ArrayList();
		ArrayList files = new ArrayList();
		
		#region Load Language
		private void AddLang(string file, string name) {
			if (hash.ContainsKey(name))
				return;
			Hashtable lang = new Hashtable(); // string->string
			hash.Add(name, lang);
			if (name != "")
				AddLang(file.Replace("." + name + ".", "."), "");
			int pos = name.LastIndexOf('.');
			if (pos > 0)
				AddLang(file.Replace("." + name + ".", "." + name.Substring(0, pos) + "."),
				        name.Substring(0, pos));
			int linenr = 0;
			using (StreamReader reader = new StreamReader(file)) {
				while(true) {
					string line = reader.ReadLine();
					linenr++;
					if (line == null) {
						break;
					}
					line = line.Trim();
					// skip empty or comment lines
					if (line.Length == 0 || line[0] == '#') {
						continue;
					}
					
					// search for a = char
					int idx = line.IndexOf('=');
					if (idx < 0) {
						MessageBox.Show("error in file " + file + " at line " + linenr);
						continue;
					}
					string key = line.Substring(0, idx).Trim();
					string val = line.Substring(idx + 1).Trim();
					
					if (val[0] == '"') { // case 1 : string value
						val = val.Trim(new char[] {'"'});
						StringBuilder tmp = new StringBuilder();
						for (int i = 0; i < val.Length; ++i) {
							switch (val[i]) { // handle the \ char
								case '\\':
									++i;
									if (i < val.Length)
										switch (val[i]) {
										case '\\':
											tmp.Append('\\');
											break;
										case 'n':
											tmp.Append("\r\n");
											break;
										case '\"':
											tmp.Append('\"');
											break;
									}
									break;
								default:
									tmp.Append(val[i]);
									break;
							}
						}
						lang.Add(key, tmp.ToString());
					} else {
						MessageBox.Show(file + " contains non-strings!");
					}
				}
			}
			
			bool direct;
			string languageName = GetString("LanguageName", name, out direct);
			if (!direct)
				languageName = ">>" + name + "<<";
			AddCol(languageName);
			languages.Add(name);
			files.Add(file);
			TabPage page = new TabPage(languageName);
			TextBox textBox = new TextBox();
			textBox.Multiline = true;
			textBox.Dock = DockStyle.Fill;
			textBox.WordWrap = false;
			textBox.ScrollBars = ScrollBars.Both;
			textBox.Tag = name;
			textBox.Text = "start_value_that_makes_the_set_to_empty_a_change";
			textBox.TextChanged += new EventHandler(BoxTextChanged);
			page.Controls.Add(textBox);
			tabControl.TabPages.Add(page);
			edit.Add(textBox);
		}
		
		private string GetString(string id, string language, out bool direct) {
			Hashtable l = (Hashtable)hash[language];
			string str = (string)l[id];
			if (str != null) {
				direct = true;
				return str;
			}
			direct = false;
			int pos = language.LastIndexOf('.');
			bool nul;
			if (pos > 0)
				return GetString(id, language.Substring(0, pos), out nul);
			else if (language != "")
				return GetString(id, "", out nul);
			else
				return "${res:" + id + "}";
		}
		
		private void AddCol(string name) {
			ColumnHeader head = new ColumnHeader();
			head.Text = name;
			listView.Columns.Add(head);
		}
		#endregion
		
		string selectedID;
		bool user = true;
		private void ListViewIndexChanged(object sender, EventArgs e) {
			user = false;
			tabControl.Visible = listView.SelectedItems.Count == 1;
			if (!tabControl.Visible) return;
			selectedID = listView.SelectedItems[0].Text;
			for (int i = 0; i < languages.Count; i++) {
				Hashtable l = (Hashtable)hash[(string)languages[i]];
				string txt = (string)l[selectedID];
				if (txt == null) txt = "";
				((TextBox)edit[i]).Text = txt;
			}
			user = true;
		}
		
		private void BoxTextChanged(object sender, EventArgs e) {
			TextBox b = (TextBox)sender;
			if (b.Text == "")
				b.BackColor = Color.LightGray;
			else
				b.ResetBackColor();
			if (user) {
				Hashtable l = (Hashtable)hash[(string)b.Tag];
				if (b.Text == "")
					l[selectedID] = null;
				else
					l[selectedID] = b.Text;
				UpdateItem(selectedID, (string)b.Tag, listView.SelectedItems[0]);
			}
		}
		
		private void UpdateItem(string id, string language, ListViewItem item) {
			ListViewSubItem sub = item.SubItems[languages.IndexOf(language) + 1];
			bool direct;
			string val = GetString(id, language, out direct);
			sub.Text = val;
			if (direct) {
				sub.ForeColor = Color.Black;
				sub.Font = defaultFont;
			} else {
				sub.ForeColor = Color.Gray;
				sub.Font = italicFont;
			}
		}
		
		private void ListViewLabelEdit(object sender, LabelEditEventArgs e) {
			string old = listView.Items[e.Item].Text;
			if (e.Label == "") {
				MessageBox.Show("An entry must have a name!");
				e.CancelEdit = true;
			}
			if (e.Label == null) return;
			if (e.Label == old) return;
			if (mainHash[e.Label] != null) {
				MessageBox.Show("The entry '" + e.Label + "' does already exist!");
				e.CancelEdit = true;
			} else {
				foreach(Hashtable l in hash.Values) {
					string val = (string)l[old];
					l[old] = null;
					l[e.Label] = val;
				}
				if (selectedID == old)
					selectedID = e.Label;
			}
		}
		
		[STAThread]
		public static void Main(string[] args)
		{
			try {
				string[] files;
				if (args.Length == 0) {
					using (OpenFileDialog dlg = new OpenFileDialog()) {
						dlg.CheckFileExists = true;
						dlg.DefaultExt = ".txtres";
						dlg.Filter = "Resource files|*.txtres";
						dlg.Multiselect = true;
						dlg.Title = "Open language files";
						dlg.InitialDirectory = Path.GetFullPath("../../../data/resources");
						if (dlg.ShowDialog() != DialogResult.OK)
							return;
						files = dlg.FileNames;
					}
				} else {
					files = args;
				}
				Application.Run(new MainForm(files));
			} catch (Exception ex) {
				Grunwald.Gui.ErrorReporter.ShowErrorReport(ex);
			}
		}
		
		#region Windows Forms Designer generated code
		/// <summary>
		/// This method is required for Windows Forms designer support.
		/// Do not change the method contents inside the source code editor. The Forms designer might
		/// not be able to load this method if it was changed manually.
		/// </summary>
		private void InitializeComponent() {
			this.listView = new System.Windows.Forms.ListView();
			this.splitter = new System.Windows.Forms.Splitter();
			this.tabControl = new System.Windows.Forms.TabControl();
			this.SuspendLayout();
			// 
			// listView
			// 
			this.listView.Dock = System.Windows.Forms.DockStyle.Top;
			this.listView.GridLines = true;
			this.listView.HeaderStyle = System.Windows.Forms.ColumnHeaderStyle.Nonclickable;
			this.listView.HideSelection = false;
			this.listView.LabelEdit = true;
			this.listView.Location = new System.Drawing.Point(0, 0);
			this.listView.MultiSelect = false;
			this.listView.FullRowSelect = true;
			this.listView.Name = "listView";
			this.listView.Size = new System.Drawing.Size(464, 168);
			this.listView.Sorting = System.Windows.Forms.SortOrder.Ascending;
			this.listView.TabIndex = 0;
			this.listView.View = System.Windows.Forms.View.Details;
			// 
			// splitter
			// 
			this.splitter.Dock = System.Windows.Forms.DockStyle.Top;
			this.splitter.Location = new System.Drawing.Point(0, 168);
			this.splitter.Name = "splitter";
			this.splitter.Size = new System.Drawing.Size(464, 4);
			this.splitter.TabIndex = 1;
			this.splitter.TabStop = false;
			// 
			// tabControl
			// 
			this.tabControl.Dock = System.Windows.Forms.DockStyle.Fill;
			this.tabControl.Location = new System.Drawing.Point(0, 172);
			this.tabControl.Name = "tabControl";
			this.tabControl.SelectedIndex = 0;
			this.tabControl.Size = new System.Drawing.Size(464, 162);
			this.tabControl.TabIndex = 2;
			this.tabControl.SelectedIndexChanged += new System.EventHandler(this.TabControlSelectedIndexChanged);
			// 
			// MainForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(464, 334);
			this.Controls.Add(this.tabControl);
			this.Controls.Add(this.splitter);
			this.Controls.Add(this.listView);
			this.Name = "MainForm";
			this.Text = "MainForm";
			this.ResumeLayout(false);
		}
		#endregion
		
		void TabControlSelectedIndexChanged(object sender, EventArgs e)
		{
			tabControl.SelectedTab.Controls[0].Focus();
		}
	}
}
